unit DataFitter;
{ Copyright (C) 2005 C M Western; see NLFit.pas for license}
interface

uses Classes, DataSource, Expr, Vector, MatrixVec, MathSup;

type

  { TDataSourceFitter }

  TDataSourceFitter = class(TComponent)
  private
    FVariables: TVariableList;
    FDataSource: TDataSource;
    FExpr: TExpr;
    FLex: TLex;
    FExpression: string;
    FUnWtNewSD: Double;
    FNewSD: Double;
    FUnWtOldSD: Double;
    FOldSD: Double;
    FSVD: TMatrix;
    Fada: TMatrix;
    Fb: TColumnVector;
    FSVDw: TColumnVector;
    FScaleChanges: Double;
    FIncrement: Double;
    FSVDThresh: Double;
    FnPar, FnObs: Integer;
    FObservedVariable, FWeightVariable: TVariable;
    FUseStdDev: Boolean;
    FLinearFit: Boolean;
    FUseSVD: Boolean;
    procedure SetExpression(const Value: string);
    function GetGlobalFit: Boolean;
    procedure SetGlobalFit(const Value: Boolean);
    function GetVariable(AName: string): Double;
    function IsSVDThreshDefault: Boolean;
  protected
    procedure FreeLex;
    procedure SetFitExprs(Expr: TExpr; nPar: Integer);
    procedure SetVariable(AName: string; const Value: Double); virtual;
  public
    Floating: TBooleanArray;
    LastError: TDoubleArray;
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    procedure FitOnce;
    property Expr: TExpr read FExpr;
    property Lex: TLex read FLex;
    property Expression: string read FExpression write SetExpression;
    property DataSource: TDataSource read FDataSource { write SetDataSource };
    property ObservedVariable: TVariable read FObservedVariable;
    property WeightVariable: TVariable read FWeightVariable;


    property Variables: TVariableList read FVariables;
    property Variable[AName: string]: Double read GetVariable write SetVariable;
    { Results of the last fit cycle }
    { Previous and estimated next standard deviations }
    property OldSD: Double read FOldSD;
    property NewSD: Double read FNewSD;
    property UnWtOldSD: Double read FUnWtOldSD;
    property UnWtNewSD: Double read FUnWtNewSD;
    property nPar: Integer read FnPar;
    property nObs: Integer read FnObs;
    { Fit details }
    property b: TColumnVector read Fb;  { Standard Deviations of the parameters}
    property ada: TMatrix read Fada; { Correlation Matrix }
    property SVD: TMatrix read FSVD; { Singular Value decomposition details }
    property SVDw: TColumnVector read FSVDw;
  published
    { Fit options }
    property Increment: Double read FIncrement write FIncrement;
    property ScaleChanges: Double read FScaleChanges write FScaleChanges;
    property SVDThresh: Double read FSVDThresh write FSVDThresh stored IsSVDThreshDefault ;
    property UseStdDev: Boolean read FUseStdDev write FUseStdDev default True;
    property GlobalFit: Boolean read GetGlobalFit write SetGlobalFit;
    property LinearFit: Boolean read FLinearFit write FLinearFit default False;
    property UseSVD: Boolean read FUseSVD write FUseSVD default True;
  end;

  TFitterDataSource = class(TDataSource)
  private
    FDataSourceFitter: TDataSourceFitter;
    function GetDataSourceFitter: TDataSourceFitter;
  protected
    function MakeDataSourceFitter: TDataSourceFitter; virtual;
  public
    property DataSourceFitter: TDataSourceFitter read GetDataSourceFitter;
  end;

var
  GlobalFitter: TDataSourceFitter;

implementation

uses SysUtils, Math, NLFit, SmoothSpline, SplineExpr;

{ TDataSourceFitter }

constructor TDataSourceFitter.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  Increment := 1E-5;
  FScaleChanges := 1;
  FSVDThresh := 1e-15;
  FUseStdDev := True;
  FUseSVD := True;
  if AOwner is TDataSource then
    FDataSource := TDataSource(AOwner);
  FObservedVariable := TVariable.Create('observed');
  FWeightVariable := TVariable.Create('weight');
end;

destructor TDataSourceFitter.Destroy;
begin
  if GlobalFitter = Self then
    GlobalFitter := nil;
  FreeLex;
  FreeAndNil(Fb);
  FreeAndNil(Fada);
  FreeAndNil(FSVD);
  FreeAndNil(FSVDw);
  FreeAndNil(FObservedVariable);
  FreeAndNil(FWeightVariable);
  inherited;
end;

procedure TDataSourceFitter.FitOnce;
var
  iObs, iPar: Integer;
  i: Integer;
  WeightedFit, WeightedFitSet: Boolean;
  Calc, x, y, w, dp: Double;

  a: TMatrix;   { d(obs)/d(par) }
  obs: TColumnVector; { normally obs - calc }
  par: TColumnVector; { Parameters - changed by this routine }
  Weights: TColumnVector; { nil or The relative standard deviation of each value in obs }

  Floated: array of Integer;
  AllParameters: array of Double;

  procedure ForAllDataPoints(Mode: Integer); { Would be easier if we could use local procedure variables }
  var
    di: Integer;
    procedure ForOneDataSource(DataSource: TDataSource);
    var
      r: Longint;
      tc: Double;
    begin
      with DataSource do
      for r := 1 to nRows do
        if GetData(r, x, y, w, Mode>0) then
          case Mode of
            0: begin
                 Inc(FnObs);
                 if WeightedFitSet then begin
                   if WeightedFit <> Weighted then
                     raise EFitError.Create('Mix of weighted and non-weighted data');
                 end else begin
                   WeightedFitSet := True;
                   WeightedFit := Weighted;
                 end;
               end;
            1: begin
                if Weighted then
                  Weights[iObs] := w;
                Calc := Expr.Value;
                obs[iObs] := y - Calc;
                Inc(iObs);
               end;
            2: begin
                tc := Expr.Value;
                a[iObs,iPar] := (tc+obs[iObs]-y) / dp;
                Inc(iObs);
               end;
            3: begin
                if Weighted then
                  Weights[iObs] := w;
                obs[iObs] := x;
                par[iObs] := y;
                Inc(iObs);
               end;
            4: begin
                tc := Expr.Value;
                a[iObs,iPar] := tc;
                Inc(iObs);
               end;
          end
    end;
  begin
    if GlobalFit then
      for di := 0 to High(DataSources) do
        ForOneDataSource(DataSources[di])
    else
      ForOneDataSource(DataSource);
  end;

var
  tp: Double;
begin
  a := nil;
  obs := nil;
  par := nil;
  Weights := nil;
  try
    FnObs := 0;
    WeightedFitSet := False;
    ForAllDataPoints(0);
    if nObs <= 0 then
      raise EFitError.Create('No Observations');

    FnPar := 0;
    SetLength(Floated,Variables.Count);
    for i := 0 to High(Floating) do
      if Floating[i] then begin
        Floated[nPar] := i;
        Inc(FnPar);
      end;
    SetLength(Floated,nPar);
    if nPar > nObs then
      raise EFitError.CreateFmt('More parameters(%d) than observations(%d)', [nPar, nObs]);
    par := TColumnVector.Create(nPar);
    for iPar := 1 to nPar do
      par[iPar] := Variables.Objects[Floated[iPar-1]].Value;
    if WeightedFit then
      weights := TColumnVector.Create(nObs);
    obs := TColumnVector.Create(nObs);
    a := TMatrix.Create(nObs, nPar);
    if not Assigned(b) then
      Fb := TColumnVector.Create(nPar)
    else
      b.Size := nPar;
    if not Assigned(ada) then
      Fada := TMatrix.Create(nPar,nPar)
    else begin
      ada.ColCount := nPar;
      ada.RowCount := nPar;
    end;
    if not UseSVD then begin
      FreeAndNil(FSVD);
      FreeAndNil(FSVDw);
    end else begin
      if not Assigned(SVD) then
        FSVD := TMatrix.Create(nPar,nPar)
      else begin
        FSVD.ColCount := nPar;
        FSVD.RowCount := nPar;
      end;
      if not Assigned(SVDw) then
        FSVDw := TColumnVector.Create(nPar)
      else
        SVDw.Size := nPar;
    end;
    try
      iObs := 1;
      if LinearFit then begin
        { Function is stictly of the form Sum(Parameter*Function) so
          derivative can be found by setting all parameters to 0 and
          then the one required to 1; In this case start from all parameters
          0 as this seems to give better results in awkward cases,
          such as high order polynomials. }
        par.Zero;
        SetLength(AllParameters, Variables.Count);
        for iPar := 0 to High(AllParameters) do begin
          AllParameters[iPar] := Variables.Objects[iPar].Value;
          Variables.Objects[iPar].Value := 0;
        end;
      end;
      SetSplineScan(nObs);
      ForAllDataPoints(1);
      if SetSplineScan(-nObs) > 0 then begin
        iObs := 1;
        ForAllDataPoints(1);  { Do it again to calculate the proper residual }
      end;
      if LinearFit then begin
        for iPar := 1 to nPar do begin
          Variables.Objects[Floated[iPar-1]].Value := 1;
          iObs := 1;
          ForAllDataPoints(4);
          Variables.Objects[Floated[iPar-1]].Value := 0;
        end;
        for iPar := 0 to High(AllParameters) do
          Variables.Objects[iPar].Value := AllParameters[iPar];
      end else
        for iPar := 1 to nPar do begin
          tp := par[iPar];
          if UseStdDev and not IsNanOrInf(LastError[Floated[iPar-1]]) then
            dp := Abs(LastError[Floated[iPar-1]])
          else
            dp := 0;
          dp := Max(dp, Abs(tp*Increment));
          if dp = 0 then
            dp := Increment;
          Variables.Objects[Floated[iPar-1]].Value := tp+dp;
          iObs := 1;
          ForAllDataPoints(2);
          Variables.Objects[Floated[iPar-1]].Value := par[iPar];
        end;
    except
      raise EFitError.Create(Concat('Invalid starting parameters: ', Exception(ExceptObject).Message));
    end;
    LeastSQ(a, obs, par, weights, FOldSD, FNewSD, FUnWtOldSD, FUnWtNewSD, b, ada,  SVDw, SVD, SVDThresh, False, ScaleChanges);
    for iPar := 1 to nPar do
      Variables.Objects[Floated[iPar-1]].Value := par[iPar];
    SetFitExprs(Expr, nPar);
  finally
    a.Free;
    obs.Free;
    par.Free;
    Weights.Free;
  end;
end;

procedure TDataSourceFitter.FreeLex;
begin
  SetFitExprs(nil,-1);
  FreeAndNil(FVariables);
  FreeAndNil(FExpr);
  FreeAndNil(FLex);
end;

function TDataSourceFitter.GetGlobalFit: Boolean;
begin
  Result := Assigned(GlobalFitter);
end;

function TDataSourceFitter.GetVariable(AName: string): Double;
var
  Index: Integer;
begin
  Index := Variables.IndexOf(AName);
  if Index >= 0 then
    Result := Variables.Objects[Index].Value
  else
    SetNan(Result);
end;

function TDataSourceFitter.IsSVDThreshDefault: Boolean;
begin
  Result := Abs(SVDThresh-1e-15) < 1e-20;
end;

procedure TDataSourceFitter.SetExpression(const Value: string);
var
  i, n: Integer;
  OldVariables: TVariableList;
  NewLastError: TDoubleArray;
  NewFloating: TBooleanArray;
begin
  FExpression := Value;
  OldVariables := Variables;
  FVariables := nil;
  FreeLex;
  try
    FLex := TLex.Create;
    Lex.Options := Lex.Options - [xFundamentalConstants];
    FExpr := TExpr.Create(Lex);
    DataSource.AddIndependentVariables(Lex);
    Lex.AddVariable(FObservedVariable);
    Lex.AddVariable(FWeightVariable);
    Lex.AddFunction(ConstantPi);
    try
      FVariables := Expr.MakeVariables(Value,False);
      Lex.ReturnUnknown := False;
      Expr.Text := Value;
    except
      { Parse failed - revert to previous }
      FreeLex;
      raise;
    end;
    SetLength(NewFloating,Variables.Count);
    for i := 0 to High(Floating) do
      Floating[i] := True; { Default is all floating }
    Setlength(NewLastError, Variables.Count);
    for i := 0 to Variables.Count - 1 do
      SetNan(NewLastError[i]);
    { Preserve values where possible }
    if Assigned(OldVariables) then
      for i := 0 to OldVariables.Count - 1 do begin
        n := Variables.IndexOf(OldVariables[i]);
        if n >= 0 then begin
          Variables.Objects[n].Value := OldVariables.Objects[i].Value;
          NewLastError[n] := LastError[i];
          NewFloating[n] := Floating[i];
        end;
    end;
    LastError := NewLastError;
    Floating := NewFloating;
  finally
    OldVariables.Free;
  end;
end;

procedure TDataSourceFitter.SetFitExprs(Expr: TExpr; nPar: Integer);
var
  i: Integer;
begin
  if GlobalFit then
    for i := 0 to High(DataSources) do
      DataSources[i].SetFitExpr(Expr, nPar)
  else
    DataSource.SetFitExpr(Expr, nPar)
end;

procedure TDataSourceFitter.SetGlobalFit(const Value: Boolean);
var
  i: Integer;
begin
  if Value then begin
    if Assigned(GlobalFitter) then begin
      if GlobalFitter <> Self then
        Free; { Global Fitter already available }
    end else begin
      GlobalFitter := Self;
      { Synchronize all other Fitters }
      for i := 0 to High(DataSources) do
        (DataSources[i] as TFitterDataSource).DataSourceFitter.GlobalFit := True;
    end;
  end else
    GlobalFitter := nil;
end;

procedure TDataSourceFitter.SetVariable(AName: string; const Value: Double);
var
  Index: Integer;
begin
  Index := Variables.IndexOf(AName);
  if Index >= 0 then
    Variables.Objects[Index].Value := Value;
end;

{ TFitterDataSource }

function TFitterDataSource.GetDataSourceFitter: TDataSourceFitter;
begin
  if Assigned(GlobalFitter) then begin
    if Assigned(FDataSourceFitter) and (FDataSourceFitter <> GlobalFitter) then
      FreeAndNil(FDataSourceFitter);
    Result := GlobalFitter;
  end else if FDataSourceFitter = nil then begin
    FDataSourceFitter := MakeDataSourceFitter;
    Result := FDataSourceFitter;
  end else
    Result := FDataSourceFitter;
end;

function TFitterDataSource.MakeDataSourceFitter: TDataSourceFitter;
begin
  Result := TDataSourceFitter.Create(Self);
end;

end.
