unit JDXLoader;
{ Copyright (C) 2005 C M Western; see NLFit.pas for license}
interface

uses ExpPlotLoader, Classes;

type

  TJDXFileLoader = class(TExperimentalPlotLoader)
  private
    c: Char;
    EOF: Boolean;
    CharPending: Boolean;
    PendingChar: Char;
    StreamPosition: Int64;
    SavedChar: Char;
    NLChar: Char;
    PlotNumber: Integer;
    function IsSpace(Ch: Char): Boolean;
    function IsDigit(Ch: Char): Boolean;
    function NextC: Char;
    function PeekChar: Char;
    procedure SaveStreamPosition;
    procedure RestoreStreamPosition;
    procedure SkipLDR;
  private
    LastLongValue, DUPoutstanding: Integer;
    LastVal, FirstX, LastX, xFactor, yFactor, LastDiff: Double;
    LastWasDiff, HadNewline, HadPeaks, HadData: Boolean;
    function DXAddToTitle: string;
    function AFFN: string;
    function AFFNtail(Start: string): string;
    function AFFNSkipnl: string;
    function ASDF:Boolean;
    procedure DiscardDXComments;
    function DXdiscardable(Ac: Char):Boolean;
    procedure DXexpect(e: Char);
    function DXGetItem(Term: Char): string;
    procedure DXNextLine;
    function GetDATA_LABEL: string;
    function IsAFFNterm(Ch: Char):Boolean;
    procedure WantAFFN(Item: string);
    procedure LoadEx(AskAboutWeights: Boolean; PartFile:Boolean);
  public
    function Load(var Log: TextFile; AskAboutWeights: Boolean): Boolean; override;
  end;

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     Units, SysUtils, Math, MathSup;

{ TJDXFileLoader }

function TJDXFileLoader.DXAddToTitle: string;
var
  FirstNonBlank, NextMarkerSeen: Boolean;
begin
  Result := '';
  while not EOF and (c in [' ',#9]) do
    NextC;
  DiscardDXComments;
  NextMarkerSeen := False;
  repeat
    { For each line }
    FirstNonBlank := True;
    while not EOF and not (c in [#10, #13]) do begin
      if not IsSpace(c) then begin
        if FirstNonBlank and (c = '#') and (PeekChar = '#') then begin
          NextMarkerSeen := True;
          Break;
        end;
        FirstNonBlank := False;
      end;
      Result := Result + c;
      NextC;
      DiscardDXComments();
    end;
    if NextMarkerSeen then Break;
    if PeekChar in [#10, #13] then
      NextC;
    Result := Result + ' ';
    NextC;
    DiscardDXComments();
  until EOF or NextMarkerSeen;
  Result := Trim(Result);
  if Connector is TExperimentalPlotLoaderConnector then
    with TExperimentalPlotLoaderConnector(Connector) do begin
      if Plots[PlotNumber].Comment <> '' then
        Result := ' ' + Result;
      Plots[PlotNumber].Comment := Plots[PlotNumber].Comment + Result;
    end
  else begin
    if Connector.Comment <> '' then
      Result := sLineBreak + Result;
      Connector.Comment := Connector.Comment + Result;
  end;
end;

function TJDXFileLoader.AFFNSkipnl: string;
begin
  repeat
    DiscardDXComments;
    if EOF or not IsSpace(c) then
      Break;
    NextC;
  until False;
  Result := AFFNtail('');
end;

function TJDXFileLoader.AFFNtail(Start: string): string;
var
  Exponent: string;
  nDigits: Integer;
  LastType: TArrayType;
begin
  LastType := tyI16;
  Result := Start;
  if not EOF and (c in ['+','-']) then begin
    Result := Result + c;
    NextC;
  end;
  while not EOF and ( (c in ['0'..'9']) or ((LastType = tyI16) and (c = '.')) ) do begin
    if c = '.' then
      LastType := tyDouble;
    Result := Result + c;
    NextC;
  end;
  if Result = '' then
    Exit;
  { look for E+nn or E+nnn only (sign must be present; 2 or 3 digits only)
    allow 1 digit and lower case e as an extension }
  SaveStreamPosition;
  if not EOF and (c in ['E', 'e']) and (PeekChar in ['+','-']) then begin
    Exponent := 'E' + NextC;
    NextC;
    nDigits := 0;
    while IsDigit(c) and (nDigits <= 3) do begin
      Exponent := Exponent + c;
      NextC;
      Inc(nDigits);
    end;
    if (nDigits in [1..3]) and IsAFFNterm(c) then begin
      Result := Result+Exponent;
      LastType := tyDouble;
    end else
      RestoreStreamPosition;
  end else if LastType <> tyDouble then begin
    if TryStrToInt(Result, LastLongValue) then begin
      LastVal := LastLongValue;
      if (LastLongValue < Low(SmallInt)) or (LastLongValue > High(SmallInt)) then
        LastType := tyI32;
    end else
      LastType := tyDouble;
  end;
  if LastType = tyDouble then
    if not TryStrToFloatEx(Result, LastVal) then
      FileErr('Bad number:' + Result);
  if not EOF and (c = ',') then
    NextC; { absorb a trailing , }
  while not EOF and (c in [' ',#9]) do
    NextC;
  LastWasDiff := False;
end;

procedure TJDXFileLoader.DiscardDXComments;
begin
  if not EOF and (PeekChar = '$') then
    repeat
      NextC;
    until EOF or (c = #10);
end;

function TJDXFileLoader.DXdiscardable(Ac: Char): Boolean;
begin
  begin
    Result := Ac in [' ', #9, '-', '/', '_'];
  end;
end;

procedure TJDXFileLoader.DXexpect(e: Char);
begin
  while not EOF and DXdiscardable(c) do
    NextC;
  if EOF then
    raise Exception.Create('File too short');
  if c <> e then begin
    if c = #0 then c := '?';
    FileErr(Format('Expecting %s found %s', [e, c]));
  end;
  while not EOF and ( c <> e ) do
    NextC; { Look for one anyway }
  NextC; { and discard it }
end;

function TJDXFileLoader.DXGetItem(Term: Char): string;
begin
  Result := '';
  repeat
    while not EOF and DXdiscardable(c) do
      NextC;
    DiscardDXComments;
    if EOF or (c = Term) or (c = #10) then
      Exit;
    if not IsSpace(c) then
      Result := Result + c;
    NextC;
  until EOF;
end;

procedure TJDXFileLoader.DXNextLine;
begin
  while not EOF and not (c in [#10, #13]) do
    NextC;
end;

function TJDXFileLoader.GetDATA_LABEL: string;
begin
  repeat
    DiscardDXComments;
    while not EOF and IsSpace(c) do
      NextC;
    if EOF then
      raise Exception.Create('File too short');
  until c <> '$';
  DXexpect('#');
  DXexpect('#');
  Result := DXGetItem('=');
  if c <> '=' then
    FileErr('= not found after ' + Result)
  else
    NextC;
end;

function TJDXFileLoader.IsAFFNterm(Ch: Char): Boolean;
begin
  Result := (Ch in [#0,',']) or IsSpace(Ch);
end;

function TJDXFileLoader.IsDigit(Ch: Char): Boolean;
begin
  Result := Ch in ['0'..'9'];
end;

function TJDXFileLoader.IsSpace(Ch: Char): Boolean;
begin
  Result := Ch in [' ', #9, #10, #13];
end;

function TJDXFileLoader.Load(var Log: TextFile; AskAboutWeights: Boolean): Boolean;
var
  buf: string[6];
begin
  Result := False;
  NLChar := #0;
  EOF := False;
  CharPending := False;
  LineNumber := 1;
  { IE6/7 with Chime enabled seems to save files as name.dx,_.dxl, so
    test for this }
  if (Ext = '') or (Ext = '.dxl') then begin
    SaveStreamPosition;
    NextC;
    while not EOF and IsSpace(c) do
      NextC;
    if (c = '#') and (PeekChar = '#') then begin
      NextC;
      SetLength(buf,Stream.Read(buf[1], 6));
      if CompareText(buf,'TITLE=') = 0 then
        Ext := '.jdx';
    end;
    RestoreStreamPosition;
  end;
  if (Ext <> '.jdx') and (Ext <> '.dx') and (Ext <> '.jcm') then
    Exit;
  Result := True;
  PlotNumber := 0;
  LoadEx(AskAboutWeights, False);
end;

procedure TJDXFileLoader.LoadEx(AskAboutWeights: Boolean; PartFile:Boolean);
var
  Item, DataForm: string;
  x, y, yValue, dx: Double;
  nPoints, i, nCh: Integer;
  LastyWasDiff, PeakTable: Boolean;
  Section: string;
  procedure SkipPunct;
  begin
    if c in [',',';'] then
      NextC;
  end;
  procedure CheckPoints(Weighted:Boolean);
  begin
    if i = nPoints then Exit;
    FileErr('Wrong number of points' + Section);
    Connector.SetLength(PlotNumber,i,Weighted,True,True);
  end;
  procedure AddFrequency(V: Double);
  begin
    if i >= nPoints then begin
      FileErr('Too many points read' + Section);
      Exit;
    end;
    SetValue(PlotNumber,0,i,V*FrequencyFactor);
  end;
begin
  { Defaults }
  SetNan(FirstX);
  SetNan(LastX);
  xFactor := 1;
  yFactor := 1;
  HadNewline := False;
  HadPeaks := False;
  HadData := False;
  DUPoutstanding := 0;
  nPoints := -1;
  NextC;
  if PartFile then
    Item := 'TITLE'
  else
    Item := GetDATA_LABEL;
  if CompareText(Item, 'TITLE') <> 0 then
    FileErr('Expected TITLE found ' + Item)
  else
    DXAddToTitle;
  Item := GetDATA_LABEL;
  if CompareText(Item, 'DATATYPE') = 0 then begin
    Item := DXAddToTitle;
    if CompareText(Trim(Item), 'LINK') = 0 then begin
      repeat
        Item := GetDATA_LABEL;
        if CompareText(Item, 'TITLE') = 0 then begin
          if PlotNumber >= Connector.nPlots then
            Connector.nPlots := PlotNumber+1;
          LoadEx(AskAboutWeights, True);
          Inc(PlotNumber);
        end else
          SkipLDR;
      until EOF;
      Exit;
    end else
      raise Exception.Create('Unexpected DATA TYPE = '+Item);
  end else if CompareText(Item, 'JCAMPDX') <> 0 then
    FileErr('Expected JCAMPDX found ' + Item)
  else begin
    DXNextLine; // Don't bother checking for 4.24
    Item := GetDATA_LABEL;
  end;
  while CompareText(Item, 'END') <> 0 do begin
    if CompareText(Item, 'DATATYPE') = 0 then begin
      Item := DXAddToTitle;
      if CompareText(Item, 'LINK') = 0 then
        raise Exception.Create('LINK blocks not implemented');
    end else if CompareText(Item, 'BLOCK') = 0 then
      raise Exception.Create('LINK blocks not implemented')
    else if CompareText(Item, 'NTUPLES') = 0 then
      raise Exception.Create('NTUPLES not implemented')
    else if (CompareText(Item, 'ORIGIN') = 0)
            or (CompareText(Item, 'YUNITS') = 0)
            or (CompareText(Item, 'OWNER') = 0) then
      DXAddToTitle
    else if CompareText(Item, 'XUNITS') = 0 then begin
      Item := UpperCase(DXGetItem(#0));
      if (Item = '1/CM') or (Item = '1CM') then
        Connector.SetUnits(PlotNumber,cm1)
      else if Item = 'NANOMETERS' then begin
        Connector.SetUnits(PlotNumber,nmAir);
      end else if Item = 'MICROMETERS' then begin
        Connector.SetUnits(PlotNumber,nmVac);
        FrequencyFactor := 1e3;
      end else if Item = 'MHZ' then
        Connector.SetUnits(PlotNumber,MHz)
      else if Item = 'KHZ' then begin
        Connector.SetUnits(PlotNumber,MHz);
        FrequencyFactor := 1e-3;
      end else if Item = 'HZ' then begin
        Connector.SetUnits(PlotNumber,MHz);
        FrequencyFactor := 1e-6;
      end else if Item = 'GHZ' then begin
        Connector.SetUnits(PlotNumber,MHz);
        FrequencyFactor := 1e3;
      end;
    end else if CompareText(Item, 'NPOINTS') = 0 then begin
      WantAFFN(Item);
      nPoints := Round(LastVal);
    end else if CompareText(Item, 'FIRSTX') = 0 then begin
      WantAFFN(Item);
      FirstX := LastVal;
    end else if CompareText(Item, 'LASTX') = 0 then begin
      WantAFFN(Item);
      LastX := LastVal;
    end else if CompareText(Item, 'XFACTOR') = 0 then begin
      WantAFFN(Item);
      xFactor := LastVal;
    end else if CompareText(Item, 'YFACTOR') = 0 then begin
      WantAFFN(Item);
      yFactor := LastVal;
    end else if (CompareText(Item, 'XYDATA') = 0)
                or (CompareText(Item, 'XYPOINTS') = 0)
                or (CompareText(Item, 'PEAKTABLE') = 0) then begin
      PeakTable := CompareText(Item, 'PEAKTABLE') = 0;
      if PeakTable then
        Connector.SetLineList(PlotNumber,True);
      if HadData then
        FileErr('More than one data block (' + Item +')');
      { TODO: Accept more than one data block in a file }
      HadData := True;
      DataForm := DXGetItem(#0);
      if CompareText(DataForm, '(X++(Y..Y))') = 0 then begin
        Connector.SetLength(PlotNumber,nPoints,False,True,True);
        Section := ' for ' + Item + ' ' + DataForm;
        i := 0;
        y := 0;
        LastyWasDiff := False;
        if IsNan(FirstX) then
          FileErr('FIRSTX missing' + Section);
        x := FirstX;
        if IsNan(LastX) then begin
          FileErr('LASTX missing' + Section);
          LastX := 1000;
        end;
        if npoints <= 1 then begin
          FileErr('NPOINTS missing' + Section);
          dx := 1;
        end else
          dx := (LastX - FirstX) / (nPoints - 1);
        while ASDF do begin
          if HadNewline then begin
            LastVal := LastVal * xFactor;
            if LastyWasDiff and (i > 0) then
              x := x - dx;
            if not SameValue(x, LastVal)
               and (Abs(x - LastVal) >= Abs(dx)*0.8) {Max(Abs(dx),1/xFactor)*0.8)} then begin
                FileErr(Format('X-Value check fails at %g - error %g: %s',[x, x-LastVal, Section]));
                x := LastVal; { Take check value }
            end;
            { x := LastVal; Spec says use value from firstx }
            if not ASDF then begin
              FileErr('Missing y value' + Section);
              Continue;
            end else begin
              yValue := LastVal * yFactor;
              if LastyWasDiff and (i > 0) then begin
                { Y-VALUE CHECK }
                if LastWasDiff then
                  FileErr('Y value check not present' + Section);
                LastyWasDiff := LastWasDiff;
                if not SameValue(y, yValue) then begin
                  FileErr('Y value check fails' + Section);
                  yValue := y;
                end;
                x := x + dx;
                Continue;
              end;
            end
          end else
            yValue := LastVal * yFactor;
          LastyWasDiff := LastWasDiff;
          y := yValue;
          AddFrequency(x);
          if i < nPoints then
            Connector.SetValue(PlotNumber,1,i,y);
          Inc(i);
          x := x + dx;
        end;
        CheckPoints(False);
      end else if (CompareText(DataForm, '(XY..XY)') = 0)
                  or (CompareText(DataForm, '(XYZ..XYZ)') = 0) then begin
        nCh := 2;
        if CompareText(DataForm, '(XY..XY)') <> 0 then
          Inc(nCh);
        i := 0;
        Connector.SetLength(PlotNumber,nPoints,nCh>2,True,not PeakTable);
        while AFFNSkipnl <> '' do begin
          AddFrequency(LastVal);
          SkipPunct;
          WantAFFN(Item);
          if (i < nPoints) then
            Connector.SetValue(PlotNumber,1,i,LastVal);
          if nCh = 3 then begin
            SkipPunct;
            WantAFFN(Item);
            if (nCh > 2) and (i < nPoints) then
              Connector.SetValue(PlotNumber,2,i,LastVal);
          end;
          SkipPunct;
          Inc(i);
        end;
        CheckPoints(nCh>2);
      end else
        raise Exception.Create('Bad dataform' + Section);
    end else
      SkipLDR; { Unwanted LDR }
    Item := GetDATA_LABEL;
  end;
end;

function TJDXFileLoader.NextC: Char;
begin
  if CharPending then begin
    CharPending := False;
    Result := PendingChar;
  end else
    EOF := Stream.Read(Result, 1) < 1;
  if NLChar = #0 then begin
    if c in [#10,#13] then
      NLChar := c;
  end;
  if (c = NLChar) and (NLChar <> #0) then
    Inc(LineNumber);
  c := Result;
end;

function TJDXFileLoader.PeekChar: Char;
begin
  if not CharPending then
    CharPending := Stream.Read(PendingChar, 1) = 1;
  if CharPending then
    Result := PendingChar
  else
    Result := #0;
end;

procedure TJDXFileLoader.RestoreStreamPosition;
begin
  Stream.Position := StreamPosition;
  CharPending := False;
  c := SavedChar;
end;

procedure TJDXFileLoader.SaveStreamPosition;
begin
  StreamPosition := Stream.Position;
  if CharPending then
    Dec(StreamPosition);
  SavedChar := c;
end;

procedure TJDXFileLoader.SkipLDR;
begin
  repeat
    DXNextLine;
    while not EOF and IsSpace(c) do NextC;
    while not EOF and DXdiscardable(c) do NextC;
  until EOF or (c = '#');
end;

procedure TJDXFileLoader.WantAFFN(Item: string);
begin
  if AFFNSkipnl = '' then
    FileErr('Missing ' + Item);
end;

function TJDXFileLoader.ASDF: Boolean;
var
  r: Double;
  procedure Diff(s: string);
  begin
    r := LastVal;
    NextC;
    AFFNtail(s);
    LastDiff := LastVal;
    LastWasDiff := True;
    LastVal := r + LastVal;
  end;
  function Dup(Ch: Char):Boolean;
  begin
    DUPoutstanding := Ord(Ch) - Ord('S') + 1;
    repeat
      NextC;
      if not EOF and IsDigit(c) then
        DUPoutstanding := DUPoutstanding*10 + Ord(c) - Ord('0')
      else
        Break;
    until False;
    Dec(DUPoutstanding);
    Result := ASDF();
  end;
  procedure Tail(s: string);
  begin
    NextC;
    AFFNtail(s);
  end;
begin
  HadNewline := False;
  Result := True;
  if DUPoutstanding <> 0 then begin
    Dec(DUPoutstanding);
    if LastWasDiff then
      LastVal := LastVal + LastDiff;
    Exit;
  end;
  while not EOF and IsSpace(c) do begin
    HadNewline := HadNewline or (c = #10);
    NextC;
  end;
  case c of
    '+','-','0'..'9', '.': AFFN;
    '@':      Tail('0');
    'A'..'I': Tail(Char(Ord(c) - Ord('A') + Ord('1')));
    'a'..'i': Tail('-' + Char(Ord(c) - Ord('a') + Ord('1')));
    'j'..'r': Diff('-' + Char(Ord(c) - Ord('j') + Ord('1')));
    '%':      Diff('0');
    'J'..'R': Diff(Char(Ord(c) - Ord('J') + Ord('1')));
    's':      Result := Dup(Char(Ord('S')+8));
    'S'..'Z': Result := Dup(c);
    '?':      begin
                LastWasDiff := False;
                LastVal := 0; { TODO: Could consider deleting point.  NaN would not be helpful }
                NextC;
              end;
    '$':      begin
                if NextC <> '$' then
                  FileErr('Invalid char after $:' + c)
                else
                  DXNextLine;
                Result := ASDF();
              end;
    else
      Result := False;
  end;
end;

function TJDXFileLoader.AFFN: string;
begin
  while not EOF and (c in [' ',#9]) do
    NextC;
  Result := AFFNtail('');
end;

initialization
  RegisterPlotLoader(TJDXFileLoader, 'JCAMP-DX Files', '*.dx;*.jdx;*.jcm');
end.
