unit NLFit;
{
Lutils - utility least squares fitting routines for PGOPHER and other programs 
Copyright (C) 2005 C M Western
School of Chemistry, University of Bristol, Bristol, UK
For further information please see http://pgopher.chm.bris.ac.uk/
or contact help-pgopher@bristol.ac.uk

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
}

interface

uses Vector, MatrixVec, MathSup;

{ Makes least squares adjustment of parameters in par to minimise
  sum (obs(i)/weights(i))**2 given d(obs)/d(par) in a;
  If weights or correlations of the observations are given, then the
  observations are transformed to be decorrelated with standard deviation 1,
  and the derivative matrix is updated to reflect this.
}
function LeastSQ(
                  a: TMatrix;   { d(obs)/d(par) - updated if parameters correlated or weighted }
                  obs: TColumnVector; { normally obs - calc }
                  par: TColumnVector; { Parameters - changed by this routine }
                  weights: TColumnVector; { nil or The relative standard deviation of each value in obs
                                            - updated if parameters correlated or weighted }
                  { Output: }
                  out OldSD, NewSD: Double; { Previous and estimated next standard deviations }
                  out UnWtOldSD, UnWtNewSD: Double; { Previous and estimated next standard deviations ignoring weights }
                  b: TColumnVector;  { Standard Deviations of the parameters}
                  ada: TMatrix; { Correlation Matrix }
                  SVDw: TColumnVector; SVD: TMatrix; { Details of Singular Value Decomposition; if nil use a QR factorisation method, rather than SVD }
                  SVDThresh: Double; { Values smaller than this are discarded }
                  Verbose: Boolean;
                  ScaleChanges: Double; { Scale changes by this - normally 1 }
                  Correlations: TDoubleArrayArray = nil { Correlation of observation with previous observation }
                 ):Integer; { Result is the number combinations of parameters discarded }

implementation

{.$DEFINE VISDEBUG}

uses {$IFDEF VISDEBUG} ShowMat, {$ENDIF}
     Worker, EigLA, Math, SVD, LapackLSQ;

function LeastSQ(a: TMatrix; obs: TColumnVector;  par: TColumnVector;  weights: TColumnVector;
                  out OldSD, NewSD, UnWtOldSD, UnWtNewSD: Double;
                  b: TColumnVector; ada: TMatrix; SVDw: TColumnVector; SVD: TMatrix;
                  SVDThresh: Double; Verbose: Boolean; ScaleChanges: Double; Correlations: TDoubleArrayArray
                 ):Integer;
var
  U: TMatrix;
  obsb: TColumnVector;
  procedure Decorrelate;
  var
    S, E, I, J, K: Integer;
    M: TMatrix;
    V: TColumnVector;
    Sum, W: Double;
  begin
    S := 1;
    M := TMatrix.Create;
    V := TColumnVector.Create;
    { Transform observations into uncorrelated ones with standard deviation 1 }
    while S <= obs.Size do begin
      { Find next correlated observations }
      { Logic assumes correlated observations in groups }
      if (S < obs.Size) and (Length(Correlations) > 0) and (Length(Correlations[S-1]) > 0) then begin
        I := S+1;
        E := S+Length(Correlations[S-1]);
        { Find last correlated observation }
        while (I < obs.Size) and ( (I < E) or (Length(Correlations[I-1]) > 0) ) do begin
          E := Max(E, I+Length(Correlations[I-1]));
          Inc(I);
        end;
        V.Size := E-S+1;
        M.Resize(V.Size,V.Size);
        M.Zero;
        for I := 1 to V.Size do begin
          M[I,I] := Sqr(weights[I+S-1]);
          if I < V.Size then
            for J := 1 to Length(Correlations[I+S-2]) do begin
              M[J+I,I] := Correlations[I+S-2][J-1]*weights[I+S-1]*weights[J+I+S-1];
              M[I,J+I] := M[J+I,I];
            end;
        end;
        if Verbose then begin
          WriteLn(Log, 'Transforming ', S, '..', E);
          M.Write(Log,'Correlation Matrix');
        end;
        Eigen(V, M, True);
        if Abs(V[1]) < 1e-14*Abs(V[V.Size]) then begin
          { Near zero eigenvalue; should/could give divide by zero error elsewhere }
          WriteLn(Log, '**** Warning: probable redundant observation');
          if not Verbose then
            V.Write(Log,'Weights');
        end;
        if Verbose then begin
          V.Write(Log,'Weights');
          M.Write(Log,'Transform');
        end;
        { Now transform parameters }
        for K := 1 to par.Size do begin
          for I := S to E do begin
            Sum := 0;
            for J := S to E do
              Sum := Sum + a[J,K]*M[J-S+1,I-S+1];
            U[I,K] := Sum/Sqrt(V[I-S+1]);
          end;
        end;
        for I := S to E do begin
          Sum := 0;
          for J := S to E do
            Sum := Sum + obs[J]*M[J-S+1,I-S+1];
          weights[I] := Sqrt(V[I-S+1]); { Revised effective weights }
          obsb[I] := Sum/weights[I];
        end;
        S := E+1;
      end else begin
        W := Abs(weights[S]);
        obsb[S] := obs[S] / W;
        for I := 1 to par.Size do
          U[S,I] := a[S,I] / W;
        Inc(S);
      end;
    end;
    if Verbose then begin
      obsb.Write(Log, 'Transformed observables');
      U.Write(Log, 'Transformed derivatives');
    end;
    a.Assign(U); { Save design matrix for transformed observations }
    V.Free;
    M.Free;
  end;
var
  w: Double;
  acum, acum1, s: Extended;
  i, j, k: Integer;
  {$IFDEF VISDEBUG}
  Uex: TMatrix;
  bex: TColumnVector;
  adaex, tex: TMatrix;
  {$ENDIF}
begin
  U := nil;
  obsb := nil;
  Result := 0;
  try
    U := TMatrix.Create(obs.Size, par.Size);
    obsb := TColumnVector.Create(obs.Size);
    if weights <> nil then
      Decorrelate
    else begin
      U.Assign(a);
      obsb.Assign(obs);
    end;
    { Intial Error }
    acum := 0;
    acum1 := 0;
    for i := 1 to obs.Size do begin
      acum := acum + Sqr(obsb[i]);
      acum1 := acum1 + Sqr(obs[i]);
    end;
    if obs.Size > par.Size then begin
      OldSD := Sqrt(acum/(obs.Size - par.Size));
      UnWtOldSD := Sqrt(acum1/(obs.Size - par.Size));
    end else begin
      OldSD := Sqrt(acum);
      UnWtOldSD := Sqrt(acum1);
    end;
    {$IFDEF VISDEBUG}
      bex := TColumnVector.Create(obsb);
      ShowMatrix(U, 'A', bex);
      UEx := TMatrix.Create(U);
      //dgelsd(UEx,bex);
      //ShowMatrix(UEx,'dgelsd',bex);
      //Uex.Assign(U);
      //bex.Assign(obsb);
      dgels(UEx, bex);
      ShowMatrix(UEx,'dgels',bex);
      adaex := TMatrix.Create(par.Size, par.Size);
      adaex.Zero;
      adaex.AddDiagonal(1);
      dtrtrs('U', 'N', 'N', Uex, adaex);
      ShowMatrix(adaex, 'adaex', nil);
      tex := adaex.Transpose;
      adaex.PostMultiply(tex);
      tex.Free;
      ShowMatrix(adaex, 'adaex', nil);
      for i := 1 to par.Size do
        for j := 1 to i-1 do
          adaex[i,j] := adaex[i,j] / Sqrt(adaex[i,i]*adaex[j,j]);
      ShowMatrix(adaex, 'correx', nil);
      adaex.Free;
    {$ENDIF}
    if Assigned(SVDw) then begin
      SV_DecomposeAlt(U, SVDw, U, SVD);
      if Verbose then begin
        WriteLn(Log);
        WriteLn(Log,'Singular Value Decomposition');
        U.Write(Log,'U');
        SVDw.Write(Log,'Sigma');
        SVD.Write(Log,'V');
      end;
      {$IFDEF VISDEBUG}
        ShowMatrix(U, 'U', nil);
        ShowMatrix(SVD, 'V', SVDw);
      {$ENDIF}
      s := 0;
      if SVDThresh > 0 then begin
        for i := 1 to par.Size do
          if s < SVDw[i] then s := SVDw[i];
        s := s * SVDThresh;
      end;
      for i := 1 to par.Size do
        if SVDw[i] <= s then begin
          SVDw[i] := 0;
          Inc(Result);
        end;
      SV_Solve(U, SVDw, SVD, True, obsb, b);
    end else begin
      dgels(U, obsb);
      for i := 1 to par.Size do
        b[i] := obsb[i];
      ada.Zero;
      ada.AddDiagonal(1);
      dtrtrs('U', 'N', 'N', U, ada);
      U.Free;
      U := ada.Transpose;
      ada.PostMultiply(U);
    end;
    {$IFDEF VISDEBUG}
      for i := 1 to par.Size do begin
        WriteLn(Log, b[i]:20, bex[i]:20, (b[i]-bex[i]):20, (b[i]-bex[i])/b[i]:20);
        b[i] := bex[i];
      end;
      Uex.Free;
      bex.Free;
    {$ENDIF}
    if Verbose then
      b.Write(Log, 'b');
    { Calculate estimated residuals
      calc =  a * b
      and standard deviation of fit }
    acum := 0;
    acum1 := 0;
    for i := 1 to obs.Size do begin
      s := 0;
      for j := 1 to par.Size do
        s := s + a[i,j]*b[j];
      if weights <> nil then begin
        // Was: acum := acum + Sqr((obs[i]-s)/weights[i]);
        acum := acum + Sqr(obsb[i]-s);
        { Unweighted sum of squares is calculated ignoring correlation in observations }
        acum1 := acum1 + Sqr((obsb[i]-s)*weights[i]);
      end else
        acum1 := acum1 + Sqr(obs[i]-s);
    end;
    if weights = nil then
      acum := acum1;
    if obs.Size > par.Size then begin
      NewSD := Sqrt(acum/(obs.Size - par.Size));
      UnWtNewSD := Sqrt(acum1/(obs.Size - par.Size));
    end;
    for i := 1 to par.Size do
      par[i] := par[i] + ScaleChanges*b[i];
    if Assigned(SVD) then begin
      for i := 1 to par.Size do
        if SVDw[i] > 0 then
          obsb[i] := 1 / Sqr(SVDw[i])
        else obsb.Valid[i] := False;
      { Variance - covariance matrix }
      for i := 1 to par.Size do
        for j := 1 to i do begin
            s := 0;
            for k := 1 to par.Size do
              if obsb.Valid[k] then
                s := s + SVD[k,i] * SVD[k,j] * obsb[k];
            ada[i,j] := s;
            ada[j,i] := s;
          end;
    end;
    {$IFDEF VISDEBUG}
      ShowMatrix(ada,'ADA',nil);
    {$ENDIF}
    for i := 1 to par.Size do
      if ada[i,i] > 0 then
        b[i] := Sqrt(ada[i,i])
      else
        b.Valid[i] := False;
    for i := 1 to par.Size do
      for j := 1 to i do begin
        if b.Valid[i] and b.Valid[j] then
          ada[i,j] := ada[i,j] / (b[i] * b[j])
        else
          ada.Valid[i,j] := False;
      end;
    for i := 1 to par.Size do
      if b.Valid[i] then
        b[i] := NewSD * b[i];
  finally
    U.Free;
    obsb.Free;
  end;
end;

end.

