unit PSLoader;
{ Copyright (C) 2007 C M Western; see NLFit.pas for license}
interface

{ Import OceanOptics .ProcSPec files }

uses
  Classes, ExpPlotLoader;

type

  { TProcSpecFileLoader }

  TProcSpecFileLoader = class(TExperimentalPlotLoader)
  private
    D: TStream;
    function ReadDirectory: Boolean;
  public
    function Load(var Log: TextFile; AskAboutWeights: Boolean): Boolean; override;
  end;

implementation

uses {$IFDEF FPC} zstream, zipper, {$ELSE} MZLib, {$ENDIF}
     SysUtils, Math, pgXMLRead, pgDOM, MathSup, Units;

{$IFNDEF FPC}

{ Partial re-implementation to allow header skip }

type

  TMyDecompressionStream = class(TStream)
  private
    FStream: TStream;
    State: TZState;
    FFullSise: Integer;
    ExpectedCRC, ActualCRC: LongWord;
  protected
    function GetSize:Int64; override;
  public
    constructor Create(Source: TStream; CompSize, FullSize: Integer; AExpectedCRC: LongWord);
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
  end;

  { TDecompressionStream }

constructor TMyDecompressionStream.Create(Source: TStream; CompSize, FullSize: Integer; AExpectedCRC: LongWord);
var
  Buffer: string;
begin
  inherited Create;
  FStream := Source;
  if Source is TMemoryStream then
    State.NextInput := PByte(PChar(TMemoryStream(Source).Memory) + Source.Position)
  else begin
    SetLength(Buffer, CompSize);
    Read(Buffer[1], CompSize);
    State.NextInput := PByte(Buffer);
  end;
  State.AvailableInput := CompSize;
  InflateInit2(State, -MAX_WBITS); { Flag to skip header }
  FFullSise := FullSize;
  ExpectedCRC := AExpectedCRC;
  ActualCRC := 0;
end;

destructor TMyDecompressionStream.Destroy;
begin
  inflateEnd(State);
  inherited;
end;

function TMyDecompressionStream.GetSize: Int64;
begin
  Result := FFullSise;
end;

function TMyDecompressionStream.Read(var Buffer; Count: Integer): Longint;
var
  Initial: Cardinal;
  Status: Integer;
begin
  State.NextOutput := @Buffer;
  State.AvailableOutput := Count;
  Initial := State.TotalOutput;
  Status := inflate(State,0);
  Result := State.TotalOutput - Initial;
  if Status >= 0 then
    ActualCRC := CRC32(ActualCRC,@Buffer,Result);
  case Status of
    Z_OK: { No action };
    Z_STREAM_END: { Check CRC }
          if ActualCRC <> ExpectedCRC then
            raise Exception.Create('CRC error');
  else
    raise Exception.Create('Decompression Error');
  end;
end;

{ Copied directly from zipper in FPC FCL, as not available in Delphi }

Const
  { Signatures }
{$ifdef FPC_BIG_ENDIAN}
  END_OF_CENTRAL_DIR_SIGNATURE  = $504B0506;
  LOCAL_FILE_HEADER_SIGNATURE   = $504B0304;
  CENTRAL_FILE_HEADER_SIGNATURE = $504B0102;
{$else FPC_BIG_ENDIAN}
  END_OF_CENTRAL_DIR_SIGNATURE  = $06054B50;
  LOCAL_FILE_HEADER_SIGNATURE   = $04034B50;
  CENTRAL_FILE_HEADER_SIGNATURE = $02014B50;
{$endif FPC_BIG_ENDIAN}

Type
   Local_File_Header_Type = Packed Record
     Signature              :  LongInt;
     Extract_Version_Reqd   :  Word;
     Bit_Flag               :  Word;
     Compress_Method        :  Word;
     Last_Mod_Time          :  Word;
     Last_Mod_Date          :  Word;
     Crc32                  :  LongWord;
     Compressed_Size        :  LongInt;
     Uncompressed_Size      :  LongInt;
     Filename_Length        :  Word;
     Extra_Field_Length     :  Word;
   end;

  { Define the Central Directory record types }

  Central_File_Header_Type = Packed Record
    Signature            :  LongInt;
    MadeBy_Version       :  Word;
    Extract_Version_Reqd :  Word;
    Bit_Flag             :  Word;
    Compress_Method      :  Word;
    Last_Mod_Time        :  Word;
    Last_Mod_Date        :  Word;
    Crc32                :  LongWord;
    Compressed_Size      :  LongInt;
    Uncompressed_Size    :  LongInt;
    Filename_Length      :  Word;
    Extra_Field_Length   :  Word;
    File_Comment_Length  :  Word;
    Starting_Disk_Num    :  Word;
    Internal_Attributes  :  Word;
    External_Attributes  :  LongInt;
    Local_Header_Offset  :  LongInt;
  End;

  End_of_Central_Dir_Type =  Packed Record
    Signature               :  LongInt;
    Disk_Number             :  Word;
    Central_Dir_Start_Disk  :  Word;
    Entries_This_Disk       :  Word;
    Total_Entries           :  Word;
    Central_Dir_Size        :  LongInt;
    Start_Disk_Offset       :  LongInt;
    ZipFile_Comment_Length  :  Word;
  end;

{$ELSE}

type
  { TMyDecompressionStream }

  TMyDecompressionStream = class(TDecompressionStream)
  private
    ExpectedCRC: LongWord;
  protected
    UncompressedSize: Int64;
    function GetSize:Int64; override;
  public
    constructor Create(ASource: TStream; CompSize, FullSize: Integer; AExpectedCRC: LongWord);
  end;

function TMyDecompressionStream.GetSize: Int64;
begin
  Result := UncompressedSize;
end;

constructor TMyDecompressionStream.Create(ASource: TStream; CompSize, FullSize: Integer; AExpectedCRC: LongWord);
begin
  inherited Create(ASource, True);
  UncompressedSize := FullSize;
  ExpectedCRC := AExpectedCRC;
end;

{$ENDIF}

{ TProcSpecFileLoader }

type
  PEnd_of_Central_Dir_Type = ^End_of_Central_Dir_Type;

function TProcSpecFileLoader.ReadDirectory: Boolean;
Var
   i,
   nEntries,
   CenDirPos : LongInt;
   FileName: string;
   CentralHdr  : Central_File_Header_Type;
   LocalHdr    : Local_File_Header_Type;
   Buffer      : array [0..4095] of Char;
begin
   Result := False;
   Stream.Seek(Max(Stream.Size-SizeOf(Buffer),0), soFromBeginning);
   i := Stream.Read(Buffer[0], SizeOf(Buffer));
   Dec(i, SizeOf(End_of_Central_Dir_Type));
   while i >= 0 do begin
     if PEnd_of_Central_Dir_Type(@Buffer[i])^.Signature = END_OF_CENTRAL_DIR_SIGNATURE then
       Break;
     Dec(i);
   end;
   if i < 0 then
     Exit;
   with PEnd_of_Central_Dir_Type(@Buffer[i])^ do begin
     CenDirPos := Start_Disk_Offset;
     nEntries := Entries_This_Disk;
   end;
   Stream.Seek(CenDirPos,soFrombeginning);
   for i:=0 to nEntries-1 do
     begin
       if Stream.Read(CentralHdr, SizeOf(CentralHdr)) <> SizeOf(CentralHdr) then
         Exit;
       With CentralHdr do
         begin
           if Signature<>CENTRAL_FILE_HEADER_SIGNATURE then
             Exit;
           SetLength(FileName,Filename_Length);
           if Stream.Read(FileName[1],Filename_Length) <> Filename_Length then
             Exit;
           Stream.Seek(Extra_Field_Length+File_Comment_Length,soCurrent);
           { Look for file with actual data; check for both name or size }
           if (CompareText(Copy(FileName,1,2), 'ps') = 0) or (Uncompressed_Size > 3000) then begin
             Stream.Seek(Local_Header_Offset,soFromBeginning);
             if Stream.Read(LocalHdr,SizeOf(LocalHdr)) <> SizeOf(LocalHdr) then
               Exit;
             if LocalHdr.Signature<>LOCAL_FILE_HEADER_SIGNATURE then
               Exit;
             Stream.Seek(LocalHdr.Filename_Length+Extra_Field_Length,soCurrent);
             if Compress_Method = 0 then begin
               D := TMemoryStream.Create;
               D.CopyFrom(Stream,Compressed_Size);
             end else { Assuming Method is 8 }
               D := TMyDecompressionStream.Create(Stream, Compressed_Size, Uncompressed_Size, Crc32);
             Result := True;
             Exit;
           end;
         end;
     end;
end;

function TProcSpecFileLoader.Load(var Log: TextFile; AskAboutWeights: Boolean): Boolean;
var
  Position: Int64;
  XMLDocument: TXMLDocument;
  function ReadValues(Node: TDOMNode; Col: Integer):Integer;
  var
    Child, e: TDOMNode;
    s: string;
  begin
    Result := 0;
    Child := Node.FirstChild;
    while Assigned(Child) do begin
      if Col >= 0 then begin
        e := (Child as TDOMElement).FirstChild;
        s := e.NodeValue;
        SetValue(-1,Col,Result,StrToFloat(s));
      end;
      Inc(Result);
      Child := Child.NextSibling;
    end;
  end;
  function DeepFindNode(Node: TDOMNode_WithChildren; const AName: string):TDomNode;
  var
    Child: TDomNode;
  begin
    Result := Node.FindNode(AName);
    Child := Node.FirstChild;
    while not Assigned(Result) and Assigned(Child) do begin
      if Child is TDOMNode_WithChildren then
        Result := DeepFindNode(TDOMNode_WithChildren(Child),AName);
      if Assigned(Result) and (ReadValues(Result,-1) < 4) then
        Result := nil; { Ignore nodes that are just references to other nodes }
      Child := Child.NextSibling;
    end;
  end;
var
  XNode, YNode: TDOMNode_WithChildren;
begin
  Result := False;
  if (Ext = '') or (Ext = '.procspec') then begin
    Position := Stream.Position;
    if ReadDirectory then
      try
        ReadXMLFile(XMLDocument, D);
        XNode := DeepFindNode(XMLDocument, 'channelWavelengths') as TDOMNode_WithChildren;
        Connector.SetUnits(-1,nmAir);
        YNode := DeepFindNode(XMLDocument, 'processedPixels') as TDOMNode_WithChildren;
        if not Assigned(YNode) then
          YNode := DeepFindNode(XMLDocument, 'pixelValues') as TDOMNode_WithChildren;
        if not (Assigned(XNode) and Assigned(YNode)) then
          Exit;
        SetupChannels(1,ReadValues(XNode,-1),False);
        ReadValues(XNode, 0);
        ReadValues(YNode, 1);
      finally
        D.Free;
      end
    else begin
      { Unrecognized }
      Stream.Position := Position;
      Exit;
    end;
  end else
    Exit; { Wrong extension }
  Result := True;
end;

initialization
  RegisterPlotLoader(TProcSpecFileLoader, 'ProcSpec Files', '*.ProcSpec');
end.

