unit SPLoader;
{ Copyright (C) 2007 C M Western; see NLFit.pas for license}
{ Import Perkin Elmer spectrometer files }
interface

uses ExpPlotLoader, Classes;

type

  TSPFileLoader = class(TExperimentalPlotLoader)
  public
    function Load(var Log: TextFile; AskAboutWeights: Boolean): Boolean; override;
  end;

implementation

uses SysUtils, Units, TextStrm, Math, MathSup;

function ReadInteger(p: PChar):Integer;
begin
  Result :=
  {$IFDEF FPC_BIG_ENDIAN}
    SwapEndian(PInteger(p)^);
  {$ELSE}
    PInteger(p)^;
  {$ENDIF}
end;

function ReadWord(p: PChar):Word;
begin
  Result :=
  {$IFDEF FPC_BIG_ENDIAN}
    SwapEndian(PWord(p)^);
  {$ELSE}
    PWord(p)^;
  {$ENDIF}
end;

type
  Tm = record
    case Integer of
      1: (r8: Double);
      {$IFDEF FPC_BIG_ENDIAN}
      2: (hi, lo: Longint);
      {$ELSE}
      2: (lo, hi: Longint);
      {$ENDIF}
      3: (i8: Int64);
    end;

function ReadDouble(p: PChar):Double;
{$IFDEF FPC_BIG_ENDIAN}
var
  v: Tm;
{$ENDIF}
begin
  {$IFDEF FPC_BIG_ENDIAN}
    v.i8 := SwapEndian(PInt64(p)^);
    Result := v.r8;
  {$ELSE}
    Result := PDouble(p)^;
  {$ENDIF}
end;

function UnpackReal(p: PChar):Double;
var
  v: Tm;
  Exponent: Integer;
begin
  v.i8 :=
  {$IFDEF FPC_BIG_ENDIAN}
    SwapEndian(PInt64(p)^);
  {$ELSE}
    PInt64(p)^;
  {$ENDIF}
  if v.i8 = 0 then
    Result := 0
  else begin
    { TODO: fix for big endian, 64 bit }
    Exponent := (v.hi shr 20) + 22; { 11 bit exponent and sign, offset 22 from IEEE standard }
    v.hi := Integer(v.hi and $800fffff) or ((Exponent and $7ff) shl 20);
    Result := v.r8;
  end;
end;

const
  Verbose: Boolean = False;

type
  { Record fields that have been read in the file }
  TSPState = (spHeader, spX, spY, spDeltaX, spNPoints, spUnits, spIUnits, spData, spDone);

{ TSPFileLoader }

function TSPFileLoader.Load(var Log: TextFile; AskAboutWeights: Boolean): Boolean;
var
  Position: Int64;
  Buffer: string[48];
  BinaryHeader, BinaryData:Boolean;
  S: TTextStream;
  Line: string;
  XScale, YScale, YOffset, FirstX, LastX, DeltaX: Double;
  i, nPoints, DataFormat: Integer;
  pi: ^Integer;
  p, e: PChar;
  Data: string;
  Field: string;
  SPState: TspState;
  procedure AddToTitle(s: string);
  begin
    if Connector.Comment <> '' then
      s := ' ' + s;
    Connector.Comment := Connector.Comment + s;
  end;
  function ReadReal: Double;
  begin
    if not S.NextLine then begin
      Line := S.RestOfLine;
      if not TryStrToFloatEx(Line, Result) then
        Connector.ReportError('Bad number:'+Line, True);
    end else
      Connector.ReportError('Unexpected end of file', True);
  end;
  procedure ReadDataBlock;
  var
    i: Integer;
  begin
    if ReadInteger(p) <> nPoints*8 then
      Connector.ReportError(Format('Inconsistent file length: %d %d',[ReadInteger(p), nPoints*8]), False);
    nPoints := ReadInteger(p) div 8;
    Inc(p,4);
    CheckAvailableChannels(2, nPoints, AskAboutWeights);
    for i := 0 to nPoints - 1 do begin
      if Verbose then
        WriteLn(Log, '0x',IntToHex(Integer(p-PChar(Line)),8), ' - ', i, ':', UnpackReal(p));
      SetValue(-1,0,i,FirstX * XScale);
      SetValue(-1,1,i,UnpackReal(p)*YScale + YOffset);
      Inc(p,8);
      FirstX := FirstX + DeltaX;
    end;
  end;
  function ReadField(Indent: string): Boolean;
  var
    Len: Integer;
    FStart: PChar;
    FieldType: Char;
    procedure ShowField;
    var
      q: PChar;
    begin
      if not Verbose then
        Exit;
      q := FStart;
      while q < p do begin
        Write(Log, IntToHex(Ord(q^),2));
        if q^ in [' '..'z'] then
          Write(Log, '(', q^, ')')
        else
          Write(Log, '   ');
        Inc(q);
      end;
    end;
  begin
    Result := True;
    FStart := p;
    FieldType := p^;
    case FieldType of
      '#': begin
            { String field }
            Inc(p,2);
            Len := ReadWord(p);
            SetLength(Field, Len);
            Move((p+2)^,PChar(Field)^,Len);
            if SPState = spUnits then begin
              if CompareText(Line,'cm-1') = 0 then
                Connector.SetUnits(-1,cm1);
              // May need this
              // else if CompareText(Line,'~M') then begin
              //  Connector.SetUnits(nmVac);
              //  XScale := 1e3;
              // end;
              SPState := spIUnits;
            end else if SPState = spIUnits then begin
              if CompareText(Line,'%T') = 0 then
                // Connector.SetIntensityUnits(Normalized);
                ;
              SPState := spData;
            end;
            Inc(p,2+Len);
           end;
      ',', '$', #$15, #$2F: begin
            { Word field }
            Inc(p,2);
            Field := 'Word (0x' + IntToHex(Ord(FieldType), 2) + '): ' + IntToStr(ReadWord(p));
            Inc(p,2);
          end;
     #$1A, #$2B: begin
            { LongWord field }
            Inc(p,2);
            if FieldType = #$2B then begin
              if not (SPState in [spNPoints, spHeader, spDone]) then
                Connector.ReportError('Unexpected nPoints field', False)
              else if SPState = spNPoints then begin
                if ReadInteger(p) <> nPoints then
                  Connector.ReportError(Format('Inconsistent nPoints: %d %d',[nPoints,ReadInteger(p)]), False);
                SPState := spUnits;
              end;
            end;
            Field := 'Word (0x' + IntToHex(Ord(FieldType), 2) + '): ' + IntToStr(ReadInteger(p));
            Inc(p,4);
          end;
      #$12, #$13:
          begin
            { number of following 3/6 field items? }
            Inc(p,2);
            Field := 'Number of items (0x' + IntToHex(Ord(FieldType), 2) + ') : ' + IntToStr(ReadInteger(p));
            Inc(p,4);
          end;
      #$16:
          begin
            { Data Block field }
            Inc(p,2);
            Field := 'Data Block';
            if SPState <> spData then
              Connector.ReportError('Early data block', False);
            ReadDataBlock;
            SPState := spDone;
          end;
      #$1B, #$1C:
          begin
            Inc(p,2);
            if FieldType = #$1B then
              if SPState = spDeltaX then begin
                DeltaX := ReadDouble(p);
                SPState := spNPoints;
                nPoints := Round((LastX-FirstX)/DeltaX + 1);
              end else
                Connector.ReportError('Unexpected DeltaX field', False);
            Field := 'Float (0x' + IntToHex(Ord(FieldType), 2) + '): ' + FloatToStr(UnpackReal(p)) + ' ' + FloatToStr(ReadDouble(p));
            Inc(p,8);
          end;
      #$1D:
          begin
            Inc(p,2);
            case SPState of
              spX: begin FirstX := ReadDouble(p); LastX := ReadDouble(p+8); SPState := spY; end;
              spY: SPState := spDeltaX ;
            else
              Connector.ReportError('Unexpected range field', False);
            end;
            Field := '0x1d u field ' // + FloatToStr(UnpackReal(p)) + ' ' + FloatToStr(UnpackReal(p+8))
                     + ' ' + FloatToStr(ReadDouble(p)) + ' ' + FloatToStr(ReadDouble(p+8));
            Inc(p,16);
          end;
    else
      Field := 'Unknown Field ' + IntToHex(Ord(FieldType),2);
      Result := False;
    end;
    if Verbose then
      WriteLn(Log, Indent,' ',Field);
  end;
  function ReadBlock(Indent: string): Boolean;
  var
    Len: Integer;
    BlockType, BlockSig: Integer;
    BlockEnd: PChar;
  begin
    Result := False;
    if Verbose then
      Write(Log, Indent, '0x',IntToHex(Integer(p-PChar(Line)),8),': ');
    if p = e then
      Exit;
    BlockSig := Ord((p+1)^);
    BlockType := Ord(p^);
    Inc(p,2);
    Len := ReadInteger(p);
    Inc(p,4);
    if Verbose then
      WriteLn(Log, 'Block start (',IntToHex(BlockType,2),') ', IntToHex(BlockSig,2), ' Length ', Len);
    if not (BlockSig in [$8B, $8C, $FF, 0]) then begin
      if Verbose then WriteLn(Log, 'Bad block start');
      Connector.ReportError('Bad block start', False);
      Exit;
    end;
    Result := True;
    BlockEnd := p + Len;
    if BlockSig in [$8B, 0] then
      case BlockType of
        $7f: Inc(p,6); { Initial block }
        $71..$7e, $fc..$ff, $81..$83:
      else
        if Verbose then
          WriteLn(Log, 'Unknown Block Type:' + IntToHex(BlockType,2));
        Connector.ReportError('Unknown Block Type:' + IntToHex(BlockType,2), False);
        p := BlockEnd;
        Exit;
      end;
//    while (p < e) and not ((p+1)^ in [#$8B,#$8C, 'u']) do begin
//      if Verbose then
//        WriteLn(Log, 'Skipping ', IntToHex(Ord(p^),2));
//      Inc(p);
//    end;
    while Result and (p < BlockEnd) do
      if (p+1)^ = 'u' then begin
        if not ReadField(Indent) then begin
          p := BlockEnd; { Unknown field - skip rest of block }
          Exit;
        end
      end else
        if not ReadBlock(Indent+' ') then
          p := BlockEnd;
  end;
begin
  Result := False;
  BinaryHeader := False;
  BinaryData := True;
  if (Ext = '') or (Ext = '.sp') or (Ext = '.asc') then begin
    Position := Stream.Position;
    if (Stream.Read(Buffer[1],48) = 48)
       and (buffer[1] = 'P') and (buffer[2] = 'E') then begin
      SetLength(buffer,48);
      if Pos('BINARY', Buffer) > 0 then
        BinaryData := True
      else if Pos('ASCII', buffer) > 0 then
        BinaryData := False
      else
        BinaryHeader := True;
    end else begin
      { Unrecognized }
      Stream.Position := Position;
      Exit;
    end;
  end else
    Exit; { Wrong extension }
  Result := True;
  Stream.Position := Position;
  XScale := 1;
  DeltaX := 1;
  YScale := IntPower(2, -22);
  YOffset := 0;
  if not BinaryHeader then begin
    S := TTextStream.Create(Stream);
    try
      while not S.NextLine and (S.RestOfLine <> '#GR') do
       { Search for spectrum parameters };
      S.NextLine;
      Line := S.RestOfLine; { X units }
      if Line = 'CM-1' then
        Connector.SetUnits(-1,cm1)
      else if Line = '~M' then begin
        Connector.SetUnits(-1,nmVac);
        XScale := 1e3;
      end;
      S.NextLine;
      Line := S.RestOfLine; { Y units }
      // Connector.SetIntensityUnits(Normalized);
      AddToTitle(Line);
      YScale := ReadReal; { Y scaling }
      YOffset := ReadReal; { Y offset }
      FirstX := ReadReal; { First X }
      DeltaX := ReadReal; { Delta X }
      nPoints := Round(ReadReal);{ number of points }
      DataFormat := Round(ReadReal); { data format; 8 = 4 byte integer }
      while not S.NextLine and (S.RestOfLine <> '#DATA') do
       { Search for data block };
      S.NextLine;
      if BinaryData then begin
        Data := S.RestOfFile;
        if Length(Data) <> nPoints*4 then
          Connector.ReportError('Wrong file length', False);
        if DataFormat <> 8 then
          Connector.ReportError('Unexpected data format', False);
        nPoints := Length(Data) div 4;
        CheckAvailableChannels(2, nPoints, AskAboutWeights);
        pi := Pointer(PChar(Data));
        for i := 0 to nPoints - 1 do begin
          SetValue(-1,0,i,FirstX * XScale);
          SetValue(-1,1,i,ReadInteger(Pointer(pi))*YScale + YOffset);
          Inc(pi);
          FirstX := FirstX + DeltaX;
        end;
      end else begin
        CheckAvailableChannels(2, nPoints, AskAboutWeights);
        for i := 0 to nPoints - 1 do begin
          SetValue(-1,0,i,S.NextDouble(0) * XScale);
          SetValue(-1,1,i,S.NextDouble(0)*YScale + YOffset);
          S.NextLine;
        end;
      end;
    finally
      S.Free;
    end;
  end else begin
    Line := '';
    repeat { Fill buffer with rest of file }
      i := Length(Line);
      SetLength(Line, i+100000);
      SetLength(Line, Stream.Read((PChar(Line)+i)^,100000)+i);
    until Length(Line) <> i+100000;
    p := PChar(Line);
    e := p + Length(Line);
    Inc(p,$32);
    SPState := spHeader;
    if not ReadBlock('') then
      Connector.ReportError('Only one data block', True);
    SPState := spX;
    while ReadBlock('') and (Verbose or (SPState <> spDone)) do
     ;
    if Verbose then
      WriteLn(Log);
  end;
end;

initialization
  RegisterPlotLoader(TSPFileLoader, 'Perkin Elmer Files', '*.sp;*.asc');
end.
