unit SplineExpr;
{ Copyright (C) 2007 C M Western; see NLFit.pas for license}

{ Spline function for use with TExpr; note that this requires some
  faking, as the required list of values can't be given directly to
  the function. For this reason it will only work inside a DataFitter
  object.
  Usage:
    spline(x,observed,weight,<control>)
  where
    x is the  independent variable
    observed  is the dependent variable
    weight    is an estimate of the relative standard deviation
              of the observed value
    <control> determines the spline form produced:
      control = 0: for a spline exactly through every point
      control > 0: then a smoothed spline is produced with a small a
                   second derivative as possible, but with an average
                   error of approximately control (in units of observed/weight)
      control < 0: Construct a spline through |control| equally spaced points,
                   determining the spline cefficients by least squares fitting
                   to the observed
}

interface

uses Expr;

var
  SplineFunction: TVaradicFunction;

{ Call SetSplineScan(nObs) at the start of a scan of
  all the observed data only, and call SetSplineScan(-nObs)
  when done; this forces re-evaluation of the spline function;
  the return value indicates the number of points in the spline function }
function SetSplineScan(nObs: Integer):Integer;

implementation

uses MathSup, SplineControl, PiecewisePoly, SysUtils;

var
  IsScanning: Boolean = False;
  x, y, w: TDoubleArray;
  cp: Integer;
  PiecewisePolynomial: TPiecewisePolynomial = nil;
  ExpectedError, ActualError: Double;

function SetSplineScan(nObs: Integer):Integer;
begin
  FreeAndNil(PiecewisePolynomial);
  if nObs > 0 then begin
    { Allocate space - note that element [0] is unused }
    SetLength(x, nObs+1);
    SetLength(y, nObs+1);
    SetLength(w, nObs+1);
    cp := 0;
    ExpectedError := 0; { Default }
  end else if cp > 0 then
    PiecewisePolynomial := MakeSmoothSpline(x, y, w, ActualError, ExpectedError);
  IsScanning := nObs > 0;
  Result := cp;
end;

function _Spline(var a: array of Double): Double;
var
  nTerms: Integer;
  xv, yv, wv: Double;
begin
  nTerms := High(a);
  xv := a[nTerms];
  Result := 0; { Default }
  if IsScanning then begin
    { Accumlate data for constructing spline }
    if nTerms >= 2 then begin
      wv := a[nTerms-2];
      if nTerms >= 3 then
        ExpectedError := a[nTerms-3];
    end else
      wv := 1;
    if wv > 0 then begin
      yv := a[nTerms-1];
      Inc(cp);
      x[cp] := xv;
      y[cp] := yv;
      w[cp] := wv;
    end;
  end else if Assigned(PiecewisePolynomial) then           
    { Calculate spline }
    Result := PiecewisePolynomial.Value(xv);
end;

initialization

  SplineFunction := TVaradicFunction.Create({$IFDEF FPC}@{$ENDIF}_Spline, 2, 'spline');
  GlobalFunctions.AddObject('spline', SplineFunction);

finalization
  FreeAndNil(PiecewisePolynomial);
end.
