unit Constrain;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Classes, Mixture, Obs, Expr, MNode, TypInfo, pgDom, Overlay;

type

  TMNodeVariable = class (TFunction)
  private
    MNode: TMNode;
    PropInfo: PPropInfo;
  public
    constructor Create(Text: string; Reader: TObservationsReader; Mixture: TMixture; Overlays: TOverlays);
    function Value(var a:array of Double):Double; override;
    procedure SetValue(AValue: Double);
  end;

  TConstraint = class (TComponent)
  private
    FExpr: TExpr;
    FLex: TLex;
    FText: string;
    LHS: TMNodeVariable;
  public
    destructor Destroy; override;
    procedure Apply;
    procedure SetText(Value: string; Reader: TObservationsReader; Mixture: TMixture; Overlays: TOverlays);
    procedure ClearExpr;
    procedure Write(var F: Text);
    property Text: string read FText;
  end;

  TVariables = class (TMNodeWithParameters)
    { Used to hold variables used in constraints }
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
  end;

procedure ApplyConstraints;
procedure ClearConstraints;
function AddConstraint(Text: string; Reader: TObservationsReader; Mixture: TMixture; Overlays: TOverlays): TConstraint;
procedure WriteConstraints(var F: Text);

implementation

uses SysUtils, Contnrs, PropHelp;

{ Constraints on parameters; Global Variable for now,
  and only applies to Mixture it was set up for; could
  make it a variable in TMixture }

var
  Constraints: TComponentList;

procedure ClearConstraints;
begin
  FreeAndNil(Constraints);
end;

procedure ApplyConstraints;
var
  I: Integer;
begin
  if Assigned(Constraints) then
    for I := 0 to Constraints.Count - 1 do
      (Constraints[i] as TConstraint).Apply;
end;

function AddConstraint(Text: string; Reader: TObservationsReader; Mixture: TMixture; Overlays: TOverlays): TConstraint;
begin
  Result := TConstraint.Create(nil);
  try
    Result.SetText(Text,Reader,Mixture,Overlays);
    Result.Apply; { Test it out }
    if not Assigned(Constraints) then
      Constraints := TComponentList.Create;
    Constraints.Add(Result);
  except
    Result.Free;
    raise;
  end;
end;

procedure WriteConstraints(var F: Text);
var
  I: Integer;
begin
  if not Assigned(Constraints) or (Constraints.Count = 0) then
    Exit;
  WriteLn(F,'Constraints:');
  for I := 0 to Constraints.Count - 1 do
    { TODO: (Constraints[I] as TConstraint).Write(F) }
    WriteLn(F, (Constraints[I] as TConstraint).Text);
end;

{ TConstraint }

procedure TConstraint.Apply;
var
  Value: Double;
begin
  Value := FExpr.Value;
  LHS.SetValue(Value);
end;

procedure TConstraint.ClearExpr;
var
  i: Integer;
begin
  if Assigned(Flex) then
    with FLex.Functions do
      for i := 0 to Count-1 do
        if Objects[i] is TMNodeVariable then
          Objects[i].Free;
  FreeAndNil(FExpr);
  FreeAndNil(FLex);
end;

destructor TConstraint.Destroy;
begin
  ClearExpr;
  inherited;
end;

procedure TConstraint.SetText(Value: string; Reader: TObservationsReader; Mixture: TMixture; Overlays: TOverlays);
var
  Variables: TVariableList;
  i, eq: Integer;
  LHSname: string;
begin
  FText := Value;
  ClearExpr;
  eq := Pos(':=', Value);
  { = on it's own is no good as v=0 is a default name! }
  if eq <= 1 then
    raise EReportException.Create('Constraint must contain :=');
  LHSname := Copy(Value,1,eq-1);
  LHSname := Trim(LHSname);
  LHS := TMNodeVariable.Create(LHSname, Reader, Mixture, Overlays);
  Delete(Value,1,eq+2);
  try
    FLex := TLex.Create;
    FExpr := TExpr.Create(FLex);
    FLex.IdentifierChars := FLex.IdentifierChars + ['.','=','>','<','|'];
    FLex.AddFunction(ConstantPi);
    FLex.AddFunction(LHS);
    Variables := FExpr.MakeVariables(Value,False);
    try
      for i := 0 to Variables.Count - 1 do begin
        FLex.AddFunction(TMNodeVariable.Create(Variables[i],Reader,Mixture,Overlays));
        FLex.RemoveVariable(Variables.Objects[i]);
      end;
    finally
      Variables.Free;
    end;
    FLex.ReturnUnknown := False;
    FExpr.Text := Value;
    with FLex.Functions do
      for i := 0 to Count - 1 do
        if Objects[i] is TMNodeVariable then
          with TMNodeVariable(Objects[i]) do
            Name := MNode.FullHeirarchicName;
  except
    ClearExpr;
    raise;
  end;
end;

procedure TConstraint.Write(var F: Text);
var
  ExprText: string;
  ExprNode: TExprNode;
begin
  ExprText := FText;
  ExprNode := nil;
  try
    try
      ExprNode := FExpr.ExprTree;
      ExprText := LHS.Name + ' := ' + ExprNode.Text(True);
    finally
      ExprNode.Free;
    end;
  except
    { Fall back to input text }
  end;
  Writeln(F, ExprText);
end;

{ TMNodeVariable }

constructor TMNodeVariable.Create(Text: string; Reader: TObservationsReader; Mixture: TMixture; Overlays: TOverlays);
var
  Variables: TVariables;
  Ignored: string;
begin
  inherited Create(Text, 0);
  PropInfo := nil;
  if Assigned(Overlays) then begin
    Variables := TVariables(Overlays.ChildWithClass(TVariables));
    if Assigned(Variables) then
      PropInfo := Variables.SubPropInfo(Text, MNode, Ignored);
  end;
  if not Assigned(PropInfo) then begin
    Variables := TVariables(Mixture.ChildWithClass(TVariables));
    if Assigned(Variables) then
      PropInfo := Variables.SubPropInfo(Text, MNode, Ignored);
  end;
  if not Assigned(PropInfo) then
    PropInfo := Mixture.SubPropInfo(Text, MNode, Ignored);
  if not Assigned(PropInfo) and (Mixture.nSpecies = 1) then
    PropInfo := Mixture.Species[0].SubPropInfo(Text, MNode, Ignored);
  if not Assigned(PropInfo) and Assigned(Reader.DefaultMolecule) then
    PropInfo := Reader.DefaultMolecule.SubPropInfo(Text, MNode, Ignored);
  if not Assigned(PropInfo) and Assigned(Reader.DefaultManifoldU) then
    PropInfo := Reader.DefaultManifoldU.SubPropInfo(Text, MNode, Ignored);
  if not Assigned(PropInfo) and Assigned(Reader.DefaultManifoldL) then
    PropInfo := Reader.DefaultManifoldL.SubPropInfo(Text, MNode, Ignored);
  if not Assigned(PropInfo) and (Reader.DefaultStateU <> nil) then
    PropInfo := Reader.DefaultStateU.SubPropInfo(Text, MNode, Ignored);
  if not Assigned(PropInfo) and (Reader.DefaultStateL <> nil) then
    PropInfo := Reader.DefaultStateL.SubPropInfo(Text, MNode, Ignored);
  if not Assigned(PropInfo) and Assigned(Overlays) then
    PropInfo := Overlays.SubPropInfo(Text, MNode, Ignored);
  if not Assigned(PropInfo) then
    raise Exception.Create('Can''t find '+Text);
end;

procedure TMNodeVariable.SetValue(AValue: Double);
begin
  if Cardinal(PropInfo) <= MNode.nParameters then
    MNode.Parameters[Cardinal(PropInfo)] := AValue
  else if (PropInfo^.Index > 0) and (PropInfo^.PropType^.Kind <> tkFloat) then
    MNode.Settings[PropInfo^.Index] := Round(AValue)
  else
    MNode.Parameters[PropInfo^.Index] := AValue;
end;

function TMNodeVariable.Value(var a: array of Double): Double;
begin
  if Cardinal(PropInfo) <= MNode.nParameters then
    Result := MNode.Parameters[Cardinal(PropInfo)]
  else if (PropInfo^.Index > 0) and (PropInfo^.PropType^.Kind <> tkFloat) then
    Result := MNode.Settings[PropInfo^.Index]
  else
    Result := MNode.Parameters[PropInfo^.Index];
end;

{ TVariables }

function TVariables.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  Child, NextChild: TDOMElement;
  Attr: TDOMAttr;
begin
  Result := inherited ReadProperties(Node, False);
  Child := Node.FirstChild as TDOMElement;
  { Read (and create) dynamic parameters containing variables }
  while Assigned(Child) do begin
    NextChild := Child.NextSibling as TDOMElement;
    if CompareText(Child.NodeName, 'Parameter') = 0 then begin
      Attr := (Child as TDOMElement).GetAttributeNode('Name');
      Result := Result + ReadFloatableDOMNode(AddParameter(Attr.NodeValue),'0',Child);
      Node.RemoveChild(Child);
    end;
    Child := NextChild;
  end;
  if CheckUnread then
    Result := Result + nUnreadProperties(Node)
end;

initialization
  TVariables.RegisterClass(TOverlays);
finalization
  ClearConstraints;
end.

