unit ExpPlot;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses {$IFNDEF CONSOLE} Gendlg, Graphs, Forms, PickDest, IPlotLoader,
                       {$IFDEF LCL_} LCLType,
     {$ENDIF} {$ENDIF}
     MathSup, Graphics, Classes, SysUtils, pgDOM, MNode, Mixture, GraphRange,
     Units, Overlay, DataSource, ExpPlotLoader, SpectrumPlot, GraphM, OverlaySumFitter;

type

  { TExperimentalPlot }

  TExperimentalPlot = class(TStretchablePlot)
  private
    FIMin, FIMax: Double;
    FDataSource: TDataSource;
    FResidualPlotter: TObject;
    FLastIntensity: TDoubleArray;

    { Cache for convoluted spectrum if required }
    CurrentSimulation: TSimulation;
    ConvolutedSpectrum: TSpectrumPlot;

    procedure SetResidualPlotter(const Value: TObject);
    function GetDataSource: TDataSource;
    procedure DoMNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer);
    function GetComponentType(const Index: Integer): TComponentType;
    procedure SetComponentType(const Index: Integer; const Value: TComponentType);
  protected
    function GetIMin: Double; override;
    function GetIMax: Double; override;
    function LocalFrequency(Index: Integer):Double;
    procedure SetModified(const Value: Boolean); override;
  public
    Changing: Boolean;
    Intensity, Frequency, Weights, OriginalFrequency: TDoubleArray;
    function StartPoint: Integer;
    function EndPoint: Integer;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function ConvertPeakPosition(Position: Double; PlotUnits: TResonanceUnits):Double; override;
    function Empty: Boolean; override;
    function Etalon(Spacing: Double):Integer;
    procedure LoadFromFile(FileName: string; DefaultUnits: TResonanceUnits; AlwaysPrompt: Boolean; DefaultIntensityUnits: TIntensityUnits); override;
    procedure WeightedLoadFromStream(Stream: TStream; DefaultUnits: TResonanceUnits; FileName:string = ''; AskAboutWeights: Boolean = True; AlwaysPrompt: Boolean = False; DefaultIntensityUnits: TIntensityUnits = Arbitrary);
    function IntensityAt(Freq: Double; out Point: Integer):Double; { Interpolate Frequency }
    function GetPoint(i: Integer; PlotUnits: TResonanceUnits; PlotIntensityUnits: TIntensityUnits; var x, y:Double):Boolean;
    procedure Sort(Ascending: Boolean = True);
    function PrepareConvolution(Simulation: TSimulation; PlotIntensityUnits: TIntensityUnits):Boolean;
    procedure SetParameter(AIndex: Integer; const Value: Double); override;
    procedure AdjustParameterNames;
    procedure AddNodeParameters(var FittingParameters: TParameters); override;

    procedure SaveForUndo; override;
    procedure Undo; override;
    function CanUndo: Boolean; override;

    { Operations }
    { TODO: Should be able to undo these }
    procedure Crop(First, Last: Integer);
    procedure ApplyCalibration(APlot:TExperimentalPlot; SamePoints: Boolean);
    procedure DiscardFrequency;
    procedure DiscardIntensity;
    procedure DiscardCalibration;
    procedure Invert; override;
    procedure ConvoluteData(Values: array of Double);
    procedure Smooth(Width, Derivative, Poly:Integer);
    procedure Compress(nAverage: Integer);
    procedure ReBin(xInterval: Double);
    procedure JoinTo(JoinPoint: Double; Plot: TExperimentalPlot; ScalePlot:Boolean; IncreasingFrequency: Boolean);
    procedure Split(SplitPoint: Double);
    procedure Separate;
    procedure SwapXandY;
    function AutoZero(Bins:Integer = -1; First: Integer = -1; Last: Integer = -1):Double;

    // function nDF: Integer; override;
    function FindPeak(AFmin, AFmax: Double; out Position: Double; PlotUnits: TResonanceUnits): Boolean; override;
    function LimitsInRange(Simulation: TSimulation; var AIMin, AImax: Double; PlotIntensityUnits: TIntensityUnits):Boolean; override;
    function FixupProperties:Exception; override;
    procedure ClearSelection; override;
    procedure SelectFrequencyRange(AFmin, AFmax: Double); override;
    {$IFNDEF CONSOLE}
    procedure OffsetBy(Offset: Double); override;
    {$ENDIF}
    function Show(G: TGraphM; Simulation: TSimulation; PlotIntensityUnits: TIntensityUnits; PlotRange: TPlotRange; var PlotRanges: TPlotRanges; OwnAxis: Boolean; ShowName: Boolean; ForceAxis: Boolean = False):Boolean; override;
    function Resolution: Double; override;
    procedure ReadData(Child: TDOMCDATASection);
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    procedure ScaleParameters(UFrom, UTo: TResonanceUnits); override;
    function SimpleSaveToStream(Stream: TStream; WithLinkedMNodes: Boolean = False): Boolean; override;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    function WriteBasicData(Parent: TDOMNode): TDOMCDATASection;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function HasDataSource: Boolean;

    property DataSource: TDataSource read GetDataSource write FDataSource;
    property ResidualPlotter: TObject read FResidualPlotter write SetResidualPlotter;

  published
    property Baseline:       Double index 3 read GetParameter write SetParameter;
    property Scale:          Double index 4 read GetParameter write SetParameter;
    { property c1.... are done dynamically }

    property FirstPoint: Integer    index  7 read GetSetting   write SetSetting;
    property LastPoint:  Integer    index  8 read GetSetting   write SetSetting;
    property IsLineList: Boolean    index  9 read GetFlag      write SetFlag;
    property Convolute:  Boolean    index 10 read GetFlag      write SetFlag;
    property ComponentType: TComponentType index 11 read GetComponentType write SetComponentType;
  end;

  { TPGOPHERPlotLoaderConnector }

  TPGOPHERPlotLoaderConnector = class({$IFNDEF CONSOLE}TInteractivePlotLoaderConnector{$ELSE}TExperimentalPlotLoaderConnector{$ENDIF})
  private
    FExperimentalPlot: TExperimentalPlot;
    FBaseName: string;
    function GetBaseName: string;
  protected
    function GetComment: string; override;
    procedure SetComment(const AValue: string); override;
  public
    property ExperimentalPlot: TExperimentalPlot read FExperimentalPlot;
    constructor Create(AnObject: TObject; AStreamName: string); override;
    destructor Destroy; override;
    procedure ReportError(Msg: string; Fatal: Boolean); override;
    procedure SetLineList(Plot: Integer; IsLineList:Boolean); override;
    procedure SetName(const AValue: string); override;
    procedure SetIntensityUnits (IntensityUnits: TIntensityUnits); override;
    procedure SetUnits(Plot: Integer; AUnits: TResonanceUnits); override;
    procedure SetScale(Scale: Double); override;
    procedure SetColour(Plot: Integer; Colour: Integer); override;
    function Another(Index: Integer): TPlotLoaderConnector; override;
    property BaseName: string read GetBaseName;
  end;

implementation

uses {$IFNDEF CONSOLE} ClipBrd, Huge,
       {$IFDEF MSWINDOWS} Windows, {$ENDIF}
       CalFrm, Main,
     {$ENDIF}
     MNodeColour, Math, TextStrm, ShellSort, SavitzkyGolay, TabbedLine,
     DatLoader, Expr, PlotDataSource, Worker, Progress, Global, Peaks;

{ TExperimentalPlot }

procedure TExperimentalPlot.WeightedLoadFromStream(Stream: TStream; DefaultUnits: TResonanceUnits; FileName: string; AskAboutWeights: Boolean; AlwaysPrompt: Boolean; DefaultIntensityUnits: TIntensityUnits);
{ If Ext is blank, try autodetecting file format }
begin
  ClearPeaks;
  Name := ExtractFileName(FileName);
  if Name = '' then begin
    Name := IntToStr(Owner.ComponentCount);
    FDefaultName := True;
  end;
  if TryPlotLoader(Log, TPGOPHERPlotLoaderConnector, Self, Stream, FileName, AskAboutWeights, AlwaysPrompt, DefaultUnits, DefaultIntensityUnits) then begin
    Sort(Frequency[High(Frequency)]>Frequency[0]); { Ensure monotonic scale }
    if Comment = '' then
      Comment := FileName;
    if (IntensityUnits <> Arbitrary) and Assigned(Owner) then
      (Owner as TOverlays).IntensityUnits := Normalized;
    Exit;
  end else
    raise EReportException.Create('Failed to load ' + FileName);
end;

function TExperimentalPlot.LimitsInRange(Simulation: TSimulation;
  var AIMin, AIMax: Double;
  PlotIntensityUnits: TIntensityUnits): Boolean;
var
  i, nPoints: Integer;
  x, y, Mi, Mx, AFmin, AFmax: Double;
begin
  nPoints := 0;
  AFmin := Simulation.Fmin-Simulation.Foffset;
  AFmax := Simulation.Fmax-Simulation.Foffset;
  InOrder(AFmin, AFmax);
  if PrepareConvolution(Simulation, PlotIntensityUnits) then
    with ConvolutedSpectrum do begin
      Result := nLinesInRange > 0;
      Mi := IMin;
      Mx := IMax;
    end
  else begin
    { Determine limits from points in range }
    for i := StartPoint to EndPoint do begin
      if GetPoint(i,Simulation.PlotUnits,PlotIntensityUnits,x,y) and ( x >= AFmin) and (x <= AFmax) and not IsNanOrInf(y) then begin
        if nPoints = 0 then begin
          Mi := y;
          Mx := Mi;
        end else begin
          if y < Mi then Mi := y;
          if y > Mx then Mx := y;
        end;
        Inc(nPoints);
      end;
    end;
    Result := nPoints > 0;
  end;
  if not Result then Exit;
  if IsLineList or (Mx = Mi) then
    Mi := Min(Mi,0);
  if Mx = Mi then
    Mx := Mi + 1; { flat }
  AIMin := Mi;
  AIMax := Mx;
end;

procedure TExperimentalPlot.LoadFromFile(FileName: string; DefaultUnits: TResonanceUnits; AlwaysPrompt: Boolean; DefaultIntensityUnits: TIntensityUnits);
var
  Stream: TStream;
begin
  Stream := nil;
  try
  {$IFNDEF CONSOLE}
    if FileName = '' then
      if ClipBoard.HasFormat(CF_TEXT) then begin
        Stream := THugeMemoryStream.Create;
        THugeMemoryStream(Stream).SetFromClipboard(CF_TEXT);
      end else
        raise Exception.Create('No Text on clipboard')
    else
  {$ENDIF}
    begin
      Stream := TMemoryStream.Create;
      TMemoryStream(Stream).LoadFromFile(FileName);
    end;
    WeightedLoadFromStream(Stream, DefaultUnits, FileName, True, AlwaysPrompt, DefaultIntensityUnits);
    Modified := False;
  finally
    Stream.Free;
  end;
end;

function TExperimentalPlot.LocalFrequency(Index: Integer): Double;
begin
  if Calibrating then
    if Length(OriginalFrequency) > 0 then
      Result := OriginalFrequency[Index]
    else
      Result := Frequency[Index]
  else
    Result := AdjustFrequency(Frequency[Index])
end;

{$IFNDEF CONSOLE}
procedure TExperimentalPlot.OffsetBy(Offset: Double);
var
  WasChanging: Boolean;
begin
  if Calibrating and Assigned(FDataSource) then begin
    WasChanging := Changing;
    Changing := True;
    try
      with DataSource as TInteractiveDataSource do begin
        DataSourceFitter.Variable['c'] := DataSourceFitter.Variable ['c'] + Offset;
        NLFitDlg.PlotFunctionClick(nil);
      end;
    finally
      Changing := WasChanging;
    end;
  end else
    inherited OffsetBy(Offset);
end;
{$ENDIF}

function TExperimentalPlot.PrepareConvolution(Simulation: TSimulation; PlotIntensityUnits: TIntensityUnits):Boolean;
var
  xs, ys: array of Double;
  i, n: Integer;
  C: TColor;
begin
  if Assigned(Simulation) and Convolute and Simulation.HasWidths then begin
    Result := True;
    C := Colour;
    if (Simulation = CurrentSimulation) then begin
      ConvolutedSpectrum.Colour := C;
      Exit;
    end;
    FreeAndNil(ConvolutedSpectrum);
    { Convolute with simulation width }
    ConvolutedSpectrum := TSpectrumPlot.Create(Simulation.PlotUnits,False);
    ConvolutedSpectrum.InitializeFromSimulation(Simulation);
    Colour := C;
    if IsLineList then begin
      for i := StartPoint to EndPoint do
        ConvolutedSpectrum.AddLine(Frequency[i],ConvertIntensity(Intensity[i],Frequency[i],Units,IntensityUnits,Units,PlotIntensityUnits,True),0,Colour,Units);
    end else begin
      { Record all points, to allow easy division by separation }
      n := 0;
      SetLength(xs, EndPoint-StartPoint+1);
      SetLength(ys, EndPoint-StartPoint+1);
      for i := StartPoint to EndPoint do
        if GetPoint(i,Simulation.PlotUnits,PlotIntensityUnits,xs[n],ys[n]) then
          Inc(n);
      for i := 1 to n - 2 do begin
        { General case - work out separation from two adjoining gaps }
        ConvolutedSpectrum.AddLine(xs[i],ConvertIntensity(xs[i],ys[i]*Abs(xs[i+1]-xs[i-1])/2,Units,IntensityUnits,Units,PlotIntensityUnits,True),0,Colour,Units);
      end;
      if n > 1 then begin
        { Add end points }
        ConvolutedSpectrum.AddLine(xs[0],ConvertIntensity(xs[i],ys[0]*Abs(xs[1]-xs[0]),Units,IntensityUnits,Units,PlotIntensityUnits,True),0,Colour,Units);
        ConvolutedSpectrum.AddLine(xs[n-1],ConvertIntensity(xs[i],ys[n-1]*Abs(xs[n-1]-xs[n-2]),Units,IntensityUnits,Units,PlotIntensityUnits,True),0,Colour,Units);
      end;
    end;
    if ConvolutedSpectrum.nLinesInRange > 0 then begin
      ConvolutedSpectrum.Convolute;
      ConvolutedSpectrum.FindLimits;
    end;
    CurrentSimulation := Simulation;
  end else begin
    CurrentSimulation := nil;
    FreeAndNil(ConvolutedSpectrum);
    Result := False;
  end;
end;

function TExperimentalPlot.FixupProperties:Exception;
begin
  AdjustParameterNames;
  Result := inherited FixupProperties;
  if Assigned(Result) then Exit;
  LastPoint := EndPoint;
  if High(Frequency) < 0 then Exit; { Empty }
  Limits(Frequency, FHmin, FHMax, StartPoint, EndPoint);
  { TODO: sort by frequency? }
  { Set defaults for vertical scaling }
  Limits(Intensity, FImin, FImax, StartPoint, EndPoint);
  if IsLineList then
    FIMin := Min(FIMin,0);
  FIMin := (FIMin-Baseline)*Scale;
  FIMax := (FIMax-Baseline)*Scale;
  InOrder(FIMin,FIMax);
end;

function TExperimentalPlot.Show(
   G: TGraphM; Simulation: TSimulation; PlotIntensityUnits: TIntensityUnits;
   PlotRange: TPlotRange; var PlotRanges: TPlotRanges; OwnAxis: Boolean;
   ShowName: Boolean; ForceAxis: Boolean):Boolean;
var
  x, y: Double;
  i, nPoints: Integer;
  xl, xu, yl, yu, Fi, Fx: Double;
  OwnRange: Boolean;
  IU: TIntensityUnits;
begin
  Result := False;
  if (Colour = clHide) or (StartPoint > EndPoint) then
    Exit;
  if (StartPoint = EndPoint) and not IsLineList then
    Exit;
  if Colour <> clNone then
    G.StyleP(Colour, G.LineStyle.Width, G.LineStyle.Pattern)
  else
    G.Style(1);
  G.MapArg(xl, xu, yl, yu);
  Fi := xl;
  Fx := xu;
  InOrder(Fi, Fx);
  OwnRange := OwnAxis { or IsArbitraryStyle(PlotIntensityUnits) }
              or IsArbitraryStyle(IntensityUnits);
              { or IsArbitraryStyle((Owner as TOverlays).IntensityUnits); }
  if OwnRange then begin
    y := PlotRange.yi/PlotRange.yx;
    LimitsInRange(Simulation, PlotRange.yi, PlotRange.yx, PlotIntensityUnits);
    { Arbitrary unit plots will normally be plotted in range 0..1, but the logic below
      will extend the lower range so that the zero levels line up }
    if not OwnAxis and ZeroValid then
      PlotRange.yi := y*PlotRange.yx;
  end else
    PrepareConvolution(Simulation, PlotIntensityUnits); { Convolution normally done by Limits routine; but limits routine not always called }
  MapFromPlotRange(G,PlotRange);
  if Assigned(CurrentSimulation) then
    ConvolutedSpectrum.Show(G, PlotRange, False, False, PlotRanges)
  else begin
    nPoints := 0;
    for i := StartPoint to EndPoint do begin
      if GetPoint(i,Simulation.PlotUnits,PlotIntensityUnits,x,y) then begin
        if IsLineList then begin
          G.ClipPoint(x, 0);
          G.ClipJoin(x, y);
        end else if nPoints = 0 then
          G.ClipPoint(x, y)
        else
          G.ClipJoin(x,y);
        Inc(nPoints);
      end;
    end;
  end;
  AddPlotRange(PlotRanges, PlotRange);
  G.Map(xl, xu, yl, yu);
  if ForceAxis or (OwnAxis and not IsArbitraryStyle(IntensityUnits)) then begin
    IU := PlotIntensityUnits;
    if IsArbitraryStyle(IU) then
      IU := IntensityUnits;
    AxisFromPlotRange(PlotRange, G, xl, NameOfUnit(IU,IsLineList,Simulation.PlotUnits));
  end;
  Result := inherited Show(G,Simulation,PlotIntensityUnits,PlotRange,PlotRanges,OwnAxis,ShowName);
  Result := Result or (nPoints > 0);
end;

function TExperimentalPlot.Resolution: Double;
begin
  if IsLineList then
    Result := 0
  else
    Result := Abs(HMax-Hmin)/Length(Frequency);
end;

function TExperimentalPlot.WriteBasicData(Parent: TDOMNode): TDOMCDATASection;
var
  Strings: TStringList;
  s: string;
  i: Integer;
begin
  Strings := TStringList.Create;
  try
    Strings.Capacity := Length(Intensity)+1;
    Strings.Add('');
    for i := 0 to High(Intensity) do begin
      if Assigned(FDataSource) and (High(OriginalFrequency) > 0) then
        s := FloatToStr(OriginalFrequency[i])
      else
        s := FloatToStr(Frequency[i]);
      s := s +' '+FloatToStr(Intensity[i]);
      if i <= High(Weights) then
        s := s + ' '+FloatToStr(Weights[i]);
      Strings.Add(s);
    end;
    Result := Parent.OwnerDocument.CreateCDATASection(Strings.Text);
  finally
    Strings.Free;
  end;
end;

function TExperimentalPlot.WriteProperties(Parent: TDOMNode; AllParameters: Boolean; WithChildren: Boolean): TDOMElement;
var
  Attr: TDOMAttr;
begin
  Result := inherited WriteProperties(Parent, AllParameters, WithChildren);
  if LastPoint >= High(Intensity) then begin
    { Remove explicit last point from written properties if it corresponds to the
      end of the data }
    Attr := (Result as TDOMElement).GetAttributeNode('LastPoint');
    if Assigned(Attr) then
      RemoveAttributeNode(Result as TDOMElement, Attr);
  end;
  if Assigned(FDataSource) then
    (DataSource as TPlotDataSource).WriteXML(Result);
  Result.AppendChild(WriteBasicData(Parent));
  WritePeaks(Result);
end;

function TExperimentalPlot.SimpleSaveToStream(Stream: TStream; WithLinkedMNodes: Boolean): Boolean;
var
  i: Integer;
  s: string;
begin
  for i := 0 to High(Intensity) do begin
    s := FloatToStr(AdjustFrequency(Frequency[i]))+#9+FloatToStr(Intensity[i]);
    if i <= High(Weights) then
      s := s + #9 +FloatToStr(Weights[i]);
    s := s + sLineBreak;
    Stream.Write(s[1],Length(s))
  end;
  Result := True;
end;

procedure TExperimentalPlot.Smooth(Width, Derivative, Poly: Integer);
var
  C: TDoubleArray;
begin
  C := SmoothingCoefficients(Width div 2, Width div 2, Derivative, Poly);
  ConvoluteData(C);
end;

procedure TExperimentalPlot.ReadData(Child: TDOMCDATASection);
var
  s: string;
  P: PChar;
  nFields, Code, nPoints, nWeights: Integer;
  // Intensity, Frequency, Weights,
begin
  s := Child.Data;
  P := @s[1];
  nPoints := 0;
  nWeights := 0;
  SetLength(Frequency, 0); { Guess at initial length }
  while P^ <> #0 do begin
    if nPoints >= High(Frequency) then begin
      if Length(Frequency) = 0 then
        SetLength(Frequency, Length(s) div 40 + 10) { Guess at length }
      else
        SetLength(Frequency, Length(Frequency)+10000);
      SetLength(Intensity,Length(Frequency));
      SetLength(Weights,Length(Frequency));
    end;
    Weights[nPoints] := 1;
    Code := TryReadTabbedLine(P, [@Frequency[nPoints],@Intensity[nPoints],@Weights[nPoints]], nFields);
    case nFields of
      3: Inc(nWeights);
      1: Intensity[nPoints] := 1;
      0: Dec(nPoints);
    end;
    Inc(nPoints);
    if Code <> 0 then begin
      Inc(P,Code-1);
      case P^ of
        #10: if (P+1)^ = #13 then
               Inc(P,2)
             else
               Inc(P);
        #13: if (P+1)^ = #10 then
               Inc(P,2)
             else
               Inc(P);
      else
        raise EReportException.Create('Unexpected character in overlay data:'+P^);
      end;
    end;
  end;
  if nWeights = 0 then
    SetLength(Weights,0)
  else
    SetLength(Weights,nPoints);
  SetLength(Intensity,nPoints);
  SetLength(Frequency,nPoints);
end;

function TExperimentalPlot.ReadProperties(Node: TDOMNode; CheckUnread: Boolean):Integer;
var
  Child, NextChild: TDOMNode;
  WasChanging: Boolean;
  Attr: TDOMAttr;
  nCoeff: Integer;
  s: string;
begin
  Result := inherited ReadProperties(Node, False);
  { Compatibility: read any unread coefficients, adjusting parent nComponents if necessary }
  Child := Node.FirstChild;
  while Assigned(Child) do begin
    NextChild := Child.NextSibling;
    if CompareText(Child.NodeName, 'Parameter') = 0 then begin
      Attr := (Child as TDOMElement).GetAttributeNode('Name');
      if Assigned(Attr) and (Length(Attr.NodeValue) > 1)
         and (Attr.NodeValue[1] = 'c') and TryStrToInt(Copy(Attr.NodeValue,2,255),nCoeff) and (nCoeff in [1..50]) then begin
        with Owner as TOverlays do
          nComponents := Max(nComponents, nCoeff);
        Result := Result + ReadFloatableValue(Node, nCoeff+FirstDynamicParameter-1, Attr.NodeValue, '0');
      end;
    end;
    Child := NextChild;
  end;
  { Compatibility: FitState -> ComponentType }
  s := ReadValue(Node,'FitState','fsNone');
  if s <> 'fsNone' then
    if s = 'fsTarget' then
      ComponentType := Target
    else if s = 'fsComponent' then
      ComponentType := Component
    else if s = 'fsFloat' then
      ComponentType := OverlaySumFitter.Floating;
  { Read text data }
  Child := Node.FirstChild;
  while Assigned(Child) do begin
    if Child is TDOMCDATASection then begin
      ReadData(TDOMCDATASection(Child));
      ReadPeaks(Child.NextSibling);
      Node.RemoveChild(Child);
      Break;
    end;
    Child := Child.NextSibling;
  end;
  Child := Node.FindNode('Calibration');
  WasChanging := Changing;
  if Assigned(Child) then
    try
      Changing := True;
      Result := Result + (DataSource as TPlotDataSource).ReadXML(Child, CheckUnread);
      if CheckUnread then
        Result := Result + nUnreadProperties(Child);
      Node.RemoveChild(Child);
    finally
      Changing := WasChanging;
    end;
  if CheckUnread then
    Result := Result + nUnreadProperties(Node);
end;

constructor TExperimentalPlot.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',LastPoint=-1,Scale=1';
  DimensionlessParameters := DimensionlessParameters + ',Baseline,Scale';
  inherited;
  AddMNodeNotification({$IFDEF FPC}@{$ENDIF}DoMNodeChange);
  AdjustParameterNames;
end;

procedure TExperimentalPlot.Crop(First, Last: Integer);
var
  i: Integer;
begin
  First := Max(First,0);
  Last := Min(Last,EndPoint);
  if First > 0 then
    for i := 0 to Last-First do begin
      Frequency[i] := Frequency[First+i];
      Intensity[i] := Intensity[First+i];
      if Length(Weights) > 0 then
        Weights[i] := Weights[First+i];
      if Length(OriginalFrequency) > 0 then
        OriginalFrequency[i] := OriginalFrequency[First+i];
    end;
  SetLength(Frequency, Last-First+1);
  SetLength(Intensity, Last-First+1);
  if Length(Weights) > 0 then
    SetLength(Weights, Last-First+1);
  if Length(OriginalFrequency) > 0 then
    SetLength(OriginalFrequency, Last-First+1);
  FirstPoint := Max(0,FirstPoint-First);
  if LastPoint > 0 then
    LastPoint := Min(High(Frequency),LastPoint-First);
  FixupProperties;
end;

procedure TExperimentalPlot.SaveForUndo;
begin
  FLastIntensity := Intensity;
end;

procedure TExperimentalPlot.ScaleParameters(UFrom, UTo: TResonanceUnits);
var
  i: Integer;
  Sum: Double;
begin
  inherited;
  Sum := 0;
  for i := 0 to High(Frequency) do begin
    Sum := Sum + Frequency[i];
    Frequency[i] := Convert(Frequency[i], UFrom, UTo);
  end;
  if Length(Frequency) > 0 then
    FrequencyOffset := ConvertWidth(FrequencyOffset, Sum/Length(Frequency), UFrom, Uto);
  CheckProperties;
end;

function TExperimentalPlot.FindPeak(AFmin, AFmax: Double; out Position: Double; PlotUnits: TResonanceUnits): Boolean;
var
  i: Integer;
  First, Last: Integer; { Range of points within given frequency range }
  Mx, Mi: Double;
  Threshold, SumI, SumIX, v, f: Double;
begin
  AFmin := Convert(AFmin, PlotUnits, Units);
  AFmax := Convert(AFmax, PlotUnits, Units);
  InOrder(AFmin, AFmax);
  Threshold := 0.5;
  if Owner is TOverlays then
    Threshold := TOverlays(Owner).PeakThreshold;
  First := -1;
  Last := -1;
  Result := False;
  for i := StartPoint to EndPoint do begin
    f := AdjustFrequency(Frequency[i]);
    if (f >= AFmin) and (f <= AFmax) then begin
      Last := i;
      if First < 0 then begin
        First := i;
        Mx := Intensity[i];
        Mi := Intensity[i];
      end else begin
        if Intensity[i] < Mi then
          Mi := Intensity[i];
        if Intensity[i] > Mx then
          Mx := Intensity[i];
      end;
    end;
  end;
  if (First = Last) or ((Mx = Mi) and not IsLineList) then begin
    { 1 or fewer points }
    if (First >= 0) and IsLineList then begin
      Position := LocalFrequency(First);
      Result := True;
    end;
  end else begin
    { More than one point, so use centre of mass algorithm
      on the points above the threshold }
    Threshold := Mi + Threshold*(Mx-Mi);
    SumI := 0;
    SumIX := 0;
    for i := First to Last do begin
      f := LocalFrequency(i);
      if IsLineList then
        v := Abs(Intensity[i])
      else
        v := Intensity[i] - Threshold;
      if v > 0 then begin
        SumI := SumI + v;
        SumIX := SumIX + v * f;
      end;
    end;
    if SumI > 0.01*(Mx-Mi) then begin
      { Check region not too flat }
      Position := SumIX / SumI;
      if not IsLineList then
        { Round to a sensible number of places }
        Position := RoundToIncrement(Position, Resolution/20);
      Result := True;
    end;
  end;
end;

procedure TExperimentalPlot.Invert;
var
  i: Integer;
begin
  Limits(Intensity, FImin, FImax); { Just in case }
  for i := 0 to High(Intensity) do
    Intensity[i] := FImax+FImin-Intensity[i];
  Modified := True;
end;

procedure TExperimentalPlot.JoinTo(JoinPoint: Double; Plot: TExperimentalPlot; ScalePlot: Boolean; IncreasingFrequency: Boolean);
var
  ThisPoint, OtherPoint: Integer;
  procedure Process(var Left: TDoubleArray; const Right: TDoubleArray; VertScale: Boolean);
  var
    i: Integer;
    v: Double;
  begin
    SetLength(Left,ThisPoint+1+(Length(Right)-OtherPoint));
    for i := ThisPoint+1 to High(Left) do
      if VertScale then begin
        v := (Right[i-ThisPoint-1+OtherPoint]-Plot.Baseline)*Plot.Scale;
        v := v/Scale+Baseline;
        Left[i] := v;
      end else
        Left[i] := Right[i-ThisPoint-1+OtherPoint];
  end;
  procedure FixOrder(APlot: TExperimentalPlot);
  begin
    with APlot do
      if IncreasingFrequency xor not IsEnergyUnit(Units) xor (FrequencyScale < 0)
         xor (Frequency[0] < Frequency[High(Frequency)]) then begin
        Sort(Frequency[0] > Frequency[High(Frequency)]);
        CheckProperties;
      end;
  end;
var
  i: Integer;
begin
  { Insist on easy case }
  if Units <> Plot.Units then
    raise EReportException.Create('Overlays have different units');
  { Logic below requires x values in same order as plot }
  FixOrder(Self);
  FixOrder(Plot);

  if Plot.HMin < HMin then
    raise EReportException.Create('Overlays out of order');
  { Ensure minimum number of points lost }
  JoinPoint := Max(Min(JoinPoint,HMax),Plot.HMin);
  IntensityAt(JoinPoint, ThisPoint);
  Plot.IntensityAt(Convert(JoinPoint,Units,Plot.Units), OtherPoint);
  Dec(ThisPoint);
  if ThisPoint <= 0 then
    raise EReportException.Create('No contribution from left overlay');
  if OtherPoint >= High(Plot.Frequency) then
    raise EReportException.Create('No contribution from right overlay');
  Process(Intensity,Plot.Intensity,ScalePlot);
//  Process(Frequency,Plot.Frequency,False);
  { Copy frequency, removing any frequency scaling }
  SetLength(Frequency,Length(Intensity));
  for i := 0 to ThisPoint do
    Frequency[i] := AdjustFrequency(Frequency[i]);
  for i := ThisPoint+1 to High(Frequency) do
    Frequency[i] := Plot.AdjustFrequency(Plot.Frequency[i-ThisPoint-1+OtherPoint]);
  DiscardCalibration;
  if Assigned(Weights) then
    Process(Weights,Plot.Weights,False);
  LastPoint := -1;
  CheckProperties;
end;

function TExperimentalPlot.IntensityAt(Freq: Double; out Point: Integer):Double; { Interpolate Frequency }
var
  i: Integer;
begin
  Freq := DeAdjustFrequency(Freq);
  if High(Frequency)<1 then begin
    Point := -1;
    Result := 0;;
    Exit;
  end;
  i := 1;
  { Find an interval i, i-1; if off end choose last interval }
  if Frequency[0] > Frequency[High(Frequency)] then
    while (Freq < Frequency[i]) and (i < High(Frequency)) do
      Inc(i)
  else
    while (Freq > Frequency[i]) and (i < High(Frequency)) do
      Inc(i);
  Result := Frequency[i]-Frequency[i-1];
  if Abs(Result) > 0 then
    Result := (Freq-Frequency[i-1])/Result
  else
    Result := Frequency[i]; { Zero interval }
  Result := Result * (Intensity[i]-Intensity[i-1]) + Intensity[i-1];
  Result := (Result-Baseline)*Scale;
  Point := i;
end;

function TExperimentalPlot.EndPoint: Integer;
begin
  if (LastPoint < 0) or (LastPoint > High(Intensity)) then
    Result := High(Intensity)
  else
    Result := LastPoint;
end;

function TExperimentalPlot.Etalon(Spacing: Double):Integer;
var
  P, Delta, OrgDelta, ADelta: Double;
  Mi, Mx: Double;
  Peak: TPeak;
  SetFirst: Boolean;
  First: Integer;
  AScale: Double;
begin
  { Trim any trailing blank peaks }
  First := nPeaks-1;
  while (First >= 0) and IsNan(Peaks[First].Position) do
    Dec(First);
  nPeaks := First+1;
  if nPeaks < 2 then
    raise EReportException.Create('Please find the first two peaks');
  Mi := LocalFrequency(0);
  Mx := LocalFrequency(High(Frequency));
  InOrder(Mi,Mx);
  { Trim any trailing blank peaks }
  First := nPeaks;
  Peak := EmptyPeak;
  Peak.Actual := 0;
  ADelta := Spacing;
  if not IsNan(Actual[nPeaks-1]) then
    if not IsNan(Actual[nPeaks-2]) then begin
      ADelta := Actual[nPeaks-1] - Actual[nPeaks-2]; { Both peaks set }
      Peak.Actual := Actual[nPeaks-2];
    end else
      Peak.Actual := Actual[nPeaks-1] - ADelta      { Only second }
  else if not IsNan(Actual[nPeaks-2]) then
      Peak.Actual := Actual[nPeaks-2];              { Only first }
  if IsNan(Actual[nPeaks-1]) then begin
    Actual[nPeaks-1] := Peak.Actual + ADelta;
    PeakChanged(nPeaks-1);
  end;
  if IsNan(Actual[nPeaks-2]) then begin
    Actual[nPeaks-2] := Peak.Actual;
    PeakChanged(nPeaks-2);
  end;
  Peak.Actual := Peak.Actual + 2*ADelta;
  P := Peaks[nPeaks-1].Position;
  OrgDelta := P - Peaks[nPeaks-2].Position;
  AScale := 1; { Progress display needs an integer }
  while Max(Abs(Mi),Abs(Mx))*AScale >= High(Integer) do
    AScale := AScale / 10;
  PushProgress('Position', Round(Mx*AScale));
  try
    repeat
      Delta := P - Peaks[nPeaks-2].Position;
      Peak.Position := P + Delta;
      Delta := Abs(Delta/2);
      if (Peak.Position-Delta/2 >= Mi) and (Peak.Position+Delta/2 <= Mx)
          and FindPeak(ConvertPeakPosition(Peak.Position-Delta,Units),
                       ConvertPeakPosition(Peak.Position+Delta,Units),
                       Peak.Position,Units) then
        if Abs(Peak.Position-P) > OrgDelta/2 then
          AddPeak(Peak)
        else
          Break
      else
        Break;
      SetProgress(Round(P*AScale));
      P := Peak.Position;
      Peak.Actual := Peak.Actual + ADelta;
    until False;
  finally
    PopProgress;
  end;
  Result := nPeaks-First;
end;

function TExperimentalPlot.StartPoint: Integer;
begin
  if FirstPoint < 0 then
    Result := 0
  else
    Result := FirstPoint;
end;

procedure TExperimentalPlot.SwapXandY;
var
  T: TDoubleArray;
begin
  T := Frequency;
  Frequency := Intensity;
  Intensity := T;
  Modified := True;
end;

procedure TExperimentalPlot.Undo;
begin
  Intensity := FLastIntensity;
  FLastIntensity := nil;
end;

procedure TExperimentalPlot.SelectFrequencyRange(AFmin, AFmax: Double);
var
  i, First, Last: Integer;
begin
  First := High(Frequency);
  Last := 0;
  for i := 0 to High(Frequency) do
    if (Frequency[i] >= AFmin) and (Frequency[i] <= AFmax) then begin
      if First > i then First := i;
      if Last < i then Last := i;
    end;
  FirstPoint := First;
  LastPoint := Last;
end;

procedure TExperimentalPlot.Separate;
  procedure CreatePlot(NewPlot: TExperimentalPlot; var Data: TDoubleArray; const Tag: string);
  var
    i: Integer;
  begin
    NewPlot.Name := Name + '_' + Tag;
    NewPlot.Comment := Comment + ' ' + Tag;
    NewPlot.Intensity := Data;
    Data := nil;
    with NewPlot do begin
      SetLength(Frequency,Length(Intensity));
      for i := 0 to High(Frequency) do
        Frequency[i] := i;
    end;
  end;
begin
  CreatePlot(TExperimentalPlot.Create(Owner), Intensity, 'Intensity');
  if Length(Weights) > 0 then
    CreatePlot(TExperimentalPlot.Create(Owner), Weights, 'Weights');
  if Length(OriginalFrequency) > 0 then
    CreatePlot(Self, OriginalFrequency, 'Frequency')
  else
    CreatePlot(Self, Frequency, 'Frequency')
end;

procedure TExperimentalPlot.SetComponentType(const Index: Integer; const Value: TComponentType);
begin
  SetSetting(Index, Ord(Value));
end;

procedure TExperimentalPlot.SetModified(const Value: Boolean);
begin
  inherited;
  if Value then
    PrepareConvolution(nil,Arbitrary);
end;

procedure TExperimentalPlot.SetParameter(AIndex: Integer; const Value: Double);
begin
  if AIndex >= 0 then
    inherited
  else { Negative parameter numbers are used for the Intensity values }
    Intensity[-AIndex-1] := Value;
end;

procedure TExperimentalPlot.SetResidualPlotter(const Value: TObject);
begin
  // Do not: FreeAndNil(FResidualPlotter);
  FResidualPlotter := Value;
end;

type
  TExperimentalPlotSorter = class(TShellSorter)
    E: TExperimentalPlot;
    Descending: Boolean;
    function InOrder(I,J:Integer):Boolean; override;
    procedure Swap(I,J: Integer); override;
  end;

procedure TExperimentalPlot.Sort(Ascending: Boolean = True);
begin
  with TExperimentalPlotSorter.Create do begin
    E := Self;
    Descending := not Ascending;
    Sort(Length(Frequency));
    Free;
  end;
end;

procedure TExperimentalPlot.Split(SplitPoint: Double);
var
  ThisPoint: Integer;
  NewPlot: TExperimentalPlot;
  procedure Process(var L, R: TDoubleArray);
  var
    i: Integer;
  begin
    SetLength(R,High(L)-ThisPoint);
    for i := ThisPoint+1 to High(L) do
      R[i-ThisPoint-1] := L[i];
    SetLength(L,ThisPoint+1);
  end;
begin
  IntensityAt(SplitPoint, ThisPoint);
  if (ThisPoint >= High(Frequency)) or (ThisPoint <= 1) then
    raise EReportException.Create('Split point off end of overlay');
  Dec(ThisPoint);
  NewPlot := TExperimentalPlot.Create(Owner);
  NewPlot.Name := Name + ' part';
  NewPlot.Baseline := Baseline;
  NewPlot.Scale := Scale;
  NewPlot.Units := Units;
  NewPlot.IntensityUnits := IntensityUnits;
  NewPlot.Colour := Colour;
  NewPlot.IsLineList := IsLineList;
  Process(Intensity,NewPlot.Intensity);
  Process(Frequency,NewPlot.Frequency);
  DiscardCalibration;
  if Assigned(Weights) then
    Process(Weights,NewPlot.Weights);
  LastPoint := -1;
  CheckProperties;
  NewPlot.CheckProperties;
end;

procedure TExperimentalPlot.AddNodeParameters(var FittingParameters: TParameters);
var
  i, n: Integer;
  s: string;
begin
  inherited;
  if (StartPoint > EndPoint) or (ComponentType <> OverlaySumFitter.Floating) then
    Exit;
  s := UniqueName;
  if s <> '' then
    s := s + ' ';
  n := Length(FittingParameters);
  SetLength(FittingParameters, n+EndPoint-StartPoint+1);
  for i := StartPoint to EndPoint do
    with FittingParameters[n+i-StartPoint] do begin
      MNode := Self;
      Index := -i-1;
      Value := Intensity[i];
      Increment := Max(Abs(Value),1);
      Inverse := False;
      Name := s + FloatToStr(Frequency[i]);
    end;
end;

procedure TExperimentalPlot.AdjustParameterNames;
var
  i, n: Integer;
begin
  if not Assigned(Owner) then
    Exit;
  n := (Owner as TOverlays).nComponents;
  { Add/remove coefficients parameters as required }
  for i := nParameters+1 to FirstDynamicParameter + n - 1 do
    AddParameter(Format('c%d',[i-FirstDynamicParameter+1]));
  for i := nParameters downto FirstDynamicParameter + n do
    DeleteParameter(i);
end;

procedure TExperimentalPlot.ApplyCalibration(APlot: TExperimentalPlot; SamePoints: Boolean);
var
  i: Integer;
begin
  if SamePoints and (High(Frequency) <> High(APlot.Frequency)) then
    raise EReportException.CreateFmt('Overlays must be the same length - actually %d and %d', [Length(Frequency), Length(APlot.Frequency)]);
  with APlot do begin
//    { Preserve previous frequency scale }
//    if Length(OriginalFrequency) = 0 then begin
//      OriginalFrequency := Frequency;
//      Frequency := nil;
//      SetLength(Frequency,Length(OriginalFrequency));
//    end;
    FrequencyScale := 1;
    FrequencyOffset := 0;
  end;
  if APlot <> Self then
    FreeAndNil(Aplot.FDataSource);
  SamePoints := SamePoints or (Self = APlot);
  with DataSource as TPlotDataSource do
    for i := 0 to High(APlot.Frequency) do begin
      if SamePoints then
        xVariable.Value := OriginalFrequency[i]
      else
        xVariable.Value := APlot.Frequency[i]; // Was OriginalFrequency
      APlot.Frequency[i] := FitExpr.Value;
    end;
  if Aplot = Self then
    FreeAndNil(FDataSource)
  else
    Aplot.Units := Units;
  APlot.OriginalFrequency := nil; { Added }
  Aplot.CheckProperties;
end;

function TExperimentalPlot.AutoZero(Bins: Integer; First, Last: Integer):Double;
var
  Counts: array of Integer;
  Sum: array of Double;
  I, Bin: Integer;
  Ii, Ix: Double;
{ Try and set the zero of a spectrum }
begin
  { Zero detetction must be on unscaled data }
  if First < 0 then
    First := StartPoint;
  if Last < 0 then
    Last := EndPoint;
  Limits(Intensity, Ii, Ix, First, Last);
  if Bins < 0 then
    Bins := (Last-First) div 20 + 1;
  if Bins > 0 then begin
    SetLength(Counts,Bins);
    SetLength(Sum,Bins);
    FillChar(Counts[0],Bins*Sizeof(Integer),0);
    FillChar(Sum[0],Bins*Sizeof(Double),0);
    for I := First to Last do begin
      Bin := Min(Floor((Intensity[I]-Ii)/(Ix-Ii)*Bins),Bins-1);
      Inc(Counts[Bin]);
      Sum[Bin] := Sum[Bin] + Intensity[I];
    end;
    Bin := 0;
    for I := 0 to Bins - 1 do
      if Counts[I] > Bin then begin
        Bin := Counts[I];
        Result := Sum[I]/Bin;
      end;
  end else if Scale < 0 then
    Result := Ix
  else
    Result := Ii;
end;

function TExperimentalPlot.CanUndo: Boolean;
begin
  Result := Length(FLastIntensity) > 0;
end;

procedure TExperimentalPlot.ClearSelection;
begin
  FirstPoint := 0;
  LastPoint := High(Intensity);
end;

procedure TExperimentalPlot.Compress(nAverage: Integer);
  procedure Process(var V: TDoubleArray);
  var
    s, d, j, n: Integer;
    Sum: Double;
  begin
    s := 0;
    d := 0;
    while s < Length(V) do begin
      n := Min(nAverage, Length(V)-s);
      Sum := 0;
      for j := 0 to n - 1 do begin
        Sum := Sum + V[s];
        Inc(s);
      end;
      V[d] := Sum/n;
      Inc(d);
    end;
    SetLength(V,d);
  end;
begin
  Process(Intensity);
  Process(Frequency);
  if Assigned(OriginalFrequency) then
    Process(OriginalFrequency);
  if Assigned(Weights) then
    Process(Weights);
  FirstPoint := 0;
  LastPoint := -1;
  CheckProperties;
end;

procedure TExperimentalPlot.ReBin(xInterval: Double);
var
  x: Double;
  s, d, e, NewN, LastPointSet: Integer;
  NewFrequency, NewIntensity, NewWeights: TDoubleArray;
  procedure Average(NewY, Y: TDoubleArray);
  var
    Sum: Double;
    i: Integer;
  begin
    if Length(Y) = 0 then
      Exit;
    if e > s then begin
      Sum := 0;
      for i := s to e - 1 do
        Sum := Sum + Y[i];
      NewY[d] := Sum/(e-s);
    end;
    if LastPointSet < d-1 then begin
      { No source points contribute to preceeding points, so interpolate }
      for i := LastPointSet+1 to d - 1 do
        NewY[i] := NewY[LastPointSet] + (NewY[d]-NewY[LastPointSet])*(i-LastPointSet)/(d-LastPointSet);
    end;
  end;
  procedure SetLengths(n: Integer);
  begin
    { Create new arrays to copy into, as can't be sure source rate is less than destination rate }
    SetLength(NewFrequency,n);
    SetLength(NewIntensity,n);
    if Length(Weights) > 0 then
      SetLength(NewWeights,n);
  end;
begin
  DiscardCalibration; { OriginalFrequency is no longer needed }
  Sort; { Check in order }
  { Centre of first new frequency interval }
  x := Trunc(Frequency[0]/xInterval)*xInterval;
  { New number of points - may not be quite right because of rounding errors }
  newN := Trunc((Frequency[High(Frequency)]-Frequency[0])/xInterval)+1;
  s := 0;
  d := 0;
  LastPointSet := 2*NewN+100;
  while s < Length(Frequency) do begin
    { Find end of interval }
    e := s;
    while (e < Length(Frequency)) and (Frequency[e] < x+xInterval/2) do
      Inc(e);
    if (d > 0) or (e > s) then begin { Test may skip first point }
      newN := Max(newN,d+1);
      if newN > Length(NewFrequency) then
        SetLengths(newN);
      NewFrequency[d] := x;
      Average(NewIntensity,Intensity);
      Average(NewWeights,Weights);
      if e > s then
        LastPointSet := d;
      Inc(d);
    end;
    s := e;
    x := x + xInterval;
  end;
  SetLengths(LastPointSet+1);
  Frequency := NewFrequency;
  Intensity := NewIntensity;
  Weights := NewWeights;
  FirstPoint := 0;
  LastPoint := -1;
  CheckProperties;
end;

function TExperimentalPlot.ConvertPeakPosition(Position: Double;
  PlotUnits: TResonanceUnits): Double;
begin
  if Calibrating and (Length(OriginalFrequency) > 0) and Assigned(FDataSource) then
    with TPlotDataSource(DataSource) do begin
      xVariable.Value := Position;
      Position := FitExpr.Value;
    end;
  Result := inherited ConvertPeakPosition(Position, PlotUnits)
end;

procedure TExperimentalPlot.ConvoluteData(Values: array of Double);
var
  i, j, n: Integer;
  V : TDoubleArray;
  Sum: Double;
begin
  SaveForUndo;
  SetLength(V,Length(Intensity));
  n := Length(Values) div 2;
  for I := 0 to High(Intensity) do begin
    Sum := 0;
    for J := -n to n do
      Sum := Sum + Intensity[Min(Max(I+J,0),High(Intensity))]*Values[J+n];
    V[I] := Sum;
  end;
  Intensity := V;
  CheckProperties;
end;

function TExperimentalPlot.GetDataSource: TDataSource;
begin
  if not Assigned(FDataSource) then begin
    FDataSource := TPlotDataSource.Create(nil);
    TPlotDataSource(FDataSource).Plot := Self;
  end;
  Result := FDataSource;
end;

function TExperimentalPlot.GetComponentType(const Index: Integer): TComponentType;
begin
  Result := TComponentType(GetSetting(Index));
end;

function TExperimentalPlot.GetIMin: Double;
begin
  Result := FIMin;
end;

function TExperimentalPlot.GetPoint(i: Integer; PlotUnits: TResonanceUnits; PlotIntensityUnits: TIntensityUnits; var x, y: Double): Boolean;
begin
  { Find and convert x, y for point i }
  x := AdjustFrequency(Frequency[i]);
  if (x > 0) or not (IsWavelengthUnit(PlotUnits) xor IsWavelengthUnit(Units)) then begin
    y := ConvertIntensity((Intensity[i]-Baseline)*Scale,x,Units,IntensityUnits,PlotUnits,PlotIntensityUnits,IsLineList);
    x := Convert(x,Units,PlotUnits);
    Result := True;
  end else
    Result := False;
end;

function TExperimentalPlot.GetIMax: Double;
begin
  Result := FIMax;
end;

const
  ExperimentalPlotSettingNames: array[7..10] of string = (
'First point to use in fitting and plotting.',
'Last point to use in fitting and plotting.',
'Set to treat data as a list of line positions, optionally with intensity.',
'Set to convolute with simulation linewidth(s)'
);

function TExperimentalPlot.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(ExperimentalPlotSettingNames))
     and (AIndex <= High(ExperimentalPlotSettingNames)) then
    Result := ExperimentalPlotSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

const
 ExperimentalPlotParameterNames: array [2..4] of string = (
'Scaling factor for horizontal scale.',
'Baseline position.',
'Vertical scale factor used in plotting.'
);

function TExperimentalPlot.HasDataSource: Boolean;
begin
  Result := Assigned(FDataSource);
end;

function TExperimentalPlot.HelpOnParameter(AIndex: Integer): string;
var
  I, J: Integer;
begin
  if Aindex >= FirstDynamicParameter then begin
    Dec(AIndex, FirstDynamicParameter);
    with Owner as TOverlays do
      for I := 0 to nPlots-1 do
        if (Plots[i] is TExperimentalPlot) and (TExperimentalPlot(Plots[i]).ComponentType in [Component, OverlaySumFitter.Floating]) then
          if AIndex = 0 then begin
            Result := 'Multiplier of '+Plots[i].Name;
            Break;
          end else
            Dec(AIndex);
  end else if (AIndex >= Low(ExperimentalPlotParameterNames))
     and (AIndex <= High(ExperimentalPlotParameterNames)) then
    Result := ExperimentalPlotParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

function TExperimentalPlot.Empty: Boolean;
begin
  Result := (High(Frequency) < 0) or ((High(Frequency) = 0) and not IsLineList);
end;

destructor TExperimentalPlot.Destroy;
begin
  FreeAndNil(FDataSource);
  FreeAndNil(FResidualPlotter);
  PrepareConvolution(nil,Arbitrary); { Clear convoluted spectrum cache }
  RemoveMNodeNotification({$IFDEF FPC}@{$ENDIF}DoMNodeChange);
  inherited;
end;

procedure TExperimentalPlot.DiscardCalibration;
begin
  OriginalFrequency := nil;
  FreeAndNil(FDataSource);
  FrequencyOffset := 0;
  FrequencyScale := 1;
end;

procedure TExperimentalPlot.DiscardFrequency;
var
  I: Integer;
begin
  for I := 0 to High(Frequency) do
    Frequency[I] := I;
  DiscardCalibration;
end;

procedure TExperimentalPlot.DiscardIntensity;
var
  I: Integer;
begin
  for I := 0 to High(Intensity) do
    Intensity[I] := 1;
  IsLineList := True;
  Modified := True;
end;

procedure TExperimentalPlot.DoMNodeChange(Sender: TMNode; Change: TMNodeChange;
  Index: Integer);
begin
  if Sender is TSimulation then
    PrepareConvolution(nil,Arbitrary); { Clear convoluted spectrum cache }
end;

{ TExperimentalPlotSorter }

function TExperimentalPlotSorter.InOrder(I, J: Integer): Boolean;
begin
  if Descending then
    Result := E.Frequency[I] >= E.Frequency[J]
  else
    Result := E.Frequency[I] <= E.Frequency[J];
end;

procedure TExperimentalPlotSorter.Swap(I, J: Integer);
var
  T: Double;
begin
  with E do begin
    T := Frequency[I];
    Frequency[I] := Frequency[J];
    Frequency[J] := T;
    T := Intensity[I];
    Intensity[I] := Intensity[J];
    Intensity[J] := T;
    if Length(Weights) > 0 then begin
      T := Weights[I];
      Weights[I] := Weights[J];
      Weights[J] := T;
    end;
    if Length(OriginalFrequency) > 0 then begin
      T := OriginalFrequency[I];
      OriginalFrequency[I] := OriginalFrequency[J];
      OriginalFrequency[J] := T;
    end;
  end;
end;

{ TPGOPHERPlotLoaderConnector }

function TPGOPHERPlotLoaderConnector.GetBaseName: string;
begin
  if FBaseName = '' then
    FBaseName := ExperimentalPlot.Name;
  Result := FBaseName;
end;

function TPGOPHERPlotLoaderConnector.GetComment: string;
begin
  Result := ExperimentalPlot.Comment;
end;

procedure TPGOPHERPlotLoaderConnector.ReportError(Msg: string; Fatal: Boolean);
begin
  if Fatal then
    raise EReportException.Create(Msg)
  else begin
    {$IFNDEF CONSOLE}
      inherited;
    {$ELSE}
      WriteLn(Log, 'Error in file load:', Msg);
    {$ENDIF}
  end;
end;

procedure TPGOPHERPlotLoaderConnector.SetColour(Plot: Integer; Colour: Integer);
var
  I: Integer;
begin
  if Plot < 0 then
    for I := 0 to nPlots - 1 do
      (Plots[I] as TPGOPHERPlotLoaderConnector).ExperimentalPlot.Colour := Colour
  else
    ExperimentalPlot.Colour := Colour;
end;

procedure TPGOPHERPlotLoaderConnector.SetComment(const AValue: string);
begin
  ExperimentalPlot.Comment := AValue;
end;

constructor TPGOPHERPlotLoaderConnector.Create(AnObject: TObject; AStreamName: string);
begin
  inherited Create(AnObject, AStreamName);
  {$IFNDEF CONSOLE}
  PickDestForm.Xname := 'Frequency';
  PickDestForm.Yname := 'Intensity';
  {$ENDIF}
  FExperimentalPlot := AnObject as TExperimentalPlot;
  Frequency := ExperimentalPlot.Frequency;
  Intensity := ExperimentalPlot.Intensity;
  Weights   := ExperimentalPlot.Weights;
end;

destructor TPGOPHERPlotLoaderConnector.Destroy;
begin
  ExperimentalPlot.Frequency := Frequency;
  ExperimentalPlot.Intensity := Intensity;
  ExperimentalPlot.Weights := Weights;
  inherited Destroy;
end;

procedure TPGOPHERPlotLoaderConnector.SetLineList(Plot: Integer; IsLineList: Boolean);
var
  i: Integer;
begin
  if Plot < 0 then
    for i := 0 to nPlots - 1 do
      (Plots[i] as TPGOPHERPlotLoaderConnector).ExperimentalPlot.IsLineList := IsLineList
  else
    (Plots[Plot] as TPGOPHERPlotLoaderConnector).ExperimentalPlot.IsLineList := IsLineList;
end;

procedure TPGOPHERPlotLoaderConnector.SetName(const AValue: string);
begin
  if FBaseName = '' then
    FBaseName := ExperimentalPlot.Name;
  if FBaseName = '' then
    ExperimentalPlot.Name := AValue
  else
    ExperimentalPlot.Name := FBaseName + ' ' + AValue;
end;

procedure TPGOPHERPlotLoaderConnector.SetIntensityUnits(IntensityUnits: TIntensityUnits);
begin
  ExperimentalPlot.IntensityUnits := IntensityUnits;
end;

procedure TPGOPHERPlotLoaderConnector.SetUnits(Plot: Integer; AUnits: TResonanceUnits);
var
  i: Integer;
begin
  with ExperimentalPlot.Owner as TOverlays do
    if Units <> AUnits then
      if (Count = 1) or (Plot < 0) then
        Units := AUnits; { Only one overlay - change overlay units }
  if Plot < 0 then
    for i := 0 to nPlots - 1 do
      (Plots[i] as TPGOPHERPlotLoaderConnector).ExperimentalPlot.Units := AUnits
  else
    (Plots[Plot] as TPGOPHERPlotLoaderConnector).ExperimentalPlot.Units := AUnits;
end;

procedure TPGOPHERPlotLoaderConnector.SetScale(Scale: Double);
begin
  ExperimentalPlot.Scale := Scale;
end;

function TPGOPHERPlotLoaderConnector.Another(Index: Integer): TPlotLoaderConnector;
var
  NewPlot: TExperimentalPlot;
begin
  NewPlot := TExperimentalPlot.Create(ExperimentalPlot.Owner);
  if ExperimentalPlot.Name = '1' then begin
    NewPlot.Name := IntToStr(Index+1);
    NewPlot.FDefaultName := True;
  end else
    NewPlot.Name := ExperimentalPlot.Name + '-' + IntToStr(Index+1);
  NewPlot.Comment := ExperimentalPlot.Comment + ' Channel ' + IntToStr(Index+1);
  NewPlot.Units := ExperimentalPlot.Units;
  Result := TPGOPHERPlotLoaderConnector.Create(NewPlot,StreamName);
  TPGOPHERPlotLoaderConnector(Result).FBaseName := BaseName;
end;

initialization
  TExperimentalPlot.RegisterClass(TOverlays);
end.
