unit ExternalList;

{ Use externally generated line list for plots }

interface
{$I DFS.inc}{Copyright (C) 2008 C M Western; see commandline.pas for license}

uses {$IFNDEF CONSOLE} Menus, {$ENDIF}
     SysUtils, Classes, MNode, QuantumNumber, Mixture, DRintensity, pgDOM,
     QuantumMath, MathSup, StaticField;

type

  TExternalEigenvalues = class(TEigenvalues)
  private
    FEnergies: TDoubleArray;
    Fg: array of Integer;
    FSize: Integer;
    FQuanta: array of string;
  protected
    function GetEnergy(AIndex: Integer): Double; override;
  public
    procedure FixMem;
    function Size: Integer; override;
    function Degeneracy(AIndex: Integer): Integer; override;
    function Qno(AIndex: Integer; QnoType: TQnoType):TQuantumNumber; override;
    function AddLevel(AEnergy: Double; g: Integer; EDiff: Double; const AQuanta: string):Integer;
    function DominantLabel(AIndex: Integer; QuantumNumberFormat: TQuantumNumberFormat):string; override;
    procedure SetStatisticalWeight(AValue: Integer);
  end;

  TExternalManifold = class(TManifold)
  private
    FDummyEigenvalues: array of TExternalEigenvalues;
    FHalfSpin: Integer;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function FixupProperties:Exception; override;
    function AddLevel(Energy: Double; g: Integer; J: TQuantumNumber; EDiff: Double; gDiv: Integer; const Quanta: string):Integer;
    procedure ClearCache; override;
//    function Eigenvalues(J: TQuantumNumber; Sym: Integer; MatElemTypes: TMatElemTypes;
//                         Wait: Boolean = True;
//                         PrintLevel: Integer = -10; SelState: TState = nil; IncludePerturbations: Boolean = True;
//                         RemoveSpins: Byte = 0): TEigenvalues;
    function nSym: Integer; override; { Number of different symmetries }
    function Population(AEigenvalues: TEigenvalues; i: Integer; Simulation: TSimulation; var Overflow: Exception; IncludeTspin: Integer): Double; override;
    procedure AddAllLevels(AState: TState); override;
    function HalfSpin(RemovedSpins: Byte = 0): TQuantumNumber; override;
    function FindJmin(Default: Integer; RemovedSpins: Byte = 0):Integer; override;
    function Eigenvalues(g: Integer; J: TQuantumNumber): TExternalEigenvalues; overload;
    function Eigenvalues(J: TQuantumNumber; Sym: Integer; MatElemTypes: TMatElemTypes;
                         const AExField: TStaticFields;
                         Wait: Boolean = True;
                         PrintLevel: TPrintLevel = plDefault; SelState: TState = nil; IncludePerturbations: Boolean = True;
                         RemoveSpins: Byte = 0; CreateIfNotFound: Boolean = True): TEigenvalues; overload; override;
  end;

  TExternalEntry = record
    LineNumber: Integer; { Line number in file }
    Frequency: Double; { Transition frequency }
    Intensity: Double;
    Width: Double;
    ELower:    Double; { Lower State energy (if available) }
    gL, gU: Integer;        { Upper and lower state degeneracy, if available }
    JL, JU: TQuantumNumber; { Upper and lower state J, if available }
    ManifoldL: TExternalManifold;
    ManifoldU: TExternalManifold;
    UpperQuanta, LowerQuanta: string;
    IndexU, IndexL: Integer; { Computed }
  end;

  TExternalMolecule = class(TMolecule)
  private
    FDummyManifold: TExternalManifold;
    FnEntries: Integer;
  public
    Entries: array of TExternalEntry;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function FixupProperties:Exception; override;
    function AddToSpectrum(Spectrum: TSpectrum; PrintLevel: TPrintLevel; const Excitation: TExcitation; var Problem: Exception;
                           Simulation: TSimulation;
                           J1Sel: Integer = Low(Integer); Sym1Sel: Integer = -1; J2Sel: Integer = Low(Integer); Sym2Sel: Integer = -1;
                           Manifold1Sel: TManifold = nil; Manifold2Sel: TManifold = nil;
                           DeltaJ: TQuantumNumber = -10000;
                           TransitionMomentsSel: TTransitionMoments = nil; TransitionMomentSel: TTransitionMoment = nil;
                           State1Sel: TState = nil; State2Sel: TState = nil; IncludePerturbations: Boolean = True; IncludeZero: Boolean = False):Boolean; override;
    // ? function Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean; virtual; abstract;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    function LastSym: Integer; override;
    function DummyManifold: TExternalManifold;
    function JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer; override;
    function QnoInfo(QnoType: TQnoType): TQnoInfo; override;
    function FindLine(var ALine: TLine): Integer;
    {$IFNDEF CONSOLE}
    procedure AddHandlers; override;
    procedure LoadFromFileClick(Sender: TObject);
    {$ENDIF}
    procedure AddEntry(const Entry: TExternalEntry);
    property nEntries: Integer read FnEntries;
  published
    property WScale: Double index 2 read GetParameter write SetParameter;
    property EDiff:  Double index 3 read GetParameter write SetParameter;
    property gDiv:  Integer index LMI read GetSetting write SetSetting;
  end;

implementation

uses {$IFNDEF CONSOLE} Forms, Dialogs, IPlotLoader, {$ENDIF}
      Units, Overlay, Worker, Hitran, ExpPlotLoader, PartFunc,
      Graphics, Math, TabbedLine;

var
  EmptyEigenvalues: TEigenvalues;

{ TExternalMolecule }

function TExternalMolecule.AddToSpectrum(Spectrum: TSpectrum;
  PrintLevel: TPrintLevel; const Excitation: TExcitation; var Problem: Exception; Simulation: TSimulation;
  J1Sel, Sym1Sel, J2Sel, Sym2Sel: Integer; Manifold1Sel,
  Manifold2Sel: TManifold; DeltaJ: TQuantumNumber;
  TransitionMomentsSel: TTransitionMoments;
  TransitionMomentSel: TTransitionMoment; State1Sel, State2Sel: TState;
  IncludePerturbations, IncludeZero: Boolean): Boolean;
var
  i: Integer;
  ALine, Line: TLine;
  wt: Double;
begin
  with Line do begin
    Width := 0;
  end;
  Line.Colour := Colour;
  wt := Abundance*Species.Concentration;
  for i := 0 to nEntries-1 do
    with Entries[i] do begin
      if Assigned(TransitionMomentsSel) then
        if not( ((TransitionMomentsSel.Bra = ManifoldU)
                  and (TransitionMomentsSel.Ket = ManifoldL))
               or ((TransitionMomentsSel.Bra = ManifoldL)
                  and (TransitionMomentsSel.Ket = ManifoldU))
                )
         then
        Continue;
      if Assigned(Manifold1Sel) and (ManifoldL <> Manifold1Sel) then
        Continue;
      if Assigned(Manifold2Sel) and (ManifoldU <> Manifold2Sel) then
        Continue;
      Line.Frequency := Frequency;
      Line.Strength := Intensity*wt;
      if (UpperQuanta <> '') and (LowerQuanta <> '') then
        Line.CachedName := UpperQuanta + ' - ' + LowerQuanta
      else
        Line.CachedName := UpperQuanta + LowerQuanta;
      { Use negative index values here to flag upper state }
      // Line.IndexU := -1-i; { Flag upper state index }
      // Line.IndexL := i;
      Line.IndexL := IndexL;
      Line.IndexU := IndexU;
      Line.EigenvaluesL := ManifoldL.Eigenvalues(gL,JL);
      Line.EigenvaluesU := ManifoldU.Eigenvalues(gU,JU);
      if Line.EigenvaluesU.Manifold.Colour <> clNone then
        Line.Colour := Line.EigenvaluesU.Manifold.Colour
      else if Line.EigenvaluesL.Manifold.Colour <> clNone then
        Line.Colour := Line.EigenvaluesL.Manifold.Colour
      else
        Line.Colour := Colour;
      Line.Width := Width*WScale;
      if PrintLevel >= plLineList then begin
        ALine := Line;
        if Mixture.Simulations[0].PlotUnits <> Mixture.Units then
          ConvertLine(ALine, Mixture.Units, Mixture.Simulations[0].PlotUnits);
        Spectrum.ShowLine(ALine, Log, PrintLevel, 0, False, Problem);
      end;
      Spectrum.AddLine(Line, Mixture.Simulations[0], cm1, Problem);
    end;
  Result := True;
end;

constructor TExternalMolecule.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',EDiff=1e-3';
  inherited;
end;

destructor TExternalMolecule.Destroy;
begin
  inherited;
end;

function TExternalMolecule.DummyManifold: TExternalManifold;
begin
  if not Assigned(FDummyManifold) then  begin
    FDummyManifold := TExternalManifold.Create(Self);
    FDummyManifold.Name := 'Dummy';
    FDummyManifold.Initial := True;
  end;
  Result := FDummyManifold;
end;

function TExternalMolecule.FixupProperties: Exception;
var
  i: Integer;
begin
  SetLength(Entries, nEntries); { Discard unused memory }
  if ActivePartitionFunction is TDefaultPartitionFunction then
    with TInterpolatedPartitionFunction.Create(Self) do
      Name := 'Partition Function';
  Mixture.LargestIndex := Max(Mixture.LargestIndex,nEntries*10);
  for i := 0 to nEntries-1 do
    with Entries[i] do
      if (UpperQuanta <> '') and (LowerQuanta <> '') then begin
        IndexL := ManifoldL.AddLevel(ELower, gL, JL, EDiff, gDiv, LowerQuanta);
        IndexU := ManifoldU.AddLevel(ELower+Frequency, gU, JU, EDiff, gDiv, UpperQuanta);
      end else begin
        IndexL := ManifoldL.AddLevel(ELower, gL, JL, EDiff, gDiv, '');
        IndexU := ManifoldU.AddLevel(ELower+Frequency, gU, JU, EDiff, gDiv, '');
      end;
  Result := inherited FixupProperties;
end;

function TExternalMolecule.JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer;
begin
  Result := Sym;
end;

function TExternalMolecule.LastSym: Integer;
begin
  Result := 0;
end;

function TExternalMolecule.FindLine(var ALine: TLine): Integer;
var
  i: Integer;
begin
  { This is intended for testing purposes only; given eigenvalue numbers find corresponding
    line list entry }
  for i := 0 to High(Entries) do
    with Entries[i] do
      if (ALine.IndexU = IndexU) and (ALine.IndexL = IndexL)
         and (ALine.EigenvaluesU = ManifoldU.Eigenvalues(gU,JU))
         and (ALine.EigenvaluesL = ManifoldL.Eigenvalues(gL,JL)) then begin
        Result := i;
        if (UpperQuanta <> '') and (LowerQuanta <> '') then
          ALine.CachedName := UpperQuanta + ' - ' + LowerQuanta
        else
          ALine.CachedName := UpperQuanta + LowerQuanta;
        ALine.Strength := Intensity;
        ALine.Frequency := Frequency;
        ALine.Width := Width;
        Exit;
      end;
  Result := -1;
end;

function TExternalMolecule.ReadProperties(Node: TDOMNode;
  CheckUnread: Boolean): Integer;
var
  Child: TDOMNode;
  b: string;
  P: PChar;
  nFields, Code: Integer;
  LineNo, gUt, gLt, JUt, JLt: Double;
  Entry: TExternalEntry;
  function NextField(Text:Boolean):string;
  var
    S: PChar;
  begin
    Inc(P);
    S := P;
    while not ( (S^ in [#0, #10, #13])
                or (not Text and (S^ in [',', '/']))
                or (Text and (S^ = #9))
                 ) do
      Inc(S);
    SetLength(Result, S-P);
    if S > P then
      StrMove(@Result[1], P, S-P);
    Result := Trim(Result);
    P := S;
  end;
begin
  Result := inherited ReadProperties(Node, False);
  Child := Node.FirstChild;
  if Assigned(Child) and (Child is TDOMCDATASection) then begin
    b := TDOMCDATASection(Child).Data;
    P := @b[1];
    FnEntries := 0;
    while P^ <> #0 do begin
      with Entry do begin
        Width := 0;
        SetNan(ELower);
        LineNo := -1; { Transfer variables - must be double }
        gUt := 0;
        gLt := 0;
        JUt := -0.5;
        JLt := -0.5;
        Code := TryReadTabbedLine(P, [@LineNo,@Frequency,@Intensity,@Width,@ELower,@gUt,@gLt,@JUt,@JLt], nFields);
        if nFields > 0 then begin
          LineNumber := Round(LineNo);
          gU := Round(gUt);
          gL := Round(gLt);
          JU := Round(JUt*2);
          JL := Round(JLt*2);
          Inc(P,Code-2);
          if P^ = ',' then
            ManifoldU := FindOrCreateNode(NextField(False), TExternalManifold) as TExternalManifold
          else
            ManifoldU := nil;
          if P^ = ',' then
            ManifoldL := FindOrCreateNode(NextField(False), TExternalManifold) as TExternalManifold
          else
            ManifoldL := nil;
          if not (P^ in [#0, #10, #13]) then
            UpperQuanta := NextField(True)
          else
            UpperQuanta := '';
          if not (P^ in [#0, #10, #13]) then
            LowerQuanta := NextField(True)
          else
            LowerQuanta := '';
          AddEntry(Entry);
        end;
      end;
      if Code <> 0 then begin
        case P^ of
          #10: if (P+1)^ = #13 then
                 Inc(P,2)
               else
                 Inc(P);
          #13: if (P+1)^ = #10 then
                 Inc(P,2)
               else
                 Inc(P);
        end;
      end;
    end;
    Node.RemoveChild(Child);
  end;
  if CheckUnread then
    Result := Result + nUnreadProperties(Node);
end;

function TExternalMolecule.WriteProperties(Parent: TDOMNode; AllParameters: Boolean; WithChildren: Boolean): TDOMElement;
var
  Strings: TStringList;
  i: Integer;
  s: string;
begin
  Result := inherited WriteProperties(Parent, AllParameters, WithChildren);
  Strings := TStringList.Create;
  try
    Strings.Capacity := Length(Entries)+1;
    Strings.Add('');
    for i := 0 to High(Entries) do begin
      with Entries[i] do begin
        s := IntToStr(LineNumber)
             +#9+FloatToStr(Frequency)
             +#9+FloatToStr(Intensity)
             +#9;
        if Width > 0 then
          s := s+FloatToStr(Width);
        s := s+#9;
        if not IsNan(ELower) then
          s := s+FloatToStr(ELower);
        s := s+#9;
        if gU > 0 then
          s := s+FloatToStr(gU);
        s := s+#9;
        if gL > 0 then
          s := s+FloatToStr(gL);
        s := s+#9;
        if JU >= 0 then
          s := s+QuantumNumber.QuantumNumberToStr(JU,qftJ);
        s := s+#9;
        if JL >= 0 then
          s := s+QuantumNumber.QuantumNumberToStr(JL,qftJ);
        s := s+',';
        if ManifoldU <> FDummyManifold then
          s := s+ManifoldU.Name;
        s := s+',';
        if ManifoldL <> FDummyManifold then
          s := s+ManifoldL.Name;
        s := s+'/'+UpperQuanta;
        if LowerQuanta <> '' then
          s := s+#9+LowerQuanta;
      end;
      Strings.Add(s);
    end;
    Result.AppendChild(Parent.OwnerDocument.CreateCDATASection(Strings.Text));
  finally
    Strings.Free;
  end;
end;

procedure TExternalMolecule.AddEntry(const Entry: TExternalEntry);
var
  tm: TTransitionMoments;
begin
  if High(Entries) < nEntries then
    SetLength(Entries,nEntries+10000);
  Entries[nEntries] := Entry;
  with Entries[nEntries] do begin
    if not Assigned(ManifoldL) then
      ManifoldL := DummyManifold;
    if not Assigned(ManifoldU) then
      ManifoldU := DummyManifold;
    tm := FindTransitionMoments(ManifoldU, ManifoldL);
    if not Assigned(tm) then
      TTransitionMoments.Create(Self).SetBraKetName(ManifoldU.Name,ManifoldL.Name);
  end;
  Inc(FnEntries);
  if not Modified then
    Modified := True; { Don't call every time - much too slow }
end;

function TExternalMolecule.QnoInfo(QnoType: TQnoType): TQnoInfo;
begin
  Result := inherited QnoInfo(QnoType);
  case QnoType of
    qtA: Result.Name := '#';
  end;
end;

{$IFNDEF CONSOLE}
procedure TExternalMolecule.AddHandlers;
begin
  inherited;
  // TODO: AddHandler('Load...', 'Load from file', {$IFDEF FPC}@{$ENDIF}LoadFromFileClick);
end;

procedure TExternalMolecule.LoadFromFileClick(Sender: TObject);
//var
//  Stream: TMemoryStream;
begin
{ TODO: Restore }
//  with TOpenDialog.Create(Application) do
//    try
//      Filter := 'Hitran file (*.par)|*.par|All Files|'+AllFilesMask;
//      Title := 'Load External Line List';
//      if Execute then begin
//        Stream := TMemoryStream.Create;
//        try
//          Stream.LoadFromFile(FileName);
//          ReadHitranFile(Stream, FileName);
//        finally
//          Stream.Free;
//        end;
//      end;
//    finally
//      Free;
//    end;
end;
{$ENDIF}

{ TExternalEigenvalues }

function TExternalEigenvalues.Qno(AIndex: Integer; QnoType: TQnoType): TQuantumNumber;
begin
  case QnoType of
    qtA: Result := AIndex;
    qtB: Result := 0;
  else
    Result := J;
  end;
end;

procedure TExternalEigenvalues.SetStatisticalWeight(AValue: Integer);
begin
  FStatisticalWeight := AValue;
end;

function TExternalEigenvalues.Size: Integer;
begin
  Result := FSize;
end;

function TExternalEigenvalues.AddLevel(AEnergy: Double; g: Integer; EDiff: Double; const AQuanta: string): Integer;
var
  i: Integer;
begin
  for i := 0 to FSize - 1 do
    if (AQuanta = FQuanta[i]) or (AQuanta = '') then
      if (Abs(FEnergies[i]-AEnergy)<EDiff) and (Fg[i] = g) then begin
        Result := i+1;
        Exit;
      end else if AQuanta <> '' then
        WriteLn(Log, 'Warning: Quanta equal, but no energy match in', Manifold.UniqueName, '#', i, ' ', AQuanta, ' ', FEnergies[i], ' ', AEnergy, ' ', Fg[i], ' ', g, ' ', Abs(FEnergies[i]-AEnergy));
  Inc(FSize);
  if FSize > Length(FEnergies) then begin
    if FSize < 10 then
      i := 10
    else
      i := 100;
    SetLength(FEnergies, FSize+i);
    SetLength(Fg, Length(FEnergies));
    SetLength(FQuanta, Length(FEnergies));
  end;
  Result := FSize;
  FEnergies[Result-1] := AEnergy;
  Fg[Result-1] := g;
  FQuanta[Result-1] := AQuanta;
end;

function TExternalEigenvalues.Degeneracy(AIndex: Integer): Integer;
begin
  Result := Fg[AIndex-1];
end;

function TExternalEigenvalues.DominantLabel(AIndex: Integer; QuantumNumberFormat: TQuantumNumberFormat): string;
begin
  Result := Manifold.UniqueName + ' ' + FQuanta[AIndex-1];
end;

procedure TExternalEigenvalues.FixMem;
begin
  SetLength(FEnergies, FSize);
  SetLength(Fg, Length(FEnergies));
  SetLength(FQuanta, Length(FEnergies));
end;

function TExternalEigenvalues.GetEnergy(AIndex: Integer): Double;
begin
  Result := FEnergies[AIndex-1];
end;

{ TExternalManifold }

procedure TExternalManifold.AddAllLevels(AState: TState);
var
  i, Ind: Integer;
  pop: Double;
  Overflow: Exception;
  Simulation: TSimulation;
  Mol: TExternalMolecule;
begin
  Simulation := Mixture.Simulations[0];
  Overflow := nil;
  Mol := Molecule as TExternalMolecule;
  for i := 0 to Mol.nEntries - 1 do
    if Mol.Entries[i].ManifoldL = Self then begin
      pop := Population(Eigenvalues(Mol.Entries[i].gL,Mol.Entries[i].JL), i, Simulation, Overflow, 0);
      Ind := PopulationIndex(0,0,i,True);
      Parameters[Ind] := pop;
    end;
end;

function TExternalManifold.AddLevel(Energy: Double; g: Integer; J:TQuantumNumber; EDiff: Double; gDiv: Integer; const Quanta: string): Integer;
begin
  Result := Eigenvalues(g,J).AddLevel(Energy, g, EDiff, Quanta);
end;

procedure TExternalManifold.ClearCache;
var
  i: Integer;
begin
  inherited;
  for i := 0 to High(FDummyEigenvalues) do
    FreeAndNil(FDummyEigenvalues[i]);
  FHalfSpin := 0;
end;

constructor TExternalManifold.Create(AOwner: TComponent);
begin
  inherited;
end;

destructor TExternalManifold.Destroy;
begin
  ClearCache;
  inherited;
end;

function TExternalManifold.Eigenvalues(g: Integer; J: TQuantumNumber): TExternalEigenvalues;
var
  OldSize, Index, i: Integer;
begin
  if (J < 0) and (g > 0) and ((Molecule as TExternalMolecule).gDiv > 0) then begin
    J := Round(g/(Molecule as TExternalMolecule).gDiv-1);
    FHalfSpin := SmallestProjection(J);
  end;
  Index := ((J or 1)+1) div 2;
  if Index > High(FDummyEigenvalues) then begin
    OldSize := Length(FDummyEigenvalues);
    SetLength(FDummyEigenvalues,Index+20);
    for i := OldSize to High(FDummyEigenvalues) do
      FDummyEigenvalues[i] := nil;
  end;
  if not Assigned(FDummyEigenvalues[Index]) then begin
    FDummyEigenvalues[Index] := TExternalEigenvalues.Create(Self, J, 0, ZeroField);
    { TODO: Fails for equivalent nuclei }
    FDummyEigenvalues[Index].SetStatisticalWeight((Molecule as TExternalMolecule).gDiv);
  end;
  Result := FDummyEigenvalues[Index];
end;

function TExternalManifold.Eigenvalues(J: TQuantumNumber; Sym: Integer;
  MatElemTypes: TMatElemTypes; const AExField: TStaticFields;
  Wait: Boolean; PrintLevel: TPrintLevel;
  SelState: TState; IncludePerturbations: Boolean;
  RemoveSpins: Byte; CreateIfNotFound: Boolean): TEigenvalues;
var
  Index: Integer;
begin
  Index := ((J or 1)+1) div 2;
  Result := nil;
  if Index <= High(FDummyEigenvalues) then
    Result := FDummyEigenvalues[Index];
  if not Assigned(Result) and CreateIfNotFound then
    Result := EmptyEigenvalues;
end;

function TExternalManifold.FindJmin(Default: Integer; RemovedSpins: Byte): Integer;
begin
  if (Length(FDummyEigenvalues) = 0) or Assigned(FDummyEigenvalues[0]) then
    Result := -1
  else
    Result := 0;
end;

function TExternalManifold.FixupProperties: Exception;
var
  i: Integer;
begin
  for i := 0 to High(FDummyEigenvalues) do
    if Assigned(FDummyEigenvalues[i]) then
      FDummyEigenvalues[i].FixMem;
  Result := inherited FixupProperties;
end;

function TExternalManifold.nSym: Integer;
begin
  Result := 1;
end;

function TExternalManifold.HalfSpin(RemovedSpins: Byte): TQuantumNumber;
begin
  Result := FHalfSpin;
end;

function TExternalManifold.Population(AEigenvalues: TEigenvalues; i: Integer;
  Simulation: TSimulation; var Overflow: Exception; IncludeTspin: Integer): Double;
  function bfp(T, E: Double): Double;
  begin
    Result := bf(T,E);
    if Result < 0 then begin
      if not Assigned(Overflow) then
        Overflow := EReportException.CreateFmt('Overflow in Bolzmann for line %d: E=%g',[i, E]);
      Result := 0;
    end;
  end;
var
  PIndex: Integer;
begin
  if Simulation.Temperature > 0 then
    { Normal Boltzman calculation }
    Result := bfp(Simulation.Temperature, AEigenvalues.Energy[i])
  else begin
    { Custom population }
    PIndex := PopulationIndex(0,0,i);
    if PIndex > 0 then
      Result := Parameters[PIndex]
    else
      Result := 0;
  end;
  { TODO: May be able to handle Tspin > 0 }
//    if i >= 0 then begin
//      Result := 1;
//      if i <= High((Molecule as TExternalMolecule).Entries) then
//        with (Molecule as TExternalMolecule).Entries[i] do
//          if gL > 0 then { TODO: Handle -ve T => External population }
//            Result := bfp(Simulation.Temperature,ELower);
//    end else begin
//      Result := 0;
//      if -i-1 <= High((Molecule as TExternalMolecule).Entries) then
//        with (Molecule as TExternalMolecule).Entries[-i-1] do
//          if gU > 0 then { TODO: Handle -ve T => External population }
//            Result := bfp(Simulation.Temperature,Frequency+ELower);
//    end
//  else if i >= 0 then begin
//    { For custom populations only use lower state }
//    PIndex := PopulationIndex(0,0,i);
//    if PIndex > 0 then
//      Result := Parameters[PIndex]
//    else
//      Result := 0;
//  end else
//    Result := 0;
end;

initialization
  TExternalMolecule.RegisterClass(TSpecies);
  TExternalManifold.RegisterClass(TExternalMolecule);
  EmptyEigenvalues := TEigenvalues.Create(nil,0,0,ZeroField);
finalization
  FreeAndNil(EmptyEigenvalues);
end.
