unit Fitter;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses MatrixVec, Mixture, Overlay, Obs, MathSup;

type
  OnShowContour = procedure (Spectra: TOverlays; Simulation: TColumnVector) of object;

function FitLines(var Parameters: TParameters; var Obs: array of TObservation; Precision: Integer; Mixture: TMixture):string;
function FitContour(Mixture: TMixture; Spectra: TOverlays; ShowContour: OnShowContour; var Parameters: TParameters; JustResiduals: Boolean):string;
function CalculateObs(var Observation: TObservation; out Value: Double; var RegeneratedLabel: string; RegenerateLabel:Boolean; AObsType: TObsType): Boolean;

type
  TFitter = class
  public
    nObs, nPar: Integer;
    a: TMatrix;   { d(obs)/d(par) }
    omc: TColumnVector; { normally obs - calc }
    par: TColumnVector; { Parameters - changed by this routine }
    weights: TColumnVector; { nil or the relative standard deviation of each value in obs }
    OldSD, NewSD: Double; { Previous and estimated next standard deviations }
    UnWtOldSD, UnWtNewSD: Double; { Previous and estimated next standard deviations ignoring weights }
    b: TColumnVector;  { Standard Deviations of the parameters}
    ada: TMatrix; { Correlation Matrix }
    SVD: TMatrix; SVDw: TColumnVector; { Details of Singular Value Decomposition }
    Correlations: TDoubleArrayArray; { Correlation of observation with following observations }
    constructor Create(AnPar, AnObs: Integer; MaxCorrelatedObservations: Integer = 0);
    destructor Destroy; override;
    function DoLeastSQ(var Parameters: TParameters; Precision: Integer; Mixture: TMixture):string;
  end;

implementation

uses Progress, MNode, Graphics, SysUtils, NLFit,
     SpectrumPlot, ExpPlot, Units, DRintensity,
     Worker, PGWork, Math, QuantumNumber, Constrain;

{ TFitter }

constructor TFitter.Create(AnPar, AnObs: Integer; MaxCorrelatedObservations: Integer);
begin
  nObs := AnObs;
  nPar := AnPar;
  a := TMatrix.Create(nObs, nPar);
  omc := TColumnVector.Create(nObs);
  weights := TColumnVector.Create(nObs); { Also used as workspace }
  if MaxCorrelatedObservations > 0 then
    SetLength(Correlations,nObs-1);
end;

destructor TFitter.Destroy;
begin
  a.Free;
  omc.Free;
  par.Free;
  b.Free;
  ada.Free;
  weights.Free;
  SVD.Free;
  SVDw.Free;
  inherited;
end;

function TFitter.DoLeastSQ(var Parameters: TParameters; Precision: Integer; Mixture: TMixture):string;
var
  i, j, nErrDig, nDiscard: Integer;
  r1: string;
  Sum, C: Double;
begin

  if Mixture.FitDetails and (nPar > 0) then begin
    WriteLn(Log);
    WriteLn(Log, 'Derivative Matrix');
    Write(Log,' ':6);
    for i := 0 to nPar-1 do
      Write(Log, ' ', Parameters[i].Name:13);
    WriteLn(Log);
    for i := 1 to nObs do begin
      Write(Log, i:6);
      for j := 1 to nPar do
        Write(Log, DTosMA(a[i,j],Max(14,Length(Parameters[j-1].Name)+1),-1));
      WriteLn(Log);
    end;
  end;

  { Fill in parameter info }
  par := TColumnVector.Create(nPar);
  for i := 0 to nPar-1 do
    par[i+1] := Parameters[i].Value;
  b := TColumnVector.Create(nPar);
  ada := TMatrix.Create(nPar, nPar);
  SVD := TMatrix.Create(nPar, nPar);
  SVDw := TColumnVector.Create(nPar);
  nDiscard := LeastSQ(a, omc, par, weights, OldSD, NewSD, UnWtOldSD, UnWtNewSD, b, ada, SVDw, SVD, Mixture.SVDThresh, Mixture.FitDetails, Mixture.ScaleChanges, Correlations);
  WriteLn(Log);
  WriteLn(Log, nObs, ' Observations,  ', nPar, ' Parameters');
  if nPar > 0 then
    Result := 'Initial '
  else
    Result := '';
  Result := Result + 'Average Error: ' + FloatToStr(OldSD);
  if OldSD <> UnWtOldSD then
    Result := Result + '  (Unweighted: ' +  FloatToStr(UnWtOldSD) + ')';
  WriteLn(Log, Result);
  { Was 10:Precision }
  if nPar > 0 then begin
    r1 := 'Predicted New Error:   ' + FloatToStr(NewSD);
    if NewSD <> UnWtNewSD then
      r1 := r1 + '  (Unweighted: ' +  FloatToStr(UnWtNewSD) + ')';
    WriteLn(Log, r1);
    Result := Result + ' - ' + r1;
    if nDiscard > 0 then
      WriteLn(Log, '**** ', nDiscard, ' combinations of parameters not floated ****');

    { For constants (such as the observed spectrum scale factor) where the residual is
      inversly proportional to the constant, the estimated change is particularly poor if
      the change is large. Adjust the change by converting to a change in the inverse,
      and then inverting. [The ratio of derivatives of a parameter p and it's inverse is -1/p^2
      so a change d in p corresponds to a change -d/p^2 in 1/p so the new p is
      1/(1/p-d/p^2) = p/(1-d/p) }
    for i := 0 to nPar - 1 do
      if Parameters[i].Inverse then begin
        WriteLn(Log, 'Linear New Value for ',Parameters[i].Name,': ',par[i+1]);
        C := Parameters[i].Value/(1-(par[i+1]-Parameters[i].Value)/Parameters[i].Value);
        { Check that parameter does not change sign; fall back to original method if it does }
        if C * Parameters[i].Value > 0 then
          par[i+1] := C;
      end;

    WriteLn(Log, 'Parameters:');
    WriteLn(Log, '  # Old                 New                 Std Dev   Change/Std Sens    Summary              Name');
    for i := 0 to nPar-1 do begin
      { Calculate parameter sensitivity }
      Sum := 0;
      for j := 1 to nObs do
        { Note - don't need Sqr(a[j,i+1]/weights[j]) as a has been adjusted to give observations with standard deviation 1 }
        Sum := Sum + Sqr(a[j,i+1]);
      { and use it to determine the number of uncertainty digits to show }
      Sum := 0.1*NewSD/nPar/Sqrt(Sum/nObs);
      nErrDig := Max(Trunc(Log10(b[i+1]/Sum))+1,2);
      WriteLn(Log, Format('%3d%s%s%s%10.4f %s %-20s %s', [i+1, DtosMA(Parameters[i].Value,20,-1),
                   DtosMA(par[i+1],20,-1), DtosMA(b[i+1],10,-1), (par[i+1]-Parameters[i].Value)/b[i+1],
                   DtosMA(Sum,8,-1),
                   DtosErr(par[i+1],b[i+1],nErrDig), Parameters[i].Name]));
      Parameters[i].Value := par[i+1];
      Parameters[i].StdDev := b[i+1];
    end;
    WriteConstraints(Log);
    if nPar > 1 then begin
      WriteLn(Log);
      WriteLn(Log, 'Correlation Matrix');
      for i := 1 to nPar do
        Write(Log, i:8);
      WriteLn(Log);
      for i := 1 to nPar do begin
        Write(Log, i:3);
        for j := 1 to i do
          Write(Log, ada[i,j]:8:3);
        WriteLn(Log);
      end;

      if Mixture.FitDetails or (nDiscard > 0) then begin
        WriteLn(Log);
        WriteLn(Log, 'Singular Value Decomposition');
        Write(Log,' ':17);
        for i := 0 to nPar-1 do
          Write(Log, ' ', Parameters[i].Name:13);
        WriteLn(Log);
        for i := 1 to nPar do begin
          Write(Log, i:3, DTosMA(SVDW[i],14,-1));
          for j := 1 to nPar do
            Write(Log, DTosMA(SVD[i,j],Max(14,Length(Parameters[j-1].Name)+1),-1));
          WriteLn(Log);
        end;
      end;
    end;
    { Copy Back Parameters }
    Mixture.Correlations.FitDetails.Parameters := Copy(Parameters);
    with Mixture.Correlations do begin
      SetLength(FitDetails.Correlation, ada.RowCount);
      for i := 0 to ada.RowCount - 1 do begin
        SetLength(FitDetails.Correlation[i],i);
        for j := 0 to i - 1 do
          FitDetails.Correlation[i][j] := ada[i+1,j+1];
      end;
    end;
    for i := 0 to nPar-1 do
      with Parameters[i] do begin
        MNode.SetParameter(Index, par[i+1]);
        MNode.StdDev[Index] := b[i+1];
      end;
    ApplyConstraints;
  end;
end;

function FitContour(Mixture: TMixture; Spectra: TOverlays; ShowContour: OnShowContour; var Parameters: TParameters; JustResiduals: Boolean):string;
var
  f: TFitter;
  nObs: Integer;
  i, j, l: Integer;
  Simulation: TFitPlot;
  vo, vn: Double;
  ConditionSimulation: TSimulation;
  procedure CalculateObs(v: TColumnVector);
  var
    i, j, l: Integer;
    Problem: Exception;
    Val: Double;
  begin
    { Copy in required frequencies }
    l := 0;
    for i := 0 to Spectra.nPlots-1 do
     if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        if not Hide then
          for j := StartPoint to EndPoint do begin
            Simulation.Frequency[l] := Convert(AdjustFrequency(Frequency[j]),Units,ConditionSimulation.PlotUnits);
            Inc(l);
          end;
    Simulation.MakeFromSimulation(ConditionSimulation, not Mixture.NoLineList, Problem);
    if Assigned(Problem) then
      if Problem is EWarningException then
        Problem.Free
      else
        raise Problem;
    l := 0; { Copy simulation, doing vertical scaling as we go }
    { Logically we should be comparing to the values in the observed TExperimentalPlot.
      However the calculated values would not then depend on the scaling parameters in
      the observed spectrum, so we apply the reverse transform to the calculated values }
    for i := 0 to Spectra.nPlots-1 do
     if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        if not Hide then
          for j := StartPoint to EndPoint do begin
            Val := ConvertIntensity(Simulation.Plot[l],Simulation.Frequency[l],ConditionSimulation.PlotUnits,ConditionSimulation.IntensityUnits,Units,IntensityUnits);
            v[l+1] := Val/Scale + Baseline; { Inverse transform }
            Inc(l);
          end;
  end;
begin
  ConditionSimulation := Mixture.Simulations[0];
  nObs := 0;
  { Find total number of observed points }
  for i := 0 to Spectra.nPlots-1 do
    if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        if not Hide then
          Inc(nObs, EndPoint-StartPoint+1);
  f := TFitter.Create(Length(Parameters), nObs);
  Simulation := nil;
  { Because of inverse scaling used in CalculateObs (see comment)
    resdiduals are inversely proportional to spectrum scale factors,
    so flag as such }
  for i := 0 to High(Parameters) do
    with Parameters[i] do
      if (MNode is TExperimentalPlot) and (Index > 0) and (MNode.ParameterNames[Index] = 'Scale') then
        Inverse := True;
  PushProgress('Derivatives', High(Parameters));
  try
    Simulation := TFitPlot.Create(ConditionSimulation.PlotUnits,IsPopDist(ConditionSimulation.IntensityUnits));
    Simulation.ndf := nObs;
    CalculateObs(f.omc);
    if Assigned(ShowContour) then
      ShowContour(Spectra, f.omc);
    if JustResiduals then Exit;
    {  Calculate Derivatives wrt each parameter }
    { TODO: Optimize for simple derivatives }
    { Note that f.weights is only used as workspace, and all weights are taken as 1 }
    for i := 0 to High(Parameters) do begin
      SetProgress(i);
      vo := Parameters[i].Value;
      if Parameters[i].Increment <= 0 then begin
        vn := vo * 1.0001;
        if vn = 0 then
          vn := 0.001; { Random Guess }
      end else
        vn := vo + Parameters[i].Increment;
      Parameters[i].MNode.SetParameter(Parameters[i].Index, vn);
      ApplyConstraints;
      try
        Parameters[i].MNode.CheckProperties;
        CalculateObs(f.weights);
      finally
        Parameters[i].MNode.SetParameter(Parameters[i].Index, vo);
        ApplyConstraints;
        Parameters[i].MNode.CheckProperties;
      end;
      for j := 1 to nObs do
        f.a[j,i+1] := (f.omc[j] - f.weights[j])/(vn-vo);
    end;
    FreeAndNil(f.Weights);
    l := 0; { Finally work out observed - calculated }
    for i := 0 to Spectra.nPlots-1 do
     if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        if not Hide then
          for j := StartPoint to EndPoint do begin
            // Note Baseline and Scale applied to calculated, not observed
            f.omc[l+1] := f.omc[l+1] - Intensity[j];
            Inc(l);
          end;
    Spectra.ClearStdDev;
    Result := f.DoLeastSQ(Parameters, Mixture.Precision, Mixture);
  finally
    PopProgress;
    f.Free;
    Simulation.Free;
  end;
end;

function CalculateObs(var Observation: TObservation; out Value: Double; var RegeneratedLabel: string; RegenerateLabel:Boolean; AObsType: TObsType): Boolean;
var
  EigL,EigU: TEigenvalues;
  LineList: TLineList;
  i: Integer;
  Mixture: TMixture;
  Molecule: TMolecule;
  TransitionMoments: TTransitionMoments;
  Intensity: Double;
  Problem: Exception;
  CalcLine: TLine;
  ConditionSimulation: TSimulation;
begin
  Result := True;
  if RegenerateLabel then
    RegeneratedLabel := '';
  with Observation do begin
    if AObsType = obParameter then
      Value := ManifoldU.Items[IndexL].Parameters[IndexU]
    else if AObsType <> obEUpper then begin
      AIndexL := FindEigenvaluesAndIndex(EigL, ManifoldL, JL, SymL, IndexL, Name, True, LineNumber, FileName, RemoveSpins, QuantumNumberFormat, ExField);
      if AIndexL > 0 then begin
        Value := EigL.E[AIndexL];
        if AObsType <> obELower then
          Value := -Value;
        if Colour = clNone then
          Colour := EigL.DominantBasis[AIndexL].State.FindColour;
        if RegenerateLabel then
          RegeneratedLabel := EigL.DominantBasis[AIndexL].StateLabel(ManifoldL.Mixture.QuantumNumberFormat);
      end else
        Result := False;
    end else
      Value := 0;
    if not (AObsType in [obELower, obParameter]) then begin
      AIndexU := FindEigenvaluesAndIndex(EigU, ManifoldU, JU, SymU, IndexU, Name, False, LineNumber, FileName, RemoveSpins, QuantumNumberFormat, ExField);
      if AIndexU > 0 then begin
        Value := Value + EigU.E[AIndexU];
        if Colour = clNone then
          Colour := EigU.DominantBasis[AIndexU].State.FindColour;
        if RegenerateLabel then begin
          if AObsType <> obEUpper then
            RegeneratedLabel := ' - ' + RegeneratedLabel;
          RegeneratedLabel := EigU.DominantBasis[AIndexU].StateLabel(ManifoldU.Mixture.QuantumNumberFormat) + RegeneratedLabel;
          if AObsType <> obEUpper then
            RegeneratedLabel := ' : ' + EigU.DominantBasis[AIndexU].TxName(EigU.DominantBasis[AIndexU],EigL.DominantBasis[AIndexL]) + ' : ' + RegeneratedLabel;
        end;
      end else
        Result := False;
    end;
    if RegenerateLabel and (AObsType = obLine) and (ManifoldU.Initial or ManifoldL.Initial) then begin
      { Consider regenerating intensity calculation }
      Molecule := ManifoldU.Molecule;
      TransitionMoments := Molecule.FindTransitionMoments(ManifoldU, ManifoldL);
      if Assigned(TransitionMoments) and TransitionMoments.Allowed(JU, SymU, JL, SymL) then begin
        TransitionMoments.MakeLine(CalcLine, EigL, EigU, AIndexL, AIndexU, ManifoldL.Mixture.Simulations[0]);
        Problem := nil;
        Mixture := ManifoldL.Mixture;
        Intensity := Mixture.Simulations[0].LineIntensity(CalcLine, Mixture.Units, Problem);
        if not Assigned(Problem) then
          RegeneratedLabel := FloatToStrF(Intensity, ffExponent, 5, 2) + RegeneratedLabel;
      end;
    end;
    if Colour = clNone then
      Colour := clBlack;
    if Result and (AObsType = obIntensity) then begin
      { Slightly inefficient - calculate all lines between these pairs of Manifold, J and Sym.
        Could be optimized by sorting observations by Manifold, J and sym }
      Mixture := ManifoldL.Mixture;
      ConditionSimulation := Mixture.Simulations[0];
      LineList := TLineList.Create(ConditionSimulation.PlotUnits,IsPopDist(ConditionSimulation.IntensityUnits));
      { TODO: IsPopDist(Mixture.IntensityUnits) may need treating as a special case }
      try
        Problem := nil;
        ManifoldU.Molecule.AddToSpectrum(LineList, Mixture.PrintLevel, RandomExcitation, Problem, ConditionSimulation,
                                         JL, ManifoldL.Molecule.JDeAdjustedSym(JL,SymL),
                                         JU, ManifoldU.Molecule.JDeAdjustedSym(JU,SymU), ManifoldL, ManifoldU,
                                         -10000, nil, nil, nil, nil, true, true);
        for i := 0 to LineList.nLines-1 do
          with LineList.Line[i] do
            if ((EigenvaluesU.Manifold = ManifoldU) and (IndexU = AindexU) and (IndexL = AindexL) and (EigenvaluesU.J = JU) and (EigenvaluesU.Sym = SymU) ) { TODO: -cField: Sym test fails with Field }
               or ((EigenvaluesU.Manifold = ManifoldL) and (IndexL = AindexU) and (IndexU = AindexL) and (EigenvaluesL.J = JU) and (EigenvaluesL.Sym = SymU)) then begin
              Value := ConditionSimulation.LineIntensity(LineList.Line[i], LineList.Units, Problem);
              if Assigned(Problem) then begin
                Result := False;
                WriteLn(Log, Problem.Message);
                Problem.Free;
              end;
              Exit;
            end;
        WriteLn(Log, 'Transition not found on line number ', LineNumber, ' of ', FileName, ':', Name);
        for i := 0 to LineList.nLines-1 do
          LineList.ShowLine(i, Log, plLineList, 0, False, Problem);
        Problem.Free;
        Result := False;
      finally
        LineList.Free;
      end;
    end;
  end;
end;

function FitLines(var Parameters: TParameters; var Obs: array of TObservation; Precision: Integer; Mixture: TMixture):string;
var
  f: TFitter;
  nObs: Integer;
  i, j: Integer;
  vo, vn: Double;
  { For determining printout format }
  LargestMolecule, MolName, LargestLower, LargestUpper: string;  { For Determining field lengths }
  SymWidth: Integer;
  MoleculeChanged, LowerChanged, UpperChanged: Boolean; { For Determining if field needs printing }
  RLab: string;
  function CalculateAllObs(v: TColumnVector): Boolean;
  var
    i: Integer;
    val: Double;
    Worker: TPgopherWorker;
  begin
    Result := True;
    if nWorkers > 0 then begin
      { Parallel algorithm }
      for i := 0 to nObs-1 do begin
        Worker := FindWorkerWait as TPgopherWorker;
        if Assigned(Worker) and Result then
          Worker.CalculateObs(@Obs[i], v.Item[i+1], @Result)
        else begin
          { Something has failed }
          if Assigned(Worker) then
            Worker.Skip;
          Break;
        end;
      end;
      WaitNoBusyWorkers;
    end else { Serial algorithm }
      for i := 0 to nObs-1 do begin
        Result := Result and CalculateObs(Obs[i], val, RLab, False, Obs[i].ObsType);
        v[i+1] := val;
      end;
  end;
var
  JHeader: string;
  HasEField, HasBField: Boolean;
  MaxCorrelation: Integer;
begin
  nObs := Length(Obs);
  if nObs <= 0 then
    raise EReportException.Create('No observations');
  if nObs < Length(Parameters) then
    raise EReportException.CreateFmt('More parameters (%d) than observations (%d)', [Length(Parameters), nObs]);
  { Work out widths of fields for printout }
  MoleculeChanged := False;
  LowerChanged := False;
  UpperChanged := False;
  LargestLower := '';
  LargestUpper := '';
  LargestMolecule := '';
  SymWidth := 0;
  HasEField := False;
  HasBField := False;
  MaxCorrelation := 0;
  for i := 0 to nObs-1 do
    with Obs[i] do begin
      if ObsType <> obEUpper then begin
        MolName := ManifoldL.Molecule.Name;
        SymWidth := Max(SymWidth, ManifoldL.Molecule.SymLength);
      end else begin
        MolName := ManifoldU.Molecule.Name;
        SymWidth := Max(SymWidth, ManifoldU.Molecule.SymLength);
      end;
      if LargestMolecule <> MolName then begin
        if Length(LargestMolecule) > 0 then
          MoleculeChanged := True;
        if Length(LargestMolecule) < Length(MolName) then
          LargestMolecule := MolName;
      end;
      if (ObsType <> obEUpper) and (LargestLower <> ManifoldL.Name) then begin
        if Length(LargestLower) > 0 then
          LowerChanged := True;
        if Length(LargestLower) < Length(ManifoldL.Name) then
          LargestLower := ManifoldL.Name
      end;
      if (ObsType <> obELower) and (LargestUpper <> ManifoldU.Name) then begin
        if Length(LargestUpper) > 0 then
          UpperChanged := True;
        if Length(LargestUpper) < Length(ManifoldU.Name) then
          LargestUpper := ManifoldU.Name
      end;
      HasEField := HasEField or (ExField.E <> 0);
      HasBField := HasBField or (ExField.B <> 0);
      MaxCorrelation := Max(MaxCorrelation, Length(Correlation));
    end;

  f := TFitter.Create(Length(Parameters), nObs, MaxCorrelation);
  PushProgress('Derivatives', High(Parameters));
  try
    { Determine obs - calc }
    if not CalculateAllObs(f.omc) then
      raise EReportException.Create('Transitions not found - see log for details');
    for i := 0 to nObs-1 do
      with Obs[i] do begin
        Calculated := f.omc[i+1];
        Flags := Flags + [ofCalculatedSet];
        if Length(Correlation) > 0 then
          f.Correlations[i] := Correlation;
      end;
    {  Calculate Derivatives wrt each parameter }
    for i := 0 to High(Parameters) do begin
      SetProgress(i);
      vo := Parameters[i].Value;
      if Parameters[i].Increment <= 0 then begin
        vn := vo * 1.0001;
        if vn = 0 then
          vn := 0.001; { Random Guess }
      end else
        vn := vo + Parameters[i].Increment;
      Parameters[i].MNode.SetParameter(Parameters[i].Index, vn);
      ApplyConstraints;
      try
        Parameters[i].MNode.CheckProperties;
        if not CalculateAllObs(f.weights) then
          raise EReportException.Create('Transitions not found - see log for details');
      finally
        Parameters[i].MNode.SetParameter(Parameters[i].Index, vo);
        ApplyConstraints;
        Parameters[i].MNode.CheckProperties;
      end;
      for j := 1 to nObs do
        f.a[j,i+1] := (f.omc[j] - f.weights[j])/(vn-vo)
    end;
    { Write out headers }
    WriteLn(Log);
    Write(Log, 'Residuals before fit');
    if MoleculeChanged then begin
      if Length(LargestMolecule) < 3 then
        LargestMolecule := 'Mol';
    end else begin
      Write(Log, ' for ', LargestMolecule);
      LargestMolecule := '';
    end;
    if not (LowerChanged or UpperChanged) then
      if (LargestLower = '') or (LargestUpper = LargestLower) then
        Write(Log, ' for the ', LargestUpper, ' manifold')
      else if LargestUpper = '' then
        Write(Log, ' for the ', LargestLower, ' manifold')
      else
        Write(Log, ' for the ', LargestUpper, ' - ', LargestLower, ' Transition');
    WriteLn(Log);
    if Length(LargestMolecule) > 0 then
      Write(Log, Copy('Molecule', 1, Length(LargestMolecule)), ' ');
    if UpperChanged then
      Write(Log, ' ':Length(LargestUpper)+1);
    Inc(SymWidth);
    if Mixture.QuantumNumberFormat = qftJ then
      JHeader := '  J'
    else
      JHeader := 'J';
    Write(Log, '  ', JHeader, '''', 'S''':SymWidth, ' #''');
    if LowerChanged then
      Write(Log, ' ':Length(LargestLower)+1);
    Write(Log, '  ', JHeader, '"', 'S"':SymWidth, ' #"');
    if HasEField then
      Write(Log, ' E        ');
    if HasBField then
      Write(Log, '  B       ');
    Write(Log, '     Observed    Calculated      Obs-Calc  StdDev');
    for i := 1 to MaxCorrelation do
      Write(Log, ' Correl');
    WriteLn(Log);
//    if ofRegenerateLabel in Flags then
//      Write(Log, ' (O-C)/Wt');
    { Work out obs-calc, weights }
    for i := 0 to nObs-1 do begin
      with Obs[i] do begin
        if Length(LargestMolecule) > 0 then
          Write(Log, ManifoldU.Molecule.Name:Length(LargestMolecule), ' ');
        if ObsType = obParameter then begin
          Write(Log, ManifoldL.Name, ' ',
                     ManifoldL.Items[IndexL].Name, ' ',
                     ManifoldL.Items[IndexL].ParameterNames[IndexU], ' =',
                     ' ':(Length(LargestUpper)+Length(LargestLower)+15+SymWidth*2-Length(ManifoldL.Name)-Length(ManifoldL.Items[IndexL].Name)-Length(ManifoldL.Items[IndexL].ParameterNames[IndexU])-6));
        end else begin
          if ObsType <> obELower then begin
            if UpperChanged then
              Write(Log, ManifoldU.Name:Length(LargestUpper), ' ');
            Write(Log, ManifoldU.Mixture.QuantumNumberToStr(JU):2+Length(JHeader),
                       ManifoldU.Molecule.JAdjustedSymName(JU,SymU):SymWidth,
                       AIndexU:3, ' ')
          end else
            Write(Log, ' ':Length(LargestUpper)+6+SymWidth+Length(JHeader));
          if ObsType <> obEUpper then begin
            if LowerChanged then
              Write(Log, ManifoldL.Name:Length(LargestLower), ' ');
              Write(Log, ManifoldL.Mixture.QuantumNumberToStr(JL):2+Length(JHeader),
                         ManifoldL.Molecule.JAdjustedSymName(JL,SymL):SymWidth,
                         AIndexL:3)
          end else
            Write(Log, ' ':Length(LargestLower)+6+SymWidth+Length(JHeader))
        end;
        if HasEField then
          Write(Log, DtosMA(ExField.E,10,-1));
        if HasBField then
          Write(Log, DtosMA(ExField.B,10,5));
        if ObsType = obIntensity then
          Write(Log, Frequency:14, f.omc[i+1]:14,
                     Frequency-f.omc[i+1]:14,
                     DtosMA(Weight,8,4))
        else
          Write(Log, AtLeast(Frequency,14,Precision), AtLeast(f.omc[i+1],14,Precision),
                     AtLeast(Frequency-f.omc[i+1],14,Precision),
                     DtosMA(Weight,8,-1));
        for j := 1 to MaxCorrelation do
          if (i < nObs-1) and (j <= Length(f.Correlations[i])) and (f.Correlations[i][j-1] <> 0) then
            Write(Log, DtosMA(f.Correlations[i][j-1],7,4))
          else
            Write(Log, ' ':7);
        if ofRegenerateLabel in Flags then begin
          CalculateObs(Obs[i], vo, RLab, True, Obs[i].ObsType);
          Write(Log, ' ', RLab, ' :':(LongestObsLabel+2-Length(Rlab)));
        end else begin
          Write(Log, ' ', Name, ' :':(LongestObsLabel+2-Length(Name)));
        end;
        Writeln(Log, LineNumber:4, ':', FileName);
      end;
      f.omc[i+1] := f.omc[i+1] - Obs[i].Frequency;
      f.weights[i+1] := Obs[i].Weight;
    end;
    Result := f.DoLeastSQ(Parameters, Precision, Mixture);
  finally
    f.Free;
    PopProgress;
  end;
end;


end.
