unit GraphRange;

interface

uses GraphM;

type
  TPlotRange = record
  { TODO: May want to add x range in for completeness, though not currently used }
   y1, y2: Double; { Fraction of current map used }
   yi, yx: Double; { Corresponding plot range }
   Tag: Integer;
  end;
  TPlotRanges = array of TPlotRange;

procedure AddPlotRange(var PlotRanges: TPlotRanges; PlotRange: TPlotRange);
function PlotRange(G: TGraphM; yu1, yu2, ya1, ya2: Double; ATag: Integer = 0):TPlotRange;
procedure MapFromPlotRange(G: TGraphM; PlotRange: TPlotRange);
procedure AxisFromPlotRange(PlotRange: TPlotRange; G: TGraphM; Xpos: Double; ALabel: string);
function FractionToValue(Frac: Double; G: TGraphM; const PlotRange: TPlotRange):Double;
function FractionToY(Frac: Double; G: TGraphM):Double;
function YFraction(YValue: Double; G:TgraphM):Double;
function RangeOverlaps(l, u: Double; const PlotRange: TPlotRange):Boolean;
function IsNullPlotRange(const PlotRange: TPlotRange):Boolean;
{$IFNDEF CONSOLE}
function FindYRange(var YValue: Double; G:TGraphM; const PlotRanges: TPlotRanges; TagMin: Integer = Low(Integer); TagMax: Integer = High(Integer)): Integer;
{$ENDIF}

const
  NullPlotRange: TPlotRange = (y1:0; y2:0; yi:0; yx:0; Tag:0);

implementation

uses GraphI, Math, MathSup;

function RangeOverlaps(l, u: Double; const PlotRange: TPlotRange):Boolean;
var
  ORange: Double;
begin
  InOrder(l,u);
  with PlotRange do
    if SameValue(l,u) then
      Result := (l >= y1) and (u <= y2)
    else begin
      ORange := Min(y2,u)-Max(y1,l); { Range in common }
      Result := (ORange/(u-l) > 0.3)    { At least 30% of given range in this plot }
                or (ORange/(y2-y1) > 0.4); { or At least 40% of plot covered by selected area }
    end;
end;

procedure AxisFromPlotRange(PlotRange: TPlotRange; G: TGraphM; Xpos: Double; ALabel: string);
begin
  with PlotRange do
    G.ScaledAxis(yi, yx, y1*(G.Ymax-G.Ymin)+G.Ymin, y2*(G.Ymax-G.Ymin)+G.Ymin, Xpos, False, ALabel);
end;

function YFraction(YValue: Double; G:TgraphM):Double;
begin
  if G.Ymax <> G.Ymin then
    Result := (YValue-G.Ymin)/(G.Ymax-G.Ymin)
  else
    Result := 0;
end;

function FractionToY(Frac: Double; G: TGraphM):Double;
begin
  with G do
    if Ymax <> Ymin then
     Result := Frac*(G.Ymax-G.Ymin) + G.Ymin
    else
      Result := Ymin;
end;

function FractionToValue(Frac: Double; G: TGraphM; const PlotRange: TPlotRange):Double;
begin
  with PlotRange do
    if y2 <> y1 then
      Result := (Frac-y1)/(y2-y1)*(yx-yi) + yi
    else
      Result := yi { Arbitrary, but avoid divide by zero }
end;

function PlotRange(G: TGraphM; yu1, yu2, ya1, ya2: Double; ATag: Integer = 0):TPlotRange;
begin
  with Result do begin
    { The actual fraction of the plot area used }
    y1 := YFraction(yu1, G);
    y2 := YFraction(yu2, G);
    { The corresponding mathematical plot range }
    if y2 > y1 then begin
      yi := ya1;
      yx := ya2;
    end else begin
      { Out of order }
      InOrder(y1, y2);
      yi := ya2;
      yx := ya1;
    end;
    Tag := ATag;
  end;
end;

procedure MapFromPlotRange(G: TGraphM; PlotRange: TPlotRange);
var
  xu, xl, yu, yl:Double;
begin
  G.MapArg(xl, xu, yl, yu);
  G.Map(xl, xu, FractionToValue(0,G,PlotRange), FractionToValue(1,G,PlotRange));
end;

procedure AddPlotRange(var PlotRanges: TPlotRanges; PlotRange: TPlotRange);
begin
  SetLength(PlotRanges, Length(PlotRanges)+1);
  PlotRanges[High(PlotRanges)] := PlotRange;
end;

function IsNullPlotRange(const PlotRange: TPlotRange):Boolean;
begin
  with PlotRange do
    Result := (Tag = 0) and (y1 = 0) and (y2 = 0);
end;

{$IFNDEF CONSOLE}
function FindYRange(var YValue: Double; G:TGraphM; const PlotRanges: TPlotRanges; TagMin: Integer; TagMax: Integer): Integer;
var
  i: Integer;
  dist, distMin: Double;
begin
  distMin := YFraction(YValue+G.PixelSize(False)*SmallY, G);
  YValue := YFraction(YValue, G);
  { Keep track of near misses, but must be reasonably close }
  distMin := Abs(YValue-distMin);
  Result := -Max(Length(PlotRanges),1);
  for i := High(PlotRanges) downto 0 do
    with PlotRanges[i] do
      if (Tag < TagMin) or (Tag > TagMax)  then
        Continue
      else if (YValue >= y1) and (YValue <= y2) then begin
        if distMin > 0 then
          Result := i { First exact hit }
        else
          Result := -Abs(Result); { Subsequent exact hit }
        distMin := 0;
      end else begin
        { keep track of closest range }
        dist := Min(Abs(YValue-y1), Abs(YValue-y2));
        if dist <= distMin then begin
          distMin := dist;
          if Result = -Length(PlotRanges) then
            Result := i { First near hit }
          else
            Result := -Abs(Result); { Subsequent near hit }
        end;
      end;
  if Result > -Length(PlotRanges) then
    YValue := FractionToValue(YValue, G, PlotRanges[Abs(Result)]);
end;
{$ENDIF}

end.
