unit ImportCALPGM;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}

{ Import data file from Herb Pickett's CALPGM program suite }
interface

uses TextStrm, Mixture;

function ImportCALPGMFile(S: TTextStream): TMixture;

implementation

uses {$IFNDEF CONSOLE} Dialogs, {$ENDIF}
     Classes, SysUtils, HundA, OAsymT, SymT, Units, MNode, MathSup,
     QuantumNumber, Worker, Math, Abelian, MatrixVec, pgDOM;

type

  TCalOptionInfo = record
    CHR: Char;
    SPIND, NVIB, KNMIN, KNMAX, IXX, IAX, WTPL, WTMN, VSYM, EWT, DIAG: Integer;
  end;

  TCALPGMReader = class
  public
    Title, Comment: string;
    OI: TCalOptionInfo;
    OIs: array of TCalOptionInfo;
    { Parameter details }
    IDPAR, FF, I2, I1, NS, TYP, KSQ, NSQ, V2, V1: Integer;
    PAR, ERRPAR: Double;
    ID: Integer; { IDPAR less vibrational number }
    nPar: Integer; { Maximum number of parameters left to read }
    S: TTextStream;
    Mixture: TMixture;
    constructor Create(Stream: TTextStream);
    function ReadOptions(FirstLine: Boolean): TCalOptionInfo;
    function ReadParameterLine(out LineOK:Boolean): Boolean;
  end;

  { TMoleculeReader }

  TMoleculeReader = class
  public
    R: TCALPGMReader;
    nProblems: Integer;
    ElectronSpin: TQuantumNumber;
    ElectronSpinIndex: Integer;
    NuclearSpins: array of TQuantumNumber;
    SpinIndex: array of Integer;
    StateMap: array of Integer;
    Deleted: array of Boolean;
    Parameters: TParameters;
    nParameters: Integer;
    DefaultElectronSpin: Integer;
    constructor Create(AReader: TCALPGMReader);
    procedure Read; virtual;
    procedure ShowProblem;
    procedure SetValues(MNode: TMNode; Index: Integer; Scale: Double; OtherPar: string = '');

    function AddState(Manifold: TManifold; Index: Integer):TState; virtual; abstract;
    function ConvertParameter(State: TState; out Scale: Double; out MNode: TMNode):Integer; virtual; abstract;
    function SParameter(IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; virtual; abstract;
    function ISParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; virtual; abstract;
    function IParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; virtual; abstract;
    function IIParameter(INuc1, INuc2: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; virtual; abstract;
    function ConvertPerturbation(State1, State2: TState; out Scale: Double; out MNode: TMNode):Integer; virtual; abstract;
    procedure AddSpins;
    function ConvertParametersWithSpin(State: TState; out Scale: Double; out MNode: TMNode):Integer;
  end;

  { TAsymmetricMoleculeReader }

  TAsymmetricMoleculeReader = class(TMoleculeReader)
  public
    Manifold: TAsymmetricManifold;
    Molecule: TAsymmetricMolecule;
    RepKnown: Boolean;
    procedure Read; override;
    function AddState(AManifold: TManifold; Index: Integer):TState; override;
    function ConvertParameter(State: TState; out Scale: Double; out MNode: TMNode):Integer; override;
    function SParameter(IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function ISParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function IParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function IIParameter(INuc1, INuc2: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function ConvertPerturbation(State1, State2: TState; out Scale: Double; out MNode: TMNode):Integer; override;
  end;

  { TSymmetricMoleculeReader }

  TSymmetricMoleculeReader = class(TMoleculeReader)
  public
    Manifold: TSymmetricManifold;
    Molecule: TSymmetricMolecule;
    LastEState: Integer;
    ZetaEtaMsg: Boolean;
    procedure Read; override;
    function AddState(AManifold: TManifold; Index: Integer):TState; override;
    function ConvertParameter(State: TState; out Scale: Double; out MNode: TMNode):Integer; override;
    function SParameter(IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function ISParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function IParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function IIParameter(INuc1, INuc2: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function ConvertPerturbation(State1, State2: TState; out Scale: Double; out MNode: TMNode):Integer; override;
  end;

  { TLinearMoleculeReader }

  TLinearMoleculeReader = class(TMoleculeReader)
  public
    Manifold: TLinearManifold;
    Molecule: TLinearMolecule;
    procedure Read; override;
    function AddState(AManifold: TManifold; Index: Integer):TState; override;
    function ConvertParameter(State: TState; out Scale: Double; out MNode: TMNode):Integer; override;
    function SParameter(IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function ISParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function IParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function IIParameter(INuc1, INuc2: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function ConvertPerturbation(State1, State2: TState; out Scale: Double; out MNode: TMNode):Integer; override;
  end;

{ TMoleculeReader }

constructor TMoleculeReader.Create(AReader: TCALPGMReader);
begin
  R := AReader;
  nProblems := 0;
  DefaultElectronSpin := 1;
end;

procedure TMoleculeReader.ShowProblem;
begin
  Inc(nProblems);
  WriteLn(Log, 'Problem with parameter on line ', R.S.Line-1, ': ',
           R.IDPAR, ' ', R.PAR, ' ', R.ERRPAR, ' ', R.Comment);
end;

procedure TMoleculeReader.SetValues(MNode: TMNode; Index: Integer; Scale: Double; OtherPar: string);
begin
  if Index < 0 then
    ShowProblem
  else begin
    if (Abs(R.ERRPAR) > 1e-10) and (Abs(R.ERRPAR) < 1e17) then begin
      if (MNode.Parameters[Index] <> 0) then begin
        ShowProblem;
        WriteLn(Log, 'Standard Deviation ignored as it provides multiple contribution to ', MNode.FullParameterName(Index));
      end;
      if OtherPar <> '' then begin
        ShowProblem;
        WriteLn(Log, 'Standard Deviation ignored as it contributes to ', MNode.FullParameterName(Index), ' and ', OtherPar);
      end;
    end;
    MNode.Floating[Index] := Abs(R.ERRPAR) > 1e-35;
    MNode.Parameters[Index] := MNode.Parameters[Index] + Scale*R.PAR;
    if MNode.Floating[Index] then
      MNode.StdDev[Index] := Sqrt(Sqr(MNode.StdDev[Index]) + Sqr(Scale*R.ERRPAR));
    Parameters[nParameters].MNode := MNode;
    Parameters[nParameters].Index := Index;
    Parameters[nParameters].Value := Sign(Scale); { Needed to convert correlation matrix }
  end;
end;

procedure TMoleculeReader.Read;
var
  Manifold: TManifold;
  procedure SeparateStates;
  var
    i, j, k: Integer;
    XMLDocument: TXMLDocument;
    DOMElement: TDOMElement;
    NewManifold: TManifold;
    NewPerturbation: TPerturbation;
  begin
    XMLDocument := TXMLDocument.Create;
    { Separate out non-connected states into separate manifolds }
    for i := 1 to Manifold.nStates-1 do begin
      DOMElement := Manifold.WriteProperties(XMLDocument,True,False);
      Manifold.State[i].WriteProperties(DOMElement);
      NewManifold := TMNodeClass(Manifold.ClassType).Create(Manifold.Owner) as TManifold;
      NewManifold.ReadProperties(XMLDocument.FirstChild);
      NewManifold.Name := 'X'+IntToStr(i);
      XMLDocument.Free;
      for j := 0 to nParameters - 1 do
        if Parameters[j].MNode = Manifold.State[i] then
          Parameters[j].MNode := NewManifold.State[0];
      for j := 0 to Manifold.nPerturbations-1 do
        if Manifold.Perturbation[j].Bra = Manifold.State[i] then begin
          XMLDocument := TXMLDocument.Create;
          NewPerturbation := TMNodeClass(Manifold.Perturbation[j].ClassType).Create(NewManifold) as TPerturbation;
          NewPerturbation.ReadProperties(XMLDocument.FirstChild);
          for k := 0 to nParameters - 1 do
            if Parameters[k].MNode = Manifold.Perturbation[j] then
              Parameters[k].MNode := NewPerturbation;
          Manifold.Perturbation[j].Free;
          XMLDocument.Free;
        end;
      Manifold.State[i].Free;
    end;
  end;
var
  i, j, StateMin, StateMax, s, p: Integer;
  State: TState;
  LineOk: Boolean;
  MNode: TMNode;
  Index, nPerturbations, nCor: Integer;
  Scale, v: Double;
  line, val: string;
  Cor, CorT: TMatrix;
begin
  R.Mixture.Species[0].Isotopomer[0].Name := Trim(Copy(R.Title,1,49));
  R.Mixture.Species[0].Isotopomer[0].Comment := R.Title;
  Manifold := R.Mixture.Species[0].Isotopomer[0].Manifold[0];
  Manifold.Name := 'X';
  Manifold.Initial := True;
  Manifold.Molecule.nNuclei := Length(NuclearSpins);
  for i := 0 to Manifold.Molecule.nNuclei - 1 do
    Manifold.Molecule.Nuclei[i].Name := 'Nucleus' + IntToStr(i+1);
  SetLength(StateMap, ABS(R.OI.NVIB));
  SetLength(Deleted, ABS(R.OI.NVIB));
  for i := 0 to Abs(R.OI.NVIB) - 1 do
    StateMap[i] := i;
  for i := 0 to Abs(R.OI.NVIB)-1 do begin
    State := AddState(Manifold,i);
    State.Name := IntToStr(i);
    State.AdjustNuclei(True);
    for j := 0 to Manifold.Molecule.nNuclei-1 do
      State.Nucleus(j).Spin := NuclearSpins[j];
  end;
  { Delete unwanted states }
  i := Abs(R.OI.NVIB)-1; { Logical index }
  repeat
    Deleted[i] := StateMap[i] <> i;
    if Deleted[i] then begin
      Manifold.State[i].Destroy;
      for j := i to Abs(R.OI.NVIB) - 1 do
        if StateMap[j] >= i then
          Dec(StateMap[j]);
    end;
    Dec(i);
  until i < 0;
  nParameters := 0;
  nPerturbations := 0;
  while R.ReadParameterLine(LineOK) do
    if LineOK then begin
      if nParameters >= High(Parameters) then
        SetLength(Parameters, nParameters+20);
      StateMin := R.V1;
      StateMax := R.V2;
      if ((Abs(R.OI.NVIB) < 10) and (R.V1 = 9) and (R.V2 = 9))
         or ((R.V1 = 99) and (R.V2 = 99)) then begin
        StateMin := 0;
        StateMax := Abs(R.OI.NVIB)-1;
      end else if R.V1 <> R.V2 then begin
        Index := ConvertPerturbation(Manifold.State[StateMap[R.V1]], Manifold.State[StateMap[R.V2]], Scale, MNode);
        SetValues(MNode, Index, Scale);
        Inc(nPerturbations);
        Continue;
      end;
      for i := StateMin to StateMax do
        if (StateMin = StateMax) or not Deleted[i] then begin
          Index := ConvertParametersWithSpin(Manifold.State[StateMap[i]], Scale, MNode);
          SetValues(MNode, Index, Scale);
        end;
      Inc(nParameters);
    end else
      Inc(nProblems);
  SetLength(Parameters, nParameters);
  Cor := TMatrix.Create(nParameters,nParameters);
  Cor.Zero;
  i := 1;
  j := 1;
  while not R.S.AtEndOfFile and (i <= nParameters) do begin
    { Read correlation coeficients }
    line := R.S.RestOfLine;
    s := 1;
    while s < Length(line) do begin
      p := s+1;
      while (p < Length(line)) and not (line[p] in [' ', '-']) do
        Inc(p);
      val := Trim(Copy(line,s,p-s));
      if val <> '' then
        if val[1] = '!' then
          Break
        else if TryStrToFloatEx(val, v) then begin
          Cor[i,j] := v;
          Inc(j);
          if j > i then begin
            Inc(i);
            if i > nParameters then
              Break;
            j := 1;
          end;
        end else begin
          Inc(nProblems);
          WriteLn(Log, 'Problem with correlation matrix on line ', R.S.Line, ':', v);
        end;
      s := p;
    end;
    R.S.NextLine;
  end;
  if (nPerturbations = 0) and (Manifold.nStates > 1) then
    SeparateStates;
  CorT := Cor.Transpose;
  CorT.PostMultiply(Cor);
  Cor.Free;
  if (i > 1) or (j > 1) then begin
    R.Mixture.Correlations.FitDetails.Parameters := Parameters;
    R.Mixture.Correlations.AdjustSize;
    for i := 0 to nParameters - 1 do
      for j := 0 to i - 1 do
        R.Mixture.Correlations.FitDetails.Correlation[i][j] := CorT[i+1,j+1]*Parameters[i].Value*Parameters[j].Value;
  end;
  CorT.Free;
end;

procedure TMoleculeReader.AddSpins;
var
  t, i: Integer;
  SpinDigits, SPIND: Integer;
  Index, SIndex: Integer;
  Buffer: string;
  function NextSpin: TQuantumNumber;
  begin
    Result := (t mod SpinDigits) - 1;
    t := t div SpinDigits;
  end;
begin
  ElectronSpin := 0;
  ElectronSpinIndex := -1;
  SPIND := ABS(R.OI.SPIND);
  for i := 0 to High(R.OIs) do
    if ABS(R.OIs[i].SPIND) <> SPIND then begin
      WriteLn(Log, 'Inconsistent SPIND values: ', SPIND, ' <> ', ABS(R.OIs[i].SPIND));
      SPIND := Max(SPIND, ABS(R.OIs[i].SPIND));
    end;
  if SPIND > 1 then begin
    Buffer := IntToStr(DefaultElectronSpin);
    {$IFNDEF CONSOLE}
    if not InputQuery('Select Electron Spin',
                      'Enter spin number for electron (Enter 0 for singlet states) (SPIND='+IntToStr(SPIND)+')', Buffer) then
      Abort;
    {$ENDIF}
    ElectronSpinIndex := StrToInt(Buffer);
    if ElectronSpinIndex = 0 then
      ElectronSpinIndex := -1;
    t := SPIND;
    if t mod 10 = 0 then begin
      SpinDigits := 100;
      t := t div 10;
    end else
      SpinDigits := 10;
    Index := 0;
    SIndex := 0;
    while t <> 0 do begin
      SetLength(SpinIndex, Index+2); { Ignore first entry }
      if Index = ElectronSpinIndex-1 then begin
        ElectronSpin := NextSpin;
        WriteLn(Log, 'Assuming electron is spin number ', ElectronSpinIndex, ' (', ElectronSpin, '/2)');
        SpinIndex[High(SpinIndex)] := -1;
      end else begin
        SetLength(NuclearSpins, Length(NuclearSpins)+1);
        NuclearSpins[High(NuclearSpins)] := NextSpin;
        SpinIndex[High(SpinIndex)] := SIndex;
        Inc(SIndex);
        WriteLn(Log, 'Assuming spin number ', Index+1, ' is a nuclear spin of ', NuclearSpins[High(NuclearSpins)], '/2');
        R.Mixture.Species[0].Isotopomer[0].JAdjustSym := False;
      end;
      Inc(Index);
    end
  end;
end;

function ImportCALPGMFile(S: TTextStream): TMixture;
var
  Reader: TCALPGMReader;
  MoleculeReader: TMoleculeReader;
  Species: TSpecies;
begin
  S.ItemSeparators := S.ItemSeparators + [isComma];
  Reader := TCALPGMReader.Create(S);
  try
    Result := TMixture.Create(nil);
    try
      Reader.Mixture := Result;
      Result.Name := 'Mixture';
      Result.Units := MHz;
      Result.Simulations[0].IntensityUnits := nm2MHzperMolecule;
      Result.Simulations[0].PlotUnits := MHz;
      Species := TSpecies.Create(Result);
      Species.Name := 'S';
      if Reader.OI.KNMIN = Reader.OI.KNMAX then
        MoleculeReader := TLinearMoleculeReader.Create(Reader)
      else if Reader.OI.SPIND >= 0 then
        MoleculeReader := TAsymmetricMoleculeReader.Create(Reader)
      else
        MoleculeReader := TSymmetricMoleculeReader.Create(Reader);
      try
        MoleculeReader.Read;
        {$IFNDEF CONSOLE}
        if MoleculeReader.nProblems > 0 then
          ShowMessage('Problems with import - see log window');
        {$ENDIF}
      finally
        MoleculeReader.Free;
      end;
      Result.FixupProperties;
    except
      Result.Free;
      Raise;
    end;
  finally
    Reader.Free;
  end;
end;

function TMoleculeReader.ConvertParametersWithSpin(State: TState; out Scale: Double; out MNode: TMNode):Integer;
var
  IDI: Integer;
begin
 Scale := 1;
 MNode := State;
 if (R.I1 <> 0) or (R.I2 <> 0) then begin
   IDI := (((R.FF*1000 + R.NS)*100 + R.TYP)*10 + R.KSQ)*10 + R.NSQ;
   if R.I1 = ElectronSpinIndex then
     if R.I2 = 0 then
       Result := SParameter(IDI, State, Scale, MNode)
     else if R.I2 = ElectronSpinIndex then
       Result := SParameter(IDI + 1100000, State, Scale, MNode)
     else
       Result := ISParameter(R.I2, IDI, State, Scale, MNode)
   else if R.I2 = ElectronSpinIndex then
     if R.I1 = 0 then
       Result := SParameter(IDI, State, Scale, MNode)
     else
       Result := ISParameter(R.I1, IDI, State, Scale, MNode)
   else if R.I2 = 0 then
     Result := IParameter(R.I1, IDI, State, Scale, MNode)
   else if R.I1 = 0 then
     Result := IParameter(R.I2, IDI, State, Scale, MNode)
   else if R.I1 = R.I2 then
     Result := IParameter(R.I2, IDI + 1100000, State, Scale, MNode)
   else
     Result := IIParameter(R.I1, R.I2, IDI, State, Scale, MNode);
 end else
   Result := ConvertParameter(State, Scale, MNode)
end;

{ TCALPGMReader }

constructor TCALPGMReader.Create(Stream: TTextStream);
var
  OIt: TCalOptionInfo;
  i: Integer;
begin
  S := Stream;
  Title := S.RestOfLine;
  S.NextLine;
  nPar := S.NextInteger(999);
  S.NextLine; { Skip NLINE, NITR, NXPAR,  THRESH, ERRTST, FRAC, CAL }
  S.FlowToNextLine := False;
  { Line 3 }
  OI := ReadOptions(True);
  SetLength(OIs, Abs(OI.NVIB));
  for i := 0 to Abs(OI.NVIB) - 1 do
    OIs[i] := OI;
  OIt := OI;
  while OIt.VSYM < 0 do begin
    OIt := ReadOptions(False);
    OIs[Abs(OIt.NVIB)] := OIt;
  end;
end;

function TCALPGMReader.ReadOptions(FirstLine: Boolean): TCalOptionInfo;
var
  c: string;
begin
  c := S.NextString('g');
  if (Length(c) = 1) and not (c[1] in ['0'..'9']) then begin
    Result.CHR := c[1];
    S.NextIntegerVar(Result.SPIND);
  end else
    Result.SPIND := StrToInt(c);
  S.NextIntegerVar(Result.NVIB);
  if FirstLine and (Result.NVIB = 0) then begin
    Result.NVIB := 1;
    WriteLn(Log, 'NVIB read as 0 - assuming 1');
  end;
  Result.KNMIN := S.NextInteger(0);
  Result.KNMAX := S.NextInteger(-1);
  Result.IXX := S.NextInteger(0);
  if FirstLine and (Result.IXX <> 0) then
    WriteLn(Log, 'IXX = ', Result.IXX, ' ignored - all off-diagonal interactions will be included');
  Result.IAX := S.NextInteger(1);
  Result.WTPL := S.NextInteger(1);
  Result.WTMN := S.NextInteger(1);
  Result.VSYM := S.NextInteger(0);
  Result.EWT := S.NextInteger(99);
  Result.DIAG := S.NextInteger(0);
  S.NextLine;
end;

function TCALPGMReader.ReadParameterLine(out LineOK:Boolean): Boolean;
var
  t: Integer;
  idstring: string;
  function ND(n: Integer = 10): Integer;
  begin
    Result := t mod n;
    t := t div n;
  end;
begin
  LineOK := True;
  Result := False;
  if (nPar <= 0) or not S.GetFreeItem or (S.Token = '') then
    Exit;
  idstring := S.Token;
  if idstring[Length(idstring)] = '.' then
    Delete(idstring, Length(idstring), 1);
  try
    IDPAR := StrToInt(idstring);
    if IDPAR = MaxInt then
      Exit;
    t := Abs(IDPAR);
    if Abs(OI.NVIB) < 10 then begin
      V1 := ND;
      V2 := ND;
    end else if Abs(OI.NVIB) < 100 then begin
      V1 := ND(100);
      V2 := ND(100);
    end else begin
      V1 := ND(1000);
      V2 := ND(1000);
    end;
    ID := t;
    NSQ := ND;
    KSQ := ND;
    TYP := ND(100);
    NS := ND;
    I1 := ND;
    I2 := ND;
    FF := ND;
    S.NextDoubleVar(PAR);
    S.NextDoubleVar(ERRPAR);
    Comment := S.RestOfLine;
  except
    on E: Exception do begin
      Writeln(Log, 'Format problem on line ', S.Line, ':', E.Message);
      LineOK := False;
    end;
  end;
  S.NextLine;
  Dec(nPar);
  Result := True; { Read a line }
end;

{ TAsymmetricMoleculeReader }

function TAsymmetricMoleculeReader.ConvertParameter(State: TState; out Scale: Double; out MNode: TMNode):Integer;
  procedure SetRep(SRep: Boolean);
  var
    i: Integer;
  begin
    if RepKnown then begin
      if TAsymmetricTop(State).Molecule.SReduction <> SRep then
        raise Exception.Create('Inconsistent reduction');
    end else begin
      TAsymmetricTop(State).Molecule.SReduction := SRep;
      if SRep then begin
        for i := 0 to Abs(R.OI.NVIB)-1 do
          with Manifold.State[i] as TAsymmetricTop do
            deltaJ := - deltaJ;
        for i := 0 to nParameters - 1 do
          if (Parameters[i].MNode is TAsymmetricTop) and (Parameters[i].MNode.ParameterNames[Parameters[i].Index] = 'deltaJ') then
            Parameters[i].Value := -Parameters[i].Value
      end;
    end;
    RepKnown := True;
  end;
var
  ParName: string;
begin
 Scale := 1;
 ParName := '';
 MNode := State;
 with State as TAsymmetricTop do
  case R.ID of
     0: ParName := 'Origin';
   100: ParName := 'A';
   200: ParName := 'B';
   300: ParName := 'C';

     1: begin
          ParName := 'BBar';
          SetValues(State, zIndex, 1, ParName);
        end;
    10: ParName := ParameterNames[zIndex];
   400: begin; ParName := 'BDelta'; Scale := 4; end;

     2: begin; ParName := 'DJ'; Scale := -1; end;
    11: begin; ParName := 'DJK'; Scale := -1; end;
    20: begin; ParName := 'DK'; Scale := -1; end;
   410: begin; ParName := 'deltaK'; SetRep(False); Scale := -1; end;
   401: begin; ParName := 'deltaJ'; if not Molecule.SReduction then Scale := -1; end;
   500: begin; ParName := 'deltaK'; SetRep(True); end;

     3: ParName := 'HJ';
    12: ParName := 'HJK';
    21: ParName := 'HKJ';
    30: ParName := 'HK';
   501: begin; ParName := 'phiJK'; SetRep(True); end;
   600: begin; ParName := 'phiK'; SetRep(True); end;
   402: ParName := 'phiJ';
   411: begin; ParName := 'phiJK'; SetRep(False); end;
   420: begin; ParName := 'phiK'; SetRep(False); end;

     4: ParName := 'LJ';
    13: ParName := 'LJJK';
    22: ParName := 'LJK';
    31: ParName := 'LKKJ';
    40: ParName := 'LK';
   403: ParName := 'llJ';
   412: begin; ParName := 'llJK'; SetRep(False); end;
   421: begin; ParName := 'llKJ'; SetRep(False); end;
   430: begin; ParName := 'llK'; SetRep(False); end;
   502: begin; ParName := 'llJK'; SetRep(True); end;
   601: begin; ParName := 'llKJ'; SetRep(True); end;
   700: begin; ParName := 'llK'; SetRep(True); end;

 10001: ParName := 'DsN';  { A Reduction }
 50001: ParName := 'DsNK'; { A Reduction }
 10010: ParName := 'DsKN'; { A Reduction }
 50010: ParName := 'DsK';  { A Reduction }
 10400: ParName := 'ds1';
 50400: ParName := 'ds2';  { A Reduction }
  end;
  if ParName <> '' then
    Result := State.IndexOfParameter(ParName)
  else
    Result := ConvertPerturbation(State, State, Scale, MNode);
end;

function TAsymmetricMoleculeReader.ConvertPerturbation(State1, State2: TState; out Scale: Double; out MNode: TMNode):Integer;
var
  zOffset, zPowerIndex, nOps: Integer;
  Div2: Boolean;
  Perturbation: TAsymmetricTopPerturbation;
begin
  Result := -1;
  Scale := 1;
  if (R.FF = 0) and (R.I2 = 0) and (R.I1 = 0) then begin
    Perturbation := TAsymmetricTopPerturbation.Create(Manifold);
    with Perturbation do begin
      SetBraKetName(State1.Name, State2.Name);
      Result := IndexOfParameter('Value');
      MNode := Perturbation;
      JPower := R.NSQ*2;
      zOffset := Ord(Molecule.abcAxis(zAxis));
      zPowerIndex := zOffset+IndexOfSetting('aPower');
      SetSetting(zPowerIndex,2*R.KSQ);
      Div2 := False;
      Comment := IntToStr(R.IDPAR);
      if R.NS > 0 then begin
        if R.NS > 4 then begin
          R.NS := R.NS - 5;
          { Add NzSz }
          SetSetting(zPowerIndex,GetSetting(zPowerIndex)+1);
          SetSetting(zOffset+IndexOfSetting('SaPower'),1);
          Div2 := True;
        end;
        NSPower := R.NS;
      end;
      case R.TYP of
        0: ;
        1: aPower := aPower + 2;
        2: bPower := bPower + 2;
        3: cPower := cPower + 2;
        4..10: pPower := 2*(R.TYP-3);
        20: aPower := aPower + 1;
        21: begin bPower := bPower + 1; cPower := cPower + 1; end;
        40: bPower := bPower + 1;
        41: begin aPower := aPower + 1; cPower := cPower + 1; end;
        60: cPower := cPower + 1;
        61: begin aPower := aPower + 1; bPower := bPower + 1; end;
      else
        Free;
        Result := -1;
      end;
      nOps := 0;
      if aPower <> 0 then Inc(nOps);
      if bPower <> 0 then Inc(nOps);
      if cPower <> 0 then Inc(nOps);
      if pPower <> 0 then Inc(nOps);
      if Div2 or (nOps > 1) then
        Scale := Scale / 2;
    end;
  end;
end;

procedure TAsymmetricMoleculeReader.Read;
begin
  Molecule := TAsymmetricMolecule.Create(R.Mixture.Species[0]);
  AddSpins;
  if R.OI.NVIB > 0 then
    Molecule.Representation := Ir
  else
    Molecule.Representation := IIIl;
  Molecule.PointGroup := C2v;
  case Abs(R.OI.IAX) of
    1: begin
        Molecule.C2zAxis := aAxis;
        Molecule.eeWt := R.OI.WTPL; Molecule.eoWt := R.OI.WTPL;
        Molecule.oeWt := R.OI.WTMN; Molecule.ooWt := R.OI.WTMN;
      end;
    2: begin
        Molecule.C2zAxis := bAxis;
        Molecule.eeWt := R.OI.WTPL; Molecule.eoWt := R.OI.WTMN;
        Molecule.oeWt := R.OI.WTMN; Molecule.ooWt := R.OI.WTPL;
      end;
    3: begin
        Molecule.C2zAxis := cAxis;
        Molecule.eeWt := R.OI.WTPL; Molecule.eoWt := R.OI.WTMN;
        Molecule.oeWt := R.OI.WTPL; Molecule.ooWt := R.OI.WTMN;
      end;
    0:
  else
    Inc(nProblems);
    WriteLn(Log, 'Unexpected value for IAX:',R.OI.IAX);
  end;
  Manifold := TAsymmetricManifold.Create(Molecule);
  RepKnown := False;
  inherited;
end;

function TAsymmetricMoleculeReader.AddState(AManifold: TManifold; Index: Integer): TState;
var
  State: TAsymmetricTop;
  t: Integer;
begin
  State := TAsymmetricTop.Create(AManifold);
  State.S := ElectronSpin;
  Result := State;
  { This bizarre construct avoids a bug in Delphi }
  t := 2*R.OI.KNMIN;
  State.Kmin := t;
  if R.OI.KNMAX >= 0 then
    State.Kmax := 2*R.OI.KNMAX;
end;

function TAsymmetricMoleculeReader.SParameter(IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer;
var
  ParName: string;
begin
  Result := -1;
  with State as TAsymmetricTop do
    case IDI of
    100: ParName := 'eaa';
    200: ParName := 'ebb';
    300: ParName := 'ecc';
   6100: ParName := 'eabbar';
      1: ParName := 'DsN';  { S Reduction }
     10: ParName := 'DsKN'; { S Reduction }
    101: ParName := 'DsNK'; { S Reduction }
    110: ParName := 'DsK';  { S Reduction }
    500: ParName := 'ds2';
1100100: begin; ParName := 'alpha'; Scale := 2/(3*S*(2*S-2)); end;
    end;
  if ParName <> '' then
    Result := State.IndexOfParameter(ParName)
end;

function TAsymmetricMoleculeReader.ISParameter(INuc, IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer;
var
  ParName: string;
  procedure SetTqq(abcAxis: TabcAxis);
  var
    xyzAxis: TxyzAxis;
  begin
    xyzAxis := (State.Molecule as TAsymmetricMolecule).xyzAxis(abcAxis);
    with State.Nucleus(SpinIndex[INuc]) as TAsymmetricTopNucleus do
      case xyzAxis of
        { NB: Earlier version had 2/3 scale factor here - removed }
        xAxis: ParName := 'Txxmyy';
        yAxis: begin; ParName := 'Txxmyy'; Scale := -1; end;
        zAxis: ParName := 'Tzz';
      end;
  end;
begin
  Result := -1;
  with State.Nucleus(SpinIndex[INuc]) as TAsymmetricTopNucleus do
    case IDI of
      0: ParName := 'aF';
    100: SetTqq(aAxis);
    200: SetTqq(bAxis);
    300: SetTqq(cAxis);
   2100: ParName := 'Tbc';
   4100: ParName := 'Tac';
   6100: ParName := 'Tab';
    400: begin; ParName := 'Txxmyy'; Scale := 4; end;
//  110: TKaa := TKaa*2/3;
//  101: TJaa := TKaa*2/3;
//  410: TKxxmyy := TKxxmyy + PAR*4;
    end;
  if ParName <> '' then
    Result := State.IndexOfParameter(ParName)
end;

function TAsymmetricMoleculeReader.IIParameter(INuc1, INuc2, IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer;
begin
  Result := -1;
end;

function TAsymmetricMoleculeReader.IParameter(INuc, IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer;
var
  ParName: string;
  procedure SetChiqq(abcAxis: TabcAxis);
  var
    xyzAxis: TxyzAxis;
  begin
    xyzAxis := (State.Molecule as TAsymmetricMolecule).xyzAxis(abcAxis);
    with State.Nucleus(SpinIndex[INuc]) as TAsymmetricTopNucleus do
      case xyzAxis of
        { NB: Earlier version had 2/3 scale factor here - removed }
        xAxis: ParName := 'CHIxxmyy';
        yAxis: begin; ParName := 'CHIxxmyy'; Scale := -1; end;
        zAxis: ParName := 'CHIzz';
      end;
  end;
begin
  Result := -1;
  with State.Nucleus(SpinIndex[INuc]) as TAsymmetricTopNucleus do
    case IDI of
    100: ParName := 'Caa';
    200: ParName := 'Cbb';
    300: ParName := 'Ccc';
//  410: chiKxxmyy := chiKxxmyy + PAR*4; ? need spin scaling
1100100: SetChiqq(aAxis);
1100200: SetChiqq(bAxis);
1100300: SetChiqq(cAxis);
1100400: begin; ParName := 'CHIxxmyy'; Scale := 4; end;
1102100: ParName := 'Chibc';
1104100: ParName := 'Chiac';
1106100: ParName := 'Chiab';
    end;
  if ParName <> '' then
    Result := State.IndexOfParameter(ParName)
end;

{ TSymmetricMoleculeReader }

procedure TSymmetricMoleculeReader.Read;
begin
  LastEState := -999;
  Molecule := TSymmetricMolecule.Create(R.Mixture.Species[0]);
  case Abs(R.OI.IAX) of
    6:     Molecule.PointGroup := C3v;
    7,8:   Molecule.PointGroup := C4v;
    9:     Molecule.PointGroup := C5v;
    10,11: Molecule.PointGroup := C6v;
  else
    { Need to scan all states }
  end;
  AddSpins;
  Manifold := TSymmetricManifold.Create(Molecule);
  Inc(nProblems);
  WriteLn(Log, 'Point group, state symmetries and statistical weights may well need attention');
  if ElectronSpin > 0 then
    WriteLn(Log, 'Electron Spin not supported in symmetric tops');
  inherited;
end;

function TSymmetricMoleculeReader.AddState(AManifold: TManifold; Index: Integer): TState;
var
  State: TSymmetricTop;
  EWT, n: Integer;
begin
  State := TSymmetricTop.Create(AManifold);
  State.S := ElectronSpin;
  EWT := R.OIs[Index].EWT;
  if (EWT > 0) and (EWT mod 100 <> 99) then begin
    n := State.Molecule.CharacterTable.n;
    EWT := (EWT div 100) * 2;
    if EWT = 0 then
      { A1/A2 Symmetry }
    else if EWT = n then
      { B1/B2 Symmetry }
    else if EWT < n then begin
      { E Symmetry }
      if LastEState > 0 then
        Writeln(Log, 'State order problem for ', Index)
      else if LastEState = -999 then begin
        LastEState := Index;
      end else begin
        StateMap[-LastEState-1] := Index;
        LastEState := -999;
      end;
      State.Symmetry := 2; { TODO - will need adjustment }
    end else begin
      { Other half of E symmetry - discard }
      if LastEState >= 0 then begin
        StateMap[Index] := LastEState;
        LastEState := -999;
      end else if LastEState = -999 then
         LastEState := -(Index+1)
      else
        WriteLn(Log, 'State ordering problem for ', Index);
      State.Symmetry := 2; { TODO - will need adjustment }
    end;
  end;
  Result := State;
end;

function TSymmetricMoleculeReader.ConvertParameter(State: TState; out Scale: Double; out MNode: TMNode):Integer;
  procedure CheckZetaEta;
  begin
    if not ZetaEtaMsg then
      WriteLn(Log, 'Zeta and eta values may need checking');
    ZetaEtaMsg := True;
  end;
var
  ParName: string;
begin
 Result := -1;
 Scale := 1;
 MNode := State;
 with State as TSymmetricTop do
  case Self.R.ID of
     0: ParName := 'Origin';
     1: begin
          ParName := 'B';
          SetValues(State,IndexOfParameter('C'), 1, 'B');
        end;
    10: ParName := 'C';

     2: begin; ParName := 'DJ'; Scale := -1; end;
    11: begin; ParName := 'DJK'; Scale := -1; end;
    20: begin; ParName := 'DK'; Scale := -1; end;

     3: ParName := 'HJ';
    12: ParName := 'HJK';
    21: ParName := 'HKJ';
    30: ParName := 'HK';

     4: ParName := 'LJ';
    13: ParName := 'LJJK';
    22: ParName := 'LJK';
    31: ParName := 'LKKJ';
    40: ParName := 'LK';

  2000: begin ParName := 'zeta'; Scale := - 1/2/C; CheckZetaEta; end;
  2001: begin ParName := 'etaJ'; CheckZetaEta; end;
  2010: begin ParName := 'etaK'; CheckZetaEta; end;

  end;
  if ParName <> '' then
    Result := State.IndexOfParameter(ParName)
  else
    Result := ConvertPerturbation(State, State, Scale, MNode);
end;

function TSymmetricMoleculeReader.SParameter(IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer;
begin
  Result := -1;
end;

function TSymmetricMoleculeReader.ISParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer;
begin
  Result := -1;
end;

function TSymmetricMoleculeReader.IIParameter(INuc1, INuc2, IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer;
begin
  Result := -1;
end;

function TSymmetricMoleculeReader.IParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer;
var
  ParName: string;
begin
  Result := -1;
  with State.Nucleus(SpinIndex[INuc]) as TSymmetricTopNucleus do
    case IDI of
1100100: begin; ParName := 'eQq'; Scale := 2/3; end;
      0: begin
           ParName := 'Cperp';
           SetValues(State, IndexOfParameter('Czz'), 1, 'Cperp');
         end;
    100: ParName := 'Czz';
    end;
  if ParName <> '' then
    Result := State.IndexOfParameter(ParName)
end;

function TSymmetricMoleculeReader.ConvertPerturbation(State1, State2: TState; out Scale: Double; out MNode: TMNode):Integer;
var
  Perturbation: TSymmetricTopPerturbation;
begin
  Result := -1;
  Scale := 1;
  if State1 = State2 then
    with State1 as TSymmetricTop do begin
      case Self.R.ID of
        { These are really diagonal, but expressed as matrix elements between
          degenerate states }
        400: begin Result := IndexOfParameter('qplus'); Scale := -2; end;
        401: begin Result := IndexOfParameter('DqJ'); Scale := -2; end;
        410: begin Result := IndexOfParameter('DqK'); Scale := -1; end;
      end;
      if Result > 0 then
        MNode := State1;
    end;
  if (Result < 0) and (R.FF = 0) and (R.I2 = 0) and (R.I1 = 0) and (R.NS = 0) then begin
    Perturbation := TSymmetricTopPerturbation.Create(Manifold);
    with Perturbation do begin
      SetBraKetName(State1.Name, State2.Name);
      Result := IndexOfParameter('Value');
      MNode := Perturbation;
      JPower := R.NSQ*2;
      zPower := 2*R.KSQ;
      Comment := IntToStr(R.IDPAR);
      case R.TYP of
        0: ;
        //1: aPower := aPower + 2;
        //2: bPower := bPower + 2;
        //3: cPower := cPower + 2;
        { TODO: lChange values should probably be adjusted to satisfy Delta(K,l) rules }
        4..10: begin
                 pPower := 2*(R.TYP-3);
                 lChange := 2;
                 Scale := -Scale;
               end;
        { TODO: lPower values are probably C3v specific }
        20..39: begin
                  pPower := (R.TYP-19) div 2 * 2;
                  lPower := 1;
                  zPower := zPower + 1;
                end;
        60..79: begin
                  pPower := (R.TYP-60) div 2 * 2 + 1;
                  lChange := -2;
                end;
        //20: aPower := aPower + 1;
        //21: begin bPower := bPower + 1; cPower := cPower + 1; end;
        //40: bPower := bPower + 1;
        //41: begin aPower := aPower + 1; cPower := cPower + 1; end;
        //60: cPower := cPower + 1;
        //61: begin aPower := aPower + 1; bPower := bPower + 1; end;
      else
        Free;
        Result := -1;
      end;
    end;
  end;
end;

{ TLinearMoleculeReader }

procedure TLinearMoleculeReader.Read;
begin
  Molecule := TLinearMolecule.Create(R.Mixture.Species[0]);
  Molecule.Symmetric := R.OI.WTMN <> R.OI.WTPL;
  AddSpins;
  Manifold := TLinearManifold.Create(Molecule);
  inherited Read;
end;

function TLinearMoleculeReader.AddState(AManifold: TManifold; Index: Integer): TState;
var
  State: TLinear;
begin
  State := TLinear.Create(AManifold);
  State.S := ElectronSpin;
  State.Lambda := R.OI.KNMIN*2;
  Result := State;
end;

function TLinearMoleculeReader.ConvertParameter(State: TState; out Scale: Double; out MNode: TMNode):Integer;
var
  ParName: string;
begin
  Result := -1;
  Scale := 1;
  MNode := State;
  with State as TLinear do
    case R.ID of
      0: ParName := 'Origin';
      1: ParName := 'B';
      2: begin; ParName := 'D'; Scale := -1; end;
      3: ParName := 'H';
      4: ParName := 'L';
      5: ParName := 'M';
      6: ParName := 'PP';

    400: begin; ParName := 'q'; Scale := 2; end;
    401: begin; ParName := 'qD'; Scale := 2; end;
    402: begin; ParName := 'qH'; Scale := 2; end;
    403: begin; ParName := 'qL'; Scale := 2; end;
    end;
  if ParName <> '' then
    Result := State.IndexOfParameter(ParName)
  else
    Result := ConvertPerturbation(State, State, Scale, MNode);
end;

function TLinearMoleculeReader.ConvertPerturbation(State1, State2: TState; out Scale: Double; out MNode: TMNode):Integer;
begin
  Result := -1;
end;

function TLinearMoleculeReader.SParameter(IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer;
var
  ParName: string;
begin
  Result := -1;
  with State as TLinear do
    case IDI of
      0: ParName := 'gamma';
      1: ParName := 'gammaD';
      2: ParName := 'gammaH';
      3: ParName := 'gammaL';

    400: begin; ParName := 'p'; Scale := -2; end;
    401: begin; ParName := 'pD'; Scale := -2; end;
    402: begin; ParName := 'pH'; Scale := -2; end;
    403: begin; ParName := 'pL'; Scale := -2; end;

    100: ParName := 'A';

    1100100: begin; ParName := 'LambdaSS'; Scale := 1/(S*(2*S-2)); end;
    1100101: begin; ParName := 'LambdaD'; Scale := 1/(S*(2*S-2)); end;
    1100400: ParName := 'o';
    1100401: ParName := 'oD';
    end;
  if ParName <> '' then
    Result := State.IndexOfParameter(ParName)
end;

function TLinearMoleculeReader.ISParameter(INuc, IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer;
var
  ParName: string;
begin
  Result := -1;
  MNode := State.Nucleus(SpinIndex[INuc]);
  with MNode as TLinearNucleus do begin
    case IDI of
      0: ParName := 'b';
    100: begin
           ParName := 'c';
           SetValues(MNode, MNode.IndexOfParameter('b'), -1/3, 'c');
         end;
    400: begin; ParName := 'd'; Scale := -2; end;
    401: begin; ParName := 'dJ'; Scale := -2; ShowProblem; WriteLn(Log, 'Conversion is approximate'); end;
    end;
    if ParName <> '' then
      Result := MNode.IndexOfParameter(ParName);
  end;
end;

function TLinearMoleculeReader.IIParameter(INuc1, INuc2, IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer;
var
  SpinSpin :TLinearSpinSpin;
  ParName: string;
begin
  Result := -1;
  MNode := State.SpinSpin(SpinIndex[INuc1], SpinIndex[INuc2]);
  with MNode as TLinearSpinSpin do
    case IDI of
        0:  ParName := 'J';
      100:  begin; ParName := 'S'; Scale := 1/3; end;
    end;
  if ParName <> '' then
    Result := MNode.IndexOfParameter(ParName)
end;

function TLinearMoleculeReader.IParameter(INuc, IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer;
var
  ParName: string;
begin
  Result := -1;
  MNode := State.Nucleus(SpinIndex[INuc]);
  with MNode as TLinearNucleus do begin
    case IDI of
            0: ParName := 'cI';
          100: ParName := 'a';
      1100100: begin; ParName := 'eQq0'; Scale := 2/3; end;
      1100400: begin; ParName := 'eQq2'; Scale := -4; end;
    end;
    if ParName <> '' then
      Result := MNode.IndexOfParameter(ParName)
  end;
end;

end.
