unit ImportJB95;
{$I DFS.inc}{Copyright (C) 2007 C M Western; see commandline.pas for license}
{ Import data from David Plusquellic's JB95 program }
interface

uses TextStrm, Mixture, Obs;

function ImportJB95File(Stream: TTextStream): TMixture;

type
  TJJB95Reader = class(TObservationsReader)
    constructor Create; override;
    function DoReadOneLine: Boolean; override;
  end;

implementation

uses {$IFNDEF CONSOLE} Dialogs, {$ENDIF}
     SysUtils, OAsymT, Worker, Units, MNode, Classes, ExpPlotLoader,
     QuantumNumber;

{ Should be class variables, but don't fit as such }
var
  S:TTextStream;
  nMsgs: Integer;

procedure Msg(const AMsg: string; Fatal: Boolean = False);
begin
  if Fatal then
    raise Exception.Create(AMsg)
  else begin
    Inc(nMsgs);
    WriteLn(Log, AMsg);
  end;
end;

procedure SkipTo(const Marker: string);
begin
  while S.GetFreeItem do
    if S.Token = Marker then
      Exit;
  Msg(Marker + ' not found', True);
end;

function NextHex:Integer;
begin
  Result := StrToInt(HexDisplayPrefix+S.NextString('0'));
end;

function NextColour:Integer;
begin
  Result := NextHex*256*256;
  Result := Result + NextHex*256;
  Result := Result + NextHex;
end;

function NextUnits:TResonanceUnits;
var
  U: string;
begin
  U := S.NextString('MHz');
  Result := MHz; { Default }
  if U = 'MHz' then
    Result := MHz
  else if U = 'CM-1' then
    Result := cm1
  else
    Msg('Unknown units:'+U);
end;

function ImportJB95File(Stream: TTextStream): TMixture;
var
  PlotUnitsSet: Boolean;

  procedure ImportSpecies(St: TTextStream; Mixture: TMixture; const BaseName: string);
  var
    Species: TSpecies;
    Molecule: TAsymmetricMolecule;
    ManifoldL, ManifoldU: TAsymmetricManifold;
    StateL, StateU: TAsymmetricTop;
    TransitionMoments: TTransitionMoments;

    procedure ReadManifold(out Manifold: TAsymmetricManifold; out State: TAsymmetricTop);
    begin
      Manifold := TAsymmetricManifold.Create(Molecule);
      State := TAsymmetricTop.Create(Manifold);
      St.NextLine;
      with State do begin
        A := St.NextDouble(0);
        B := St.NextDouble(0);
        C := St.NextDouble(0);
        { Could also read fit indicators }
        { TODO: <PZ>(A), <PX>(B) & <PY>(C) CONSTANTS }
        { TODO: <PXPY + PYPX>, <PXPZ + PZPX> & <PYPZ + PZPX> }
        SkipTo('<DJ>');
        St.NextLine;
        DJ := St.NextDouble(0);
        DJK := St.NextDouble(0);
        DK := St.NextDouble(0);
        SkipTo('<dJ>');
        St.NextLine;
        deltaJ := St.NextDouble(0);
        deltaK := St.NextDouble(0);
      end;
    end;
  
    procedure ReadTMom(AnAxis: TabcAxis);
    var
      Str: Double;
    begin
      Str := St.NextDouble(0);
      if Str <> 0 then
        with TCartesianTransitionMoment.Create(TransitionMoments) do begin
          Strength := Str;
          Axis := AnAxis;
          SetBraKetName(StateU.Name, StateL.Name);
        end;
    end;

  var
    i: Integer;
    OldStream: TTextStream;
  begin
    OldStream := S;
    S := St;
    Species := TSpecies.Create(Mixture);
    if BaseName = '' then
      Species.Name := 'Species'
    else
      Species.Name := BaseName;
    Molecule := TAsymmetricMolecule.Create(Species);
    if BaseName = '' then
      Molecule.Name := 'Molecule'
    else
      Molecule.Name := BaseName;

    St.Comments := False; { Want # comments }
    SkipTo('VERSION');
    if St.NextString('') <> 'v5.0' then
      Msg('Version not 5.0');
    SkipTo('COMMENT');
    St.NextLine;
    Molecule.Comment := Trim(St.RestOfLine);
    if Molecule.Comment = 'NO COMMENT' then
      Molecule.Comment := '';
    SkipTo('GROUND');
    ReadManifold(ManifoldL, StateL);
    with ManifoldL do begin
      Name := 'Ground';
      Initial := True;
    end;
    StateL.Name := 'Lower';
    SkipTo('EXCITED');
    ReadManifold(ManifoldU, StateU);
    ManifoldU.Name := 'Excited';
    StateU.Name := 'Upper';
    SkipTo('ORIGIN');
    St.NextLine;
    StateU.Origin := St.NextDouble(0);
    if StateU.Origin = 0 then begin
      ManifoldU.Free;
      ManifoldU := ManifoldL;
      StateU := StateL;
    end;
    SkipTo('REDUCTION');
    St.NextLine;
    St.NextInteger(0); { One State fit statue }
    i := St.NextInteger(1);
    Molecule.SReduction := i = 0;
    SkipTo('NUCLEAR');
    St.NextLine;
    Molecule.eeWt := St.NextInteger(1);
    Molecule.eoWt := St.NextInteger(1);
    Molecule.oeWt := St.NextInteger(1);
    Molecule.ooWt := St.NextInteger(1);
    SkipTo('TRANSITION');
    St.Nextline;
    if St.NextInteger(0) <> 1 then
      Msg('Angular transition moment not implemented');
    SkipTo('TRANSITION');
    St.NextLine;
    TransitionMoments := TTransitionMoments.Create(Molecule);
    TransitionMoments.SetBraKetName(ManifoldU.Name, ManifoldL.Name);
    ReadTMom(aAxis);
    ReadTMom(bAxis);
    ReadTMom(cAxis);
    SkipTo('TEMPERATURES(K)');
    St.NextLine;
    Mixture.Simulations[0].Temperature := St.NextDouble(3);
    St.NextDouble(0);
    if St.NextDouble(0) <> 0 then
      Msg('T2 not implemented');
    Mixture.Units := NextUnits;
    if PlotUnitsSet and (Mixture.Simulations[0].PlotUnits <> Mixture.Units) then
      Msg('Mixed units not implemented')
    else
      PlotUnitsSet := True;
    Mixture.Simulations[0].PlotUnits := Mixture.Units;
    S := OldStream;
  end;


var
  Line, Name: string;
  SubStream: TTextStream;
  FileStream: TFileStream;
  I, n: Integer;
  Active: array [0..8] of Boolean;
begin
  nMsgs := 0;
  Result := nil;
  PlotUnitsSet := False;
  try
    Result := TMixture.Create(nil);
    S := Stream;
    S.Comments := False; { Want # comments }
    S.NextLine;
    Line := S.RestOfLine;
    if Pos('IAR', Line) > 0 then begin
      S.Position := 0;
      ImportSpecies(S, Result,'');
    end else begin
      SkipTo('SIMULATION');
      SkipTo('CHN');
      for I := 0 to High(Active) do begin
        S.NextLine;
        Active[I] := False;
        Line := S.NextString('');
        if (Line = '') or (Line = '#') then
          Break;
        if S.NextInteger(0) > 0 then begin
          Name := S.NextString('');
          Line := Name +'.in';
          if FileExists(Line) then begin
            SubStream := nil;
            FileStream := TFileStream.Create(Line,fmOpenRead or fmShareDenyWrite);
            try
              SubStream := TFreeFormatStream.Create(FileStream);
              ImportSpecies(SubStream,Result,Name);
              Active[I] := True;
            finally
              SubStream.Free;
              FileStream.Free;
            end;
          end else
            Msg('Skipping '+Line);
        end;
      end;
      SkipTo('COLOR:RGB');
      n := 0;
      for I := 0 to High(Active) do begin
        S.NextLine;
        if Active[I] then begin
          S.NextInteger(0); { Channel number }
          S.NextDouble(0);  { Gaussian Width }
          S.NextDouble(0);  { Lorentzian Width }
          S.NextInteger(0); { Widths }
          S.NextInteger(0); { Assoc }
          S.NextInteger(0); { Pen }
          Result.Species[n].Isotopomer[0].Colour := NextColour;
          Inc(n);
        end;
      end;
    end;
    {$IFNDEF CONSOLE}
    if nMsgs > 0 then
      ShowMessage('Problems with import - see log for details');
    {$ENDIF}
  except
    Result.Free;
    raise;
  end;
end;

type
  TJB95PlotLoader = class(TExperimentalPlotLoader)
  public
    function Load(var Log: TextFile; AskAboutWeights: Boolean): Boolean; override;
  end;

{ TJB95PlotLoader }

function TJB95PlotLoader.Load(var Log: TextFile; AskAboutWeights: Boolean): Boolean;
var
  I, j, nActive, nPlot: Integer;
  Files: array [0..8] of string;
  Lengths: array [0..8] of Integer;
  Line: string;
  FileStream: TFileStream;
  Start, FSR, PTS: Double;
begin
  Result := False;
  if Ext <> '.jbw' then
    Exit;
  Result := True;
  S := TFreeFormatStream.Create(Stream);
  S.Comments := False;
  try
    SkipTo('EXPT');
    SkipTo('CHN');
    nActive := 0;
    for I := 0 to High(Files) do begin
      S.NextLine;
      Line := S.NextString('');
      if (Line = '') or (Line = '#') then
        Break;
      S.NextInteger(0); { DIG }
      if S.NextInteger(0) <> 0 { DSP } then begin
        Files[I] := S.NextString('');
        if not FileExists(Files[I]) then begin
          Files[I] := '';
          Msg('Skipping '+Line);
        end else
          Inc(nActive);
      end;
    end;
    SetupChannels(nActive, 0, True); { 0 is placeholder }
    nPlot := 0;
    for I := 0 to High(Files) do
      if Files[I] <> '' then begin
        FileStream := TFileStream.Create(Files[I], fmOpenRead or fmShareDenyWrite);
        try
          Lengths[I] := FileStream.Size div SizeOf(Single);
          Connector.SetLength(nPlot, Lengths[I], False, True, True);
          ReadAndCopyArray(nPlot,1,tySingle,-1,1,FileStream);
        finally
          FileStream.Free;
        end;
        Inc(nPlot);
      end;
    SkipTo('CHN');
    nPlot := 0;
    for I := 0 to High(Files) do begin
      S.NextLine;
      Line := S.NextString('');
      if (Line = '') or (Line = '#') then
        Break;
      if Files[I] <> '' then begin
        if S.NextInteger(0) <> Lengths[I] then
          raise Exception.CreateFmt('Inconsistent plot lengths for plot %d',[I]);
         { Plot length - should be already set  }
        S.NextInteger(0); { JBI SIZE }
        S.NextString(''); { AVG }
        S.NextInteger(0); { JBM SIZE }
        S.NextString(''); { AVG }
        S.NextInteger(0); { JBO SIZE }
        S.NextString(''); { AVE }
        S.NextInteger(0); { Pen }
        Connector.SetColour(nPlot,NextColour);
        Inc(nPlot);
      end;
    end;
    SkipTo('CHN');
    nPlot := 0;
    for I := 0 to High(Files) do begin
      S.NextLine;
      Line := S.NextString('');
      if (Line = '') or (Line = '#') then
        Break;
      if Files[I] <> '' then begin
        Start := S.NextDouble(0);
        FSR := S.NextDouble(0);
        PTS := S.NextDouble(0);
        Connector.SetUnits(nPlot,NextUnits);
        for j := 0 to Lengths[I] do
          Connector.SetValue(nPlot,0,j,Start + FSR/PTS*2*j); { Not sure why 2 is needed }
        Inc(nPlot);
      end;
    end;
  finally
    FreeAndNil(S);
  end;
end;

{ TJJB95Reader }

constructor TJJB95Reader.Create;
begin
  inherited;

end;

function TJJB95Reader.DoReadOneLine: Boolean;
var
  Kap, Kcp, Ka, Kc: TQuantumNumber;
  Int, Calc, Obsf: Double;
  np: Smallint; { Upper State }
  n: SmallInt; { Lower State }
  StrU, StrL: string;
begin
  Result := False;
  if not f.GetFreeItem then
    Exit;
  { Ignore a, b or c }
  if (Length(f.Token) = 1) and (f.Token[1] in ['A'..'C','a'..'c']) then
    f.GetFreeItem;
  ReadQuantumNumber(Obs.JU, f.Token);
  ReadQuantumNumber(Kap,NextItem); ReadQuantumNumber(Kcp,NextItem);
  ReadQuantumNumber(Obs.JL,NextItem);
  ReadQuantumNumber(Ka,NextItem); ReadQuantumNumber(Kc,NextItem);
  f.NextDoubleVar(Int); f.NextDoubleVar(Obsf); f.NextDoubleVar(Calc);
  Obs.Frequency := Calc;
  case ObsType of
    obLine: case Take of
              tkObs:  Obs.Frequency := Obsf;
              tkAll:  if Obsf <> 0 then
                        Obs.Frequency := Obsf;
            end;
    obIntensity: Obs.Frequency := Int;
  end;
  Obs.LineNumber := f.Line;
  Obs.FileName := FileName;
  StrU := DefaultManifoldU.FindStateLabel(np, Obs.SymU, Obs.JU, [Kap, Kcp], qftJ, ExField, -1, -1, -1, DefaultStateU);
  if np < 0 then
    ReportError('Upper state not found:'+StrU)
  else begin
    StrL := DefaultManifoldL.FindStateLabel(n, Obs.SymL, Obs.JL, [Ka, Kc], qftJ, ExField, -1, -1, -1, DefaultStateL);
    if n < 0 then
      ReportError('Lower state not found:'+StrL)
    else begin
      Obs.IndexU := 0;
      Obs.IndexL := 0;
      Obs.Name := ':' + StrU + ' - ' + StrL + ':' + Obs.Name;
      Obs.Weight := 1;
      Result := (Obsf <> 0) or (Take <> tkCalc);
    end;
  end;
end;

initialization
  RegisterPlotLoader(TJB95PlotLoader, 'JB95 files', '*.jbw');
end.

