unit MatrixElements;
{$I DFS.inc}{Copyright (C) 2006 C M Western; see commandline.pas for license}
interface

uses Expr, QuantumNumber;

function rlE(J, K: TExprNode; dK: Integer): TExprNode; overload;
function rlE(J: TExprNode; Kp, K: TQuantumNumber): TExprNode; overload;
function rlxE(n: Integer; J, K: TExprNode; dK: TQuantumNumber):TExprNode;
function rlyE(n: Integer; J, K: TExprNode; dK: TQuantumNumber):TExprNode;
function Halve(Q: TQuantumNumber):TExprNode;
function ReducedMatrixElementE(Rank: TQuantumNumber; J:TExprNode): TExprNode;

type
  PExprNode = ^TExprNode;
  
  TMatrixElement = record
    Signature: array of Integer; { Key quantum numbers (or change in them) to allow matrix element to be identified }
    Expr: TExprNode;
    Evaluator: TEvaluator;
  end;
  
  TMatrixElements = array of TMatrixElement;
  
  TSortedVariableNode = class(TVariableNode)
    Index: Integer;
    function DependsOn: String; override;
    function SortText(ConstantsHigh:Boolean = False): string; override;
  end;

  TSortedVariable = class (TVariable)
    Template: TSortedVariableNode;
    Index: Integer;
    function CreateExprNode:TExprNode; override;
    constructor Create(s:String; AIndex: Integer);
    destructor Destroy; override;
  end;

var
  ExprChecking: Boolean = False; { Set to evaluate matrix elements symbolically as well as numerically }

function MakeSortedVariables(const Names: array of string; const Nodes: array of PExprNode):TVariableList;
procedure DiscardMatrixElements(var MatrixElements: TMatrixElements);

implementation

uses SysUtils, Math;

{ TSortedVariableNode }

function TSortedVariableNode.DependsOn: String;
begin
  { DependsOn is used for sorting, so force order here by prefixing with number }
  Result := IntToHex(Index,8) + inherited DependsOn;
end;

function TSortedVariableNode.SortText(ConstantsHigh:Boolean = False): string;
begin
  Result := IntToHex(Index,8) + inherited SortText(ConstantsHigh);
end;

{ TSortedVariable }

function TSortedVariable.CreateExprNode: TExprNode;
var
  R: TSortedVariableNode;
begin
  R := TSortedVariableNode.Create(Self);
  R.Index := Index;
  Result :=  R;
end;

constructor TSortedVariable.Create(s: String; AIndex: Integer);
begin
  inherited Create(s);
  Index := AIndex;
  Template := CreateExprNode as TSortedVariableNode;
end;

destructor TSortedVariable.Destroy;
begin
  Template.Free;
  inherited Destroy;
end;
  
function MakeSortedVariables(const Names: array of string; const Nodes: array of PExprNode):TVariableList;
var
  i: Integer;
  V: TSortedVariable;
begin
  Result := TVariableList.Create;
  for i := 0 to High(Names) do begin
    V := TSortedVariable.Create(Names[i], 1000+i);
    Result.Add(V);
    Nodes[i]^ := V.Template;
  end;
end;

procedure DiscardMatrixElements(var MatrixElements: TMatrixElements);
var
  i: Integer;
begin
  for i := 0 to High(MatrixElements) do begin
    FreeAndNil(MatrixElements[i].Expr);
    FreeAndNil(MatrixElements[i].Evaluator);
  end;
  SetLength(MatrixElements,0);
end;

function rlE(J, K: TExprNode; dK: Integer): TExprNode;
{ Work out matrix element of arbitrary power of raising/lowering operator }
{ J, K are Ket variables; dK is change in K }
var
  JSq, Kp, Node: TExprNode;
  i: Integer;
begin
  Result := TIntegerNode.Create(1);
  JSq := EB([J, '+', 1, '*', J]);
  K := K.Duplicate;
  Kp := nil;
  for i := 1 to Abs(dK) do begin
    if dK > 0 then
      Kp := EB([K,'+',1])
    else
      Kp := EB([K,'-',1]);
    Kp := Kp.Simplify;
    Node := EB([K, '*', Kp]);
    Node := OperatorMinus.CreateExprNode(JSq.Duplicate, Node);
    Result := OperatorTimes.CreateExprNode(Result, Node);
    K.Free;
    K := Kp;
  end;
  Result := SqrtFunction.CreateExprNode(Result);
  JSq.Free;
  Kp.Free;
end;

function Halve(Q: TQuantumNumber):TExprNode;
begin
  Result := EN([Q, '/', 2]);
end;

function rlE(J: TExprNode; Kp, K: TQuantumNumber): TExprNode;
{ Work out matrix element of arbitrary power of raising/lowering operator }
var
  JSq, Node: TExprNode;
  dK: TQuantumNumber;
begin
  Result := TIntegerNode.Create(1);
  JSq := EB([J, '+', 1, '*', J]);
  dK := Sign(Kp-K)*2;
  while Kp <> K do begin
    Node := EN([Halve(K), '*', Halve(K+dK)]);
    Node := OperatorMinus.CreateExprNode(JSq.Duplicate, Node);
    Result := OperatorTimes.CreateExprNode(Result, Node);
    Inc(K, dK);
  end;
  Result := SqrtFunction.CreateExprNode(Result);
  JSq.Free;
end;

function rlxy(n: Integer; J, K: TExprNode; dK: Integer; Sign:Integer):TExprNode;
{ Helper function for matrix elements of Jx^n and Jy^n; K is Ket }
var
  Node, eK: TExprNode;
begin
  if (n < Abs(dK)) or Odd(n+dK) then
    Result := TIntegerNode.Create(0)
  else if n = Abs(dK) then
    { Easy case - equivalent to all raising/all lowering operator }
    if n = 0 then
      Result := TIntegerNode.Create(1)
    else begin
      Result := rlE(J,K,dK);
      Result := OperatorDivide.CreateExprNode(Result, EB([2, '^', n]));
      if (dK < 0) and Odd(n) and (Sign < 0) then
        Result := MonadicMinus.CreateExprNode(Result);
    end
  else begin
    { Need to consider combinations;
      e.g. Jx^2 can give DeltaK = 0, +/-2 Matrix elements }
      eK := EB([K,'+',1]);
      Result := EN([rlxy(n-1,J,eK,dK-1,Sign),'*',rlxy(1,J,K,1,Sign)]);
      eK.Free;
      eK := EB([K,'-',1]);
      Node := EN([rlxy(n-1,J,eK,dK+1,Sign),'*',rlxy(1,J,K,-1,Sign)]);
      Result := OperatorPlus.CreateExprNode(Result,Node);
      eK.Free;
  end;
end;

function rlxE(n: Integer; J, K: TExprNode; dK: TQuantumNumber):TExprNode;
{ Matrix elements of Jx^n; K is Ket }
begin
  Result := rlxy(n,J,K,dK,1);
end;

function rlyE(n: Integer; J, K: TExprNode; dK: TQuantumNumber):TExprNode;
{ Matrix elements of Jy^n; K is Ket; for odd powers the result is multiplied by i }
var
  Phase: Integer;
begin
  Result := rlxy(n,J,K,dK,-1);
  { Multiply by i^n }
  if Odd(n) then
    Phase := Minus(n+1) { Multiply by a bonus i to give a real result }
  else
    Phase := Minus(n);
  if Phase < 0 then
    Result := MonadicMinus.CreateExprNode(Result);
end;

function ReducedMatrixElementE(Rank: TQuantumNumber; J:TExprNode): TExprNode;
{ Return the reduced matrix element <J||T^k(J)||J> formed by
  coupling J with J k times. }
begin
//  case Rank of
//    2: Result := EN([EB([J, '+', 1, '*', J]), '*', EB([2, '*', J, '+', 1])]);
//    4: Result := EN([EB([2, '*', J, '-', 1, '*', 2, '*', J]),
//                     '*', EB([2, '*', J, '+', 1]), '*', EB([2, '*', J, '+', 2]),
//                     '*', EB([2, '*', J, '+', 3]), '/', 24]);
//  else
//    raise Exception.Create('ReducedMatrixElementE');
//  end;
//  Result := SqrtFunction.CreateExprNode(Result);
//  Exit;
  { Brown and Carrington 5.134 }
  Result := EN([ 2, '^', Rank div 2]);
  Result := EN([ Result,
                 '*', EB([J, '*', 2, '-', Rank div 2, 'factorial']),
                 '*', EN([Rank, 'factorial'])],[],1);
  Result := EN([EB([J, '*', 2, '+', Rank div 2, '+', 1, 'factorial']),
               '/', Result, 'sqrt']);
  Result := EN([Rank div 2, 'factorial', '*', Result]);
{ Specific results for reference:
<J||T^1(J)||J>=sqrt(J*(2*J+1)*(2*J+2))
<J||T^2(J)||J>=sqrt(J*(2*J-1)*(2*J+1)*(2*J+2)*(2*J+3))/(2*sqrt(3))
<J||T^3(J)||J>=sqrt(J*(2*J-1)*(2*J-2)*(2*J+1)*(2*J+2)*(2*J+3)*(2*J+4))/(4*sqrt(5))
<J||T^4(J)||J>=sqrt(J*(2*J-1)*(2*J-2)*(2*J-3)*(2*J+1)*(2*J+2)*(2*J+3)*(2*J+4)*(2*J+5))/(4*sqrt(35))
<J||T^5(J)||J>=sqrt(J*(2*J-1)*(2*J-2)*(2*J-3)*(2*J-4)*(2*J+1)*(2*J+2)*(2*J+3)*(2*J+4)*(2*J+5)*(2*J+6))/(24*sqrt(7))
<J||T^6(J)||J>=sqrt(J*(2*J-1)*(2*J-2)*(2*J-3)*(2*J-4)*(2*J-5)*(2*J+1)*(2*J+2)*(2*J+3)*(2*J+4)*(2*J+5)*(2*J+6)*(2*J+7))/(8*sqrt(462))
}
end;

end.
