unit Obs;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Graphics, Classes, Mixture, QuantumNumber, TextStrm, MNode, Units,
     MathSup, Contnrs, StaticField;

type

  TTakeWeight = (tkAll, tkObs, tkCalc);

  TObservationReaderProc = procedure(const S: string; ObsType: TObsType; TakeWeight: TTakeWeight) of object;

  TObservationsReader = class(TObservationsReaderBase)
  private
    FIncludedFrom: TObservationsReader;
  protected
    Obs: TObservation; { Current Observation }
    s: string;         { Current Token }
    LastAutoCD: Integer; { Number of observations after last AutoCD directive; only set in base Observations reader }
    LastObsType: TObsType;
    function FindMNodeStartingWith(Start: string; MNode: TMnode; MNodeClass: TMNodeClass): TMNode;
    function FindMoleculeStartingWith(Start: string; Mixture:TMixture): TMNode;
  public
    nErrors, nObs: Integer;
    Mixture: TMixture;
    DefaultMolecule: TMolecule;
    DefaultManifoldU, DefaultManifoldL: TManifold;
    StateU, StateL: TState;
    f: TTextStream;
    FileName: string;
    Take: TTakeWeight;
    QuantumNumberFormat: TQuantumNumberFormat;
    OverrideWeight, ScaleWeight: Double;
    IndexOffsetsL, IndexOffsetsU: TIntegerArray;
    Units: TResonanceUnits;
    FrequencyOffset, FrequencyScale: Double;
    DefaultColour: TColor;
    ObsType: TObsType;
    DefaultFlags: TObsFlags;
    NameOverride: string;
    InputPgopherVersion: Integer;
    Selections: TObjectList;
    ExField: TStaticFields;
    GlobalNQN: Integer;
    UpperMap, LowerMap: TIntegerArray;

    constructor Create; overload; virtual;
    constructor Create(AMixture: TMixture); overload;
    constructor Create(AReader: TObservationsReader); overload; virtual;
    destructor Destroy; override;
    procedure ResetDefaultMolecule;
    procedure ResetDefaultManifoldU;
    procedure ResetDefaultManifoldL;
    function DefaultStateU: TState;
    function DefaultStateL: TState;

    function ReadStream(AStream: TTextStream; AName: string):Boolean; overload;
    function ReadStream(AStream: TStream; AName: string):Boolean; overload;
    function ReadInteger(var a: Integer): Boolean; overload;
    function ReadInteger(var a: Smallint): Boolean; overload;
    function ReadInteger(var a: Shortint): Boolean; overload;
    function ReadQuantumNumber(var a: Integer; const si: string): Boolean; override;
    function ReadSymmetry(var Sym: Shortint; J: TQuantumNumber; Molecule: TMolecule; Adjust: Boolean = False): Boolean;
    function ReadInt(var a: Integer; const si: string): Boolean;
    function NextItem: string; override;
    procedure ReportError(es: string); override;
    function ReadOptionalManifold(Molecule: TMolecule): TManifold;
    function ReadOptionalMolecule: TMolecule;
    function ReadOneLine: Boolean; virtual;
    function DoReadOneLine: Boolean; virtual; abstract;
    function HandleDirective(var Handled: Boolean):Boolean; virtual;
    procedure AddObservation(AObs: TObservation);
    procedure ResetObs;
    procedure SwitchObsType(NewObsType: TObsType; ForceRebase: Boolean);
    function CheckSelected(const AObs: TObservation): Boolean;

    procedure AutoCommonDifference(Upper: Boolean);
    function BaseReader: TObservationsReader;

    property IncludedFrom: TObservationsReader read FIncludedFrom;
  end;

  TObservationsReaderClass = class of TObservationsReader;

var
  Observations: array of TObservation; { Global Variable for now }
  nObsFiles: Integer;        { Number of diferent files }
  LongestObsLabel: Integer;  { Length of longest observation description }

function AddToObservations(Mixture: TMixture; FileName:string; Stream: TStream = nil; ObsType: TObsType = obLine; TakeWeight: TTakeWeight = tkObs; ReaderClass: TObservationsReaderClass = nil; CurrentReader: TObservationsReader = nil; TrueName: string = ''): Boolean;
procedure ClearObservations;
procedure LoadObservations(Mixture: TMixture; Stream: TStream; FileName:string; ObsType:TObsType = obLine; TakeWeight: TTakeWeight = tkObs);
function FindIndex(Eigenvalues: TEigenvalues; Name: string; Lower: Boolean; LineNumber: Integer; const FileName: string; QuantumNumberFormat: TQuantumNumberFormat; ShowError:Boolean):Integer;
function FindEigenvaluesAndIndex(out Eigenvalues: TEigenvalues; Manifold: TManifold; J: TQuantumNumber;
                                 var Sym: Shortint; Index: Integer; Name: string;
                                 Lower: Boolean; LineNumber: Integer; const FileName: string;
                                 RemovedSpins: Byte; QuantumNumberFormat: TQuantumNumberFormat;
                                 ExField: TStaticFields):Integer;

function AutoCommonDifferenceRange(Upper, Sort: Boolean; First: Integer; var nObs: Integer; NonCD: Integer):Integer;

implementation

uses Math, SysUtils, JPLCat, Hitran, Profile, Progress, MNodeColour, PropHelp, SimpleQnos,
     Worker, UCompat, Global, ObsSelect, ImportJB95, Constrain, ImportACDD, ShellSort;

procedure ClearObservations;
begin
  SetLength(Observations, 0);
  ClearConstraints;
  nObsFiles := 0;
  LongestObsLabel := 0;
  PerformMNodeNotification(nil, mnObsCleared, -1);
end;

type

  TPgopherReader = class(TObservationsReader)
    BranchLabels: array of string;
    function DoReadOneLine: Boolean; override;
    function HandleDirective(var Handled: Boolean):Boolean; override;
  end;

  TOldPgopherReader = class(TObservationsReader)
    function DoReadOneLine: Boolean; override;
  end;

procedure SplitObsName(Name: string; out Pre, Upper, Lower, Post: string);
var
  e, Sep: Integer;
begin
  { Identify section with transition in - use presence of " - " to flag this }
  Sep := Pos(' - ', Name);
  if Sep <= 0 then
    Sep := Pos(#9'-'#9, Name);
  if (Sep <= 0) and (Pos(' -',Name) = Length(Name)-1) then
    Sep := Length(Name)-1;
  if (Sep <= 0) and (Pos('- ',Name) = 1) then begin
    Delete(Name, 1, 2);
    Name := Trim(Name);
    Sep := 1;
  end;
  e := Pos(':', Name);
  { Note that if no " - " state separator, the level name must be first }
  while (e < Sep) and (e > 0) do begin
    if Pre <> '' then
      Pre := Pre + ' : ';
    Pre := Pre + Trim(Copy(Name,1,e-1));
    Delete(Name, 1, e);
    Dec(Sep,e);
    e := Pos(':', Name);
  end;
  if Sep < e then begin
    Post := Trim(Copy(Name,e+1,Length(Name)));
    Delete(Name, e, Length(Name))
  end;
  if Sep <= 0 then begin
    Lower := Trim(Name);
    Upper := Lower; { Only one level specified - assume it is the right one }
  end else begin
    if (Sep = 1) and (Name[1] = '-') then
      Dec(Sep); { Special case - starts with "- " }
    Upper := Trim(Copy(Name,1,Sep-1));
    Lower := Trim(Copy(Name,Sep+2,Length(Name)));
  end;
end;

function FindIndex(Eigenvalues: TEigenvalues; Name: string; Lower: Boolean; LineNumber: Integer; const FileName: string; QuantumNumberFormat: TQuantumNumberFormat; ShowError:Boolean):Integer;
var
  e, Sep: Integer;
begin
  { Identify section with transition in - use presence of " - " to flag this }
  repeat
    Sep := Pos(' - ', Name);
    if Sep <= 0 then
      Sep := Pos(#9'-'#9, Name);
    if (Sep <= 0) and not Lower and (Pos(' -',Name) = Length(Name)-1) then
      Sep := Length(Name)-1;
    if (Sep <= 0) and Lower and (Pos('- ',Name) = 1) then begin
      Delete(Name, 1, 2);
      Name := Trim(Name);
    end;
    e := Pos(':', Name);
    if e <= 0 then
      Break;
    { Note that if no " - " state separator, the level name must be first }
    if Sep < e then
      Delete(Name, e, Length(Name))
    else
      Delete(Name, 1, e);
    Name := Trim(Name);
  until False;
  { Identify upper or lower state name }
  if Sep > 0 then
    if Lower then
      Delete(Name, 1, Sep+2)
    else
      Delete(Name, Sep, Length(Name));
  for Result := 1 to Eigenvalues.Size do
    if Eigenvalues.DominantBasis[Result].StateLabel(QuantumNumberFormat) = Name then
      Exit;
  Result := 0;
  if not ShowError then Exit;
  Writeln(Log);
  Writeln(Log, 'No match for ' + Name + ' on line number ' + IntToStr(LineNumber) + ' of ' + FileName + ' Possible matches:');
  for e := 1 to Eigenvalues.Size do
    Writeln(Log, Eigenvalues.DominantBasis[e].StateLabel(QuantumNumberFormat));
end;

function FindEigenvaluesAndIndex(out Eigenvalues: TEigenvalues; Manifold: TManifold; J: TQuantumNumber;
                                 var Sym: Shortint; { If negative, overwrite with found value }
                                 Index: Integer; Name: string;
                                 Lower: Boolean;
                                 LineNumber: Integer; const FileName: string;
                                 RemovedSpins: Byte; QuantumNumberFormat: TQuantumNumberFormat;
                                 ExField: TStaticFields):Integer;
  procedure Error(s: string);
  begin
    WriteLn(Log, s, ' on line number ', LineNumber, ' of ', FileName);
    Result := 0;
  end;
var
  s: String;
  iSym, SymMin, SymMax: Integer;
begin
  EigenValues := nil;
  if IsNonZero(ExField) then begin
    SymMin := 0;
    SymMax := 0;
  end else if Sym < 0 then begin
    { Search for symmetry }
    SymMin := 0;
    SymMax := Manifold.nSym-1;
    Index := -1;
  end else begin
    SymMin := Sym;
    SymMax := Sym;
  end;
  try
    for iSym := SymMin to SymMax do begin
      Eigenvalues := Manifold.Eigenvalues(J, iSym, [], ExField, True, plDefault, nil, True, RemovedSpins);
      Result := Index;
      if Result <= 0 then
        Result := FindIndex(Eigenvalues, Name, Lower, LineNumber, FileName, QuantumNumberFormat, SymMin = SymMax);
      if Result > 0 then Break;
    end;
    if (Result > 0) and (Result > Eigenvalues.Size) then begin
      if Lower then
        s := 'Lower'
      else
        s := 'Upper';
      Error(Format('%s eigenvalue not found (only %d present)', [s, Eigenvalues.Size]));
    end else if (Result <= 0) { and (SymMin <> SymMax) Why? } then
      Error('State not found')
    else if Sym < 0 then
      Sym := Eigenvalues.Sym;
  except
    on E:Exception do
      Error(E.Message);
  end;
end;

{ TObservationsReader }

constructor TObservationsReader.Create;
begin
  inherited;
  nErrors := 0;
  Obs.ManifoldU := DefaultManifoldU;
  Obs.ManifoldL := DefaultManifoldL;
end;

constructor TObservationsReader.Create(AMixture: TMixture);
begin
  Mixture := AMixture;
  ResetDefaultMolecule;
  OverrideWeight := -1;
  ScaleWeight := 1;
  Take := tkObs;
  Units := Mixture.Simulations[0].PlotUnits;
  ExField := Mixture.Simulations[0].ExField;
  FrequencyOffset := 0;
  FrequencyScale := 1;
  RemoveSpins := 0;
  DefaultColour := clDefault; { Copy from Manifold }
  QuantumNumberFormat := Mixture.QuantumNumberFormat;
  DefaultFlags := [];
  NameOverride := '';
  StateU := nil;
  StateL := nil;
  InputPgopherVersion := High(Integer); { By default, assume highest }
  BranchRank := 2; { Default is dipole transitions }
  DefaultSymmetry := -1; { No default }
  GlobalNQN := 0;
  UpperMap := nil;
  LowerMap := nil;
  Create;
end;

function CompareManifolds(MA, MB: TManifold): Integer;
begin
  if MA = MB then
    Result := 0
  else begin
    { Different manifolds - use text compare for consistent order }
    Result := MA.NodeIndex - MB.NodeIndex;
    if Result = 0 then
      Result := CompareText(MA.UniqueName, MB.UniqueName);
    if Result = 0 then
      raise NotYet;
  end;
end;

function CompareTextQNOs(const A, B: TObservation; OnUpper: Boolean): Integer;
var
  APre, AUn, ALn, APost: string;
  BPre, BUn, BLn, BPost: string;
begin
  SplitObsName(A.Name, APre, AUn, ALn, APost);
  SplitObsName(B.Name, BPre, BUn, BLn, BPost);
  if OnUpper then
    Result := CompareText(AUn,BUn)
  else
    Result := CompareText(ALn,BLn)
end;

function CompareUpper(const A, B: TObservation): Integer;
begin
  Result := CompareManifolds(A.ManifoldU, B.ManifoldU);
  if Result = 0 then
    Result := A.JU - B.JU;
  if Result = 0 then
    Result := A.SymU - B.SymU;
  if Result = 0 then
    if A.SymU < 0 then
      Result := CompareTextQNOs(A,B,True)
    else begin
      Result := A.IndexU - B.IndexU;
      if (Result = 0) and (A.IndexU <= 0) then
        Result := CompareTextQNOs(A,B,True);
    end;
end;

function CompareLower(const A, B: TObservation): Integer;
begin
  Result := CompareManifolds(A.ManifoldL, B.ManifoldL);
  if Result = 0 then
    Result := A.JL - B.JL;
  if Result = 0 then
    Result := A.SymL - B.SymL;
  if Result = 0 then
    if A.SymL < 0 then
      Result := CompareTextQNOs(A,B,False)
    else begin
      Result := A.IndexL - B.IndexL;
      if (Result = 0) and (A.IndexL <= 0) then
        Result := CompareTextQNOs(A,B,False);
    end;
end;

function CompareControl(const A, B: TObservation):Integer;
begin
  Result := Ord(FieldsDiff(A.ExField, B.ExField));
  if Result = 0 then
    Result := A.RemoveSpins - B.RemoveSpins;
  if Result = 0 then
    Result := Ord(A.QuantumNumberFormat) - Ord(B.QuantumNumberFormat);
  if Result = 0 then
    Result := Ord(A.ObsType) - Ord(B.ObsType);
end;

function CompareObservations(const A, B: TObservation; UpperFirst: Boolean):Integer;
begin
  Result := CompareControl(A,B);
  if Result = 0 then
    if UpperFirst then
      Result := CompareUpper(A,B)
    else
      Result := CompareLower(A,B);
  if Result = 0 then
    if not UpperFirst then
      Result := CompareUpper(A,B)
    else
      Result := CompareLower(A,B);
end;

type
  TObsShellSorter = class(TShellSorter)
    OnUpper: Boolean;
    Offset: Integer;
    function InOrder(I,J:Integer):Boolean; override;
    function Compare(const A, B: TObservation):Integer;
    procedure Swap(I,J: Integer); override;
  end;

{ TObsShellSorter }

function TObsShellSorter.Compare(const A, B: TObservation): Integer;
begin
  Result := CompareObservations(A, B, OnUpper);
end;

function TObsShellSorter.InOrder(I, J: Integer): Boolean;
begin
  Result := Compare(Observations[I+Offset], Observations[J+Offset]) <= 0;
end;

procedure TObsShellSorter.Swap(I, J: Integer);
begin
  SwapValues(Observations[I+Offset], Observations[J+Offset], SizeOf(TObservation));
end;

procedure TObservationsReader.AutoCommonDifference(Upper : Boolean);
var
  NonCD: Integer;
  CDstatus: string;
begin
  NonCD := 0; { default to force error }
  if f.GetFreeItem then begin
    { Possible actions on transitions that do not involve common differences:
      Ignore - Ignore the transitions
      Force - Display an error ( default )
      Include - Include in obseravtions as a normal transition
    }
    CDstatus := LowerCase(f.Token);
    if CDstatus = 'ignore' then
      NonCD := -1
    else if CDstatus = 'force' then
      NonCD := 0
    else if CDstatus = 'include' then
      NonCD := 1
    else
      ReportError('Common difference action not recognized:'+f.Token);
  end;
  nErrors := nErrors + AutoCommonDifferenceRange(Upper, True, BaseReader.LastAutoCD, nObs, NonCD);
  BaseReader.LastAutoCD := nObs;
end;

function MergeObservations(const A, B: TObservation):TObservation;
var
  Wt: Double;
  APre, AUn, ALn, APost: string;
  BPre, BUn, BLn, BPost: string;
begin
  { Weighted average of two otherwise identical observations }
  Result := A;
  Wt := Sqrt(1/(1/Sqr(A.Weight) + 1/Sqr(B.Weight)));
  Result.Frequency := (A.Frequency/Sqr(A.Weight) + B.Frequency/Sqr(B.Weight))*Sqr(Wt);
  Result.Weight := Wt;
  SplitObsName(A.Name, APre, AUn, ALn, APost);
  SplitObsName(B.Name, BPre, BUn, BLn, BPost);
  Result.Name := '';
  if Length(APre) > 0 then
    Result.Name := APre+' : ';
  if Length(AUn)+Length(ALn) > 0 then
    Result.Name := Result.Name + AUn+' - '+ALn+' : ';
  if APost <> BPost then
    Result.Name := Result.Name + APost+'/'+BPost
  else
    Result.Name := Result.Name + APost;
end;

function MakeCombinationDifference(const U, L: TObservation; out Reversed: Boolean; Upper: Boolean): TObservation;
var
  O: TObservation;
  procedure CopyFields(const U, L: TObservation);
  var
    UPre, UUn, ULn, UPost: string;
    LPre, LUn, LLn, LPost: string;
  begin
    SplitObsName(U.Name, UPre, UUn, ULn, UPost);
    SplitObsName(L.Name, LPre, LUn, LLn, LPost);
    if Upper then begin
      O := U;
      O.ManifoldL := L.ManifoldU;
      O.JL := L.JU;
      O.SymL := L.SymU;
      O.IndexL := L.IndexU;
      O.Name := '';
      if Length(UPre) + Length(LPre) > 0 then
        O.Name := UPre+'/'+LPre+' : ';
      if Length(ULn)+Length(LLn) > 0 then
        O.Name := O.Name + UUn+' - '+LUn+' : ';
      if UPost <> LPost then
        O.Name := O.Name + UPost+'/'+LPost
      else
        O.Name := O.Name + UPost;
      O.Name := O.Name + ' '+IntToStr(L.LineNumber)+':'+L.FileName;
    end else begin
      O := L;
      O.ManifoldU := U.ManifoldL;
      O.JU := U.JL;
      O.SymU := U.SymL;
      O.IndexU := U.IndexL;
      O.Name := '';
      if Length(UPre) + Length(LPre) > 0 then
        O.Name := UPre+'/'+LPre+' : ';
      if Length(ULn)+Length(LLn) > 0 then
        O.Name := O.Name + ULn+' - '+LLn+' : ';
      if UPost <> LPost then
        O.Name := O.Name + UPost+'/'+LPost
      else
        O.Name := O.Name + UPost;
      O.Name := O.Name + ' '+IntToStr(U.LineNumber)+':'+U.FileName;
    end;
    O.Weight := Sqrt(Sqr(U.Weight)+Sqr(L.Weight));
    O.Frequency := Abs(U.Frequency - L.Frequency);
    O.ObsType := obLine;
  end;
begin
  Reversed := (U.Frequency < L.Frequency) xor Upper;
  if Reversed then
    CopyFields(U, L)
  else
    CopyFields(L,U);
  Result := O; { Copy required to avoid loosing values }
end;

function AutoCommonDifferenceRange(Upper, Sort: Boolean; First: Integer; var nObs: Integer; NonCD: Integer):Integer;
  { Reduce First..nObs-1 observations to common differences }
  { Assumed appropriately sorted on entry }
  { NonCD indicates possible actions on transitions that do not involve common differences:
    -1: Ignore - Ignore the transitions
     0: Force - Display an error
     1: Include - Include in obseravtions as a normal transition
  }

  procedure ReportError(es: string; n: Integer);
  begin
    Inc(Result);
    WriteLn(Log, 'Error on line ', Observations[n].LineNumber, ' of ', Observations[n].FileName, ': ', es);
  end;
var
  LastObs: TObservation;
  LastFile: string;
  so, de: Integer; { Read and Write pointer into observations }
  LastLine, nNotCD: Integer;
  procedure LastObsNotCD;
  begin
    { This observation cannot be converted to common difference }
    if NonCD = 0 then
      ReportError('Not common difference: '+IntToStr(LastObs.LineNumber)+' of '+LastObs.FileName, so)
    else if NonCD < 0 then begin
      { Discard observation }
    end else if de <> so then begin
      { Keep as normal observation }
      Observations[de] := LastObs;
      Inc(de);
    end;
    LastFile := LastObs.FileName;
    LastLine := LastObs.LineNumber;
    Inc(nNotCD);
  end;
var
  IsCD, Reversed, InCD, LastReversed, FirstTime: Boolean;
begin
  Result := 0;
  if nObs = First then
    Exit; { No observations }
  if Sort then
    with TObsShellSorter.Create do
      try
        OnUpper := not Upper;
        Offset := First;
        Sort(nObs-Offset)
      finally
        Free;
      end;
  de := First;
  so := de+1;
  { Pass 1 - merge observations with identical upper and lower states }
  while so < nObs do begin
    if CompareObservations(Observations[so-1], Observations[so], True) = 0 then begin
      Observations[de] := MergeObservations(Observations[so-1], Observations[so]);
      Inc(so);
    end else if so <> de+1 then
      Observations[de+1] := Observations[so];
    Inc(so);
    Inc(de);
  end;
  nObs := de+1;
  if nObs - First = 1 then
    ReportError('Common difference expected', First);
  { Pass 2 - merge observations identical in required (upper or lower) state }
  de := First;
  so := de;
  InCD := True; { Prevents error on first line }
  nNotCD := 0; { Number of observations that are not common differences; start at second observation }
  FirstTime := True;
  while so < nObs do begin
    IsCD := not FirstTime and (CompareControl(LastObs, Observations[so]) = 0);
    if IsCD then
      if Upper then
        IsCD := IsCD and (CompareLower(LastObs, Observations[so]) = 0)
      else
        IsCD := IsCD and (CompareUpper(LastObs, Observations[so]) = 0);
    if IsCD then begin
      if Length(LastObs.Correlation) + Length(Observations[so].Correlation) > 0 then
        ReportError('Common differences must not be correlated', so);
      Observations[de] := MakeCombinationDifference(LastObs, Observations[so], Reversed, Upper);
      if InCD then
      { For two common differences in a row, need to consider correlation between observations
        Theory:
        For Obs' = L Obs (L is a matrix; elements will be 1, 0 or -1 in this case )
        then the variance-covariance matrix is related by:
        V' = L V Transpose(L)
        so covariance is
        Vij = Lik Vkk Ljk assuming orignial line positions are uncorrelated.
        covariance thus depends only on variance of line in common, with sign dependent on the relative
        signs of difference taken. }
        with Observations[de-1] do begin
          if Length(Correlation) = 0 then
            SetLength(Correlation,1);
          Correlation[0] := Sqr(LastObs.Weight)/(Observations[de].Weight*Weight);
          if LastReversed = Reversed then
            Correlation[0] := -Correlation[0];
        end;
      Inc(de);
      LastReversed := Reversed;
      InCD := True;
    end else begin
      { Not merging with previous observation }
      if not InCD then
        LastObsNotCD;
      InCD := False;
    end;
    LastObs := Observations[so];
    Inc(so);
    FirstTime := False;
  end;
  if not InCD then
    LastObsNotCD;
  if (NonCD <> 0) and (nNotCD <> 0) then begin
    WriteLn(Log, IntToStr(nNotCD), ' observation(s) not converted to common differences');
    WriteLn(Log, 'Last unconverted observation at line ', LastLine, ' of ', LastFile);
  end;
  nObs := de;
end;

function TObservationsReader.BaseReader: TObservationsReader;
begin
  Result := Self;
  while Assigned(Result.IncludedFrom) do
    Result := Result.IncludedFrom;
end;

function TObservationsReader.CheckSelected(const AObs: TObservation): Boolean;
begin
  Result := CheckSelections(AObs, Selections);
  if Assigned(IncludedFrom) then
    Result := Result and IncludedFrom.CheckSelected(AObs);
end;

constructor TObservationsReader.Create(AReader: TObservationsReader);
begin
  Mixture := AReader.Mixture;
  DefaultMolecule := AReader.DefaultMolecule;
  DefaultManifoldU := AReader.DefaultManifoldU;
  DefaultManifoldL := AReader.DefaultManifoldL;
  OverrideWeight := AReader.OverrideWeight;
  ScaleWeight := AReader.ScaleWeight;
  IndexOffsetsU := Copy(AReader.IndexOffsetsU);
  IndexOffsetsL := Copy(AReader.IndexOffsetsL);
  Take := AReader.Take;
  Units := AReader.Units;
  ExField := AReader.ExField;
  FrequencyOffset := AReader.FrequencyOffset;
  FrequencyScale := AReader.FrequencyScale;
  RemoveSpins := AReader.RemoveSpins;
  DefaultColour := AReader.DefaultColour;
  ObsType := AReader.ObsType;
  QuantumNumberFormat := AReader.QuantumNumberFormat;
  DefaultFlags := AReader.DefaultFlags;
  NameOverride := AReader.NameOverride;
  StateU := AReader.StateU;
  StateL := AReader.StateL;
  InputPgopherVersion := AReader.InputPgopherVersion;
  BranchRank := AReader.BranchRank;
  DefaultSymmetry := AReader.DefaultSymmetry;
  GlobalNQN := AReader.GlobalNQN;
  UpperMap := Copy(AReader.UpperMap);
  LowerMap := Copy(AReader.LowerMap);
  Create;
  FIncludedFrom := AReader;
end;

function TObservationsReader.DefaultStateL: TState;
begin
  if Assigned(StateL) then
    Result := StateL
  else if Assigned(DefaultManifoldL) then
    Result := DefaultManifoldL.State[0]
  else
    Result := nil;
end;

function TObservationsReader.DefaultStateU: TState;
begin
  if Assigned(StateU) then
    Result := StateU
  else if Assigned(DefaultManifoldU) then
    Result := DefaultManifoldU.State[0]
  else
    Result := nil;
end;

destructor TObservationsReader.Destroy;
begin
  FreeAndNil(Selections);
  inherited;
end;

function TObservationsReader.FindMNodeStartingWith(Start: string; MNode: TMnode; MNodeClass: TMNodeClass): TMNode;
var
  i: Integer;
begin
  for i := 0 to MNode.ComponentCount-1 do begin
    Result := MNode.Items[i];
    if (Result is MNodeClass) and (Pos(Start,Result.Name) = 1)
       and (Length(Result.Name) > Length(Start)) and (Result.Name[Length(Start)+1] = ' ') then
      Exit;
  end;
  Result := nil;
end;

function TObservationsReader.FindMoleculeStartingWith(Start: string; Mixture: TMixture): TMNode;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to Mixture.nSpecies - 1 do begin
    Result := FindMNodeStartingWith(Start,Mixture.Species[i],TMolecule);
    if Assigned(Result) then
      Exit;
  end;
end;

procedure TObservationsReader.ResetDefaultManifoldU;
begin
  if Assigned(DefaultMolecule) and (DefaultMolecule.nManifolds <=2) then
    with DefaultMolecule do
      if (nManifolds = 1) { Take likely default manifold }
       or not Manifold[0].Initial
       or (Manifold[0].LowestStateOrigin > Manifold[1].LowestStateOrigin) then
        DefaultManifoldU := Manifold[0]
      else
        DefaultManifoldU := Manifold[1]
  else
    DefaultManifoldU := nil;
end;

procedure TObservationsReader.ResetDefaultManifoldL;
begin
  if Assigned(DefaultMolecule) and (DefaultMolecule.nManifolds <=2) then
    with DefaultMolecule do
      if (nManifolds = 1) { Take likely default manifold }
       or Manifold[0].Initial
       or (Manifold[1].LowestStateOrigin > Manifold[0].LowestStateOrigin) then
        DefaultManifoldL := Manifold[0]
      else
        DefaultManifoldL := Manifold[1]
  else
    DefaultManifoldL := nil;
end;

procedure TObservationsReader.ResetDefaultMolecule;
begin
  if (Mixture.nMolecules = 1) and (Mixture.Species[0].Count = 1) then
    DefaultMolecule := Mixture.Species[0].Isotopomer[0] { Default to only molecule }
  else
    DefaultMolecule := nil;
  ResetDefaultManifoldU;
  ResetDefaultManifoldL;
end;

procedure TObservationsReader.ResetObs;
begin
  Obs.RemoveSpins := RemoveSpins;
  Obs.Colour := DefaultColour;
  Obs.ObsType := ObsType;
  Obs.Calculated := 0;
  Obs.Flags := DefaultFlags - [ofCalculatedSet];
  Obs.QuantumNumberFormat := QuantumNumberFormat;
  Obs.Name := '';
  Obs.ExField := ExField;
  if DefaultStateU <> nil then
    Obs.ManifoldU := DefaultStateU.Manifold
  else
    Obs.ManifoldU := nil;
  if DefaultStateL <> nil then
    Obs.ManifoldL := DefaultStateL.Manifold
  else
    Obs.ManifoldL := nil;
  SetLength(Obs.Correlation,0);
end;

procedure TObservationsReader.SwitchObsType(NewObsType: TObsType; ForceRebase: Boolean);
begin
  ObsType := NewObsType;
  if BaseReader.LastObsType in [obCDLower, obCDUpper] then begin
    nErrors := nErrors + AutoCommonDifferenceRange(BaseReader.LastObsType = obCDUpper, False, BaseReader.LastAutoCD, nObs, 0);
    BaseReader.LastAutoCD := nObs;
  end;
  if ForceRebase then
    BaseReader.LastAutoCD := nObs;
end;

procedure TObservationsReader.ReportError(es: string);
begin
  Inc(nErrors);
  WriteLn(Log, 'Error on line ', f.Line, ' of ', FileName, ': ', es);
end;

function TObservationsReader.NextItem: string;
begin
  if not f.GetFreeItem then
    ReportError('Not enough items on line');
  Result := f.Token;
end;

function TObservationsReader.ReadInt(var a: Integer; const si: string): Boolean;
begin
  Result := not TryStrToInt(si, a);
  if Result then
    ReportError('Bad integer: '+ si);
end;

function TObservationsReader.ReadInteger(var a: Integer): Boolean;
var
  si: string;
begin
  si := NextItem;
  Result := ReadInt(a, si);
end;

function TObservationsReader.ReadInteger(var a: Smallint): Boolean;
var
  n: Integer;
begin
  Result := ReadInteger(n);
  if not Result and (n >= Low(a)) and (n <= High(a)) then
    a := n
  else if not Result then begin
    Result := True;
    ReportError('Bad small integer:'+IntToStr(n));
  end;
end;

function TObservationsReader.ReadInteger(var a: Shortint): Boolean;
var
  n: Integer;
begin
  Result := ReadInteger(n);
  if not Result and (n >= Low(a)) and (n <= High(a)) then
    a := n
  else if not Result then begin
    Result := True;
    ReportError('Bad short integer:'+IntToStr(n));
  end;
end;

function TObservationsReader.ReadSymmetry(var Sym: Shortint; J: TQuantumNumber; Molecule: TMolecule; Adjust: Boolean): Boolean;
var
  si: string;
  sy: Integer;
begin
  Result := False;
  si := NextItem;
  if TryStrToInt(si, sy) then begin
    Result := sy > Molecule.LastSym;
    if Result then
      ReportError('Symmetry number not in range 0..'+IntToStr(Molecule.LastSym))
    else
      if Adjust then
        Sym := Molecule.FixSym(J,sy,InputPgopherVersion)
      else
        Sym := sy;
    Exit;
  end;
  if Molecule.TryStrToAnySymName(J,si,sy) then begin
    Sym := sy;
    Exit;
  end;
  Result := True;
  ReportError('Bad symmetry: ' + si);
end;

function TObservationsReader.ReadStream(AStream: TTextStream; AName: string): Boolean;
var
  Cyc: Integer;
begin
  PushProgress(ExtractFileName(AName));
  try
    SetProgress(0,AStream.Size);
    f := AStream;
    FileName := AName;
    nObs := Length(Observations);
    Cyc := 0;
    AStream.Comments := True;
    AStream.CommentChars := AStream.CommentChars-['/'];
    while not f.AtEndOfFile do begin
      if ReadOneLine then
        Break;
      f.NextLine;
      Inc(Cyc);
      if Cyc > 100 then begin
        SetProgress(AStream.Position);
        Cyc := 0;
      end;
    end;
    if Assigned(IncludedFrom) then
      SwitchObsType(IncludedFrom.ObsType, False)
    else
      SwitchObsType(ObsType, False);
    BaseReader.LastObsType := obLine;
    SetLength(Observations, nObs);
    Result := nErrors > 0;
  finally
    PopProgress;
  end;
end;

threadvar
  IncludeFileDepth: Integer;

var
  LoadingObservationsTime: string = 'Loading Observations';

function AddToObservations(
  Mixture: TMixture; FileName:string;
  Stream: TStream = nil;
  ObsType: TObsType = obLine;
  TakeWeight: TTakeWeight = tkObs;
  ReaderClass: TObservationsReaderClass = nil;
  CurrentReader: TObservationsReader = nil;
  TrueName: string = ''): Boolean;
var
  Reader: TObservationsReader;
  FileStream: TFileStream;
  Ext: string;
begin
  DoDirSeparators(FileName);
  if TrueName = '' then
    TrueName := FileName;
  Ext := LowerCase(ExtractFileExt(FileName));
  if ReaderClass = nil then
    if (Ext = '.cat') or (Ext = '.mrg') then
      ReaderClass := TJPLCatReader
    else if Ext = '.par' then
      ReaderClass := THITRANReader
    else if Ext = '.acdd9' then
      ReaderClass := TACDD9Reader
    else
      ReaderClass := TPgopherReader;
  if Assigned(CurrentReader) then
    Reader := ReaderClass.Create(CurrentReader)
  else
    Reader := ReaderClass.Create(Mixture);
  Reader.ObsType := ObsType;
  Reader.Take := TakeWeight;
  Result := True;
  try
    if IncludeFileDepth = 0 then
      TStart(LoadingObservationsTime);
    Inc(nObsFiles);
    Inc(IncludeFileDepth);
    if IncludeFileDepth > 20 then
      raise EReportException.Create('Include files nested more than 20 deep');
    if Assigned(Stream) then
      Result := Reader.ReadStream(Stream, TrueName)
    else begin
      FileStream := TFileStream.Create(FileName,fmOpenRead or fmShareDenyWrite);
      try
        Result := Reader.ReadStream(FileStream, TrueName);
      finally
        FileStream.Free;
      end;
    end;
  finally
    Dec(IncludeFileDepth);
    Reader.Free;
    if IncludeFileDepth = 0 then
      TEnd(LoadingObservationsTime);
  end
end;

function TObservationsReader.ReadStream(AStream: TStream; AName: string): Boolean;
var
  TextStream: TTextStream;
begin
  if AStream is TTextStream then
    Result := ReadStream(AStream as TTextStream, AName)
  else begin
    TextStream := TTextStream.Create(AStream);
    try
      Result := ReadStream(TextStream, AName);
    finally
      TextStream.Free;
    end;
  end;
end;

function TObservationsReader.ReadOneLine: Boolean;
var
  Handled: Boolean;
  {$IFDEF FPC}
  Frames: PPointer;
  FrameCount, FrameNumber: Integer;
  {$ENDIF}
begin
  Result := False;
  while not f.GetFreeItem do
    if f.AtEndOfLine then
      Exit; { Blank line }
  if f.token[1] in ['/', '#', ';'] then
    Exit; { comment }
  s := f.Token;
  Handled := True;
  try
    Result := HandleDirective(Handled);
    if Handled then Exit;
    ResetObs;
    if DoReadOneLine then
      AddObservation(Obs);
  except
    on E:Exception do begin
      if not (E is EReportException) then begin
        {$IFDEF FPC}
        WriteLn(Log,E.Message,' at:');
        Writeln(Log,BackTraceStrFunc(ExceptAddr));
        FrameCount:=ExceptFrameCount;
        Frames:=ExceptFrames;
        for FrameNumber := 0 to FrameCount-1 do
          WriteLn(Log,BackTraceStrFunc(Frames[FrameNumber]));
        Writeln(Log);
        {$ENDIF}
        { TODO : Delphi }
      end;
      ReportError('Exception:'+E.Message)
    end;
  end;
end;

function TObservationsReader.HandleDirective(var Handled: Boolean): Boolean;
var
  Cmd: string;
  procedure ReadFile(IncName: string; ReaderClass: TObservationsReaderClass);
  var
    Stream: TStream;
    DirName, TrueName: string;
  begin
    SetLength(Observations, nObs);
    if IncName = '*' then
      IncName := '';
    try
      Stream := nil;
      if IncName = '' then begin
        Stream := f;
        TrueName := FileName;
      end else if not (IncName[1] in ['\','/']) then begin
        DirName := ExtractFileDir(FileName);
        if DirName <> '' then
          IncName :=  DirName + PathDelim + IncName;
        TrueName := IncName;
      end;
      if AddToObservations(Mixture, IncName, Stream, ObsType, Take, ReaderClass, Self, TrueName) then
        Inc(nErrors);
    except
      on E: Exception do
        ReportError(E.Message);
    end;
    nObs := Length(Observations);
  end;
  procedure ReadIndexOffsets(var IndexOffsets: TIntegerArray);
  var
    i, Offset: Integer;
  begin
    i := 0;
    while f.GetFreeItem do
      if TryStrToInt(f.Token,Offset) then begin
        if i > High(IndexOffsets) then begin
          SetLength(IndexOffsets, i+1);
          IndexOffsets[i] := Offset;
        end else
          Inc(IndexOffsets[i],Offset);
        Inc(i);
      end else if LowerCase(f.Token) = 'search' then begin
        SetLength(IndexOffsets, 1);
        IndexOffsets[0] := MaxInt;
        Exit;
      end else if LowerCase(f.Token) = 'searchall' then begin
        SetLength(IndexOffsets, 1);
        IndexOffsets[0] := -MaxInt;
        Exit;
      end;
    SetLength(IndexOffsets, i);
  end;
  function ReadManifold(var Manifold: TManifold):Boolean;
  begin
    Result := True;
    if not f.GetFreeItem then
      Exit;
    s := f.Token;
    if not Assigned(DefaultMolecule) then
      ReportError('Default molecule must be assigned')
    else begin
      Manifold := ReadOptionalManifold(DefaultMolecule);
      if (Manifold = nil) and (s <> 'none') then
        ReportError('Unknown Manifold:'+s)
      else
        Result := False;
    end;
  end;
  function ReadState(const State:string; Manifold:TManifold):TState;
  begin
    Result := nil;
    if not Assigned(Manifold) then
      ReportError('Manifold must be specified for state '+State)
    else begin
      Result := Manifold.StateWithName(State);
      if not Assigned(Result) then
        ReportError('State not found:'+State);
    end;
  end;
  procedure LoadArray(var Map: TIntegerArray);
  var
    i: Integer;
    s: String;
  begin
    if f.GetFreeItem then begin
      i := f.IntegerToken;
      s := f.NextString(IntToStr(i));
      AddToMap(Map, i, s);
    end else
      SetLength(Map, 0);
  end;
var
  Item: string;
  Value: Double;
  nO, Offset: Integer;
  OldNQN: Integer;
begin
  Result := False;
  Cmd := LowerCase(s);
  if (s = '-1') or (Cmd = 'stop') or (Cmd = 'end') then
    Result := True
  else if Cmd = 'include' then
    if f.GetFreeItem then
      ReadFile(f.Token, nil)
    else
      ReadFile('', nil)
  else if Cmd = 'includeoldpgopher' then begin
    if f.GetFreeItem then
      ReadFile(f.Token, TOldPgopherReader)
    else
      ReadFile('', TOldPgopherReader);
  end else if Cmd = 'includejb95' then begin
    if f.GetFreeItem then
      ReadFile(f.Token, TJJB95Reader)
    else
      ReadFile('', TJJB95Reader);
  end else if Cmd = 'includespfit' then begin
    Item := f.NextString('');
    OldNQN := GlobalNQN;
    GlobalNQN := f.NextInteger(0);
    if Abs(GlobalNQN) > 10 then
      ReportError('Bad value for NQN for spfit reader:'+IntToStr(GlobalNQN))
    else
      ReadFile(Item, TSPFITReader);
    GlobalNQN := OldNQN;
  end else if (Cmd = 'nqn') or (Cmd = 'nquantumnumbers') then
    GlobalNQN := f.NextInteger(0)
  else if Cmd = 'molecule' then begin
    if f.GetFreeItem then
      s := f.Token
    else
      s := '';
    DefaultMolecule := Mixture.FindMolecule(s);
    if (DefaultMolecule = nil) and (s <> 'none') then begin
      ReportError('Unknown Molecule:'+s);
      ResetDefaultMolecule;
    end else begin
      ResetDefaultManifoldU;
      ResetDefaultManifoldL;
    end;
  end else if Cmd = 'uppermanifold' then begin
    if ReadManifold(DefaultManifoldU) then
        ResetDefaultManifoldU;
  end else if Cmd = 'lowermanifold' then begin
    if ReadManifold(DefaultManifoldL) then
        ResetDefaultManifoldL;
  end else if (Cmd = 'overrideweight') or (Cmd = 'overridestddev') then
    if f.GetFreeItem then
      OverrideWeight := f.RealToken
    else
      OverrideWeight := -1
  else if (Cmd = 'scaleweights') or (Cmd = 'scalestddev') then
    if f.GetFreeItem then
      ScaleWeight := f.RealToken
    else
      ScaleWeight := 1
  else if Cmd = 'frequencyoffset' then
    if f.GetFreeItem then
      FrequencyOffset := f.RealToken { override frequency offset }
    else
      FrequencyOffset := 0
  else if Cmd = 'offsetfrequencyoffset' then
    if f.GetFreeItem then
      FrequencyOffset := FrequencyOffset + f.RealToken { add to frequency offset }
    else
      FrequencyOffset := 0
  else if Cmd = 'frequencyscale' then
    if f.GetFreeItem then
      FrequencyScale := f.RealToken
    else
      FrequencyScale := 1
  else if Cmd = 'takeall' then
    Take := tkAll
  else if Cmd = 'takeobs' then
    Take := tkObs
  else if Cmd = 'takecalc' then
    Take := tkCalc
  else if Cmd = 'upperindexoffsets' then
    ReadIndexOffsets(IndexOffsetsU)
  else if Cmd = 'lowerindexoffsets' then
    ReadIndexOffsets(IndexOffsetsL)
  else if Cmd = 'indexoffsets' then begin
    ReadIndexOffsets(IndexOffsetsU);
    IndexOffsetsL := Copy(IndexOffsetsU);
  end else if Cmd = 'units' then begin
    if f.GetFreeItem then
      Units := StrToUnits(f.Token)
    else
      Units := Mixture.Units;
  end else if Cmd = 'efield' then begin
    if f.GetFreeItem then
      ExField.E := f.RealToken
    else
      ExField.E := Mixture.Simulations[0].ExField.E;
  end else if Cmd = 'bfield' then begin
    if f.GetFreeItem then
      ExField.B := f.RealToken
    else
      ExField.B := Mixture.Simulations[0].ExField.B;
  end else if Cmd = 'angle' then begin
    if f.GetFreeItem then
      ExField.Angle := f.RealToken
    else
      ExField.Angle := Mixture.Simulations[0].ExField.Angle;
  end else if Cmd = 'removespins' then begin
    if f.GetFreeItem then
      RemoveSpins := RemoveSpins + f.IntegerToken
    else
      RemoveSpins := 0
  end else if (Cmd = 'colour') or (Cmd = 'color') then begin
    if f.GetFreeItem then
      DefaultColour := StrToColour(f.Token,False)
    else
      DefaultColour := clDefault;
  end else if Cmd = 'quantumnumberformat' then begin
    if f.GetFreeItem then
      if f.Token = '2J' then
        QuantumNumberFormat := qft2J
      else if f.Token = 'J' then
        QuantumNumberFormat := qftJ
      else if f.Token = 'J+1/2' then
        QuantumNumberFormat := qftJph
      else if f.Token = 'J-1/2' then
        QuantumNumberFormat := qftJmh
      else
        ReportError('Bad QuantumNumberFormat:'+f.Token)
    else
      ReportError('QuantumNumberFormat not specified');
  end else if Cmd = 'regeneratelabels' then begin
    if f.GetFreeItem and not StrToBoolean(f.Token) then
      DefaultFlags := DefaultFlags - [ofRegenerateLabel]
    else
      DefaultFlags := DefaultFlags + [ofRegenerateLabel]
  end else if Cmd = 'nameoverride' then
    NameOverride := Trim(F.RestOfLine)
  else if Cmd = 'upperstate' then begin
    Item := Trim(f.RestOfLine);
    if Item <> '' then
      StateU := ReadState(Item,DefaultManifoldU)
    else
      StateU := nil;
  end else if Cmd = 'lowerstate' then begin
    Item := Trim(f.RestOfLine);
    if Item <> '' then
      StateL := ReadState(Item,DefaultManifoldL)
    else
      StateL := nil;
  end else if Cmd = 'pgopherversion' then begin
    if f.GetFreeItem then
      InputPgopherVersion := DecodePgopherVersion(f.Token)
    else
      InputPgopherVersion := High(Integer);
  end else if Copy(Cmd,1,9) = 'constrain' then begin
    AddConstraint(Trim(f.RestOfLine),Self,Mixture,GlobalOverlays);
  end else if Copy(Cmd,1,6) = 'select' then begin
    Item := Trim(f.RestOfLine);
    if Item <> '' then begin
      Item := AddSelection(Selections, Item, QuantumNumberFormat);
      if Item <> '' then
        ReportError(Item);
    end else
      FreeAndNil(Selections);
  end else if Cmd = 'branchrank' then begin
    if not (f.GetFreeItem and not ReadQuantumNumber(BranchRank, f.Token)) then
      BranchRank := 2
  end else if Cmd = 'symmetry' then
    if f.GetFreeItem then begin
      if not Assigned(DefaultMolecule) then
        raise EReportException.Create('DefaultMolecule required for DefaultSymmetry');
      if not DefaultMolecule.TryStrToJAdjustedSymName(0,f.Token,DefaultSymmetry) then
        raise EReportException.Create('Symmetry not recognized:'+f.Token);
    end else
      DefaultSymmetry := -1
  else if Cmd = 'line' then
    SwitchObsType(obLine, False)
  else if Cmd = 'intensity' then
    SwitchObsType(obIntensity, False)
  else if Cmd = 'elower' then
    SwitchObsType(obELower, False)
  else if Cmd = 'eupper' then
    SwitchObsType(obEUpper, False)
  else if Cmd = 'cdlower' then
    SwitchObsType(obCDLower, True)
  else if Cmd = 'cdupper' then
    SwitchObsType(obCDUpper, True)
  else if Cmd = 'cdstart' then
    SwitchObsType(ObsType, True) { Force purge of any current common differences }
  else if (Cmd = 'autocdupper') or (Cmd = 'autocdlower') then
    AutoCommonDifference(Cmd = 'autocdupper')
  else if Cmd = 'correlation' then begin
    SwitchObsType(ObsType, True); { Force purge of any current common differences }
    Value := f.NextDouble(0.5);
    Offset := f.NextInteger(-1);
    if Value <> 0 then begin
      if Offset < 0 then
        nO := nObs-1+Offset
      else
        nO := nObs-1; { Most recent observation correlated with future observation }
      Offset := Abs(Offset);
      with Observations[nO] do begin
        SetLength(Correlation, Max(Length(Correlation), Offset));
        if (Correlation[Offset-1] <> 0) and ((Offset <> 1) or not (BaseReader.LastObsType in [obCDLower, obCDUpper])) then
          ReportError('Duplicate correlation');
        Correlation[Offset-1] := Value;
      end;
    end;
  end else if s = 'mapupper' then begin
    LoadArray(UpperMap);
    Handled := True;
  end else if s = 'maplower' then begin
    LoadArray(LowerMap);
    Handled := True;
  end else
    Handled := False;
end;

function TObservationsReader.ReadOptionalManifold(Molecule: TMolecule): TManifold;
begin
  Result := Molecule.ManifoldWithName(s);
  { TODO: Fails if both "A0" and "A0 hot band" exist as separate manifolds }
  if not Assigned(Result) and not (s[1] in ['.', '+', '-', '0'..'9']) then begin
    { Might well be a Manifold name, as it is not a number;
      Allow for Manifold name to have spaces in it }
    while FindMNodeStartingWith(s,Molecule,TManifold) <> nil do begin
      s := s + ' ' + NextItem;
      Result := Molecule.ManifoldWithName(s);
      if Assigned(Result) then Break;
    end;
    // if not Assigned(Result) and (Pos('(',s) <= 0) then
    //   ReportError('Manifold not found: ' + s);
  end;
  if Assigned(Result) then
    if f.GetFreeItem then
      s := f.Token
    else
      s := '';
end;

function TObservationsReader.ReadOptionalMolecule: TMolecule;
begin
  Result := Mixture.FindMolecule(s);
  { TODO: Fails if both "A0" and "A0 hot band" exist as separate molecules }
  if not Assigned(Result) and not (s[1] in ['.', '+', '-', '0'..'9']) then begin
    { Might well be a Molecule name, as it is not a number;
      Allow for Molecule name to have spaces in it }
    while FindMoleculeStartingWith(s,Mixture) <> nil do begin
      s := s + ' ' + NextItem;
      Result := Mixture.FindMolecule(s);
      if Assigned(Result) then Break;
    end;
    // if not Assigned(Result) and (Pos('(',s) <= 0) then
    //   ReportError('Manifold not found: ' + s);
  end;
  if Assigned(Result) then
    if f.GetFreeItem then
      s := f.Token
    else
      s := '';
end;

function TObservationsReader.ReadQuantumNumber(var a: Integer; const si: string): Boolean;
begin
  if not TryStrToQuantumNumber(si, a, QuantumNumberFormat) then begin
    ReportError('Bad quantum number:'+si);
    Result := True;
  end else
    Result := False;
end;

procedure TObservationsReader.AddObservation(AObs: TObservation);
  function FixOffset(J: TQuantumNumber; var Sym: Shortint; var Index: Smallint; const IndexOffsets: array of Integer; Manifold: TManifold; State: TState):string;
  var
    i: Integer;
    Ev: TEigenvalues;
  begin
    Result := '';
    if (Sym >= 0) and (Index > 0) and (Length(IndexOffsets) > 0) then begin
      if (Length(IndexOffsets) = 1) and (Abs(IndexOffsets[0]) = MaxInt) then begin
        Index := 0;
        if IndexOffsets[0] < 0 then
          Sym := -1;
        Exit;
      end;
      i := Manifold.nSym*(J div 2);
      i := i + Manifold.Molecule.JAdjustedSym(J,Sym);
      while i > High(IndexOffsets) do
        Dec(i, Manifold.nSym);
      if i < 0 then
        i := 0;
      // WriteLn(Log, J:4, Sym:4, Index:4, IndexOffsets[i]:4);
      Inc(Index,IndexOffsets[i]);
      if Index <= 0 then
        ReportError('Offset index = '+IntToStr(Index));
    end;
    if Assigned(State) and (Index <> 0) then begin
      Ev := TEigenvalues.Create(Manifold,J,Sym,ExField,True,State);
      try
        if (Index > 0) and (Index <= Ev.Size) then begin
          Result := Ev.Basis[Index].StateLabel(AObs.QuantumNumberFormat);
          Index := 0;
        end else
          ReportError(Format('Index too large (%d > %d)', [Index, Ev.Size]));
      finally
        Ev.Free;
      end;
    end;
  end;
var
  ExtraName: string;
  Molecule: TMolecule;
begin
  if AObs.ObsType in [obEUpper, obCDUpper] then
    Molecule := AObs.ManifoldU.Molecule
  else
    Molecule := AObs.ManifoldL.Molecule;
  if (Molecule.nNuclei = 0) and (AObs.RemoveSpins > 0) then
    AObs.RemoveSpins := 1; { Only 1 spin that could be ignored }
  if AObs.Colour = clDefault then begin
    if AObs.Colour = clNone then
      AObs.Colour := AObs.ManifoldL.Colour;
    if AObs.Colour = clNone then
      AObs.Colour := AObs.ManifoldU.Colour;
  end;
  AObs.LineNumber := f.Line;
  if NameOverride = '' then
    AObs.FileName := FileName
  else
    AObs.FileName := NameOverride;
  AObs.Frequency := (AObs.Frequency + FrequencyOffset)*FrequencyScale;
  if (AObs.ObsType <> obIntensity) and (Units <> Mixture.Units) then begin
    if AObs.Weight > 0 then begin
      AObs.Weight := ConvertWidth(AObs.Weight*FrequencyScale, AObs.Frequency, Units, Mixture.Units);
    end;
    AObs.Frequency := Convert(AObs.Frequency,Units,Mixture.Units);
  end;
  if AObs.Weight = 0 then
    AObs.Weight := -1;
  if OverrideWeight >= 0 then
    AObs.Weight := OverrideWeight;
  AObs.Weight := AObs.Weight*ScaleWeight;
  if (Take <> tkObs) or (AObs.Weight > 0) then begin
    if AObs.ObsType <> obParameter then begin
      ExtraName := FixOffset(AObs.JL, AObs.SymL, AObs.IndexL, IndexOffsetsL, AObs.ManifoldL, StateL);
      ExtraName := FixOffset(AObs.JU, AObs.SymU, AObs.IndexU, IndexOffsetsU, AObs.ManifoldU, StateU)
                   + ' - ' + ExtraName;
      if ExtraName <> ' - ' then
        AObs.Name := ':' + ExtraName + ':' + Obs.Name;
    end;
    try
      if not CheckSelected(Obs) then
        Exit;
    except
      ReportError('Selection failure');
      Exit;
    end;
    if nObs >= Length(Observations) then
      SetLength(Observations, Length(Observations)+1000);
    if BaseReader.LastObsType <> AObs.ObsType then
      SwitchObsType(ObsType, False); { Force purge of any current common differences }
    BaseReader.LastObsType := AObs.ObsType;
    if AObs.ObsType in [obCDLower, obCDUpper] then
      AObs.ObsType := obLine;
    Observations[nObs] := AObs;
    LongestObsLabel := Max(LongestObsLabel, Length(Observations[nObs].Name));
    Inc(nObs);
  end;
end;

{ TPgopherReader }

function TPgopherReader.DoReadOneLine: Boolean;
var
  Molecule: TMolecule;
  Jtag: string;
  i: TQuantumNumber;
  StateForParameter: TState;
  ForceObParameter, NotKZero: Boolean;
  j: Integer;
  Qnos: TSimpleQnos;
  AStateU, AStateL: TState;
begin
  Result := False;
  ForceObParameter := False;

  { BranchTable Handling - multiple transitons per line }
  if Length(BranchLabels) > 0 then begin
    if ReadQuantumNumber(i,s) then
      ReportError('Bad quantum number:'+s)
    else begin
      Jtag := '(' + s + ')';
      for i := 0 to High(BranchLabels) do
        { TODO: tab is not necessarily end of line }
        if f.GetFreeItem then begin
          if f.Token <> '-' then begin
            Obs.ManifoldU := DefaultManifoldU;
            Obs.ManifoldL := DefaultManifoldL;
            if not DefaultStateL.ReadBranchFormat(BranchLabels[i]+JTag,Obs,DefaultStateU, Self) then
              ReportError('Bad manifold or branch:'+Obs.Name)
            else if not TryStrToFloatEx(f.Token, Obs.Frequency) then
               ReportError('Bad frequency:' + s)
            else begin
              Obs.Weight := 1;
              AddObservation(Obs);
              ResetObs;
            end;
          end;
        end else
          Break;
    end;
    Exit;
  end;

  if (GlobalNQN <> 0) and ((s = '-') or TryStrToQuantumNumber(s,Qnos.QNP[0],QuantumNumberFormat)) then begin
    { Simple format }
    if s = '-' then begin
      { Level rather than transition fit. In copy from linelist
        window the quantum numbers show as '-' if blank }
      Obs.ObsType := obELower;
      Qnos.QNP[0] := -1;
      for j := 2 to GlobalNQN do begin
        s := NextItem; { Skip other quantum numbers }
        Qnos.QNP[j-1] := -1;
      end;
    end else
      for j := 2 to GlobalNQN do begin
        s := NextItem;
        if not TryStrToQuantumNumber(s,Qnos.QNP[j-1],QuantumNumberFormat) then
          ReportError('Bad quantum number:'+s);
      end;
    s := NextItem;
    if s = '-' then begin
      { Upper state level rather than transition fit }
      Obs.ObsType := obEUpper;
      Qnos.QN[0] := -1;
      for j := 2 to GlobalNQN do begin
        s := NextItem; { Skip other quantum numbers }
        Qnos.QN[j-1] := -1;
      end;
    end else
      for j := 1 to GlobalNQN do begin
        if j <> 1 then
          s := NextItem;
        if not TryStrToQuantumNumber(s,Qnos.QN[j-1],QuantumNumberFormat) then
          ReportError('Bad quantum number:'+s);
      end;
    AStateU := DefaultStateU;
    AStateL := DefaultStateL;
    if ConvertQuantumNumbers(Obs.ManifoldU, AStateU, Obs.JU, Obs.SymU, Obs.IndexU,
                             Obs.ManifoldL, AStateL, Obs.JL, Obs.SymL, Obs.IndexL, Obs.Name,
                             ComputeQ(DefaultStateU, DefaultStateL, GlobalNQN, RemoveSpins), GlobalNQN, Qnos,
                             ExField, UpperMap, LowerMap, QuantumNumberFormat, RemoveSpins, BranchRank, NotKZero) then
      ReportError('Bad Quantum numbers');
  end else begin
    { Normal one transition per line }
    if LowerCase(s) = 'parameter' then begin
      ForceObParameter := True;
      s := NextItem;
    end;

    Molecule := ReadOptionalMolecule;
    if Molecule = nil then
      Molecule := DefaultMolecule;
    if Molecule = nil then begin
      ReportError('Molecule ' + s + ' not found');
      Exit;
    end;

    Obs.ManifoldU := ReadOptionalManifold(Molecule);
    if Obs.ManifoldU = nil then
      Obs.ManifoldU := DefaultManifoldU;
    if Obs.ManifoldU = nil then begin
      ReportError('Manifold ' + s + ' not found');
      Exit;
    end;
    if ForceObParameter or not TryStrToQuantumNumber(s,Obs.JU,QuantumNumberFormat) then begin
      StateForParameter := Obs.ManifoldU.StateWithName(s);
      if ForceObParameter and not Assigned(StateForParameter) then begin
        ReportError('State ' + s + ' not found for parameter');
        Exit;
      end;
    end else
      StateForParameter := nil;
    if not Assigned(StateForParameter) then
      { Optionally allow manifoldL here }
      Obs.ManifoldL := ReadOptionalManifold(Molecule);

    if Assigned(StateForParameter) then begin
      { Parameter as observation }
      s := NextItem; { Parameter name }
      Obs.IndexU := StateForParameter.IndexOfParameter(s);
      if Obs.IndexU <= 0 then begin
        ReportError('Paramter ' + s + ' not found in ' + StateForParameter.Name);
        Exit;
      end;
      s := NextItem;
      if s <> '='  then begin
        ReportError('= expected, found ' + s);
        Exit;
      end;
      Obs.ObsType := obParameter;
      Obs.IndexL := StateForParameter.ComponentIndex;
      Obs.ManifoldL := Obs.ManifoldU;
      Obs.Flags := Obs.Flags - [ofRegenerateLabel];
    end else if Pos('(',s) > 0 then begin
      { Check for branch name }
      if Obs.ManifoldL = nil then
        Obs.ManifoldL := DefaultManifoldL;
      if DefaultStateL = nil then begin
        ReportError('Default lower state not set');
        Exit;
      end;
      if not DefaultStateL.ReadBranchFormat(s,Obs,DefaultStateU, Self) then begin
        ReportError('Bad manifold or branch:'+s+' '+Obs.Name);
        Exit;
      end;
    end else begin
      if s = '-' then begin
        { Level rather than transition fit. In copy from linelist
          window the quantum numbers show as '-' if blank }
        Obs.ObsType := obELower;
        s := NextItem;
        s := NextItem;
        Obs.ManifoldU := nil;
      end else begin
        if ReadQuantumNumber(Obs.JU, s) then Exit;
        if ReadSymmetry(Obs.SymU, Obs.JU, Molecule, True) then Exit;
        if ReadInteger(Obs.IndexU) then Exit;
        if Obs.IndexU < 0 then ReportError('(Upper) Eigenvalue number must be >=0');
      end;

      s := NextItem;
      if Obs.ManifoldL = nil then begin
        Obs.ManifoldL := ReadOptionalManifold(Molecule);
        if Obs.ManifoldL = nil then
          Obs.ManifoldL := DefaultManifoldL;
      end;
      if s = '-' then begin
        { Upper state level rather than transition fit }
        Obs.ObsType := obEUpper;
        s := NextItem;
        s := NextItem;
        Obs.ManifoldL := nil;
      end else begin
        if Obs.ManifoldL = nil then begin
          ReportError('Manifold ' + s + ' not found');
          Exit;
        end;
        if ReadQuantumNumber(Obs.JL,s) then Exit;
        if ReadSymmetry(Obs.SymL, Obs.JL, Molecule, True) then Exit;
        if ReadInteger(Obs.IndexL) then Exit;
        if Obs.IndexL < 0 then ReportError('(Lower) Eigenvalue number must be >=0');
      end;
    end;
  end;

  s := NextItem;
  if not TryStrToFloatEx(s, Obs.Frequency) then begin
    ReportError('Bad frequency:' + s);
    Exit;
  end;
  if not f.GetFreeItem then begin
    Obs.Weight := 1 { End of line - take default weight }
  end else begin
    s := f.Token;
    if s = '-' then
      Exit; { Not weighted - skip }
    if s = '' then
    else if not TryStrToFloatEx(s, Obs.Weight) then begin
      ReportError('Bad (relative) standard deviation:' + s);
      Exit;
    end;
    s := TrimRight(f.RestOfLine);
    if (s <> '') and (s[1] in [#9, ' ']) then
      Delete(s, 1, 1);
    if (Obs.Name <> '') and (s <> '') then
      Obs.Name := Obs.Name + ': ' + s
    else
      Obs.Name := Obs.Name + s;
  end;
  Result := True;
end;

procedure LoadObservations(Mixture: TMixture; Stream: TStream; FileName:string;
  ObsType: TObsType; TakeWeight: TTakeWeight);
begin
  ClearObservations;
  if AddToObservations(Mixture,FileName,Stream,ObsType,TakeWeight) then
    raise EReportException.Create({$IFNDEF CONSOLE}'See log windows for '+{$ENDIF}'errors detected in loading ' + Filename);
end;

function TPgopherReader.HandleDirective(var Handled: Boolean): Boolean;
begin
  Result := inherited HandleDirective(Handled);
  if not Handled and (LowerCase(s) = 'branchtable') then begin
    SetLength(BranchLabels,0);
    while f.GetFreeItem do begin
      SetLength(BranchLabels,Length(BranchLabels)+1);
      BranchLabels[High(BranchLabels)] := f.Token;
    end;
    Handled := True;
  end;
end;

{ TOldPgopherReader }

function TOldPgopherReader.DoReadOneLine: Boolean;
begin
  Result := False;
  if ReadInt(Obs.JU, s) then Exit;
  if ReadInteger(Obs.SymU) then Exit;
  Obs.SymU := (1 - Obs.SymU) div 2;
  if (Obs.SymU < 0) or (Obs.SymU > DefaultMolecule.LastSym) then
    ReportError('Upper Symmetry number not in range 0..'+IntToStr(DefaultMolecule.LastSym));
  if ReadInteger(Obs.IndexU) then Exit;
  if Obs.IndexU <= 0 then ReportError('(Upper) Eigenvalue number must be > 0');

  if ReadInteger(Obs.JL) then Exit;
  if ReadInteger(Obs.SymL) then Exit;
  Obs.SymL := (1 - Obs.SymL) div 2;
  if (Obs.SymL < 0) or (Obs.SymL > DefaultMolecule.LastSym) then
    ReportError('Lower Symmetry number not in range 0..'+IntToStr(DefaultMolecule.LastSym));
  if ReadInteger(Obs.IndexL) then Exit;
  if Obs.IndexL <= 0 then ReportError('(Lower) Eigenvalue number must be > 0');

  s := NextItem;
  if not TryStrToFloatEx(s, Obs.Frequency) then begin
    ReportError('Bad frequency:' + s);
    Exit;
  end;
  s := NextItem;
  if s = '-' then
    Exit; { Not weighted - skip }
  if not TryStrToFloatEx(s, Obs.Weight) then begin
    ReportError('Bad (relative) standard deviation:' + s);
    Exit;
  end;
  Obs.Name := TrimRight(f.RestOfLine);
  if (Obs.Name <> '') and (Obs.Name[1] in [#9, ' ']) then
    Delete(Obs.Name, 1, 1);
  Result := True;
end;

end.

