unit Overlay;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses {$IFNDEF CONSOLE} {$IFDEF MSWINDOWS} Windows, {$ENDIF} {$ENDIF}
     Units, Graphics, GraphM, GraphRange, SysUtils, Classes, MNode, Mixture, pgDOM, Math, MathSup, Peaks;

type

  { TPlot }

  TPlot = class(TUnitsMNode)
  private
    FPeaks: array of TPeak;
    function GetActual(AIndex: Integer): Double;
    procedure SetActual(AIndex: Integer; const Value: Double);
    function GetnPeaks: Integer;
    procedure SetnPeaks(Value: Integer);
    function GetPeak(AIndex: Integer): TPeak;
    procedure SetPeak(AIndex: Integer; const Value: TPeak);
  protected
    FDefaultName: Boolean;
    function GetHMax: Double; virtual; abstract;
    function GetHMin: Double; virtual; abstract;
    function GetIMin: Double; virtual; abstract;
    function GetIMax: Double; virtual; abstract;
    procedure PeakChanged(PeakNumber:Integer);
    procedure SetName(const Value: TComponentName); override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure AddPeak(Position: Double; DisplayUnits: TResonanceUnits); overload;
    procedure AddPeak(Peak: TPeak); overload;
    function AdjustFrequency(Freq: Double):Double; virtual;
    function DeAdjustFrequency(Freq: Double):Double; virtual;
    function Calibrating: Boolean;
    procedure ClearPeaks;
    procedure SortPeaks(Field:Integer; Ascending: Boolean);
    procedure ClearSelection; virtual; abstract;
    function ConvertPeakPosition(Position: Double; PlotUnits: TResonanceUnits):Double; virtual;
    procedure CycleUnits; override;
    procedure DeletePeak(AIndex:Integer);
    function Empty: Boolean; virtual; abstract;
    function FindPeak(AFmin, AFmax: Double; out Position: Double; PlotUnits: TResonanceUnits): Boolean; virtual; abstract;
    function LimitsInRange(Simulation: TSimulation; var AIMin, AImax: Double; PlotIntensityUnits: TIntensityUnits):Boolean; virtual; abstract;
    procedure LoadFromFile(FileName: string; DefaultUnits: TResonanceUnits; AlwaysPrompt: Boolean; DefaultIntensityUnits: TIntensityUnits); virtual; abstract;
    procedure Invert; virtual; abstract;
    procedure SaveForUndo; virtual; abstract;
    procedure Undo; virtual; abstract;
    function CanUndo: Boolean; virtual;
    function IsSettingColour(AIndex: Integer): Boolean; override;
    procedure SelectFrequencyRange(AFmin, AFmax: Double); virtual; abstract;
    procedure ScaleParameters(UFrom, UTo: TResonanceUnits); override;
    procedure OffsetBy(Offset: Double); virtual;
    function Show(G: TGraphM; Simulation: TSimulation; PlotIntensityUnits: TIntensityUnits;
                  PlotRange: TPlotRange; var PlotRanges: TPlotRanges; OwnAxis: Boolean;
                  ShowName: Boolean; ForceAxis: Boolean = False):Boolean; virtual;
    function ShowPeaks(G: TGraphM; AImin, AImax: Double; Simulation: TSimulation):Boolean;
    function Resolution: Double; virtual; abstract;
    procedure ReadPeaks(Child: TDOMNode);
    procedure WritePeaks(Result: TDOMNode);
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    property HMin: Double read GetHMin;
    property HMax: Double read GetHMax;
    property IMin: Double read GetIMin;
    property IMax: Double read GetIMax;
    property Actual[AIndex:Integer]: Double read GetActual write SetActual;
    property Peaks[AIndex:Integer]: TPeak read GetPeak write SetPeak;
    property nPeaks: Integer read GetnPeaks write SetnPeaks;
    property DefaultName:Boolean read FDefaultName;
  published
    property FrequencyOffset: Double index 1 read GetParameter write SetParameter; { Frequency Offset }

    property Units:  TResonanceUnits index 1 read GetUnits  write SetUnits;
    property IntensityUnits: TIntensityUnits index 2 read GetIntensityUnits write SetIntensityUnits;
    property Colour: TColor          index 3 read GetColour write SetColour;
    property Hide:   Boolean         index 4 read GetFlag   write SetFlag;
    property ZeroValid: Boolean      index 5 read GetFlag   write SetFlag;
  end;

  TPlotClass = class of TPlot;

  TStretchablePlot = class(TPlot)
  protected
    FHMin, FHMax: Double;
    function GetHMin: Double; override;
    function GetHMax: Double; override;
  public
    constructor Create(AOwner: TComponent); override;
    function AdjustFrequency(Freq:Double):Double; override;
    function DeAdjustFrequency(Freq: Double):Double; override;
    function HelpOnSetting(AIndex: Integer): string; override;
  published
    property FrequencyScale: Double index 2 read GetParameter write SetParameter;
    property FixCentre: Boolean     index 6 read GetFlag write SetFlag;
  end;

  { TOverlays }

  TOverlays = class(TRootMNode)
  private
    FHasLimits: Boolean;
    FnPlots: Integer;
    FPlots: TMNodeArray;
    function GetPlot(AIndex: Integer): TPlot;
    procedure SetnComponents(const Index, Value: Integer);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    HMin, HMax: Double; { Units as set below }
    Imin, Imax: Double;
    LastShowOffset: Boolean;
    constructor Create(AOwner: TComponent); override;
    procedure CycleUnits; override;
    procedure Show(G: TGraphM; Simulation: TSimulation; PlotUnits: TResonanceUnits; PlotIntensityUnits: TIntensityUnits;
                   PlotRange: TPlotRange; var PlotRanges: TPlotRanges; OwnAxis: Boolean; Offset: Boolean = false);
    {$IFNDEF CONSOLE}
    function FindPeak(AArea: TDRect; var Position: Double; var Plot: TPlot; PlotUnits: TResonanceUnits; const PlotRanges: TPlotRanges): Integer;
    {$ENDIF}
    function FixupProperties:Exception; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    function LimitsInRange(Simulation: TSimulation; PlotUnits: TResonanceUnits; var AImin, AImax: Double; PlotIntensityUnits: TIntensityUnits; out AllArbitrary: Boolean):Boolean;
    procedure Invert;
    procedure ChooseUniqueColourFor(Plot: TPlot);
    function nVisible: Integer; { Number of non-hidden plots }
    property Plots[AIndex: Integer]: TPlot read GetPlot; default;
    procedure SaveToFile(AFileName: string); override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    property HasLimits: Boolean read FHasLimits;
    property nPlots: Integer read FnPlots;
  published
    property Units:          TResonanceUnits index 1 read GetUnits write SetUnits;
    property IntensityUnits: TIntensityUnits index 2 read GetIntensityUnits write SetIntensityUnits;
    property ContourFile:    Boolean index 3 read GetFlag write SetFlag; { if set, read an observation file when contour fitting, mainly for constraints }
    property nComponents:    Integer index 4 read GetSetting write SetnComponents;
    property PeakThreshold:  Double index 1 read GetParameter write SetParameter;
  end;

{ This routine adds to the given overlay, or creates the overlay if necessary;
  It return true if a new overlay hes been created }
function AddToOverlay(var Overlay: TOverlays; FileName: string; Merge: Boolean; DefaultUnits: TResonanceUnits; AlwaysPrompt: Boolean; DefaultIntensityUnits: TIntensityUnits):Boolean; overload;

implementation

uses {$IFNDEF CONSOLE} Clipbrd, ExpBitmap, MNodeClip, CalFrm, LineFrm, {$ENDIF}
     MNodeColour, ExpPlot, ShellSort, Global, CalSource, I2, Profile, TabbedLine;

var
  LoadOverlayTime: string = 'Load Overlay';

function AddToOverlay(var Overlay: TOverlays; FileName: string; Merge: Boolean; DefaultUnits: TResonanceUnits; AlwaysPrompt: Boolean; DefaultIntensityUnits: TIntensityUnits):Boolean;
var
  Plot: TPlot;
  i: Integer;
  Stream: TFileStream;
  NewOverlay: TOverlays;
  function TryLoading(PlotClass: TPlotClass): TPlot;
  begin
    Result := PlotClass.Create(NewOverlay);
    try
      Result.LoadFromFile(FileName, DefaultUnits, AlwaysPrompt, DefaultIntensityUnits);
    except
      FreeAndNil(Result);
    end;
  end;
begin
  Result := not Merge or not Assigned(Overlay);
  Plot := nil;
  try
    if Result then begin
      NewOverlay := TOverlays.Create(nil);
      NewOverlay.Name := 'Overlays';
      // if Assigned(Mixture) then
      //  Overlays.Units := Mixture.Units;
    end else
      NewOverlay := Overlay;
    TStart(LoadOverlayTime);
    i := Pos(' Calibration>',FileName);
    if (i > 0) and (i = Length(FileName)-12) then
      MakeCalibrationSource(Copy(FileName, 2, Pos(' Calibration>',FileName)-2),NewOverlay)
    else if (LowerCase(ExtractFileExt(FileName)) = '.ovr') or (LowerCase(ExtractFileExt(FileName)) = '.xml') then begin
      Stream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
      try
        if Result then
          NewOverlay.LoadFromStream(Stream, FileName)
        else
          NewOverlay.MergeFromStream(Stream, FileName)
      finally
        Stream.Free;
      end;
    end else {$IFNDEF CONSOLE} if (Filename <> '') or (PasteXMLNodeOnto(NewOverlay) = nil) then {$ENDIF} begin
      Plot := TryLoading(TExperimentalPlot);
      {$IFNDEF CONSOLE}
      if not Assigned(Plot) then
        Plot := TryLoading(TBitmapPlot);
      {$ENDIF}
      if not Assigned(Plot) then begin
        if FileName = '' then
          FileName := 'clipboard';
        raise EReportException.Create('Overlay Load from ' + FileName + ' failed');
      end;
//      if Result then
//        NewOverlay.IntensityUnits := Plot.IntensityUnits;
      for i := Plot.ComponentIndex to NewOverlay.nPlots-1 do
        NewOverlay.ChooseUniqueColourFor(NewOverlay.Plots[i]);
      Plot.Modified := False;
    end;
    NewOverlay.CheckProperties;
    TEnd(LoadOverlayTime);
    if Result then begin
      Overlay.Free;
      Overlay := NewOverlay;
      Overlay.Modified := False;
      PerformMNodeNotification(NewOverlay, mnNewBaseNode, 0);
    end;
  except
    if Result then
      FreeAndNil(NewOverlay)
    else
      Plot.Free;
    raise;
  end;
end;

{ TPlot }

function TPlot.DeAdjustFrequency(Freq: Double): Double;
begin
  Result := Freq - FrequencyOffset;
end;

function TPlot.AdjustFrequency(Freq: Double): Double;
begin
  Result := Freq + FrequencyOffset;
end;

function TPlot.Show(G: TGraphM; Simulation: TSimulation; PlotIntensityUnits: TIntensityUnits;
                    PlotRange: TPlotRange; var PlotRanges: TPlotRanges; OwnAxis: Boolean;
                    ShowName: Boolean; ForceAxis: Boolean):Boolean;
var
  xl, xr, gl, gr, x0, x1, y0, y1: Double;
  LeftJustify: Boolean;
  Justification : TAlignment;
begin
  G.MapArg(x0, x1, y0, y1);
  MapFromPlotRange(G, PlotRange);
  if ShowName then begin
    xl := Convert(HMin,Units,Simulation.PlotUnits);
    xr := Convert(HMax,Units,Simulation.PlotUnits);
    InOrder(xl,xr);
    gl := G.Xmin;
    gr := G.Xmax;
    InOrder(gl,gr);
    LeftJustify := G.Xmax > G.Xmin;
    if xl - gl > gr - xr then
      { More space on left }
      LeftJustify := not LeftJustify
    else { More space on right }
      xl := xr;
    if LeftJustify then
      Justification := taLeftJustify
    else
      Justification := taRightJustify;
    G.TextJ(xl, (PlotRange.yi+PlotRange.yx)/2, Justification, Name)
  end;
  Result := ShowPeaks(G, PlotRange.yi, PlotRange.yx, Simulation);
  G.Map(x0, x1, y0, y1);
end;

function TPlot.ShowPeaks(G: TGraphM; AImin, AImax: Double; Simulation: TSimulation):Boolean;
var
  i, nShown: Integer;
  Mi, Mx, Fi, Fx, s, ds: Double;
  Peak: TPeak;
  Lab: string;
begin
  G.MapArg(Fi, Fx, Mi, Mx);
  if Fi > Fx then begin
    s := Fi;
    Fi := Fx;
    Fx := s;
  end;
  nShown := 0;
  with Owner as TOverlays do
    if LastShowOffset or (Count = 1) then
      G.Style(4);
  ds := (Fx-Fi)/G.Width;
  for i := 0 to nPeaks-1 do begin
    Peak := Peaks[i];
    if IsNanOrInf(Peak.Position) then
      Continue;
    s := ConvertPeakPosition(Peak.Position, Simulation.PlotUnits);
    if (s >= Fi) and (S <= Fx) then begin
      G.ClipPoint(s, AImin);
      G.ClipJoin(s, AImax);
      G.TextAngle(90);
      {$IFNDEF CONSOLE}
      if CalibratingPlot = Self then
        Lab := FloatToStr(Peak.Position)
      else
      {$ENDIF}
        Lab := RoundedStr(s, ds);
      G.TextJ(S, AImax, taRightJustify, Lab);
      Inc(nShown);
    end;
  end;
  G.TextAngle(0);
  Result := nShown > 0;
end;

procedure TPlot.AddPeak(Peak: TPeak);
begin
  Setlength(FPeaks, Length(FPeaks)+1);
  FPeaks[High(FPeaks)] := Peak;
  PeakChanged(High(FPeaks));
end;

procedure TPlot.AddPeak(Position: Double; DisplayUnits: TResonanceUnits);
var
  Peak: TPeak;
  Lab1, Lab2, Lab: string;
begin
  Peak := EmptyPeak;
  if Calibrating then
    Peak.Position := Position
  else
    Peak.Position := DeAdjustFrequency(Position);
  AddPeak(Peak);
  Lab1 := ExtractFileName(Name);
  if (Lab1 = '1') and ((Owner as TOverlays).nPlots = 1) then
    Lab1 := '';
  Lab2 := ExtractFileName((Owner as TMNode).Name);
  if (Lab1 <> '') and (Lab2 <> '') then
    Lab := Lab1 + ' - ' + Lab2
  else
    Lab := Lab1 + Lab2;
  {$IFNDEF CONSOLE}
  PeakMeasured(Convert(Position, Units, DisplayUnits), DisplayUnits, Lab,Self);
  {$ENDIF}
end;

function TPlot.Calibrating: Boolean;
begin
  {$IFNDEF CONSOLE}
  Result := Self = CalibratingPlot;
  {$ELSE}
  Result := False;
  {$ENDIF}
end;

function TPlot.CanUndo: Boolean;
begin
  Result := False;
end;

procedure TPlot.ClearPeaks;
begin
  nPeaks := 0;
end;

function TPlot.ConvertPeakPosition(Position: Double;
  PlotUnits: TResonanceUnits): Double;
begin
  Result := Convert(AdjustFrequency(Position), Units, PlotUnits);
end;

type
  TPeakSorter = class (TShellSorter)
    Plot: TPlot;
    Fld: Integer;
    Asc: Boolean;
    function InOrder(I,J:Integer):Boolean; override;
    procedure Swap(I,J: Integer); override;
  end;

{ TPeakSorter }

function TPeakSorter.InOrder(I, J: Integer): Boolean;
var
  PI, PJ: PDouble;
begin
  PI := PDouble(@Plot.FPeaks[I]);
  Inc(PI,Fld);
  PJ := PDouble(@Plot.FPeaks[J]);
  Inc(PJ,Fld);
  { Relies on contiguous allocation; explicit check in SortPeaks }
  if Asc then
    Result := PI^ <= PJ^
  else
    Result := PI^ >= PJ^
end;

procedure TPeakSorter.Swap(I, J: Integer);
begin
  SwapValues(Plot.FPeaks[I],Plot.FPeaks[J],Sizeof(TPeak));
end;

procedure TPlot.SortPeaks(Field: Integer; Ascending: Boolean);
begin
  with TPeakSorter.Create do begin
    if SizeOf(TPeak) <> 4*SizeOf(Double) then
      raise NotYet;
    Plot := Self;
    Fld := Min(Max(Field,0),3);
    Asc := Ascending;
    Sort(nPeaks);
    Free;
    PeakChanged(-1);
  end;
end;

constructor TPlot.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Colour=None';
  inherited;
  FDefaultName := True;
end;

procedure TPlot.DeletePeak(AIndex: Integer);
var
  i: Integer;
begin
  if (AIndex < 0) or (AIndex >= nPeaks) then
    Exit;
  for i := AIndex to nPeaks - 2 do
    FPeaks[i] := FPeaks[i+1];
  nPeaks := nPeaks - 1;
end;

function TPlot.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := AIndex = 3;
  if not Result then
    Result := inherited IsSettingColour(AIndex);
end;

procedure TPlot.OffsetBy(Offset: Double);
begin
  FrequencyOffset := FrequencyOffset + Offset;
end;

procedure TPlot.PeakChanged(PeakNumber: Integer);
begin
  {$IFNDEF CONSOLE}
  if Assigned(CalForm) then
    CalForm.PeakChanged(Self,PeakNumber);
  {$ENDIF}
end;

function TPlot.HelpOnSetting(AIndex: Integer): string;
begin
  case AIndex of
    1: Result := 'Horizontal Axis units.';
    2: Result := 'Verical units of the data; select arbitrary to autoscale the plot';
    3: Result := 'Colour for plot.';
    4: Result := 'Set to hide this plot.';
    5: Result := 'Set if zero in data is correct.';
  else
    Result := inherited HelpOnSetting(AIndex);
  end;
end;

function TPlot.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Frequency offset used when displaying plot.'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

procedure TPlot.WritePeaks(Result: TDOMNode);
var
  Strings: TStringList;
  i: Integer;
begin
  if nPeaks = 0 then
    Exit;
  Strings := TStringList.Create;
  try
    Strings.Capacity := nPeaks+1;
    Strings.Add('');
    for i := 0 to nPeaks-1 do
      Strings.Add(PeakToString(Peaks[i]));
    Result.AppendChild(Result.OwnerDocument.CreateCDATASection(Strings.Text));
  finally
    Strings.Free;
  end;
end;

procedure TPlot.ReadPeaks(Child: TDOMNode);
var
  Strings: TStringList;
  i: Integer;
  Peak: TPeak;
  NextChild: TDOMNode;
begin
  while Assigned(Child) do begin
    NextChild := Child.NextSibling;
    if Child is TDOMCDATASection then begin
      Strings := TStringList.Create;
      try
        Strings.Text := TDOMCDATASection(Child).Data;
        Child.ParentNode.RemoveChild(Child);
        for i := 0 to Strings.Count-1 do
          if TryStrToPeak(Strings[i], Peak) then begin
            if not IsNan(Peak.Position) then
              AddPeak(Peak);
          end else
            raise EReportException.Create('Bad line in peak list:'+Strings[i]);
      finally
        Strings.Free;
      end;
    end;
    Child := NextChild;
  end;
end;

function TPlot.GetActual(AIndex: Integer): Double;
begin
  Result := Peaks[AIndex].Actual;
end;

function TPlot.GetnPeaks: Integer;
begin
  Result := Length(FPeaks);
end;

function TPlot.GetPeak(AIndex: Integer): TPeak;
begin
  Result := FPeaks[AIndex]
end;

procedure TPlot.CycleUnits;
var
  NewUnits: TResonanceUnits;
begin
  if Units = High(TResonanceUnits) then
    NewUnits := Low(TResonanceUnits)
  else
    NewUnits := Succ(Units);
  ScaleParameters(Units, NewUnits);
  Units := NewUnits;
end;

procedure TPlot.ScaleParameters(UFrom, UTo: TResonanceUnits);
var
  i: Integer;
begin
  for i := 0 to nPeaks-1 do begin
    FPeaks[i].Position := Convert(FPeaks[i].Position, UFrom, UTo);
    if not IsNan(FPeaks[i].Weight) then
      FPeaks[i].Weight := Convert(FPeaks[i].Weight, UFrom, UTo);
    PeakChanged(i);
  end;
end;

procedure TPlot.SetActual(AIndex: Integer; const Value: Double);
begin
  FPeaks[AIndex].Actual := Value;
end;

procedure TPlot.SetName(const Value: TComponentName);
begin
  inherited;
  FDefaultName := Value = '';
end;

procedure TPlot.SetnPeaks(Value: Integer);
var
  OldNPeaks, i: Integer;
begin
  OldNPeaks := nPeaks;
  Value := Max(Value,0);
  if Value <> OldNPeaks then begin
    SetLength(FPeaks, Value);
    for i := OldNPeaks to Value - 1 do
      Peaks[i] := EmptyPeak;
    PeakChanged(-1);
  end;
end;

procedure TPlot.SetPeak(AIndex: Integer; const Value: TPeak);
begin
  FPeaks[AIndex] := Value;
  PeakChanged(AIndex);
end;

{ TOverlays }

constructor TOverlays.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',PeakThreshold=0.5';
  DimensionlessParameters := DimensionlessParameters + ',PeakThreshold=0.5';
  inherited;
end;

function TOverlays.FixupProperties:Exception;
var
  yStarted: Boolean;
  vi, vx, ix: Double;
  i: Integer;
  function GetIntensity(Plot: TPlot; Value: Double):Double;
  begin
    { Stick spectra intensities are independent of the frequency units }
    Result := ConvertIntensity(Value, (Plot.HMin+Plot.HMax)/2, Plot.Units, Plot.IntensityUnits, Units, IntensityUnits,
                               (Plot is TExperimentalPlot) and TExperimentalPlot(Plot).IsLineList);
  end;
begin
  FHasLimits := False;
  Result := inherited FixupProperties;
  if Assigned(Result) or (nPlots = 0) then Exit;
  yStarted := False;
  for i := 0 to nPlots-1 do
    if not (Plots[i] is TCalibrationSource) and not Plots[i].Empty and not Plots[i].Hide then begin
      vi := Convert(Plots[i].HMin, Plots[i].Units, Units);
      vx := Convert(Plots[i].HMax, Plots[i].Units, Units);
      InOrder(vi,vx);
      if not FHasLimits or (HMin > vi) then HMin := vi;
      if not FHasLimits or (HMax < vx) then HMax := vx;
      FHasLimits := True;
      if not IsArbitraryStyle(Plots[i].IntensityUnits) then begin
        ix := GetIntensity(Plots[i], Plots[i].IMin);
        if not yStarted or (IMin > ix) then IMin := ix;
        ix := GetIntensity(Plots[i], Plots[i].IMax);
        if not yStarted or (IMax < ix) then IMax := ix;
        yStarted := True;
      end;
    end;
  if not yStarted then begin
    IMin := 0; { All arbitrary units }
    IMax := 1;
  end;
  if FHasLimits and (HMin = HMax) then begin
    HMin := Max(HMin - 1,HMin/10); { Avoid zero }
    HMax := HMax + 1
  end;
end;

function TOverlays.GetPlot(AIndex: Integer): TPlot;
begin
  Result := FPlots[AIndex] as TPlot;
end;

procedure TOverlays.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  FnPlots := AdjustNodeIndexes(TPlot, FPlots, AComponent, Operation);
end;

{$IFNDEF CONSOLE}
function TOverlays.FindPeak(AArea: TDRect; var Position: Double; var Plot: TPlot;
                            PlotUnits: TResonanceUnits; const PlotRanges: TPlotRanges): Integer;
{ Return number of overlays scanned; < 0 indicates a peak found }
var
  i, j, PlotIndex, nFound: Integer;
  p, Frac, CFrac, CHMin, CHMax: Double;
begin
  { Track fraction of plot area covered }
  Result := 0;
  Frac := 0;
  nFound := 0;
  for i := 0 to nPlots - 1 do begin
    PlotIndex := -1;
    { Find matching plot range }
    for j := 0 to High(PlotRanges) do
      if PlotRanges[j].Tag = i+10 then begin
        PlotIndex := j;
        Break;
      end;
    if (PlotIndex < 0) or not RangeOverlaps(AArea.Bottom, AArea.Top, PlotRanges[PlotIndex]) then
      Continue;
    Inc(Result); { Track number of overlays considered }
    with Plots[i] do
      if FindPeak(AArea.Left, AArea.Right, p, PlotUnits) then
        if GetPlot(i) is TCalibrationSource then begin
          Inc(nFound);
          Plot := GetPlot(i);
          Position := p;
        end else begin
          CHMin := Convert(HMax,Units,PlotUnits);
          CHMax := Convert(HMin,Units,PlotUnits);
          InOrder(CHMin, CHMax);
          CFrac := (Min(AArea.Right,CHMax) - Max(CHMin,AArea.Left))/(AArea.Right - AArea.Left);
          if CFrac > Frac then begin
            Position := p;
            Plot := GetPlot(i);
            if CFrac > 1.5*Frac then
              nFound := 1
            else
              Inc(nFound);
            Frac := CFrac;
          end;
        end;
  end;
  if (nFound = 1) or ((nFound > 1) and LastShowOffset) then begin
    Plot.AddPeak(Position, PlotUnits);
    Result := -Result; { Flag a peak actually found }
  end;
end;
{$ENDIF}

procedure TOverlays.Show(G: TGraphM; Simulation: TSimulation; PlotUnits: TResonanceUnits;
                         PlotIntensityUnits: TIntensityUnits;
                         PlotRange: TPlotRange;  { Input: y1..y2 is screen plot range to use }
                                                 {        yi..yx is logical plot range }
                                                 { Mainly ignored if OwnAxis }
                         var PlotRanges: TPlotRanges;
                         OwnAxis: Boolean;       { If set, set compute PlotRange from overlay spectra }
                         Offset: Boolean = false { Offset individual plots } );
var
  i: Integer;
  yl, dy: Double;
  AllArbitrary: Boolean;
  UnitsCaption: string;
begin
  LastShowOffset := Offset;
  if nVisible = 0 then Exit;
  yl := 0;
  if Offset then
    dy := 1/nVisible
  else
    dy := 1;
  if OwnAxis then begin
    OwnAxis := OwnAxis and ((nVisible = 1) or Offset);
    if not OwnAxis then
      if IsArbitraryStyle(PlotIntensityUnits) then begin
        { Spectra will all have same vertical appearance on plot, apart from those with AlignZero set }
        LimitsInRange(Simulation, PlotUnits, PlotRange.yi, PlotRange.yx, PlotIntensityUnits, AllArbitrary);
      end else begin
        { Show normal vertical axis, allowing spectra to have different vertical ranges }
        LimitsInRange(Simulation, PlotUnits, PlotRange.yi, PlotRange.yx, PlotIntensityUnits, AllArbitrary);
        // Was PlotRange.yi := Imin;
        // PlotRange.yx := Imax;
        if AllArbitrary then
          UnitsCaption := '' { Normalised but no real units - Show scale, but leave units blank }
        else
          { TODO: Strictly should check for mixture of linelist/non linelist styles }
          UnitsCaption := NameOfUnit(PlotIntensityUnits,False,PlotUnits);
        AxisFromPlotRange(PlotRange, G, G.Xmin, UnitsCaption);
      end;
  end;
  for i := 0 to nPlots - 1 do
    if not Plots[i].Hide then begin
      if OwnAxis then
        PlotRange := GraphRange.PlotRange(G, yl, yl+dy, 0, 1);
      G.StyleP(clBlack, G.LineStyle.Width, G.LineStyle.Pattern);
      if Plots[i].Show(G, Simulation, PlotIntensityUnits, PlotRange, PlotRanges, OwnAxis, Offset and (nVisible > 1)) then
        PlotRanges[High(PlotRanges)].Tag := 10+i;
      if Offset then
        yl := yl + dy;
    end;
end;

function TOverlays.WriteProperties(Parent: TDOMNode; AllParameters: Boolean; WithChildren: Boolean): TDOMElement;
begin
  Result := inherited WriteProperties(Parent, AllParameters, WithChildren);
  WriteValue(Result, 'Version', VersionInfo, '');
end;

procedure TOverlays.ChooseUniqueColourFor(Plot: TPlot);
var
  i, ci: Integer;
  Clash: Boolean;
  NewColour: TColor;
begin
  for ci := 1 to High(MNodeColourValues)-3 do begin
    NewColour := MNodeColourValues[ci];
    Clash := False;
    for i := 0 to nPlots - 1 do
      if (Plots[i] <> Plot)
         and (Plots[i].Colour = NewColour) then begin
        Clash := True;
        Break;
      end;
    if not Clash then Break;
  end;
  { Can't avoid a clash, so choose a random value }
  if Clash then
    NewColour := MNodeColourValues[Random(High(MNodeColourValues)-3)+1];
  Plot.Colour := NewColour;
end;

procedure TOverlays.SaveToFile(AFileName: string);
var
  Stream: TStream;
  i: Integer;
  s: string;
begin
  { Add meaningful names on initial save }
  s := ExtractFileName(AFileName);
  s := Copy(s,1,Length(s)-Length(ExtractFileExt(s)));
  for i := 0 to nPlots - 1 do
    if Plots[i].DefaultName then
      if nPlots = 1 then
        Plots[i].Name := s
      else
        Plots[i].Name := s + '-' + Items[i].Name;
  if LowerCase(ExtractFileExt(AFileName)) = '.dat' then begin
    Stream := TFileStream.Create(AFileName, fmCreate);
    try
      for i := 0 to nPlots - 1 do
        if Plots[i] is TExperimentalPlot then
          TExperimentalPlot(Plots[i]).SimpleSaveToStream(Stream);
    finally
      Stream.Free;
    end;
  end else
    inherited;
end;

procedure TOverlays.SetnComponents(const Index, Value: Integer);
var
  i: Integer;
begin
  SetSetting(Index,Value);
  for i := 0 to ComponentCount-1 do
    if Components[i] is TExperimentalPlot then
      TExperimentalPlot(Components[i]).AdjustParameterNames;
end;

procedure TOverlays.Invert;
var
  i: Integer;
begin
  for i := 0 to nPlots - 1 do
    Plots[i].Invert;
end;

function TOverlays.LimitsInRange(Simulation: TSimulation;
  PlotUnits: TResonanceUnits; var AImin, AImax: Double;
  PlotIntensityUnits: TIntensityUnits;
  out AllArbitrary: Boolean): Boolean;
var
  i, pass: Integer;
  yi, yx: Double;
begin
  Result := False;
  AllArbitrary := True;
  for pass := 0 to 1 do
    { Double pass arrangement allows non arbitrary unit spectra to be correctly
      scaled with respect to each other, with arbitrary ones autoscaled }
    for i := 0 to nPlots - 1 do
      if not Plots[i].Hide and ((pass = 0) xor IsArbitraryStyle(Plots[i].IntensityUnits))
         and Plots[i].LimitsInRange(Simulation, yi, yx, PlotIntensityUnits) then begin
        if IsArbitraryStyle(Plots[i].IntensityUnits) then begin
          if not Result then begin
            AImin := 0;
            AImax := 1;
          end;
          { Arbitrary autoscales, but limit scaling when zero is flagged as good }
          if Plots[i].ZeroValid then
            yi := Min(yi/yx*AImax,0)
          else
            yi := AImin;
          yx := AImax;
        end else
          AllArbitrary := False;
        if Result then begin
          AImin := Min(AImin, yi);
          AImax := Max(AImax, yx);
        end else begin
          AImin := yi;
          AImax := yx;
          Result := True;
        end;
      end;
end;

const
  OverlaySettingNames: array[1..3] of string = (
'Frequency/Wavelength Units for plot.',
'Intensity unitss for plot',
'Set to read an observation file when contour fitting, typically to set constraints.'
);

function TOverlays.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(OverlaySettingNames))
     and (AIndex <= High(OverlaySettingNames)) then
    Result := OverlaySettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TOverlays.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Threshold for peak finder.'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

procedure TOverlays.CycleUnits;
begin
  { Note - just change top level - do not change units of individual plots }
  if Units = High(TResonanceUnits) then
    Units := Low(TResonanceUnits)
  else
    Units := Succ(Units);
end;

function TOverlays.nVisible: Integer;
var
  i: Integer;
begin
  Result := 0;
  for i := 0 to nPlots - 1 do
    if not Plots[i].Hide then
      Inc(Result);
end;

function TOverlays.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
begin
  { Version info first written in 5.2.860 }
  FReadVersion :=  DecodePgopherVersion(ReadValue(Node, 'Version', ''));
  Result := inherited ReadProperties(Node, False);
end;

{ TStretchablePlot }

function TStretchablePlot.AdjustFrequency(Freq: Double): Double;
begin
  if FixCentre then begin
    { Choose function so that centre of plot is unaffected by FrequencyScale }
    Result := (FHMax+FHMin)/2;
    Result := (Freq-Result)*FrequencyScale + Result + FrequencyOffset;
  end else
    Result := Freq*FrequencyScale + FrequencyOffset;
end;

function TStretchablePlot.DeAdjustFrequency(Freq: Double): Double;
begin
  if FixCentre then begin
    Result := (FHMax+FHMin)/2;
    Result := (Freq - Result - FrequencyOffset)/FrequencyScale + Result;
  end else
    Result := (Freq-FrequencyOffset)/FrequencyScale ;
end;

constructor TStretchablePlot.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',FrequencyScale=1';
  DimensionlessParameters := DimensionlessParameters + ',FrequencyScale';
  inherited;
end;

function TStretchablePlot.GetHMax: Double;
begin
  if FrequencyScale >= 0 then
    Result := AdjustFrequency(FHMax)
  else
    Result := AdjustFrequency(FHMin);
end;

function TStretchablePlot.GetHMin: Double;
begin
  if FrequencyScale < 0 then
    Result := AdjustFrequency(FHMax)
  else
    Result := AdjustFrequency(FHMin);
end;

function TStretchablePlot.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 6 then
    Result := 'Apply FrequencyScale so that centre does not move if set'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

initialization
  TOverlays.RegisterClass(nil);
end.

