unit PgopherImport;
{ Routines for importing and checking calcualtions against the previous pgopher
  version for linear molecules }

  {$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Classes, SysUtils, Mixture, TextStrm;

function ImportPgopherFile(F: TTextStream): TMixture;
function CheckOldPgopherFile(Mixture: TMixture; S: TTextStream; var nTried: Integer; Tolerance: Double; var Failures: Integer; OmegaFromFn: Boolean; SkipF1: Boolean): Double;

implementation

uses {$IFNDEF CONSOLE} Dialogs, {$ENDIF}
     MNode, Math, MathSup, MatrixVec, Life, HundA, Units, QuantumNumber, IntVec,
     Worker, StaticField, DRintensity;

function CheckOldPgopherFile(Mixture: TMixture; S: TTextStream; var nTried: Integer; Tolerance: Double; var Failures: Integer; OmegaFromFn: Boolean; SkipF1: Boolean):Double;
var
  Jp, Symp, Omp, NNp, fnp, np, F1p: Integer; { Upper State }
  J,  Sym,  Om, NN, fn, n, F1: Integer; { Lower State }
  f, str, w, wCalc, Ep, E: Double;
  Buffer: string;
  ManifoldL, ManifoldU: TLinearManifold;
  Molecule: TLinearMolecule;
  TransitionMoments: TTransitionMoments;
  Simulation: TSimulation;
  i, po: Integer;
  D: TMatrix;
  Line: TLine;
  EMult: Double;
  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn(Log, 'Line ', S.Line, ':', Name);
      WriteLn(Log, Jp:3, NNp:3, Symp:2, np:2, J:5, NN:3, Sym:2, n:2, f:12:4, str:8:4, w:8:3, Ep:12:4, E:12:4);
    end;
  end;
  function evNo(EigenValues: TEigenvalues; n: Integer): Integer;
  begin
    Result := 0;
    if (n <= 0) or (n > EigenValues.Size) then
      Fail(Format('Not found %d', [n]))
    else
      Result := n;
  end;
  procedure CheckNo(n1, n2, tol: Double; Name: string);
  var
    Err: Double;
    Test: string;
  begin
    if isNanOrInf(n2) then
      Fail(Format('%s fails: %g Nan', [Name, n1]))
    else begin
      Test := 'Diff';
      if tol < 0 then begin
        if Abs(n1-n2) < 1e-15 then
          Err := Abs(n1-n2) { Permit small absolute errors }
        else begin
          Err := 2*Abs(n1-n2)/(Abs(n1)+Abs(n2));
          Test := 'RelDiff';
        end
      end else
        Err := Abs(n1-n2);
      if Err > Abs(tol) then
        Fail(Format('%s out of tolerance: %g %g %s= %g', [Name, n1, n2, Test, Err]));
      if Err > Result then
        Result := Err;
    end;
  end;
  procedure CheckQno(n1, n2: Integer; Name: string);
  begin
    if n1 <> n2 then
      Fail(Format('Quantum number %s mismatch: Expected %d, Calculated %d', [name, n1, n2]));
  end;
  function GetField(S, E: Integer):string;
  begin
    Result := Copy(Buffer, S+1, E-S+1);
  end;
  function GetInt(S, E: Integer):Integer;
  begin
    if not TryStrToInt(GetField(S,E),Result) then
      Fail('Bad integer:'+GetField(S,E));
  end;

  function ReadOddFloat: Double;
  var
    p: Integer;
    t: string;
  begin
    if S.GetFreeItem then begin
      WriteLn(Log, S.Token);
      if TryStrToFloatEx(S.Token, Result) then
        Exit;
      p := LastDelimiter('-', S.Token);
      if p > 1 then begin
        t := Copy(S.Token, 1, p-1) + 'e' + Copy(S.Token, p, Length(S.Token));
        WriteLn(Log, S.Token, ' ', t);
        if TryStrToFloatEx(t, Result) then
          Exit;
      end;
    end;
    raise Exception.Create('No double');
  end;
  function GetOddDouble(S, E: Integer):Double;
  { Allow for .1055-153 meaning .1055e-153 }
  var
    p: Integer;
    t: string;
  begin
    t := Trim(GetField(S,E));
    if TryStrToFloatEx(t, Result) then
      Exit;
    p := LastDelimiter('-', t);
    if p > 1 then begin
      t := Copy(t, 1, p-1) + 'e' + Copy(t, p, Length(t));
      // WriteLn(Log, GetField(S,E), ' ', t);
      if TryStrToFloatEx(t, Result) then
        Exit;
    end;
    raise Exception.Create('Bad double:'+GetField(S,E));
  end;
  function GetDouble(S, E: Integer):Double;
  begin
    Result := StrToFloatEx(Trim(GetField(S,E)));
  end;
  function GetF1(A:TBasisState):TQuantumNumber;
  var
    B: TLinearBasisState;
  begin
    B := A as TLinearBasisState;
    Result := B.J + B.HyperfineState[0];
  end;
  function FixOm(Value: TQuantumNumber; fn: Integer; Manifold: TManifold):TQuantumNumber;
  var
    Offset: Integer;
  begin
    Result := Value;
    Offset := 0;
    while Result > 500 do begin
      Result := Result - 1000;
      Inc(Offset);
    end;
    if OmegaFromFn then
      with Manifold.State[Offset] as TLinear do
        if Lambda <= 0 then
          if Inverted then
            Result := S - ((fn-1) and not 1)
          else
            Result := fn and not 1
        else if Inverted then
          Result := Lambda + S - (fn-1)*2
        else
          Result := Lambda - S + (fn-1)*2;
  end;
var
  Problem: Exception;
begin
  Result := 0;
  if Mixture.Units = Mhz then
    EMult := 100
  else
    EMult := 1;
  Molecule := Mixture.Species[0].Isotopomer[0] as TLinearMolecule;
  if Molecule.nTransitionMoments <> 1 then Fail(Format('%d transition moments', [Molecule.nTransitionMoments]));
  i := 0;
  if not Molecule.Manifold[0].Initial then Inc(i);
  ManifoldL := Molecule.Manifold[i] as TLinearManifold;
  if Molecule.nManifolds = 1 then begin
    ManifoldU := ManifoldL;
    TransitionMoments := Molecule.TransitionMoments[0,0];
  end else begin
    ManifoldU := Molecule.Manifold[1-i] as TLinearManifold;
    TransitionMoments := Molecule.TransitionMoments[0,1];
  end;
  Simulation := Mixture.Simulations[0];
  po := 52;
  while not S.AtEndOfFile do begin
    po := Pos('ransition Str', S.RestOfLine)-4;
    if po > 0 then Break;
    S.NextLine;
  end;
  S.NextLine;
  while not S.AtEndOfFile do begin
    Buffer := S.RestOfLine;
    if Trim(GetField(0,3)) = '' then begin
      S.NextLine;
      Continue;
    end;
    Jp := GetInt(0,3);   Symp := (1-GetInt(14,15)) div 2; Np := GetInt(16,18);
    if not TryStrToInt(Buffer[46],fnp) then
      fnp := -1;
    Omp := FixOm(GetInt(6,11), fnp, ManifoldU);
    J  := GetInt(21,24); Sym  := (1-GetInt(35,36)) div 2;
    N  := GetInt(37,39);
    if not TryStrToInt(Buffer[48],fn) then
      fn := -1;
    Om := FixOm(GetInt(27,32), fn, ManifoldL);
    NN := GetInt(49,51); NNp := 2*(Ord(Buffer[44])-Ord('q')) + NN;
    if Molecule.nNuclei = 2 then begin
      F1p := GetInt(52,54); F1 := GetInt(55,57);
    end else begin
      F1p := -1;
      F1 := -1;
    end;
    f := GetDouble(po, po+11); str := GetOddDouble(po+12,po+21);
    Ep := GetDouble(po+22,po+33); E := GetDouble(po+34,po+45);
    w := GetDouble(po+56,po+65);
    Inc(nTried);
    Line.EigenvaluesL := ManifoldL.Eigenvalues(J, Sym, [meHrot, meWidth], ZeroField);
    Line.EigenvaluesU := ManifoldU.Eigenvalues(Jp, Symp, [meHrot, meWidth], ZeroField);
    Line.IndexL := evNo(Line.EigenvaluesL, n);
    if Line.IndexL > 0 then begin
      CheckNo(E, Line.EigenvaluesL.E[Line.IndexL], EMult*Tolerance, 'E"');
      CheckQNo(Om, (Line.EigenvaluesL.DominantBasis[Line.IndexL] as TLinearBasisState).Omega, 'Omega"');
      CheckQNo(NN, (Line.EigenvaluesL.DominantBasis[Line.IndexL] as TLinearBasisState).N, 'N"');
      if fn >= 0 then CheckQNo(fn, (Line.EigenvaluesL.DominantBasis[Line.IndexL] as TLinearBasisState).Fn, 'f"');
      if (F1 >= 0) and not SkipF1 then
        if Molecule.Symmetric then
          CheckQNo(F1, Line.EigenvaluesL.DominantBasis[Line.IndexL].HyperfineState[0], 'I12"')
        else
          CheckQNo(F1, GetF1(Line.EigenvaluesL.DominantBasis[Line.IndexL]), 'F1"');
      Line.IndexU := evNo(Line.EigenvaluesU, np);
      if Line.IndexU > 0 then begin
        CheckNo(Ep, Line.EigenvaluesU.E[Line.IndexU], EMult*Tolerance, 'E''');
        CheckQNo(Omp, (Line.EigenvaluesU.DominantBasis[Line.IndexU] as TLinearBasisState).Omega, 'Omega''');
        CheckQNo(NNp, (Line.EigenvaluesU.DominantBasis[Line.IndexU] as TLinearBasisState).N, 'N''');
        if fnp >= 0 then CheckQNo(fnp, (Line.EigenvaluesU.DominantBasis[Line.IndexU] as TLinearBasisState).Fn, 'f''');
        if (F1p >= 0) and not SkipF1 then
          if Molecule.Symmetric then
            CheckQNo(F1p, Line.EigenvaluesU.DominantBasis[Line.IndexU].HyperfineState[0], 'I12''')
          else
            CheckQNo(F1p, GetF1(Line.EigenvaluesU.DominantBasis[Line.IndexU]), 'F1''');
        D := TransitionMoments.TransitionMatrix(Line.EigenValuesL, Line.EigenValuesU, True, True, nil, Simulation);
        TransitionMoments.MakeLine(Line,D,Line.EigenvaluesL,Line.EigenvaluesU,Line.IndexL,Line.IndexU,RandomExcitation,True);
        Problem := nil;
        CheckNo(str, Simulation.LineIntensity(Line, Mixture.Units, Problem), -5*Tolerance, 'Intensity');
        if Assigned(Problem) then begin
          Fail(Problem.Message);
          Problem.Free;
        end;
        // WriteLn(Log, Line.Strength:8:4, LineIntensity(Line, Mixture.Temperature, Mixture.Tvib)/5:8:4, str:8:4);
        wCalc := 0;
        if Assigned(Line.EigenvaluesL.Widths) then
          wCalc := Line.EigenvaluesL.Widths[Line.IndexL];
        if Assigned(Line.EigenvaluesU.Widths) then
          wCalc := wCalc + Line.EigenvaluesU.Widths[Line.IndexU];
        CheckNo(w, wCalc, -5*Tolerance, 'Width');
      end;
    end;
    S.NextLine;
  end;
end;

{$IFDEF VER170}
{$OPTIMIZATION OFF}
{$ENDIF}

function ImportPgopherFile(F: TTextStream): TMixture;
var
  nManifolds, i, i1, i2: Integer;
  Rank: TQuantumNumber;
  t: string;
  Species: TSpecies;
  Molecule: TLinearMolecule;
  ManifoldL, ManifoldU: TLinearManifold;
  Old: Boolean;
  MagneticDipole: Boolean;
  nMagnetic, nErrs: Integer;
  procedure Record1;
  begin
    ManifoldL := TLinearManifold.Create(Molecule);
    ManifoldL.Name := 'Lower';
    ManifoldL.Initial := True;
    if nManifolds = 2 then begin
      ManifoldU := TLinearManifold.Create(Molecule);
      ManifoldU.Name := 'Upper';
    end else if nManifolds = 1 then
      ManifoldU := ManifoldL
    else
      raise Exception.CreateFmt('nstates: %d', [nManifolds]);
    { LimitSearch by default - gives better F1/F2/F3 quantum number values }
    ManifoldL.LimitSearch := True;
    ManifoldU.LimitSearch := True;
    Species.SetSetting(2, F.NextInteger(0)*2); { JMax }
    Rank := F.NextInteger(2);
    Result.Simulations[0].Temperature := F.NextDouble(300);
    F.NextDouble(0); { STRMIN }
    Result.Simulations[0].Fmin := F.NextDouble(0);
    Result.Simulations[0].Fmax := F.NextDouble(0);
    Result.Simulations[0].Foffset := F.NextDouble(0);
    Result.Simulations[0].nDF := F.NextInteger(Result.Simulations[0].nDF);
    Result.PrintLevel := TPrintLevel(Min(Max(Ord(plNone) + F.NextInteger(-3)+3,0),Ord(High(TPrintLevel))));
    F.NextInteger(0); { skip Itype }
    Species.SetSetting(1, F.NextInteger(Species.Jmin)); { Jmin }
    F.NextInteger(0); { skip jumin }
    F.NextInteger(0); { skip jumax }
    i := F.NextInteger(-1);
    { Control how to handle population in simulating intensities: }
    case i of
      { TODO:   2 => Emission Specrum }
      1: if nManifolds = 2 then
           ManifoldU.Initial := True;
    end;
    F.NextInteger(0); { skip lineset }
    F.NextInteger(2); { skip branch }
    if F.NextInteger(0) = 0 { rotrn } then
      Molecule.RSquaredH := True;
    F.NextDouble(-1); { TODO satn }
    Result.Simulations[0].Tvib := F.NextDouble(-1);
    { Nuclear spins }
    i1 := F.NextInteger(0);
    i2 := F.NextInteger(0);
    if i1 > 0 then begin
      Molecule.nNuclei := 1;
      Molecule.Nuclei[0].Spin := i1;
    end;
    if i2 > 0 then begin
      Molecule.nNuclei := Molecule.nNuclei + 1;
      Molecule.Nuclei[Molecule.nNuclei-1].Spin := i2;
    end;
    if F.NextInteger(0) <> 0 then { in_MHz }
      Result.Units := mhz
    else
      Result.Units := cm1;
    Result.Simulations[0].PlotUnits := Result.Units;
    F.NextInteger(0); { skip nstatpr = number of basis states to print }
    { Skip eigenvalues limits }
    F.NextInteger(-1); { skip grmin }
    F.NextInteger(999); { skip grmax }
    F.NextInteger(-1); { skip umin }
    F.NextInteger(999); { skip umax }
    MagneticDipole := F.NextInteger(0) > 0;
    Result.Simulations[0].LifeModel := TLifeModel(F.NextInteger(0));
    Result.Simulations[0].RefWidth := F.NextDouble(0);
    F.NextInteger(999); { skip maxobs }
    F.NextInteger(999); { skip maxpar }
    F.NextInteger(1); { TODO: add_spectra }
    F.NextInteger(2); { TODO: nerrdigits }
    F.NextInteger(1); { TODO: multiple }
    F.NextInteger(0); { TODO: nucoupl - force I = I1+I2 coupling }
    F.NextRecord;
  end;
  procedure ReadPert(Manifold: TLinearManifold; Old: Boolean; ReadWts: Boolean);
  var
    i, nStates, State1, State2: Integer;
    Value: Double;
    Perturbation: TLinearPerturbation;
    Op: TLinearOperator;
    function ReadLabel(ReadWts: Boolean; State: TLinear):Integer;
    var
      ugsym: Integer;
    begin
      Result := F.NextInteger(-2); { ABS(Lambda) }
      if Result >= 0 then begin
        if not Assigned(State) then
          State := TLinear.Create(Manifold);
        State.Lambda := Result;
        if F.NextInteger(0) < 0 then
          State.Lambda := -1;
        State.S := F.NextInteger(0);
        F.NextInteger(0); { always skipped }
        ugsym := F.NextInteger(0);
        Molecule.Symmetric := ugsym <> 0;
        State.gerade := not (ugsym < 0);
      end else
        F.SkipItems(4);
      if ReadWts then begin
        Molecule.SymWt := F.NextInteger(1);
        Molecule.AsymWt := F.NextInteger(1);
      end else
        F.SkipItems(2);
    end;
    procedure ReadParams(StateNumber: Integer; Bvalue: Double);
      procedure ReadNucleus(Nucleus: TLinearNucleus);
      begin
        with Nucleus do begin
          a := F.NextDouble(0);
          b := F.NextDouble(0);
          c := F.NextDouble(0);
          d := F.NextDouble(0);
          eQq0 := F.NextDouble(0);
          eQq2 := F.NextDouble(0);
          cI := F.NextDouble(0);
          aJ := F.NextDouble(0);
          bJ := F.NextDouble(0);
          cJ := F.NextDouble(0);
          dJ := F.NextDouble(0);
          eQq0J := F.NextDouble(0);
          eQq2J := F.NextDouble(0);
        end;
      end;
    begin
      with Manifold.State[StateNumber] as TLinear do begin
        Name := 'State' + IntToStr(StateNumber+1);
        B := Bvalue;
        D := F.NextDouble(0);
        H := F.NextDouble(0);
        A := F.NextDouble(0);
        AD := F.NextDouble(0);
        Gamma := F.NextDouble(0);
        p := F.NextDouble(0);
        q := F.NextDouble(0);
        Origin := F.NextDouble(0);
        LambdaSS := F.NextDouble(0);
        o := F.NextDouble(0);
        lambdaD := F.NextDouble(0);
        oD  := F.NextDouble(0);
        if not Old then begin
          oH      := F.NextDouble(0); { TODO: Check sign }
          oL      := F.NextDouble(0); { TODO: Check sign }
          pD      := F.NextDouble(0);
          pH      := F.NextDouble(0);
          pL      := F.NextDouble(0);
          qD      := F.NextDouble(0);
          qH      := F.NextDouble(0);
          qL      := F.NextDouble(0);
          gammaD  := F.NextDouble(0);
          gammaH  := F.NextDouble(0);
          gammaL  := F.NextDouble(0);
          L       := F.NextDouble(0);
          M       := F.NextDouble(0);
          PP {P}  := F.NextDouble(0);
          Width   := F.NextDouble(0);
          theta   := F.NextDouble(0);
          if Molecule.nNuclei > 0 then begin
            AdjustNuclei(True);
            if Molecule.Symmetric and (Molecule.nNuclei = 2) then
              Nucleus(0).AsNext := True;
            ReadNucleus(Nucleus(0) as TLinearNucleus);
            if (Molecule.nNuclei > 1) and not Molecule.Symmetric then
              ReadNucleus(Nucleus(0) as TLinearNucleus);
          end else
            F.SkipItems(26); { Skip Hyperfine }
          Comment := Trim(F.NextString(''));
          F.NextRecord;
        end;
      end;
    end;
  begin
    if Old then begin
      nStates := 0;
      repeat
        Value := F.NextDouble(0);
        if Value <> 0 then begin
          TLinear.Create(Manifold);
          ReadParams(nStates, Value);
          Inc(nStates);
        end;
      until Value = 0;
      F.NextRecord;
      nStates := ReadLabel(ReadWts, Manifold.State[0] as TLinear);
      if nStates < 0 then
        { Multiple States }
        for i := 1 to -nStates do
          ReadLabel(False, Manifold.State[i-1] as TLinear );
      F.NextRecord;
    end else begin
      nStates := ReadLabel(ReadWts, nil);
      F.NextRecord;
      if nStates < 0 then
        { Multiple States }
        for i := 1 to -nStates do begin
          ReadLabel(False, nil);
          F.NextRecord;
        end;
      for i := 0 to Manifold.Count-1 do
        ReadParams(i, F.NextDouble(0));
    end;
    for i := 0 to Manifold.Count-1 do
      with Manifold.State[i] as TLinear do
        { Convert J independent lambda doubling parameter }
        if (S >= 2) and (Lambda = 2) then begin
          o := o - p - q;
          oD := oD - pD - qD;
          oH := oH - pH - qH;
          oL := oL - pL - qL;
        end else if Lambda <> 4 then begin
          o := 0;
          oD := 0;
          oH := 0;
          oL := 0;
        end;
    if nStates < 0 then
      { Perturbations }
      repeat
        i := F.NextInteger(0);
        if i <> 0 then begin
          Perturbation := TLinearPerturbation.Create(Manifold);
          Perturbation.Value := F.NextDouble(0);
          case i of
             6: begin Op := Homog; Inc(nerrs);
                      WriteLn(Log, 'Warning: Conversion of perturbation type 6 (Homog) is not exact, and may require scaling by Sqrt(2) or a sign change');
                      { TODO: Try some test cases }
                end;
            -2: Op := Luncouple;
            -3: begin Op := Luncouple; Perturbation.n := 2; end;
            -4: Op := Suncouple;
            -9: Op := LS;
            -500: begin Op := aL; Perturbation.Nucleus := 1; end;
            -501: begin Op := e; Perturbation.Nucleus := 1; end;
            -502: begin Op := eQq1; Perturbation.Nucleus := 1; end;
            80..120: begin Op := Homog; Perturbation.OmegaSelect := i-100; Perturbation.n := 2; end;
            180..220: begin Op := Homog; Perturbation.OmegaSelect := i-200; Perturbation.n := 2; Perturbation.SymSelect := 0; end;
            280..320: begin Op := Homog; Perturbation.OmegaSelect := i-300; Perturbation.n := 2; Perturbation.SymSelect := 1; end;
            380..420: begin Op := Homog; Perturbation.OmegaSelect := i-400; Perturbation.n := 0; end;
            480..520: begin Op := Homog; Perturbation.OmegaSelect := i-500; Perturbation.n := 0; Perturbation.SymSelect := 0; end;
            580..620: begin Op := Homog; Perturbation.OmegaSelect := i-600; Perturbation.n := 0; Perturbation.SymSelect := 1; end;
          else
            Inc(nErrs);
            WriteLn(Log, 'Perturbation type ', i, ' not implemented (Value =', Perturbation.Value, ')');
            FreeAndNil(Perturbation);
          end;
          if Assigned(Perturbation) then begin
            Perturbation.Op := Op;
            State1 := F.NextInteger(0); { State index }
            State2 := F.NextInteger(0); { State index }
            Perturbation.SetBraKetName(Manifold.State[State1].Name, Manifold.State[State2].Name);
          end;
        end;
        F.NextRecord;
      until i = 0;
  end;
  procedure ReadTrans;
  var
    Component, StateG, StateU: Integer;
    Value: Double;
    M: TSphericalTransitionMoment;
    MS: TTransitionMoments;
  begin
    MS := nil;
    repeat
      Component := F.NextInteger(-99);
      if Component <> -99 then begin
        StateG := F.NextInteger(0);
        StateU := F.NextInteger(0);
        Value := F.NextDouble(0);
        { Note - put upper state as Bra for Phase consistency }
        if not Assigned(MS) then
          MS := TTransitionMoments.Create(Molecule, ManifoldU.Name, ManifoldL.Name);
        if MagneticDipole then begin
          if Component in [0..1] then begin
            M := TSpinTransitionMoment.Create(MS,ManifoldU.State[StateU].Name, ManifoldL.State[StateG].Name);
          end else begin
            M := TOrbitalTransitionMoment.Create(MS,ManifoldU.State[StateU].Name, ManifoldL.State[StateG].Name);
          end;
          Inc(nMagnetic);
          if (nMagnetic = 2) or (Value <> 1) then begin
            WriteLn(Log, 'Warning: Relative magnetic dipole strengths are not consistent with provious pgopher');
            Inc(nErrs);
            nMagnetic := 3; { Only show message once }
          end;
          M.Component := (Component mod 2) *2;
        end else begin
          M := TSphericalTransitionMoment.Create(MS,ManifoldU.State[StateU].Name, ManifoldL.State[StateG].Name);
          M.Component := Component*2;
        end;
        M.Strength := Value;
        M.Rank := Rank;
      end;
      F.NextRecord;
    until Component = -99;
  end;
begin
  nMagnetic := 0;
  nErrs := 0;
  Result := TMixture.Create(nil);
  try
    Result.Name := 'Mixture';
    t := LowerCase(F.RestOfLine);
    Old := Pos('pgopher',t) <= 0;
    if Old then
      F.Position := 0;
    Species := TSpecies.Create(Result);
    Species.Name := 'S';
    Molecule := TLinearMolecule.Create(Species);
    Molecule.Name := 'M';
    nManifolds := F.NextInteger(2);
    Record1;
    ReadPert(ManifoldL, Old, True);
    if nManifolds > 1 then
      ReadPert(ManifoldU, Old, False);
    ReadTrans;
    F.NextRecord; { Skip emin, emax, a0, a1, a2,
                    x0, x1, x2, pmin, pmax, efm(1), efm(2), arkmag, nbaspr,
                    junk, parsweep }
    Result.Simulations[0].Lorentzian := F.NextDouble(0);
    Result.Simulations[0].Gaussian := F.NextDouble(0);
    F.NextDouble(0); { skip sine }
    F.NextDouble(0); { skip tau }
    F.NextDouble(0); { skip contoff }
    F.NextDouble(1); { skip contscale }
    F.SkipItems(4); { skip intensity }
    F.NextDouble(1); { skip stretch }
    F.NextDouble(0); { skip overlay_baseline }
    F.NextInteger(0); { skip overlay_baseline_fix }
    F.NextDouble(1); { skip overlay_scale }
    F.NextInteger(0); { skip overlay_scale_fix }
    Result.Simulations[0].SMargin := F.NextDouble(Result.Simulations[0].SMargin);
    {$IFNDEF CONSOLE}
    if nErrs > 0 then
      ShowMessage(IntToStr(nErrs) + ' problems found. See log window for details');
    {$ENDIF}
  except
    Result.Free;
    raise;
  end;
end;

end.
