unit SpectrumPlot;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses GStyle, Gmarks, GraphI, GraphM, GraphRange,
     Units, Graphics, MNode, Mixture, Classes, Life,
     SysUtils, QuantumNumber, MathSup, QuantumMath;

type

  TAbstractPlot = class(TSpectrum)
  private
    FFOffset: Double;
    FRefWidth: Double;
    FLifeModel: TLifeModel;
    FIntensityUnits: TIntensityUnits;
  protected
    procedure AdjustWidth(var Width, Strength: Double);
  public
    IMin, IMax: Double; { Vertical limits used in plotting }
    procedure GetColoursUsed(var Colours: TColours); virtual;
    procedure InitializeFromSimulation(Simulation: TSimulation); virtual;
    function FindLimits:Boolean; virtual; { Return True if not Flat }
    procedure MakeFromSimulation(Simulation: TSimulation; MakeLineList: Boolean; out Problem: Exception);
    property FOffset: Double read FFOffset write FFOffset;
    property RefWidth: Double read FRefWidth write FRefWidth;
    property LifeModel: TLifeModel read FLifeModel write FLifeModel;
    property IntensityUnits: TIntensityUnits read FIntensityUnits write FIntensityUnits;
    procedure Show(G: TGraphM; PlotRange:TPlotRange; Offset: Boolean; ShowYLabel: Boolean; var PlotRanges:TPlotRanges); virtual; abstract;
  end;

  TConvolutionParameters = record
    Gaussian: Double;
    Lorentzian: Double;
    WidthMult: Integer;
    SMargin: Double;
    Doppler: Double;
  end;
  { TSpectrumPlot }

  TSpectrumPlot = class(TAbstractPlot)
  private
    FIntensityInRange: Double;
    FnDF: Integer; { Number of points spectrum required at }
    WorknDF: Integer; { Total number of points spectrum will be calculated at }
    FnExtra: Integer; { Number of additional points calculated at either side }
    FnLinesInRange: Integer;
    LineShape: TDoubleArray;
    FConvoluted: Boolean;
    FNoConvolution: Boolean;
    FConvolutionParameters: TConvolutionParameters;
    FSomeWidthsSet: Boolean;
    FSomeWidthsNotSet: Boolean;
    FPlot: TDoubleArray; { The Spectrum }
    function GetPlot(AIndex: Integer): Double;
  protected
    OThreshold: Double;
    procedure SetnDF(const Value: Integer); virtual;
  public
    WidthSet: Boolean;
    NoBaseLine: Boolean; { If a zero width spectrum is shown with a baseline }
    procedure InitializeFromSimulation(Simulation: TSimulation); override;
    property nDF: Integer read FnDF write SetnDF; { Number of points to simulate spectrum at }
    property ConvolutionParameters:TConvolutionParameters read FConvolutionParameters;
    property Convoluted: Boolean read FConvoluted;
    { TODO: Tau, Sine / Generalize }
    procedure AddLine(AFrequency, AStrength, AWidth: Double; AColour: TColor; AUnits: TResonanceUnits); overload; override;
    procedure AddLine(const Line: TLine; Simulation: TSimulation; AUnits: TResonanceUnits; var Problem: Exception); overload; override;
    procedure Clear; override;
    procedure Convolute;
    function FindLimits:Boolean; override;
    procedure MakeLineShape(const AConvolutionParameters:TConvolutionParameters);
    function ConsitentUnits: Boolean;
    function IntegratedUints: Boolean;
    procedure Show(G: TGraphM; PlotRange:TPlotRange; Offset: Boolean; ShowYLabel: Boolean; var PlotRanges:TPlotRanges); override;
    property nLinesInRange: Integer read FnLinesInRange;
    property IntensityInRange: Double read FIntensityInRange;
    property Plot[AIndex: Integer]:Double read GetPlot;
  end;

  TMultiColouredSpectrumPlot = class(TSpectrumPlot)
  private
    FSimulation: TSimulation;
    Plots: array of TSpectrumPlot;
  public
    constructor Create(ASimulation: TSimulation);
    destructor Destroy; override;
    procedure Clear; override;
    procedure AddLine(AFrequency, AStrength, AWidth: Double; AColour: TColor; AUnits: TResonanceUnits); override;
    function FindLimits:Boolean; override;
    procedure GetColoursUsed(var Colours: TColours); override;
    procedure Show(G: TGraphM; PlotRange:TPlotRange; Offset: Boolean; ShowYLabel: Boolean; var PlotRanges:TPlotRanges); override;
  end;

  TFortratPlot = class(TAbstractPlot)
  private
    JMin, JMax: Integer;
    Stream: TMemoryStream;
    MaxIntensity: Double;
    MaxSym: Integer;
    MaxDeltaJ: Integer;
    MaxStateNo: Smallint;
    MoleculeType: TMoleculeClass; { If only one molecue, use that for symmetry labels }
    SampleMolecule: TMolecule;
    ColoursUsed: TColours;
    GraphStyle: TGraphStyle;
    PlotOffset: array of Double;
  public
    MinI: Double;
    ShowSymmetry: Boolean;
    ShowDeltaJ: Boolean;
    ScaleMarkSize: Boolean;
    UseUpper: Boolean;
    UseSymmetry: Boolean;
    UseStateNumber: Boolean;
    PlotQno: TQnoType;
    constructor Create(AUnits: TResonanceUnits);
    destructor Destroy; override;
    procedure Clear; override;
    procedure AddLine(const Line: TLine; Simulation: TSimulation; AUnits: TResonanceUnits; var Problem: Exception); override;
    procedure GetColoursUsed(var Colours: TColours); override;
    procedure InitializeFromSimulation(Simulation: TSimulation); override;
    {$IFNDEF CONSOLE}
    procedure GetPlotY(Line: TLine; out PlotNumber: Integer; out Y: Double);
    {$ENDIF}
    procedure Show(G: TGraphM; PlotRange:TPlotRange; Offset: Boolean; ShowYLabel: Boolean; var PlotRanges:TPlotRanges); override;
  end;

  TFitPlot = class(TSpectrumPlot)
  private
    FnDFLockout: Boolean;
  protected
    procedure SetnDF(const Value: Integer); override;
  public
    MaxIntensity: Double;
    Frequency: TDoubleArray; { The required frequencies }
    procedure Clear; override;
    procedure AddLine(AFrequency, AStrength, AWidth: Double; AColour: TColor; AUnits: TResonanceUnits); override;
    procedure InitializeFromSimulation(Simulation: TSimulation); override;
    procedure AddIn(LineList: TLineList; Simulation: TSimulation; var Problem: Exception; FirstLine: Integer = 0; Increment: Integer = 1); override;
  end;


implementation

uses {$IFNDEF CONSOLE} ClipBrd, Huge, {$IFDEF MSWINDOWS} Windows, {$ENDIF} {$ENDIF}
     Progress, TextStrm, Math, Wofz, MNodeColour, Worker, Sysinfo, PGWork;

{ TAbstractPlot }

function TAbstractPlot.FindLimits:Boolean;
begin
  { Would be abstract, only no action might well be OK }
  Result := False;
end;

procedure TAbstractPlot.GetColoursUsed(var Colours: TColours);
begin
  Colours := nil;
end;

procedure TAbstractPlot.InitializeFromSimulation(Simulation: TSimulation);
begin
  FOffset := Simulation.FOffset;
  FMin := Simulation.FMin;
  FMax := Simulation.FMax;
  Temperature := Simulation.Temperature;
  RefWidth := Simulation.RefWidth;
  LifeModel := Simulation.LifeModel;
  Units := Simulation.PlotUnits;
  FIntensityUnits := Simulation.IntensityUnits;
  Clear;
end;

procedure TAbstractPlot.MakeFromSimulation(Simulation: TSimulation; MakeLineList: Boolean; out Problem: Exception);
begin
  PushProgress('Simulating Spectrum');
  try
    Problem := nil;
    if not Assigned(Simulation.Mixture.LineList) and MakeLineList then
      Simulation.MakeLineList(Problem);
    InitializeFromSimulation(Simulation);
    if not Assigned(Problem) then
      if Assigned(Simulation.Mixture.LineList) then
        AddIn(Simulation.Mixture.LineList, Simulation, Problem)
      else
        Simulation.Mixture.AddToSpectrum(Self, Problem, Simulation);
  finally
    PopProgress;
  end;
end;

procedure TAbstractPlot.AdjustWidth(var Width, Strength: Double);
begin
  case LifeModel of
    lmProduct, lmProductWidth:
      { Scale Peak intensity as width/(width + RefWidth) }
      if RefWidth > 0 then
        Strength := Strength * Width / (Width + RefWidth)
      else
        Strength := Strength * Width;
    lmParent, lmParentWidth:
      { Scale Peak intensity as 1/(width + RefWidth) }
      if Width + RefWidth > 0 then
        Strength := Strength / ( Width + RefWidth);
    lmGate:
      { Scale Peak intensity as exp(-width * RefWidth) }
      Strength := Strength * Exp(-Width*RefWidth);
  end;
  if LifeModel in [lmNone, lmProduct, lmParent, lmGate] then
    Width := 0; { Discard width if required }
end;

{ TSpectrumPlot }

procedure TSpectrumPlot.AddLine(AFrequency, AStrength, AWidth: Double; AColour: TColor; AUnits: TResonanceUnits);
var
  Centre, nl, i, Lines, nParts: Integer;
  df, a, w1, Offset: Double;
begin
  if (AStrength = 0) or (AColour = clHide) then Exit;
  AdjustWidth(AWidth, AStrength);
  AWidth := ConvertWidth(AWidth, AFrequency, AUnits, Units);
  AFrequency := Convert(AFrequency, AUnits, Units);
  df := (FMax-FMin)/(nDF-1);
  { Add each line twice if Doppler split }
  if ConvolutionParameters.Doppler <> 0 then begin
    if IsEnergyUnit(Units) then
      Offset := AFrequency*ConvolutionParameters.Doppler
    else begin
      a := Convert(AFrequency,Units,cm1);
      Offset := ConvertWidth(a*ConvolutionParameters.Doppler,a,cm1,Units);
    end;
    AFrequency := AFrequency - Offset;
    nParts := 2;
    AStrength := AStrength / 2;
  end else begin
    Offset := 0;
    nParts := 1;
  end;
  for Lines := 1 to nParts do begin
    if Lines = 2 then
      AFrequency := AFrequency + 2*Offset;
    a := (AFrequency - FMin)/df + FnExtra;
    { Prevent overflows on conversion to integer }
    if Abs(a) < High(Integer) div 2 then
      Centre := Round(a)
    else
      Centre := High(Integer) div 2;
    if AWidth > 0 then
      FSomeWidthsSet := True
    else
      FSomeWidthsNotSet := True;
    if AWidth > Abs(3*df) then begin
      { Add in Lorentzian to plot }
      nl := Round(ConvolutionParameters.WidthMult * 2 * AWidth / Abs(df) + 4);
      if (Centre >= -nl) and (Centre-nl < WorknDF) then begin
        WidthSet := True;
        a := AStrength * AWidth / 2 / Pi / Abs(df);
        if Length(LineShape) = 1 then
          a := a / Abs(df) / LineShape[0]; { Full convolution done here }
        w1 := Sqr(AWidth) / 4 / df / df;
        for i := Max(Centre-nl,0) to Min(Centre+nl,WorknDF-1) do
          FPlot[i] := FPlot[i] + a/(Sqr(i-Centre) + w1);
      end else
        Continue;
    end else if (Centre >= 0) and (Centre < WorknDF) then begin
      if (AWidth > 0) and (Length(LineShape) = 1) then
        { Stick spectrum - make intensities consistent by taking intensity at centre of peak }
        { Do full convolution here to take account of case where both widths non-zero,
          but still a stick spectrum }
        AStrength := AStrength * Voigt(0,AWidth+ConvolutionParameters.Lorentzian, ConvolutionParameters.Gaussian) / LineShape[0];
      FPlot[Centre] := FPlot[Centre] + AStrength;
    end else
      Continue;
    Inc(FnLinesInRange);
    FIntensityInRange := FIntensityInRange + AStrength;
  end;
end;

procedure TSpectrumPlot.AddLine(const Line: TLine; Simulation: TSimulation; AUnits: TResonanceUnits; var Problem: Exception);
var
  Centre, df: Double;
begin
  { Speed up only - throw away out of range lines quickly, without calculating intensity }
  if Line.Width = 0 then begin
    Centre := Convert(Line.Frequency, AUnits, Units);
    df := (FMax-FMin)/(nDF-1);
    Centre := (Centre - FMin)/df + FnExtra;
    if (Centre < 0) or (Centre >= WorknDF) then
      Exit;
  end else if Line.Width < 0 then begin
    if not Assigned(Problem) then
      Problem := EWarningException.Create('Lines with negative width discarded');
    Exit;
  end;
  AddLine(Line.Frequency, Simulation.LineIntensity(Line, AUnits, Problem), Line.Width, Line.Colour, AUnits);
end;

procedure TSpectrumPlot.Clear;
begin
  inherited;
  FConvoluted := False;
  if Length(FPlot) > 0 then
    FillChar(FPlot[0], Length(FPlot)*SizeOf(Double), 0);
  WidthSet := False;
  FSomeWidthsSet := False;
  FSomeWidthsNotSet := False;
  FnLinesInRange := 0;
  FIntensityInRange := 0;
end;

procedure TSpectrumPlot.Convolute;
var
  x: TDoubleArray;
  i, k, nLw: Integer;
begin
  if Convoluted then Exit;
  nLw := (Length(LineShape)-1) div 2;
  SetLength(x, Length(FPlot));
  FillChar(x[0], Length(FPlot)*SizeOf(Double), 0);
  for i := 0 to High(FPlot) do
    if FPlot[i] <> 0 then
      for k := Max(i-nLw,0) to Min(i+nLw, High(FPlot)) do
        x[k] := x[k] + FPlot[i]*LineShape[nLW+k-i];
  for i := 0 to High(FPlot) do
    FPlot[i] := x[i];
  FConvoluted := True;
end;

function TSpectrumPlot.FindLimits:Boolean;
begin
  Convolute;
  Limits(FPlot, IMin, IMax, FnExtra, FnExtra + nDf - 1);
  Result := IMax <> IMin;
  If not Result then
    IMax := IMin + 1; { Flat spectrum - quick fix to avoid divide by zero }
end;

function TSpectrumPlot.GetPlot(AIndex: Integer): Double;
begin
  Result := FPlot[AIndex+FnExtra];
end;

procedure TSpectrumPlot.InitializeFromSimulation(Simulation: TSimulation);
begin
  inherited;
  OThreshold := Simulation.OThreshold;
  with FConvolutionParameters do begin
    SMargin := Simulation.SMargin;
    Lorentzian := Simulation.Lorentzian;
    Gaussian := Simulation.Gaussian;
    WidthMult := Simulation.WidthMult;
    Doppler := Simulation.Doppler;
  end;
  SetnDF(Simulation.nDF);
end;

procedure TSpectrumPlot.MakeLineShape(const AConvolutionParameters:TConvolutionParameters);
var
  n, i: Integer;
  df: Double;
begin
  df := (FMax-FMin)/(nDF-1);
  FNoConvolution := False;
  if AConvolutionParameters.Lorentzian+AConvolutionParameters.Gaussian <= Abs(3*df) then begin
    SetLength(LineShape, 1);
    { Makes transition between stick and full lineshape consistent by }
    if AConvolutionParameters.Lorentzian+AConvolutionParameters.Gaussian > 0 then
      LineShape[0] := Voigt(0, AConvolutionParameters.Lorentzian, AConvolutionParameters.Gaussian) { approximating stick by value at centre }
    else begin
      LineShape[0] := 1;
      FNoConvolution := True;
    end;
  end else begin
{  TODO: IF (sine .NE. 0) THEN
     n = MIN((dfmult*(2*lor+gau)+ABS(sine))/ABS(df)+4, DBLE(nmax))
   ELSE }
    n := Round(AConvolutionParameters.WidthMult*(AConvolutionParameters.Lorentzian+AConvolutionParameters.Gaussian)/ABS(df)+4);
    if n > nDF then
      n := nDF+1; { Very large lineshape; however only nDF points either
                    side of lineshape will be used }
    SetLength(LineShape, 2*n+1);
    for i := 0 to n do begin
      LineShape[i+n] := Voigt(i*df, AConvolutionParameters.Lorentzian, AConvolutionParameters.Gaussian);
      LineShape[n-i] := LineShape[i+n];
    end;
    WidthSet := True;
  end;
end;

function TSpectrumPlot.ConsitentUnits: Boolean;
begin
  Result := FNoConvolution or not (FSomeWidthsSet and FSomeWidthsNotSet);
end;

function TSpectrumPlot.IntegratedUints: Boolean;
begin
  Result := FNoConvolution and FSomeWidthsNotSet;
end;

procedure TSpectrumPlot.SetnDF(const Value: Integer);
var
  df, dp: Double;
begin
  if Value > 1 then begin
    FnDF := Value;
    df := (FMax-FMin)/(nDF-1);
    if ConvolutionParameters.SMargin >= 0 then
      FnExtra := Round(ConvolutionParameters.SMargin)
    else begin
      dp := Max(Convert(FMin,Units,cm1),Convert(FMin,Units,cm1));
      dp := ConvertWidth(dp*ConvolutionParameters.Doppler,dp,cm1,Units);
      FnExtra := -Round(ConvolutionParameters.WidthMult*(ConvolutionParameters.Lorentzian+ConvolutionParameters.Gaussian+dp)/ABS(df)*ConvolutionParameters.SMargin);
    end;
    if FnExtra > 0 then
      FnExtra := Max(FnExtra, 5);
    if FnExtra > nDf then
      FnExtra := nDF; { TODO: Switch algorithm in this limit }
    WorknDF := nDF+2*FnExtra;
    SetLength(FPlot, WorknDF);
    { This is called here to allow stick spectrum corrections to work properly }
    MakeLineShape(ConvolutionParameters);
  end else
    raise EReportException.Create('nDF must be > 1');
end;

procedure TSpectrumPlot.Show(G: TGraphM; PlotRange:TPlotRange; Offset: Boolean; ShowYLabel: Boolean; var PlotRanges:TPlotRanges);
var
  i: Integer;
  df, x: Double;
  xl, xu, yl, yu: Double;
  AutoScale: Boolean;
begin
  if Colour = clHide then Exit;
  Convolute;
  if Colour <> clNone then
    G.StyleP(Colour, G.LineStyle.Width, G.LineStyle.Pattern);
  G.MapArg(xl,xu,yl,yu);
  AutoScale := IsNullPlotRange(PlotRange);
  if AutoScale then begin
    FindLimits;
    PlotRange := GraphRange.PlotRange(G, 0, 1, Imin, Imax, -9);
  end;
  MapFromPlotRange(G,PlotRange);
  df := (FMax-FMin)/(nDF-1);
  x := FMin-Foffset;
  if WidthSet then begin
    G.Point(x, FPlot[FnExtra]);
    for i := FnExtra+1 to FnExtra+nDF-1 do begin
      x := x + df;
      G.Join(x, FPlot[i]);
    end;
  end else begin
    { Special case for stick spectra - make sure lines are vertical and single }
    if not NoBaseLine then
     G.Point(x, 0);
    for i := FnExtra to FnExtra+nDF-1 do begin
      if FPlot[i] > 0 then
        if NoBaseline then begin
          G.Point(x, 0);
          G.Join(x, FPlot[i]);
        end else begin
          G.Join(x, 0);
          G.Join(x, FPlot[i]);
          G.Join(x, 0);
        end;
      x := x + df;
    end;
    if not NoBaseLine then
      G.Join(x-df, 0);
  end;
  G.Map(xl,xu,yl,yu);
  if ShowYLabel and ConsitentUnits then
    AxisFromPlotRange(PlotRange, G, G.Xmin, NameOfUnit(IntensityUnits, IntegratedUints, Units));
  AddPlotRange(PlotRanges, PlotRange);
end;

{ TFitPlot }

procedure TFitPlot.AddIn(LineList: TLineList; Simulation: TSimulation; var Problem: Exception; FirstLine: Integer; Increment: Integer);
var
  I: Integer;
  Worker: TPgopherWorker;
  nPlots: Integer;
  Plots: array of TFitPlot;
  j: Integer;
  SubProblem: Exception;
begin
  nPlots := Min(nWorkers, nCPUs);
  if (nPlots <= 1) or (Increment > 1) then
    inherited
  else begin
    { Parallel algorithm }
    if IsNormalized(Simulation.IntensityUnits) then
      Simulation.Mixture.PreCalculatePartitionFunctions(Simulation, Problem);
    SetLength(Plots, nPlots);
    Plots[0] := Self;
    SubProblem := nil;
    for I := 1 to nPlots - 1 do begin
      Plots[I] := TFitPlot.Create(Units, IsDistribution);
      Plots[I].InitializeFromSimulation(Simulation);
      Plots[I].Frequency := Frequency;
      Plots[I].nDF := nDF;
      Plots[I].Clear;
      Worker := FindWorkerWait as TPgopherWorker;
      if Assigned(Worker) then
        Worker.AddIn(Plots[I], LineList, Simulation, SubProblem, I, nPlots)
      else begin
        { Something has failed }
        if Assigned(Worker) then
          Worker.Skip;
        Break;
      end;
    end;
    { Use main thread for one list, so that progress bar can be used }
    inherited AddIn(LineList, Simulation, Problem, 0, nPlots);
    { Combine exception flag }
    if not Assigned(Problem) then
      Problem := SubProblem
    else if Assigned(SubProblem) then
      FreeAndNil(SubProblem);
    WaitNoBusyWorkers;
    for I := 1 to nPlots - 1 do begin
      for j := 0 to High(FPlot) do
        FPlot[j] := FPlot[j] + Plots[I].FPlot[j];
      FreeAndNil(Plots[I]);
    end;
  end;
end;

procedure TFitPlot.AddLine(AFrequency, AStrength, AWidth: Double; AColour: TColor; AUnits: TResonanceUnits);
var
  i: Integer;
  x: Double; { Smallest offset from peak centre }
  y: Double; { Correspondig height }
  Offset: Double; { For Doppler splitting convolution }
  Line, nParts: Integer;
begin
  if AColour = clHide then
    Exit;
  AdjustWidth(AWidth, AStrength);
  AWidth := ConvertWidth(AWidth, AFrequency, AUnits, Units);
  AFrequency := Convert(AFrequency, AUnits, Units);
  if AStrength <= MaxIntensity then
    Exit;
  if ConvolutionParameters.Doppler <> 0 then begin
    if IsEnergyUnit(Units) then
     Offset := AFrequency*ConvolutionParameters.Doppler
    else begin
      Offset := Convert(AFrequency,Units,cm1);
      Offset := ConvertWidth(Offset*ConvolutionParameters.Doppler,Offset,cm1,Units);
    end;
    AFrequency := AFrequency - Offset;
    AStrength := AStrength / 2;
    nParts := 2;
  end else begin
    Offset := 0;
    nParts := 1;
  end;
  for Line := 1 to nParts do begin
    if Line = 2 then
      AFrequency := AFrequency + 2*Offset;
    if AFrequency <= FMin then
      x := FMin - AFrequency
    else if AFrequency >= FMax then
      x := AFrequency - FMax
    else
      x := 0;
    y := AStrength*Voigt(x, AWidth + ConvolutionParameters.Lorentzian, ConvolutionParameters.Gaussian);
    if y <= MaxIntensity then
      Continue;
    if y*OThreshold > MaxIntensity then
      MaxIntensity := y*OThreshold;
    { TODO: Lots of possible optimizations here }
    for i := 0 to High(FPlot) do
      FPlot[i] := FPlot[i] + AStrength*Voigt(Frequency[i]-AFrequency, AWidth + ConvolutionParameters.Lorentzian, ConvolutionParameters.Gaussian);
  end;
end;

procedure TFitPlot.Clear;
begin
  inherited Clear;
  Limits(Frequency, Fmin, FMax);
  MaxIntensity := 0;
end;

procedure TFitPlot.InitializeFromSimulation(Simulation: TSimulation);
begin
  FnDFLockout := True;
  try
    inherited;
  finally
    FnDFLockout := False;
  end;
end;

procedure TFitPlot.SetnDF(const Value: Integer);
begin
  if FnDFLockout then Exit;
  FConvolutionParameters.SMargin := 0;
  inherited;
  SetLength(Frequency, nDF);
end;

{ TMultiColouredSpectrumPlot }

procedure TMultiColouredSpectrumPlot.AddLine(AFrequency, AStrength, AWidth: Double; AColour: TColor; AUnits: TResonanceUnits);
var
  i: Integer;
  CurrentPlot: TSpectrumPlot;
begin
  if AColour = clHide then
    Exit;
  inherited; { Always calculate sum, even if the sum is not displayed so that scaling is consistent }
  CurrentPlot := nil;
  for i := 0 to High(Plots) do
    if Plots[i].Colour = AColour then begin
      CurrentPlot := Plots[i];
      Break;
    end;
  if not Assigned(CurrentPlot) then begin
    SetLength(Plots, Length(Plots)+1);
    CurrentPlot := TSpectrumPlot.Create(FSimulation.PlotUnits, IsDistribution);
    CurrentPlot.InitializeFromSimulation(FSimulation);
    CurrentPlot.Colour := AColour;
    Plots[Length(Plots)-1] := CurrentPlot;
  end;
  CurrentPlot.AddLine(AFrequency, AStrength, AWidth, AColour, AUnits);
end;

procedure TMultiColouredSpectrumPlot.Clear;
var
  i: Integer;
begin
  inherited;
  for i := 0 to High(Plots) do
    Plots[i].Free;
  SetLength(Plots,0);
end;

procedure TMultiColouredSpectrumPlot.GetColoursUsed(var Colours: TColours);
var
  i: Integer;
begin
  SetLength(Colours, Length(Plots));
  for i := 0 to High(Plots) do
    Colours[i] := Plots[i].Colour;
end;

constructor TMultiColouredSpectrumPlot.Create(ASimulation: TSimulation);
begin
  inherited Create(ASimulation.PlotUnits, IsPopDist(ASimulation.IntensityUnits));
  FSimulation := ASimulation;
end;

destructor TMultiColouredSpectrumPlot.Destroy;
begin
  Clear;
  inherited;
end;

function TMultiColouredSpectrumPlot.FindLimits: Boolean;
var
  i: Integer;
begin
  Result := inherited FindLimits;
  if not Result then
    IMax := IMin;
  for i := 0 to High(Plots) do begin
    if Plots[i].FindLimits then
      IMax := Max(IMax, Plots[i].IMax);
    IMin := Min(IMin, Plots[i].IMin);
  end;
  Result := IMax <> IMin;
  If not Result then
    IMax := IMin + 1; { Flat spectrum - quick fix to avoid divide by zero }
end;

procedure TMultiColouredSpectrumPlot.Show(G: TGraphM; PlotRange:TPlotRange;
            Offset: Boolean; ShowYLabel: Boolean; var PlotRanges:TPlotRanges);
var
  i, nPlots: Integer;
  AutoScale: Boolean;
  function PlotRangeFor(nPlot: Integer; Tag:Integer):TPlotRange;
  begin
    if not AutoScale then
      Result := PlotRange
    else
      if Offset then
        Result := GraphRange.PlotRange(G, nPlot/nPlots, (nPlot+0.95)/nPlots, Imin, Imax, Tag)
      else
        Result := GraphRange.PlotRange(G, 0, 1, Imin, Imax, Tag);
  end;
begin
  AutoScale := IsNullPlotRange(PlotRange);
  if AutoScale then
    FindLimits;
  { Calculate number of plots }
  nPlots := Length(Plots);
  if FSimulation.ShowSum then
    Inc(nPlots);
  if nPlots = 0 then Exit;
  for i := 0 to High(Plots) do
    Plots[i].Show(G, PlotRangeFor(i,-10-i), False, ShowYLabel and (Offset or (nPlots = 1) or ConsitentUnits), PlotRanges);
  if FSimulation.ShowSum then begin
    G.Style(0);
    inherited Show(G, PlotRangeFor(nPlots-1,-9), Offset, ShowYLabel, PlotRanges);
  end;
end;

{ TFortratPlot }

type
  TFortratRec = packed record
    Frequency: Double;
    Strength:  Double;
    Colour:    TColor;
    J:         Smallint;
    StateNo:   Smallint;
    DeltaJ:    Shortint;
    Sym:       Shortint;
  end;

procedure TFortratPlot.AddLine(const Line: TLine; Simulation: TSimulation; AUnits: TResonanceUnits; var Problem: Exception);
var
  c: Double;
  FortratRec: TFortratRec;
  Width, f: Double;
  i: Integer;
  Molecule: TMolecule;
  procedure LoadRec(Ev: TEigenvalues; Index: Integer);
  begin
    Molecule := Ev.Manifold.Molecule;
    FortratRec.Sym := Molecule.JAdjustedSym(Ev.J, Ev.Sym);
    FortratRec.StateNo := Index;
    FortratRec.J := Ev.DominantBasis[Index].Qno(PlotQno);
  end;
begin
  if Line.Colour = clHide then
    Exit;
  f := Convert(Line.Frequency, AUnits, Units);
  c := (f - FMin)/(FMax - FMin);
  if (c>=0) and (c <=1) and Assigned(Line.EigenvaluesL) then begin
    FortratRec.Frequency := f - FOffset;
    if UseUpper and Assigned(Line.EigenvaluesU) then
      LoadRec(Line.EigenvaluesU, Line.IndexU)
    else if Assigned(Line.EigenvaluesL) then
      LoadRec(Line.EigenvaluesL, Line.IndexL);
    if MoleculeType = nil then begin
      MoleculeType := TMoleculeClass(Molecule.ClassType);
      SampleMolecule := Molecule;
    end else if Assigned(SampleMolecule) and (MoleculeType <> Molecule.ClassType) then
      SampleMolecule := nil;
    if Assigned(Line.EigenvaluesU) and Assigned(Line.EigenvaluesL) then
      FortratRec.DeltaJ := Line.EigenvaluesU.DominantBasis[Line.IndexU].Qno(PlotQno)
                           - Line.EigenvaluesL.DominantBasis[Line.IndexL].Qno(PlotQno)
    else
      FortratRec.DeltaJ := 0;
    if Abs(FortratRec.DeltaJ) > MaxDeltaJ then MaxDeltaJ := Abs(FortratRec.DeltaJ);
    if FortratRec.StateNo > MaxStateNo then MaxStateNo := FortratRec.StateNo;
    FortratRec.Colour := Line.Colour;
    if FortratRec.Colour = clNone then
      FortratRec.Colour := clBlack;
    if MaxSym < FortratRec.Sym then
      MaxSym := FortratRec.Sym;
    if not Assigned(Simulation) then
      Simulation := Line.EigenvaluesL.Manifold.Mixture.Simulations[0];
    FortratRec.Strength := Simulation.LineIntensity(Line, AUnits, Problem);
    Width := Line.Width;
    AdjustWidth(Width, FortratRec.Strength);
    // Not needed as only Strength required: Width := ConvertWidth(Width, Line.Frequency, Mixture.Units, Units);
    Stream.Write(FortratRec, SizeOf(FortratRec));
    if (FnLines = 0) or (JMin > FortratRec.J) then begin
      JMin := FortratRec.J;
      IMin := FortratRec.J / 2.;
    end;
    if (FnLines = 0) or (FortratRec.J > JMax) then begin
      JMax := FortratRec.J;
      IMax := JMax / 2.;
    end;
    if MaxIntensity < FortratRec.Strength then
      MaxIntensity := FortratRec.Strength;
    Inc(FnLines);
    for i := 0 to High(ColoursUsed) do
      if FortratRec.Colour = ColoursUsed[i] then
        Exit;
    SetLength(ColoursUsed, Length(ColoursUsed)+1);
    ColoursUsed[Length(ColoursUsed)-1] := FortratRec.Colour;
  end;
end;

procedure TFortratPlot.Clear;
begin
  inherited;
  IMin := 0;
  IMax := 0;
  JMax := -1;
  MaxIntensity := 0;
  MaxDeltaJ := 0;
  MaxSym := -1;
  MaxStateNo := 0;
  Stream.Free;
  Stream := TMemoryStream.Create;
  SetLength(ColoursUsed, 0);
  MoleculeType := nil;
  SampleMolecule := nil;
end;

constructor TFortratPlot.Create(AUnits: TResonanceUnits);
begin
  inherited Create(AUnits,False);
end;

destructor TFortratPlot.Destroy;
begin
  Clear;
  FreeAndNil(Stream);
  inherited;
end;

procedure TFortratPlot.GetColoursUsed(var Colours: TColours);
begin
  Colours := ColoursUsed;
end;

procedure TFortratPlot.InitializeFromSimulation(Simulation: TSimulation);
begin
  inherited;
  MinI := Simulation.MinI;
  ShowSymmetry := Simulation.ShowSymmetry;
  ShowDeltaJ := Simulation.ShowDeltaJ;
  ScaleMarkSize := Simulation.ScaleMarkSize;
  UseUpper := Simulation.UseUpper;
  UseSymmetry := Simulation.UseSymmetry;
  UseStateNumber := Simulation.UseStateNumber;
  PlotQno := Simulation.FortratQno;
  GraphStyle := Simulation.GraphStyle;
end;

{$IFNDEF CONSOLE}
procedure TFortratPlot.GetPlotY(Line: TLine; out PlotNumber: Integer; out Y: Double);
var
  i: Integer;
  Offset: Boolean;
begin
  if (UseUpper and not Assigned(Line.EigenvaluesU))
     or (not UseUpper and not Assigned(Line.EigenvaluesL)) then
    Exit; { Eigenvalues not available }
  { Compute plot number }
  PlotNumber := 0;
  Offset := GraphStyle in [gsOffsetAll, gsInvertedAll];
  if Offset then
    for i := 0 to High(ColoursUsed) do
      if Line.Colour = ColoursUsed[i] then begin
        PlotNumber := i;
        Break;
      end;
  if UseSymmetry and (MaxSym >= 0) then
    PlotNumber := PlotNumber*(MaxSym+1) + IfThen(UseUpper, Line.EigenvaluesU.Sym, Line.EigenvaluesL.Sym);
  if UseStateNumber then
    PlotNumber := PlotNumber*MaxStateNo + IfThen(UseUpper, Line.IndexU, Line.IndexL) - 1;
  { Compute effective J }
  if UseUpper then
    Y := Line.EigenvaluesU.DominantBasis[Line.IndexU].Qno(PlotQno)
  else
    Y := Line.EigenvaluesL.DominantBasis[Line.IndexL].Qno(PlotQno);
  Y := (Y/2 - PlotOffset[0])/Length(PlotOffset) + PlotOffset[PlotNumber];
end;
{$ENDIF}

procedure TFortratPlot.Show(G: TGraphM; PlotRange:TPlotRange; Offset: Boolean; ShowYLabel: Boolean; var PlotRanges:TPlotRanges);
var
  FortratRec: TFortratRec;
  MarkStyle: TMarkStyle;
  BaseSize: Single;
  i, m, iSym, MarkNumber, nPlots: Integer;
  o: Double;
  s: string;
  function SymLabel(Sym: Integer): string;
  begin
    if Assigned(SampleMolecule) then begin
      Result := SampleMolecule.JAdjustedSymName(0,Sym);
      if Result <> IntToStr(Sym) then
        Result := IntToStr(Sym) + '/' + Result;
    end else
      Result := IntToStr(Sym);
  end;
begin
  if nLines = 0 then
    Exit;
  if not IsNullPlotRange(PlotRange) then
    raise NotYet; { Actually not required }
  m := 0;
  if ShowSymmetry and (MaxSym >= 0) then begin
    s := 'Symmetry: ';
    for iSym := 0 to MaxSym do
      if ShowDeltaJ then
        for i := -MaxDeltaJ div 2 to MaxDeltaJ div 2 do begin
          s := s + #1 + Char(Ord(FirstMark)+m) + #0':' + SymLabel(iSym) + DeltaJName(i*2) + ' ';
          Inc(m);
        end
      else
        s := s + #1 + Char(Ord(FirstMark)+iSym) + #0':' + SymLabel(iSym) + ' ';
  end else if ShowDeltaJ then begin
    s := 'DeltaJ: ';
    for i := -MaxDeltaJ div 2 to MaxDeltaJ div 2 do begin
      s := s + #1 + Char(Ord(FirstMark)+m) + #0':' + DeltaJName(i*2) + ' ';
      Inc(m);
    end;
  end;
  G.Title2(s);
  Stream.Position := 0;
  MarkStyle := TMarkStyle.Create;
  MarkStyle.DotAtCentre := True;
  MarkStyle.Size := 2;
  BaseSize := G.MarkStyles[0].Size;
  nPlots := 1;
  if Offset then
    nPlots := Length(ColoursUsed);
  if UseSymmetry and (MaxSym >= 0) then
    nPlots := nPlots * (1+MaxSym);
  if UseStateNumber then
    nPlots := nPlots * MaxStateNo;
  SetLength(PlotOffset, nPlots);
  o := G.Ymin;
  for i := 0 to High(PlotOffset) do begin
    PlotOffset[i] := o;
    o := o + (G.Ymax-G.Ymin)/nPlots;
    G.ScaledAxis(IMin, G.Ymax, PlotOffset[i] + (IMin-G.Ymin)/nPlots, o, G.Xmin, False, QnoTypeNames[Ord(PlotQno)]);
    AddPlotRange(PlotRanges, GraphRange.PlotRange(G, PlotOffset[i] + (IMin-G.Ymin)/nPlots, o, IMin, G.Ymax, -10-i));
    { TODO: Separate Y axis labels here }
  end;
  if nPlots = 1 then
    PlotRanges[0].Tag := -9; { only one plot, so need not select on this }
  o := G.Ymin;
  while Stream.Read(FortratRec, SizeOf(FortratRec)) = SizeOf(FortratRec) do
    with FortratRec do
      if Strength >= MinI*MaxIntensity then begin
        MarkStyle.Colour := Colour;
        if ShowDeltaJ then begin
          MarkNumber := (DeltaJ + MaxDeltaJ) div 2;
          if ShowSymmetry then
            Inc(MarkNumber, Sym * (MaxDeltaJ + 1));
        end else if ShowSymmetry then
          MarkNumber := Max(Sym,0)
        else
          MarkNumber := 0;
        MarkStyle.MarkType := TMarkType(Ord(FirstMark) + MarkNumber mod (Ord(LastMark)-Ord(FirstMark)+1));
        if ScaleMarkSize and (MaxIntensity > 0) then
          MarkStyle.Size := 2 * BaseSize * Sqrt(Strength / MaxIntensity);
        if nPlots > 1 then begin
          { If the logic changes here, change GetPlotY to match }
          m := 0;
          if Offset then
            for i := 0 to High(ColoursUsed) do
              if Colour = ColoursUsed[i] then begin
                m := i;
                Break;
              end;
          if UseSymmetry and (MaxSym >= 0) then
            m := m*(MaxSym+1) + Sym;
          if UseStateNumber then
            m := m*MaxStateNo + FortratRec.StateNo - 1;
          o := PlotOffset[m];
        end;
        G.SymbolT(Frequency, (J/2.-G.Ymin)/nPlots+o, MarkStyle);
      end;
  MarkStyle.Free;
  { Currently only plot once, so free up space }
  FreeAndNil(Stream);
end;

end.
