unit SymT;
{ Symmetric Top }

{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Classes, SysUtils, MathSup, MNode, Mixture, QuantumNumber, QuantumMath, pgDOM;

const
  WTI = LMI+2; { Index of first statistical weight }

type

  TPointGroup = (C3v=0, C4v, C5v, C6v, C7v, C8v, D3h, D4h, D5h, D6h, D7h, D8h, D2d, D3d, D4d, D5d, D6d, D7d, D8d,
                 C3h, C4h, C5h, C6h, S4, S6, S8, S10, S12);

  TCharacterTableEntry = record
    Name: AnsiString;
    g, rho, eta: Integer;
  end;

  TCharacterTable = record
    GroupName: string;
    Size: Integer;
    n: Integer;         { Highest order rotation axis }
    gKmult: Integer;    { multiplier of k in expression for g}
    rhoJmult: Integer;  { multiplier of J in expression for rho}
    rhoKmult: Integer;  { multiplier of k in expression for rho}
    etaKmult: Integer;  { multiplier of k in expression for eta}
    DipoleSym: Integer; { Symmetry of space fixed dipole moment component }
    HasSeparableDegeneracy: Boolean; { See C3h discussion below }
    Names: array of PChar;
    IReps: array of TCharacterTableEntry;
    En: array of Integer; { 1 for E/E1, 2 for E2, ... }
    Group: array of Integer; { Statistical weight group number for each irep }
    GroupNames: array of string;
  end;

  TSymmetricMolecule = class;

  { TSymmetricTop }

  TSymmetricTop = class (TState)
  private
    FMolecule: TSymmetricMolecule;
    FProlate: Boolean;
  public
    property Prolate: Boolean read FProlate;
    constructor Create(AOwner: TComponent); override;
    function FixupProperties:Exception; override;
    function AddOneJ(Eigenvalues: TEigenvalues; J: TQuantumNumber; Sym: Integer; HyperfineState:THyperfineBasis; Seffective: TQuantumNumber; out StatWtGroup: Integer): Integer; override;
    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; override;
    function HRot(J, Kp, lp, K, l: TQuantumNumber): Double;
    function HWidth(J, Kp, lp, K, l: TQuantumNumber): Double;
    function HDip(TransitionMoment: TSingleTransitionMoment; Jp, Kp, lp, J, K, l: TQuantumNumber; Drho: Integer): Double;
    function HyperfineMatrixElement(ABra, AKet: TBasisState; Signp, Sign: Integer; MNode: TMNode; RankWanted: TQuantumNumber): Complex; override;
    function RotationalMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Complex; override;
    function RotationalPerturbationMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMNode; Flag: Integer):Complex; override;
    function RotationalTransitionMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMNode; F: TQuantumNumber):Complex; override;
    function AllowedSphericalComponent(TransitionMoment: TSphericalTransitionMoment; State: TState; out Component: TQuantumNumber):Exception; override;

    property Molecule: TSymmetricMolecule read FMolecule;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    function ReadBranchFormat(Text: string; var Obs: TObservation; UpperState: TState; Reader: TObservationsReaderBase):Boolean; override;
    function SettingToStr(AIndex: Integer): string; override;
    procedure SetSettingFromStr(AIndex: Integer; const Str: string); override;
  published
    property Symmetry: Integer index LSI read GetSetting write SetSetting;
    property Kmin:     TQuantumNumber  index LSI+1 read GetSetting write SetSetting;
    property Kmax:     TQuantumNumber  index LSI+2 read GetSetting write SetSetting;

    property B: Double index LSP   read GetParameter write SetParameter;
    property C: Double index LSP+1 read GetParameter write SetParameter;
    { Quartic Centrifugal Distortion }
    property DJ:      Double index LSP+2 read GetParameter write SetParameter;
    property DJK:     Double index LSP+3 read GetParameter write SetParameter;
    property DK:      Double index LSP+4 read GetParameter write SetParameter;
    { Coriolis }
    property zeta:    Double index LSP+5 read GetParameter write SetParameter;
    property etaJ:    Double index LSP+6 read GetParameter write SetParameter;
    property etaK:    Double index LSP+7 read GetParameter write SetParameter;
    { l doubling }
    property qplus:   Double index LSP+ 8 read GetParameter write SetParameter;
    property qminus:  Double index LSP+ 9 read GetParameter write SetParameter;
    property r:       Double index LSP+10 read GetParameter write SetParameter;
    property DqJ:     Double index LSP+11 read GetParameter write SetParameter;
    property DqK:     Double index LSP+12 read GetParameter write SetParameter;
    property DrJ:     Double index LSP+13 read GetParameter write SetParameter;
    property DrK:     Double index LSP+14 read GetParameter write SetParameter;
    { Sextic Centrifugal Distortion }
    property HJ:    Double index LSP+15 read GetParameter write SetParameter;
    property HJK:   Double index LSP+16 read GetParameter write SetParameter;
    property HKJ:   Double index LSP+17 read GetParameter write SetParameter;
    property HK:    Double index LSP+18 read GetParameter write SetParameter;
    { Octic Centrifugal Distortion }
    property LJ:    Double index LSP+19 read GetParameter write SetParameter;
    property LJJK:  Double index LSP+20 read GetParameter write SetParameter;
    property LJK:   Double index LSP+21 read GetParameter write SetParameter;
    property LKKJ:  Double index LSP+22 read GetParameter write SetParameter;
    property LK:    Double index LSP+23 read GetParameter write SetParameter;
    { Linewidth parameters - initial stab }
    property wK:    Double index LSP+24 read GetParameter write SetParameter;
    property wJ:    Double index LSP+25 read GetParameter write SetParameter;
  end;

  TSymmetricTopPerturbation = class (TPerturbation)
  public
    constructor Create(AOwner: TComponent); override;
    function MatrixElement(J, Kp, lp, K, l: TQuantumNumber): Double;
    function PairName: string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    function SettingToStr(AIndex: Integer): string; override;
    procedure SetSettingFromStr(AIndex: Integer; const s: string); override;
  published
    property JPower:Integer  index LPI   read GetSetting write SetSetting;
    property zPower:Integer  index LPI+1 read GetSetting write SetSetting;
    property pPower:Integer  index LPI+2 read GetSetting write SetSetting;
    property lPower:Integer  index LPI+3 read GetSetting write SetSetting;
    property lChange:Integer index LPI+4 read GetSetting write SetSetting;
    property KSelect:TQuantumNumber index LPI+5 read GetSetting write SetSetting;
  end;

  TsymmetricTopNucleus = class(TNucleus)
  private
    CHI: array[-2..2] of Double; { Spherical tensor equivalent of quadrupole terms }
    QuadI: Double; { = 1/thrj(I, 4, I, -I, 0, I) }
    TC00, TC20: Double;
  public
    function FixupProperties: Exception; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function HasIdentical: Boolean; override;
  published
    { Quadrupole }
    property eQq:                Double index  1 read GetParameter write SetParameter;
    { Nuclear Spin - Rotation }
    property Czz:                Double index  2 read GetParameter write SetParameter;
    property Cperp:{=(Cxx+Cyy)/2}Double index  3 read GetParameter write SetParameter;
  end;

  TSymmetricTopBasisState = class(TBasisState)
  public
    K, l: TQuantumNumber;
    function klLabel(out ExtraTag: string): string;
    function TrackIndex: Integer; override;
    function BasisLabel: string; override; { Basis State Label }
    function BasisLabelKey: string; override; { Explanation of Basis State Label }
    function StateLabel(AQuantumNumberFormat: TQuantumNumberFormat): string; override; { State name, assuming energy ordering preserved }
    function StateLabelKey: string; override; { Explanation of State Label }
    function Qno(QnoType: TQnoType): TQuantumNumber; override; { extra quantum numbers - |K|, l }
    function TxName(Upper, Lower: TBasisState): string; override;    { Make transition label such as qRK(3) }
    function TxNameKey(Upper, Lower: TBasisState): string; override; { String explaining transition label }
  end;

  TSymmetricManifold = class(TManifold)
  public
    function nSym: Integer; override; { Number of different symmetries }
  end;

  { TSymmetricMolecule }

  TSymmetricMolecule = class(TMolecule)
  private
    FCharacterTable: TCharacterTable;
    function GetPointGroup(const Index: Integer): TPointGroup;
    procedure SetPointGroup(const Index: Integer; const AValue: TPointGroup);
  protected
    function GetSettingNames(AIndex: Integer): string; override;
  public
    { Symmetric Top Specific }
    property CharacterTable: TCharacterTable read FCharacterTable;
    { General }
    constructor Create(AOwner: TComponent); override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function FixupProperties: Exception; override;
    function IsValidChildClass(MNodeClass: TMNodeClass):Boolean; override;
    function Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean; override;
    function DipoleAllowed(Sym1, Sym2: Integer): Boolean;
    function JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer; override;
    function LastSym: Integer; override;
    function SymName(Sym: Integer): string; override;
    function QnoInfo(QnoType: TQnoType): TQnoInfo; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function VibronicallyAllowed(TransitionMoment: TSphericalTransitionMoment):TVibronicallyAllowed; overload;
    function VibronicallyAllowed(TransitionMoment: TSphericalTransitionMoment; Component: TQuantumNumber; BraSym, KetSym: Integer):TVibronicallyAllowed; overload;
  published
    property PointGroup:     TPointGroup index LMI read GetPointGroup write SetPointGroup;
    property AlternateBasis: Boolean  index LMI+1  read GetFlag write SetFlag;
    property Wt0:            Integer  index WTI+0  read GetSetting write SetSetting;
    property Wt1:            Integer  index WTI+1  read GetSetting write SetSetting;
    property Wt2:            Integer  index WTI+2  read GetSetting write SetSetting;
    property Wt3:            Integer  index WTI+3  read GetSetting write SetSetting;
    property Wt4:            Integer  index WTI+4  read GetSetting write SetSetting;
    property Wt5:            Integer  index WTI+5  read GetSetting write SetSetting;
    property Wt6:            Integer  index WTI+6  read GetSetting write SetSetting;
    property Wt7:            Integer  index WTI+7  read GetSetting write SetSetting;
    property Wt8:            Integer  index WTI+8  read GetSetting write SetSetting;
    property Wt9:            Integer  index WTI+9  read GetSetting write SetSetting;
    property Wt10:           Integer  index WTI+10 read GetSetting write SetSetting;
    property Wt11:           Integer  index WTI+11 read GetSetting write SetSetting;
    property Wt12:           Integer  index WTI+12 read GetSetting write SetSetting;
    property Wt13:           Integer  index WTI+13 read GetSetting write SetSetting;
  end;

const
  MaxnStatWeight = 14; { Maximum number of statistical weights }

implementation

uses Math, nj, Worker, StaticField;

{ This implements Table 12-14 and the surrounding logic of
  P.R. Bunker and Per Jensen, Molecular Symmetry and Spectroscopy, 2nd Edition.
  See also http://symmetry.jacobs-university.de/
}

const

  CharacterTableNames: array [TPointGroup] of PChar = ('C3v', 'C4v', 'C5v', 'C6v', 'C7v', 'C8v',
                                                       'D3h', 'D4h', 'D5h', 'D6h', 'D7h', 'D8h',
                                                       'D2d', 'D3d', 'D4d', 'D5d', 'D6d', 'D7d', 'D8d',
                                                       'C3h', 'C4h', 'C5h', 'C6h',
                                                       'S4',  'S6',  'S8',  'S10', 'S12');

  C3vCharacterTable: array [0..2] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'E';  g:1; rho:0; eta:0));

  C5vCharacterTable: array [0..3] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0));

  C7vCharacterTable: array [0..4] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0));

  C4vCharacterTable: array [0..4] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:2; rho:0; eta:0), { Was 3 }
   (Name:'B2'; g:2; rho:1; eta:0), { Was 3 }
   (Name:'E';  g:1; rho:0; eta:0));

  C6vCharacterTable: array [0..5] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:3; rho:0; eta:0),
   (Name:'B2'; g:3; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0));

  C8vCharacterTable: array [0..6] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:4; rho:0; eta:0),
   (Name:'B2'; g:4; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0));

  D3hCharacterTable: array [0..5] of TCharacterTableEntry =
  ((Name:'A1'''; g:0; rho:0; eta:0),
   (Name:'A2'''; g:0; rho:1; eta:0),
   (Name:'E''';  g:1; rho:0; eta:0),
   (Name:'A1"';  g:0; rho:0; eta:1),
   (Name:'A2"';  g:0; rho:1; eta:1),
   (Name:'E"';   g:1; rho:0; eta:1));

  D5hCharacterTable: array [0..7] of TCharacterTableEntry =
  ((Name:'A1'''; g:0; rho:0; eta:0),
   (Name:'A2'''; g:0; rho:1; eta:0),
   (Name:'E1'''; g:1; rho:0; eta:0),
   (Name:'E2'''; g:2; rho:0; eta:0),
   (Name:'A1"';  g:0; rho:0; eta:1),
   (Name:'A2"';  g:0; rho:1; eta:1),
   (Name:'E1"';  g:1; rho:0; eta:1),
   (Name:'E2"';  g:2; rho:0; eta:1));

  D7hCharacterTable: array [0..9] of TCharacterTableEntry =
  ((Name:'A1'''; g:0; rho:0; eta:0),
   (Name:'A2'''; g:0; rho:1; eta:0),
   (Name:'E1'''; g:1; rho:0; eta:0),
   (Name:'E2'''; g:2; rho:0; eta:0),
   (Name:'E3'''; g:3; rho:0; eta:0),
   (Name:'A1"';  g:0; rho:0; eta:1),
   (Name:'A2"';  g:0; rho:1; eta:1),
   (Name:'E1"';  g:1; rho:0; eta:1),
   (Name:'E2"';  g:2; rho:0; eta:1),
   (Name:'E3"';  g:3; rho:0; eta:1));

  D4hCharacterTable: array [0..9] of TCharacterTableEntry =
  ((Name:'A1g'; g:0; rho:0; eta:0),
   (Name:'A2g'; g:0; rho:1; eta:0),
   (Name:'B1g'; g:2; rho:0; eta:0),
   (Name:'B2g'; g:2; rho:1; eta:0),
   (Name:'Eg';  g:1; rho:0; eta:0),
   (Name:'A1u'; g:0; rho:0; eta:1),
   (Name:'A2u'; g:0; rho:1; eta:1),
   (Name:'B1u'; g:2; rho:0; eta:1),
   (Name:'B2u'; g:2; rho:1; eta:1),
   (Name:'Eu';  g:1; rho:0; eta:1));

  D6hCharacterTable: array [0..11] of TCharacterTableEntry =
  ((Name:'A1g'; g:0; rho:0; eta:0),
   (Name:'A2g'; g:0; rho:1; eta:0),
   (Name:'B1g'; g:3; rho:0; eta:0),
   (Name:'B2g'; g:3; rho:1; eta:0),
   (Name:'E1g'; g:1; rho:0; eta:0),
   (Name:'E2g'; g:2; rho:0; eta:0),
   (Name:'A1u'; g:0; rho:0; eta:1),
   (Name:'A2u'; g:0; rho:1; eta:1),
   (Name:'B1u'; g:3; rho:0; eta:1),
   (Name:'B2u'; g:3; rho:1; eta:1),
   (Name:'E1u'; g:1; rho:0; eta:1),
   (Name:'E2u'; g:2; rho:0; eta:1));

  D8hCharacterTable: array [0..13] of TCharacterTableEntry =
  ((Name:'A1g'; g:0; rho:0; eta:0),
   (Name:'A2g'; g:0; rho:1; eta:0),
   (Name:'B1g'; g:4; rho:0; eta:0),
   (Name:'B2g'; g:4; rho:1; eta:0),
   (Name:'E1g'; g:1; rho:0; eta:0),
   (Name:'E2g'; g:2; rho:0; eta:0),
   (Name:'E3g'; g:3; rho:0; eta:0),
   (Name:'A1u'; g:0; rho:0; eta:1),
   (Name:'A2u'; g:0; rho:1; eta:1),
   (Name:'B1u'; g:4; rho:0; eta:1),
   (Name:'B2u'; g:4; rho:1; eta:1),
   (Name:'E1u'; g:1; rho:0; eta:1),
   (Name:'E2u'; g:2; rho:0; eta:1),
   (Name:'E3u'; g:3; rho:0; eta:1));

  D3dCharacterTable: array [0..5] of TCharacterTableEntry =
  ((Name:'A1g'; g:0; rho:0; eta:0),
   (Name:'A2g'; g:0; rho:1; eta:0),
   (Name:'Eg';  g:1; rho:0; eta:0),
   (Name:'A1u'; g:0; rho:0; eta:1),
   (Name:'A2u'; g:0; rho:1; eta:1),
   (Name:'Eu';  g:1; rho:0; eta:1));

  D5dCharacterTable: array [0..7] of TCharacterTableEntry =
  ((Name:'A1g'; g:0; rho:0; eta:0),
   (Name:'A2g'; g:0; rho:1; eta:0),
   (Name:'E1g'; g:1; rho:0; eta:0),
   (Name:'E2g'; g:2; rho:0; eta:0),
   (Name:'A1u'; g:0; rho:0; eta:1),
   (Name:'A2u'; g:0; rho:1; eta:1),
   (Name:'E1u'; g:1; rho:0; eta:1),
   (Name:'E2u'; g:2; rho:0; eta:1));

  D7dCharacterTable: array [0..9] of TCharacterTableEntry =
  ((Name:'A1g'; g:0; rho:0; eta:0),
   (Name:'A2g'; g:0; rho:1; eta:0),
   (Name:'E1g'; g:1; rho:0; eta:0),
   (Name:'E2g'; g:2; rho:0; eta:0),
   (Name:'E3g'; g:3; rho:0; eta:0),
   (Name:'A1u'; g:0; rho:0; eta:1),
   (Name:'A2u'; g:0; rho:1; eta:1),
   (Name:'E1u'; g:1; rho:0; eta:1),
   (Name:'E2u'; g:2; rho:0; eta:1),
   (Name:'E3u'; g:3; rho:0; eta:1));

  D2dCharacterTable: array [0..4] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:2; rho:0; eta:0),
   (Name:'B2'; g:2; rho:1; eta:0),
   (Name:'E';  g:1; rho:0; eta:0));

  D4dCharacterTable: array [0..6] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:4; rho:0; eta:0),
   (Name:'B2'; g:4; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0));

  D6dCharacterTable: array [0..8] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:6; rho:0; eta:0),
   (Name:'B2'; g:6; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0),
   (Name:'E4'; g:4; rho:0; eta:0),
   (Name:'E5'; g:5; rho:0; eta:0));

  D8dCharacterTable: array [0..10] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:8; rho:0; eta:0),
   (Name:'B2'; g:8; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0),
   (Name:'E4'; g:4; rho:0; eta:0),
   (Name:'E5'; g:5; rho:0; eta:0),
   (Name:'E6'; g:6; rho:0; eta:0),
   (Name:'E7'; g:7; rho:0; eta:0));

 { For point groups like C3h E functions are actually pairs
   of separate functions that are complex conjugates of each other;
   The E functions must therefore be counted twice, as they are
   independent functions and the nuclear spin function can be chosen
   independently for each function.
   Also, there is no R- operation, which converts K -> -K
   This has the consequence that for K a multiple of n, the functions
   have the same symmetry, and are not necessarily degenerate.
   (The E functions still are, but this arises from time reversal symmetry,
    not standard point group theory).
 }
 C3hCharacterTable: array [0..3] of TCharacterTableEntry =
  ((Name:'A'''; g:0; rho:0; eta:0),  { Don't use rho at all; no R'+ }
   (Name:'E'''; g:1; rho:0; eta:0),
   (Name:'A"';  g:0; rho:0; eta:1),
   (Name:'E"';  g:1; rho:0; eta:1));

 C4hCharacterTable: array [0..5] of TCharacterTableEntry =
  ((Name:'Ag'; g:0; rho:0; eta:0),  { Don't use rho at all; R'+ = i }
   (Name:'Bg'; g:2; rho:0; eta:0),
   (Name:'Eg'; g:1; rho:0; eta:0),
   (Name:'Au'; g:0; rho:0; eta:1),
   (Name:'Bu'; g:2; rho:0; eta:1),
   (Name:'Eu'; g:1; rho:0; eta:1));

 C5hCharacterTable: array [0..5] of TCharacterTableEntry =
  ((Name:'A''';  g:0; rho:0; eta:0), { Don't use rho at all; R'+ = sigma }
   (Name:'E1'''; g:1; rho:0; eta:0),
   (Name:'E2'''; g:2; rho:0; eta:0),
   (Name:'A"';   g:0; rho:0; eta:1),
   (Name:'E1"';  g:1; rho:0; eta:1),
   (Name:'E2"';  g:2; rho:0; eta:1));

 C6hCharacterTable: array [0..7] of TCharacterTableEntry =
  ((Name:'Ag';  g:0; rho:0; eta:0),  { Don't use rho at all; R'+ = i }
   (Name:'Bg';  g:3; rho:0; eta:0),
   (Name:'E1g'; g:1; rho:0; eta:0),
   (Name:'E2g'; g:2; rho:0; eta:0),
   (Name:'Au';  g:0; rho:0; eta:1),
   (Name:'Bu';  g:3; rho:0; eta:1),
   (Name:'E1u'; g:1; rho:0; eta:1),
   (Name:'E2u'; g:2; rho:0; eta:1));

 S4CharacterTable: array [0..2] of TCharacterTableEntry =
  ((Name:'A'; g:0; rho:0; eta:0),  { Don't use rho at all; no R'+ }
   (Name:'B'; g:2; rho:0; eta:0),
   (Name:'E'; g:1; rho:0; eta:0));

 S6CharacterTable: array [0..3] of TCharacterTableEntry =
  ((Name:'Ag'; g:0; rho:0; eta:0),  { Don't use rho at all; R'+ = i }
   (Name:'Eg'; g:1; rho:0; eta:0),
   (Name:'Au'; g:0; rho:0; eta:1),
   (Name:'Eu'; g:1; rho:0; eta:1));

 S8CharacterTable: array [0..4] of TCharacterTableEntry =
  ((Name:'A';  g:0; rho:0; eta:0),  { Don't use rho at all; no R'+ }
   (Name:'B';  g:4; rho:0; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0));

 S10CharacterTable: array [0..5] of TCharacterTableEntry =
  ((Name:'Ag';  g:0; rho:0; eta:0),  { Don't use rho at all; R'+ = i }
   (Name:'E1g'; g:1; rho:0; eta:0),
   (Name:'E2g'; g:2; rho:0; eta:0),
   (Name:'Au';  g:0; rho:0; eta:1),
   (Name:'E1u'; g:1; rho:0; eta:1),
   (Name:'E2u'; g:2; rho:0; eta:1));

 S12CharacterTable: array [0..6] of TCharacterTableEntry =
  ((Name:'A';  g:0; rho:0; eta:0),  { Don't use rho at all; no R'+ }
   (Name:'B';  g:6; rho:0; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0),
   (Name:'E4'; g:4; rho:0; eta:0),
   (Name:'E5'; g:5; rho:0; eta:0));

var
  CharacterTables: array [TPointGroup] of TCharacterTable;

function AddG(G,n:Integer):Integer;
begin { Reduce G quantum number }
  if g < 0 then
    g := g - g*n; { Want result of mod on next line to be positive }
  g := g mod n;
  if g > n div 2 then
    Dec(g,n);
  Result := g;
end;

{ TSymmetricMolecule }

function TSymmetricMolecule.DipoleAllowed(Sym1, Sym2: Integer): Boolean;
var
  Gdipole, G1, G2: TCharacterTableEntry;
  g: Integer;
begin
  Result := False;
  { Does Sym1 * DipoleSym = Sym2? }
  { En Match }
  { TODO: Speed up? Gdipole.g is only non-zero for D2d, D4d, ... }
  Gdipole := CharacterTable.IReps[CharacterTable.DipoleSym];
  G1 := CharacterTable.IReps[Sym1];
  G2 := CharacterTable.IReps[Sym2];
  g := AddG(Gdipole.g + G1.g, CharacterTable.n);
  if Abs(g) <> G2.g then Exit;
  { g/u '/'' Match }
  if Odd(Gdipole.eta + G1.eta + G2.eta) then Exit;
  if (CharacterTable.En[sym2] = 0) { Not degenerate }
     and Odd(Gdipole.rho + G1.rho + G2.rho) then Exit;
  Result := True;
end;

function TSymmetricMolecule.FixupProperties: Exception;
var
  I: Integer;
begin
  Result := Inherited FixupProperties;
  if not Assigned(Result) and JAdjustSym and (nNuclei > 0) then
    for i := 0 to nNuclei - 1 do
      if Nuclei[i].Spin > 0 then
        Result := EReportException.Create('J adjusted symmetry for '+Name+' not implemented for hyperfine structure');
end;

function TSymmetricMolecule.Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean;
begin
  Result := False;
  case TransitionMoment.VibronicallyAllowed of
    vaMaybe: if (TransitionMoment is TSphericalTransitionMoment)
                and (VibronicallyAllowed(TSphericalTransitionMoment(TransitionMoment)) = vaNo) then begin
               TransitionMoment.VibronicallyAllowed := vaNo;
               Exit;
             end else
               TransitionMoment.VibronicallyAllowed := vaYes;
    vaNo: Exit;
  end;

  if not TransitionMoment.Magnetic and  (Minus(TransitionMoment.Rank) < 0) then
    Result := DipoleAllowed(Sym1, Sym2)
  else
    Result := Sym1 = Sym2;
end;

constructor TSymmetricMolecule.Create(AOwner: TComponent);
var
  i: Integer;
begin
  SavedProperties := SavedProperties + ',PointGroup=C3v';
  for i := 0 to MaxnStatWeight-1 do
    SavedProperties := SavedProperties + Format(',wt%d=1', [i]);
  inherited Create(AOwner);
end;

function TSymmetricMolecule.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  s: string;
  p: Integer;
begin
  { AlternateBasis was represented in point group name before 6.0.199, read it here }
  s := ReadValue(Node, 'PointGroup', 'C3v', '', False);
  p := Pos('alt', LowerCase(s));
  if p > 0 then begin
    s := Copy(s,1,Length(s)-3);
    WriteValue(Node,'PointGroup',s,'C3v');
  end;
  Result:=inherited ReadProperties(Node, CheckUnread);
  if p > 0 then
    AlternateBasis := True;
end;

function TSymmetricMolecule.GetSettingNames(AIndex: Integer): string;
begin
  Result := inherited GetSettingNames(AIndex);
  if (AIndex >= WTI) and (AIndex < WTI+MaxnStatWeight) then
    if AIndex <= WTI+High(CharacterTable.GroupNames) then
      Result := CharacterTable.GroupNames[AIndex-WTI]
    else
      Result := '(' + Result + ')';
end;

function TSymmetricMolecule.GetPointGroup(const Index: Integer): TPointGroup;
begin
  Result := TPointGroup(GetSetting(Index));
end;

procedure TSymmetricMolecule.SetPointGroup(const Index: Integer; const AValue: TPointGroup);
begin
  FCharacterTable := CharacterTables[AValue];
  FnSpinSpecies := Length(CharacterTable.GroupNames);
  if PointGroup = AValue then Exit;
  SetSetting(Index, Ord(AValue));
  PerformMNodeNotification(Self, mnPropNameChange, Index);
end;

function TSymmetricMolecule.JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer;
begin
  Result := Sym;
  if (Sym >= 0) and JAdjustSym and (Minus(J) < 0) and (CharacterTable.IReps[Sym].Name[1] <> 'E') then
    if CharacterTable.IReps[Sym].rho = 1 then
      Dec(Result)
    else if CharacterTable.IReps[Result+1].rho = 1 then
      { No J adjustment for Separably Degenerate groups }
      Inc(Result);
end;

function TSymmetricMolecule.SymName(Sym: Integer): string;
begin
  if Sym < 0 then
    Result := '-'
  else
    Result := CharacterTable.Names[Sym];
end;

function TSymmetricMolecule.VibronicallyAllowed(TransitionMoment: TSphericalTransitionMoment): TVibronicallyAllowed;
var
  Bra, Ket: TSymmetricTop;
  Component: TQuantumNumber;
begin
  Bra := TransitionMoment.Bra as TSymmetricTop;
  Ket := TransitionMoment.Ket as TSymmetricTop;
  if (Bra.S = Ket.S) then begin
    Component := TransitionMoment.Component;
    if Component <= -9999 then
      Component := TransitionMoment.ActualComponent;
    Result := VibronicallyAllowed(TransitionMoment, Component, Bra.Symmetry, Ket.Symmetry);
  end else
    Result := vaMaybe;
end;

function TSymmetricMolecule.VibronicallyAllowed(TransitionMoment: TSphericalTransitionMoment; Component: TQuantumNumber; BraSym, KetSym: Integer):TVibronicallyAllowed;
var
  g, rho, eta, g1: Integer;
begin
  Result := vaNo;
  { Symmetry of Rovibronic Transition Moment }
  if (Minus(TransitionMoment.Rank) < 0) and not (TransitionMoment is TMagneticTransitionMoment) then begin
    g := CharacterTable.IReps[CharacterTable.DipoleSym].g;
    rho := CharacterTable.IReps[CharacterTable.DipoleSym].rho;
    eta := CharacterTable.IReps[CharacterTable.DipoleSym].eta;
  end else begin
    g := 0;
    rho := 0;
    eta := 0;
  end;
  g := AddG(CharacterTable.gKmult*Component div 2 + g, CharacterTable.n);
  eta := eta + CharacterTable.etaKmult*Component div 2;
  if Odd(eta + CharacterTable.IReps[BraSym].eta + CharacterTable.IReps[KetSym].eta) then
    Exit;
  g1 := AddG(g + CharacterTable.IReps[BraSym].g, CharacterTable.n);
  if Abs(g1) <> CharacterTable.IReps[KetSym].g then begin
    { Check alternate - identical unless ExE }
    g1 := AddG(g - CharacterTable.IReps[BraSym].g, CharacterTable.n);
    if Abs(g1) <> CharacterTable.IReps[KetSym].g then
      Exit;
  end;
  if (g1 = 0) or (2*g1 = CharacterTable.n) then begin
    if (g = 0) or (2*g = CharacterTable.n) then begin
      rho := rho + (CharacterTable.rhoJmult*TransitionMoment.Rank + CharacterTable.rhoKmult*Component) div 2;
      if Odd(rho + CharacterTable.IReps[BraSym].rho + CharacterTable.IReps[KetSym].rho) then
        Exit;
    end;
    { E x E gives both values of rho, so don't check }
  end;
  Result := vaYes;
end;

function TSymmetricMolecule.LastSym: Integer;
begin
  Result := CharacterTable.Size-1;
end;

function TSymmetricMolecule.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = LMI then
    Result := 'Point group of molecule.'
  else if AIndex = LMI+1 then
    Result := 'Alternate Basis - Separably degenerate groups only.'
  else if (AIndex >= WTI) and (AIndex < WTI+MaxnStatWeight) then
    if AIndex <= WTI+High(CharacterTable.GroupNames) then begin
      Result := CharacterTable.GroupNames[AIndex-WTI];
      Result := 'Statistical weight for ' + Copy(Result,4,Length(Result)-4) + ' rotational levels.'
    end else
      Result := 'Unused'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TSymmetricMolecule.IsValidChildClass(MNodeClass: TMNodeClass): Boolean;
begin
  Result := inherited IsValidChildClass(MNodeClass);
  if Result and MNodeClass.InheritsFrom(TTransitionMoment) then
    Result := (MNodeClass = TSphericalTransitionMoment)
               or (MNodeClass = TCartesianTransitionMoment);
end;

function TSymmetricMolecule.QnoInfo(QnoType: TQnoType): TQnoInfo;
begin
  Result := inherited QnoInfo(QnoType);
  case QnoType of
    qtA: Result.Name := '|K|';
    qtB: Result.Name := 'Kl';
  end;
end;

{ TSymmetricManifold }

function TSymmetricManifold.nSym: Integer;
begin
  Result := TSymmetricMolecule(Owner).CharacterTable.Size;
end;

{ TSymmetricTop }

function SameStatisticalWeightGroup(PointGroup: TPointGroup; sym, sym1: Integer):Boolean;
var
  CharacterTable: TCharacterTable;
  function Twog(sym: Integer):Integer;
  begin
    { Apply g for applying R+ twice }
    Result := 2*CharacterTable.IReps[sym].g mod CharacterTable.n;
    if Result > CharacterTable.n div 2 then
      Dec(Result,CharacterTable.n);
    Result := Abs(Result);
  end;
begin
  CharacterTable := CharacterTables[PointGroup];
  { For each symmetric top point group, one of the generating operations corresponds to
    a permutation/inversion, so the symmetry of the nuclear+rovibronic wavefunction is not selected
    by this operation. Symmetries that only differ in this operation are therefore in the same
    statistical weight group }
  { R+ Operation }
  if PointGroup in [D2d, D4d, D6d, D8d, S4, S8, S12] then
    { Operation is Sn; that need not match but Sn^2 must }
    Result := Twog(sym) = Twog(sym1)
  else { Operation is Cn - just match }
    Result := CharacterTable.IReps[sym].g = CharacterTable.IReps[sym1].g;
  { Does R'+ Operation only interchange equivalent nuclei? } { TODO: Always false? }
  if not (PointGroup in [D3h, D4h, D5h, D6h, D7h, D8h, D3d, D5d, D7d, C3h, C4h, C5h, C6h, S4, S6, S8, S10, S12]) then
    Result := Result and (CharacterTable.IReps[sym].eta = CharacterTable.IReps[sym1].eta);
  if not (PointGroup in [C3v, C4v, C5v, C6v, C7v, C8v]) then { R- Operation }
    Result := Result and (CharacterTable.IReps[sym].rho = CharacterTable.IReps[sym1].rho);
end;

procedure DetermineStatisticalWeightGroups(PointGroup: TPointGroup);
var
  sym, sym1, nGroups: Integer;
begin
  with CharacterTables[PointGroup] do begin
    SetLength(Group, Size);
    SetLength(GroupNames, Size); { Initial guess - contract later }
    nGroups := 0;
    FillChar(Group[0], SizeOf(Group), 0);
    for sym := 0 to Size-1 do { For each irreducible representation }
      if Group[sym] = 0 then begin { if not assigned to statistical weight group }
        GroupNames[nGroups] := 'wt(' + Names[sym];
        Group[sym] := nGroups+1;
        for sym1 := sym+1 to Size-1 do
          if SameStatisticalWeightGroup(PointGroup, sym, sym1) then begin
            Group[sym1] := nGroups+1;
            GroupNames[nGroups] := GroupNames[nGroups] + ',' + Names[sym1];
          end;
        GroupNames[nGroups] := GroupNames[nGroups] + ')';
        Inc(nGroups);
      end;
    SetLength(GroupNames,nGroups);
//    Write(GroupName, ': ', Names[DipoleSym], ' : ');
//    for i := 0 to nGroups-1 do
//      Write(' ', GroupNames[i]);
//    Writeln;
  end;
end;

function FindIrep(PointGroup: TPointGroup; g, rho, eta:Integer):Integer;
var
  i: Integer;
begin
  with CharacterTables[PointGroup] do
    for i := 0 to Size-1 do
      if (Ireps[i].g = g) and (Ireps[i].rho = rho) and (Ireps[i].eta = eta) then begin
        Result := i;
        Exit;
      end;
   raise Exception.Create('FindIrep');
end;

procedure DetermineDipoleSymmetry(PointGroup: TPointGroup);
begin
  with CharacterTables[PointGroup] do
    if PointGroup in [C3v, C4v, C5v, C6v, C7v, C8v] then
      DipoleSym := FindIrep(PointGroup, 0, 1, 0)
    else if PointGroup in [D2d, D4d, D6d, D8d, S4, S8, S12] then
      DipoleSym := FindIrep(PointGroup, n div 2, 0, 0)
    else
      DipoleSym := FindIrep(PointGroup, 0, 0, 1);
end;

procedure AddTables;
var
  p: TPointGroup;
  {.$DEFINE CHECK}
  {$IFDEF CHECK}
  procedure SetStatWtSets(PointGroup: TPointGroup; DipoleName: string; GroupsDef: array of PChar);
  var
    i, p, sym: Integer;
    s, Irep: string;
    function IrepIndex(Name: string): Integer;
    var
      j: Integer;
    begin
      with CharacterTables[PointGroup] do
        for j := 0 to Size-1 do
          if Names[j] = Name then begin
            Result := j;
            Exit;
          end;
      raise Exception.Create('SymT:3');
    end;
  begin
    with CharacterTables[PointGroup] do begin
      {$IFDEF CHECK}
        if Length(Group) <> Size then raise NotYet;
        if Length(GroupNames) <> Length(GroupsDef) then raise NotYet;
      {$ELSE}
        SetLength(Group, Size);
        SetLength(GroupNames, Length(GroupsDef));
        FillChar(Group[0], SizeOf(Group[0])*Size, 0);
      {$ENDIF}
      for i := 0 to High(GroupsDef) do begin
        s := GroupsDef[i];
        {$IFDEF CHECK}
          if GroupNames[i] <> 'wt(' + s + ')' then raise NotYet;
        {$ELSE}
          GroupNames[i] := 'wt(' + s + ')';
        {$ENDIF}
        while s <> '' do begin
          Irep := s;
          p := Pos(',',s);
          if p > 0 then begin
            Delete(Irep, p, Length(Irep));
            Delete(s,1,p);
          end else
            s := '';
          sym := IrepIndex(Irep);
          {$IFDEF CHECK}
          if Group[sym] <> i+1 then raise NotYet;
          {$ELSE}
          if Group[sym] <> 0 then
            raise Exception.Create('SymT:1')
          else
            Group[sym] := i+1;
          {$ENDIF}
        end;
      end;
      { Check for unset groups }
      for sym := 0 to Size-1 do
        if Group[sym] = 0 then
          raise Exception.Create('SymT:2');
      {$IFDEF CHECK}
      if DipoleSym <> IrepIndex(DipoleName) then raise NotYet;
      {$ELSE}
      DipoleSym := IrepIndex(DipoleName);
      {$ENDIF}
    end;
  end;
  {$ENDIF}

  procedure A(T: array of TCharacterTableEntry; Order:Integer; rhomult: Integer; gMult:Integer = -1; etamult: Integer = 0;
              ArhoJmult: Integer = 1; AHasSeparableDegeneracy: Boolean = False);
  var
    i: Integer;
  begin
    with CharacterTables[p] do begin
      GroupName := CharacterTableNames[p];
      Size := Length(T);
      n := Order;
      gKmult := gMult;
      rhoJmult := ArhoJmult;
      rhoKmult := rhomult;
      etaKmult := etamult;
      HasSeparableDegeneracy := AHasSeparableDegeneracy;
      Setlength(IReps, Size);
      Setlength(Names, Size+1);
      SetLength(En,Size);
      Names[Size] := nil;
      for i := 0 to Size-1 do begin
        Names[i] := PChar(T[i].Name);
        IReps[i] := T[i];
        if Names[i][0] = 'E' then
          if Names[i][1] in ['1'..'9'] then
            En[i] := StrToInt(Names[i][1])
          else
            En[i] := 1 { E = E1 }
        else
          En[i] := 0; { A/B }
      end;
      DetermineDipoleSymmetry(p);
      DetermineStatisticalWeightGroups(p);
      if p < High(TPointGroup) then
        Inc(p);
    end;
  end;
begin
  p := Low(TPointGroup);
  { Must be in ordinal order }
  A(C3vCharacterTable,3,1); A(C4vCharacterTable,4,1); A(C5vCharacterTable,5,1); A(C6vCharacterTable,6,1); A(C7vCharacterTable,7,1); A(C8vCharacterTable,8,1);
  A(D3hCharacterTable,3,0,-1,1); A(D4hCharacterTable,4,0); A(D5hCharacterTable,5,0,-1,1); A(D6hCharacterTable,6,0); A(D7hCharacterTable,7,0,-1,1); A(D8hCharacterTable,8,0);
  A(D2dCharacterTable, 4,0,-3);  A(D3dCharacterTable,3,0); A(D4dCharacterTable, 8,0,-5);  A(D5dCharacterTable,5,0);
  A(D6dCharacterTable,12,0,-7);  A(D7dCharacterTable,7,0); A(D8dCharacterTable,16,0,-9);
  A(C3hCharacterTable,3,0,-1,1,0,True);
  A(C4hCharacterTable,4,0,-1,0,0,True);
  A(C5hCharacterTable,4,0,-1,1,0,True);
  A(C6hCharacterTable,6,0,-1,0,0,True);
  A(S4CharacterTable,4,0,-1,0,0,True);
  A(S6CharacterTable,3,0,-1,0,0,True);
  A(S8CharacterTable,8,0,-1,0,0,True);
  A(S10CharacterTable,5,0,-1,0,0,True);
  A(S12CharacterTable,12,0,-1,0,0,True);
  {$IFDEF CHECK}
  SetStatWtSets(C3v, 'A2',  ['A1,A2', 'E']);
  SetStatWtSets(D3h, 'A1"', ['A1'',A1"', 'A2'',A2"', 'E'',E"']);
  SetStatWtSets(D6h, 'A1u', ['A1g,A1u', 'A2g,A2u', 'B1g,B1u', 'B2g,B2u', 'E1g,E1u', 'E2g,E2u']);
  {$ENDIF}
end;

function TSymmetricTop.AddOneJ(Eigenvalues: TEigenvalues; J: TQuantumNumber; Sym: Integer; HyperfineState: THyperfineBasis; Seffective: TQuantumNumber; out StatWtGroup: Integer): Integer;
var
  EVsym, EVRSym: TCharacterTableEntry;
  l, K, Kinc: TQuantumNumber;
  CharacterTable: TCharacterTable;
  procedure Consider(K, l: Integer; ConsiderBoth: Boolean);
  var
    BasisState: TSymmetricTopBasisState;
    g, rho, eta, Phase: Integer;
  begin
    if (Kmin >= 0) and (Abs(K) < Kmin) then
      Exit;
    if (Kmax >= 0) and (Abs(K) > Kmax) then
      Exit;
    { Does EVsym * Sym(K) = EVRSym? }
    { En Match }
    if l < 0 then  { TODO: Check - l < 0 should only be needed when HasSeparableDegeneracy }
      g := AddG(CharacterTable.gKmult*K div 2 - EVsym.g, CharacterTable.n)
    else
      g := AddG(CharacterTable.gKmult*K div 2 + EVsym.g, CharacterTable.n);
    if Abs(g) <> EVRsym.g then Exit;
    { g/u '/'' Match }
    eta := K*CharacterTable.etaKmult div 2 + EVsym.eta;
    if Odd(eta+EVRSym.eta) then Exit;
    if CharacterTable.HasSeparableDegeneracy then begin
      { Point group operations do not change sign of K, so need to explicitly
        include, say K=1 and K=-1 for non E functions }
      Phase := 0;
      if ((K <> 0) or (l <> 0)) and ((g=0) or (2*g = CharacterTable.n)) then begin
        if Molecule.AlternateBasis then begin
          { Take |K,l> and |-K,-l> as basis functions }
          if ConsiderBoth then
            Consider(-K,-l,False); { Add other K function }
        end else
          { To give better match with C3v style calculations, take
            |K,l> +/- |-K,-l> as basis functions for non E functions;
            this makes K=1 labels in particular more meaningful }
          if ConsiderBoth then begin
            Phase := -1;
            Consider(K,l,False);
          end else
            Phase := 1;
      end;
    end else if (g = 0) or (2*g = CharacterTable.n) then begin
      { A1 or A2 / B1 or B2 }
      rho := (CharacterTable.rhoJmult*J+K*CharacterTable.rhoKmult) div 2 + EVsym.rho;
      if Odd(rho+EVRSym.rho) then begin
        Phase := -1;
        if (l=0) and (K=0) then
          Exit; { |J00>-|J00> }
      end else begin
        Phase := 1;
        if (l=0) and (K=0) then
          Phase := 0; { |J00>+|J00> }
      end;
    end else begin
      { E - two levels, but only work out the energy of one of them }
      Phase := 0;
      { In general, choose the basis state with G > 0, preserving sign of kl }
      { There should be no matrix elements off diagonal in G }
      { This also fixes transition matrix elements (apart from D2nd)
        as these are also diagonal in G }

      { Special case for D2d, D4d, D6d, D8d as Transition moment changes G by n }
      { Flipping the sign of G for B states or the unique E state means that dipole transitions
        are calculated with the correct combination of K values }
//      if (EVsym.g > 0) and ((CharacterTable.En[Symmetry] = 0) { or (CharacterTable.En[Symmetry] = CharacterTable.n div 4) })
//         and (CharacterTable.IReps[CharacterTable.DipoleSym].g <> 0) then
//        g := -g;
      if g < 0 then begin
        K := -K;
        l := -l;
      end;
    end;
    BasisState := TSymmetricTopBasisState.Create(Eigenvalues,J,Sym);
    BasisState.State := Self;
    BasisState.J := J;
    BasisState.N := J;
    BasisState.K := K;
    BasisState.l := l;
    BasisState.Phase := Phase;
    BasisState.StateIndex := Eigenvalues.LastStateIndex;
    BasisState.HyperfineState := HyperfineState;
    BasisState.S := Seffective;
  end;
begin
  { K=2 missing for C4v }
  CharacterTable := Molecule.CharacterTable;
  StatWtGroup := CharacterTable.Group[Sym];
  Result := Molecule.GetSetting(WTI+StatWtGroup-1);
  if Result = 0 then
    Exit;
  EVsym := CharacterTable.IReps[Symmetry];
  if EVsym.Name[1] = 'E' then
    l := 2
  else
    l := 0;
  EVRsym := CharacterTable.IReps[Sym];
  if Prolate then begin
    K := 0;
    Kinc := 2
  end else begin
    Kinc := -2;
    K := J;
  end;
  while (K <= J) and (K >= 0) do begin
    Consider(K,l,True);
    if (l<>0) and (K>0) then
      Consider(-K,l,True);
    Inc(K,Kinc);
  end;
  Inc(Eigenvalues.LastStateIndex);
end;

function TSymmetricTop.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  Result := inherited GetSettingsNames(AIndex,LowestValue);
  if AIndex = LSI then
    if Molecule = nil then { Should never happen }
      Result := @CharacterTables[D3h].Names[0]
    else
      Result := @CharacterTables[Molecule.PointGroup].Names[0];
end;

function TSymmetricTop.AllowedSphericalComponent(
  TransitionMoment: TSphericalTransitionMoment; State: TState;
  out Component: TQuantumNumber): Exception;
var
  Ket: TSymmetricTop;
  q, qMatch: TQuantumNumber;
begin
  Ket := State as TSymmetricTop;
  if S <> Ket.S then
    Result := EReportException.Create(TransitionMoment.UniqueName+': Spin forbidden transition - component must be specified')
  else if TransitionMoment is TMagneticTransitionMoment then
    Result := EReportException.Create(TransitionMoment.UniqueName+': auto not implemented for magnetic dipole transitions')
  else begin
    q := 0;
    qMatch := -1;
    Result := nil;
    while q <= TransitionMoment.Rank do begin
      if Molecule.VibronicallyAllowed(TransitionMoment, q, Symmetry, Ket.Symmetry) = vaYes then
        if qMatch >= 0 then begin
          if not Assigned(Result) then
            Result := EReportException.Create(TransitionMoment.UniqueName+': More than one possible component for an allowed transition - please set component ' + QuantumNumberToStr(qMatch,qftJ));
          Result.Message := Result.Message + ' ' + QuantumNumberToStr(q,qftJ);
        end else
          qMatch := q;
      Inc(q,2);
    end;
    if not Assigned(Result) then
      if qMatch >= 0 then
        Component := qMatch
      else
        Result := EReportException.Create(TransitionMoment.UniqueName+': No allowed component');
  end;
end;

constructor TSymmetricTop.Create(AOwner: TComponent);
begin
  if Assigned(AOwner) then
    FMolecule := AOwner.Owner as TSymmetricMolecule;
  SavedProperties := SavedProperties + ',Kmin=all,Kmax=all';
  DimensionlessParameters := DimensionlessParameters + ',zeta';
  inherited;
end;

function TSymmetricTop.FixupProperties: Exception;
begin
  { Force legal symmetry, in case point group changed }
  Symmetry := Min(Symmetry, Molecule.LastSym);
  FProlate := C > B; { Precomputation - not a speed up, but work(ed) round a FPC PPC bug }
  FFullRotationalSymmetry := not Molecule.CharacterTable.HasSeparableDegeneracy;
  Result:=inherited FixupProperties;
end;

{ TSymmetricTopBasisState }

function TSymmetricTopBasisState.BasisLabel: string;
begin
  Result := Format('%s %2s', [inherited BasisLabel, State.Manifold.Mixture.QuantumNumberToStr(K)]);
  if l <> 0 then
    Result := Result + Format(' %s', [State.Manifold.Mixture.QuantumNumberToStr(l)]);
  Result := Result + HyperfineStateLabel;
end;

function TSymmetricTopBasisState.BasisLabelKey: string;
begin
  Result := inherited BasisLabelKey + '  K l' + HyperfineStateLabelKey;
end;

function TSymmetricTopBasisState.klLabel(out ExtraTag: string): string;
var
  kl: Integer;
  CharacterTable: TCharacterTable;
begin
  if l <> 0 then begin
    if K = 0 then
      Result := ' '
    else begin
      kl := K*l;
      CharacterTable := TSymmetricTop(State).Molecule.CharacterTable;
      if CharacterTable.IReps[TSymmetricTop(State).Symmetry].g
         mod 3 = 2 then
        { this is sgn(k*sin(2 pi Gve /3)) as per Hoy and Mills J Mol Spec 1973 46, 333 }
        kl := -kl;
      if kl > 0 then
        Result := '+1'
      else
        Result := '-1'
    end;
  end else
    Result := '';
  with TSymmetricTop(State) do
    if Molecule.CharacterTable.HasSeparableDegeneracy then
      if Molecule.AlternateBasis then begin
        with Molecule.CharacterTable.IReps[Sym] do
          if (K <> 0) and ((g = 0) or (2*g = Molecule.CharacterTable.n)) then
            if K > 0 then
              ExtraTag := '+'
            else
              ExtraTag := '-'
          else
            ExtraTag := ' ';
      end else begin
        if Phase > 0 then
          ExtraTag := '+'
        else if Phase < 0 then
          ExtraTag := '-'
        else
          ExtraTag := ' ';
      end
    else
      ExtraTag := '';
end;

function TSymmetricTopBasisState.Qno(QnoType: TQnoType): TQuantumNumber;
var
  ExtraTag: string;
begin
  case QnoType of
    qtA: Result := Abs(K);
    qtB: if l <> 0 then
           if K = 0 then
             Result := 2
           else if klLabel(ExtraTag) = '+1' then
             Result := 2
           else
             Result := -2
         else
           Result := 0;
  else
    Result := inherited Qno(QnoType);
  end;
end;

function TSymmetricTopBasisState.StateLabel(AQuantumNumberFormat: TQuantumNumberFormat): string;
var
  CharacterTable: TCharacterTable;
  kl, ExtraTag: string;
begin
  CharacterTable := TSymmetricTop(State).Molecule.CharacterTable;
  Result := Format('%s %2s %2s', [inherited StateLabel(AQuantumNumberFormat),
                                  QuantumNumberToStr(J,AQuantumNumberFormat),
                                  QuantumNumberToStr(Abs(K),AQuantumNumberFormat)]);
  kl := kllabel(ExtraTag);
  Result := Result + ExtraTag;
  if l<>0 then
    Result := Result + ' ' + kl;
  Result := Result + Format(' %*s', [Length(CharacterTable.Ireps[Sym].Name), TSymmetricTop(State).Molecule.SymName(Sym)])
            + HyperfineStateLabel;
end;

function TSymmetricTopBasisState.StateLabelKey: string;
begin
  Result := Format('%s J K (kl) Sym', [inherited StateLabelKey]) + HyperfineStateLabelKey;
end;

function TSymmetricTopBasisState.TrackIndex: Integer;
begin
  Result := Abs(K); { TODO: May want to include l also }
end;

function TSymmetricTopBasisState.TxName(Upper, Lower: TBasisState): string;
var
  Molecule: TSymmetricMolecule;
  U, LL: TSymmetricTopBasisState;
  Utag, LTag, ExtraTag: string;
begin
  Molecule := State.Manifold.Molecule as TSymmetricMolecule;
  U := Upper as TSymmetricTopBasisState;
  LL := Lower as TSymmetricTopBasisState;
  Result := LowerCase(DeltaJName(Abs(U.K)-Abs(LL.K)))+DeltaJName(U.J-LL.J);
  if Molecule.CharacterTable.HasSeparableDegeneracy then begin
    U.kllabel(Utag);
    LL.klLabel(LTag);
    if Utag = Ltag then
      ExtraTag := Utag
    else
      ExtraTag := UTag+LTag;
    Result := Result + Trim(ExtraTag);
  end;
  if LL.l <> 0 then
    Result := Result + Trim(Copy(LL.klLabel(LTag),1,1) { Only take sign of kl label }
                     + Molecule.QuantumNumberToStr(Abs(LL.K)))
  else
    Result := Result + Molecule.QuantumNumberToStr(Abs(LL.K));
  Result := Result + Format('(%s)',[Molecule.QuantumNumberToStr(LL.J)]);
  if (Length(Lower.HyperfineState) > 0) or (Owner as TEigenvalues).HasFields then
    Result := Result + Upper.HyperfineStateLabel(False) + ',' + Lower.HyperfineStateLabel(False);
  { TODO: Open Shell }
end;

function TSymmetricTopBasisState.TxNameKey(Upper, Lower: TBasisState): string;
begin
  Result := 'delta|K| deltaJ K"(J")' + HyperfineTxNameKey(False);
end;

function TSymmetricTop.HWidth(J, Kp, lp, K, l: TQuantumNumber): Double;
var
  JSq, KSq: Double;
begin
  if (K = Kp) and (l = lp) then begin
    JSq := J*(J+2) / 4;
    KSq := K*K / 4;
    Result := Width + wJ*JSq + (wK-wJ)*KSq
  end else
    Result := 0;
end;

function TSymmetricTop.HyperfineMatrixElement(ABra, AKet: TBasisState; Signp, Sign: Integer; MNode: TMNode; RankWanted: TQuantumNumber): Complex;
{ TODO: Abstract more from asymmetric top equivalent }
var
  Bra, Ket: TSymmetricTopBasisState;
  Nuc: TSymmetricTopNucleus;
  JP, J, Np, N, Kp, K, Sp: TQuantumNumber;
  t, Sum: Double;
  Rank: Integer;
begin
  Bra := ABra as TSymmetricTopBasisState;
  Ket := AKet as TSymmetricTopBasisState;
  Nuc := MNode as TSymmetricTopNucleus;
  JP := Bra.J;
  J := Ket.J;
  NP := Bra.N;
  N := Ket.N;
  KP := SignP*Bra.K;
  K := Sign*Ket.K;
  Sp := Bra.S;
  Result := z0;
  if Abs(K-Kp) > 4 then
    Exit;
  { Section below mainly copoed from OAsymT }
  if RankWanted = 2 then begin
    { Magnetic Dipole }
    { TODO }
    { Nuclear Spin - Rotation }
    Sum := 0;
    for Rank := 0 to 1 do begin { Actually a flag: Rank = 1 is k=2 }
      if K = Kp then begin
        if Rank = 0 then
          t := Nuc.TC00
        else
          t := Nuc.TC20; { Note T22 component missing for a symmetric top }
        Sum := Sum
               + t * Sqrt(4*Rank+One) * Minus(Np-Kp)
                 * threej(Np, -Kp, Rank*4, Kp-K, N, K)
                 *(ReducedMatrixElement(2,N)*sixj(Rank*4, 2, 2, N, Np, N)
                   +ReducedMatrixElement(2,Np)*sixj(Rank*4, 2, 2, Np, N, Np))
      end;
    end;
    if Sum <> 0 then
      Result.re := Result.re - Sum / 2 * sixj(N, J, Sp, Jp, Np, 2)
                               * Sqrt((J+1)*(JP+1)*(N+One)*(Np+One))
                         * Minus(N+Sp+J) * ReducedMatrixElement(2, Nuc.Spin);
  end else if RankWanted = 4 then begin
    { Quadrupole }
    if Abs(K-Kp) <= 4 then begin
      t := sixj(NP, JP, Sp, J, N, 4)* threej(Np, -Kp, 4, Kp-K, N, K);
      Result.re := Result.re + t * Sqrt((N+1)*(NP+One)) / 2
                               * Minus(NP+S+J+Np-Kp) * Nuc.CHI[(Kp-K) div 2];
      Result.re := Result.re * Sqrt((J+1)*(JP+One)) * Nuc.QuadI;
    end;
  end;
  if IsNan(Result.re) then
    raise Exception.Create('Help');
end;

function TSymmetricTop.HRot(J, Kp, lp, K, l: TQuantumNumber): Double;
var
  JSq, KSq: Double;
begin
  if (K = Kp) and (l = lp) then begin
    JSq := J*(J+2) / 4;
    KSq := K*K / 4;
    Result := Origin + B*JSq + (C-B)*KSq
              - DJ*Sqr(JSq) - DJK*JSq*KSq - DK*Sqr(KSq)
              + HJ*Sqr(JSq)*JSq + HJK*Sqr(JSq)*KSq + HKJ*Jsq*Sqr(KSq) + HK*Sqr(KSq)*KSq;
    Result := Result + LJ*Sqr(JSq)*Sqr(JSq) + LJJK*Sqr(JSq)*JSq*KSq + LJK*Sqr(Jsq)*Sqr(KSq) + LKKJ*Sqr(KSq)*KSq*JSq + LK*Sqr(KSq)*Sqr(KSq);
    Result := Result + (-2*C*zeta + etaj*JSq + etak*KSq) * l*K/4;
  end else if (Abs(K-Kp) = 4) and (K-Kp = l-lp) then begin
    JSq := J*(J+2) / 4;
    KSq := K*K / 4;
    { TODO: Similar to Cartwright and Mills; JMS 34 415 (1970); qplus = -q as used in NH3 Rydberg work }
    { Aliev and Papousek suggest qt+ has extra factors involving v and l }
    Result := (qplus + DqJ*JSq + DqK*(Sqr(Kp)/4+KSq))/2;
    Result := Result * Sqrt((JSq - K*(K+Kp)/8)*(JSq - Kp*(K+Kp)/8));
  end else if (Abs(K-Kp) = 4) and (K-Kp = -(l-lp)) then begin
    JSq := J*(J+2) / 4;
    Result := qminus/2;
    Result := Result * Sqrt((JSq - K*(K+Kp)/8)*(JSq - Kp*(K+Kp)/8));
  end else if (Abs(K-Kp) = 2) and (2*(K-Kp) = -(l-lp)) then begin
    JSq := J*(J+2) / 4;
    Result := r + DrJ*JSq + DrK*(Sqr(Kp)+Sqr(K))/4;
    Result := Result * Sqrt((JSq - K*Kp/4)) * (K+Kp) / 2;
  end else
    Result := 0;
end;

function TSymmetricTop.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex in [LSI+1..LSI+2];
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

function TSymmetricTop.ReadBranchFormat(Text: string; var Obs: TObservation;
  UpperState: TState; Reader: TObservationsReaderBase): Boolean;
var
  K, DeltaK, DeltaJ, kl, JU, JL: TQuantumNumber;
  p, pe, Sym: Integer;
  Lname: string;
  HyperfineL, HyperfineU: THyperfineBasis;
  HasFields: Boolean;
begin
  Result := False;
  Obs.Name := ' ' + Text;
  if not IsDeltaJName(Text[1],DeltaK) then
    Exit
  else
    Delete(Text,1,1);
  if (Length(Text) = 0) or not IsDeltaJName(Text[1],DeltaJ) then begin
    DeltaJ := DeltaK;
    DeltaK := 0;
  end else
    Delete(Text,1,1);
  if Length(Text) = 0 then begin
    Result := True; { Force Branch }
    Text := Reader.NextItem;
  end;
  kl := 0;
  if Text[1] = '+' then
    kl := 2
  else if Text[1] = '-' then
    kl := -2;
  if kl <> 0 then
    Delete(Text,1,1);
  p := Pos('(',Text);
  pe := Pos(')',Text);
  if (p > 1) and (pe > p+1)
     and not Reader.ReadQuantumNumber(K,Copy(Text,1,p-1))
     and not Reader.ReadQuantumNumber(Obs.JL,Copy(Text,p+1,pe-p-1))
      then begin
    Delete(Text,1,pe);
    Obs.JU :=  Obs.JL+DeltaJ;
    JL := Obs.JL;
    JU := Obs.JU;
    HasFields := IsNonZero(Obs.ExField);
    if HasFields or (Molecule.nNuclei > Reader.RemoveSpins) then begin
      { Some hyperfine structure to be found }
      if ReadHyperfineBranch(Text, Obs, HyperfineU, HyperfineL, Reader, HasFields) then
        Exit;
    end else if Text <> '' then
      Exit;
    Lname := Manifold.FindStateLabel(Obs.IndexL, Obs.SymL, Obs.JL,
                                     [K,kl], Obs.QuantumNumberFormat, Obs.ExField,
                                     -1, JL, -1, Self, Reader.RemoveSpins, HyperfineL);
    if (Obs.IndexL = -3) and (Reader.DefaultSymmetry >= 0) then
      { Two matches - try using default symmetry }
      Lname := Manifold.FindStateLabel(Obs.IndexL, Obs.SymL, Obs.JL,
                                       [K,kl], Obs.QuantumNumberFormat, Obs.ExField,
                                       Reader.DefaultSymmetry, JL, -1, Self, Reader.RemoveSpins, HyperfineL);
    { TODO: This section needs to look for allowed transitions to select sign of kl }
    { See e.g. tests/C3v_A1_E.pgo tests/C3v_A1_E.lin which should be readable, but currently are not }
    if Obs.IndexL = -1 then
      { No matches - try without kl }
      Lname := Manifold.FindStateLabel(Obs.IndexL, Obs.SymL, Obs.JL,
                                       [K,Low(TquantumNumber)], Obs.QuantumNumberFormat, Obs.ExField,
                                       -1, JL, -1, Self, Reader.RemoveSpins, HyperfineL);
    if Obs.SymL >= 0 then begin
      if Odd(Reader.BranchRank div 2) then begin
        for Sym := 0 to Molecule.LastSym do
          if Molecule.DipoleAllowed(Sym, Obs.SymL) then begin
            Obs.SymU := Sym;
            Break;
          end;
      end else
        Obs.SymU := Obs.SymL;
      Obs.Name := Obs.Name + ' : '
                  + UpperState.Manifold.FindStateLabel(Obs.IndexU, Obs.SymU, Obs.JU,
                                                       [K+DeltaK,Low(TquantumNumber)], Obs.QuantumNumberFormat, Obs.ExField,
                                                       Obs.SymU, JU, -1, UpperState, Reader.RemoveSpins, HyperfineU)
                  + ' - ' + Lname;

      if Obs.SymU >= 0 then begin
        Result := True;
        Obs.IndexU := 0;
        Obs.IndexL := 0;
        Exit;
      end;
    end else
      Obs.Name := Obs.Name + ' : ' + Lname
  end;
  if Result then
    Reader.ReportError('Bad Branch format');
end;

function TSymmetricTop.RotationalMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Complex;
var
  Bra, Ket: TSymmetricTopBasisState;
  H: function (J, Kp, lp, K, l: TQuantumNumber): Double of object;
  Kp, lp, K, l: TQuantumNumber;
begin
  Bra := ABra as TSymmetricTopBasisState;
  Ket := AKet as TSymmetricTopBasisState;
  case TMatElemType(Flag) of
    meHrot:  H := {$IFDEF FPC}@{$ENDIF}Hrot;
    meWidth: H := {$IFDEF FPC}@{$ENDIF}HWidth;
  else
    raise NotYet;
  end;
  if Bra.J <> Ket.J then
    raise Exception.Create('TSymmetricTop.RotationalMatrixElement');
  Kp := SignA*Bra.K;
  lp := SignA*Bra.l;
  K := SignB*Ket.K;
  l := SignB*Ket.l;
  { TODO: may require all four matrix elements or explicit test for identical phase? }
  Result.re := H(Bra.J, Kp, lp, K, l);
  Result.im := 0;
  { TODO: S <> 0 terms }
end;

function TSymmetricTop.HDip(TransitionMoment: TSingleTransitionMoment; Jp, Kp, lp, J, K, l: TQuantumNumber; Drho: Integer): Double;
var
  q: TQuantumNumber;
  Phase: Integer;
  mm: Double;
begin
  Phase := 0;
  Result := TransitionMoment.Strength;
  if TransitionMoment is TSphericalTransitionMoment then begin
    q := TSphericalTransitionMoment(TransitionMoment).ActualComponent;
  end else if TransitionMoment is TCartesianTransitionMoment then begin
    { Convert to Spherical equivalent }
    q := 2;
    case TCartesianTransitionMoment(TransitionMoment).Axis of
      aAxis: if Prolate then q := 0;
      cAxis: if not Prolate then q := 0;
    end;
    if q = 2 then
      Result := Result / Sqrt(Two);
  end else
    raise NotYet;
  if Abs(q) <> Abs(Kp-K) then begin
    Result := 0;
    Exit;
  end;
  with TransitionMoment as TBaseRotationalTransitionMoment do begin
    mm := (Jp*(Jp+2)+ J*(J+2))/8;
    Result := Result*Sqrt(1 + C*mm + D*mm*mm);
  end;
  { If the other sign of q is required, apply symmtery operation }
  { For Delta K = 0, need to choose a sign of l; in this case choose:
      l' > l or l' > 0
    as the defined one }
  if (q <> K-Kp)
     or ((q = 0) and ( (lp < l) or (lp < 0) ) ) then
    phase := Drho*2 + TransitionMoment.Rank*Molecule.CharacterTable.rhoJmult + q*Molecule.CharacterTable.rhoKmult;
  Result := Result * Sqrt((J+1)*(Jp+1))
            * Minus(Jp-Kp+phase) * threej(Jp, -Kp, TTransitionMoment(TransitionMoment).Rank, Kp-K, J, K);
  //Was:    * Minus(J-K+q+Phase) * threej(Jp, Kp, TransitionMoment.Rank, K-Kp, J, -K) * Result
end;

function TSymmetricTop.RotationalTransitionMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMNode; F: TQuantumNumber):Complex;
var
  Bra, Ket: TSymmetricTopBasisState;
  TransitionMoment: TTransitionMoment;
  Kp, lp, K, l: TQuantumNumber;
  Drho: Integer;
begin
  Bra := ABra as TSymmetricTopBasisState;
  Ket := AKet as TSymmetricTopBasisState;
  TransitionMoment := MNode as TTransitionMoment;
  { The transition moment uses tensor notation T(k,q) for evaluation, and a key
    point is that the matrix elements of T(k,q) and T(k,-q) are symmetry related.
    This is worked out here as follows. The matrix element required is of the form:
        <A'|Htran|A>
    which is split into separate vibronic (EV) and rotational(R) parts:
        <A'|Htran|A> = <EV'|HE(k,q)|EV><R'|HR(k,q)|R>
    <EV'|HE(k,q)|EV> is the strength parameter of the transition moment. HE(k,q) is the body
    fixed transition operator and HR(k,q) is essentially the Wigner D matrix transforming
    it from the space to body fixed axis systems. Now, if Hdip finds q has the wrong
    sign to give a non zero matrix element, it actually works out:
        <EV'|HE(k,q)|EV><R'|HR(k,-q)|R>
    whereas what is actually wanted is
        <-EV'|HE(k,-q)|-EV><R'|HR(k,-q)|R>
    The -EV indicates that some quanum numbers (l) in the vibronic part may also need
    changing to suit. We can generate <EV'|HE(k,q)|EV> by using the R- operator,
    which flips the sign of q. The characters with respect to this operator are rho
    in the CharaterTable.Irep array. A rotational function transforms as:
        R- |J,K> = rhoJK(J,K)|J,-K>
    where rhoJK = (-1)^(J+rhoKmult*K)
          where rhoKmult is 1 for Cnv and 0 otherwise.
    HR(K,q) transforms as the rotational function |K,q> so
        R- HR(K,q) = rhoJK(K,q)HR(K,-q)
    Now the overall transition moment operator transforms as:
        R- Htran = rhoTran Htran
    where rhoTran is given by Molecule.CharacterTable.DipoleSym for dipole transitions
    and 1 otherwise. As
        Htran = HE(k,q) HR(k,q)
    we have
        R- Htran = rhoTran Htran(-) = R- HE(k,q) R- HR(k,q)
                 = rhoHE HE(k,-q) rhoJK(K,q)HR(K,-q)
    where rhoHE, the multiplying factor on HE(k,q) is
        rhoHE = rhoTran rhoJK(K,q)
    Applying R- to a number has no effect, so
        <EV'|HE(k,q)|EV> = R- <EV'|HE(k,q)|EV> = <R- EV'|R- HE(k,q)|R- EV>
                         = rho' rhoTran rhoJK(K,q) rho <-EV'|HE(k,-q)|-EV>
    This gives the required phase factor; rhoJK(K,q) is actually included
    in HDip.
    Note that the above does not apply to C3h...,S4... as there is no operation that
    transforms K -> -K.
  }
  if not TransitionMoment.Magnetic and  (Minus(TransitionMoment.Rank) < 0) then begin
    Drho := Molecule.CharacterTable.IReps[Molecule.CharacterTable.DipoleSym].rho;
    if (Molecule.CharacterTable.IReps[Molecule.CharacterTable.DipoleSym].g > 0)
       and (Bra.Phase = 0) and (Ket.Phase = 0)
       and (Molecule.CharacterTable.En[Bra.Sym] > 0)
       and (Molecule.CharacterTable.En[Ket.Sym] > 0) then begin
      SignB := -SignB;
    end;
  end else
    Drho := 0;
  Drho := Molecule.CharacterTable.IReps[(Bra.State as TSymmetricTop).Symmetry].rho
          - Molecule.CharacterTable.IReps[(Ket.State as TSymmetricTop).Symmetry].rho
          + Drho;
  Kp := SignA*Bra.K;
  lp := SignA*Bra.l;
  K := SignB*Ket.K;
  l := SignB*Ket.l;
  Result.re := HDip(TransitionMoment as TSingleTransitionMoment, Bra.J, Kp, lp, Ket.J, K, l, Drho);
  Result.im := 0;
end;

procedure TSymmetricTop.SetSettingFromStr(AIndex: Integer; const Str: string);
begin
  if ((AIndex = LSI+1) or (AIndex = LSI+2)) and (Str = 'all') then
    SetSetting(AIndex,-9999)
  else
    inherited;
end;

function TSymmetricTop.SettingToStr(AIndex: Integer): string;
begin
  if ((AIndex = LSI+1) or (AIndex = LSI+2)) and (GetSetting(AIndex) < 0) then
      Result := 'all'
  else
    Result := inherited SettingToStr(AIndex);
end;

function TSymmetricTop.RotationalPerturbationMatrixElement(ABra, AKet: TBasisState;
  SignA, SignB: Integer; MNode: TMNode; Flag: Integer): Complex;
var
  Bra, Ket: TSymmetricTopBasisState;
  P: TSymmetricTopPerturbation;
begin
  if Flag <> Ord(meHrot) then
    raise NotYet;
  Bra := ABra as TSymmetricTopBasisState;
  Ket := AKet as TSymmetricTopBasisState;
  P := MNode as TSymmetricTopPerturbation;
  if Bra.J <> Ket.J then
    raise Exception.Create('TSymmetricTop.PerturbationMatrixElement');
  Result.re := P.MatrixElement(Bra.J, SignA*Bra.K, SignA*Bra.l,
                                      SignB*Ket.K, SignB*Ket.l);
  Result.im := 0;
end;

const
  SymmetricTopSettingNames: array[LSI..LSI+2] of string = (
'Symmetry',
'Minimum |K| to use; "all" (default) for no lower limit.',
'Maximum |K| to use; "all" (default) for no upper limit.'
);

function TSymmetricTop.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(SymmetricTopSettingNames))
     and (AIndex <= High(SymmetricTopSettingNames)) then
    Result := SymmetricTopSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

const
 SymmetricTopParameterNames: array [LSP..LSP+25] of string = (
'Rotational constant perpendicular to symmetry axis.',
'Rotational constant about symmetry axis (A or C).',
'J^2(J+1)^2 Quartic Centrifugal Distortion',
'J(J+1)K^2 Quartic Centrifugal Distortion',
'K^4 Quartic Centrifugal Distortion',
'Coriolis coupling constant.',
'J(J+1) dependence of Coriolis coupling constant.',
'K^2 dependence of Coriolis coupling constant.',
'l doubling constant.',
'l doubling constant.',
'l doubling constant.',
'J(J+1) dependence of qplus l doubling constant.',
'K^2 dependence of qplus l doubling constant.',
'J(J+1) dependence of r l doubling constant.',
'K^2 dependence of r l doubling constant.',
'J^3(J+1)^3 Sextic Centrifugal Distortion',
'J^2(J+1)^2K^2 Sextic Centrifugal Distortion',
'J(J+1)K^4 Sextic Centrifugal Distortion',
'K^6 Sextic Centrifugal Distortion',
'J^4(J+1)^4 Octic Centrifugal Distortion',
'J^3(J+1)^3K^2 Octic Centrifugal Distortion',
'J^2(J+1)^2K^4 Octic Centrifugal Distortion',
'J(J+1)K^6 Octic Centrifugal Distortion',
'K^8 Octic Centrifugal Distortion',
'multiplier of <K^2> for linewidth',
'multiplier of <J(J+1)-K^2> for linewidth'
);

function TSymmetricTop.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(SymmetricTopParameterNames))
     and (AIndex <= High(SymmetricTopParameterNames)) then
    Result := SymmetricTopParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

{ TSymmetricTopPerturbation }

const
  SymmetricTopPerturbationSettingNames: array[LPI..LPI+5] of string = (
'Twice power of J(J+1)',
'Power of Jz',
'Power of J+-',
'Power of l',
'Power of l+-; < 0 implies opposite sign to change in K',
'K the perturbation applies to'
);

constructor TSymmetricTopPerturbation.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',KSelect=all';
  inherited;
end;

function TSymmetricTopPerturbation.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(SymmetricTopPerturbationSettingNames))
     and (AIndex <= High(SymmetricTopPerturbationSettingNames)) then
    Result := SymmetricTopPerturbationSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TSymmetricTopPerturbation.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex = LPI+5;
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

function TSymmetricTopPerturbation.MatrixElement(J, Kp, lp, K, l: TQuantumNumber): Double;
begin
  Result := 0;
  if KSelect > -9999 then begin
    { For the moment, require K selection to be diagonal in |K| }
    { TODO: Allow K select to work for elements off diagonal in K }
    if (Abs(K) <> KSelect) or (Abs(Kp) <> KSelect) then
      Exit;
  end;
  if Abs(lChange)*2 <> Abs(l-lp) then
    Exit;
  if pPower*2 <> Abs(K-Kp) then
    Exit;
  if lChange*pPower*(l-lp)*(K-Kp) < 0 then
    Exit; { lChange < 0 => Change in l has opposite sign to change in K }
  Result := rl(J, Kp, K);
  if zPower > 0 then
    if pPower <> 0 then
      Result := Result * (IntPower(K,zPower) + IntPower(Kp,zPower))
    else
      Result := Result * IntPower(K,zPower);
  if JPower > 0 then
    Result := Result * IntPower(J*(J+2), JPower div 2);
  if lPower <> 0 then
    if lChange <> 0 then
      Result := Result * Sign(lp-l) * lPower
    else
      Result := Result * Sign(lp) * lPower;
  Result := Result * Value / IntPower(2,zPower + JPower);
end;

function TSymmetricTopPerturbation.PairName: string;
  function Show(const Letter: string; Power: Integer): string;
  begin
    case Power of
      0: Result := '';
      1: Result := Letter;
    else
      Result := Letter + '^' + IntToStr(Power)
    end;
  end;
begin
  Result := Show('J',JPower) + Show('Jz',zPower) + Show('J+-', pPower) + Show('l', lPower);
  if lChange < 0 then
    Result := Result + Show('l+-',Abs(lChange))
  else
    Result := Result + Show('l-+',lChange);
  if KSelect > -9999 then
    Result := Result + '(K=' + (RootNode as TMixture).QuantumNumberToStr(KSelect) + ')';
end;

procedure TSymmetricTopPerturbation.SetSettingFromStr(AIndex: Integer;
  const s: string);
begin
  if (AIndex = LPI+5) and (s = 'all') then
    KSelect := -9999
  else
    inherited;
end;

function TSymmetricTopPerturbation.SettingToStr(AIndex: Integer): string;
begin
  if (AIndex = LPI+5) and (KSelect <= -9999) then
      Result := 'all'
  else
    Result := inherited SettingToStr(AIndex);
end;

{ TsymmetricTopNucleus }

function TsymmetricTopNucleus.FixupProperties: Exception;
begin
  CHI[ 0] := eQq/2;
  if Spin > 1 then
    QuadI := 1/threej(Spin, -Spin, 4, 0, Spin, Spin)
  else
    QuadI := 0;
  { Nuclear spin rotation }
  { Note that off-axis nuclei do not require Cxx = Cyy, though
    the difference between them will rarely be determinable }
  TC00 := -(Czz + 2*Cperp);
  TC20 := (3*Czz + TC00)/Sqrt(Six);
  TC00 := TC00/Sqrt(Three);
  Result:=inherited FixupProperties;
end;

function TsymmetricTopNucleus.HasIdentical: Boolean;
begin
  Result := True;
end;

const
  SymmetricTopNucleusParameterNames: array[1..3] of string = (
'Qudrupole Coupling Constant',
'Component along symmetry axis of Nuclear Spin - Rotation interaction.',
'Average of components perpendicular to symmetry axis of Nuclear Spin - Rotation interaction.'
);

function TsymmetricTopNucleus.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(SymmetricTopNucleusParameterNames))
     and (AIndex <= High(SymmetricTopNucleusParameterNames)) then
    Result := SymmetricTopNucleusParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

initialization
  AddTables;
  TSymmetricMolecule.RegisterClass(TSpecies);
  TSymmetricManifold.RegisterClass(TSymmetricMolecule);
  TSymmetricTop.RegisterClass(TSymmetricManifold);
  TSymmetricTopPerturbation.RegisterClass(TSymmetricManifold);
  TSymmetricTopNucleus.RegisterClass(TSymmetricTop);
end.
