unit Vib;
{ Vibrational Energy Levels }

{$I DFS.inc}{Copyright (C) 2008 C M Western; see commandline.pas for license}

{ TODO: Plot normal modes }
{ TODO: Update DushinskyFrm if calculate following change in l matrix }
{ TODO: Should allow direct entry of Dushinsky matrix }

{ TODO: Odd Spin }

interface

uses SysUtils, Classes, MNode, Mixture, QuantumNumber, MathSup, pgDOM,
     Abelian, { Note here in uses list so point group names not confused }
     MatrixVec, Vector, MatrixElements, QuantumMath, Interlock, Expr;

type

  TPointGroup = (C1=0, Ci, C2, Cs, D2, C2v, C2h, D2h, CInfinityv, DInfinityh);

  TPointGroupInfo = class(TPersistent)
  private
    FPointGroup: TPointGroup;
    FLastSymmetry: Integer;
    FHasDegenerateModes: Boolean;
  public
    Name: string;
    SettingsNames: PPChar;
    constructor Create(APointGroup: TPointGroup);
    function Allowed(Bra, Dipole, Ket: Integer):Boolean; virtual; abstract;
    function TryStrToSymmetry(Str: string; out Symmetry:Integer):Boolean; virtual;
    function StrToSymmetry(Str: string):Integer;
    function SymmetryToStr(Symmetry:Integer):string; virtual;
    property HasDegenerateModes:Boolean read FHasDegenerateModes;
    property LastSymmetry:Integer read FLastSymmetry;
  published
    property PointGroup: TPointGroup read FPointGroup; { To allow conversion to string }
  end;

  TVibratingMolecule = class;

  TBaseVibrationalMode = class(TMNode)
  protected
    function GetName: TComponentName; override;
    procedure SetName(const Value: TComponentName); override;
  public
    constructor Create(AOwner: TComponent); override;
    function FixupProperties: Exception; override;
    function Molecule: TVibratingMolecule;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    procedure SetSetting(AIndex: Integer; const AValue: Integer); override;
    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; override;
    function SettingToStr(AIndex: Integer): string; override;
    procedure SetSettingFromStr(AIndex: Integer; const Str: string); override;
    function HelpOnSetting(AIndex: Integer): string; override;
  published
    property Symmetry: Integer index 1 read GetSetting write SetSetting;
  end;

  TVibrationalOperator = (x, q, qSigned, Gaussian);

  TVibrationalPerturbation = class (TPerturbation)
  private
    FMolecule: TVibratingMolecule;
    FTotalChange: Integer;
    FRTMode: Integer; { Mode containing a Renner-Teller type operator mixing electronic states }
    procedure SetOperator(const Index: Integer; const AValue: TVibrationalOperator);
    function GetOperator(const Index: Integer): TVibrationalOperator;
  public
    constructor Create(AOwner: TComponent); override;
    function FixupProperties: Exception; override;
    function PairName: string; override;
    procedure AdjustSettingNames;
    function MatrixElement(ABra, AKet: TIntegerArray; SignA, SignB: Integer):Double;
    function CheckMatrixElementsAvailable: Boolean; override;
    function SymbolicMatrixElements:TMatrixElements; override;
    function EffectiveQuantumNumber(v: array of Integer):Integer;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    property Molecule: TVibratingMolecule read FMolecule;
  published
    { Exponent is only used for Gaussian type perturbation }
    property Exponent:  Double index 2 read GetParameter write SetParameter;
    property Op:TVibrationalOperator    index LPI read GetOperator write SetOperator;
    property EffectiveQno:Boolean       index LPI+1 read GetFlag write SetFlag;
  end;

  TVibPertEquiv = record
    Op:TVibrationalOperator;
    Powers: array of Integer;
  end;


  TVibrationalMode = class(TBaseVibrationalMode)
  public
    constructor Create(AOwner: TComponent); override;
    procedure AdjustParameterNames;
    function PertEquiv(AIndex: Integer):TVibPertEquiv;
    procedure GlobalShift(Shift:Integer);
    procedure GlobalMoveUp(Sender: TObject);
    procedure GlobalMoveDown(Sender: TObject);
    procedure Sort(Sender: TObject);
    procedure AddHandlers; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
  published
    property Omega:    Double index  1 read GetParameter write SetParameter;
    property xOmega:   Double index  2 read GetParameter write SetParameter;
    property yOmega:   Double index  3 read GetParameter write SetParameter;
    property zOmega:   Double index  4 read GetParameter write SetParameter;
    property gll:      Double index  5 read GetParameter write SetParameter;
    property epsOmega: Double index  6 read GetParameter write SetParameter;

    property vMin: Integer     index 2 read GetSetting write SetSetting;
    property vMax: Integer     index 3 read GetSetting write SetSetting;
  end;

  TElectronicState = class (TState)
  private
    FMolecule: TVibratingMolecule;
    FModes: TMNodeArray;
    FEffectiveQuanta: array of TVibrationalPerturbation;
  public
    adm, Madm, Itrans: TMatrix;
    Ixx: array [0..2] of Double; { Diagonal moments of inertia }

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function FixupProperties:Exception; override;
    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; override;
    function SettingToStr(AIndex: Integer): string; override;
    procedure SetSettingFromStr(AIndex: Integer; const Str: string); override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;

    function AddOneJ(Eigenvalues: TEigenvalues; J: TQuantumNumber; Sym: Integer; HyperfineState:THyperfineBasis; Seffective: TQuantumNumber; out StatWtGroup: Integer): Integer; override;
    function RotationalMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Complex; override;
    function RotationalTransitionMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; TransitionMoment: TMNode; F: TQuantumNumber):Complex; override;
    function RotationalPerturbationMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMNode; Flag: Integer):Complex; override;
    function ReadBranchFormat(Text: string; var Obs: TObservation; UpperState: TState; Reader: TObservationsReaderBase):Boolean; override;
    function StateOrigin: Double; override;
    function SymmetryOf(v: array of Integer; out AlternateSym:Integer):Integer;

    function ConvertParameter(AIndex: Integer; Zero: Boolean):TPerturbation; override;
    function MatrixElementType(const Changes: array of Integer; const H: string):string; override;
    function SymbolicRotationalMatrixElement(AIndex: Integer):TMatrixElements; override; { For given parameter }
    function ParameterVariable(AIndex: Integer):TSortedVariableNode; override;
    function nParameterVariables: Integer; override;

    procedure AddHandlers; override;
    procedure PrintMolecularInfo(Sender: TObject);
    procedure FixCentreOfMass(Sender: TObject);
    procedure Normalize(Sender: TObject);
    procedure Orthogonalize(Sender: TObject);
    procedure CentreOfMass(Fix: Boolean; Print: Boolean);
    procedure PrepareMomentsOfInteria(Print: Boolean);
    procedure SetLmatrix(Sender: TObject);
    procedure AdjustModes(FixNumber: Boolean);

    procedure PrepareMatrices(Print: Boolean; AllModes:Boolean = True);

    function AddMode:TVibrationalMode;
    function Mode(Index:Integer): TVibrationalMode;
    property Molecule: TVibratingMolecule read FMolecule;
  published
    property A:        Double index  LSP   read GetParameter write SetParameter; { Spin orbit coupling }
    property LambdaSS: Double index  LSP+1 read GetParameter write SetParameter; { Spin-spin coupling }
    property o:        Double index  LSP+2 read GetParameter write SetParameter; { Omega = 0 lambda doubling }
    property Emax:     Double index  LSP+3 read GetParameter write SetParameter; { Energy cut off for basis }

    property Symmetry: Integer index LSI   read GetSetting write SetSetting;
  end;

  TMolecularNucleus = class(TNucleus)
  protected
    procedure SetComponentIndex(const Value: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure AddHandlers; override;
    procedure DoFindExactMass(Sender: TObject);
    procedure SetParameter(AIndex: Integer; const Value: Double); override;
    procedure GlobalShift(Shift:Integer); override;
    function HelpOnParameter(AIndex: Integer): string; override;
  published
    property Mass:     Double index  1 read GetParameter write SetParameter;
    property x:        Double index  2 read GetParameter write SetParameter;
    property y:        Double index  3 read GetParameter write SetParameter;
    property z:        Double index  4 read GetParameter write SetParameter;
  end;

  TVibrationalManifold = class(TManifold)
  public
    function nSym: Integer; override; { Number of different symmetries }
    function FindJmax(Default: Integer; RemovedSpins: Byte = 0):Integer; override;
  end;

  TModeInfo = record
    Symmetry: Integer;
    Name: string;
    dIndex: Integer; { Index in basis set v array of l quantum number }
    Degeneracy: Integer; { use v + Degeneracy / 2 in quantum number expressions }
    ShowQno: Boolean; { If this quantum number is to be displayed }
  end;

  TVibratingMolecule = class(TMolecule)
  private
    FNuclearMasses: array of Double;
    FnQuantumNumbers: Integer;
    FHasElectronicAngularMomentum: Boolean;
    FMass: Double;
    function GetPointGroup(const Index: Integer): TPointGroup;
    procedure SetPointGroup(const Index: Integer; const AValue: TPointGroup);
    procedure SetnModes(const Index, Value: Integer);
  protected
    procedure SetnNuclei(const Index, Value: Integer); override;
    procedure FixMass;
  public
    ModeInfo: array of TModeInfo;
    constructor Create(AOwner: TComponent); override;
    function IsValidChildClass(MNodeClass: TMNodeClass):Boolean; override;
    function FixupProperties:Exception; override;
    function Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean; override;
    function JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer; override;
    function LastSym: Integer; override;
    function QnoInfo(QnoType: TQnoType): TQnoInfo; override;
    function SymName(Sym: Integer):string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function IsDegenerate(Symmetry: Integer):Boolean;
    function AngularMomentum(Symmetry: Integer):TQuantumNumber;
    property nQuantumNumbers: Integer read FnQuantumNumbers;
    property HasElectronicAngularMomentum: Boolean read FHasElectronicAngularMomentum;
    property Mass: Double read FMass;
  published
    property PointGroup:     TPointGroup index LMI read GetPointGroup write SetPointGroup;
    property nModes:         Integer     index LMI+1 read GetSetting write SetnModes;
    property ShowAllQnos:    Boolean     index LMI+2 read GetFlag write SetFlag;
  end;

  { TVibrationalBasisState }

  TVibrationalBasisState = class(TBasisState)
  private
    FOriginalPosition: Integer; { Position in array sorted by vibrational quantum numbers }
    FOffset: Integer; { Position within basis }
  public
    v: TIntegerArray; { v1,l1, v2,l2, ... vn, ln, Lambda, Omega;
                        l only present for degenerate modes;
                        Lambda, Omega only present if any state in molecule has Lambda > 0 or S > 0 }
    function Qno(QnoType: TQnoType): TQuantumNumber; override;
    function VibLabel: string;
    function VibLabelKey: string;
    function BasisLabel: string; override; { Basis State Label }
    function BasisLabelKey: string; override; { Explanation of Basis State Label }
    function StateLabel(AQuantumNumberFormat: TQuantumNumberFormat): string; override;
    function StateLabelKey: string; override; { Explanation of State Name }
    function TxName(Upper, Lower: TBasisState): string; override;
    function TxNameKey(Upper, Lower: TBasisState): string; override;
    function MakeSignature(SignA:Integer; AKet: TBasisState; SignB: Integer): TIntegerArray; override;
    procedure SetSymbolicVariables(SignA:Integer; AKet: TBasisState; SignB: Integer); override;
    property OriginalPosition: Integer read FOriginalPosition;
    property Offset: Integer read FOffset;
  end;

  TVibrationalModeChange = class(TBaseVibrationalMode)
  public
    procedure AdjustParameterNames;
    function CanShift(Shift:Integer):Boolean; override;
    function HelpOnParameter(AIndex: Integer): string; override;
  published
    property Displacement:  Double index  1 read GetParameter write SetParameter;
  end;

  TCachedTransitionMatrix = class(TInterlockedComponent)
  public
    BraSample,KetSample: TVibrationalBasisState; { Only valid during construction }
    BraOffsets, KetOffsets: TIntegerArray;
    TransitionMatrix: TMatrix;
    TransitionMoment: TTransitionMoment;
    destructor Destroy; override;
    procedure DoMake(RequiredAttributes: Pointer); override;
    function HasCorrectAttributes(Attributes: Pointer):Boolean; override;
  end;

  TMultidimensionalFCF = class(TSingleTransitionMoment)
  private
    FMolecule: TVibratingMolecule;
    FModes: TMNodeArray;
    FActualSymmetry: Integer;
    FMatrixCache: array of TCachedTransitionMatrix;
    CacheLock: TMultiReadExclusiveWriteSynchronizer;
    procedure ClearMatrices;
  public
    d: TColumnVector;
    S, Q1, P1, R: TMatrix;
    Rdelta, P1Delta: TColumnVector;
    I00: Double;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function FixupProperties:Exception; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    procedure ClearCache; override;
    procedure PrepareDushinsky(nModes: Integer; Print: Boolean; SetModeInfo: Boolean);
    procedure PrepareMatrices(Print: Boolean; SetModeInfo, AllModes: Boolean);
    procedure ShowDushinsky(Sender: TObject);
    procedure AddHandlers; override;
    procedure PrintInfo(Sender: TObject);
    procedure SetModeInfo(Sender: TObject);
    function PairName: string; override;
    procedure AdjustModes(FixNumber: Boolean);
    procedure SwapModes(nMode1, nMode2: Integer);
    function Mode(Index:Integer): TVibrationalModeChange;
    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; override;
    function SettingToStr(AIndex: Integer): string; override;
    procedure SetSettingFromStr(AIndex: Integer; const Str: string); override;
    function HelpOnSetting(AIndex: Integer): string; override;

    function Overlap(var ABra, AKet: array of Integer):Double;
    procedure InvertTransformation;
    function TransitionMatrix(BraSample,KetSample: TVibrationalBasisState; Cache: Boolean):TMatrix;

    property Molecule: TVibratingMolecule read FMolecule;
    property ActualSymmetry: Integer read FActualSymmetry;
  published
    property Symmetry:  Integer index 1 read GetSetting write SetSetting;
    property Calculate: Boolean index 2 read GetFlag write SetFlag;
  end;

var
  ReduceFrequency: Double; { 4 pi^2 10^-21 c / h / NA = 0.02966016141352712 }

implementation

uses {$IFNDEF CONSOLE} LmatrixFrm, DushinskyFrm, {$ENDIF}
     NuclearInfo, Units, Worker, MatrixOps, Math, TextStrm, Profile,
     Global, PropHelp, EigLA, Decomp, Progress;

var
  PointGroupInfo: array [TPointGroup] of TPointGroupInfo;

  eLambda, eOmega: TSortedVariableNode;
  elVariables: TVariableList; { Contains eLambda, eOmega }
  vVariables, lVariables: TVariableList; { These contain v1, v2, ... and l1, l2, ..., filled as required using the functions below }

function FindVariable(Prefix: Char; Index: Integer; var VariableList: TVariableList):TSortedVariableNode;
var
  Name: string;
  i: Integer;
begin
  if not Assigned(VariableList) then
    VariableList := TVariableList.Create;
  Name := Prefix+IntToStr(Index+1);
  i := VariableList.IndexOf(Name);
  if i < 0 then { Just create as needed }
    VariableList.Add(TSortedVariable.Create(Name,Index));
  i := VariableList.IndexOf(Name);
  Result := (VariableList.Objects[i] as TSortedVariable).Template;
end;

function vVariable(Index: Integer):TSortedVariableNode;
begin
  Result := FindVariable('v',Index,vVariables);
end;

function lVariable(Index: Integer):TSortedVariableNode;
begin
  Result := FindVariable('l',Index,lVariables);
end;

function vnvp(v, n, vp: Integer):Double;
{ Matrix elements of <v|q^n|v'> in a harmonic basis }
var
  vv: Double;
begin
  Result := 0;
  if (Abs(v-vp) > n) or Odd(v+n+vp) then
    Exit;
  if n < 0 then
    raise NotYet;
  case n of
    0: Result := 1;
    1: Result := Sqrt(Max(v,vp)/2);
    2: if v = vp then
         Result := v + 0.5
       else begin
         vv := Max(v, vp);
         vnvp := Sqrt( vv*(vv-1) / 4 );
       end;
    3: begin
         vv := Max(v, vp);
         if Abs(v-vp) = 1 then
           Result := 3 * Sqrt( vv*vv*vv / 8 )
         else
           Result := Sqrt( vv*(vv-1)*(vv-2) / 8 );
       end;
    4: begin
         vv := Max(v, vp);
         if v = vp then
           Result := 3/4 * (2*vv*vv + 2*vv + 1)
         else if Abs(v-vp) = 2 then
           Result := (2*vv-1) * sqrt( vv*(vv-1) /4 )
         else
           Result := Sqrt( vv*(vv-1)*(vv-2)*(vv-3) / 16 )
       end;
    else
      { Use recursion: }
      { <v|q^n|v'> = sqrt((v+1)/2) * <v+1|q^n-1|v'> + sqrt(v/2) * <v-1|q^n-1|v'> }
      Result := Sqrt((v+1)/2) * vnvp(v+1,n-1,vp);
      if v > 0 then
        Result := Result + Sqrt(v/2) * vnvp(v-1,n-1,vp);
    end;
end;

function vnvpE(n: Integer; v: TExprNode; dv: Integer):TExprNode;
{ Symbolic Matrix elements of <v+dv|q^n|v> in a harmonic basis }
var
  nm: TExprNode;
begin
  Result := nil;
  if (Abs(dv) > n) or Odd(dv+n) then
    Exit;
  if n < 0 then
    raise NotYet;
  case n of
    0: Result := TIntegerNode.Create(1);
    1: begin
         if dv > 0 then
           Result := EB([v, '+', dv, '/', 2])
         else
           Result := EB([v, '/', 2]);
         Result := SqrtFunction.CreateExprNode(Result);
       end;
  else
    { Use recursion }
    nm := EB([v, '+', dv-1]);
    Result := EN([vnvpE(1,nm,1), '*', vnvpE(n-1,v,dv-1)]);
    nm.Free;
    nm := EB([v, '+', dv+1]);
    Result := EN([vnvpE(1,nm,-1), '*', vnvpE(n-1,v,dv+1), '+', Result]);
    nm.Free;
  end;
end;

function vnvpl(v, n, vp, ll: Integer):Double;
{ Matrix elements of <v ll|q^n|vp ll> in a 2D harmonic basis. }
var
  vv: Integer;
begin
  Result := 0;
  if (Abs(v-vp) > n) or Odd(v+n+vp) then
    Exit;
  if (n < 0) or Odd(n) then
    raise NotYet;
  case n of
    0: Result := 1;
    2: if v = vp then
         Result := v+1
       else begin
         vv := Max(v, vp);
         Result := Sqrt((vv*vv-ll*ll) / 4 );
       end;
  else { Recurse }
    vv := Max(v-2, Abs(ll));
    while vv <= v+2 do begin
      Result := Result + vnvpl(v,2,vv,ll)*vnvpl(vv,n-2,vp,ll);
      Inc(vv,2);
    end;
  end;
end;

function vnvplE(n: Integer; v, ll: TExprNode; dv: Integer):TExprNode;
{ Symbolic matrix elements of <v+dv ll|q^n|v ll> in a 2D harmonic basis. }
var
  ddv: Integer;
  np: TExprNode;
begin
  Result := nil;
  if (Abs(dv) > n) or Odd(dv+n) then
    Exit;
  if (n < 0) or Odd(n) then
    raise NotYet;
  case n of
    0: Result := TIntegerNode.Create(1);
    2: if dv = 0 then
         Result := EB([v, '+', 1])
       else begin
         if dv < 0 then
           Result := v.Duplicate
         else
           Result := EB([v, '+', dv]);
         Result := EN([SqrFunction.CreateExprNode(Result), '-', SqrFunction.CreateExprNode(ll.Duplicate), '/', 4]);
         Result := SqrtFunction.CreateExprNode(Result);
       end;
  else { Recurse }
    for ddv := -1 to 1 do begin
      np := EB([v, '+', dv+ddv*2]);
      Result := EN([vnvplE(2,np,ll,-ddv*2), '*', vnvplE(n-2,v,ll,dv+ddv*2), '+', Result]);
      np.Free;
    end;
  end;
end;

function vnvpll(vp, llp, n, v, ll: Integer):Double;
{ Matrix elements of <v ll|q+-^n|vp llp> in a 2D harmonic basis. }
var
  vi, li: Integer;
begin
  Result := 0;
  if (Abs(v-vp) > n) or Odd(v+n+vp) then
    Exit;
  if n = 0 then begin
    if ll = llp then
      Result := 1
  end else if n = 1 then begin
    { q+- }
    if vp = v+1 then begin
      if llp = ll+1 then
        Result := sqrt((v+2+ll)/2)
      else if llp = ll-1 then
        Result := sqrt((v+2-ll)/2)
    end else if vp = v-1 then begin
      if llp = ll+1 then
        Result := sqrt((v-ll)/2)
      else if llp = ll-1 then
        Result := sqrt((v+ll)/2)
    end;
  end else begin
    { General case by recursion }
    vi := vp-1;
    while vi <= vp+1 do begin
      if llp > ll then
        li := llp - 1
      else
        li := llp + 1;
      if Abs(li) <= vi THEN
        Result := Result + vnvpll(vp,llp,1,vi,li)*vnvpll(vi,li,n-1,v,ll);
      Inc(vi,2);
    end;
  end;
end;

function vnvpllE(n: Integer; v, ll:TExprNode; dv, dl: Integer):TExprNode;
{ Matrix elements of <v+dv ll+dl|q+-^n|v ll> in a 2D harmonic basis. }
var
  ddl: Integer;
  n1, n2, np: TExprNode;
begin
  Result := nil;
  if (Abs(dv) > n) or Odd(dv+n) then
    Exit;
  if n < 0 then
    raise NotYet;
  case n of
    0: if dl = 0 then
         Result := TIntegerNode.Create(1);
    1: if dv = -1 then begin
         if dl = -1 then
           Result := EB([v, '+', ll, '/', 2, 'sqrt'])
         else if dl = 1 then
           Result := EB([v, '-', ll, '/', 2, 'sqrt']);
       end else if dv = 1 then begin
         if dl = -1 then
           Result := EB([v, '-', ll, '+', 2, '/', 2, 'sqrt'])
         else if dl = 1 then
           Result := EB([v, '+', ll, '+', 2, '/', 2, 'sqrt']);
       end;
  else
    { General case by recursion }
    if dl > 0 then
      ddl := 1
    else
      ddl := -1;
    n1 := EB([v, '+', dv-1]);
    n2 := EB([v, '+', dv+1]);
    np := EB([ll, '+', dl-ddl]);
    Result := EN([vnvpllE(1,n1,np,1,ddl), '*', vnvpllE(n-1,v,ll,dv-1,dl-ddl)]);
    Result := EN([vnvpllE(1,n2,np,-1,ddl), '*', vnvpllE(n-1,v,ll,dv+1,dl-ddl), '+', Result]);
    n1.Free;
    n2.Free;
    np.Free;
  end;
end;

{ Matrix elements of exp(-bq^2)
  Use recursion formula in
  J. B. Coon, N. W. Naugle and R. D. McKenzie
  JMS 20, 107 (1966)
  See also (Rather slower):
  S.I. Chan and D. Stelman. J. Chem. Phys. 39, 545 (1963)
}
{ TODO: This routine becomes very slow for larger values of v;
  may need to cache values }
function vGaussian(v, vp:Integer; b:Double):Double;
//var
//  t: Integer;
//  function vGO(n, m:Integer):Double;
//  begin
//    if n = 0 then
//      if m = 0 then
//        Result := 1/Sqrt(1+b)
//      else
//        Result := -vGO(0,m-2)*Sqrt((m-1)/m)*b/(1+b)
//    else begin
//      Result := Sqrt(m)*vGO(n-1,m-1)+Sqrt(m+1)*vGO(n-1,m+1);
//      if n > 1 then
//        Result := (Result-Sqrt(n-1)*vGO(n-2,m))/Sqrt(n)
//    end;
//  end;
  function vG(m, n:Integer):Double;
  begin
    if n = 0 then
      if m = 0 then
        Result := 1/Sqrt(1+b)
      else
        Result := -vG(m-2,0)*Sqrt((m-1)/m)*b/(1+b)
    else if m = 0 then
        Result := -vG(0,n-2)*Sqrt((n-1)/n)*b/(1+b)
    else begin
      Result := Sqrt(n/m)/(1+b)*vG(m-1,n-1);
      if m > 1 then
        Result := Result - Sqrt((m-1)/m)*b/(1+b)*vG(m-2,n);
    end;
  end;
begin
  Result := 0;
  if Odd(v-vp) then
    Exit;
  if v > vp then
    Result := vG(v,vp)
  else
    Result := vG(vp,v);
end;

{$IFDEF TEST}
function vnvptest(v, n, vp: Integer):Double;
begin
  Result := 0;
  if (Abs(v-vp) > n) or Odd(v+n+vp) then
    Exit;
  case n of
    0: Result := 1;
    1: Result := Sqrt(Max(v,vp)/2);
    else
      { Use recursion: }
      { <v|q^n|v'> = sqrt((v+1)/2) * <v+1|q^n-1|v'> + sqrt(v/2) * <v-1|q^n-1|v'> }
      Result := Sqrt((v+1)/2) * vnvp(v+1,n-1,vp);
      if v > 0 then
        Result := Result + Sqrt(v/2) * vnvp(v-1,n-1,vp);
    end;
end;

procedure TestVib;
var
  v, v1, n: Integer;
  o, p: Double;
begin
  for v := 0 to 10 do
    for n := 0 to 10 do
      for v1 := 0 to 10 do begin
        o := vnvp(v,n,v1);
        p := vnvptest(v,n,v1);
        if not SameValue(o,p) then begin
          WriteLn(Log,v,n,v1,o,p);
          Abort;
        end;
      end;
end;
{$ENDIF}

{ TVibrationalPerturbation }

procedure TVibrationalPerturbation.AdjustSettingNames;
var
  i: Integer;
begin
  for i := nSettings+1 to FirstDynamicSetting + Molecule.nModes - 1 do
    AddSetting(Molecule.ModeInfo[i-FirstDynamicSetting].Name,'0');
  for i := nSettings downto FirstDynamicSetting + Molecule.nModes do
    DeleteSetting(i);
  for i := FirstDynamicSetting to nSettings do
    SetSettingName(i, Molecule.ModeInfo[i-FirstDynamicSetting].Name);
end;

function TVibrationalPerturbation.CheckMatrixElementsAvailable: Boolean;
begin
  if Op = Gaussian then
    Result := False { Not implementable in closed form }
  else
    Result := inherited CheckMatrixElementsAvailable;
end;

constructor TVibrationalPerturbation.Create(AOwner: TComponent);
begin
  if Assigned(AOwner) then
    FMolecule := AOwner.Owner as TVibratingMolecule;
  inherited;
  if Assigned(AOwner) then
    AdjustSettingNames;
end;

function TVibrationalPerturbation.EffectiveQuantumNumber(v: array of Integer): Integer;
var
  m, p: Integer;
begin
  Result := 0;
  for m := 1 to Molecule.nModes do begin
    p := GetSetting(FirstDynamicSetting+m-1);
    if p <> 0 then
      Result := Result + v[m-1]*(FTotalChange-Abs(p));
  end;
end;

function TVibrationalPerturbation.FixupProperties: Exception;
var
  m, p: Integer;
begin
  Result := inherited FixupProperties;
  AdjustSettingNames;
  FTotalChange := 0;
  FRTMode := -1;
  for m := 0 to Molecule.nModes-1 do begin
    p := GetSetting(FirstDynamicSetting+m);
    Inc(FTotalChange,Abs(p));
    if (Op = Gaussian) and (p <> 0) then begin
      if (FRTMode >= 0) or Molecule.IsDegenerate(Molecule.ModeInfo[m].Symmetry) then
        Result := EReportException.Create('Only 1D Gaussian barrier implemented in '+Name);
      FRTMode := m;
    end else if (p < 0) and (Op = q) then
      if FRTMode >= 0 then
        Result := EReportException.Create('Only one Renner-Teller mode implemented in '+Name)
      else if not Molecule.IsDegenerate(Molecule.ModeInfo[m].Symmetry) then
        Result := EReportException.Create('Renner-Teller operator must involve degenerate mode in '+Name)
      else
        FRTMode := m;
  end;
  if EffectiveQno and (FTotalChange = 1) then
    FTotalChange := 2; { This allows a single quantum number to be selected }
end;

function TVibrationalPerturbation.GetOperator(const Index: Integer): TVibrationalOperator;
begin
  Result := TVibrationalOperator(GetSetting(Index));
end;

const
 VibrationalPerturbationParameterNames: array [2..2] of string = (
'Exponent for Gaussian type perturbation.'
);

function TVibrationalPerturbation.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(VibrationalPerturbationParameterNames))
     and (AIndex <= High(VibrationalPerturbationParameterNames)) then
    Result := VibrationalPerturbationParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

const
 VibrationalPerturbationSettingNames: array [LPI..LPI+1] of string = (
'Type of perturbation.',
'Set to group together states connected by this perturbation when assigning quantum numbers.'
);

function TVibrationalPerturbation.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex > High(VibrationalPerturbationSettingNames) then
    Result := 'Power of ' + (Bra as TElectronicState).Mode(AIndex-High(VibrationalPerturbationSettingNames)-1).Name
              + ' in perturbation.'
  else if AIndex >= Low(VibrationalPerturbationSettingNames) then
    Result := VibrationalPerturbationSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

procedure TVibrationalPerturbation.SetOperator(const Index: Integer; const AValue: TVibrationalOperator);
begin
  SetSetting(Index, Ord(AValue));
end;

function SymbolicMatrixElements(PertEquiv: TVibPertEquiv; Molecule: TVibratingMolecule): TMatrixElements;
var
  RTMode: Integer;
  procedure AddVibChanges(vn, p: Integer);
    { Add all possible changes for given vibrational mode }
    { p is power of operator, and possible changes are -p .. p in steps of 2 }
  var
    Old, nNew, i, j, Base, dind: Integer;
    Expr: TExprNode;
  begin
    Old := Length(Result);
    nNew := Abs(p)+1; { Number of different matrix elements from this mode }
    if Old = 0 then
      SetLength(Result, nNew)
    else
      SetLength(Result, Length(Result)*nNew);
    for i := Max(Old-1,0) downto 0 do begin
      { For each old set of quantum numbers, duplicate and concatenate new set of changes }
      Base := i*nNew; { Start of block to be updated }
      if vn = 0 then begin
        SetLength(Result[0].Signature,Molecule.nQuantumNumbers);
        Result[0].Expr := TIntegerNode.Create(1);
      end;
      for j := Base+nNew-1 downto Base+1 do begin
        Result[j].Signature := Copy(Result[i].Signature);
        if Assigned(Result[i].Expr) then
          Result[j].Expr := Result[i].Expr.Duplicate;
      end;
      if Base <> i then { This test should not be required, byt FPC 2.2.3 generates bad code for this }
        Result[Base].Signature := Result[i].Signature; { Can use current memory for one }
      Result[Base].Expr := Result[i].Expr;
      for j := Base to Base+nNew-1 do begin
        Result[j].Signature[vn] := -Abs(p) + (j-Base)*2;
        if ((PertEquiv.Op = qSigned) and (Abs(p) <> Abs(Result[j].Signature[vn])))
           or not Assigned(Result[j].Expr) then begin
          { This filters out some of the extra elements for qSigned, those with
            the wrong magnitude of the change in v }
          FreeAndNil(Result[j].Expr);
          Expr := nil;
        end else begin
          if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
            dind := Molecule.ModeInfo[vn].dIndex;
            if (p < 0) and (PertEquiv.Op = q) then begin
              if RTMode < 0 then
                RTMode := vn
              else if RTMode <> vn then                   
                raise EReportException.Create('Mode than one RT active mode in operator');
              { Delta Lambda = - Delta p = +/- p }
              Result[j].Signature[dind] := -p;
              Result[j].Signature[Molecule.nQuantumNumbers-2] := p;
              Result[j].Signature[Molecule.nQuantumNumbers-1] := p; { Delta sigma = 0 }
              Expr := vnvpllE(Abs(p),vVariable(vn),lVariable(vn),Result[j].Signature[vn],-p);
              { Need to duplicate this with opposite signs - see DuplicateRTChanges }
            end else begin
              Expr := vnvplE(Abs(p),vVariable(vn),lVariable(vn),Result[j].Signature[vn]);
              Result[j].Signature[dind] := 0;
            end;
          end else
            Expr := vnvpE(Abs(p),vVariable(vn),Result[j].Signature[vn]);
          Expr := EN([Result[j].Expr, '*', Expr]);
        end;
        Result[j].Expr := Expr;
      end;
    end;
  end;
  procedure CheckSigns;
  var
    r, w, vn, p, RelSign: Integer;
  begin
    { Check for relative signes of changes for qSigned;
      also remove completely zero elements }
    w := 0;
    for r := 0 to High(Result) do begin
      if Assigned(Result[r].Expr) then begin
        RelSign := 0;
        for vn := 0 to Molecule.nModes-1 do begin
          p := PertEquiv.Powers[vn];
          if p <> 0 then begin
            if RelSign = 0 then
              if p*Result[r].Signature[vn] > 0 then
                RelSign := 1
              else
                RelSign := -1;
            if p * RelSign <> Result[r].Signature[vn] then
              FreeAndNil(Result[r].Expr);
          end;
        end;
        if Assigned(Result[r].Expr) then begin
          Result[w] := Result[r];
          Inc(w);
        end;
      end;
    end;
    SetLength(Result, w);
  end;
  procedure DuplicateRTChanges;
  var
    Old, r, p: Integer;
  begin
    { Each matrix element needs to be duplicated with opposite sign changes in the angular momenta }
    Old := Length(Result);
    SetLength(Result, Old*2);
    for r := Old to High(Result) do begin
      Result[r].Signature := Copy(Result[r-Old].Signature);
      p := -Result[r].Signature[Molecule.nQuantumNumbers-2];
      Result[r].Signature[Molecule.ModeInfo[RTMode].dIndex] := -p;
      Result[r].Signature[Molecule.nQuantumNumbers-2] := p;
      Result[r].Signature[Molecule.nQuantumNumbers-1] := p;
      Result[r].Expr := vnvpllE(Abs(p),vVariable(RTMode),lVariable(RTMode),Result[r].Signature[RTMode],-p);
    end;
  end;
  function xElements: TExprNode;
  var
    vn, p: Integer;
  begin
    Result := TIntegerNode.Create(1);
    for vn := 0 to Molecule.nModes-1 do begin
      p := PertEquiv.Powers[vn];
      if p > 0 then
        Result := EN([Result, '*', EB([Molecule.ModeInfo[vn].Degeneracy, '/', 2, '+', vVariable(vn), '^', p])]);
    end;
  end;
var
  vn: Integer;
begin
  SetLength(Result,0);
  case PertEquiv.Op of
    x:begin
        SetLength(Result,1); { No change in any quantum number }
        SetLength(Result[0].Signature,Molecule.nQuantumNumbers);
        Result[0].Expr := xElements;
      end;
    q, qSigned:
      begin
         RTMode := -1;
         for vn := 0 to Molecule.nModes-1 do
           AddVibChanges(vn, PertEquiv.Powers[vn]);
         if PertEquiv.Op = qSigned then
           CheckSigns
         else if RTMode >= 0 then
           DuplicateRTChanges;
      end;
    // Gaussian: ;
  else
    raise NotYet;
  end;
end;

function TVibrationalPerturbation.MatrixElement(ABra, AKet: TIntegerArray; SignA, SignB: Integer): Double;
var
  p, vn, dind: Integer;
  RelSign, DLambda: Integer;
begin
  Result := Value;
  case Op of
    x: if (ABra = AKet) and (SignA = SignB) then
         for vn := 0 to Molecule.nModes-1 do begin
           p := GetSetting(FirstDynamicSetting+vn);
           if p > 0 then
             Result := Result * IntPower(AKet[vn]+Molecule.ModeInfo[vn].Degeneracy/2,p);
           if Result = 0 then
             Exit;
         end
       else
         Result := 0;
    q, qSigned:
       begin
         { Check electronic part first }
         if Molecule.HasElectronicAngularMomentum then begin
           p := High(ABra)-1;
           if SignA*(ABra[p+1]-ABra[p]) <> SignB*(AKet[p+1]-AKet[p]) then
             { Sigma must not change }
             Result := 0;
           DLambda := (SignA*ABra[p]-SignB*AKet[p]) div 2;
           if FRTMode >= 0 then begin
             { Require Delta lambda = - Delta l = +-p }
             { First work out delta lambda }
             p := Abs(GetSetting(FirstDynamicSetting+FRTMode)); { Power of operator }
             if p <> Abs(DLambda) then
               Result := 0
             else begin
               { check delta l }
               dind := Molecule.ModeInfo[FRTMode].dIndex;
               if (SignA*ABra[dind]-SignB*AKet[dind]) <> -DLambda then
                 Result := 0;
             end;
           end else if DLambda <> 0 then
             Result := 0; { lambda must not change }
           if Result = 0 then
             Exit;
         end;
         RelSign := 0;
         for vn := 0 to Molecule.nModes-1 do begin
           p := GetSetting(FirstDynamicSetting+vn);
           if (p <> 0) and (Op = qSigned) then begin
             { Require specific change in v }
             if Abs(ABra[vn]-AKet[vn]) = Abs(p) then
               if RelSign = 0 then
                 if (ABra[vn] - AKet[vn])*p > 0 then
                   RelSign := 1
                 else
                   RelSign := -1;
             if (ABra[vn]-AKet[vn])*RelSign <> p then begin
               Result := 0;
               Exit;
             end;
             p := Abs(p);
           end;
           if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
             dind := Molecule.ModeInfo[vn].dIndex;
             if p < 0 then
               Result := Result * vnvpll(ABra[vn],SignA*ABra[dind],Abs(p),AKet[vn],SignB*AKet[dind])
             else if SignA*ABra[dind] = SignB*AKet[dind] then
               Result := Result * vnvpl(ABra[vn],p,AKet[vn],SignB*AKet[dind])
             else
               Result := 0;
           end else
             Result := Result * vnvp(ABra[vn],p,AKet[vn]);
           if Result = 0 then
             Exit;
         end;
       end;
    Gaussian:
       begin
         { Electronic part must not change }
         if Molecule.HasElectronicAngularMomentum then begin
           p := High(ABra)-1;
           if (SignA*ABra[p] <> SignB*AKet[p]) or (SignA*ABra[p+1] <> SignB*AKet[p+1]) then begin
             Result := 0;
             Exit;
           end;
         end;
         for vn := 0 to Molecule.nModes-1 do begin
           p := GetSetting(FirstDynamicSetting+vn);
           if p <> 0 then
             Result := Result * vGaussian(ABra[vn],AKet[vn],Exponent)
           else if ABra[vn] <> AKet[vn] then begin
             Result := 0;
             Exit;
           end else if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
             dind := Molecule.ModeInfo[vn].dIndex;
             if SignA*ABra[dind] <> SignB*AKet[dind] then begin
               Result := 0;
               Exit;
             end;
           end;
         end;
       end;
  end;
end;

function TVibrationalPerturbation.SymbolicMatrixElements: TMatrixElements;
var
  PertEquiv: TVibPertEquiv;
  vn: Integer;
begin
  if SymSelect >= 0 then
    raise NotYet;
  PertEquiv.Op := Op;
  SetLength(PertEquiv.Powers, Molecule.nModes);
  for vn := 0 to Molecule.nModes - 1 do
    PertEquiv.Powers[vn] := GetSetting(FirstDynamicSetting+vn);
  Result := Vib.SymbolicMatrixElements(PertEquiv, Molecule);
end;

function TVibrationalPerturbation.PairName: string;
var
  m, p: Integer;
begin
  case op of
    x: begin
         Result := 'x';
         for m := 1 to Molecule.nModes do begin
           p := GetSetting(FirstDynamicSetting+m-1);
           if p > 0 then
             if m < 10 then
               Result := Result + StringOfChar(IntToStr(m)[1],p)
             else begin
               Result := Result + '(' + IntToStr(m);
               if p > 1 then
                 Result := Result + '^' + IntToStr(p);
               Result := Result + ')';
             end;
         end;
       end;
    q: begin
         Result := '';
         for m := 1 to Molecule.nModes do begin
           p := GetSetting(FirstDynamicSetting+m-1);
           if p <> 0 then begin
             Result := Result + 'q' + IntToStr(m);
             if Abs(p) > 1 then
               Result := Result + '^' + IntToStr(Abs(p));
             if p < 0 then
               Result := Result + 'L^' + IntToStr(Abs(p));
           end;
         end;
       end;
    qSigned:
       begin
         Result := '';
         for m := 1 to Molecule.nModes do begin
           p := GetSetting(FirstDynamicSetting+m-1);
           if p <> 0 then
             Result := Result + Format('v%d(%d)',[m,p]);
         end;
       end;
    Gaussian:
       begin
         Result := '';
         for m := 1 to Molecule.nModes do begin
           p := GetSetting(FirstDynamicSetting+m-1);
           if p <> 0 then
             if Result <> '' then
               Result := '??'
             else
               Result := Result + Format('exp(-q%d^2)',[m]);
         end;
       end;
  end;
end;

{ TVibratingMolecule }

function TVibratingMolecule.Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean;
begin
  if PointGroupInfo[PointGroup].HasDegenerateModes then
    Result := Abs(AngularMomentum(Sym1)-AngularMomentum(Sym2)) <= TransitionMoment.Rank
  else
    Result := True; { TODO: fix }
end;

procedure TVibratingMolecule.FixMass;
var
  i: Integer;
begin
  FMass := 0;
  for i := 0 to nNuclei - 1 do
    FMass := FMass + FNuclearMasses[i];
end;

function TVibratingMolecule.FixupProperties: Exception;
var
  vn, i, j: Integer;
begin
  nNuclei := nNuclei; { Trigger add/remove of nuclei as required }
  FnQuantumNumbers := nModes;
  i := nModes;
  for vn := 0 to nModes - 1 do begin
    ModeInfo[vn].dIndex := -1;
    ModeInfo[vn].Degeneracy := 1;
    ModeInfo[vn].ShowQno := ShowAllQnos;
    if IsDegenerate(ModeInfo[vn].Symmetry) then begin
      ModeInfo[vn].dIndex := i;
      ModeInfo[vn].Degeneracy := 2;
      Inc(FnQuantumNumbers);
      Inc(i);
    end;
  end;
  FHasElectronicAngularMomentum := False;
  for i := 0 to nManifolds-1 do
    for j := 0 to Manifold[i].nStates - 1 do begin
      FHasElectronicAngularMomentum := FHasElectronicAngularMomentum
                                       or (AngularMomentum((Manifold[i].State[j] as TElectronicState).Symmetry) > 0)
                                       or (Manifold[i].State[j].S > 0);
      for vn := 0 to nModes - 1 do
        with (Manifold[i].State[j] as TElectronicState).Mode(vn) do
          ModeInfo[vn].ShowQno := ModeInfo[vn].ShowQno or (vMin <> 0) or (vMax <> 0);
    end;
  if HasElectronicAngularMomentum then
    Inc(FnQuantumNumbers,2);
  Result := inherited FixupProperties;
  for i := 0 to ComponentCount - 1 do
    if Components[i] is TTransitionMoments then
      TTransitionMoments(Components[i]).DisableRankCheck := True;
end;

function TVibratingMolecule.GetPointGroup(const Index: Integer): TPointGroup;
begin
  Result := TPointGroup(GetSetting(Index));
end;

const
  VibratingMoleculeSettingNames: array[LMI..LMI+2] of string = (
'Point group of molecule.',
'Number of vibrational modes to include in the calculation.',
'Set false to hide vibrational quantum numbers with vmin=vmax=0'
);

function TVibratingMolecule.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(VibratingMoleculeSettingNames))
     and (AIndex <= High(VibratingMoleculeSettingNames)) then
    Result := VibratingMoleculeSettingNames[AIndex]
  else if AIndex = IndexOfSetting('nNuclei') then
    Result := 'Number of atoms; Only required for l matrix calculations'
  else if AIndex = IndexOfSetting('Jmin') then
    Result := 'Not used'
  else if AIndex = IndexOfSetting('Jmax') then
    Result := 'Maximum vibrational angular momentum'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TVibratingMolecule.AngularMomentum(Symmetry: Integer): TQuantumNumber;
begin
  case PointGroup of
    CInfinityv: Result := Max(Symmetry-1,0)*2;
    DInfinityh: Result := Max(Symmetry div 2 - 1,0)*2;
  else
    Result := 0;
  end;
end;

constructor TVibratingMolecule.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',ShowAllQnos=True';
  inherited;
end;

function TVibratingMolecule.IsDegenerate(Symmetry: Integer): Boolean;
begin
  Result := AngularMomentum(Symmetry) > 0;
end;

function TVibratingMolecule.IsValidChildClass(MNodeClass: TMNodeClass): Boolean;
begin
  Result := inherited IsValidChildClass(MNodeClass);
  if Result and MNodeClass.InheritsFrom(TTransitionMoment) then
    Result := MNodeClass = TMultidimensionalFCF;
end;

function TVibratingMolecule.JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer;
begin
  Result := Sym;
end;

function TVibratingMolecule.LastSym: Integer;
begin
  Result := PointGroupInfo[PointGroup].LastSymmetry;
  if Result = High(Result) then
    Result := JMax
  else
    Exit; { Not a linear point group }
  if Result < 0 then
    Result := Species.Jmax;
  if Result < 0 then
    Result := 20;
  Result := Result div 2;
end;

function TVibratingMolecule.QnoInfo(QnoType: TQnoType): TQnoInfo;
var
  i: TQnoType;
  vn: Integer;
begin
  { qtA is v1;
    order is v1, (l1), v2, (l2), .... Lambda, Omega }
  if QnoType < qtA then
    Result := inherited QnoInfo(QnoType)
  else begin
    { Find relevant quantum numbers }
    i := qtA;
    if Ord(i)-Ord(qtA) < nModes then
      with Result do begin
        MinValue := 0;
        MaxValue := High(Integer);
        StepSize := 1;
        Name := ModeInfo[Ord(i)-Ord(qtA)].Name;
        Exit;
      end;
    Inc(i,nModes);
    for vn := 0 to nModes - 1 do begin
      if IsDegenerate(ModeInfo[vn].Symmetry) then begin
        if i = QnoType then
          with Result do begin
            MinValue := -High(Integer);
            MaxValue := High(Integer);
            StepSize := 1;
            Name := 'l'+Copy(ModeInfo[vn].Name,2,255);
            Exit;
          end;
        Inc(i);
      end;
    end;
    if i = QnoType then
      with Result do begin
        MinValue := -High(Integer);
        MaxValue := High(Integer);
        StepSize := 1;
        Name := 'Lambda';
        Exit;
      end;
    Inc(i);
    if i = QnoType then
      with Result do begin
        MinValue := -High(Integer);
        MaxValue := High(Integer);
        StepSize := 1;
        Name := 'Omega';
        Exit;
      end;
    Inc(i);
    if i = QnoType then
      with Result do begin
        MinValue := -1;
        MaxValue := 1;
        StepSize := 0;
        Name := 'Phase';
        Exit;
      end;
    Result := inherited QnoInfo(QnoType); { Not found }
  end;
end;

procedure TVibratingMolecule.SetnModes(const Index, Value: Integer);
var
  i, j: Integer;
begin
  inherited SetSetting(Index, Value);
  j := Length(ModeInfo);
  SetLength(ModeInfo,Value);
  for i := j to Value - 1 do
    ModeInfo[i].Name := Format('v%d', [i+1]); { Fill in default name }
  { Make sure each state has the right number of modes }
  for i := 0 to nManifolds - 1 do
    with Manifold[i] do begin
      for j := 0 to nStates - 1 do
        (State[j] as TElectronicState).AdjustModes(True);
      for j := 0 to nPerturbations - 1 do
        (Perturbation[j] as TVibrationalPerturbation).AdjustSettingNames;
    end;
  for i := 0 to ComponentCount-1 do
    if Items[i] is TTransitionMoments then
      with TTransitionMoments(Items[i]) do
        for j := 0 to ComponentCount - 1 do
          (TransitionMoment[j] as TMultidimensionalFCF).AdjustModes(True);
end;

procedure TVibratingMolecule.SetnNuclei(const Index, Value: Integer);
begin
  SetLength(FNuclearMasses, Value);
  inherited;
  nModes := nModes; { Trigger update of Mode information }
  FixMass;
end;

procedure TVibratingMolecule.SetPointGroup(const Index: Integer; const AValue: TPointGroup);
begin
  SetSetting(Index, Ord(AValue));
end;

function TVibratingMolecule.SymName(Sym: Integer): string;
begin
  if Sym < 0 then
    Result := '-'
  else
    Result := PointGroupInfo[PointGroup].SymmetryToStr(Sym);
end;

{ TVibrationalManifold }

function TVibrationalManifold.FindJmax(Default: Integer; RemovedSpins: Byte): Integer;
begin
  if PointGroupInfo[(Molecule as TVibratingMolecule).PointGroup].HasDegenerateModes then
    Result := inherited FindJmax(Default, RemovedSpins)
  else
    Result := HalfSpin(RemovedSpins);
end;

function TVibrationalManifold.nSym: Integer;
begin
  Result := Molecule.LastSym+1;
end;

{ TMolecularNucleus }

procedure TMolecularNucleus.AddHandlers;
begin
  inherited;
  AddHandler('Set Exact Mass', 'Try to set exact mass from mass and name',{$IFDEF FPC}@{$ENDIF}DoFindExactMass);
end;

constructor TMolecularNucleus.Create(AOwner: TComponent);
begin
  inherited;
  { Mass should be a per molecule setting, but is actually
    in the per state object. Copy from the per molecule setting on
    initialisation }
  ModifyParameter(1, (AOwner.Owner.Owner as TVibratingMolecule).FNuclearMasses[NodeIndex], False);
end;

procedure TMolecularNucleus.DoFindExactMass(Sender: TObject);
var
  EMass: Double;
begin
  if Copy(Name,1,7) = 'Nucleus' then
    EMass := FindExactMass(Mass)
  else
    EMass := FindExactMass(Mass,Name);
  if Emass > 0 then
    Mass := EMass
  else
    raise EReportException.Create('Mass not found');
end;

procedure TMolecularNucleus.GlobalShift(Shift: Integer);
var
  Target: TMolecularNucleus;
begin
  if (NodeIndex+Shift < 0) or (NodeIndex+Shift >= Molecule.nNuclei) then
    Exit;
  Target := State.Nucleus(NodeIndex+Shift) as TMolecularNucleus;
  inherited;
  { Reset mass, to ensure copied to global array }
  SetParameter(1,GetParameter(1));
  Target.SetParameter(1,Target.GetParameter(1));
end;

const
 MolecularNucleusParameterNames: array [1..4] of string = (
'Mass of nucleus/amu.',
'Equilibrium x co-ordinate / Angstrom.',
'Equilibrium y co-ordinate / Angstrom.',
'Equilibrium z co-ordinate / Angstrom.'
);

function TMolecularNucleus.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(MolecularNucleusParameterNames))
     and (AIndex <= High(MolecularNucleusParameterNames)) then
    Result := MolecularNucleusParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

procedure TMolecularNucleus.SetComponentIndex(const Value: Integer);
var
  Displacements: array of array of Double;
  Original: array of TNucleus;
  OldNuc, NewNuc, i, j, nModes: Integer;
  EState: TElectronicState;
begin
  { If nuclei numbering changed, move nuclear displacements also }
  { Save original }
  EState := State as TElectronicState;
  SetLength(Original,Molecule.nNuclei);
  nModes := (Molecule as TVibratingMolecule).nModes;
  for i := 0 to Molecule.nNuclei-1 do
    Original[i] := State.Nucleus(i);
  SetLength(Displacements,nModes,3*Molecule.nNuclei);
  for i := 0 to nModes-1 do
    with EState.Mode(i) do
      for j := 0 to 3*Molecule.nNuclei-1 do
        Displacements[i,j] := GetParameter(FirstDynamicParameter+j);
  inherited;
  { Look for changed nodes }
  for OldNuc := 0 to Molecule.nNuclei-1 do
    if Original[OldNuc].NodeIndex <> OldNuc then begin
      NewNuc := Original[OldNuc].NodeIndex;
      for i := 0 to nModes-1 do
        with EState.Mode(i) do
          for j := 0 to 2 do
            SetParameter(FirstDynamicParameter+j+NewNuc*3,Displacements[i,j+OldNuc*3]);
    end;
end;

procedure TMolecularNucleus.SetParameter(AIndex: Integer; const Value: Double);
var
  i, j, MyNodeIndex: Integer;
  Changed: Boolean;
  NewValue: Double;
begin
  inherited SetParameter(AIndex, Value);
  { See notes in Create method; set per molecule settting also }
  if NodeIndex < 0 then
    Exit;
  Changed := False;
  NewValue := GetParameter(AIndex); { Handle reset to default case }
  case AIndex of
    1: begin
         Changed := (Molecule as TVibratingMolecule).FNuclearMasses[NodeIndex] <> NewValue;
         (Molecule as TVibratingMolecule).FNuclearMasses[NodeIndex] := NewValue;
       end;
  end;
  if not Changed then
    Exit;
  { As GetParameter is not virtual, changes must be copied to all other nueclei }
  (Molecule as TVibratingMolecule).FixMass;
  MyNodeIndex := NodeIndex;
  for i := 0 to Molecule.nManifolds - 1 do
    with Molecule.Manifold[i] do
      for j := 0 to nStates - 1 do
        State[j].Nucleus(MyNodeIndex).ModifyParameter(AIndex, NewValue, True);
end;

{ TElectronicState }

procedure TElectronicState.AddHandlers;
begin
  inherited;
  if Molecule.nNuclei > 0 then begin
    AddHandler('l Matrix...', 'Enter/Edit l matrix', {$IFDEF FPC}@{$ENDIF}SetLmatrix);
    AddHandler('Print Geometry', 'Print geometrical information', {$IFDEF FPC}@{$ENDIF}PrintMolecularInfo);
    AddHandler('Fix Centre of Mass', 'Adjust atomic co-ordinates so centre of mass is at the origin', {$IFDEF FPC}@{$ENDIF}FixCentreOfMass);
    AddHandler('Normalize', 'Normalize vibrational co-ordinates', {$IFDEF FPC}@{$ENDIF}Normalize);
    AddHandler('Orthogonalize', 'Adjust vibrational co-ordinates to be orthogonal and correctly normalized', {$IFDEF FPC}@{$ENDIF}Orthogonalize);
  end;
end;

function TElectronicState.AddMode:TVibrationalMode;
begin
  Result := TVibrationalMode.Create(Self);
end;

function TElectronicState.AddOneJ(Eigenvalues: TEigenvalues; J: TQuantumNumber; Sym: Integer; HyperfineState: THyperfineBasis; Seffective: TQuantumNumber; out StatWtGroup: Integer): Integer;
var
  v: TIntegerArray;
  Position, Phase: Integer;
  EQMin, EQMax: array of Integer; { Range of effective quantum numbers }
  procedure AddState;
  var
    BasisState: TVibrationalBasisState;
    FirstSym, AltSym, i, q: Integer;
  begin
    { Determine symmetry }
    Phase := 0;
    FirstSym := SymmetryOf(v, AltSym);
    if AltSym >=0 then
      Phase := 1;
    if FirstSym <> Sym then
      if AltSym = Sym then
        Phase := -1
      else
        Exit;
    BasisState := TVibrationalBasisState.Create(Eigenvalues,0,Sym);
    BasisState.State := Self;
    BasisState.HyperfineState := HyperfineState;
    BasisState.StateIndex := Eigenvalues.LastStateIndex;
    BasisState.v := Copy(v,0,Length(v));
    BasisState.FOriginalPosition := Position;
    BasisState.Phase := Phase;
    for i := 0 to High(FEffectiveQuanta) do begin
      q := FEffectiveQuanta[i].EffectiveQuantumNumber(v);
      EQMin[i] := Min(EQMin[i],q);
      EQMax[i] := Max(EQMax[i],q);
    end;
  end;
var
  i, vn, StartIndex, nQ, dind: Integer;
  OK: Boolean;
begin
  if Molecule.IsDegenerate(Sym) then
    Result := 2
  else
    Result := 1;
  StatWtGroup := 1; { Statistical weights not significant for vibration }
  if J <> 0 then
    Exit;
  StartIndex := Eigenvalues.Size;
  { Cycle through all available vibrational quantum numbers }
  SetLength(v, Molecule.nQuantumNumbers);
  Position := 1;
  for i := 0 to Molecule.nQuantumNumbers - 1 do
    v[i] := 0;
  SetLength(EQMin,Length(FEffectiveQuanta));
  for i := 0 to Length(FEffectiveQuanta) - 1 do
    EQMin[i] := High(Integer);
  SetLength(EQMax,Length(FEffectiveQuanta));
  repeat { For every mode combination }
    OK := True;
    for vn := 0 to Molecule.nModes - 1 do
      if v[vn] < Mode(vn).vmin then begin
        OK := False;
        Break;
      end;
    if OK then
      { Cycle though electronic angular momentum }
      if Molecule.HasElectronicAngularMomentum then begin
        dind := High(v)-1;
        v[dind] := Molecule.AngularMomentum(Symmetry);
        { Try both + and - values of electronic orbital angular momentum }
        repeat
          for i := 0 to S do begin
            v[dind+1] := v[dind] + 2*i - S; { Omega }
            AddState;
          end;
          v[dind] := -v[dind];
        until v[dind] >= 0;
      end else
        AddState;
    { Cycle though vibrational angular momentum; must do it here
      so that Postion variable is correct }
    vn := Molecule.nModes - 1;
    while vn >= 0 do begin
      if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
        { Bump up angular momentum }
        dind := Molecule.ModeInfo[vn].dIndex;
        Inc(v[dind],2);
        if v[dind] > v[vn] then begin
          { run out of l values, so reset and move to next }
          v[dind] := -v[vn];
        end else
          Break;
      end;
      Dec(vn);
    end;
    if vn >= 0 then
      Continue; { New combination of angular momenta; Add }
    Inc(Position);
    { Now find next set of v, cycling highest numbered modes fastest }
    vn := Molecule.nModes - 1;
    while vn >= 0 do begin
      Inc(v[vn]);
      if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
        dind := Molecule.ModeInfo[vn].dIndex;
        v[dind] := -v[vn]; { Reset angular momentum to minimum }
      end;
      if v[vn] > Mode(vn).vMax then begin
        v[vn] := 0; { Overflow to next }
        if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then
          v[dind] := 0; { Reset angular momentum to minimum }
      end else
        Break; { Combination OK }
      Dec(vn);
    end;
  until vn < 0;
  { Put effective quantum numbers in state index }
  nQ := 1;
  EigenValues.Fixup;
  if StartIndex+1 <= Eigenvalues.Size then
    for i := 0 to High(FEffectiveQuanta) do begin
      for j := StartIndex+1 to Eigenvalues.Size do
        with EigenValues.Basis[j] as TVibrationalBasisState do
          Inc(StateIndex,(FEffectiveQuanta[i].EffectiveQuantumNumber(v)-EQMin[i])*nQ);
      nQ := nQ * (EQMax[i]-EQMin[i]+1);
    end;
  EigenValues.SortFrom(StartIndex,Emax);
  Inc(Eigenvalues.LastStateIndex,nQ);
end;

procedure TElectronicState.AdjustModes(FixNumber: Boolean);
var
  j, n: Integer;
begin
  if FixNumber then
    n := AdjustNodeIndexes(TVibrationalMode, FModes, nil, opInsert, Molecule.nModes)
  else
    n := AdjustNodeIndexes(TVibrationalMode, FModes);
  for j := 0 to n - 1 do
    Mode(j).AdjustParameterNames;
end;

function TElectronicState.ConvertParameter(AIndex: Integer; Zero: Boolean): TPerturbation;
var
  PertEquiv: TVibPertEquiv;
  vn, i: Integer;
begin
  Result := nil; { Leave by default }
  if AIndex > nParameters then begin
    { Parameter is mode specific }
    Dec(AIndex, nParameters+1);
    vn := AIndex div Mode(0).nParameters ;
    AIndex := AIndex mod Mode(vn).nParameters + 1;
    if AIndex = 1 then
      Exit; { Omega values must not be converted, as they determine the basis }
    PertEquiv := Mode(vn).PertEquiv(AIndex);
    if Length(PertEquiv.Powers) = 0 then
      Exit; { Not implemented - leave }
    Result := TVibrationalPerturbation.Create(Owner);
    Result.SetBraKetName(Name,Name);
    with Result as TVibrationalPerturbation do begin
      Op := PertEquiv.Op;
      for i := 0 to Molecule.nModes - 1 do
        SetSetting(FirstDynamicSetting+i,PertEquiv.Powers[i]);
      Value := Mode(vn).Parameters[AIndex];
    end;
    if Zero then
      Mode(vn).Parameters[AIndex] := 0;
  end;
  { TODO: Spin parameters as perturbations }
end;

constructor TElectronicState.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Emax=1e99';
  if Assigned(AOwner) then
    FMolecule := AOwner.Owner as TVibratingMolecule;
  inherited;
end;

destructor TElectronicState.Destroy;
begin
  FreeAndNil(adm);
  FreeAndNil(Madm);
  FreeAndNil(Itrans);
  inherited;
end;

function TElectronicState.FixupProperties: Exception;
var
  nPQ, i: Integer;
begin
  AdjustNuclei(True); { Explicit call here, so that nuclei typically appear before vibrations, though this ordere is not essential }
  AdjustModes(True);
  Result := inherited FixupProperties;
  nPQ := 0;
  for i := 0 to Manifold.Count - 1 do
    if Manifold.Components[i] is TVibrationalPerturbation then
      with TVibrationalPerturbation(Manifold.Components[i]) do
        if (Bra = Self) and (Ket = Self) and EffectiveQno then begin
          if Length(FEffectiveQuanta) <= nPQ then
            SetLength(FEffectiveQuanta,nPQ+1);
          FEffectiveQuanta[nPQ] := TVibrationalPerturbation(Manifold.Components[i]);
          Inc(nPQ);
        end;
  SetLength(FEffectiveQuanta,nPQ);
end;

function TElectronicState.MatrixElementType(const Changes: array of Integer; const H: string): string;
  procedure AddVariables(Bra: Boolean);
    procedure AddVariable(const Variable: string; Change: Integer);
    begin
      Result := Result + Variable;
      if not Bra then
        Exit;
      if Change > 0 then
        Result := Result + '+';
      if Change <> 0 then
        Result := Result + IntToStr(Change);
    end;
    procedure AddNumberedVariable(const Variable: string; AIndex, Change: Integer);
    begin
      AddVariable(Variable+IntToStr(AIndex+1),Change);
    end;
  var
    vn, dind: Integer;
  begin
    for vn := 0 to Molecule.nModes - 1 do begin
      if vn > 0 then
        Result := Result + ',';
      AddNumberedVariable('v',vn,Changes[vn]);
      if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
        dind := Molecule.ModeInfo[vn].dIndex;
        Result := Result + '(';
        AddNumberedVariable('l',vn,Changes[dind]);
        Result := Result + ')';
      end;
    end;
    if Molecule.HasElectronicAngularMomentum then begin
      if Molecule.nModes > 0 then
        Result := Result + ',';
      AddVariable('Lambda',Changes[Molecule.nQuantumNumbers-2]);
      Result := Result + ',';
      AddVariable('Omega',Changes[Molecule.nQuantumNumbers-1]);
    end;
  end;
begin
  Result := '<';
  AddVariables(True);
  Result := Result + '|' + H + '|';
  AddVariables(False);
  Result := Result + '>';
end;

function TElectronicState.Mode(Index: Integer): TVibrationalMode;
begin
  if (Index >= Length(FModes)) and (Molecule.nModes <> Length(FModes)) then
    AdjustModes(True);
  Result := FModes[Index] as TVibrationalMode;
end;

function TElectronicState.nParameterVariables: Integer;
begin
  Result := nParameters;
  if Molecule.nModes > 0 then
    Result := Result + Molecule.nModes*Mode(0).nParameters;
end;

function TElectronicState.ParameterVariable(AIndex: Integer): TSortedVariableNode;
var
  vn, i: Integer;
begin
  if AIndex <= nParameters then
    Result := inherited ParameterVariable(AIndex)
  else begin
    if not Assigned(VariableList) then
      inherited ParameterVariable(1); { Force creation of list }
    if VariableList.Count < nParameterVariables then
      for vn := 0 to Molecule.nModes - 1 do
        with Mode(vn) do
          for i := 1 to nParameters do
            VariableList.Add(TSortedVariable.Create(ParameterNames[i]+IntToStr(vn+1),VariableList.Count+1));
    Dec(AIndex,nParameters+1);
    vn := AIndex div Mode(0).nParameters;
    AIndex := AIndex mod Mode(vn).nParameters + 1;
    i := VariableList.IndexOf(Mode(vn).ParameterNames[AIndex]+IntToStr(vn+1));
    Result := (VariableList.Objects[i] as TSortedVariable).Template;
    Result.Value := Mode(vn).Parameters[AIndex];
  end;
end;

procedure TElectronicState.FixCentreOfMass(Sender: TObject);
begin
  CentreOfMass(True, True);
end;

procedure TElectronicState.CentreOfMass(Fix, Print: Boolean);
var
  i, j, p: Integer;
  Sum: Double;
  v: array [0..2] of Double;
  N: TMolecularNucleus;
begin
  Molecule.CheckProperties;
  Sum := 0;
  v[0] := 0; v[1] := 0; v[2] := 0;
  for i := 0 to Molecule.nNuclei - 1 do begin
    N := Nucleus(i) as TMolecularNucleus;
    Sum := Sum + N.Mass;
    for j := 0 to 2 do
      v[j] := v[j] + N.GetParameter(N.IndexOfParameter('x')+j)*N.Mass;
  end;
  for i := 0 to 2 do
    v[i] := v[i]/Sum;
  if Print then begin
    ReShowLogForm;
    WriteLn(Log);
    WriteLn(Log, 'Geometrical Information for ', UniqueName);
    if Fix then
      Write(Log,'Original ');
    WriteLn(Log, 'Centre of Mass: x=', v[0], ', y=', v[1], ', z=', v[2]);
  end;
  if Fix then
    for i := 0 to Molecule.nNuclei - 1 do begin
      N := Nucleus(i) as TMolecularNucleus;
      for j := 0 to 2 do begin
        p := N.IndexOfParameter('x')+j;
        N.SetParameter(p, N.GetParameter(p) - v[j]);
      end;
    end;
end;

procedure TElectronicState.PrintMolecularInfo(Sender: TObject);
begin
  CentreOfMass(False,True);
  PrepareMatrices(True);
end;

procedure TElectronicState.PrepareMomentsOfInteria(Print: Boolean);
var
  I: TMatrix;
  RotConsts: TColumnVector;
  p, q, n, offset: Integer;
  Nuc: TMolecularNucleus;
begin
  if Molecule.nNuclei = 0 then begin
    FreeAndNil(Itrans);
    Exit;
  end;
  I := TMatrix.Create(3,3);
  I.Zero;
  for n := 0 to Molecule.nNuclei - 1 do begin
    Nuc := Nucleus(n) as TMolecularNucleus;
    offset := Nuc.IndexOfParameter('x')-1;
    for p := 1 to 3 do begin
      for q := 1 to p-1 do begin
        I[p,q] := I[p,q] - Nuc.Mass * Nuc.Parameters[offset+p] * Nuc.Parameters[offset+q];
        I[q,p] := I[p,q];
      end;
      for q := 1 to 3 do
        if p <> q then
          I[p,p] := I[p,p] + Nuc.Mass * Sqr(Nuc.Parameters[offset+q]);
    end;
  end;
  for p := 1 to 3 do
    Ixx[p-1] := I[p,p];
  if Print then
    I.Write(Log, 'Inertial Tensor');
  RotConsts := TColumnVector.Create(3);
  Eigen(RotConsts,I,True);
  if Print then
    RotConsts.Write(Log,'Principal Moments of Intertia');
  for p := 1 to 3 do
    if RotConsts[p] <> 0 then
      RotConsts[p] := MomentOfInertiaTimesRotationalConstant[cm1]/RotConsts[p];
  FreeAndNil(Itrans);
//  Itrans := I.Transpose;
  Itrans := TMatrix.Create(I);
  if Print then begin
    RotConsts.Write(Log,'Rotational Constants / cm-1');
    I.Write(Log,'Transformation to principal axis system')
  end;
  I.Free;
  RotConsts.Free;
end;

procedure TElectronicState.Normalize(Sender: TObject);
var
  i, k, offset: Integer;
  T: TMatrix;
  Sc: Double;
begin
  PrepareMatrices(False,False);
  if not Assigned(adm) then
    Exit;
  T := Madm.Transpose;
  T.PostMultiply(adm);
  offset := Mode(0).FirstDynamicParameter;
  for i := 0 to Molecule.nModes - 1 do begin { For each vibration }
    Sc := T[i+1,i+1];
    for k := 0 to adm.RowCount-1 do
      Mode(i).Parameters[offset+k] := Mode(i).Parameters[offset+k] / Sqrt(Sc);
  end;
  T.Free;
end;

procedure TElectronicState.Orthogonalize(Sender: TObject);
var
  i, j, k, offset: Integer;
  T: TMatrix;
  Sc: Double;
  procedure UpdateT(All:Boolean);
  begin
    PrepareMatrices(False,All);
    T := Madm.Transpose;
    T.PostMultiply(adm);
  end;
begin
  CentreOfMass(True, True);
  T := nil;
  UpdateT(True);
  if not Assigned(adm) then
    Exit;
  offset := Mode(0).FirstDynamicParameter;
  { Remove components along T, R }
  for i := 0 to Molecule.nModes - 1 do { For each vibration }
    for j := Molecule.nModes to adm.ColCount-1 do begin
      { For each T or R motion }
      Sc := T[i+1,j+1];
      for k := 0 to adm.RowCount-1 do
        Mode(i).Parameters[offset+k] := Mode(i).Parameters[offset+k] - adm[k+1,j+1] * Sc;
    end;
  { Now remove components along other vibrations }
  for i := 0 to Molecule.nModes - 1 do begin { For each vibration }
    for j := 0 to i - 1 do begin
      UpdateT(False);
      Sc := T[i+1,j+1];
      for k := 0 to adm.RowCount-1 do
        Mode(i).Parameters[offset+k] := Mode(i).Parameters[offset+k] - adm[k+1,j+1] * Sc;
    end;
    UpdateT(False);
    Sc := T[i+1,i+1];
    for k := 0 to adm.RowCount-1 do
      Mode(i).Parameters[offset+k] := Mode(i).Parameters[offset+k] / Sqrt(Sc);
  end;
  T.Free;
  PrepareMatrices(True,True);
end;

procedure TElectronicState.PrepareMatrices(Print: Boolean; AllModes:Boolean);
var
  i, j, k, k1, k2, b, bx: Integer;
  N: TMolecularNucleus;
  T: TMatrix;
begin
  PrepareMomentsOfInteria(Print);
  if Molecule.nNuclei = 0 then begin
    FreeAndNil(adm);
    FreeAndNil(Madm);
    Exit;
  end;
  if not Assigned(Madm) then begin
    adm := TMatrix.Create(0,0);
    Madm := TMatrix.Create(0,0);
  end;
  adm.RowCount := 3*Molecule.nNuclei;
  if AllModes then
    adm.ColCount := adm.RowCount
  else
    adm.ColCount := Molecule.nModes;
  adm.Zero;
  Madm.RowCount := 3*Molecule.nNuclei;
  Madm.ColCount := adm.ColCount;
  for i := 0 to Molecule.nNuclei - 1 do begin
    N := Nucleus(i) as TMolecularNucleus;
    for j := 0 to 2 do
      for k := 0 to Molecule.nModes - 1 do
        with Mode(k) do
          adm[3*i+j+1,k+1] := GetParameter(FirstDynamicParameter+3*i+j);
    if AllModes then begin
      b := Molecule.nModes+1;
      bx := N.IndexOfParameter('x');
      { Add rotations }
      { See Hirota 2.2.9, though note adm matrix ls mass weighted: el/sqrt(m) }
      for k := 0 to 2 do
        { Check for zero moments of Inertia }
        if Abs(Ixx[k]) > 1e-10 then begin
          k1 := (k+1) mod 3;
          k2 := (k1+1) mod 3;
          adm[3*i+k2+1,b] := N.GetParameter(bx+k1) / Sqrt(Ixx[k]);
          adm[3*i+k1+1,b] := -N.GetParameter(bx+k2) / Sqrt(Ixx[k]);
          Inc(b);
        end;
      { Add translations }
      b := 3*Molecule.nNuclei-3; { This may overwrite one of the rotations }
      if Molecule.Mass > 0 then { Avoid errors during data entry }
        for k := 0 to 2 do
          adm[3*i+k+1,b+k+1] := 1/Sqrt(Molecule.Mass);
    end;
    for j := 0 to 2 do
      for k := 0 to adm.ColCount-1 do
        Madm[3*i+j+1,k+1] := N.Mass * adm[3*i+j+1,k+1];
  end;
  if Print then begin
    WriteLn(Log,'Matrices for ',UniqueName);
    adm.Write(Log,'adm');
    Madm.Write(Log,'Madm');
    T := Madm.Transpose;
    T.PostMultiply(adm);
    T.Write(Log,'Orthogonality check: Madm^T adm');
    T.Free;
  end;
end;

function TElectronicState.ReadBranchFormat(Text: string; var Obs: TObservation;
  UpperState: TState; Reader: TObservationsReaderBase): Boolean;
  function ReadQnos(var vs: TIntegerArray; State: TElectronicState): Boolean;
  var
    vn, p, dind: Integer;
    function Absorb(c: Char):Boolean;
    begin
      Result := (p <= Length(Text)) and (Text[p] = c);
      if Result then
        Inc(p);
    end;
    function NextNumber(var Value: Integer):Boolean;
    var
      e: Integer;
    begin
      e := p;
      Absorb('-');
      while (p <= Length(Text)) and (Text[p] in ['0'..'9']) do
        Inc(p);
      Result := (e = p) or Reader.ReadQuantumNumber(Value,Copy(Text,e,p-e));
    end;
  begin
    Result := False;
    p := 1;
    if not Absorb('(') then
      Exit;
    SetLength(vs,Molecule.nQuantumNumbers+1);
    for vn := 0 to Molecule.nModes - 1 do begin
      if NextNumber(vs[vn]) then
        Exit;
      vs[vn] := vs[vn] div 2;
      if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
        dind := Molecule.ModeInfo[vn].dIndex;
        if Absorb('(') then begin
          { Read optional l value }
          if NextNumber(vs[dind]) then
            Exit;
          vs[dind] := vs[dind] div 2;
          Absorb(')');
        end else
          vs[dind] := vs[vn] and 1; { Default to lowest possible value }
      end;
      Absorb(',');
    end;
    if Molecule.HasElectronicAngularMomentum then begin
      dind := Length(vs)-3; { Extra qno here (not present in normal basis) is phase }
      vs[dind] := 0; { Default to no electronic angular momentum }
      vs[dind+1] := 0;
      if Absorb('[') then begin
        if Molecule.IsDegenerate(State.Symmetry) then begin
          if Absorb('+') then
            vs[dind] := Molecule.AngularMomentum(State.Symmetry)
          else if Absorb('-') then
            vs[dind] := -Molecule.AngularMomentum(State.Symmetry)
          else
            Exit;
          Absorb(',');
        end;
        if State.S > 0 then begin
          if NextNumber(vs[dind+1]) then
            Exit;
        end else
          vs[dind+1] := vs[dind]; { Omega = Lambda }
        Absorb(']')
      end;
    end;
    dind := High(vs);
    if Absorb('+') then
      vs[dind] := 1
    else if Absorb('-') then
      vs[dind] := -1
    else
      vs[dind] := Low(Integer);
    if Absorb(')') then begin
      Delete(Text,1,p-1);
      Result := True;
    end;
  end;
  function FindState(AManifold:TManifold; var Index: Smallint; var Sym: Shortint; v: array of Integer; State: TElectronicState):string;
  var
    J: Integer;
    AltSym: Integer;
  begin
    Sym := State.SymmetryOf(v,AltSym);
    Result := AManifold.FindStateLabel(Index, Sym, J, { Not used }
                                       v, Obs.QuantumNumberFormat, Obs.ExField,
                                       Sym, 0, -1, State, Reader.RemoveSpins);
    if (Sym < 0) and (AltSym >= 0) then begin
      Sym := AltSym;
      Result := AManifold.FindStateLabel(Index, Sym, J, { Not used }
                                         v, Obs.QuantumNumberFormat, Obs.ExField,
                                         Sym, 0, -1, State, Reader.RemoveSpins);
    end;
  end;
var
  Up, Lo: TIntegerArray;
  Lname: string;
begin
  Result := False;
  Obs.JL := 0;
  Obs.JU := 0;
  if (Length(Text) > 2) and (Text[1] = '-') and (Text[2] = '(') then begin
    { Lower state fit }
    Delete(Text,1,1);
    if not (ReadQnos(Lo, Self) and (Text = '')) then
      Exit;
    Obs.ObsType := obELower;
    Lname := FindState(Manifold, Obs.IndexL, Obs.SymL, Lo, Self);
    if Obs.SymL >= 0 then
      Obs.Name := Obs.Name + ' : . - ' + Lname
    else
      obs.Name := Obs.Name + LName;
    Obs.SymU := 0;
  end else if not ReadQnos(Up, UpperState as TElectronicState) then
    Exit
  else if (Text = '') or (Text = '-') then begin
    { Upper state fit }
    Obs.ObsType := obEUpper;
    Lname := FindState(UpperState.Manifold,Obs.IndexU, Obs.SymU,Up,UpperState as TElectronicState);
    if Obs.SymU >= 0 then
      Obs.Name := Obs.Name + ' : ' + Lname + ' - .'
    else
      obs.Name := Obs.Name + LName;
    Obs.SymL := 0;
  end else begin
    { Normal Transition fit }
    if Text[1] <> '-' then
      Exit;
    Delete(Text,1,1);
    if ReadQnos(Lo, Self) and (Text = '') then begin
      Lname := FindState(Manifold, Obs.IndexL, Obs.SymL, Lo, Self);
      if Obs.SymL >= 0 then begin
        Obs.Name := Obs.Name + ' : '
                    + FindState(UpperState.Manifold,Obs.IndexU, Obs.SymU,Up,UpperState as TElectronicState)
                     + ' - ' + Lname;
      end else
        Obs.Name := Obs.Name + LName;
    end;
  end;
  Result := (Obs.SymL >=0) and (Obs.SymU >= 0);
  if Result then begin
    Obs.IndexU := 0; { Force search }
    Obs.IndexL := 0;
  end;
end;

function TElectronicState.SymbolicRotationalMatrixElement(AIndex: Integer): TMatrixElements;
var
  Sigma, Svar: TExprNode;
  dLam, vn: Integer;
  PertEquiv: TVibPertEquiv;
begin
  if AIndex > nParameters then begin
    { Parameter is mode specific }
    Dec(AIndex, nParameters+1);
    vn := AIndex div Mode(0).nParameters;
    AIndex := AIndex mod Mode(vn).nParameters + 1;
    if AIndex >= Mode(vn).FirstDynamicParameter then
      SetLength(Result,0)
    else if AIndex = 5 then begin { Special case gll, as not implemented as perturbation (yet) }
      SetLength(Result,1);
      SetLength(Result[0].Signature,Molecule.nQuantumNumbers);
      Result[0].Expr := EB([lVariable(vn), '^', 2]);
    end else begin
      PertEquiv := Mode(vn).PertEquiv(AIndex);
      if Length(PertEquiv.Powers) = 0 then
        raise NotYet;
      Result := Vib.SymbolicMatrixElements(PertEquiv, Molecule);
    end;
  end else if Molecule.IsDegenerate(Symmetry) or (S > 0) or (AIndex = 1)then begin
    SetLength(Result,1);
    SetLength(Result[0].Signature,Molecule.nQuantumNumbers);
    Sigma := EB([eOmega, '-', eLambda]);
    case AIndex of
      1: { Origin } Result[0].Expr := TIntegerNode.Create(1) ;
      LSP: { A - Spin orbit coupling }
        if S > 0 then
          Result[0].Expr := EB([eLambda, '*', Sigma])
        else
          Result[0].Expr := TIntegerNode.Create(0);
      LSP+1: { LambdaSS - Spin-spin coupling }
        if S > 1 then begin
          Result[0].Expr := EB([Sigma, '^', 2]);
          Result[0].Expr := EN([3, '*', Result[0].Expr, '-', EN([S*(S+2), '/', 4]), '*', 2, '/', 3]);
        end else
          Result[0].Expr := TIntegerNode.Create(0);
      LSP+2: { o - Omega = 0 lambda doubling }
        if Molecule.AngularMomentum(Symmetry) >= S then begin
          { This changes sign of Lambda }
          dLam := Molecule.AngularMomentum(Symmetry);
          Result[0].Signature[Molecule.nQuantumNumbers-2] := dLam;
          Svar := EN([S, '/', 2]);
          Result[0].Expr := EN([rlE(Svar,dLam,-dLam), '/', 2]);
          Svar.Free;
          { Also need opposite sign change }
          SetLength(Result,2);
          Result[1].Signature := Copy(Result[0].Signature);
          Result[1].Signature[Molecule.nQuantumNumbers-2] := -dLam;
          Result[1].Expr := Result[0].Expr.Duplicate;
        end else
          Result[0].Expr := TIntegerNode.Create(0);
    else
      SetLength(Result,0);
    end;
    Sigma.Free;
  end else
    SetLength(Result,0);
end;

function TElectronicState.RotationalMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Complex;
var
  Bra, Ket: TVibrationalBasisState;
  vn, j, RTMode, dind: Integer;
  vh, Sum: Double;
  Lambda, Sigma: TQuantumNumber;
begin
  Bra := ABra as TVibrationalBasisState;
  Ket := AKet as TVibrationalBasisState;
  Result := z0;
  if (Bra.Sym <> Ket.Sym) or (Bra.J <> Ket.J) then
    raise Exception.Create('Bad TElectronicState.RotationalMatrixElement call');
  if (Bra.v = Ket.v) and (SignA = SignB) then begin
    { Diagonal Matrix elements }
    Result.re := Origin;
    for vn := 0 to Molecule.nModes - 1 do
      with Mode(vn) do begin
        vh := Bra.v[vn]+Molecule.ModeInfo[vn].Degeneracy/2;
        if Molecule.IsDegenerate(Symmetry) then
          Result.re := Result.re + gll*Sqr(Bra.v[Molecule.ModeInfo[vn].dIndex]); { g l^2; note that l is not doubled }
        Sum := 0;
        for j := 4 { zOmega } downto 1 { Omega } do
          Sum := (Sum + GetParameter(j))*vh;
        Result.re := Result.re + Sum;
      end;
    if Molecule.IsDegenerate(Symmetry) or (S > 0) then begin
      Lambda := SignA*Bra.v[High(Bra.v)-1];
      Sigma := SignA*Bra.v[High(Bra.v)] - Lambda;
      Result.re := Result.re + A*Lambda*Sigma / 4
                      { 2/3 lambda (3Sz^2-S^2) }
                      + 2/3 * LambdaSS*(3*Sigma*Sigma - S*(S+2)) / 4;
    end;
  end else if Molecule.IsDegenerate(Symmetry) then begin
    { Look for Renner-Teller style matrix elements }
    dind := High(Bra.v)-1;
    if SignA*Bra.v[dind] <> -SignB*Ket.v[dind] then
      Exit; { Lambda must change sign }
    if SignA*(Bra.v[dind+1]-Bra.v[dind]) <> SignB*(Ket.v[dind+1]-Ket.v[dind]) then begin
      { Sigma changes - possibly Lambda doubling term,
        for which vibrations must not change }
      if SignA*Bra.v[dind+1] <> SignB*Ket.v[dind+1] then
        Exit; { Omega changes }
      for vn := 0 to Molecule.nModes - 1 do begin
        if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
          dind := Molecule.ModeInfo[vn].dIndex;
          if SignA*Bra.v[dind] <> SignB*Ket.v[dind] then
            Exit;
        end;
        if Bra.v[vn] <> Ket.v[vn] then
          Exit;
      end;
      dind := High(Bra.v)-1;
      Result.re := o/2 * rl(S, SignA*(Bra.v[dind+1]-Bra.v[dind]), SignB*(Ket.v[dind+1]-Ket.v[dind]));
      Exit;
    end;
    { Now look for exactly one change in l of the right magnitude }
    RTMode := -1; { Flag no mode found yet }
    for vn := 0 to Molecule.nModes - 1 do begin
      if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
        dind := Molecule.ModeInfo[vn].dIndex;
        if SignA*Bra.v[dind] <> SignB*Ket.v[dind] then
          if RTMode = -1 then
            RTMode := vn { Found candidate }
          else
            RTMode := -2 { Off diagonal in more than one mode }
        else if Bra.v[dind] <> Ket.v[dind] then
          RTMode := -2; { non RT active mode changes }
      end else if Bra.v[vn] <> Ket.v[vn] then
        RTMode := -2; { non RT active mode changes }
      if RTMode = -2 then
        Break;
    end;
    if RTMode >= 0 then begin
      dind := Molecule.ModeInfo[RTMode].dIndex;
      Result.re := Mode(RTMode).epsOmega * vnvpll(Bra.v[RTMode],SignA*Bra.v[dind],Molecule.AngularMomentum(Symmetry),Ket.v[RTMode],SignB*Ket.v[dind]);
    end;
  end;
end;

function TElectronicState.RotationalPerturbationMatrixElement(ABra,
  AKet: TBasisState; SignA, SignB: Integer; MNode: TMNode;
  Flag: Integer): Complex;
var
  Bra, Ket: TVibrationalBasisState;
  P: TVibrationalPerturbation;
begin
  if Flag <> Ord(meHrot) then
    raise NotYet;
  Bra := ABra as TVibrationalBasisState;
  Ket := AKet as TVibrationalBasisState;
  P := MNode as TVibrationalPerturbation;
  if Bra.J <> Ket.J then
    raise Exception.Create('TElectronicState.PerturbationMatrixElement');
  Result.re := P.MatrixElement(Bra.v, Ket.v, SignA, SignB);
  Result.im := 0;
end;

{.$DEFINE VERYSLOWTEST}
function TElectronicState.RotationalTransitionMatrixElement(
  ABra, AKet: TBasisState; SignA, SignB: Integer;
  TransitionMoment: TMNode; F: TQuantumNumber): Complex;
var
  Left, Right: TVibrationalBasisState;
  Dmat: TMatrix;
  dind, vn: Integer;
  Lambda, LambdaP: TQuantumNumber;
  {$IFDEF VERYSLOWTEST}
  r1, r2: Double;
  BraV,KetV: TIntegerArray;
  {$ENDIF}
begin
  Left := ABra as TVibrationalBasisState;
  Right := AKet as TVibrationalBasisState;
  if not (TransitionMoment is TMultidimensionalFCF) then
    raise NotYet;
  Result := z0;
  with TMultidimensionalFCF(TransitionMoment) do
    if (VibronicallyAllowed = vaYes) and PointGroupInfo[Molecule.PointGroup].Allowed(Left.Sym, ActualSymmetry, Right.Sym) then begin
      { First check for no change in electronic quantum numbers }
      if Molecule.HasElectronicAngularMomentum then begin
        dind := Length(Left.v)-2;
        LambdaP := SignA*Left.v[dind];
        Lambda  := SignB*Right.v[dind];
        if (SignA*Left.v[dind+1]-LambdaP) <> (SignB*Right.v[dind+1]-Lambda) then
          Exit; { Sigma must not change }
        if Lambda*LambdaP < 0 then { Check for allowed change in Lambda }
          Exit; { TODO: This is only correct for one photon transitions }
      end;
      { Check for no change in l }
      for vn := 0 to Molecule.nModes - 1 do
        if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
          dind := Molecule.ModeInfo[vn].dIndex;
          if SignA*Left.v[dind] <> SignB*Right.v[dind] then
            Exit; { l must not change }
        end;
      Result.re := TMultidimensionalFCF(TransitionMoment).Strength;
      Dmat := TransitionMatrix(Left,Right,True);
      {$IFDEF VERYSLOWTEST}
      r1 := Dmat[Left.Offset,Right.Offset];
      Brav := Copy(Left.v);
      Ketv := Copy(Right.v);
      for vn := 0 to Molecule.nModes - 1 do
        if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
          dind := Molecule.ModeInfo[vn].dIndex;
          Brav[dind] := SignA*Brav[dind];
          Ketv[dind] := SignB*Ketv[dind];
        end;
      r2 := Overlap(Brav, Ketv);
      if not SameValue(r1, r2) then begin
        r2 := Overlap(Brav, Ketv);
        raise Exception.Create('Help '+Left.BasisLabel+Right.BasisLabel);
      end;
      {$ENDIF}
      Result.re := Result.re * Dmat[Left.Offset,Right.Offset];
    end;
  // No - reduction occurs outside this routine. Result.re := Result.re * Sqrt(Three); { Pretend it is a reduced matrix element }
end;

procedure TElectronicState.SetLmatrix(Sender: TObject);
begin
  {$IFNDEF CONSOLE}
  ConnectLmatrixFrm(Self)
  {$ENDIF}
end;

function TElectronicState.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  if Aindex = LSI then begin
    LowestValue := 0;
    Result := PointGroupInfo[Molecule.PointGroup].SettingsNames;
  end else
    Result := inherited GetSettingsNames(AIndex, LowestValue);
end;

const
 ElectronicStateParameterNames: array [LSP..LSP+3] of string = (
'Spin-orbit coupling constant.',
'Spin - spin coupling constant',
'Spin-spin coupling constant giving lambda doubling for Omega = 0 states.',
'Maximum energy to include in the vibrational basis.'
);

function TElectronicState.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(ElectronicStateParameterNames))
     and (AIndex <= High(ElectronicStateParameterNames)) then
    Result := ElectronicStateParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

function TElectronicState.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = LSI then
    Result := 'Symmetry of electronic state'
  else
    Result := inherited HelpOnSetting(AIndex)
end;

procedure TElectronicState.SetSettingFromStr(AIndex: Integer; const Str: string);
begin
  if AIndex = LSI then
    Symmetry := PointGroupInfo[Molecule.PointGroup].StrToSymmetry(Str)
  else
    inherited;
end;

function TElectronicState.SettingToStr(AIndex: Integer): string;
begin
  if AIndex = LSI then
    Result := Molecule.SymName(Symmetry)
  else
    Result := inherited SettingToStr(AIndex);
end;

function TElectronicState.StateOrigin: Double;
var
  Ket: TVibrationalBasisState;
  Sigma: TQuantumNumber;
  E: Double;
begin
  { Estimate energy of lowest level - Used to stop Boltzman factors blowing up
    at low temperatures, or if there are many modes, when the total zero point
    energy can become very large }
  { Just use the simple matrix elements, so this may fail if perturbations are large }
  Ket := TVibrationalBasisState.Create(nil,0,0);
  try
    SetLength(Ket.v, Molecule.nQuantumNumbers);
    { Estimate lowest spin orbit + spin-spin energy }
    Sigma := -S;
    if Molecule.HasElectronicAngularMomentum then
      Ket.v[High(Ket.v)-1] := Molecule.AngularMomentum(Symmetry); { Lambda }
    while Sigma <= S do begin
      if Molecule.HasElectronicAngularMomentum then
        Ket.v[High(Ket.v)] := Ket.v[High(Ket.v)-1] + Sigma; { Omega }
      E := RotationalMatrixElement(Ket,Ket,1,1,nil,0).Re;
      if (E < Result) or (Sigma = -S) then
        Result := E;
      Inc(Sigma, 2);
    end;
  finally
    Ket.Free;
  end;
end;

{ Return symmetry of given particular combination of quantum numbers in v.
  Result is -1 if particular combination is invalid
   -2 if redundant;
  for a redundant wavefunction the symmetry must be determined by changing the sign of the signed quantum numbers.
  Where two symmetries are possible, the alternate symmetry is returned in AlternateSym;
  for this case phase=1 will have the returned symmetry and phase=-1 will have the AlternateSym
}
function TElectronicState.SymmetryOf(v: array of Integer; out AlternateSym:Integer): Integer;
  function AbelianSymmetry: Integer;
  var
    i: Integer;
  begin
    { Abelian groups with no degenerate modes }
    Result := Symmetry;
    for i := 0 to Molecule.nModes - 1 do
      if Odd(v[i]) then
        Result := Result xor Molecule.ModeInfo[i].Symmetry;
  end;
  function Linear(Centre:Boolean): Integer;
  var
    vn, l, Phase: Integer;
    K, Sigma: TQuantumNumber;
    AllAngularMomentumZero: Boolean;
    LastProjectionIsPositive: Boolean;
  begin
    if Centre then
      Result := Symmetry and 1 { u/g symmetry of state }
    else
      Result := 0;
    if Molecule.HasElectronicAngularMomentum then begin
      K := v[Molecule.nQuantumNumbers-1];
      Sigma := Molecule.AngularMomentum(Symmetry) - K;
    end else begin
      K := 0;
      Sigma := 0;
    end;
    AllAngularMomentumZero := not Molecule.IsDegenerate(Symmetry) and (Sigma = 0);
    LastProjectionIsPositive := False;
    for vn := 0 to Molecule.nModes - 1 do begin
      if Centre and Odd(Molecule.ModeInfo[vn].Symmetry) and Odd(v[vn]) then
        Result := Result xor 1; { Track u/g symmetry }
      if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
        l := v[Molecule.ModeInfo[vn].dIndex]; { Vibrational angular momentum }
        Inc(K,2*l);
        AllAngularMomentumZero := AllAngularMomentumZero and (l = 0);
        if l <> 0 then
          LastProjectionIsPositive := l > 0;
      end;
    end;
    if Odd(K) then
      raise NotYet { TODO: The code below assumes even spin }
    else begin
      if K < 0 then
        { States with overall angular momentum < 0 are redundant - these will be the same as those with the other sign }
        Result := -2
      else if K > 0 then
        K := K div 2 + 1
      else if AllAngularMomentumZero then
        { Must be Sigma+ or Sigma-, reflecting electronic state,
          as vibrations can't be Sigma- }
        if Centre then
          K := Symmetry div 2
        else
          K := Symmetry
      else begin
        { Sigma+ or Sigma- with non-zero angular momenta;
          In this case both +ve and -ve combinations muct be considered, so return two answers }
        { First, remove redundancies, choose the state with:
            Lambda > 0,
            else Omega > 0
            else with the highest numbered l value > 0; note that
              LastProjectionIsPositive is already set for this last condition }
          if Molecule.HasElectronicAngularMomentum then
            if v[High(v)-1] <> 0 then
              LastProjectionIsPositive := v[High(v)-1] > 0
            else if v[High(v)] <> 0 then
              LastProjectionIsPositive := v[High(v)] > 0;
          if not LastProjectionIsPositive then begin
            Result := -1; { Redundant }
            Exit;
          end;
        { Now work out correct phase factor to give Sigma+ or Sigma- }
        { Effect of Sigma(v) is (-1)^(l1+l2+...)(-1)^(Lambda+s)(-1)^(S-Sigma) * (state with all angular momenta reversed)
          where s is 1 for a Sigma- state, and zero otherwise.
          This is the same as (-1)^(K+s-S); if this is +ve then the + combination is Sigma+
          This section is only reached if K = 0 }
        if Molecule.AngularMomentum(Symmetry) = 0 then
          { Work out s }
          if Centre then
            Phase := Symmetry div 2
          else
            Phase := Symmetry
        else
          Phase := 0;
        Phase := Minus(2*Phase-S);
        if Phase < 0 then begin
          K := 1; { + combination is Sigma- }
          AlternateSym := Result; { Sigma+ g/u }
        end else
          if Centre then
            AlternateSym := 2 + Result { Sigma- g/u }
          else
            AlternateSym := 1;    { Sigma- }
      end;
      if Centre then  { Add back g/u symmetry }
        Result := K*2 + Result
      else
        Result := K;
    end;
  end;
begin
  AlternateSym := -1; { Flag only one result possible }
  case Molecule.PointGroup of
    CInfinityv: Result := Linear(False);
    DInfinityh: Result := Linear(True);
  else
    Result := AbelianSymmetry;
  end;
end;

{ TVibrationalMode }

procedure TVibrationalMode.AdjustParameterNames;
var
  i: Integer;
begin
  { Add/remove l matrix parameters as required }
  for i := nParameters+1 to FirstDynamicParameter + 3*Molecule.nNuclei - 1 do
    AddParameter(Format('l%s%d',[Char(Ord('x')+(i-FirstDynamicParameter) mod 3),(i-FirstDynamicParameter) div 3]));
  for i := nParameters downto FirstDynamicParameter + 3*Molecule.nNuclei do
    DeleteParameter(i);
end;

function TVibrationalMode.PertEquiv(AIndex: Integer):TVibPertEquiv;
begin
  SetLength(Result.Powers, Molecule.nModes);
  case AIndex of
    1..4: { Omega .. zOmega }
      begin
        Result.Op := x;
        Result.Powers[NodeIndex] := AIndex;
      end;
//    5: { gll }
    6: { epsOmega }
      begin
        Result.Op := q;
        Result.Powers[NodeIndex] := -Molecule.AngularMomentum(Symmetry);
      end;
  else
    SetLength(Result.Powers,0) { Flag not (yet) implemented }
  end;
end;

constructor TVibrationalMode.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',vMax=5';
  inherited Create(AOwner);
end;

procedure TVibrationalMode.AddHandlers;
begin
  inherited;
  AddHandler('Sort', 'Sort modes by symmetry and frequency', {$IFDEF FPC}@{$ENDIF}Sort);
  AddHandler('Global Shift Up', 'Shift mode in all states', {$IFDEF FPC}@{$ENDIF}GlobalMoveUp);
  AddHandler('Global Shift Down', 'Shift mode in all states', {$IFDEF FPC}@{$ENDIF}GlobalMoveDown);
end;

procedure TVibrationalMode.Sort(Sender: TObject);
var
  i, d: Integer;
  Change: Boolean;
  State: TElectronicState;
begin
  State := Owner as TElectronicState;
  repeat
    Change := False;
    for i := 0 to Molecule.nModes - 2 do begin
      d := State.Mode(i).Symmetry - State.Mode(i+1).Symmetry;
      if (d > 0) or ((d=0) and (State.Mode(i).Omega < State.Mode(i+1).Omega)) then begin
        Change := True;
        State.Mode(i).GlobalShift(1);
      end;
    end;
  until not Change;
end;

procedure TVibrationalMode.GlobalMoveDown(Sender: TObject);
begin
  GlobalShift(1);
end;

procedure TVibrationalMode.GlobalMoveUp(Sender: TObject);
begin
  GlobalShift(-1);
end;

procedure TVibrationalMode.GlobalShift(Shift: Integer);
var
  Source, Target: TVibrationalMode;
  State: TElectronicState;
  s: string;
  i, j, MyIndex, Temp: Integer;
begin
  if (NodeIndex+Shift < 0) or (NodeIndex+Shift >= Molecule.nModes) then
    Exit;
  State := Owner as TElectronicState;
  Target := State.Mode(NodeIndex+Shift);
  { Move global values once }
  s := Name;
  if (s <> Format('v%d', [NodeIndex+1])) and (Target.Name <> Format('v%d', [Target.NodeIndex+1])) then begin
    { Only exchange names if using non-default names }
    Name := Target.Name;
    Target.Name := s;
  end;
  MyIndex := NodeIndex;
  { Now move all other settings }
  for i := 0 to Molecule.nManifolds - 1 do
    with Molecule.Manifold[i] do begin
      for j := 0 to nStates - 1 do
        with State[j] as TElectronicState do begin
          Source := Mode(MyIndex);
          Target := Mode(MyIndex+Shift);
          if Shift > 0 then
            Source.ComponentIndex := Target.ComponentIndex
          else
            Target.ComponentIndex := Source.ComponentIndex;
          AdjustModes(False);
        end;
      for j := 0 to nPerturbations - 1 do
        with Perturbation[j] do begin
          { Shift operator powers }
          Temp := Settings[FirstDynamicSetting+MyIndex];
          Settings[FirstDynamicSetting+MyIndex] := Settings[FirstDynamicSetting+MyIndex+Shift];
          Settings[FirstDynamicSetting+MyIndex+Shift] := Temp;
        end;
    end;
  for i := 0 to Molecule.nTransitionMoments - 1 do
    with Molecule.GetTransitionMoments(i) do
      for j := 0 to ComponentCount - 1 do
        if TransitionMoment[j] is TMultidimensionalFCF then
          with TMultidimensionalFCF(TransitionMoment[j]) do begin
            SwapModes(MyIndex,MyIndex+Shift);
            AdjustModes(False);
          end;
  { Reset symmetry, to ensure copied to global array }
  SetSetting(1,GetSetting(1));
  Target.SetSetting(1,Target.GetSetting(1));
end;

const
 VibrationalModeParameterNames: array [1..6] of string = (
'Vibrational Frequency.',
'(v+1/2)^2 anharmonicity.',
'(v+1/2)^3 anharmonicity.',
'(v+1/2)^4 anharmonicity.',
'l^2 term.',
'Renner-Teller term'
);

function TVibrationalMode.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex > High(VibrationalModeParameterNames) then begin
    AIndex := AIndex-High(VibrationalModeParameterNames)-1;
    Result := 'l matrix element for ' + Char(Ord('x') + AIndex mod 3)
              + ' co-ordinate for nucleus ' + (Owner as TState).Nucleus(AIndex div 3).Name
              + ' / Sqrt(amu)';
  end else if AIndex >= Low(VibrationalModeParameterNames) then
    Result := VibrationalModeParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

const
 VibrationalModeSettingNames: array [2..3] of string = (
'Minimum vibrational quantum number for this mode to include in the basis.',
'Maximum vibrational quantum number for this mode to include in the basis.'
);

function TVibrationalMode.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(VibrationalModeSettingNames))
     and (AIndex <= High(VibrationalModeSettingNames)) then
    Result := VibrationalModeSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

{ TVibrationalBasisState }

function TVibrationalBasisState.VibLabel: string;
var
  vn, dind: Integer;
  Mol: TVibratingMolecule;
  FirstQno: Boolean;
begin
  Mol := (State as TElectronicState).Molecule;
  FirstQno := True;
  for vn := 0 to Mol.nModes - 1 do
    if Mol.ModeInfo[vn].ShowQno then begin
      if not FirstQno then
        Result := Result + ',';
      FirstQno := False;
      Result := Result + IntToStr(v[vn]);
      if Mol.IsDegenerate(Mol.ModeInfo[vn].Symmetry) then
        Result := Result + '(' + IntToStr(v[Mol.ModeInfo[vn].dIndex]) +')';
    end;
  if Mol.HasElectronicAngularMomentum then begin { Lambda, Omega }
    dind := High(v)-1;
    if (v[dind] <> 0) or (State.S > 0)then begin
      Result := Result + '[';
      if v[dind] <> 0 then begin
        if v[dind] > 0 then
          Result := Result + '+'
        else
          Result := Result + '-';
        if State.S > 0 then
          Result := Result + ','
      end;
      if State.S > 0 then
        Result := Result + State.Manifold.Mixture.QuantumNumberToStr(v[dind+1]);
      Result := Result + ']';
    end;
  end else if FirstQno then
    Result := '0'; { Must put something here }
  if Phase > 0 then
    Result := Result + '+'
  else if Phase < 0 then
    Result := Result + '-';
  Result := Result + HyperfineStateLabel;
  { TODO: Make option  Result := Result + Format('%d',[StateIndex]); }
end;

function TVibrationalBasisState.BasisLabel: string;
begin
  if State.Manifold.Molecule.nManifolds > 1 then
    Result := State.Manifold.Name + ' '
  else
    Result := '';
  Result := Result + State.Name + ' ' + VibLabel;
end;

function TVibrationalBasisState.BasisLabelKey: string;
begin
  Result := 'Name ' + VibLabelKey;
end;

function TVibrationalBasisState.VibLabelKey: string;
var
  vn: Integer;
  Mol: TVibratingMolecule;
  FirstQno: Boolean;
begin
  Mol := (State as TElectronicState).Molecule;
  FirstQno := True;
  for vn := 0 to Mol.nModes - 1 do
    if Mol.ModeInfo[vn].ShowQno then begin
      if not FirstQno then
        Result := Result + ',';
      FirstQno := False;
      Result := Result + Mol.ModeInfo[vn].Name;
      if Mol.IsDegenerate(Mol.ModeInfo[vn].Symmetry) then begin
        Result := Result + '(l' + Copy(Mol.ModeInfo[vn].Name,2,255) +')';
      end;
    end;
  if Mol.HasElectronicAngularMomentum then
    Result := Result + '[Lambda,Omega]'
  else if FirstQno then
    Result := 'v';
  Result := Result + '+-';
end;

function TVibrationalBasisState.MakeSignature(SignA: Integer; AKet: TBasisState; SignB: Integer): TIntegerArray;
var
  vn, i: Integer;
  Mol: TVibratingMolecule;
  Ket: TVibrationalBasisState;
begin
  { Signature is array of changes in vibronic quantum numbers }
  SetLength(Result,Length(v));
  Mol := (State as TElectronicState).Molecule;
  Ket := AKet as TVibrationalBasisState;
  for vn := 0 to Mol.nModes - 1 do begin
    Result[vn] := v[vn] - Ket.v[vn];
    if Mol.IsDegenerate(Mol.ModeInfo[vn].Symmetry) then begin
      i := Mol.ModeInfo[vn].dIndex;
      Result[i] := (SignA*v[i]-SignB*Ket.v[i]);
    end;
  end;
  if Mol.HasElectronicAngularMomentum then begin { Lambda, Omega }
    i := High(v)-1;
    Result[i] := (SignA*v[i]-SignB*Ket.v[i]) div 2;
    Result[i+1] := (SignA*v[i+1]-SignB*Ket.v[i+1]) div 2;
  end;
end;

function TVibrationalBasisState.Qno(QnoType: TQnoType): TQuantumNumber;
var
  o: Integer;
begin
  o := Ord(QnoType)-Ord(qtA);
  if o = Length(v) then
    Result := Phase
  else if (o < Length(v)) and (o >= 0) then
    Result := v[o]
  else
    Result := inherited Qno(QnoType);
end;

procedure TVibrationalBasisState.SetSymbolicVariables(SignA: Integer; AKet: TBasisState; SignB: Integer);
var
  vn, dind: Integer;
  Mol: TVibratingMolecule;
  Ket: TVibrationalBasisState;
begin
  Mol := (State as TElectronicState).Molecule;
  Ket := AKet as TVibrationalBasisState;
  for vn := 0 to Mol.nModes - 1 do begin
    vVariable(vn).Value := Ket.v[vn];
    if Mol.IsDegenerate(Mol.ModeInfo[vn].Symmetry) then begin
      dind := Mol.ModeInfo[vn].dIndex;
      lVariable(vn).Value := SignB*Ket.v[dind];
    end;
  end;
  if Mol.HasElectronicAngularMomentum then begin { Lambda, Omega }
    dind := High(v)-1;
    eLambda.Value := SignB*Ket.v[dind] / 2;
    eOmega.Value := SignB*Ket.v[dind+1] / 2;
  end;
end;

function TVibrationalBasisState.StateLabel(AQuantumNumberFormat: TQuantumNumberFormat): string;
begin
  Result := BasisLabel;
end;

function TVibrationalBasisState.StateLabelKey: string;
begin
  Result := BasisLabelKey;
end;

function TVibrationalBasisState.TxName(Upper, Lower: TBasisState): string;
begin
  Result := '(' + (Upper as TVibrationalBasisState).VibLabel + ')-(' + (Lower as TVibrationalBasisState).VibLabel + ')';
end;

function TVibrationalBasisState.TxNameKey(Upper, Lower: TBasisState): string;
begin
  Result := '(' + (Upper as TVibrationalBasisState).VibLabelKey + ')-(' + (Lower as TVibrationalBasisState).VibLabelKey + ')';
end;

{ TMultidimensionalFCF }

procedure TMultidimensionalFCF.AddHandlers;
begin
  inherited;
  AddHandler('Dushinsky Matrix...', 'View and edit Dushinsky transformation', {$IFDEF FPC}@{$ENDIF}ShowDushinsky);
  AddHandler('Print Matrices', 'Print detailed information about Franck Condon calculation', {$IFDEF FPC}@{$ENDIF}PrintInfo);
  AddHandler('Set Modes From Calculated', 'Set mode displacements and rotations from l matrices', {$IFDEF FPC}@{$ENDIF}SetModeInfo);
end;

procedure TMultidimensionalFCF.SetModeInfo(Sender: TObject);
begin
  PrepareMatrices(True,True,False);
end;

procedure TMultidimensionalFCF.AdjustModes(FixNumber: Boolean);
var
  j, n: Integer;
begin
  if FixNumber then
    n := AdjustNodeIndexes(TVibrationalModeChange, FModes, nil, opInsert, Molecule.nModes)
  else
    n := AdjustNodeIndexes(TVibrationalModeChange, FModes);
  for j := 0 to n - 1 do
    Mode(j).AdjustParameterNames;
end;

procedure TMultidimensionalFCF.ClearCache;
var
  i: Integer;
begin
  inherited;
  ClearMatrices;
  for i := 0 to High(FMatrixCache) do
    FMatrixCache[i].Free;
  SetLength(FMatrixCache,0);
end;

procedure TMultidimensionalFCF.ClearMatrices;
begin
  FreeAndNil(Q1);
  FreeAndNil(P1);
  FreeAndNil(P1);
  FreeAndNil(R);
  FreeAndNil(Rdelta);
  FreeAndNil(P1Delta);
end;

constructor TMultidimensionalFCF.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Symmetry=Auto';
  inherited;
  if Assigned(AOwner) then
    FMolecule := Owner.Owner as TVibratingMolecule;
end;

destructor TMultidimensionalFCF.Destroy;
begin
  ClearCache;
  FreeAndNil(d);
  FreeAndNil(S);
  FreeAndNil(CacheLock);
  inherited;
end;

function TMultidimensionalFCF.FixupProperties: Exception;
var
  i: Integer;
  Left, Right: TElectronicState;
begin
  if (nWorkers > 0) and not Assigned(CacheLock) then
    CacheLock := TMultiReadExclusiveWriteSynchronizer.Create;
  Result := inherited FixupProperties;
  if Molecule.nNuclei = 0 then
    Calculate := False;
  AdjustModes(True);
  if not Assigned(Result) then
    try
      FActualSymmetry := Symmetry;
      Left := Bra as TElectronicState;
      Right := Ket as TElectronicState;
      if Symmetry < 0 then begin
        { Work out symmetry of transition moment }
        FActualSymmetry := -1;
        for i := 0 to Molecule.LastSym do
          if PointGroupInfo[Molecule.PointGroup].Allowed(Left.Symmetry, i, Right.Symmetry) then begin
            if FActualSymmetry >= 0 then
              raise Exception.Create('Auto transition moment not possible for '+UniqueName);
            FActualSymmetry := i;
            VibronicallyAllowed := vaYes;
          end
      end else
        if PointGroupInfo[Molecule.PointGroup].Allowed(Left.Symmetry, ActualSymmetry, Right.Symmetry) then
          VibronicallyAllowed := vaYes
        else
          VibronicallyAllowed := vaNo;
      PrepareMatrices(False,False,True);
    except
      on E: Exception do begin
        Result := Exception.Create(E.Message);
      end;
    end;
end;

function TMultidimensionalFCF.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  if Aindex = 1 then begin
    LowestValue := -1;
    Result := PointGroupInfo[Molecule.PointGroup].SettingsNames;
    Dec(Result); { Allow 'Auto' also }
  end else
    Result := inherited GetSettingsNames(AIndex, LowestValue);
end;

const
  MultidimensionalFCFSettingNames: array[1..2] of string = (
'Symmetry of electronic transition moment; default is ''auto''.',
'True to calculate normal co-ordinate displacements and Dushinsky matrix from l matrix.'
);

function TMultidimensionalFCF.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(MultidimensionalFCFSettingNames))
     and (AIndex <= High(MultidimensionalFCFSettingNames)) then
    Result := MultidimensionalFCFSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

procedure TMultidimensionalFCF.InvertTransformation;
begin
  { TODO: This has no effect, as d and S are overwritten } 
  PrepareMatrices(True,False,False);
  Invert(S);
  d.PreMultiply(S);
  d.Scale(-1);
  Writeln(Log,'Inverse Transform');
  d.Write(Log,'d');
  S.Write(Log,'S');
  Modified := True;
end;

function TMultidimensionalFCF.Mode(Index: Integer): TVibrationalModeChange;
begin
  if (Index >= Length(FModes)) and (Index < Molecule.nModes) then
    AdjustModes(True);
  Result := FModes[Index] as TVibrationalModeChange;
end;

{ Degenerate modes from:
  E V Doktorov, I A Malkin and V I Man'ko
  J Phys. B: Atom. Molec. Phys. 9, 507 (1976) }
{ Note that the sign of l is ignored in this rourine }
function TMultidimensionalFCF.Overlap(var ABra, AKet: array of Integer): Double;
var
  vp, v, l, vn, vn1: Integer;
  Scale: Double;
  OmegaP, Omega, Vm: Double;
begin
  { Note that arguments are not constants to avoid multiple copies of quantum numbers }
  { TODO: This makes it not thread safe }
  Result := 0;
  for vn := 0 to Molecule.nModes-1 do
    { Pick a mode to reduce }
    if AKet[vn] > 0 then
      if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
        { Special handling for reducing degenerate modes }
        { TODO: currently does not take account of mode mixing for degenerate modes }
        v := AKet[vn];
        l := Abs(AKet[Molecule.ModeInfo[vn].dIndex]);
        if v > l then begin
          vp := ABra[vn];
          Omega := (Ket as TElectronicState).Mode(vn).Omega;
          OmegaP := (Bra as TElectronicState).Mode(vn).Omega;
          Scale := Omega + OmegaP;
          { Reduce v }
          if v > 3 then begin
            Vm := (Omega-OmegaP)/Scale*Sqrt((v-2-l)*(v-2+l)/(v-l)/(v+l));
            if Vm <> 0 then begin
              Dec(AKet[vn],4);
              Result := Vm*Overlap(ABra,AKet);
              Inc(AKet[vn],4);
            end;
          end;
          if vp > 1 then begin
            Vm := 2*Omega/Scale*Sqrt((vp-l)*(vp+l)/(v-l)/(v+l));
            if Vm <> 0 then begin
              Dec(AKet[vn],2);
              Dec(ABra[vn],2);
              Result := Result + Vm*Overlap(ABra,AKet);
              Inc(ABra[vn],2);
              Inc(AKet[vn],2);
            end;
          end;
          Dec(AKet[vn],2);
          Result := Result + 2*(OmegaP*(v-1)-Omega*(vp+1))/Scale/Sqrt((v-l)*(v+l))*Overlap(ABra,AKet);
          Inc(AKet[vn],2);
          Exit;
        end else if v < l then
          raise Exception.Create('Overlap1') { Sanity check }
        else
          Continue; { This degnerate mode at minmum value for l }
      end else begin
        { Non degenerate mode }
        Scale := AKet[vn];
        Dec(AKet[vn]);
        for vn1 := 0 to Molecule.nModes-1 do
          if ABra[vn1] > 0 then begin
            if (vn=vn1) or (Molecule.ModeInfo[vn1].Symmetry = 0) then begin
              Dec(ABra[vn1]);
              Result := Result + 2*R[vn+1,vn1+1]*Sqrt((ABra[vn1]+1)/Scale)*Overlap(ABra,AKet);
              Inc(ABra[vn1]);
            end;
          end;
        for vn1 := 0 to Molecule.nModes-1 do
          if AKet[vn1] > 0 then begin
            if (vn=vn1) or (Molecule.ModeInfo[vn1].Symmetry = 0) then begin
              Dec(AKet[vn1]);
              Result := Result + Q1[vn+1,vn1+1]*Sqrt((AKet[vn1]+1)/Scale)*Overlap(ABra,AKet);
              Inc(AKet[vn1]);
            end;
          end;
        if Molecule.ModeInfo[vn].Symmetry = 0 then
          Result := Result - Rdelta[vn+1]*Sqrt(2/Scale)*Overlap(ABra,AKet);
        Inc(AKet[vn]);
        Exit;
      end;
  { Ket all zero; reduce Bra }
  for vn := 0 to Molecule.nModes-1 do
    { Pick a mode to reduce }
    if ABra[vn] > 0 then begin
      if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
        { Special handling for reducing degenerate modes }
        vp := ABra[vn];
        l := Abs(ABra[Molecule.ModeInfo[vn].dIndex]);
        if vp > l then begin
          v := AKet[vn];
          Omega := (Ket as TElectronicState).Mode(vn).Omega;
          OmegaP := (Bra as TElectronicState).Mode(vn).Omega;
          Scale := OmegaP + Omega;
          { Reduce vp }
          if vp > 3 then begin
            Vm := (OmegaP-Omega)/Scale*Sqrt((vp-2-l)*(vp-2+l)/(vp-l)/(vp+l));
            if Vm <> 0 then begin
              Dec(ABra[vn],4);
              Result := Vm*Overlap(ABra,AKet);
              Inc(ABra[vn],4);
            end;
          end;
          if v > 1 then begin
            Vm := 2*OmegaP/Scale*Sqrt((v-l)*(v+l)/(vp-l)/(vp+l));
            if Vm <> 0 then begin
              Dec(AKet[vn],2);
              Dec(ABra[vn],2);
              Result := Result + Vm*Overlap(ABra,AKet);
              Inc(ABra[vn],2);
              Inc(AKet[vn],2);
            end;
          end;
          Dec(ABra[vn],2);
          Result := Result + 2*(Omega*(vp-1)-OmegaP*(v+1))/Scale/Sqrt((vp-l)*(vp+l))*Overlap(ABra,AKet);
          Inc(ABra[vn],2);
          Exit;
        end else if vp < l then
          raise Exception.Create('Overlap2') { Sanity check }
        else
          Continue; { This degnerate mode at minmum value for l }
      end else begin
        Scale := ABra[vn];
        Dec(ABra[vn]);
        for vn1 := 0 to Molecule.nModes-1 do { Note that this is redundant code }
          if AKet[vn1] > 0 then begin
            if (vn=vn1) or (Molecule.ModeInfo[vn1].Symmetry = 0) then begin
              Dec(AKet[vn1]);
              Result := Result + 2*R[vn+1,vn1+1]*Sqrt((AKet[vn1]+1)/Scale)*Overlap(ABra,AKet);
              Inc(AKet[vn1]);
            end;
          end;
        for vn1 := 0 to Molecule.nModes-1 do
          if ABra[vn1] > 0 then begin
            if (vn=vn1) or (Molecule.ModeInfo[vn1].Symmetry = 0) then begin
              Dec(ABra[vn1]);
              Result := Result + P1[vn+1,vn1+1]*Sqrt((ABra[vn1]+1)/Scale)*Overlap(ABra,AKet);
              Inc(ABra[vn1]);
            end;
          end;
        if Molecule.ModeInfo[vn].Symmetry = 0 then
          Result := Result + P1Delta[vn+1]*Sqrt(2/Scale)*Overlap(ABra,AKet);
        Inc(ABra[vn]);
        Exit;
      end;
    end;
  Result := I00;
  for vn := 0 to Molecule.nModes-1 do
    { Special action needed for degenerate modes - these will have v' = v = l, rather than 0;
      in addition I00 is wrong, as mode should have been counted twice }
    if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
      l := Abs(ABra[Molecule.ModeInfo[vn].dIndex]);
      if l <> Abs(AKet[Molecule.ModeInfo[vn].dIndex]) then begin
        Result := 0;
        Exit;
      end;
      l := Abs(l);
      Omega := (Ket as TElectronicState).Mode(vn).Omega;
      OmegaP := (Bra as TElectronicState).Mode(vn).Omega;
      Result := Result * Sqrt(IntPower(2*Sqrt(Omega*Omegap)/(Omega+Omegap),2*l+1))
    end;
end;

function TMultidimensionalFCF.PairName: string;
begin
  Result := 'FCF';
end;

procedure TMultidimensionalFCF.PrepareDushinsky(nModes: Integer; Print: Boolean; SetModeInfo: Boolean);
var
  C, T, Rc, BestS: TMatrix;
  slam, Req, Rpeq, SumV: TColumnVector;
  BestTrace, Det: Double;
  nSMatrix, BestSMatrix, nNuclei: Integer;
  Left, Right: TElectronicState;
  procedure TryS(const signs: array of Integer);
  var
    S, Cinv, Z, D: TMatrix;
    i, j, m: Integer;
    Trace: Double;
  begin
    { Eigenvalue check }
{
    if Print then begin
      S := Rc.Transpose;
      Cinv := C.Transpose;
      S.PostMultiply(Cinv);
      Cinv.Free;
      S.PostMultiply(C);
      S.PostMultiply(Rc);
      S.Write(Log,'RCtCRt');
      S.Free;
    end;
}
    S := TMatrix.Create(Rc);
    for i := 1 to 3 do
      for j := 1 to 3 do
        S[i,j] := S[i,j] * Sqrt(slam[j]) * signs[j-1];
//    if Print then
//      S.Write(Log,'Rt Lambda Sqrt(lambda)');
    Cinv := Rc.Transpose;
    S.PostMultiply(Cinv);
    Cinv.Free;
//    if Print then
//      S.Write(Log,'Rt Lambda Sqrt(lambda) R');
    Cinv := TMatrix.Create(C);
    Invert(Cinv);
    if Print then
      Cinv.Write(Log,'C-1');
    S.PostMultiply(Cinv);
    Cinv.Free;
    Trace := S[1,1] + S[2,2] + S[3,3];
    Det := Determinant(S);
    Inc(nSMatrix);
    if Print then begin
      S.Write(Log,'S trial #'+IntToStr(nSMatrix)+', trace = '+FloatToStr(Trace)+', determinant = '+FloatToStr(Det));
      Z := TMatrix.Create(3*nNuclei,3*nNuclei);
      Z.Zero;
      for m := 0 to nNuclei - 1 do
        for i := 1 to 3 do
          for j := 1 to 3 do
            Z[3*m + i, 3*m+j] := S[i,j];
      D := Left.Madm.Transpose;
      D.PostMultiply(Z);
      D.PostMultiply(Right.adm);
      D.Write(Log,'Trial Dushinsky');
      D.Free;
      Z.Free;
    end;
    if (Det > 0) and (not Assigned(BestS) or (Trace > BestTrace)) then begin
      BestS.Free;
      BestS := S;
      BestTrace := Trace;
      BestSMatrix := nSMatrix;
    end else
      S.Free;
  end;
var
  i, j, k, m, xBase: Integer;
  LeftN, RightN: TMolecularNucleus;
  Sum, v: Double;
  sn, cs: Extended;
  AllZero: Boolean;
  Z, De: TMatrix;
begin
  { Prepare Matrices as per Gruner and Brumer }
  ClearMatrices;
  if not Assigned(d) then begin
    { Initial creation }
    d := TColumnVector.Create(0);
    S := TMatrix.Create(0,0);
  end;
  { Update sizes as required }
  d.SetSizeAndClear(nModes);
  m := S.RowCount;
  S.Resize(nModes,nModes);
  if m < nModes then
    for i := m+1 to nModes do
      S[i,i] := 1;
  if Calculate and (Molecule.nNuclei > 0) then begin
    Left := Bra as TElectronicState;
    xBase := Left.Nucleus(0).IndexOfParameter('x');
    nNuclei := Left.Molecule.nNuclei;
    Right := Ket as TElectronicState;
    { d ( = K Chen ) = displacement between normal co-ordinates in the two states }
    for m := 0 to nModes - 1 do begin
      sum := 0;
      for i := 0 to nNuclei - 1 do begin
        LeftN := Left.Nucleus(i) as TMolecularNucleus;
        RightN := Right.Nucleus(i) as TMolecularNucleus;
        for k := 0 to 2 do
          sum := sum + (RightN.GetParameter(xBase+k) - LeftN.GetParameter(xBase+k))
                   *Left.Madm[3*i+k+1,m+1];
        end;
      d[m+1] := sum;
      if SetModeInfo then
        Mode(m).Displacement := d[m+1];
    end;
    { Chen eq 22 }
    C := TMatrix.Create(3,3);
    for i := 0 to 2 do { x, y, z }
      for j := 0 to 2 do begin { x, y, z }
        Sum := 0;
        AllZero := True;
        for m := 0 to nNuclei - 1 do begin
          LeftN := Left.Nucleus(m) as TMolecularNucleus;
          RightN := Right.Nucleus(m) as TMolecularNucleus;
          { TODO: check i j order }
          v := LeftN.Mass * LeftN.GetParameter(xBase+j) * RightN.GetParameter(xBase+i);
          Sum := Sum + v;
          AllZero := AllZero and (Abs(v) < 1e-10)
        end;
        if AllZero and (i=j) then
          Sum := 1; { Avoid issues with planar molecules }
        C[i+1,j+1] := Sum;
      end;
    if Print then
      C.Write(Log,'C matrix');
    Rc := C.Transpose;
    Rc.PostMultiply(C);
    if Print then
      Rc.Write(Log,'CtC matrix');
    slam := TColumnVector.Create(3);
    Eigen(slam,Rc,True); { This generates Rt in Chen's notation }
    if Print then begin
      slam.Write(Log,'lambda');
      Rc.Write(Log,'Rt matrix');
    end;
    { Display 4 possible S matrices }
    BestS := nil;
    nSMatrix := 0;
    TryS([1,1,1]);
    TryS([-1,-1,1]);
    TryS([-1,1,-1]);
    TryS([1,-1,-1]);
    TryS([-1,1,1]);
    TryS([1,-1,1]);
    TryS([1,1,-1]);
    TryS([-1,-1,-1]);
    slam.Free;
    Rc.Free;
    C.Free;
    if Print then
      WriteLn(Log, 'Using matrix ' + IntToStr(BestSMatrix));
    if Molecule.PointGroup in [C1, Ci, Vib.Cs, C2h] then begin
      Z := TMatrix.Create(3*nNuclei,3*nNuclei);
      Z.Zero;
      for m := 0 to nNuclei - 1 do
        for i := 1 to 3 do
          for j := 1 to 3 do
            Z[3*m + i, 3*m+j] := BestS[i,j];
      if Print then begin
        Z.Write(Log, 'Z');
        { Check Chen eq 21 }
        SumV := TColumnVector.Create(3);
        SumV.Zero;
        Req := TColumnVector.Create(3);
        Rpeq := TColumnVector.Create(3);
        for m := 0 to nNuclei - 1 do begin
          LeftN := Left.Nucleus(m) as TMolecularNucleus;
          RightN := Right.Nucleus(m) as TMolecularNucleus;
          for i := 0 to 2 do begin
            Req[i+1] := RightN.GetParameter(xBase+i);
            Rpeq[i+1] := LeftN.GetParameter(xBase+i);
          end;
          Req.PreMultiply(BestS);
          Rpeq.CrossProduct(Req);
          for i := 1 to 3 do
            SumV[i] := SumV[i] + Rpeq[i] * LeftN.Mass;
        end;
        Sumv.Write(Log, 'Rotation Check');
        Req.Free;
        Rpeq.Free;
        SUmV.Free;
      end;
    end else
      Z := nil;
    { S ( = J Chen ) = rotation between co-ordinate systems }
    S.Free;
    S := Left.Madm.Transpose;
    if Assigned(Z) then
      S.PostMultiply(Z);
    S.PostMultiply(Right.adm);
    Z.Free;
    BestS.Free;
  end else begin
    S.Zero;
    S.AddDiagonal(1);
    T := TMatrix.Create(nModes,nModes);
    for m := 1 to nModes do begin
      d[m] := Mode(m-1).Displacement;
      for i := 1 to m-1 do begin
        with Mode(m-1) do
          v := GetParameter(FirstDynamicParameter+i-1);
        if v <> 0 then begin
          SinCos(v*Pi/180,sn,cs);
          T.Zero;
          T.AddDiagonal(1);
          T[m,m] := cs;
          T[i,i] := cs;
          T[m,i] := sn;
          T[i,m] := -sn;
          S.PostMultiply(T);
        end;
      end;
    end;
    T.Free;
  end;
  if Print or SetModeInfo then begin
    WriteLn(Log,'Decomposition into equivalent rotations');
    try
      De := FindRotations(S);
      if SetModeInfo then
        for m := 1 to nModes do
          for i := 1 to m-1 do
            with Mode(m-1) do
              SetParameter(FirstDynamicParameter+i-1,De[m,i]);
    except
      on E: EOutOfMemory do
        if SetModeInfo then
          raise
        else
          WriteLn(Log,' **** Skipped: too many modes');
    end;
    De.Free;
  end;
end;

procedure TMultidimensionalFCF.PrepareMatrices(Print: Boolean; SetModeInfo, AllModes: Boolean);
var
  Left, Right: TElectronicState;
  PreExp: Double;
  m, k, nModes: Integer;
  J: TMatrix;
begin
  Left := Bra as TElectronicState;
  Right := Ket as TElectronicState;
  Left.PrepareMatrices(Print, AllModes);
  Right.PrepareMatrices(Print, AllModes);
  nModes := Left.Molecule.nModes;

  PrepareDushinsky(nModes, Print, SetModeInfo);
  if Print then
    d.Write(Log,'Mode Displacements');
  if Print then
    S.Write(Log, 'Dushinsky Matrix');
  J := S.Transpose;
  J.PostMultiply(S);
  if Print then { Is the matrix necessarily Orthogonal? }
    J.Write(Log, 'Dushinsky Orthogonality Check');
  { J = lambda S lambda }
  J.Assign(S);
  J.Resize(nModes, nModes); { Only want vibrations }
  for m := 0 to nModes - 1 do
    for k := 0 to nModes - 1 do
      if Right.Mode(k).Omega = 0 then
        J[m+1,k+1] := 0 { Avoid divide by sero }
      else
        J[m+1,k+1] := Sqrt(Abs(Left.Mode(m).Omega / Right.Mode(k).Omega))*J[m+1,k+1];
  { Q = Inverse(1-J Jt) }
  Q1 := J.Transpose;
  Q1.PostMultiply(J);
  Q1.AddDiagonal(1);
  if Print then
    Q1.Write(Log,'1+J Jt');
  Invert(Q1);
  if Print then
    Q1.Write(Log,'Q');
  { R = Q Jt}
  R := TMatrix.Create(Q1);
  P1 := J.Transpose;
  R.PostMultiply(P1);
  if Print then
    R.Write(Log,'R');
  { P = J Q Jt }
  P1.Assign(J);
  P1.PostMultiply(R);
  if Print then
    P1.Write(Log,'P');
  { delta = lambda' d }
  Rdelta := TColumnVector.Create(d);
  for m := 1 to nModes do
    { Abs is not strictly necessary in next line, but avoid unhelpful crash in later working
      if Omega values are accidentally set negative }
    Rdelta[m] := Rdelta[m]*Sqrt(Abs(Left.Mode(m-1).Omega*ReduceFrequency));
  if Print then
    Rdelta.Write(Log,'delta');
  { P1Delta = (1-P) delta }
  P1Delta := TColumnVector.Create(Rdelta);
  P1Delta.PreMultiply(P1);
  P1Delta.Scale(-1);
  P1Delta.Add(Rdelta);
  if Print then
    P1Delta.Write(Log,'(1-P)delta');
  { I00 = -1/2 delta^T (1-P) delta}
  I00 := -Rdelta.DotProduct(P1Delta)/2;
  PreExp := 1;
  for m := 0 to nModes - 1 do
    PreExp := PreExp * Left.Mode(m).Omega/Right.Mode(m).Omega;
  PreExp := Sqrt(IntPower(2,nModes)*Sqrt(PreExp)*Determinant(Q1));
  if Print then
    WriteLn(Log,'PreExp=', PreExp, ', ExpArg = ', I00, ', I00=', PreExp*Exp(I00));
  I00 := PreExp*Exp(I00);

  RDelta.PreMultiply(R);
  if Print then
    RDelta.Write(Log,'R delta');
  Q1.Scale(2);
  Q1.AddDiagonal(-1);
  if Print then
    Q1.Write(Log,'2Q-1');
  P1.Scale(2);
  P1.AddDiagonal(-1);
  if Print then
    P1.Write(Log,'2P-1');
  J.Free;
end;

type
  TVibrationalBasisArray = array of TVibrationalBasisState;

{ Profiling Variables }
var
 FCFMatrixTime: string = 'Franck Condon Factors';

function TMultidimensionalFCF.TransitionMatrix(BraSample,KetSample: TVibrationalBasisState; Cache: Boolean):TMatrix;
{ This method relies on the vibrational quantum numbers begin sorted as follows:
    000
    001
    002
    010
    011
    012
    020
    ...
    For non-totally symmetric modes, there will be members in the sequence missing from a particular basis;
    "OriginalPosition" in the basis state indicates where in the complete sequence the basis lies, and thus some values will
     typically be missing
    The "Offset" value in the basis state indicates the row or column of TransitionMatrix result that
    the overlap integrals for that state lie; it runs continuously from 1, but is not necessarily in order
    because of the sorting of the basis state. This means that this routine must be called each time the basis
    state is changed.
}
{ TODO: Check cache cleared as required }
var
  BraOffsets, KetOffsets: TIntegerArray;
  OrderedBra, OrderedKet: TVibrationalBasisArray;
  EvLeft, EvRight: TEigenvalues;
  Ovr: TMatrix;
  BraState, KetState: TElectronicState;
  CurrentCacheIndex: Integer;
  BraCopy, KetCopy: TIntegerArray;
  function CacheIndex(v: TIntegerArray):Integer;
  var
    CacheSpan, vn, MaxSym, Sym: Integer;
  begin
    Result := 0;
    Sym := 0;
    MaxSym := 0;
    for vn := Molecule.nModes - 1 downto 0 do
      if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
        CacheSpan := Min(BraState.Mode(vn).vMax, KetState.Mode(vn).vMax)+1;
        Result := Result*CacheSpan + Abs(v[Molecule.ModeInfo[vn].dIndex]);
      end else begin
        { TODO: The stuff below will require revsion for non-abelian groups }
        MaxSym := MaxSym or Molecule.ModeInfo[vn].Symmetry;
        if Odd(v[vn]) then
          Sym := Sym xor Molecule.ModeInfo[vn].Symmetry;
      end;
    Result := Result*(MaxSym+1) + Sym;
  end;
  function PrepareOffsets(var Offsets: TIntegerArray; var OrderedStates: TVibrationalBasisArray; Ev:TEigenvalues; AState: TElectronicState):Integer;
  var
    i, o: Integer;
    State, oState: TVibrationalBasisState;
  begin
    SetLength(Offsets, Molecule.nModes+1);
    o := 1;
    for i := Molecule.nModes downto 1 do begin
      Offsets[i] := o;
      o := o*(AState.Mode(i-1).vmax+1);
    end;
    Offsets[0] := o;
    SetLength(OrderedStates, o+1);
    Result := 1;
    for i := 1 to High(Ev.Basis) do begin
      State := Ev.Basis[i] as TVibrationalBasisState;
      if (State.State = AState) and (CacheIndex(State.v)=CurrentCacheIndex) then begin
        oState := OrderedStates[State.OriginalPosition];
        if not Assigned(oState) then begin
          OrderedStates[State.OriginalPosition] := State;
          State.FOffset := Result;
          Inc(Result);
        end else begin
          { Basis state with the same set of vibrational quantum numbers,
            but different electronic ones }
          State.FOffset := oState.FOffset;
        end;
      end;
    end;
    Dec(Result);
  end;
  procedure SaveOffsets(out Offsets: TIntegerArray; SourceSample: TVibrationalBasisState);
  var
    i, o: Integer;
    SourceState: TElectronicState;
    Ev: TEigenValues;
    Basis: TVibrationalBasisState;
  begin
    { Save mapping between original position in standrad order and index into FC array }
    SourceState := SourceSample.State as TElectronicState;
    o := 1;
    for i := Molecule.nModes downto 1 do begin
      o := o*(SourceState.Mode(i-1).vmax+1);
    end;
    SetLength(Offsets, o+1);
    Ev := SourceSample.Owner as TEigenvalues;
    for i := 1 to High(Ev.Basis) do begin
      Basis := Ev.Basis[i] as TVibrationalBasisState;
      if (Basis.State = SourceState) and (CacheIndex(Basis.v)=CurrentCacheIndex) then
        Offsets[Basis.OriginalPosition] := Basis.Offset;
    end;
  end;
  procedure SetOffsets(const Offsets: TIntegerArray; State: TState; Dest: TEigenValues);
  var
    i: Integer;
    Basis: TVibrationalBasisState;
  begin
    for i := 1 to High(Dest.Basis) do begin
      Basis := Dest.Basis[i] as TVibrationalBasisState;
      if (Basis.State = State) and (CacheIndex(Basis.v)=CurrentCacheIndex) then
        Basis.FOffset := Offsets[Basis.OriginalPosition];
    end;
  end;
//  procedure CopyOffsets(SourceSample: TVibrationalBasisState; Dest: TEigenvalues);
//  var
//    i, o: Integer;
//    OrderedStates: TVibrationalBasisArray;
//    SourceState: TElectronicState;
//    Basis: TVibrationalBasisState;
//    Ev: TEigenValues;
//  begin
//   { Copy offsets from one vibrational basis to another;
//      number of vibrational states should be the same, but ordering may differ }
//    SourceState := SourceSample.State as TElectronicState;
//    o := 1;
//    for i := Molecule.nModes downto 1 do begin
//      o := o*(SourceState.Mode(i-1).vmax+1);
//    end;
//    SetLength(OrderedStates, o+1);
//    Ev := SourceSample.Owner as TEigenvalues;
//    for i := 1 to High(Ev.Basis) do begin
//      Basis := Ev.Basis[i] as TVibrationalBasisState;
//      if (Basis.State = SourceState) and (CacheIndex(Basis.v)=CurrentCacheIndex) then
//        OrderedStates[Basis.OriginalPosition] := Basis;
//    end;
//    for i := 1 to High(Dest.Basis) do begin
//      Basis := Dest.Basis[i] as TVibrationalBasisState;
//      if (Basis.State = SourceState) and (CacheIndex(Basis.v)=CurrentCacheIndex) then
//        Basis.FOffset := OrderedStates[Basis.OriginalPosition].Offset;
//    end;
//  end;
  function FillOverlap(BB, KK: Integer): Double;
  var
    Bra, Ket: TVibrationalBasisState;
    Scale, Vm, Omega, OmegaP: Double;
    vn, vn1, Bn, Bnn, Kn, Knn, l, v, vp, v1: Integer;
    Reduced, Prepared: Boolean;
    procedure PrepareCall;
    begin
      if Prepared then
        Exit;
      { Make a copy of quantum number arrays so that values can be changes
        and therefore make routine threadsafe }
      Move(Bra.v[0],BraCopy[0],Length(BraCopy)*SizeOf(Integer));
      Move(Ket.v[0],KetCopy[0],Length(BraCopy)*SizeOf(Integer));
      Prepared := True;
    end;
  begin
    { TODO: Not thread safe }
    Bra := OrderedBra[BB];
    Ket := OrderedKet[KK];
    Result := 0;
    Reduced := False;
    Prepared := False;
    for vn := 0 to Molecule.nModes-1 do
      { Pick a mode to reduce }
      if Ket.v[vn] > 0 then
        if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
          { Special handling for reducing degenerate modes }
          { TODO: currently does not take account of mode mixing for degenerate modes }
          v := Ket.v[vn];
          l := Abs(Ket.v[Molecule.ModeInfo[vn].dIndex]);
          if v > l then begin
            vp := Bra.v[vn];
            Omega := KetState.Mode(vn).Omega;
            OmegaP := BraState.Mode(vn).Omega;
            Scale := Omega + OmegaP;
            { Reduce v }
            if v > 3 then begin
              Kn := Ket.OriginalPosition - 4*KetOffsets[vn+1];
              Vm := (Omega-OmegaP)/Scale*Sqrt((v-2-l)*(v-2+l)/(v-l)/(v+l));
              if Vm <> 0 then
                if Assigned(OrderedKet[Kn]) then
                  Result := Vm*Ovr[Bra.Offset,OrderedKet[Kn].Offset]
                else begin
                  PrepareCall;
                  KetCopy[vn] := Ket.v[vn]-4;
                  Result := Vm*Overlap(BraCopy,KetCopy);
                end;
            end;
            Kn := Ket.OriginalPosition - 2*KetOffsets[vn+1];
            if vp > 1 then begin
              Bn := Bra.OriginalPosition - 2*BraOffsets[vn+1];
              Vm := 2*Omega/Scale*Sqrt((vp-l)*(vp+l)/(v-l)/(v+l));
              if Vm <> 0 then
                if Assigned(OrderedBra[Bn]) and Assigned(OrderedKet[Kn]) then
                  Result := Result + Vm*Ovr[OrderedBra[Bn].Offset,OrderedKet[Kn].Offset]
                else begin
                  PrepareCall;
                  KetCopy[vn] := Ket.v[vn]-2;
                  BraCopy[vn] := Bra.v[vn]-2;
                  Result := Result + Vm*Overlap(BraCopy,KetCopy);
                end;
            end;
            Vm := 2*(OmegaP*(v-1)-Omega*(vp+1))/Scale/Sqrt((v-l)*(v+l));
            if Assigned(OrderedKet[Kn]) then
              Result := Result + Vm*Ovr[Bra.Offset,OrderedKet[Kn].Offset]
            else begin
              PrepareCall;
              KetCopy[vn] := Ket.v[vn]-2;
              Result := Result + Vm*Overlap(BraCopy,KetCopy);
            end;
            Reduced := True;
            Break;
          end else
            Continue; { This degnerate mode at minmum value for l }
        end else begin
          Scale := Ket.v[vn];
          Kn := Ket.OriginalPosition - KetOffsets[vn+1];
          for vn1 := 0 to Molecule.nModes-1 do
            if Bra.v[vn1] > 0 then begin
              if (vn=vn1) or (Molecule.ModeInfo[vn1].Symmetry = 0) then begin
                Vm := R[vn+1,vn1+1];
                if Vm <> 0 then begin
                  Bn := Bra.OriginalPosition - BraOffsets[vn1+1];
                  Vm := 2 * Vm * Sqrt(Bra.v[vn1]/Scale);
                  if Assigned(OrderedBra[Bn]) and Assigned(OrderedKet[Kn]) then begin
                    Result := Result + Vm*Ovr[OrderedBra[Bn].Offset,OrderedKet[Kn].Offset];
                  end else begin
                    PrepareCall;
                    KetCopy[vn] := Ket.v[vn]-1;
                    BraCopy[vn1] := Bra.v[vn1]-1;
                    Result := Result + Vm*Overlap(BraCopy,KetCopy);
                  end;
                end;
              end;
            end;
          for vn1 := 0 to Molecule.nModes-1 do begin
            v1 := Ket.v[vn1];
            if vn=vn1 then
              Dec(v1);
            if v1 > 0 then begin
              if (vn=vn1) or (Molecule.ModeInfo[vn1].Symmetry = 0) then begin
                Vm := Q1[vn+1,vn1+1];
                if Vm <> 0 then begin
                  Vm := Vm * Sqrt(v1/Scale);
                  Knn := Kn - KetOffsets[vn1+1];
                  if Assigned(OrderedKet[Knn]) then begin
                    Result := Result + Vm*Ovr[Bra.Offset,OrderedKet[Knn].Offset];
                  end else begin
                    PrepareCall;
                    KetCopy[vn] := Ket.v[vn]-1;
                    KetCopy[vn1] := v1-1; { This is correct for vn=vn1 case }
                    Result := Result + Vm*Overlap(BraCopy,KetCopy);
                  end;
                end;
              end;
            end;
          end;
          if Molecule.ModeInfo[vn].Symmetry = 0 then begin
            Vm := Rdelta[vn+1];
            if Vm <> 0 then begin
              Vm := - Vm * Sqrt(2/Scale);
              if Assigned(OrderedKet[Kn]) then
                Result := Result + Vm*Ovr[Bra.Offset,OrderedKet[Kn].Offset]
              else begin
                PrepareCall;
                KetCopy[vn] := Ket.v[vn]-1;
                Result := Result + Vm*Overlap(BraCopy,KetCopy);
              end;
            end;
          end;
          Reduced := True;
          Break;
        end;
    if not Reduced then
      { Ket all zero; reduce Bra }
      for vn := 0 to Molecule.nModes-1 do
        { Pick a mode to reduce }
        if Bra.v[vn] > 0 then
          if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
            { Special handling for reducing degenerate modes }
            vp := Bra.v[vn];
            l := Abs(Bra.v[Molecule.ModeInfo[vn].dIndex]);
            if vp > l then begin
              v := Ket.v[vn];
              Omega := KetState.Mode(vn).Omega;
              OmegaP := BraState.Mode(vn).Omega;
              Scale := OmegaP + Omega;
              { Reduce vp }
              if vp > 3 then begin
                Bn := Bra.OriginalPosition - 4*BraOffsets[vn+1];
                Vm := (OmegaP-Omega)/Scale*Sqrt((vp-2-l)*(vp-2+l)/(vp-l)/(vp+l));
                if Vm <> 0 then
                  if Assigned(OrderedBra[Bn]) then
                    Result := Vm*Ovr[OrderedBra[Bn].Offset,Ket.Offset]
                  else begin
                    PrepareCall;
                    BraCopy[vn] := Bra.v[vn]-4;
                    Result := Vm*Overlap(BraCopy,KetCopy);
                  end;
              end;
              Bn := Bra.OriginalPosition - 2*BraOffsets[vn+1];
              if v > 1 then begin
                Kn := Ket.OriginalPosition - 2*KetOffsets[vn+1];
                Vm := 2*OmegaP/Scale*Sqrt((v-l)*(v+l)/(vp-l)/(vp+l));
                if Vm <> 0 then
                  if Assigned(OrderedBra[Bn]) and Assigned(OrderedKet[Kn]) then
                    Result := Result + Vm*Ovr[OrderedBra[Bn].Offset,OrderedKet[Kn].Offset]
                  else begin
                    PrepareCall;
                    BraCopy[vn] := Bra.v[vn]-2;
                    KetCopy[vn] := Ket.v[vn]-2;
                    Result := Result + Vm*Overlap(BraCopy,KetCopy);
                  end;
              end;
              Vm := 2*(Omega*(vp-1)-OmegaP*(v+1))/Scale/Sqrt((vp-l)*(vp+l));
              if Assigned(OrderedBra[Bn]) then
                Result := Result + Vm*Ovr[OrderedBra[Bn].Offset,Ket.Offset]
              else begin
                PrepareCall;
                BraCopy[vn] := Bra.v[vn]-2;
                Result := Result + Vm*Overlap(BraCopy,KetCopy);
              end;
              Reduced := True;
              Break;
            end else
              Continue; { This degnerate mode at minmum value for l }
          end else begin
            Scale := Bra.v[vn];
            Bn := Bra.OriginalPosition - BraOffsets[vn+1];
            for vn1 := 0 to Molecule.nModes-1 do { Note that this is redundant code }
              if Ket.v[vn1] > 0 then begin
                if (vn=vn1) or (Molecule.ModeInfo[vn1].Symmetry = 0) then begin
                  Vm := R[vn+1,vn1+1];
                  if Vm <> 0 then begin
                    Vm := 2*Vm*Sqrt((Ket.v[vn1]+1)/Scale);
                    Kn := Ket.OriginalPosition - KetOffsets[vn1+1];
                    if Assigned(OrderedBra[Bn]) and Assigned(OrderedKet[Kn]) then
                      Result := Result + Vm*Ovr[OrderedBra[Bn].Offset,OrderedKet[Kn].Offset]
                    else begin
                      PrepareCall;
                      BraCopy[vn] := Bra.v[vn]-1;
                      KetCopy[vn1] := Ket.v[vn1]-1;
                      Result := Result + Vm*Overlap(BraCopy,KetCopy);
                    end;
                  end;
                end;
              end;
            for vn1 := 0 to Molecule.nModes-1 do begin
              v1 := Bra.v[vn1];
              if vn = vn1 then
                Dec(v1);
              if v1 > 0 then begin
                if (vn=vn1) or (Molecule.ModeInfo[vn1].Symmetry = 0) then begin
                  Vm := P1[vn+1,vn1+1];
                  if Vm <> 0 then begin
                    Vm := Vm * Sqrt(v1/Scale);
                    Bnn := Bn - BraOffsets[vn1+1];
                    if Assigned(OrderedBra[Bnn]) then
                      Result := Result + Vm*Ovr[OrderedBra[Bnn].Offset,Ket.Offset]
                    else begin
                      PrepareCall;
                      BraCopy[vn] := Bra.v[vn]-1;
                      BraCopy[vn1] := v1-1;
                      Result := Result + Vm*Overlap(BraCopy,KetCopy);
                    end;
                  end;
                end;
              end;
            end;
            if Molecule.ModeInfo[vn].Symmetry = 0 then begin
              Vm := P1Delta[vn+1];
              if Vm <> 0 then begin
                Vm := Vm*Sqrt(2/Scale);
                if Assigned(OrderedBra[Bn]) then
                  Result := Result + Vm*Ovr[OrderedBra[Bn].Offset,Ket.Offset]
                else begin
                  PrepareCall;
                  BraCopy[vn] := Bra.v[vn]-1;
                  Result := Result + Vm*Overlap(BraCopy,KetCopy);
                end;
              end;
            end;
            Reduced := True;
            Break;
          end;
    if not Reduced then begin
      Result := I00;
      for vn := 0 to Molecule.nModes-1 do
        { Special action needed for degenerate modes - these will have v' = v = l, rather than 0;
          in addition I00 is wrong, as mode should have been counted twice }
        if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
          l := Abs(Bra.v[Molecule.ModeInfo[vn].dIndex]);
          Omega := (Ket.State as TElectronicState).Mode(vn).Omega;
          OmegaP := (Bra.State as TElectronicState).Mode(vn).Omega;
          Result := Result * Sqrt(IntPower(2*Sqrt(Omega*Omegap)/(Omega+Omegap),2*l+1))
        end;
      end;
    Ovr[Bra.Offset,Ket.Offset] := Result;
//    if not SameValue(Result, Overlap(Bra.v,Ket.v)) then
    if IsNan(Result) then
      Raise Exception.Create('FCFOverlap');
  end;
var
  i, j, BraMax, Ketmax: Integer;
  CacheEntry: TCachedTransitionMatrix;
begin
  Result := nil;
  if Molecule.nModes = 0 then
    Exit;
  { Compute the vibrational overlap matrix; this depends on the vibrational quantum numbers only;
    it will have the form <v1',v2'...,l1,l2...|v1,v2,...,l1,l2>, i.e. diagonal in the angular momentum
    quantum numbers. The angular momentum quantum numbers are therfore used as the key to a cache.
    In the absence of mode mixing between the degenerate modes the sign of l should be unimportant,
    so each l can have the range 0..vmax. For the other modes the overall vibrational symmetry is
    used to determine the appropriate index }
  BraState := Bra as TElectronicState;
  KetState := Ket as TElectronicState;
  CurrentCacheIndex := CacheIndex(BraSample.v);
  if Cache then begin
    { The cache is implicitly a multidimensional array indexed by Abs(l1), Abs(l2), ... but
      collapse to a one dimensional array using fortran style indexing }
    if Assigned(CacheLock) then
      CacheLock.BeginRead;
    if (CurrentCacheIndex < Length(FMatrixCache)) and Assigned(FMatrixCache[CurrentCacheIndex]) then
      CacheEntry := FMatrixCache[CurrentCacheIndex]
    else
      CacheEntry := nil;
    if Assigned(CacheLock) then
      CacheLock.EndRead;
    if not Assigned(CacheEntry) then
      try
        if Assigned(CacheLock) then
          CacheLock.BeginWrite;
        if not((CurrentCacheIndex < Length(FMatrixCache)) and Assigned(FMatrixCache[CurrentCacheIndex])) then begin
          CacheEntry := TCachedTransitionMatrix.Create(nil);
          SetLength(FMatrixCache,Max(Length(FMatrixCache),CurrentCacheIndex+1));
          FMatrixCache[CurrentCacheIndex] := CacheEntry;
          CacheEntry.BraSample := BraSample;
          CacheEntry.KetSample := KetSample;
          CacheEntry.TransitionMoment := Self;
        end;
      finally
        if Assigned(CacheLock) then
          CacheLock.EndWrite;
      end;
    CacheEntry := FMatrixCache[CurrentCacheIndex];
    CacheEntry.Make(True);
    Result := FMatrixCache[CurrentCacheIndex].TransitionMatrix;
    if Length(CacheEntry.BraOffsets) = 0 then begin
      { Matrix has just been made }
      SaveOffsets(CacheEntry.BraOffsets,BraSample);
      SaveOffsets(CacheEntry.KetOffsets,KetSample);
    end;
    { If offsets not set, then same basis used for more than one symmetry }
    { TODO: Can fail if BraSample/KetSample are deallocated, as by TranInfo }
    if BraSample.Offset = 0 then begin
//      WriteLn(Log,'Bra: CacheIndex='+IntToStr(CurrentCacheIndex));
//      PrepareOffsets(BraOffsets, OrderedBra, BraSample.Owner as TEigenvalues, BraState);
//      CopyOffsets(CacheEntry.BraSample, BraSample.Owner as TEigenvalues); { TODO: Remove this routine }
      SetOffsets(CacheEntry.BraOffsets, BraSample.State, BraSample.Owner as TEigenvalues);
//      with BraSample.Owner as TEigenvalues do
//        for i := 1 to High(Basis) do begin
//          Write(Log, i:4, (Basis[i] as TVibrationalBasisState).Offset:4, ' ', Basis[i].BasisLabel);
//          if (Basis[i] as TVibrationalBasisState).Offset <> 0 then
//            WriteLn(Log, ' ', Result[(Basis[i] as TVibrationalBasisState).Offset,1])
//          else
//            WriteLn(Log);
//        end;
    end;
    if KetSample.Offset = 0 then begin
//      WriteLn(Log,'Ket: CacheIndex='+IntToStr(CurrentCacheIndex));
//      PrepareOffsets(KetOffsets, OrderedKet, KetSample.Owner as TEigenvalues, KetState);
//      CopyOffsets(CacheEntry.KetSample, KetSample.Owner as TEigenvalues);
      SetOffsets(CacheEntry.KetOffsets, KetSample.State, KetSample.Owner as TEigenvalues);
//      with BraSample.Owner as TEigenvalues do
//        for i := 1 to High(Basis) do begin
//          Write(Log, i:4, (Basis[i] as TVibrationalBasisState).Offset:4, ' ', Basis[i].BasisLabel);
//          if (Basis[i] as TVibrationalBasisState).Offset <> 0 then
//            WriteLn(Log, ' ', Result[1,(Basis[i] as TVibrationalBasisState).Offset])
//          else
//            WriteLn(Log);
//        end;
    end;
    Exit;
  end;
  TStart(FCFMatrixTime);
  EvLeft := BraSample.Owner as TEigenvalues;
  EvRight := KetSample.Owner as TEigenvalues;
  BraMax := PrepareOffsets(BraOffsets, OrderedBra, EvLeft, BraState);
  KetMax := PrepareOffsets(KetOffsets, OrderedKet, EvRight, KetState);
  SetLength(BraCopy,Length(BraSample.v));
  SetLength(KetCopy,Length(KetSample.v));
  Ovr := TMatrix.Create(BraMax, KetMax);
  { Now fill matrix, starting from smallest quantum numbers }
  PushProgress('Franck Condon factors', High(OrderedBra));
  try
    for i := 1 to High(OrderedBra) do
      { Is this basis state present? }
      if Assigned(OrderedBra[i]) then begin
        SetProgress(i);
        for j := 1 to High(OrderedKet) do
          if Assigned(OrderedKet[j]) then
            FillOverlap(i,j);
      end;
  finally
    PopProgress;
  end;
  Result := Ovr;
  TEnd(FCFMatrixTime);
//  Result.Write(Log,'CacheIndex='+IntToStr(CurrentCacheIndex),6);
//  for i := 0 to High(BraOffsets) do
//    WriteLn(Log,i:4, BraOffsets[i]:8);
//  for i := 0 to High(OrderedBra) do
//    if Assigned(OrderedBra[i]) then
//      WriteLn(Log, i:4, OrderedBra[i].Offset:4, ' ', OrderedBra[i].BasisLabel, ' ', Result[OrderedBra[i].Offset,1]);
//  for i := 0 to High(KetOffsets) do
//    WriteLn(Log,i:4, KetOffsets[i]:8);
//  for i := 0 to High(OrderedKet) do
//    if Assigned(OrderedKet[i]) then
//      WriteLn(Log, i:4, OrderedKet[i].Offset:4, ' ', OrderedKet[i].BasisLabel, ' ', Result[1,OrderedKet[i].Offset]);
end;

procedure TMultidimensionalFCF.PrintInfo(Sender: TObject);
var
  nModes: Integer;
  DD: TColumnVector;
  C: TMatrix;
  procedure MakeGamma(var G, Gs: TMatrix; State: TElectronicState);
  var
    i: Integer;
  begin
    G := TMatrix.Create(nModes,nModes);
    G.Zero;
    Gs := TMatrix.Create(G);
    for i := 1 to nModes do begin
      G[i,i] := State.Mode(i-1).Omega*ReduceFrequency;
      Gs[i,i] := Sqrt(G[i,i]);
    end;
  end;
{$IFDEF TEST}
  function FCF(m, n: Integer):Double; { FCF from Chen paper }
  var
    p, q, r: Integer;
    S: Double;
    fac: array of Double;
  begin  { Chen Eq 19 for testing only }
    SetLength(fac, Max(m,n)+1);
    fac[0] := 1;
    for p := 1 to High(fac) do
      fac[p] := fac[p-1] * p;
    S := 0;
    for p := 0 to Min(m,n) do
      for q := 0 to (m-p) div 2 do
        for r := 0 to (n-p) div 2 do
          S := S + IntPower(2,p)*IntPower(DD[1],m-2*q-p)*IntPower(DD[2],n-2*r-p)
                   * IntPower(C[1,1],q) * IntPower(C[2,2],r) * IntPower(C[1,2],p)
                   /(Fac[p]*Fac[q]*Fac[r]*Fac[m-2*q-p]*Fac[n-2*r-p]);
    Result := S * Sqrt(Fac[m]*Fac[n]/IntPower(2,m+n));
  end;
{$ENDIF}
var
  Left, Right: TElectronicState;
  Strim, SS, GammaP, Gamma, sGammaP, sGamma, JSJ, T, T1: TMatrix;
  k: TColumnVector;
  m, i: Integer;
  PreExp, ExpArg: Double;
  {$IFDEF TEST}
  bb, kk: array of Integer;
  {$ENDIF}
begin
  ReShowLogForm;
  PrepareMatrices(True,False,True);
  Left := Bra as TElectronicState;
  Right := Ket as TElectronicState;
  nModes := Left.Molecule.nModes;

  { Show Ruhoff style info }
  MakeGamma(GammaP, sGammaP, Right);
  GammaP.Write(Log,'GammaP');
  MakeGamma(Gamma, sGamma, Left);
  Gamma.Write(Log,'Gamma');
  Strim := TMatrix.Create(S);
  Strim.Resize(nModes, nModes); { Discard rotations and vibrations }
  SS := Strim.Transpose; { J in Ruhoff notation }
  SS.Write(Log, 'Jt');
  SS.PostMultiply(GammaP);
  SS.Write(Log, 'S = Jt GammaP');
  SS.PostMultiply(Strim);
  for m := 1 to nModes do
    SS[m,m] := SS[m,m] + Gamma[m,m];
  SS.Write(Log, 'S = Jt GammaP J + Gamma');

  PreExp := 1;
  for m := 1 to nModes do
    PreExp := PreExp * Gamma[m,m] * GammaP[m,m];
  T := TMatrix.Create(Strim);
  T.PostMultiply(SS);
  PreExp := Sqrt(PreExp)/Determinant(T);
  PreExp := Sqrt(PreExp * IntPower(2,nModes));
  T.Free;

  Invert(SS);
  SS.Write(Log, 'S^-1 = (Jt GammaP J + Gamma)^-1');
  JSJ := TMatrix.Create(Strim);
  JSJ.PostMultiply(SS);
  T := Strim.Transpose;
  JSJ.PostMultiply(T);
  JSJ.Write(Log, 'J S^-1 Jt');

  T.Assign(GammaP);
  T.PostMultiply(JSJ);
  for m := 1 to nModes do
    T[m,m] := T[m,m] - 1;
  T.Write(Log, 'GammaP J S^-1 Jt - 1');
  T.PostMultiply(GammaP);
  T.Write(Log, '(GammaP J S^-1 Jt - 1)GammaP');
  k := TColumnVector.Create(d);
  k.PreMultiply(T);
  ExpArg := k.DotProduct(d)/2;
  WriteLn(Log,'PreExp=', PreExp, ', ExpArg=', ExpArg, ', I00=', PreExp*Exp(ExpArg));

  T.Assign(sGammaP);
  T.PostMultiply(JSJ);
  T.PostMultiply(sGammaP);
  for m := 1 to nModes do begin
    for i := 1 to nModes do
      T[m,i] := 2*T[m,i];
    T[m,m] := T[m,m] - 1;
  end;
  T.Write(Log, 'A');

  T.Assign(JSJ);
  T.PostMultiply(GammaP);
  for m := 1 to nModes do
    T[m,m] := 1 - T[m,m];
  k.Assign(d);
  k.PreMultiply(T);
  k.PreMultiply(sGamma);
  k.Scale(2);
  k.Write(Log, 'b=');

  C := TMatrix.Create(sGamma);
  C.PostMultiply(SS);
  C.PostMultiply(sGamma);
  for m := 1 to nModes do begin
    for i := 1 to nModes do
      C[m,i] := 2*C[m,i];
    C[m,m] := C[m,m] - 1;
  end;
  C.Write(Log, 'C');

  DD := TColumnVector.Create(d);
  DD.PreMultiply(GammaP);
  T.Free;
  T := Strim.Transpose;
  DD.PreMultiply(T);
  DD.PreMultiply(SS);
  DD.PreMultiply(sGamma);
  DD.Scale(-2);
  DD.Write(Log, 'd=');

  T.Assign(sGamma);
  T.PostMultiply(SS);
  T1 := Strim.Transpose;
  T.PostMultiply(T1);
  T1.Free;
  T.PostMultiply(sGammaP);
  T.Write(Log, 'E');

{$IFDEF TEST}
  PreExp := PreExp*Exp(ExpArg);
  T.RowCount := 5;
  T.ColCount := 5;
  if nModes > 1 then begin
    for m := 1 to T.RowCount do
      for i := 1 to T.RowCount do
        T[m,i] := FCF(m-1,i-1)*PreExp;
    T.Write(Log, 'FCF');
    SetLength(bb,2);
    SetLength(kk,2);
    kk[0] := 0; kk[1] := 0;
    for m := 1 to T.RowCount do begin
      bb[0] := m-1;
      for i := 1 to T.RowCount do begin
        bb[1] := i-1;
        T[m,i] := Overlap(bb,kk);
      end;
    end;
    T.Write(Log, '<n''n|00>');
    for m := 1 to T.RowCount do begin
      bb[0] := m-1;
      for i := 1 to T.RowCount do begin
        bb[1] := i-1;
        T[m,i] := Overlap(kk,bb);
      end;
    end;
    T.Write(Log, '<00|n''n>');
    bb[1] := 0;
  end else begin
    SetLength(bb,1);
    SetLength(kk,1);
  end;

  for m := 1 to T.RowCount do begin
    bb[0] := m-1;
    for i := 1 to T.RowCount do begin
      kk[0] := i-1;
      T[i,m] := Overlap(kk,bb);
    end;
  end;
  T.Write(Log, '<v''|v>');
{$ENDIF}

  DD.Free; C.Free; Strim.Free;
  SS.Free; GammaP.Free; Gamma.Free; sGammaP.Free; sGamma.Free; JSJ.Free;

  k.Free;
  T.Free
end;

function TMultidimensionalFCF.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
begin
  Result := inherited ReadProperties(Node, CheckUnread);
  if (RootNode as TMixture).ReadVersion < 7000106 then
    Strength := Strength * Sqrt(3);
end;

procedure TMultidimensionalFCF.SetSettingFromStr(AIndex: Integer; const Str: string);
begin
  if AIndex = 1 then
    if CompareText(Str, 'Auto') = 0 then
      Symmetry := -1
    else
      Symmetry := PointGroupInfo[Molecule.PointGroup].StrToSymmetry(Str)
  else
    inherited;
end;

function TMultidimensionalFCF.SettingToStr(AIndex: Integer): string;
begin
  if AIndex = 1 then
    if Symmetry < 0 then
      Result := 'Auto'
    else
      Result := PointGroupInfo[Molecule.PointGroup].SymmetryToStr(Symmetry)
  else
    Result := inherited SettingToStr(AIndex);
end;

procedure TMultidimensionalFCF.ShowDushinsky(Sender: TObject);
begin
  {$IFNDEF CONSOLE}
  ConnectDushinskyForm(Self);
  {$ENDIF}
end;

procedure TMultidimensionalFCF.SwapModes(nMode1, nMode2: Integer);
var
  T: Double;
  i, f: Integer;
  Mode1, Mode2: TVibrationalModeChange;
begin
  if Abs(nMode1-nMode2) <> 1 then
    raise NotYet; { This routine will only exchange adjacent modes }
  Mode1 := Mode(nMode1);
  Mode2 := Mode(nMode2);
  { Given modes have been exchanged, so move correspondign values also }
  T := Mode1.Displacement;
  Mode1.Displacement := Mode2.Displacement;
  Mode2.Displacement := T;
  { Permute rotations }
  { TODO: may require adjustment to give mathematically equivalent results }
  f := Mode1.FirstDynamicParameter;
  for i := 0 to Min(nMode1,nMode2) - 1 do begin
    T := Mode1.Parameters[f+i];
    Mode1.Parameters[f+i] := Mode2.Parameters[f+i];
    Mode2.Parameters[f+i] := T;
  end;
  for i := Max(nMode1,nMode2)+1 to Molecule.nModes-1 do begin
    T := Mode(i).Parameters[f+nMode1];
    Mode(i).Parameters[f+nMode1] := Mode(i).Parameters[f+nMode2];
    Mode(i).Parameters[f+nMode2] := T;
  end;
end;

{ TBaseVibrationalMode }

constructor TBaseVibrationalMode.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if AOwner is TElectronicState then
    TElectronicState(AOwner).AdjustModes(False)
  else
    (AOwner as TMultidimensionalFCF).AdjustModes(False);
  { Symmetry should be a per molecule setting, but is actually
    in the per state object. Copy from the per molecule setting on
    initialisation }
  ModifySetting(1, (AOwner.Owner.Owner as TVibratingMolecule).ModeInfo[NodeIndex].Symmetry, False);
end;

function TBaseVibrationalMode.FixupProperties: Exception;
begin
  { Copy back symmetry from global variable }
  ModifySetting(1, Molecule.ModeInfo[NodeIndex].Symmetry, False);
  Result := inherited FixUpProperties;
end;

function TBaseVibrationalMode.GetName: TComponentName;
begin
  Result := Molecule.ModeInfo[NodeIndex].Name;
end;

function TBaseVibrationalMode.Molecule: TVibratingMolecule;
begin
  Result := Owner.Owner.Owner as TVibratingMolecule;
end;

function TBaseVibrationalMode.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  i: Integer;
begin
  Result := inherited ReadProperties(Node, False);
  for i := FirstDynamicParameter to nParameters do
    ReadFloatableValue(Node,i,ParameterNames[i],'0');
  if CheckUnread then
    Result := Result + nUnreadProperties(Node)
end;

procedure TBaseVibrationalMode.SetName(const Value: TComponentName);
var
  i: Integer;
begin
  for i := 1 to Length(Value) do
    if not (Value[i] in ['0'..'9','a'..'z','A'..'Z','_','(',')',' ','''','"']) then
      raise EReportException.Create('Invalid character in vibrational mode name:'+Value[i]);
  inherited;
  if NodeIndex >= 0 then
    Molecule.ModeInfo[NodeIndex].Name := Value;
end;

procedure TBaseVibrationalMode.SetSetting(AIndex: Integer; const AValue: Integer);
var
  Changed: Boolean;
  procedure CopySetting(Node: TMNode; MyIndex: Integer);
  var
    i: Integer;
  begin
    if Node is TBaseVibrationalMode then
      with TBaseVibrationalMode(Node) do
        if NodeIndex = MyIndex then
          ModifySetting(AIndex, AValue, True);
    for i := 0 to Node.ComponentCount - 1 do
      CopySetting(Node.Components[i] as TMNode, MyIndex);
  end;
begin
  inherited SetSetting(AIndex, AValue);
  { See notes in Create method; set per molecule settting also }
  if NodeIndex < 0 then
    Exit;
  Changed := False;
  case AIndex of
    1: begin
         Changed := Molecule.ModeInfo[NodeIndex].Symmetry <> AValue;
         Molecule.ModeInfo[NodeIndex].Symmetry := AValue;
       end;
  end;
  if not Changed then
    Exit;
  { As GetSetting is not virtual, changes must be copied to all other copies of this mode }
  CopySetting(Molecule, NodeIndex);
end;

function TBaseVibrationalMode.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  if Aindex = 1 then begin
    LowestValue := 0;
    Result := PointGroupInfo[Molecule.PointGroup].SettingsNames;
  end else
    Result := inherited GetSettingsNames(AIndex, LowestValue);
end;

function TBaseVibrationalMode.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Symmetry of this vibrational mode. Note that this setting is shared with every state in the molecule.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

procedure TBaseVibrationalMode.SetSettingFromStr(AIndex: Integer; const Str: string);
begin
  if AIndex = 1 then
    Symmetry := PointGroupInfo[Molecule.PointGroup].StrToSymmetry(Str)
  else
    inherited;
end;

function TBaseVibrationalMode.SettingToStr(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := PointGroupInfo[Molecule.PointGroup].SymmetryToStr(Symmetry)
  else
    Result := inherited SettingToStr(AIndex);
end;

{ TVibrationalModeChange }

procedure TVibrationalModeChange.AdjustParameterNames;
var
  i: Integer;
begin
  { Add/remove rotation matrix parameters as required }
  for i := nParameters+1 to FirstDynamicParameter + NodeIndex - 1 do
    AddParameter(Molecule.ModeInfo[i-FirstDynamicParameter].Name);
  for i := nParameters downto FirstDynamicParameter + NodeIndex do
    DeleteParameter(i);
  for i := FirstDynamicParameter to nParameters do
    SetParameterName(i, Molecule.ModeInfo[i-FirstDynamicSetting].Name);
end;

function TVibrationalModeChange.CanShift(Shift: Integer): Boolean;
begin
  Result := False;
end;

function TVibrationalModeChange.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex > 1 then
    Result := 'Mixing of this mode with ' + (Owner as TMultidimensionalFCF).Mode(AIndex-1).Name + ' /degrees'
  else if AIndex = 1 then
    Result := 'Displacement along normal mode between the two electronic states'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

{ TPointGroupInfo }

constructor TPointGroupInfo.Create(APointGroup: TPointGroup);
begin
  FPointGroup := APointGroup;
  Name := PropertyToStr(Self,'PointGroup');
  PointGroupInfo[PointGroup] := Self;
  FHasDegenerateModes := False;
end;

procedure FreePointGroupInfo;
var
  I: TPointGroup;
begin
  for I := Low(TPointGroup) to High(TPointGroup) do
    PointGroupInfo[I].Free;
end;

function TPointGroupInfo.StrToSymmetry(Str: string): Integer;
begin
  if not TryStrToSymmetry(Str, Result) then
    raise Exception.Create('Symmetry ' + Str + ' not valid for ' + Name);
end;

function TPointGroupInfo.SymmetryToStr(Symmetry: Integer): string;
var
  P: PPChar;
begin
  if (Symmetry < 0) or (Symmetry > LastSymmetry) then
    raise Exception.CreateFmt('Bad Symmetry for %s: %d', [Name,Symmetry]);
  P := SettingsNames;
  Inc(P,Symmetry);
  Result := P^;
end;

function TPointGroupInfo.TryStrToSymmetry(Str: string; out Symmetry: Integer): Boolean;
var
  P: PPChar;
begin
  P := SettingsNames;
  Symmetry := 0;
  while P^ <> nil do begin
    if CompareText(P^,Str) = 0 then begin
      Result := True;
      Exit;
    end;
    Inc(P);
    Inc(Symmetry);
  end;
  Result := False;
  Symmetry := -1;
end;

const
  LambdaNames: array [0..7] of PChar = ('Sigma+', 'Sigma-', 'Pi', 'Delta', 'Phi', 'Gamma', 'H', 'I');

{ TCinfinityv }

type
  TCinfinityv = class(TPointGroupInfo)
  private
    FIrepNames: array of string;
  public
    constructor Create(APointGroup: TPointGroup);
    function TryStrToSymmetry(Str: string; out Symmetry:Integer):Boolean; override;
    function SymmetryToStr(Symmetry:Integer):string; override;
    function Allowed(Bra, Dipole, Ket: Integer):Boolean; override;
  end;

function TCinfinityv.Allowed(Bra, Dipole, Ket: Integer): Boolean;
begin
  if Bra <= 1 then
    if Ket <= 1 then { Sigma - Sigma }
      Result := (Dipole <= 1) and (Bra xor Ket xor Dipole = 0)
    else if Dipole <= 1 then { Sigma Sigma not Sigma }
      Result := False
    else
      Result := Ket = Dipole { Sigma not Sigma not Sigma }
  else if Ket <= 1 then
    if Dipole <= 1 then
      Result := False { not Sigma Sigma Sigma }
    else
      { not Sigma not Sigma Sigma }
      Result := Bra = Dipole
  else if Dipole <= 1 then
    Result := Bra = Ket { not Sigma Sigma not Sigma }
  else { not Sigma not Sigma not Sigma }
    Result := (Abs(Bra-Dipole+1) = Ket) or (Abs(Bra+Dipole-1) = Ket);
end;

constructor TCinfinityv.Create(APointGroup: TPointGroup);
var
  i: Integer;
begin
  inherited;
  SetLength(FIrepNames, (Ord(APointGroup)-Ord(CInfinityv)+1)*Length(LambdaNames)+4);
  for i := 1 to High(FIrepNames) - 1 do
    FIRepNames[i] := SymmetryToStr(i-1);
  FIrepNames[0] := 'Auto';
  SettingsNames := @FIrepNames[1];
  FLastSymmetry := High(Integer);
  FHasDegenerateModes := True;
end;

function TCinfinityv.SymmetryToStr(Symmetry: Integer): string;
begin
  if Symmetry <= High(LambdaNames) then
    Result := LambdaNames[Symmetry]
  else
    Result := QuantumNumberToStr(2*Symmetry,qftJ);
end;

function TCinfinityv.TryStrToSymmetry(Str: string; out Symmetry: Integer): Boolean;
var
  i: Integer;
begin
  Result := True;
  for i := 0 to High(LambdaNames) do
    if CompareText(Str,LambdaNames[i]) = 0 then begin
      Symmetry := i;
      Exit;
    end;
  if TryStrToQuantumNumber(Str,Symmetry,qftJ) then begin
    Symmetry := Symmetry div 2;
    Exit;
  end;
  Result := False;
end;

{ TDInfinityh }

type
  TDInfinityh = class(TCinfinityv)
    function TryStrToSymmetry(Str: string; out Symmetry:Integer):Boolean; override;
    function SymmetryToStr(Symmetry:Integer):string; override;
    function Allowed(Bra, Dipole, Ket: Integer):Boolean; override;
  end;

function TDInfinityh.Allowed(Bra, Dipole, Ket: Integer): Boolean;
begin
  Result := not Odd(Bra xor Dipole xor Ket);
  if Result then
    Result := inherited Allowed(Bra div 2, Dipole div 2,Ket div 2);
end;

function TDInfinityh.SymmetryToStr(Symmetry: Integer): string;
begin
  Result := inherited SymmetryToStr(Symmetry div 2);
  if Odd(Symmetry) then
    Result := Result + 'u'
  else
    Result := Result + 'g'
end;

function TDInfinityh.TryStrToSymmetry(Str: string; out Symmetry: Integer): Boolean;
var
  ug: Integer;
begin
  Result := False;
  Symmetry := -1;
  if Length(Str) < 2 then
    Exit;
  case Str[Length(Str)] of
    'g', 'G': ug := 0;
    'u', 'U': ug := 1;
  else
    Exit;
  end;
  Delete(Str,Length(Str),1);
  Result := inherited TryStrToSymmetry(Str,Symmetry);
  if Result then
    Symmetry := Symmetry*2 + ug;
end;

type
  TAbelianPointGroupInfo = class(TPointGroupInfo)
    constructor Create(APointGroup: TPointGroup);
    function Allowed(Bra, Dipole, Ket: Integer):Boolean; override;
  end;

{ TAbelianPointGroupInfo }

function TAbelianPointGroupInfo.Allowed(Bra, Dipole, Ket: Integer): Boolean;
begin
  Result := Bra xor Dipole xor Ket = 0;
end;

constructor TAbelianPointGroupInfo.Create(APointGroup: TPointGroup);
var
  P: PPChar;
begin
  inherited Create(APointGroup);
  SettingsNames := IrepNames(TAbelianPointGroup(PointGroup));
  P := SettingsNames;
  FLastSymmetry := -1;
  while P^ <> nil do begin
    Inc(P);
    Inc(FLastSymmetry);
  end;
end;

{ TCachedTransitionMatrix }

destructor TCachedTransitionMatrix.Destroy;
begin
  FreeAndNil(TransitionMatrix);
  inherited;
end;

procedure TCachedTransitionMatrix.DoMake(RequiredAttributes: Pointer);
begin
  TransitionMatrix := (TransitionMoment as TMultidimensionalFCF).TransitionMatrix(BraSample,KetSample,False)
end;

function TCachedTransitionMatrix.HasCorrectAttributes(Attributes: Pointer): Boolean;
begin
  Result := Assigned(TransitionMatrix);
end;

initialization
  TVibratingMolecule.RegisterClass(TSpecies);
  TVibrationalManifold.RegisterClass(TVibratingMolecule);
  TElectronicState.RegisterClass(TVibrationalManifold);
  TMolecularNucleus.RegisterClass(TElectronicState);
  TVibrationalMode.RegisterClass(TElectronicState);
  TMultidimensionalFCF.RegisterClass(TTransitionMoments);
  TVibrationalModeChange.RegisterClass(TMultidimensionalFCF);
  TVibrationalPerturbation.RegisterClass(TVibrationalManifold);

  ReduceFrequency := 4 * Pi * Pi * SpeedOfLight * 1e-21 / Planck / Avagadro;

  TAbelianPointGroupInfo.Create(C1);
  TAbelianPointGroupInfo.Create(Ci);
  TAbelianPointGroupInfo.Create(C2);
  TAbelianPointGroupInfo.Create(Cs);
  TAbelianPointGroupInfo.Create(D2);
  TAbelianPointGroupInfo.Create(C2v);
  TAbelianPointGroupInfo.Create(C2h);
  TAbelianPointGroupInfo.Create(D2h);
  TCinfinityv.Create(CInfinityv);
  TDInfinityh.Create(DInfinityh);

  elVariables := MakeSortedVariables(['Lambda','Omega'], [@eLambda, @eOmega]);

finalization
  FreePointGroupInfo;
  FreeAndNil(vVariables);
  FreeAndNil(lVariables);
  FreeAndNil(elVariables);
end.
