unit Vib2D;
{ Various support functions for 2D harmonic oscillators }

{$I DFS.inc}{Copyright (C) 2008 C M Western; see commandline.pas for license}

interface

function Overlap(l: Integer; Omegap, Omega: Double):Double; overload;
function Overlap(vp, v, l: Integer; Omegap, Omega: Double):Double; overload;

implementation

uses Math, WignerD, SysUtils;

{ Various ways of working out 2D Franck-Condon factors from
  E V Doktorov, I A Malkin and V I Man'ko
  J Phys. B: Atom. Molec. Phys. 9, 507 (1976) }

function Overlap(l: Integer; Omegap, Omega: Double):Double; overload;
begin
  { Equation 2.3a of Doktorov et al for v' = v = l }
  Result := IntPower(2*Sqrt(Omega*Omegap)/(Omega+Omegap),l+1)
end;

function Overlap(vp, v, l: Integer; Omegap, Omega: Double):Double; overload;
begin
  { Equation 2.5 of Doktorov et al }
  Result := 0;
  if v > l then begin
    { Reduce v }
    if v > 3 then
      Result := (Omega-Omegap)/(Omega+Omegap)*Sqrt((v-2-l)*(v-2+l)/(v-l)/(v+l))*Overlap(vp,v-4,l,Omegap,Omega);
    if vp > 1 then
      Result := Result + 2*Omega/(Omega+Omegap)*Sqrt((vp-l)*(vp+l)/(v-l)/(v+l))*Overlap(vp-2,v-2,l,Omegap,Omega);
    Result := Result + 2*(Omegap*(v-1)-Omega*(vp+1))/(Omega+Omegap)/Sqrt((v-l)*(v+l))*Overlap(vp,v-2,l,Omegap,Omega);
  end else if vp > l then begin
    { Reduce vp }
    if vp > 3 then
      Result := (Omegap-Omega)/(Omegap+Omega)*Sqrt((vp-2-l)*(vp-2+l)/(vp-l)/(vp+l))*Overlap(vp-4,v,l,Omegap,Omega);
    if v > 1 then
      Result := Result + 2*Omegap/(Omegap+Omega)*Sqrt((v-l)*(v+l)/(vp-l)/(vp+l))*Overlap(vp-2,v-2,l,Omegap,Omega);
    Result := Result + 2*(Omega*(vp-1)-Omegap*(v+1))/(Omegap+Omega)/Sqrt((vp-l)*(vp+l))*Overlap(vp-2,v,l,Omegap,Omega);
  end else
    { v' = v = l }
    Result := Overlap(l,Omegap,Omega);
end;

{$IFDEF NOTWORKING}
{ This version fails - it produces numbers > 1 and inconsistent with the other
  methods, implying the paper is wrong }
function OverlapRR(vp, v, l: Integer; Omegap, Omega: Double):Double; overload;
begin
  { Equation 2.6 of Doktorov et al }
  Result := 0;
  if v > l then begin
    { Reduce v }
    if v > 3 then
      Result := -Sqrt((v-2-l)*(v-2+l)/(v-l)/(v+l))*OverlapRR(vp,v-4,l,Omegap,Omega);
    Result := Result + 4*((Sqr(Omega)+Sqr(Omegap))*(v-1)-Omega*Omegap*(vp+1))
                       /(Sqr(Omega)-Sqr(Omegap))/Sqrt((v-l)*(v+l)) *OverlapRR(vp,v-2,l,Omegap,Omega);
  end else if vp > l then begin
    { Reduce vp }
    if vp > 3 then
      Result := -Sqrt((vp-2-l)*(vp-2+l)/(vp-l)/(vp+l))*OverlapRR(vp-4,v,l,Omegap,Omega);
    Result := Result + 4*(Omega*Omegap*(v+1)-(Sqr(Omega)+Sqr(Omegap))*(vp-1))
                       /(Sqr(Omega)-Sqr(Omegap))/Sqrt((vp-l)*(vp+l)) *OverlapRR(vp-2,v,l,Omegap,Omega);
  end else
    { v' = v = l }
    Result := Overlap(l,Omegap,Omega);
end;
{$ENDIF}

{$IFDEF TEST}
function OverlapW(vp, v, l: Integer; Omegap, Omega: Double):Double; overload;
var
  CosHalfTheta: Double;
begin
  { Note that symmetry requires selection rules of Delta l = 0, Delta v even }
  { Equation 2.3b of Doktorov et al }
  CosHalfTheta := 2*Sqrt(Omega*Omegap)/(Omega+Omegap);
  Result := CosHalfTheta * djmnb((vp+v) div 2, l + (v-vp) div 2, l - (v-vp) div 2, 2*ArcCos(CosHalfTheta) );
end;

var
  vp, v, l: Integer;
  a, b: Double;
  F: Text;
initialization
  AssignFile(F,'wibble.txt');
  Rewrite(F);
  for vp := 0 to 5 do begin
    for v := 0 to 5 do
      Write(F, ' ', Overlap(vp*2,v*2,0,305.7792116,51.33078804));
    WriteLn(F);
  end;
  Close(F);
  a := Overlap(2,0,0,305.7792116,51.33078804);
  b := Overlap(0,2,0,305.7792116,51.33078804);
  a := Overlap(2,2,0,305.7792116,51.33078804);
  b := Overlap(0,4,0,305.7792116,51.33078804);
  b := Overlap(4,0,0,305.7792116,51.33078804);
  for vp := 0 to 10 do begin
    a := Overlap(vp,100,123);
    b := Overlap(vp,vp,vp,100,123);
    if not SameValue(a,b) then
      raise Exception.CreateFmt('Help %d',[vp]);
    for v := 0 to 10 do
      if not Odd(vp-v) then
        for l := 0 to Min(v,vp) do
          if not Odd(v-l) then begin
            a := Overlap(vp,v,l,100,123);
            b := OverlapW(vp,v,l,100,123);
            if not SameValue(a,b) then
              raise Exception.CreateFmt('Help %d %d %d',[vp, v, l]);
          end;
  end;
{$ENDIF}
end.
