unit Interp;
{ based on routines from Gnuplot }
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

type
  TVector2 = array of array [0..1] of Double;

function CubicSpline(const V:TVector2;
                     Closed: Boolean; { If Curve is closed - last and first point should be equal }
                     UnitX, UnitY: Double;
                     nInterp: Integer { Number of points to add between each value }
                     ): TVector2;

function BSpline(const V:TVector2;
                 Closed: Boolean; { If Curve is closed - last and first point should be equal }
                 nInterp: Integer; { Number of points to add between each value }
                 Order: Integer { Order of spline }
                 ): TVector2;

implementation

uses SysUtils;

type

  TTriDiag = array of array [0..2] of Double;

(*
 * The following two procedures solve the special linear system which arise
 * in cubic spline interpolation. If x is assumed cyclic ( x[i]=x[n+i] ) the
 * equations can be written as (i=0,1,...,n-1):
 *     m[i][0] * x[i-1] + m[i][1] * x[i] + m[i][2] * x[i+1] = b[i] .
 * In matrix notation one gets M * x = b, where the matrix M is tridiagonal
 * with additional elements in the upper right and lower left position:
 *   m[i][0] = M_{i,i-1}  for i=1,2,...,n-1    and    m[0][0] = M_{0,n-1} ,
 *   m[i][1] = M_{i, i }  for i=0,1,...,n-1
 *   m[i][2] = M_{i,i+1}  for i=0,1,...,n-2    and    m[n-1][2] = M_{n-1,0}.
 * M should be symmetric (m[i+1][0]=m[i][2]) and positiv definite.
 * The size of the system is given in n (n>=1).
 *
 * In the first procedure the Cholesky decomposition M = C^T * D * C
 * (C is upper triangle with unit diagonal, D is diagonal) is calculated.
 * Return TRUE if decomposition exist.
 *)
function SolveCubic1(var m: TTriDiag): Boolean;
var
  i, n: Integer;
  m_ij, m_n, m_nn, d: Double;
begin
  n := Length(m);
  Result := False;
  if n < 1 then
    Exit; { Dimension should be at least 1 }

  d := m[0][1];		{ D_(0,0) = M_(0,0) }
  if d <= 0 then
    Exit;		{ M (or D) should be positiv definite }
  m_n := m[0][0];		{  M_(0,n-1)  }
  m_nn := m[n - 1][1];		{ M_(n-1,n-1) }
  for i := 0 to n - 3 do begin
      m_ij := m[i][2];		         { M_(i,1)  }
      m[i][2] := m_ij / d;	         { C_(i,i+1) }
      m[i][0] := m_n / d;	         { C_(i,n-1) }
      m_nn := m_nn - m[i][0] * m_n;      { to get C_(n-1,n-1) }
      m_n := -m[i][2] * m_n;	         { to get C_(i+1,n-1) }
      d := m[i + 1][1] - m[i][2] * m_ij; { D_(i+1,i+1) }
      if d <= 0 then
        Exit;	{ Elements of D should be positive }
      m[i + 1][1] := d;
  end;
  if n >= 2 then begin		 { Complete last column }
      m_n := m_n + m[n - 2][2];  { add M_(n-2,n-1) }
      m[n - 2][0] := m_n / d;	 { C_(n-2,n-1) }
      d := m_nn - m[n - 2][0] * m_n;	{ D_(n-1,n-1) }
      m[n - 1][1] := d;
      if d <= 0 then
        Exit;
  end;
  Result := True;
end;

(*
 * The second procedure solves the linear system, with the Choleky
 * decomposition calculated above (in m[][]) and the right side b given
 * in x[]. The solution x overwrites the right side in x[].
 *)
procedure SolveCubic2(const m: TTriDiag; var x: array of Double);
var
  i, n: Integer;
  x_n: Double;
begin
  n := Length(m);
  { Division by transpose of C : b = C^(-T) * b }
  x_n := x[n - 1];
  for i := 0 to n - 3 do begin
    x[i + 1] := x[i + 1] - m[i][2] * x[i];  { C_(i,i+1) * x_(i) }
    x_n := x_n - m[i][0] * x[i];	    { C_(i,n-1) * x_(i) }
  end;
  if n >= 2 then
    x[n - 1] := x_n - m[n - 2][0] * x[n - 2];	{ C_(n-2,n-1) * x_(n-1) }

  { Division by D: b = D^(-1) * b }
  for i := 0 to n - 1 do
    x[i] := x[i] / m[i][1];

  { Division by C: b = C^(-1) * b }
  x_n := x[n - 1];
  if n >= 2 then
      x[n - 2] := x[n - 2] - m[n - 2][0] * x_n;	{ C_(n-2,n-1) * x_(n-1) }
  for i := n - 3 downto 0 do
      {      C_(i,i+1) * x_(i+1) + C_(i,n-1) * x_(n-1) }
      x[i] := x[i] - m[i][2] * x[i + 1] + m[i][0] * x_n;
end;

{ Find second derivatives (x''(t_i),y''(t_i)) of cubic spline interpolation
  through list of points (x_i,y_i). The parameter t is calculated as the
  length of the linear stroke. The number of points must be at least 3.
   Note: For closed contours the first and last point must be equal. }
function GenerateCubicSpline(const V: TVector2; Closed:Boolean;
                             var DeltaT, D2x, D2y: array of Double;
                             UnitX, UnitY:Double): Boolean;
var
  m: TTriDiag;
  np, n, i: Integer;
  Norm: Double;
begin
  Result := False;
  np := length(V);
  { Calculate first differences in (d2x[i], d2y[i]) and interval lengths in delta_t[i] }
  for i := 0 to np-2 do begin
    D2x[i] := V[i+1][0] - V[i][0];
    D2y[i] := V[i+1][1] - V[i][1];
    { The norm of a linear stroke is calculated in "normal coordinates"
       and used as interval length: }
    DeltaT[i] := Sqrt(Sqr(D2x[i] / UnitX) + Sqr(D2y[i] / UnitY));

    D2x[i] := D2x[i] / DeltaT[i];	{ first difference, with unit norm: }
    D2y[i] := D2y[i] / DeltaT[i];	{   || (d2x[i], d2y[i]) || = 1      }
  end;
  { Setup linear system:  m * x = b }
  n := Length(V)-2; { Without first and last point }
  if Closed then begin
    {First and last points must be equal for closed contours }
    DeltaT[np - 1] := DeltaT[0];
    D2x[np - 1] := D2x[0];
    D2y[np - 1] := D2y[0];
    Inc(n); {Add last point (= first point) }
  end;
  SetLength(m, n);
  for i := 0 to n - 1 do begin
    { Matrix M, mainly tridiagonal with cyclic second index ("j = j+n mod n") }
    m[i][0] := DeltaT[i];	                 { Off-diagonal element M_(i,i-1) }
    m[i][1] := 2. * (DeltaT[i] + DeltaT[i + 1]); { M_(i,i) }
    m[i][2] := DeltaT[i + 1];	                 { Off-diagonal element M_(i,i+1) }

    { Right side b_x and b_y }
    D2x[i] := (D2x[i + 1] - D2x[i]) * 6;
    D2y[i] := (D2y[i + 1] - D2y[i]) * 6;

    { If the linear stroke shows a cusps of more than 90 degree, the right
      side is reduced to avoid oscillations in the spline: }

    Norm := Sqrt(Sqr(D2x[i] / UnitX) + Sqr(D2y[i] / UnitY)) / 8.5;

    if Norm > 1 then begin
      D2x[i] := D2x[i] / Norm;
      D2y[i] := D2y[i] / Norm;
      { The first derivative will not be continuous }
    end;
  end;
  if not Closed then begin
    { Third derivative is set to zero at both ends }
    m[0][1] := m[0][1] + m[0][0];	      { M_(0,)     }
    m[0][0] := 0;		              { M_(0,n-1)   }
    m[n - 1][1] := m[n - 1][1] + m[n - 1][2]; { M_(n-1,n-1) }
    m[n - 1][2] := 0;	                      { M_(n-1,0)   }
  end;

  { Solve linear systems for d2x[] and d2y[] }

  if SolveCubic1(m) { Calculate Cholesky decomposition } then begin
    SolveCubic2(m, d2x);	{ solve M * d2x = b_x }
    SolveCubic2(m, d2y);	{ solve M * d2y = b_y }
  end else
    Exit;			{ Should not happen, but who knows ... }

  { Shift all second derivatives one place right and abdate end points }
  for i := n downto 1 do begin
    D2x[i] := D2x[i - 1];
    D2y[i] := D2y[i - 1];
  end;
  if Closed then begin
    D2x[0] := D2x[n];
    D2y[0] := D2y[n];
  end else begin
    D2x[0] := D2x[1];	{ Third derivative is zero in }
    D2y[0] := d2y[1];	{     first and last interval }
    D2x[n + 1] := D2x[n];
    D2y[n + 1] := D2y[n];
  end;
  Result := True;
end;

(*
 * Calculate interpolated values of the spline function (defined via p_cntr
 * and the second derivatives d2x[] and d2y[]). The number of tabulated
 * values is n. On an equidistant grid n_intpol values are calculated.
 *)
function InterpolateCubicSpline(const V:TVector2; d2x, d2y, delta_t: array of Double; nInterp:Integer):TVector2;
var
  i, iV, iR, n: Integer;
  t, t_skip, t_max: Double;
  x0, x1, y0, y1: Double;
  d, hx, dx0, dx01, hy, dy0, dy01: Double;
begin
  n := Length(d2x);
  { The length of the total interval }
  t_max := 0;
  for i := 0 to n - 2 do
    t_max := t_max + delta_t[i];

  { The distance between interpolated points }
  t_skip := (1 - 1e-7) * t_max / (nInterp - 1);

  t := 0;			{ Parameter value }
  x1 := V[0][0];
  y1 := V[0][1];
  SetLength(Result, nInterp+20); { Leave room }
  Result[0] := V[0];
  iV := 0;
  iR := 1;
  t := t + t_skip;

  for i := 0 to n - 2 do begin
    d := delta_t[i];		{ Interval length }
    x0 := x1;
    y0 := y1;
    Inc(iV);
    x1 := V[iV][0];
    y1 := V[iV][1];
    hx := (x1 - x0) / d;
    hy := (y1 - y0) / d;
    dx0 := (d2x[i + 1] + 2 * d2x[i]) / 6;
    dy0 := (d2y[i + 1] + 2 * d2y[i]) / 6;
    dx01 := (d2x[i + 1] - d2x[i]) / (6 * d);
    dy01 := (d2y[i + 1] - d2y[i]) / (6 * d);
    while t <= delta_t[i] do begin	{ t in current interval ? }
      if iR >= Length(Result) then
        SetLength(Result, iR+100);
      Result[iR][0] := x0 + t * (hx + (t - d) * (dx0 + t * dx01));
      Result[iR][1] := y0 + t * (hy + (t - d) * (dy0 + t * dy01));
      Inc(iR, 1);
      t := t + t_skip;
    end;
    t := t - delta_t[i];	{ Parameter t relative to start of next interval }
  end;
  SetLength(Result, iR);
end;

{ Generate a cubic spline curve through the points (x_i,y_i)
  The spline is defined as a nd-function s(t) = (x(t),y(t)), where the
  parameter t is the length of the linear stroke. }
function CubicSpline(const V:TVector2;
                     Closed: Boolean; { If Curve is closed - last and first point should be equal }
                     UnitX, UnitY: Double;
                     nInterp: Integer { Number of points to add between each value }
                     ): TVector2;
var
  n: Integer;
  DeltaT, D2x, D2y: array of Double;
begin
  n := Length(V);
  SetLength(DeltaT, n); { Interval length }
  SetLength(D2x, n);    { Second derivatives }
  SetLength(D2y, n);
  if n > 2 then begin
    if not GenerateCubicSpline(V, Closed, DeltaT, D2x, D2y, UnitX, UnitY) then begin
      Result := V; { Leave unchanged }
      Exit;
    end;
  end else if n = 2 then begin
    DeltaT[0] := 1;
    D2x[0] := 0;
    D2x[1] := 0;
    D2y[0] := 0;
    D2y[1] := 0;
  end else begin
    Result := V;
    Exit;
  end;
  Result := InterpolateCubicSpline(V, D2x, D2y, DeltaT, 1 + (n-1)*nInterp);
end;

(*
 * Find Bspline approximation for this data set.
 * Algorithm: using deBoor algorithm
 * Uses nInterp to determine number of samples per
 * interval, where the knot vector intervals are assumed to be uniform
 * Note: if Curvekind is open contour than Open end knot vector is assumed,
 *       else (closed contour) Float end knot vector is assumed.
 *)
function BSpline(const V:TVector2;
                 Closed: Boolean; { If Curve is closed - last and first point should be equal }
                 nInterp: Integer; { Number of points to add between each value }
                 Order: Integer { Order of Bspline - 1 }
                 ): TVector2;
var
  n, nEff, knot_index, iV, nPoints: Integer;
  t, next_t, dt: Double;
  dx, dy: array of Double;
  { Routine to get the i knot from uniform knot vector. The knot vector
    might be float (Knot(i) = i) or open (where the first and last "order"
    knots are equal). Closed determines knot kind - open contour means
    open knot vector, and closed contour selects float knot vector.
    Note the knot vector is not exist and this routine simulates it existance
    Also note the indexes for the knot vector starts from 0. }
  function fetch_knot(i: Integer):Double;
  begin
    if not Closed then begin
      if i <= order then
        Result := 0
      else if i <= nEff then
        Result := i - order
      else
        Result := nEff - order;
    end else
      Result := i;
  end;
   { The routine to evaluate the B-spline value at the current point, t,
    using knot vector from function fetch_knot(), and the control points starting
    from V[iV]. Adds the generated point to the overall result. }
  procedure EvalBspline;
  var
    i, iiV, p: Integer;
    ti, tikp: Double;
  begin
    iiV := iV;
    { Set the dx/dy - [0] iteration step, control points (p=0 iterat.): }
    for i := knot_index - order to knot_index do begin
      dx[i] := V[iiV][0];
      dy[i] := V[iiV][1];
      Inc(iiV);
      if iiV = n then
        iiV := 1; { wrap, skipping duplicate point }
    end;

    for p := 1 to order do begin { Iteration (b-spline level) counter. }
      for i := knot_index downto knot_index - order + p do begin	{ Control points indexing. }
        ti := fetch_knot(i);
        tikp := fetch_knot(i + order + 1 - p);
        if ti <> tikp then begin { Should be different but just in case... }
          dx[i] := dx[i] * (t - ti) / (tikp - ti) +	{ Calculate x. }
                   dx[i - 1] * (tikp - t) / (tikp - ti);
          dy[i] := dy[i] * (t - ti) / (tikp - ti) +	{ Calculate y. }
                   dy[i - 1] * (tikp - t) / (tikp - ti);
        end;
      end;
    end;
    Result[nPoints][0] := dx[knot_index];
    Result[nPoints][1] := dy[knot_index];
    Inc(nPoints);
  end;
begin
  n := Length(V);
  if n < 2 then begin
    Result := V;
    Exit;
  end;
  if Order > n - 1 then { Order must be less than number of points in curve }
    Order := n - 1;
  nEff := n;
  { If the contour is Closed increase the effective number of points
    by order-1 so contour will be closed }
  if Closed then
    Inc(nEff, order-1);
  { Workspace }
  SetLength(dx, nEff+1);
  SetLength(dy, nEff+1);
  { Find first (t_min) and last (t_max) t value to eval: }
  t := fetch_knot(order);
  next_t := t + 1.0;
  knot_index := order;
  dt := 1 / (nInterp+1);{ Number of points per one section. }
  iV := 0;              { Index for input values }
  nPoints := (nInterp+1) * (nEff - order) + 1;
  if not Closed then
    Dec(nPoints);
  SetLength(Result, nPoints);
  nPoints := 0;         { Size of result so far }
  while nPoints < Length(Result) do begin
    if t > next_t then begin
      { Next order ctrl. pt. to blend. }
      Inc(iV);
      if iV = n then
        iV := 1; { wrap, skipping duplicate point }
      Inc(knot_index);
      next_t := next_t + 1;
    end;
    EvalBspline;
    t := t + dt;
  end;
end;

end.
