unit LapackLSQ;
{$I DFS.inc}{ Copyright (C) 2010 C M Western; see UCompat.pas for license}
interface

uses MatrixVec;

{ Interface to LAPACK least squares routines }

function dgels(
  A: TMatrix; { On entry, the M-by-N matrix A.
                On exit,
                  if M >= N, A is overwritten by details of its QR
                             factorization as returned by DGEQRF;
                  if M <  N, A is overwritten by details of its LQ
                             factorization as returned by DGELQF. }
  B :TColumnVector; { Can also be TMatrix;
                      On entry, the matrix B of right hand side vectors, stored
                      columnwise; B is M-by-NRHS if TRANS = 'N', or N-by-NRHS if TRANS = 'T'.
                      On exit, if INFO = 0, B is overwritten by the solution vectors, stored columnwise:
                      if TRANS = 'N' and m >= n, rows 1 to n of B contain the least squares solution vectors;
                      the residual sum of squares for the solution in each column is given by the sum of squares of elements N+1 to M in that column;
                      if TRANS = 'N' and m < n, rows 1 to N of B contain the minimum norm solution vectors;
                      if TRANS = 'T' and m >= n, rows 1 to M of B contain the minimum norm solution vectors;
                      if TRANS = 'T' and m < n, rows 1 to M of B contain the least squares solution vectors;
                      the residual sum of squares for the solution in each column is given by the sum of squares of elements M+1 to N in that column. }
  Trans: Char = 'N' { = 'N': the linear system involves A;
                      = 'T': the linear system involves A**T. }
  ): Integer;
{ Result: = 0:  successful exit
          < 0:  if INFO = -i, the i-th argument had an illegal value
          > 0:  if INFO =  i, the i-th diagonal element of the
                 triangular factor of A is zero, so that A does not have
                 full rank; the least squares solution could not be computed. }

function dgelsd(
  A: TMatrix; { On entry, the M-by-N matrix A. On exit, A has been destroyed. }
  B :TColumnVector  { Can also be TMatrix;
                      On entry, the M-by-NRHS right hand side matrix B.
                      On exit, B is overwritten by the N-by-NRHS solution matrix X.
                      If m >= n and RANK = n, the residual sum-of-squares for the
                      solution in the i-th column is given by the sum of squares of elements n+1:m in that column. }
): Integer;
{ Result : = 0:  successful exit
           < 0:  if INFO = -i, the i-th argument had an illegal value.
           > 0:  the algorithm for computing the SVD failed to converge;
                 if INFO = i, i off-diagonal elements of an intermediate
                bidiagonal form did not converge to zero. }

function dtrtrs(uplo, trans, diag: Char; A: TMatrix; B: TColumnVector):Integer;

implementation

uses Math, Lapack;

function dtrtrs(uplo, trans, diag: Char; A: TMatrix; B: TColumnVector):Integer;
var
  N, Nrhs, LDA: Integer;
begin
  Nrhs := 1;
  N := B.Size;
  if B is TMatrix then begin
    NRhs := TMatrix(B).ColCount;
    N := TMatrix(B).RowCount;
  end;
  LDA := A.RowCount;
  dtrtrs_(uplo, trans, diag, N, Nrhs, A.FirstItem^, LDA, B.FirstItem^, N, Result, 1, 1, 1);
end;

function dgels(
  A: TMatrix; { On entry, the M-by-N matrix A.
                On exit,
                  if M >= N, A is overwritten by details of its QR
                             factorization as returned by DGEQRF;
                  if M <  N, A is overwritten by details of its LQ
                             factorization as returned by DGELQF. }
  B :TColumnVector; { Can also be TMatrix;
                      On entry, the matrix B of right hand side vectors, stored
                      columnwise; B is M-by-NRHS if TRANS = 'N', or N-by-NRHS if TRANS = 'T'.
                      On exit, if INFO = 0, B is overwritten by the solution vectors, stored columnwise:
                      if TRANS = 'N' and m >= n, rows 1 to n of B contain the least squares solution vectors;
                      the residual sum of squares for the solution in each column is given by the sum of squares of elements N+1 to M in that column;
                      if TRANS = 'N' and m < n, rows 1 to N of B contain the minimum norm solution vectors;
                      if TRANS = 'T' and m >= n, rows 1 to M of B contain the minimum norm solution vectors;
                      if TRANS = 'T' and m < n, rows 1 to M of B contain the least squares solution vectors;
                      the residual sum of squares for the solution in each column is given by the sum of squares of elements M+1 to N in that column. }
  Trans: Char = 'N' { = 'N': the linear system involves A;
                      = 'T': the linear system involves A**T. }
  ): Integer;
var
  M, N, Nrhs, LDB: Integer;
  AWork: Double;
  Work: array of Double;
  Lwork: Integer;
begin
  M := A.RowCount;
  N := A.ColCount;
  Nrhs := 1;
  LDB := B.Size;
  if B is TMatrix then begin
    NRhs := TMatrix(B).ColCount;
    LDB := TMatrix(B).RowCount;
  end;
  Lwork := -1;
  dgels_(Trans, M, N, Nrhs, A.FirstItem^, M, B.FirstItem^, LDB, AWork, Lwork, Result, 1);
  if Result = 0 then begin
    Lwork := round(AWork);
    SetLength(Work, Lwork);
    dgels_(Trans, M, N, Nrhs, A.FirstItem^, M, B.FirstItem^, LDB, Work[0], Lwork, Result, 1);
  end;
end;

function dgelsd(
  A: TMatrix; { On entry, the M-by-N matrix A. On exit, A has been destroyed. }
  B :TColumnVector  { Can also be TMatrix;
                      On entry, the M-by-NRHS right hand side matrix B.
                      On exit, B is overwritten by the N-by-NRHS solution matrix X.
                      If m >= n and RANK = n, the residual sum-of-squares for the
                      solution in the i-th column is given by the sum of squares of elements n+1:m in that column. }
): Integer;
{ Result : = 0:  successful exit
           < 0:  if INFO = -i, the i-th argument had an illegal value.
           > 0:  the algorithm for computing the SVD failed to converge;
                 if INFO = i, i off-diagonal elements of an intermediate
                bidiagonal form did not converge to zero. }
var
  M, N, Nrhs, LDB: Integer;
  AWork, RCOND: Double;
  Work, S: array of Double;
  Lwork, Rank: Integer;
  Iwork: array of Integer;
begin
  M := A.RowCount;
  N := A.ColCount;
  Nrhs := 1;
  LDB := B.Size;
  if B is TMatrix then begin
    NRhs := TMatrix(B).ColCount;
    LDB := TMatrix(B).RowCount;
  end;
  Lwork := -1;
  SetLength(S, Min(M,N));
  RCOND := 0;
  SetLength(Iwork, Min(N,M)*(11+3*100)+100); { TODO: This is ony a guess }
  dgelsd_(M, N, Nrhs, A.FirstItem^, M, B.FirstItem^, LDB, S[0], RCOND, Rank, AWork, Lwork, Iwork[0], Result);
  if Result = 0 then begin
    Lwork := round(AWork);
    SetLength(Work, Lwork);
    dgelsd_(M, N, Nrhs, A.FirstItem^, M, B.FirstItem^, LDB, S[0], RCOND, Rank, Work[0], Lwork, Iwork[0], Result);
  end;
end;

end.

