unit WignerD;

{ Evaluation of the reduced Wigner d function,
   j
  d (beta)
   mn
  using the recursive method described in
  [1] "Accurate computation of the rotation matrices"
      Jorge Navaza Acta Cryst. A46, 619-620 (1990).
  
  See also
  [2] Stefano Trapania and Jorge Navaza
      Acta Cryst. A62, 262-269 (2006).
      
  Tested against DJMNB routine from CERN MATLIB;
  consistent (Max error < 1.3e-8 for J <= 25).
  This routine should be better, and gives much better
  orthogonality test results. }
  { TODO: Compare Phase conventions }

{$I DFS.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}

interface

uses QuantumNumber;

function djmnb(J, m, n: TQuantumNumber; Beta: Double): Double; overload;

implementation

uses Math;

function djm(J, m: TQuantumNumber; Beta: Double): Extended;
{ Special case evaluation of the n=j case }
var
  s, c: Extended;
  i, Jmm, Jpm: Integer;
begin
  { Note that Equation (3) in ref [1] appears to be wrong;
    Using Equation (16) of ref [2] }
  { Evaluate (2J)!/(J+m)!/(J-m)! taking cancelation into account to avoid overflow }
  Result := 1;
  Jmm := (J-m) div 2;
  for i := 2 to Jmm do
    Result := Result / i;
  Jpm := (J+m) div 2;
  for i := Jpm + 1 to J do
    Result := Result * i;
  sincos(Beta/2, s, c);
  Result := { Minus(J-m) * } Sqrt(Result) * IntPower(c, Jpm);
  if Jmm <> 0 then
    Result := Result * IntPower(s, Jmm); { Avoid 0^0 }
end;

function djmnb(J, m, n: TQuantumNumber; Beta: Double): Double;
var
  s, c, tn, tnp1, tnm1: Extended;
  ni: Integer;
begin
  { Use symmetry to find the shortest recursion length on n;
    this requires n as large as possible  }
  if Abs(n) < Abs(m) then begin
    { Swap arguments }
    Result := minus(n-m);
    ni := n;
    n := m;
    m := ni;
  end else
    Result := 1;
  { and n positive }
  if n < 0 then begin
    Result := Result * minus(n-m);
    n := -n;
    m := -m;
  end;
  tn := djm(J, m, Beta);
  if J > n then begin
    { Set up recusion relation }
    tnp1 := 0;
    sincos(Beta,s,c);
    if s = 0 then begin
      { Recursion relation fails; these are a multiple of pi,
        so treat as special cases }
      if c > 0 then
        { beta = 0 }
        if m = n then
          tn := 1
        else
          tn := 0
      else
        { beta = pi; note that this case seems difficult to trigger in practise }
        if m = n then
          tn := Minus(J+m)
        else
          tn := 0
    end else begin
      ni := J;
    { Note that Equation (5) in ref [1] appears to be wrong;
      Using Equation (15) of ref [2] with a sign changd (?sign convention change) }
      while n < ni do begin
        tnm1 :=  (-Sqrt((J+ni+2)*(J-ni))*tnp1 - 2*(m-ni*c)/s*tn) / Sqrt((J-ni+2)*(J+ni));
        Dec(ni,2);
        tnp1 := tn;
        tn := tnm1;
      end;
    end;
  end;
  Result := Result*tn;
end;

end.

