unit Worker;
{$I DFS.inc}{ Copyright (C) 2006 C M Western; see UCompat.pas for license}
{ Worker threads }
interface

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     Classes, SyncObjs, SysUtils, UCompat, TextAssign, Math;

type

  { TWorkerThread }

  TWorkerThread = class (TThread)
  private
    FNumber: Integer;
    FBusy: Boolean;
    FGoEvent: TEvent;
    PLog: PTextRec;
    FDone: TEvent;
    FExceptionMask: TFPUExceptionMask;
    procedure SetBusy(const Value: Boolean);
  protected
    procedure Execute; override;
    procedure DoWork; virtual; abstract;
    procedure DoTerminate; override;
  public
    LastException: string;
    constructor Create(CreateSuspended: Boolean; ANumber: Integer; var ALog: TextFile);
    destructor Destroy; override;
    procedure StartWork;
    procedure Discard(InDestroy:Boolean = False);
    property Number: Integer read FNumber;
    property Busy: Boolean read FBusy write SetBusy;
    property Done: TEvent read FDone;
  end;
  TWorkerClass = class of TWorkerThread;

procedure StartWorkers(nWorkers: Integer; WorkerClass: TWorkerClass);
function FindWorker:TWorkerThread;
function FindWorkerWait:TWorkerThread;
procedure WaitNoBusyWorkers;
function nWorkers:Integer;
procedure AtThreadEnd(Proc: TProcedure);

function SimpleWait(var Event: TEvent; Timeout: Cardinal): TWaitResult;

type
  TWaitFunction = function (var Event: TEvent; Timeout: Cardinal): TWaitResult;

var
  Aborted: Boolean = False;
  DoWait: TWaitFunction = {$IFDEF FPC}@{$ENDIF}SimpleWait; { Wait for event, with timeout in ms; variable so a different function can be used
                                         for the main thread of a gui app to process messages while doing so }

threadvar
  Log: Text; { Log output }

{$IFDEF UNIX}
const
  INFINITE = Cardinal(-1);
{$ENDIF}

function IsExpectedException(ExceptionType: TClass):Boolean;
procedure AddExpectedException(ExceptionType: ExceptClass);

implementation

uses {$IFDEF VCL_} {$IFNDEF CONSOLE} StackDump, {$ENDIF} {$ENDIF}
     Sysinfo, Profile, Clock;

var
  Stopping: string = 'Stopping';
  cc: string = 'cc';

var
  Workers: array of TWorkerThread;
  FnWorkers: Integer = 0;
  nFreeWorkers: Integer = 0;
  AnyWorkerFree: TEvent;
  AllWorkersFree: TEvent;
  BusyLock: {$IFDEF FPC}System.{$ENDIF}TRTLCriticalSection;

function nWorkers: Integer;
begin
  Result := FnWorkers;
end;

{$IFDEF UNIX}
// function sem_timedwait(var __sem: sem_t; var __abs_timeout: TTimeSpec):cint;cdecl;external libthreads;
{$ENDIF}

var
  Initialized: Boolean = False;

procedure InitWorker;
begin
  if not Initialized then begin
    AnyWorkerFree := TEvent.Create(nil,True,False,'');
    AllWorkersFree := TEvent.Create(nil,True,False,'');
    InitCriticalSection(BusyLock);
  end;
  Initialized := True;
end;

{ TWorkerThread }

constructor TWorkerThread.Create(CreateSuspended: Boolean; ANumber: Integer; var ALog: TextFile);
begin
  InitWorker;
  { FPC note: Stack size for worker threads should be the same size as the main thread }
  inherited Create(CreateSuspended{$IFDEF FPC},StackLength{$ENDIF});
  FExceptionMask := GetExceptionMask;
  FNumber := ANumber;
  PLog := PTextRec(@ALog);
  FGoEvent := TEvent.Create(nil,False,False,'');
  FBusy := True;
  FDone := TEvent.Create(nil,True,True,'');
  EnterCriticalSection(BusyLock);
  try
    InterLockedIncrement(FnWorkers);
    SetBusy(True);
  finally
    LeaveCriticalsection(BusyLock);
  end;
end;

destructor TWorkerThread.Destroy;
begin
  Discard(True);
  inherited;
  FGoEvent.Free;
  FDone.Free;
  InterLockedDecrement(FnWorkers);
  SetBusy(Busy); { Make sure Anyworker/AllWorkers in correct state }
end;

procedure MakeBigStack(n: Integer);
var
  a: array [0..50] of Double;
begin
  a[0] := 1;
  a[High(a)] := 1;
  if n > 0 then
    MakeBigStack(n-1);
end;
{$IFDEF OLD}
procedure MakeBigStack(n: Integer);
var
//  Eigenvals: TColumnVector;
//  Matrix: TMatrix;
//  i: Integer;
  a: array [0..10000] of Double;
begin
  a[0] := 1;
  a[High(a)] := n;
  n := n div 2;
//  Eigenvals := TColumnVector.Create(n);
//  Matrix := TMatrix.Create(n, n);
//  Matrix.Zero;
//  for i := 1 to n do begin
//    Matrix[i,i] := i;
//  end;
//  Eigen(Eigenvals, Matrix, True);
//  Eigenvals.Free;
//  Matrix.Free;
end;
{$ENDIF}

procedure TWorkerThread.Execute;
begin
  SetThreadName('Worker '+IntToStr(Number));
  AssignTextFile(Log,TextFile(PLog^));
  Rewrite(Log);
  SetExceptionMask(FExceptionMask);  { Copy exception mask from main thread }
  { TODO: This may no longer be required }
  try
    MakeBigStack(200);
  except
    on E:Exception do
      WriteLn(Log, E.Message);
    { ??? Needed to kickstart the stack expansion mechanism }
  end;
  repeat
    Busy := False;
    FGoEvent.WaitFor(INFINITE);
    if not Terminated then begin
      LastException := '';
      try
        DoWork;
      except
        on E:Exception do
          LastException := E.Message;
      end;
      if LastException <> '' then
        WriteLn(Log, '*** Exception ', LastException, ' in thread ', ThreadName);
    end;
  until Terminated;
  // WriteLn(Log, Clocks, ' Going: ', FnWorkers, ' ', nFreeWorkers, ' ', Terminated);
  CloseFile(Log);
  SetThreadName('');
end;

var
  ThreadEndProcs: array of TProcedure;

procedure AtThreadEnd(Proc: TProcedure);
begin
  SetLength(ThreadEndProcs, Length(ThreadEndProcs)+1);
  ThreadEndProcs[High(ThreadEndProcs)] := Proc;
end;

procedure TWorkerThread.DoTerminate;
var
  i: Integer;
begin
  for i := 0 to High(ThreadEndProcs) do
    ThreadEndProcs[i];
  ThreadName := '';
  inherited DoTerminate;
end;

procedure TWorkerThread.SetBusy(const Value: Boolean);
begin
  EnterCriticalSection(BusyLock);
  try
    if FBusy <> Value then
      if Value then begin
        FBusy := True;
        Dec(nFreeWorkers);
        Done.ResetEvent;
        if nFreeWorkers < 0 then
          raise Exception.Create('ChangeBusy-');
      end else begin
        FBusy := False;
        Inc(nFreeWorkers);
        Done.SetEvent;
        if nFreeWorkers > nWorkers then
          raise Exception.Create('ChangeBusy+');
      end;
    if nFreeWorkers = 0 then
      AnyWorkerFree.ResetEvent
    else
      AnyWorkerFree.SetEvent;
    if nFreeWorkers >= nWorkers then
      AllWorkersFree.SetEvent
    else
      AllWorkersFree.ResetEvent;
    //sem_getvalue(PSemaphore(AllWorkersFree.Handle)^, i1);
    //sem_getvalue(PSemaphore(AnyWorkerFree.Handle)^, i2);

    // Can only do this if writing to file:
    // WriteLn(log, Clocks, ' ', FnWorkers, ' ', nFreeWorkers, ' ', FBusy, ' ',ThreadName, ' All', i1, ' Any', i2);
  finally
    LeaveCriticalSection(BusyLock);
  end;
  // WriteLn(log, Clocks, ' ', FnWorkers, ' ', nFreeWorkers, ' ', FBusy, ' ',ThreadName, ' All', i1, ' Any', i2);
end;

procedure TWorkerThread.StartWork;
begin
  FGoEvent.SetEvent;
end;

procedure TWorkerThread.Discard(InDestroy:Boolean);
begin
  FreeOnTerminate := not InDestroy;
  EnterCriticalSection(BusyLock);
  try
    Terminate;
    if not Busy then begin
      Busy := True;
      StartWork; { If waiting (as it probably will be), stop the wait }
    end;
    FDone.SetEvent; { In case anyone is waiting }
  finally
    LeaveCriticalsection(BusyLock);
  end;
end;

procedure StartWorkers(nWorkers: Integer; WorkerClass: TWorkerClass);
{ Assumed called from main thread }
var
  OldnWorkers, I: Integer;
begin
  if nWorkers < 0 then
    nWorkers := nCPUs
  else if nWorkers < 2 then
    nWorkers := 0; { Only one worker is not useful }
  OldnWorkers := Length(Workers);
  if nWorkers > OldnWorkers then begin
    {$IFDEF FPC}
    // IsMultiThread := True;  { TODO: Remove? }
    {$ENDIF}
    SetLength(Workers, nWorkers);
    for I := OldnWorkers to nWorkers - 1 do begin
    //WriteLn(Log, Clocks, ' ', FnWorkers, ' ', nFreeWorkers);
        Workers[I] := WorkerClass.Create(True, I, Log);
    //WriteLn(Log, Clocks, ' ', FnWorkers, ' ', nFreeWorkers);
        Workers[I].Resume;
    //WriteLn(Log, Clocks, ' ', FnWorkers, ' ', nFreeWorkers);
    end;
    //WriteLn(Log, Clocks, ' ', FnWorkers, ' ', nFreeWorkers);
  end else if nWorkers < OldnWorkers then begin
    TStart(Stopping);
    //WriteLn(Log, Clocks, ' ', FnWorkers, ' ', nFreeWorkers);
    for I := OldnWorkers - 1 downto nWorkers do
      Workers[I].Discard;
    //WriteLn(Log, Clocks, ' ', FnWorkers, ' ', nFreeWorkers);
    AllWorkersFree.WaitFor(INFINITE);
    //WriteLn(Log, Clocks, ' ', FnWorkers, ' ', nFreeWorkers);
    EnterCriticalsection(BusyLock); { This ensures all the threads have left the BusyLock section, }
    LeaveCriticalsection(BusyLock); { so must be exiting SetBusy at the end of the Destroy method }
    //WriteLn(Log, Clocks, ' ', FnWorkers, ' ', nFreeWorkers);
    Sleep(1); { Allow a little more time for the threads to really finish - probably only an issue for leak detection }
    //WriteLn(Log, Clocks, ' ', FnWorkers, ' ', nFreeWorkers);
    SetLength(Workers, nWorkers);
    TEnd(Stopping);
  end;
end;

function FindWorker:TWorkerThread;
{ Assumed called from main thread }
var
  I: Integer;
begin
  if GetCurrentThreadID <> MainThreadID then
    raise Exception.Create('FindWorker called from worker');
  Result := nil;
  if Aborted then
    Exit;
  for I := 0 to High(Workers) do
    if not Workers[I].Busy then begin
      Result := Workers[I];
      Result.Busy := True;
      Exit;
    end;
end;

function SimpleWait(var Event: TEvent; Timeout: Cardinal): TWaitResult;
begin
  if not Assigned(Event) then
    Result := wrAbandoned
  else begin
    Timeout := INFINITE;
    Result := Event.WaitFor(Timeout);
    if Result = wrError then
      RaiseLastOSError;
  end;
end;

function FindWorkerWait:TWorkerThread;
begin
  while nFreeWorkers = 0 do
    if (DoWait(AnyWorkerFree, 1000) = wrSignaled) or Aborted then
      Break;
  Result := FindWorker;
  if not Assigned(Result) then begin
    { Don't try again, as must be called from main thread }
    raise Exception.Create('No worker found');
    // Result := FindWorkerWait;
  end;
end;

procedure WaitNoBusyWorkers;
//var
//  i1, i2: Integer;
begin
  //sem_getvalue(PSemaphore(AllWorkersFree.Handle)^, i1);
  //WriteLn(Log, Clocks, ' WNB: ', FnWorkers, ' ', nFreeWorkers, ' ', i1);
  while nFreeWorkers < nWorkers do begin
//    sem_getvalue(PSemaphore(AllWorkersFree.Handle)^, i1);
    DoWait(AllWorkersFree, 1000);
//    sem_getvalue(PSemaphore(AllWorkersFree.Handle)^, i2);
//    WriteLn(Log, Clocks, ' WNB: ', FnWorkers, ' ', nFreeWorkers, ' ', i1, '>', i2);
  end;
  //sem_getvalue(PSemaphore(AllWorkersFree.Handle)^, i1);
  //WriteLn(Log, Clocks, ' WNB: ', FnWorkers, ' ', nFreeWorkers, ' ', i1);
end;

{ TODO: May want to trap runerror or add an exit proc,
        as some thread failures call exit, rather than raise an exception }

var
  ExpectedExceptions: array of ExceptClass;

function IsExpectedException(ExceptionType: TClass):Boolean;
var
  i: Integer;
begin
  Result := True;
  for i := 0 to High(ExpectedExceptions) do
    if ExpectedExceptions[i] = ExceptionType then
      Exit;
  Result := false;
end;

procedure AddExpectedException(ExceptionType: ExceptClass);
begin
  if IsExpectedException(ExceptionType) then
    Exit;
  SetLength(ExpectedExceptions, Length(ExpectedExceptions)+1);
  ExpectedExceptions[High(ExpectedExceptions)] := ExceptionType;
end;

initialization
finalization
  if Initialized then begin
    StartWorkers(0,nil);
    FreeAndNil(AnyWorkerFree);
    FreeAndNil(AllWorkersFree);
    DoneCriticalSection(BusyLock);
  end;
end.

