unit Mixture;
{ Generic Types required for overall description of the spectrum of a mixture of compounds }

{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Graphics, Contnrs, Classes, SysUtils, pgDOM, Vector, MatrixVec, IntVec, Life,
     MNode, Units, QuantumNumber, DRintensity, Expr, MatrixElements, MNodeColour,
     Interlock, MathSup, QuantumMath, TypInfo, ComplexMatrix, StaticField;

const
  { Various special index values }
  TemperatureIndex = 3;
  GaussianIndex = 4;
  LorentzianIndex = 5;
  FoffsetIndex = 6;
  ShowSumIndex = 6;
  ShowPartsIndex = 7;
  ShowFortratIndex = 8;
  QuantumNumberFormatIndex = 6;
  LSP = 3; { First unused parameter index in TState }
  LSI = 4; { First unused setting index in TState }
  LMI = 7; { First unused setting index in TMolecule }
  LPI = 4; { First unused setting index in TPerturbation }
  LTI = 4; { First unused setting index in TBaseRotationalTransitionMoment }

type

  TEigenvalues = class;
  TBasisState = class;

  TLine = record
    Frequency: Double;
    Strength:  Double;
    Width:     Single;
    EigenvaluesU, EigenvaluesL: TEigenvalues; { nil for EigenvaluesU implies a population distribution }
    IndexU, IndexL: Integer;  { Index into Eigenvalues }
    Colour:    TColor;
    CachedName:string; { Not necessarily set; if blank call MakeTxName }
  end;

  TOnCheckLine = function (const Line: TLine): Boolean of Object;

  TParameter = record
    MNode: TMNode;
    Index: Integer;
    Value: Double;
    Increment: Double;
    StdDev: Double;
    Name: string;
    Inverse: Boolean;
  end;

  TFitType = (ftLine, ftContour, ftOverlaySum);

  TParameters = array of TParameter;

  TFitDetails = record
    FitType: TFitType;
    Parameters: TParameters;
    Correlation: TDoubleArrayArray;
  end;

  TMNodeWithParameters = class(TMNode)
  public
    { Descend from this if special handling for the following routine is required }
    procedure AddNodeParameters(MNode: TMNode; var FittingParameters: TParameters); overload;
    procedure AddNodeParameters(var FittingParameters: TParameters); overload; virtual;
  end;

  TMixtureMNode = class(TMNodeWithParameters)
    function BoltzmannConstant: Double; { k in units of cm-1/K or MHz/K }
    function Precision: Integer;
    function QuantumNumberToStr(AQuantumNumber: TQuantumNumber): string;
    function StrToQuantumNumber(s: string):TQuantumNumber;
    function QuantumNumberFormat: TQuantumNumberFormat;
  end;

  TStreamableMNode = class(TMixtureMNode)
  private
    FDirectoryName: string;
  public
    property DirectoryName: string read FDirectoryName;
    procedure LoadFromFile(AFileName: string); virtual;
    procedure LoadFromStream(Stream: TStream; AFileName: string = '<Stream>');
    procedure MergeFromStream(Stream: TStream; AFileName: string = '<Stream>');
    procedure SaveToFile(AFileName: string); virtual;
  end;

  TUnitsMNode = class(TStreamableMNode)
  protected
    procedure SetIntensityUnits(const Index: Integer; const AValue: TIntensityUnits);
    function GetIntensityUnits(const Index: Integer): TIntensityUnits;
    procedure SetUnits(const Index: Integer; const AValue: TResonanceUnits);
    function GetUnits(const Index: Integer): TResonanceUnits;
  public
    function UnitsName: string;
    function UnitsTypeName: string;
  end;

  TRootMNode = class(TUnitsMNode)
  protected
    FReadVersion: Integer; { Only valid during read; version set in input file }
  public
    procedure BeginPaste;
    property ReadVersion: Integer read FReadVersion; { Only valid during read; version set in input file }
    procedure PrepareForFit(FitType: TFitType; OtherNode: TRootMNode); virtual; abstract;
  end;

  TBasisOrder = (boDefault, boDiagonal, boStateDiagonal); { How to order the basis states }

  TMixture = class;     { Note logical hierarchy }
   TSimulation = class;
   TSpecies = class;
    TMolecule = class;
     TManifold = class;
      TState = class;

  TQuantumNumberReader = function (var a: Integer; const si: string):Boolean of object;

  TObsType = (obLine, obContour, obIntensity, obELower, obEUpper, obParameter, obCDLower, obCDUpper);
  TObsFlag = (ofCalculatedSet, ofRegenerateLabel, ofCD);
  TObsFlags = set of TObsFlag;

  TObservation = record
    Frequency:        Double; { Or intensity - see IntensityValue below }
    Weight:           Double;
    Calculated:       Double;
    ExField:          TStaticFields;
    Correlation:      TDoubleArray; { Correlation with following observations }
    ManifoldU, ManifoldL: TManifold;
    JU, JL:           TQuantumNumber;
    SymU, SymL:       Shortint;
    IndexU, IndexL:   Smallint; { State number - 0 to search; For obParameter ParameterIndex, StateIndex }
    AIndexU, AIndexL: Smallint; { Actual state number used }
    Name:             string;
    FileName:         string;  { Name of source file }
    LineNumber:       Integer;
    Colour:           TColor;  { Note that this can change following a fit if the colour depends on the State numbers }
    RemoveSpins:      Byte;    { Ignore this number of spins, starting with highest numbered nucleus }
    ObsType:          TObsType;{ Type of observation - never obContour }
    Flags:            TObsFlags;
    QuantumNumberFormat: TQuantumNumberFormat;
  end;
  PObservation = ^TObservation;

  TLineList = class;

  TSpectrum = class
  private
    FColor: TColor;
    FUnits: TResonanceUnits;
    FHeadingPrinted: Boolean;
    FIsDistribution: Boolean;
    procedure SetHeadingPrinted(const Value: Boolean);
  protected
    FnLines: Integer;
  public
    FMin: Double;
    FMax: Double;
    Temperature: Double;
    ITotal: Double; { TODO: Check/automate usage - this should depend on temperature }
    nPrinted: Integer;
    constructor Create(AUnits: TResonanceUnits; AIsDistribution: Boolean);
    procedure Clear; virtual;
    procedure AddLine(Frequency, Strength, Width: Double; AColour: TColor; AUnits: TResonanceUnits); overload; virtual;
    procedure AddLine(const Line: TLine; Simulation: TSimulation; AUnits: TResonanceUnits; var Problem: Exception); overload; virtual; abstract;
    procedure AddIn(LineList: TLineList; Simulation: TSimulation; var Problem: Exception; FirstLine: Integer = 0; Increment: Integer = 1); overload; virtual;
    procedure AddIn(LineList: TLineList); overload;
    procedure ShowFooter(var F: TextFile; PrintLevel: TPrintLevel);
    procedure ShowLine(var Line: TLine; var F: TextFile; PrintLevel: TPrintLevel; Threshold: Double; LinFormat: Boolean; var Problem: Exception; Footer: Boolean = False);
    property nLines: Integer read FnLines;
    property HeadingPrinted: Boolean read FHeadingPrinted write SetHeadingPrinted; { If heading for transtion list has been printed; clear to indicate scan though line list }
    property Colour: TColor read FColor write FColor;
    property Units: TResonanceUnits read FUnits write FUnits;
    property IsDistribution: Boolean read FIsDistribution;
  end;

  TLineList = class(TSpectrum)
  private
    FLines: array of TLine;
    function GetLine(AIndex: Integer): TLine;
  public
    Selected: Boolean;
    procedure AddLine(const ALine: TLine; Simulation: TSimulation; AUnits: TResonanceUnits; var Problem: Exception); overload; override;
    procedure Clear; override;
    procedure Delete(n: Integer);
    procedure DeleteSelected(UpStateSel, LoStateSel: Integer);
    function LargestIntensity: Double;
    {$IFNDEF CONSOLE}
    procedure FindLines(Centre, Range: Double; AUnits: TResonanceUnits; AColour: TColor = clHide; OnCheckLine: TOnCheckLine = nil);
    {$ENDIF}
    function FindPeak(Left, Right: Double; out Position: Double; AUnits: TResonanceUnits; AColour: TColor):Boolean;
    procedure ShowLine(LineNumber: Integer; var F: TextFile; PrintLevel: TPrintLevel; Threshold:Double; LinFormat: Boolean; var Problem: Exception); overload;
    property Line[AIndex: Integer]: TLine read GetLine;
  end;

  TVibronicallyAllowed = (vaMaybe, vaYes, vaNo);

  TTransitionMoment = class(TMNodePair)
  private
    FVibronicallyAllowed: TVibronicallyAllowed;
  protected
    FRank: TQuantumNumber;
    FMagnetic: Boolean;
    FNuclearMagneton: Boolean;
  public
    function BraOwner: TMNode; override;
    function KetOwner: TMNode; override;
    class function BraKetClass: TMNodeClass; override;
    constructor Create(AOwner: TComponent); overload; override;
    constructor Create(AOwner: TComponent; State1Name, State2Name: string); overload;
    function FixupProperties:Exception; override;
    function Allowed(Sym1, Sym2: Integer): Boolean;
    function CheckProjections(K1, K2: TQuantumNumber):Boolean; virtual;
    function MatrixElement(ABra, AKet: TBasisState; Intensity: Boolean): Complex;
    property Magnetic: Boolean read FMagnetic;
    property NuclearMagneton: Boolean read FNuclearMagneton;
    property VibronicallyAllowed: TVibronicallyAllowed read FVibronicallyAllowed write FVibronicallyAllowed ;
    property Rank: TQuantumNumber read FRank;
  end;

  TTransitionMomentClass = class of TTransitionMoment;

  TSingleTransitionMoment = class(TTransitionMoment)
  public
    constructor Create(AOwner: TComponent); overload; override;
    function HelpOnParameter(AIndex: Integer): string; override;
  published
    property Strength: Double  index 1 read GetParameter write SetParameter;
  end;

  TNuclearMagneticMoment = class(TTransitionMoment)
  public
    constructor Create(AOwner: TComponent); overload; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function PairName: string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
  published
    property gN: Double index 1 read GetParameter write SetParameter;
    property Nucleus:  Integer index 1 read GetSetting write SetSetting;
  end;

  TBaseRotationalTransitionMoment = class(TSingleTransitionMoment)
  public
    constructor Create(AOwner: TComponent); overload; override;
  published
    property C:         Double index LTI-2 read GetParameter write SetParameter;
    property D:         Double index LTI-1 read GetParameter write SetParameter;
  end;

  TRotationalTransitionMoment = class(TSingleTransitionMoment)
  public
    constructor Create(AOwner: TComponent); overload; override;
    function PairName: string; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
  end;

  { TSphericalTransitionMoment }

  TSphericalTransitionMoment = class(TBaseRotationalTransitionMoment)
  private
    FActualComponent: TQuantumNumber;
  public
    constructor Create(AOwner: TComponent); override;
    function PairName: string; override;
    procedure SetSetting(AIndex: Integer; const AValue: Integer); override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function FixupProperties:Exception; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function SettingToStr(AIndex: Integer): string; override;
    procedure SetSettingFromStr(AIndex: Integer; const Str: string); override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    { Same as component, unless component = 'auto' in which case the symmetry allowed value }
    property ActualComponent: TQuantumNumber read FActualComponent;
  published
    property Rank:      Integer index 1 read GetSetting write SetSetting;
    property Component: Integer index 2 read GetSetting write SetSetting;
  end;

  { TMagneticTransitionMoment }

  TMagneticTransitionMoment = class(TSphericalTransitionMoment)
  public
    function PairName: string; override;
    constructor Create(AOwner: TComponent); overload; override;
  end;

  { TOrbitalTransitionMoment }

  TOrbitalTransitionMoment = class(TMagneticTransitionMoment)
    function PairName: string; override;
  end;

  { TSpinTransitionMoment }

  TSpinTransitionMoment = class(TMagneticTransitionMoment)
  public
    function PairName: string; override;
    constructor Create(AOwner: TComponent); overload; override;
  end;

  TabcAxis = (aAxis=0, bAxis, cAxis);
  TxyzAxis = (zAxis=0, yAxis, xAxis);

  TCartesianTransitionMoment = class(TBaseRotationalTransitionMoment)
  public
    function PairName: string; override;
  private
    function GetAxis(const Index: Integer): TabcAxis;
    procedure SetAxis(const Index: Integer; const AValue: TabcAxis);
  public
    constructor Create(AOwner: TComponent); override;
    function FixupProperties:Exception; override;
    function GetSettingsNames(AIndex: Integer; out LowestValue:Integer): PPChar; override;
    function HelpOnSetting(AIndex: Integer): string; override;
  published
    property Axis: TabcAxis index 1 read GetAxis write SetAxis;
  end;

  TTransitionMoments = class(TMNodePair)
  private
    FnRanks: Integer;
    function GetTransitionMoment(AIndex: Integer): TTransitionMoment;
  protected
    CacheLock: TMultiReadExclusiveWriteSynchronizer;
  public
    DisableRankCheck: Boolean; { TODO: Fix this fudge for vibrational modes }
    class function BraKetClass: TMNodeClass; override;
    constructor Create(AOwner: TComponent); overload; override;
    constructor Create(AOwner: TComponent; Manifold1Name, Manifold2Name: string); overload;
    destructor Destroy; override;
    function FixupProperties:Exception; override;
    function IsSettingColour(AIndex: Integer): Boolean; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    procedure AddToSpectrum(Spectrum: TSpectrum; Ev1, Ev2: TEigenvalues; ATransitionMoment: TTransitionMoment; Simulation: TSimulation; Wait: Boolean; PrintLevel: TPrintLevel; const Excitation: TExcitation; var Problem: Exception; IncludeZero: Boolean = False); virtual;
    function MakeLine(out Line: TLine; D: TMatrix; Ev1, Ev2: TEigenvalues; Index1, Index2: Integer; const Excitation: TExcitation; IncludeZero: Boolean = false):Boolean; overload; virtual;
    function MakeLine(out Line: TLine; Ev1, Ev2: TEigenvalues; Index1, Index2: Integer; Simulation: TSimulation):Boolean; overload;
    procedure AddMatrixToSpectrum(D: TMatrix; Spectrum: TSpectrum; Ev1, Ev2: TEigenvalues; PrintLevel: TPrintLevel; const Excitation: TExcitation; var Problem: Exception; IncludeZero: Boolean = false);
    function OriginalTransitionMatrix(Ev1, Ev2: TEigenvalues; Rank: TQuantumNumber; ATransitionMoment: TTransitionMoment; Intensity, Electric, Magnetic: Boolean; Simulation: TSimulation): TComplexMatrix;
    function TransitionMatrix(Ev1, Ev2: TEigenvalues; Wait, Cache: Boolean; TransitionMomentSel: TTransitionMoment; Simulation: TSimulation; APrintLevel: TPrintLevel = plDefault): TComplexMatrix;
    procedure TransformTransitionMatrix(D: TComplexMatrix; Ev1, Ev2: TEigenvalues);
    function Allowed(J1: TQuantumNumber; Sym1: Integer; J2: TQuantumNumber; Sym2: Integer): Boolean; virtual;
    function PairName: string; override;
    function MaxRank: TQuantumNumber; virtual;
    function Mixture: TMixture;
    function Molecule: TMolecule;
    { TODO: Add checks in loader for duplicate or erroneous TransitionMoment(s) }
    property nRanks: Integer read FnRanks;
    property TransitionMoment[AIndex: Integer]: TTransitionMoment read GetTransitionMoment;
  published
    property Colour: TColor index 1 read GetColour write SetColour;
  end;

  { TNucleus }

  TNucleus = class(TMNodeWithParameters)
  private
    FState: TState;
  protected
    function GetName: TComponentName; override;
    procedure SetName(const Value: TComponentName); override;
    procedure SetFloating(AIndex: Integer; const Value: Boolean); override;
    procedure SetIncrement(AIndex: Integer; const Value: Double); override;
    procedure SetComponentIndex(const Value: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure SetSetting(AIndex: Integer; const AValue: Integer); override;
    procedure SetParameter(AIndex: Integer; const Value: Double); override;
    function Molecule: TMolecule;
    function HelpOnSetting(AIndex: Integer): string; override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    function HasIdentical: Boolean; virtual;
    function MatchingNucleus: TNucleus;
    procedure GlobalShift(Shift:Integer); virtual;
    procedure GlobalMoveUp(Sender: TObject);
    procedure GlobalMoveDown(Sender: TObject);
    procedure AddHandlers; override;
    procedure AddNodeParameters(var FittingParameters: TParameters); override;
    property State: TState read FState;
  published
    property Spin: Integer   index 1 read GetSetting write SetSetting;
    property AsNext: Boolean index 2 read GetFlag write SetFlag;
  end;

  TNucleusClass = class of TNucleus;

  TSpinSpin = class(TMNodePair)
  public
    class function BraKetClass: TMNodeClass; override;
    function PairName: string; override;
  end;

  TSpinSpinClass = class of TSpinSpin;

  THyperfineBasis = array of TQuantumNumber;

  { TPartitionFunction - for generating the partition function as a function of temperture }

  TPartitionFunction = class(TMNode)
  public
    constructor Create(AOwner: TComponent); override;
    function CalculatePartitionFunction(Molecule: TMolecule; Simulation: TSimulation; out Overflow: Exception; var SpinSpeciesAbundance: TDoubleArray;
                                        Print: Boolean = false; ShowMolecule: Boolean = false; ShowHeader: Boolean = false):Double; virtual; abstract;
    function EnergyZero(LowestOrigin: Double):Double; virtual;
    function HelpOnSetting(AIndex: Integer): string; override;
  published
    property Active: Boolean index 1 read GetFlag write SetFlag default True;
  end;

  { TMolecule }

  TMolecule = class(TMixtureMNode) { Isotopomer }
  private
    FLongestManifoldName: string;
    FnStates, FnManifolds: Integer;
    FManifolds: TMNodeArray;
    FnTransitionMoments: Integer;
    FHalfSpin, FMaxRank, FSumNuclearSpin: TQuantumNumber;
    FLowestOrigin: Double;
    FLastPartitionFunction, FLastTemperature, FLastTvib, FLastTspin: Double;
    FLastProblem: string;
    FAllSinglet: Boolean;
    FSymLength: Integer;
    FActivePartitionFunction, FDefaultPartitionFunction: TPartitionFunction;
    FSpinSpeciesAbundance: TDoubleArray;
    function GetTransitionMomentsFunction(AIndex, BIndex: Integer): TTransitionMoments;
    function GetMixture: TMixture;
    function GetManifold(AIndex: Integer):TManifold;
    function GetActivePartitionFunction: TPartitionFunction;
  protected
    FnSpinSpecies: Integer;
    procedure SetnNuclei(const Index, Value: Integer); virtual;
    procedure SetModified(const Value: Boolean); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    Nuclei: array of TNuclearInfo;
    HyperfineBasisStates: array of array of THyperfineBasis;
      { First index is number of nuclear spins, to allow for partial ignoring of spins }
      { Second index is for the basis }
      { Third index is nuclear spin }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    class function GetChildTypes(AIndex: Integer): TMNodeClass; override;
    class function nChildTypes: Integer; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    procedure ClearCache; override;
    function FixupProperties:Exception; override;
    procedure PrecomputeSymNames; virtual;
    function IsSettingColour(AIndex: Integer): Boolean; override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function SettingToPrettyStr(AIndex: Integer): string; override;
    function GetTransitionMoments(AIndex: Integer): TTransitionMoments;
    property HalfSpin: TQuantumNumber read FHalfSpin;
    property AllSinglet: Boolean read FAllSinglet;
    function Species: TSpecies;
    function EinsteinA(const Line: TLine; AUnits: TResonanceUnits): Double;
    function AddToSpectrum(Spectrum: TSpectrum; PrintLevel: TPrintLevel; const Excitation: TExcitation; var Problem: Exception;
                            Simulation: TSimulation;
                            J1Sel: Integer = Low(Integer); Sym1Sel: Integer = -1; J2Sel: Integer = Low(Integer); Sym2Sel: Integer = -1;
                            Manifold1Sel: TManifold = nil; Manifold2Sel: TManifold = nil;
                            DeltaJ: TQuantumNumber = -10000;
                            TransitionMomentsSel: TTransitionMoments = nil; TransitionMomentSel: TTransitionMoment = nil;
                            State1Sel: TState = nil; State2Sel: TState = nil; IncludePerturbations: Boolean = True; IncludeZero: Boolean = False):Boolean; virtual;
    function PartitionFunction(Simulation: TSimulation; var Overflow: Exception):Double;
    function Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean; virtual; abstract;
    function JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer; virtual; abstract;
    function JDeAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer; virtual;
    function LastSym: Integer; virtual; abstract;
    function SymName(Sym: Integer): string; virtual;
    function JAdjustedSymName(J: TQuantumNumber; Sym: Integer): string;
    function JAdjustedSymNameBase(J: TQuantumNumber; Sym: Integer): string; virtual;
    function BothSymNames(J: TQuantumNumber; Sym: Integer): string;
    function TryStrToJAdjustedSymName(J: TQuantumNumber; const s: string; out Sym: Integer): Boolean;
    function TryStrToAnySymName(J: TQuantumNumber; const s: string; out Sym: Integer): Boolean;
    function FixSym(J: TQuantumNumber; Sym:Integer; InputPgopherVersion: Integer):Integer; virtual;
    function QnoInfo(QnoType: TQnoType): TQnoInfo; virtual;
    property LongestManifoldName: string read FLongestManifoldName; { For formatting only }
    function ManifoldWithName(AName: string): TManifold;
    property LowestOrigin: Double read FLowestOrigin;
    property nManifolds: Integer read FnManifolds;
    property nStates: Integer read FnStates;
    property Manifold[AIndex: Integer]: TManifold read GetManifold;
    property MaxRank: TQuantumNumber read FMaxRank;
    property Mixture: TMixture read GetMixture;
    function FindTransitionMoments(ManifoldA, ManifoldB: TManifold):TTransitionMoments;
    property TransitionMoments[AIndex, BIndex:Integer]: TTransitionMoments read GetTransitionMomentsFunction;
    property nSpinSpecies: Integer read FnSpinSpecies;
    function SpinSpeciesAbundance(Simulation: TSimulation; Index: Integer; var Overflow: Exception):Double;
    property nTransitionMoments: Integer read FnTransitionMoments;
    property SumNuclearSpin: TQuantumNumber read FSumNuclearSpin;
    property SymLength: Integer read FSymLength;
    property ActivePartitionFunction: TPartitionFunction read GetActivePartitionFunction;
  published
    property Abundance:  Double index 1 read GetParameter write SetParameter;
    property nNuclei:    Integer index 1 read GetSetting write SetnNuclei;
    property Jmin:       Integer index 2 read GetSetting write SetSetting;
    property Jmax:       Integer index 3 read GetSetting write SetSetting;
    property Colour:     TColor  index 4 read GetColour write SetColour;
    property JAdjustSym: Boolean index 5 read GetFlag write SetFlag;
    property BlockMatrix:Boolean index 6 read GetFlag write SetFlag;
  end;

  TMoleculeClass = class of TMolecule;

  { TBasisState }

  TBasisState = class(TComponent)
  public
    State: TState;
    StateIndex: Integer; { For help with eigenvalue searches }
    HyperfineState: THyperfineBasis;
    F, J, N, S, M: TQuantumNumber; { Note than N is not guaranteed, particularly for linear molecules }
                                   { S is included here to allow it to be excluded from calculation if required }
    Sym: Integer; { Note that this F and M are included here, as they will be different from the owning TEigenvalues for non zero external fields }
    Phase: Integer;
    constructor Create(AOwner: TComponent; AF: TQuantumNumber; ASym: Integer);
    function TrackIndex:Integer; virtual; { A number >= 0 for tracking related states, mainly in energy level plots. }
    function BasisLabel: string; virtual; { Basis State Label }
    function BasisLabelKey: string; virtual; { Explanation of Basis State Label }
    function StateLabel(QuantumNumberFormat: TQuantumNumberFormat): string; virtual; { State name that goes with the given basis state
                                            (assuming match between basis state and final state is OK) }
    function StateLabelKey: string; virtual; { Explanation of State Name }
    function Qno(QnoType: TQnoType): TQuantumNumber; virtual;
    function MakeSignature(SignA:Integer; Ket: TBasisState; SignB: Integer): TIntegerArray; virtual; abstract;
    procedure SetSymbolicVariables(SignA:Integer; Ket: TBasisState; SignB: Integer); virtual; abstract;
    function HyperfineStateLabel(WithBlanks: Boolean = True): string;
    function HyperfineStateLabelKey(WithBlanks: Boolean = True; ExtraChar: Char = ' '): string;
    function TxName(Upper, Lower: TBasisState): string; virtual;    { Make transition label such as R(3) }
    function TxNameKey(Upper, Lower: TBasisState): string; virtual; { String explaining transition label }
    function HyperfineTxNameKey(WithBlanks: Boolean):string;
  end;

  { Type of matrix elements required }
  TMatElemType = (meHrot {Normal Hamiltonian}, meWidth {Operator for lifetime}, meFloatedDerivatives, meAllDerivatives);
  TMatElemTypes = set of TMatElemType;
  TDArray = array of Double;
  TComplexMatrices = array of TComplexMatrix;

  TCachedTransitionMatrix = class(TInterlockedComponent)
  public
    Eigenvalues, OtherEigenvalues:TEigenvalues;
    TransitionMatrix: TComplexMatrix;
    PrintLevel: TPrintLevel;
    TransitionMoments: TTransitionMoments;
    Simulation: TSimulation;
    destructor Destroy; override;
    procedure DoMake(RequiredAttributes: Pointer); override;
    function HasCorrectAttributes(Attributes: Pointer):Boolean; override;
  end;

  { TEigenvalues }

  TDerivative = record
    MNode: TMNode;
    Index: Integer;
    Values: TColumnVector;
  end;

  TDerivatives = array of TDerivative;

  TEigenvalues = class(TInterlockedComponent)
  private
    FE: TColumnVector;
    FWidths: TColumnVector;
    FC: TComplexMatrix;
    FDominantBasisIndex: TIntColumnVector; { Dominant basis state for each eigenvalue }
    CachedTransitionMatrices: array of TCachedTransitionMatrix;
    FHasFields: Boolean;
    FStatisticalWeights, FStatWtGroups: array of Integer;
    FDerivativesCalculated: Boolean;
    function GetDominantBasisState(AIndex: Integer): TBasisState; {$IFDEF INLINE} inline; {$ENDIF}
  protected
    FStatisticalWeight, FStatWtGroup: Integer;
    function GetEnergy(AIndex: Integer): Double; virtual;
  public
    Manifold: TManifold;
    J: TQuantumNumber;
    Sym: Integer;
    LastStateIndex: Integer;
    SelectedState: TState;    { Eigenvalues calculated with this State only in basis }
    NoPerturbations: Boolean; { Eigenvalues calculated excluding perturbations }
    RemovedSpins:    Byte;    { Number of spins not included in calculation }
    Basis: array of TBasisState;
    PrintLevel: TPrintLevel;
    ExField: TStaticFields;   { External fields; note that if either is not zero, J is interpreted as MJ and Sym is <0 }
    Derivatives: TDerivatives; { Derivatives with respect to parameters }
    function Degeneracy(AIndex: Integer): Integer; virtual;
    function Size: Integer; virtual;
    function StatisticalWeight(Index: Integer):Integer;
    function SpinSpecies(Index: Integer):Integer;
    procedure MatrixElements(out Matrices: TComplexMatrices; BlockMatrix: Boolean; MatElemType: TMatElemType; IncludePerturbations: Boolean = True);
    procedure CalculateDerivatives(MatElemTypes: TMatElemTypes);
    procedure DoEigenSearch(Co: TComplexMatrix; Offset: Integer = 0);
    procedure DoLimitedEigenSearch(Co: TComplexMatrix; Offset: Integer = 0);
    function DiagonalMatrixElements(MatElemType: TMatElemType; Base: Integer = 0; IncludePerturbations: Boolean = False): TDArray;
    procedure Diagonalize(var Matrices: TComplexMatrices; EigenvectorsWanted, EigenSearch, LimitSearch: Boolean);
    procedure Fixup;
    procedure SortBasisOn(V: TDarray; ByState: Boolean; Base: Integer = 0);
    procedure SortFrom(Start: Integer; Emax: Double);
    procedure Multiply(M: TComplexMatrix);
    procedure MakeHamiltonian(out Matrices: TComplexMatrices; BlockMatrix: Boolean; BasisOrder: TBasisOrder);
    procedure Make(Wait:Boolean; MatElemTypes: TMatElemTypes); overload;
    procedure DoMake(RequiredAttributes: Pointer); override;
    function HasCorrectAttributes(Attributes: Pointer):Boolean; override;
    procedure MakeWidths(APrintLevel: TPrintLevel);
    procedure ShowBasisMatrix(M: TComplexMatrix; const Title: string; Offset: Integer);
    procedure ShowTransitionMatrix(Ev: TEigenValues; M: TComplexMatrix; const Title: string; Offset: Integer; ShowBasis: Boolean = True);
    function FindDerivatives(const Parameter: TParameter):Integer;

    function FieldLabel: string;
    procedure ShowEigenvalues;
    function Transform(M: TComplexMatrix): TColumnVector;
    function Qno(AIndex: Integer; QnoType: TQnoType):TQuantumNumber; virtual;
    function DominantLabel(AIndex: Integer; QuantumNumberFormat: TQuantumNumberFormat):string; virtual;
    constructor Create(AManifold: TManifold; AJ: TQuantumNumber; ASym: Integer; const AExField: TStaticFields;
                       IncludePerturbations: Boolean = True;
                       AState: TState = nil; RemoveSpins: Byte = 0);
    destructor Destroy; override;
    // property Basis[AIndex: Integer]: TBasisState read GetBasisState;
    property DominantBasis[AIndex: Integer]: TBasisState read GetDominantBasisState;
    property E: TColumnVector read FE;
    property C: TComplexMatrix read FC;
    property Widths: TColumnVector read FWidths;
    property Energy[AIndex:Integer]:Double read GetEnergy; { Same as E[AIndex], but allows override }
    property HasFields: Boolean read FHasFields;
  end;

  { TPerturbation }

  TPerturbation = class(TMNodePair)
  private
    FMatrixElements: TMatrixElements; { Matrix elements in symbolic form }
    FMatrixElementsEmpty: Boolean;
    FParameterVariable: TSortedVariable;
  public
    NuclearRank: TQuantumNumber; { Rank of nuclear spin operator }
    property MatrixElements: TMatrixElements read FMatrixElements;
    property ParameterVariable: TSortedVariable read FParameterVariable;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetParameter(AIndex: Integer; const Value: Double); override;
    class function BraKetClass: TMNodeClass; override;
    function Connects(State1, State2: TState): Boolean;
    function FixupProperties:Exception; override;
    function MatrixElement(State1, State2: TBasisState; MatElemType: TMatElemType): Complex;
    function SymbolicMatrixElements:TMatrixElements; virtual;
    function CheckMatrixElementsAvailable: Boolean; virtual;
    function PairName: string; override;
    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; override;
    procedure PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); override;
    procedure PrettyPrintSettings(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
  published
    property Value:  Double index 1 read GetParameter write SetParameter;
    { Index, starting from 1, of nuclear spin involved in perturbation
      0 (default) for those not involving a nuclear spin }
    property Nucleus:Integer   index LPI-3 read GetSetting write SetSetting;
    property SymSelect:Integer index LPI-2 read GetSetting write SetSetting;
    property ScalePrev:Integer index LPI-1  read GetSetting write SetSetting;
  end;

  TMatrixElementFunction = function (A, B: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer):Complex of object;
  TSymCheck = (MustEqual, MayCancel, NoCheck, FullSum);

  TObservationsReaderBase = class
  public
    BranchRank: TQuantumNumber;
    RemoveSpins: Integer;
    DefaultSymmetry: Integer;
    function ReadQuantumNumber(var a: Integer; const si: string): Boolean; virtual; abstract;
    procedure ReportError(es: string); virtual; abstract;
    function NextItem: string; virtual; abstract;
  end;

  { TState }

  TState = class(TMNode)
  private
    FMolecule: TMolecule;
    FNuclei: TMNodeArray;
    FMatrixElements: TMatrixElements;
    FMatrixElementsEmpty: Boolean;
  protected
    VariableList: TVariableList;
    FFullRotationalSymmetry: Boolean;
    AllowComplex: Boolean;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function HalfSpin(RemovedSpins: Byte = 0): TQuantumNumber; virtual;
    function AddBasis(Eigenvalues: TEigenvalues; F, M: TQuantumNumber; Sym: Integer; out StatWtGroup: Integer): Integer; { Returns Statistical Weight }
    function AddOneJ(Eigenvalues: TEigenvalues; J: TQuantumNumber; Sym: Integer; HyperfineState:THyperfineBasis; Seffective: TQuantumNumber; out StatWtGroup: Integer): Integer; virtual; abstract;
    class function SpinSpinType: TSpinSpinClass; virtual; abstract;
    procedure AdjustNuclei(FixNumber: Boolean);
    function SpinSpin(Nuc1, Nuc2: Integer):TSpinSpin;
    function FixupProperties:Exception; override;
    procedure ClearCache; override;
    function HasWidth: Boolean; virtual;

    { Functions working on symmetrized basis functions }
    function MatrixElement(A, B: TBasisState; MatElemType: TMatElemType): Complex;
    function SymmetrizedRotationalMatrixElement(A, B: TBasisState; MNode: TMnode; Flag: Integer; MatrixElementFunction: TMatrixElementFunction; SymCheck: TSymCheck = MustEqual): Complex;
    function SymmetrizedHyperfineMatrixElement(A, B: TBasisState; Rank: TQuantumNumber;
                                    MatrixElementFunction: TMatrixElementFunction; MNode: TMNode): Complex;
    { Note that transition moment matrix elements can be implemented in TState or TransitionMoment descendents }
    function TransitionMatrixElement(TransitionMoment: TTransitionMoment; A, B: TBasisState): Complex; virtual;
    function PerturbationMatrixElement(Perturbation: TPerturbation; A, B: TBasisState; MatElemType: TMatElemType): Complex;

    { Functions working on unsymmetrized basis functions }
    // function RotSpinMatrixElement(A, B: TBasisState; MatElemType: TMatElemType): Double; virtual; abstract;
    function RotationalMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Complex; virtual; abstract;
    function RotationalTransitionMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; TransitionMoment: TMNode; F: TQuantumNumber):Complex; virtual; abstract;
    function RotationalPerturbationMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMNode; Flag: Integer):Complex; virtual; abstract;
    function HyperfineMatrixElement(ABra, AKet: TBasisState; Signp, Sign: Integer; MNode: TMnode; RankWanted: TQuantumNumber): Complex; virtual; abstract;
    function SymbolicRotationalMatrixElement(AIndex: Integer):TMatrixElements; virtual; { For given parameter }

    procedure CheckedMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; const AMatrixElements: TMatrixElements; Numeric: Complex; const AName: string);
    function CheckedRotationalPerturbationMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMNode; Flag: Integer):Complex;
    function CheckedRotationalMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Complex;
    procedure CheckMatrixElementsAvailable;
    function AllowedSphericalComponent(TransitionMoment: TSphericalTransitionMoment; State: TState; out Component: TQuantumNumber):Exception; virtual;

    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function ReadBranchFormat(Text: string; var Obs: TObservation; UpperState: TState; Reader: TObservationsReaderBase):Boolean; virtual;
    function ReadHyperfineBranch(Text: string; var Obs: TObservation; var HyperfineU, HyperfineL: THyperfineBasis; Reader: TObservationsReaderBase; HasFields:Boolean): Boolean;
    function ConvertParameter(AIndex: Integer; Zero: Boolean):TPerturbation; virtual;
    function ParameterVariable(AIndex: Integer):TSortedVariableNode; virtual;
    function nParameterVariables: Integer; virtual; { Normally nParameters, but Vib has a special case to allow for subsidiary parameters }
    function MatrixElementType(const Changes: array of Integer; const H: string):string; virtual; abstract;
    procedure ConvertToPerturbations; virtual;
    function Manifold: TManifold;
    function FindColour: TColor;
    function IsSettingColour(AIndex: Integer): Boolean; override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; override;
    function SettingToPrettyStr(AIndex: Integer): string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    procedure PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); override;

    procedure SymmetryTable(var F: Text); virtual;
    function StateOrigin: Double; virtual; { Default is just Origin, but ideally choose lowest J=0 energy }
                                           { Used to stop Boltzman factors blowing up at low temperatures }
    function Nucleus(Index:Integer): TNucleus;
    property Molecule: TMolecule read FMolecule;
    property FullRotationalSymmetry: Boolean read FFullRotationalSymmetry; { If point group symmetry used is not a subgroup of the rotational Hamiltonian }

    property MatrixElements: TMatrixElements read FMatrixElements;
    property MatrixElementsEmpty: Boolean read FMatrixElementsEmpty;
  published
    property Origin: Double index 1 read GetParameter write SetParameter;
    property Width:  Double index 2 read GetParameter write SetParameter;
    property Colour: TColor index 1 read GetColour write SetColour;
    property RveSelect: Integer index 2 read GetSetting write SetSetting; { Select overall symmetry }
    property S: TQuantumNumber index 3 read GetSetting write SetSetting; { Spin }
  end;

  TEigenValueCache = class(TComponentList)
  private
    FLock: TMultiReadExclusiveWriteSynchronizer;
  public
    destructor Destroy; override;
    property Lock: TMultiReadExclusiveWriteSynchronizer read FLock;
  end;

  { TManifold }

  TManifold = class(TMixtureMNode)
  private
    FEigenValueCache: TEigenValueCache;
    FHasWidth: Boolean;
    FnPerturbations: Integer;
    FnStates: Integer;
    FLowestStateOrigin: double;
    FMaxSpin: TQuantumNumber;
    FPerturbations: TMNodeArray;
    FStates: TMNodeArray;
    PNumbers: array of Integer;
    function GetState(AIndex: Integer): TState;
    function GetPerturbation(AIndex: Integer): TPerturbation; inline;
    function FindEigenvalues(StartFrom: Integer; out LastSearched: Integer; J: TQuantumNumber; Sym: Integer;
                             const AExField: TStaticFields;
                             SelState: TState = nil; IncludePerturbations: Boolean = True;
                             RemoveSpins: Byte = 0): TEigenvalues;
    procedure UpdateCounts(AComponent: TComponent; Operation: TOperation);
  protected
    function bf(T, E: Double): Double;
    procedure SetModified(const Value: Boolean); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    property HasWidth: Boolean read FHasWidth;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function HalfSpin(RemovedSpins: Byte = 0): TQuantumNumber; virtual;

    { These provide the implementation of the selective population stuff }
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    function PopulationIndex(J: TQuantumNumber; Sym, EvNo: Integer; AutoCreate: Boolean = False):Integer;
    function HelpOnParameter(AIndex: Integer): string; override;
    procedure DecodeParameterName(s: string; out J: TQuantumNumber; out Sym, EvNo: Integer; AQuantumNumberFormat: TQuantumNumberFormat);
    function EncodeParameterName(J: TQuantumNumber; Sym, EvNo: Integer):string;
    function nPopulations: Integer;
    function Population(Eigenvalues: TEigenvalues; i: Integer; Simulation: TSimulation; var Overflow: Exception; IncludeTspin: Integer): Double; virtual;
    procedure AddAllLevels(AState: TState); virtual;

    procedure ClearCache; override;
    function Eigenvalues(J: TQuantumNumber; Sym: Integer; MatElemTypes: TMatElemTypes;
                         const AExField: TStaticFields;
                         Wait: Boolean = True;
                         PrintLevel: TPrintLevel = plDefault; SelState: TState = nil; IncludePerturbations: Boolean = True;
                         RemoveSpins: Byte = 0;
                         CreateIfNotFound: Boolean = True
                         ): TEigenvalues; virtual;
    function FindJmin(Default: Integer; RemovedSpins: Byte = 0):Integer; virtual;
    function FindJmax(Default: Integer; RemovedSpins: Byte = 0):Integer; virtual;
    function FindColour: TColor;
    function FirstSym(J: TQuantumNumber):Integer; virtual;
    function FixupProperties:Exception; override;
    function IsSettingColour(AIndex: Integer): Boolean; override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    function SettingToPrettyStr(AIndex: Integer): string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function NextSym(J: TQuantumNumber; Sym: Integer):Integer; virtual;
    function nSym: Integer; virtual; abstract; { Number of different symmetries }
    function FindStateLabel(out IndexF: Smallint; out SymF: Shortint; FW: TQuantumNumber; OtherQno: array of TQuantumNumber; AQuantumNumberFormat: TQuantumNumberFormat; const ExField: TStaticFields; SymW: Integer = -1; JW: TQuantumNumber = -1; NW: TQuantumNumber = -1; StateW: TState = nil; RemoveSpins: Byte = 0; HyperfineW: THyperfineBasis = nil; ShowPossible: Boolean = False): string; virtual;
    function FindIndex(out EigenvaluesF: TEigenvalues; out IndexF: Smallint; out SymF: Shortint; FW: TQuantumNumber; OtherQno: array of TQuantumNumber; AQuantumNumberFormat: TQuantumNumberFormat; const ExField: TStaticFields; SymW: Integer = -1; JW: TQuantumNumber = -1; NW: TQuantumNumber = -1; StateW: TState = nil; RemoveSpins: Byte = 0; HyperfineW: THyperfineBasis = nil; EigenvaluesWanted: Boolean = True; ShowPossible: Boolean = False): string; virtual;
    function Mixture: TMixture;
    function Molecule: TMolecule;
    function PageName: string; override;
    procedure ConvertToPerturbations;
    function StateWithName(AName: string): TState;
    property nStates: Integer read FnStates;
    property State[AIndex: Integer]: TState read GetState;
    property LowestStateOrigin: double read FLowestStateOrigin;
    property nPerturbations: Integer read FnPerturbations;
    property Perturbation[AIndex: Integer]: TPerturbation read GetPerturbation; default;
    property MaxSpin: TQuantumNumber read FMaxSpin;
  published
    property Jmin:             Integer index 1 read GetSetting write SetSetting;
    property Jmax:             Integer index 2 read GetSetting write SetSetting;
    property Initial:          Boolean index 3 read GetFlag write SetFlag;
    property Colour:           TColor  index 4 read GetColour write SetColour;
    property EigenSearch:      Boolean index 5 read GetFlag write SetFlag;
    property LimitSearch:      Boolean index 6 read GetFlag write SetFlag;
    property AutoQConverge:    Boolean index 7 read GetFlag write SetFlag;
    { TODO: property ShowStateNumbers: Boolean index 7 read GetFlag write SetFlag; }
  end;

  TSpecies = class(TMNode)
  private
    function GetIsotopomer(AIndex: Integer): TMolecule;
  public
    constructor Create(AOwner: TComponent); override;
    function AddToSpectrum(Spectrum: TSpectrum; var Problem: Exception; Simulation: TSimulation): Boolean; virtual;
    function FindMolecule(AName: string): TMolecule;
    function QnoInfo(QnoType: TQnoType): TQnoInfo;
    function IsSettingColour(AIndex: Integer): Boolean; override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function SettingToPrettyStr(AIndex: Integer): string; override;
    procedure SetModified(const Value: Boolean); override;
    property Isotopomer[AIndex: Integer]: TMolecule read GetIsotopomer;
  published
    property Concentration: Double index 1 read GetParameter write SetParameter;
    property Jmin:      Integer index 1 read GetSetting write SetSetting;
    property Jmax:      Integer index 2 read GetSetting write SetSetting;
    property Colour:    TColor  index 3 read GetColour write SetColour;
  end;

  { TSimulation }

  TColours = array of TColor;
  TGraphStyle = (gsNoOffset,   { All spectra on one scale }
                 gsOffset,     { Separate simulation and experiment }
                 gsOffsetAll,  { Separate all spectra }
                 gsInverted,   { Show simulation inverted below experiment }
                 gsInvertedAll { Separate all traces, with simulated traces inverted }
                );

  TSimulation = class(TUnitsMNode)
  private
    FSettingOverridden, FParameterOverridden: array of Boolean; { Track variables that have been set }
    FBaseNode: TSimulation;
    FWritten: Boolean;
    procedure SetLifeModel(const Index: Integer; const AValue: TLifeModel);
    function GetLifeModel(const Index: Integer): TLifeModel;
    procedure SetFortratQno(const Index: Integer; const Value: TQnoType);
    function GetFortratQno(const Index: Integer): TQnoType;
    function GetExField: TStaticFields;
  protected
    function GetDefaultSettings(AIndex: Integer): string; override;
  public
    HasPlotted: Boolean;
    GraphStyle: TGraphStyle;

    constructor Create(AOwner: TComponent); override;
    function Mixture: TMixture;
    procedure ChangePlotUnits(NewPlotUnits: TResonanceUnits);
    procedure SetSetting(AIndex: Integer; const AValue: Integer); override;
    procedure SetParameter(AIndex: Integer; const Value: Double); override;
    function ParameterOverridden(AIndex: Integer):Boolean;
    function GetSettingsNames(AIndex: Integer; out LowestValue:Integer): PPChar; override;
    function SettingToStr(AIndex: Integer): string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function MakeLineList(var Problem: Exception): TLineList;
    procedure UpdateBaseNode(ABaseNode: TSimulation);
    procedure WriteFloatableValue(AIndex: Integer; Deflt: string; DOMNode: TDOMNode; DefaultCheck: Boolean); override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    procedure SetSettingFromStr(AIndex: Integer; const s: string); override;

    function LineIntensity(const Line: TLine; AUnits: TResonanceUnits; var Problem: Exception): Double;
    function HasWidths:Boolean;
    function HasFields:Boolean;
    property ExField: TStaticFields read GetExField;

  published
    property IntensityUnits: TIntensityUnits index 1 read GetIntensityUnits write SetIntensityUnits;
    property LifeModel:     TLifeModel index 2 read GetLifeModel write SetLifeModel;
    property PlotUnits:     TResonanceUnits index 3 read GetUnits write SetUnits;
    property nDF:           Integer index 4 read GetSetting write SetSetting;
    property WidthMult:     Integer index 5 read GetSetting write SetSetting;     { Multiple of linewidth to extend convolution over }
    property ShowSum:       Boolean index ShowSumIndex read GetFlag write SetFlag;
    property ShowParts:     Boolean index ShowPartsIndex read GetFlag write SetFlag;
    property ShowFortrat:   Boolean index ShowFortratIndex read GetFlag write SetFlag;
    property UseUpper:      Boolean index ShowFortratIndex+1 read GetFlag write SetFlag;
    property ShowSymmetry:  Boolean index 10 read GetFlag write SetFlag; { ... in Fortrat plots }
    property ShowDeltaJ:    Boolean index 11 read GetFlag write SetFlag; { ... in Fortrat plots }
    property ScaleMarkSize: Boolean index 12 read GetFlag write SetFlag; { ... in Fortrat plots }
    property UseSymmetry:   Boolean index 13 read GetFlag write SetFlag; { ... in Fortrat plots }
    property UseStateNumber:Boolean index 14 read GetFlag write SetFlag; { ... in Fortrat plots }
    property FortratQno:    TQnoType  index 15 read GetFortratQno write SetFortratQno; { Quantum number to use in Fortrat plots }

    property Fmin:        Double index  1 read GetParameter write SetParameter;
    property Fmax:        Double index  2 read GetParameter write SetParameter;
    property Temperature: Double index TemperatureIndex read GetParameter write SetParameter;
    property Gaussian:    Double index GaussianIndex read GetParameter write SetParameter;
    property Lorentzian:  Double index LorentzianIndex read GetParameter write SetParameter;
    property Foffset:     Double index FoffsetIndex read GetParameter write SetParameter;
    property SMargin:     Double index  7 read GetParameter write SetParameter;
    { Relative Size of peak to ignore when convoluting for contour fitting }
    property OThreshold:  Double index  8 read GetParameter write SetParameter;
    property RefWidth:    Double index  9 read GetParameter write SetParameter;
    property Tvib:        Double index 10 read GetParameter write SetParameter;
    property MinI:        Double index 11 read GetParameter write SetParameter; { as a fraction of maximum intensity; For Fortrat plots; TODO: Use elsewhere }
    property Saturation:  Double index 12 read GetParameter write SetParameter;
    property Tspin:       Double index 13 read GetParameter write SetParameter;
    property EField:      Double index 14 read GetParameter write SetParameter;
    property BField:      Double index 15 read GetParameter write SetParameter;
    { TODO: property Angle:       Double index 16 read GetParameter write SetParameter; }
    property Doppler:     Double index 16 read GetParameter write SetParameter;
  end;

  TCorrelationEntry = record
    Row, Col: string;
    Value: Double;
  end;

  TUnpackedCorrelations = array of TCorrelationEntry;

  TCorrelations = class(TMNode)
  public
    FitDetails: TFitDetails;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Clear;
    function nEntries: Integer;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    function EntryName(i: Integer; Quote: Boolean):string;
    function ReadData(UnpackedCorrelations: TUnpackedCorrelations; FirstPass: Integer; CheckUnread: Boolean; OtherNode:TRootMNode):Integer;
    procedure Resync(Sender: TObject);
    procedure AdjustSize;
    procedure AddHandlers; override;
    procedure UnpackCorrelations(var UnpackedCorrelations: TUnpackedCorrelations);
    procedure PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); override;
  end;

  { TMixture }

  TMixture = class(TRootMNode)
  private
    FnMolecules: Integer;
    FnSpecies: Integer;
    FnSimulations: Integer;
    FSavedParameters: array of TFitDetails;
    FUndoPosition: Integer; { Index within the above of current set of parameters }
    FLongestMoleculeName: string; { For formatting }
    FLongestManifoldName: string; { For formatting }
    FLongestSymLength: Integer; { For formatting }
    FLargestIndex: Integer;     { For formatting }
    FLoadedFormSettings: TDOMDocument; { required between loading and applying to forms only }
    FLineList: TLineList;
    function GetQuantumNumberFormat(AIndex: Integer): TQuantumNumberFormat;
    function GetSimulation(AIndex: Integer): TSimulation;
    function GetSpecies(AIndex: Integer): TSpecies;
    procedure SetBasisOrder(const Index: Integer; const AValue: TBasisOrder);
    function GetBasisOrder(const Index: Integer): TBasisOrder;
    procedure SetPrintLevel(const Index: Integer; const AValue: TPrintLevel);
    function GetPrintLevel(const Index: Integer): TPrintLevel;
    procedure LoadCurrentSavedParameters;
    procedure SetQuantumNumberFormat(const Index: Integer; const Value: TQuantumNumberFormat);
    procedure SetLineList(const Value: TLineList);
    procedure SetEstUnc(const Index: Integer; const Value: Boolean);
  protected
    procedure SetModified(const Value: Boolean); override;
    procedure SetEUnits(const Index: Integer; const AValue: TEnergyUnits);
    function GetEUnits(const Index: Integer): TEnergyUnits;
  public
    FittingParameters: TParameters;
    FCorrelations: TCorrelations;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure FillParameters; { Set up parameters array }
    procedure PushParameters(FitType: TFitType; SetCurrent: Boolean);
    procedure DiscardFit;
    procedure UndoFit;
    procedure ClearUndo;
    function CanUndoFit: Boolean;
    procedure RedoFit;
    function CanRedoFit: Boolean;
    procedure CycleUnits; override;
    procedure ClearCache; override;
    function FixupProperties:Exception; override;
    function SubPropInfo(const Token: string; out MNode: TMNode; out ErrMsg: string): PPropInfo; override;
    function IsValidChildClass(MNodeClass: TMNodeClass):Boolean; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    procedure PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); override;
    function QnoInfo(QnoType: TQnoType): TQnoInfo;
    procedure PreCalculatePartitionFunctions(Simulation: TSimulation; var Overflow: Exception);
    procedure ConvertToPerturbations;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    procedure ClearFormSettings;
    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; override;
    procedure SetSetting(AIndex: Integer; const AValue: Integer); override;
    function AddToSpectrum(Spectrum: TSpectrum; var Problem: Exception; Simulation: TSimulation): Boolean;
    function Correlations: TCorrelations;
    procedure PrepareForFit(FitType: TFitType; OtherNode: TRootMNode); override;

    property Species[AIndex: Integer]: TSpecies read GetSpecies; default;
    function FindMolecule(AName: string): TMolecule;
    property nMolecules: Integer read FnMolecules;
    property nSpecies: Integer read FnSpecies;
    property LongestMoleculeName: string read FLongestMoleculeName;        { For formatting }
    property LongestManifoldName: string read FLongestManifoldName;        { For formatting }
    property LongestSymLength: Integer read FLongestSymLength;             { For formatting }
    property LargestIndex: Integer read FLargestIndex write FLargestIndex; { For formatting }

    property Simulations[AIndex: Integer]: TSimulation read GetSimulation;
    property nSimulations: Integer read FnSimulations;
    property LoadedFormSettings: TDOMDocument read FLoadedFormSettings;
    property LineList: TLineList read FLineList write SetLineList;
  published
    property Units:               TEnergyUnits index 1 read GetEUnits write SetEUnits;
    property PrintLevel:          TPrintLevel index 2 read GetPrintLevel write SetPrintLevel;
    property Precision:           Integer index 3 read GetSetting write SetSetting;     { Precision (of printout) }
    property BasisOrder:          TBasisOrder index 4 read GetBasisOrder write SetBasisOrder;
    property QuantumNumberFormat: TQuantumNumberFormat index 5 read GetQuantumNumberFormat write SetQuantumNumberFormat; { How to display quantum numbers }
    property FitCycles:           Integer index 6 read GetSetting write SetSetting;
    property FitDetails:          Boolean index 7 read GetFlag write SetFlag;
    property AutoReplot:          Boolean index 8 read GetFlag write SetFlag; { Turn off if replot is rather slow }
    property AllowComplex:        Boolean index 9 read GetFlag write SetFlag; { Set to use complex wavefunctions if required }
    property NoLineList:          Boolean index 10 read GetFlag write SetFlag; { Disable use of linelist }
    property ShowEstUnc:          Boolean index 11 read GetFlag write SetEstUnc; { Show estimated uncertainties in line positions }
    property ShowDerivatives:     Boolean index 12 read GetFlag write SetFlag; { Show derivatives in line listings }

    property SVDThresh:   Double index 1 read GetParameter write SetParameter;
    property ScaleChanges:Double index 2 read GetParameter write SetParameter;

  end;

function NotYet:Exception;
procedure MustBeSerial;

procedure ConvertLine(var Line: TLine; UFrom, UTo: TResonanceUnits);
function MakeTxName(var Line: TLine): string;
function AxisName(AAxis: TabcAxis): string;
function LineListWanted(PrintLevel: TPrintLevel):Boolean;
function EstErr(EvU, EvL: TEigenvalues; IndexU, IndexL: Integer):Double;

var
 IndividualSymbolicRotationalMatrixElementTime: string = 'Work out individual symbolic matrix elements';
 SymbolicRotationalMatrixElementTime: string = 'Work out overall symbolic matrix elements';

implementation

uses {$IFNDEF CONSOLE} Fitter, LineFrm, Main, Dialogs, SettingsForm, {$ENDIF}
     {$IFDEF EXPLICITTEST} ExplicitTest, {$ENDIF}
     {$IFDEF MSWINDOWS} Windows, {$ENDIF} { For TRTLCriticalSection }
     Progress, nj, Math, EigLA, CEig, pgXMLWrite, pgXMLRead, Huge,
     Profile, ShellSort, Global, Constrain, PartFunc, OverlaySumFitter,
     UCompat, Worker, Clock, Obs, Blas, OscFields;

function NotYet:Exception;
begin
  Result := Exception.Create('Not Yet Implemented');
end;

procedure MustBeSerial;
begin
  if GetCurrentThreadId <> MainThreadID then
    raise Exception.Create('This procedure is not suitable for more than one processor ('+ThreadName+')');
end;

function LineListWanted(PrintLevel: TPrintLevel):Boolean;
begin
  Result := (PrintLevel >= plLineList) or (PrintLevel in [plCSV, plText, plBinary]);
end;

{ Profiling Variables }
var
 AddMatrixToSpectrumTime: string = 'AddMatrixToSpectrum';
 EigenvaluesTime: string = 'Make and Diagonalize Hamiltonian';
 MakeHamiltonianTime: string = 'Make Hamiltonian Matrix';
 PerturbationMatrixElementTime: string = 'Perturbation Matrix Elements';
 EigenSearchTime: string = 'Identifying eigenvalues';
 EigenvaluesSearchTime: string = 'EigenvaluesSearch';
 MakeLineListTime: string = 'Make Line List';
 AddInLinesTime: string = 'AddInLines';
 TransitionMatrixTime: string = 'Make and Transform Transition Matrix';
 MakeTransitionMatrixTime: string = 'Make Transition Matrix';
 MatrixMultiplyTime: string = 'Matrix Multiply';
 PostMultiplyTime: string = 'Post Multiply';
 TxNameTime: string = 'Transition Name';
 FixupPropertiesTime: string = 'Fixup Properties';
 MakeBasisTime: string = 'Make Basis';
const
  ColourHelpString = 'Colour - set to "None" to take value from elsewhere.';

threadvar
  SymSelNames: array of string; { Must be in static variable }

function SymSettingsNames(MNode: TMnode; out LowestValue: Integer): PPChar;
var
  Molecule: TMolecule;
  i: Integer;
begin
  LowestValue := -1;
  Molecule := TMolecule(MNode.AncestorWithClass(TMolecule));
  if Assigned(Molecule) then begin
    Molecule.PrecomputeSymNames;
    SetLength(SymSelNames, Molecule.LastSym+3);
    for i := 0 to Molecule.LastSym do
      if Molecule.JAdjustSym then
        SymSelNames[i+1] := Molecule.JAdjustedSymName(0,i)
      else
        SymSelNames[i+1] := Molecule.SymName(i);
  end else
    SetLength(SymSelNames, 1);
  SymSelNames[0] := 'all';
  Result := PPChar(@SymSelNames[0]);
end;

{ TLine }

function MakeTxName(var Line: TLine): string;
var
  Mixture: TMixture;
begin
  TStart(TxNameTime);
  with Line do
    if (Length(CachedName) = 0) and Assigned(EigenvaluesL) then begin
      if Assigned(EigenvaluesU) then begin
        Result := EigenvaluesU.Basis[1].TxName(EigenvaluesU.DominantBasis[IndexU], EigenvaluesL.DominantBasis[IndexL]);
        if Length(Result) > 0 then
          Result := Result + ' : ';
      end else
        Result := '';
      Mixture := EigenvaluesL.Manifold.Mixture;
      if Mixture.nMolecules > 1 then
        Result := Result + EigenvaluesL.Manifold.Molecule.Name + ' ';
      if Assigned(EigenvaluesU) then
        Result := Result + EigenvaluesU.DominantBasis[IndexU].StateLabel(Mixture.QuantumNumberFormat) + ' - ';
      Result := Result + EigenvaluesL.DominantBasis[IndexL].StateLabel(Mixture.QuantumNumberFormat);
      CachedName := Result;
    end else
      Result := CachedName;
  TEnd(TxNameTime);
end;

procedure ConvertLine(var Line: TLine; UFrom, UTo: TResonanceUnits);
begin
  if UFrom <> UTo then begin
    Line.Width := ConvertWidth(Line.Width, Line.Frequency, UFrom, UTo);
    Line.Frequency := Convert(Line.Frequency, UFrom, UTo);
  end;
end;

{ TManifold }

function TManifold.GetState(AIndex: Integer): TState;
begin
  {$IFOPT R+}
  Result := FStates[AIndex] as TState;
  {$ELSE}
  Result := TState(FStates[AIndex]);
  {$ENDIF}
end;


function TManifold.GetPerturbation(AIndex: Integer): TPerturbation;
begin
  {$IFOPT R+}
  Result := FPerturbations[AIndex] as TPerturbation;
  {$ELSE}
  Result := TPerturbation(FPerturbations[AIndex]);
  {$ENDIF}
end;

function TManifold.StateWithName(AName: string): TState;
var
  i: Integer;
begin
  Result := nil;
  { NB: can't use FindComponent as Name property has been overridden }
  for i := 0 to ComponentCount-1 do
    if (Items[i] is TState) and (TState(Items[i]).Name = AName) then begin
      Result := TState(Items[i]);
      Exit;
    end;
end;

procedure TManifold.UpdateCounts(AComponent: TComponent; Operation: TOperation);
begin
  FnStates := AdjustNodeIndexes(TState,FStates,AComponent,Operation);
  FnPerturbations := AdjustNodeIndexes(TPerturbation,FPerturbations,AComponent,Operation);
end;

constructor TManifold.Create(AOwner: TComponent);
begin
  FEigenValueCache := TEigenValueCache.Create;
  FEigenValueCache.OwnsObjects := True;
  SavedProperties := SavedProperties + ',AutoQConverge=True,Colour=None,Jmin=-1,Jmax=-1,EigenSearch=True';
  inherited Create(AOwner);
end;

destructor TManifold.Destroy;
begin
  ClearCache;
  FreeAndNil(FEigenValueCache);
  inherited Destroy;
end;

function TManifold.HalfSpin(RemovedSpins: Byte): TQuantumNumber;
begin
  Result := State[0].HalfSpin(RemovedSpins);
end;

procedure TManifold.ClearCache;
begin
  inherited;
  PerformMNodeNotification(Self, mnCacheCleared, 0);
  if Assigned(FEigenValueCache) then
    FEigenValueCache.Clear;
end;

function TManifold.FindColour: TColor;
begin
  Result := Colour;
  if Result = clNone then
    Result := Molecule.Colour;
  if Result = clNone then
    Result := Molecule.Species.Colour;
end;

function TManifold.FindEigenvalues(
  StartFrom: Integer; out LastSearched: Integer;
  J: TQuantumNumber; Sym: Integer;
  const AExField: TStaticFields;
  SelState: TState; IncludePerturbations: Boolean;
  RemoveSpins: Byte): TEigenvalues;
var
  i: Integer;
begin
  { Have we already calculated it ? }
  TStart(EigenvaluesSearchTime);
  { TODO: Optimize Search }
  LastSearched := FEigenValueCache.Count - 1;
  try
    if Assigned(FEigenValueCache.Lock) then
      FEigenValueCache.Lock.BeginRead;
    for i := StartFrom to FEigenValueCache.Count - 1 do begin
      Result := TEigenValues(FEigenValueCache[i]);
      if (Result.J = J) and (Result.Sym = Sym)
         and (Result.SelectedState = SelState)
         and (Result.NoPerturbations <> IncludePerturbations)
         and (Result.RemovedSpins = RemoveSpins)
         and (Result.ExField.E = AExField.E)
         and (Result.ExField.B = AExField.B)
         and (Result.ExField.Angle = AExField.Angle)
        then begin
        TEnd(EigenvaluesSearchTime);
        Exit;
      end;
    end;
  finally
    if Assigned(FEigenValueCache.Lock) then
      FEigenValueCache.Lock.EndRead;
  end;
  Result := nil;
  TEnd(EigenvaluesSearchTime);
end;

function TManifold.Eigenvalues(J: TQuantumNumber; Sym: Integer; MatElemTypes: TMatElemTypes;
                               const AExField: TStaticFields; Wait: Boolean; PrintLevel: TPrintLevel;
                               SelState: TState; IncludePerturbations: Boolean; RemoveSpins: Byte;
                               CreateIfNotFound: Boolean): TEigenvalues;
var
  LastSearched: Integer;
begin
  if PrintLevel = plDefault then
    PrintLevel := Mixture.PrintLevel;
  with Mixture do begin
    if not ShowDerivatives and (meAllDerivatives in MatElemTypes) then
      MatElemTypes := MatElemTypes - [meAllDerivatives];
    if not ShowEstUnc and (meFloatedDerivatives in MatElemTypes) then
      MatElemTypes := MatElemTypes - [meFloatedDerivatives];
  end;
  { Turn off selection if it makes no difference }
  if nStates = 1 then
    SelState := nil;
  if nPerturbations = 0 then
    IncludePerturbations := True;
  if (Molecule.nNuclei = 0) and (RemoveSpins > 0) then
    RemoveSpins := 1; { Electron spin is the only thing that can be removed }
  if IsNonZero(AExField) then
    Sym := -1;
  Result := FindEigenvalues(0, LastSearched, J, Sym, AExField, SelState, IncludePerturbations, RemoveSpins);
  if not Assigned(Result) then
    if CreateIfNotFound then
      try
        if Assigned(FEigenValueCache.Lock) then
          FEigenValueCache.Lock.BeginWrite;
        { Check that it hasn't been created by another thread }
        Result := FindEigenvalues(LastSearched+1, LastSearched, J, Sym, AExField, SelState, IncludePerturbations, RemoveSpins);
        if not Assigned(Result) then begin
          Result := TEigenvalues.Create(Self, J, Sym, AExField, IncludePerturbations, SelState, RemoveSpins);
          FEigenValueCache.Add(Result);
        end;
      finally
        if Assigned(FEigenValueCache.Lock) then
          FEigenValueCache.Lock.EndWrite;
      end
    else
      Exit;
  { We now have the eigenvalue structure, but not necessarily complete }
  Result.PrintLevel := PrintLevel;
  if Result.Size > 0 then
    Result.Make(Wait, MatElemTypes);
end;

function TManifold.FindJmin(Default: Integer; RemovedSpins: Byte): Integer;
var
  HS: Integer;
begin
  Result := Jmin;
  if Result < 0 then
    Result := Molecule.Jmin;
  if Result < 0 then
    Result := Molecule.Species.Jmin;
  if Result < 0 then
    Result := Default;
  HS := HalfSpin(RemovedSpins);
  if Odd(HS+Result) then
    Inc(Result);
  if Result < HS then
    Result := HS
end;

function TManifold.FindJmax(Default: Integer; RemovedSpins: Byte): Integer;
var
  HS: Integer;
begin
  Result := Jmax;
  if Result < 0 then
    Result := Molecule.Jmax;
  if Result < 0 then
    Result := Molecule.Species.Jmax;
  if Result < 0 then
    Result := Default;
  HS := HalfSpin(RemovedSpins);
  if Odd(HS+Result) then
    Dec(Result);
end;

function TManifold.FirstSym(J: TQuantumNumber):Integer;
begin
  Result := 0;
end;

function TManifold.NextSym(J: TQuantumNumber; Sym: Integer): Integer;
begin
  Result := Sym+1;
  if Result >= nSym then
    Result := -1;
end;

function TManifold.Molecule: TMolecule;
begin
  Result := Owner as TMolecule;
end;

procedure TManifold.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited;
  UpdateCounts(AComponent,Operation);
end;

procedure TManifold.ConvertToPerturbations;
var
  i: Integer;
begin
  for i := 0 to nStates-1 do
    State[i].ConvertToPerturbations;
end;

function TManifold.Mixture: TMixture;
begin
  Result := Owner.Owner.Owner as TMixture;
end;

procedure TManifold.SetModified(const Value: Boolean);
begin
  inherited;
  if Value then
    ClearCache;
end;

function TManifold.SettingToPrettyStr(AIndex: Integer): string;
begin
  Result := '';
  if IsSettingQuantum(AIndex) and (Settings[AIndex] < 0) then
    Exit;
  Result := inherited SettingToPrettyStr(AIndex);
end;

function TManifold.FixupProperties:Exception;
var
  i: Integer;
  LastPert, Pert: TPerturbation;
begin
  if (nWorkers > 0) and not Assigned(FEigenValueCache.FLock) then
    FEigenValueCache.FLock := TMultiReadExclusiveWriteSynchronizer.Create;
  Result := inherited FixupProperties;
  UpdateCounts(nil,opInsert);
  FHasWidth := False;
  FMaxSpin := 0;
  for i := 0 to nStates-1 do
    with State[i] do begin
      if HasWidth then
        FHasWidth := True;
      if i = 0 then
        FLowestStateOrigin := StateOrigin
      else
        FLowestStateOrigin := Min(FLowestStateOrigin, StateOrigin);
      FMaxSpin := Max(FMaxSpin, S);
    end;
  LastPert := nil;
  if not Assigned(Result) then
    for i := 0 to ComponentCount - 1 do
      if Components[i] is TPerturbation then begin
        Pert := Components[i] as TPerturbation;
        if Pert.ScalePrev <> 0 then
          if LastPert = nil then
            Result := EReportException.Create(UniqueName+': First perturbation scaled to previous')
          else begin
            Pert.Value := Pert.ScalePrev*LastPert.Value;
            Pert.Floating[1] := False;
          end;
        LastPert := Pert;
      end;
end;

function TManifold.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := AIndex = 4;
  if not Result then
    Result := inherited IsSettingColour(AIndex);
end;

function TManifold.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex in [1,2];
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

const
  ManifoldSettingNames: array[1..7] of string = (
'Minimum J in calculation - set negative to take from molecule or manifold.',
'Maximum J in calculation - set negative to take from molecule or manifold.',
'True to include population in this state when calculating spectra.',
'Colour for spectra - set to "None" to get colour from elsewhere.',
'Identify state by looking for largest coefficient in eigenvector.',
'Set to assume the energy ordering within an individual state does not change, but the eigenvectors are used to select the state.',
'Partition function (Q) sum extends until converged (if true) or Jmax (if false).'
);

function TManifold.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(ManifoldSettingNames))
     and (AIndex <= High(ManifoldSettingNames)) then
    Result := ManifoldSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TManifold.PageName: string;
begin
  Result := 'Manifold'; { These are currently not overridden }
end;

function TManifold.FindIndex(out EigenvaluesF: TEigenvalues;
  out IndexF: Smallint; out SymF: Shortint; FW: TQuantumNumber;
  OtherQno: array of TQuantumNumber; AQuantumNumberFormat: TQuantumNumberFormat;
  const ExField: TStaticFields;
  SymW:{-1} Integer; JW {-1}, NW {-1}: TQuantumNumber; StateW{nil}: TState;
  RemoveSpins:{0} Byte; HyperfineW:{[]} THyperfineBasis;
  EigenvaluesWanted:{True} Boolean; ShowPossible:{False} Boolean): string;
var
  nMatches: Integer;
  Sym, Index: Integer;
  Ev: TEigenvalues;
  function MatchOther(BasisState: TBasisState): Boolean;
  var
    i: Integer;
  begin
    Result := True;
    i := 0;
    while Result and (i <= High(OtherQno)) do begin
      Result := (OtherQno[i] = Low(TQuantumNumber)) or (OtherQno[i] = BasisState.Qno(TQnoType(Ord(qtA)+i)));
      Inc(i);
    end;
    i := 0;
    while Result and (i < High(HyperfineW)) do begin
      Result := (HyperfineW[i] = Low(TQuantumNumber)) or (BasisState.HyperfineState[i] = HyperfineW[i]);
      Inc(i);
    end;
  end;
  function ShowOther(Key: string; A: array of TQuantumNumber):string;
  var
    i: Integer;
    Show: Boolean;
  begin
    Show := False;
    for i := 0 to High(A) do begin
      Show := A[i] <> Low(TQuantumNumber);
      if Show then
        Break;
    end;
    if Show then begin
      Result := Key+'(';
      for i := 0 to High(A) do begin
       if A[i] <> Low(TQuantumNumber) then
         Result := Result + QuantumNumber.QuantumNumberToStr(A[i],AQuantumNumberFormat)
       else
         Result := Result + '?';
       if i <> High(A) then
         Result := Result + ' ';
      end;
      Result := Result + ')';
    end else
      Result := ''
  end;
var
  s, sy: string;
  KeyShown, HasFields, LocalEv: Boolean;
  SymEnd: Integer;
begin
  nMatches := 0; { More than one match is an error }
  if ShowPossible then
    WriteLn(Log, 'Possible Matches:');
  KeyShown := not ShowPossible;
  EigenvaluesF := nil;
  HasFields := IsNonZero(ExField);
  if HasFields then begin
    Sym := -1;
    SymEnd := Sym;
  end else if SymW < 0 then begin
    { SymW = -1 to search all; more negative
      to start search at higher symmetry numbers }
    Sym := -SymW-1;
    SymEnd := Molecule.LastSym;
  end else begin
    Sym := SymW;
    SymEnd := SymW;
  end;
  repeat
    Ev := Eigenvalues(FW, Sym, [meHrot], ExField, True, Mixture.PrintLevel, nil, True, RemoveSpins, EigenvaluesWanted);
    LocalEv := not Assigned(Ev);
    if LocalEv then
      Ev := TEigenvalues.Create(Self, FW, Sym, ExField, True, nil, RemoveSpins);
    try
      for Index := 1 to Ev.Size do
        { Look for matches }
        with Ev.DominantBasis[index] do begin
          if ShowPossible then begin
            if not KeyShown then
              WriteLn(Log, StateLabelKey);
            KeyShown := True;
            Writeln(Log, StateLabel(AQuantumNumberFormat));
          end;
          if ((JW < 0) or (J = JW))
             and ((NW < 0) or (N = NW))
             and (not Assigned(StateW) or (State = StateW))
             and MatchOther(Ev.DominantBasis[index]) then begin
            Inc(nMatches);
            IndexF := Index;
            if nMatches = 1 then
              SymF := Sym; { Record first match }
            Result := StateLabel(AQuantumNumberFormat);
            if EigenvaluesWanted then
              EigenvaluesF := Ev;
          end;
        end;
    finally
      if LocalEv then
        Ev.Free;
    end;
    Inc(Sym);
  until Sym > SymEnd;
  if nMatches <> 1 then begin
    EigenvaluesF := nil;
    IndexF := -1-nMatches; {No unique result found}
    if nMatches=0 then
      SymF := -1
    else
      SymF := -1 - SymF; { Return first match }
    if Assigned(StateW) then
      s := StateW.UniqueName
    else
      s := UniqueName;
    if SymW >= 0 then
      sy := Molecule.SymName(SymW)
    else
      sy := 'Sym=any';
    Result := Format(' %d matches for %s Jtot=%s %s', [nMatches, s, QuantumNumber.QuantumNumberToStr(FW,AQuantumNumberFormat), sy]);
    if JW >= 0 then
      Result := Result + Format(' J=%s', [QuantumNumber.QuantumNumberToStr(JW,AQuantumNumberFormat)]);
    if NW >= 0 then
      Result := Result + Format(' N=%s', [QuantumNumber.QuantumNumberToStr(NW,AQuantumNumberFormat)]);
    Result := Result + ShowOther(' ('+Molecule.QnoInfo(qtA).Name+ ' '+Molecule.QnoInfo(qtB).Name+')=', OtherQno);
    Result := Result + ShowOther(' Hyperfine=', HyperfineW);
  end;
end;

{ Given some information about a state, try and find a complete state label and symmetry
  for identifying that state. Notes:
  1. The hyperfine basis excludes the total angular momentum
  2. An index is returned, but does not use the full eigenvalue matching;
     Use FindIndex for this puprpose }
function TManifold.FindStateLabel(
  out IndexF: Smallint; out SymF: Shortint;
  FW: TQuantumNumber; OtherQno: array of TQuantumNumber;
  AQuantumNumberFormat: TQuantumNumberFormat;
  const ExField: TStaticFields;
  SymW:{-1}Integer; JW {-1}, NW {-1}: TQuantumNumber; StateW:{nil}TState;
  RemoveSpins:{0}Byte; HyperfineW: THyperfineBasis;
  ShowPossible:{False} Boolean): string;
var
  Unused: TEigenvalues;
begin
  Result := FindIndex(Unused,IndexF, SymF, FW, OtherQno, AQuantumNumberFormat, ExField,
                      SymW, JW, NW, StateW, RemoveSpins, HyperfineW, False, ShowPossible);
end;

function TManifold.EncodeParameterName(J: TQuantumNumber; Sym,
  EvNo: Integer): string;
begin
  Result := Format('%s %s %d', [QuantumNumberToStr(J), Molecule.JAdjustedSymName(J,Sym), EvNo]);
end;

function TManifold.nPopulations: Integer;
begin
  Result := nParameters;
end;

function TManifold.PopulationIndex(J: TQuantumNumber; Sym, EvNo: Integer; AutoCreate: Boolean = False): Integer;
var
  PNumber, t: LongInt;
  f, l, n: Integer;
begin
  { Find index corresponding to this parameter, or create it if it doesn't exist }
  { Calculate sorting index for speed }
  if EvNo >= 32768 then
    raise Exception.Create('PopulationIndex: EvNo too large');
  PNumber := (J*(Molecule.LastSym+1)+Molecule.JAdjustedSym(J,Sym))*32768+EvNo;
  Result := -1;
  if nParameters > 0 then begin
    { Binary search for index }
    f := 1;
    l := nParameters;
    while f <= l do begin
      n := (f + l) div 2;
      t := PNumber - PNumbers[n];
      if t = 0 then begin
        Result := n;
        Break;
      end else if t < 0 then
        l := n-1 { In first half }
      else
        f := n+1;
    end;
    if Result < 0 then
      if t < 0 then
        Result := -n
      else
        Result := -(n+1);
  end;
  if (Result <= 0) and AutoCreate then begin
    SetLength(PNumbers, nParameters+2);
    Result := -Result;
    AddParameter(EncodeParameterName(J,Sym,EvNo), 0, Result);
    if Result < nParameters then
      Move(PNumbers[Result], PNumbers[Result+1], (nParameters-Result)*SizeOf(PNumbers[Result]));
    PNumbers[Result] := PNumber;
//    WriteLn(Result:5, PNumber:20, ' ', EncodeParameterName(J,Sym,EvNo));
//    for f := 1 to High(Pnumbers) do
//      Write(' ', PNumbers[f]);
//    WriteLn;
  end;
//   else if (Result > 0) and (EncodeParameterName(J,Sym,EvNo) <> ParameterNames[Result]) then
//    raise Exception.Create('Help')
//  else if (Result <=0) and (IndexOfParameter(EncodeParameterName(J,Sym,EvNo)) > 0) then
//     Result := PopulationIndex(J, Sym, EvNo, False);
end;

var
  BoltzMax: Double; { = Ln(MaxDouble)/2 }

procedure TManifold.AddAllLevels(AState: TState);
var
  J, JUp: TQuantumNumber;
  Sym, l, Ind: Integer;
  Ev: TEigenvalues;
  Pops: array of Double;
  Syms: array of Integer;
  Overflow: Exception;
  Simulation: TSimulation;
begin
  Simulation := Mixture.Simulations[0];
  Overflow := nil;
  J := FindJmin(HalfSpin);
  JUp := FindJmax(J+40);
  while J <= JUp do begin
    Sym := 0;
    while Sym < nSym do begin
      Ev := Eigenvalues(J, Molecule.JAdjustedSym(J, Sym), [meHrot], Simulation.ExField);
      { Calculate populations and required states here, as adding
        custom populations will discard eigenvalues }
      SetLength(Pops, Ev.Size);
      SetLength(Syms, Ev.Size);
      for l := 1 to Ev.Size do begin
        Pops[l-1] := Population(Ev, l, Simulation, Overflow, 0);
        if not Assigned(AState) or (AState = Ev.DominantBasis[l].State) then
          Syms[l-1] := Molecule.JAdjustedSym(J, Sym)
        else
          Syms[l-1] := -1;
      end;
      for l := 1 to Ev.Size do
        if Syms[l-1] >= 0 then begin
          Ind := PopulationIndex(J,Syms[l-1],l,True);
          Parameters[Ind] := Pops[l-1];
        end;
      Inc(Sym);
    end;
    Inc(J, 2);
  end;
  if Assigned(Overflow) then
    raise Overflow;
end;

function TManifold.bf(T, E: Double): Double;
begin
  Result := -E/T/BoltzmannConstant;
  if Result > BoltzMax then
    Result := -1
  else
    Result := Exp(Result);
end;

function TManifold.Population(Eigenvalues: TEigenvalues; i: Integer; Simulation: TSimulation; var Overflow: Exception; IncludeTspin: Integer): Double;
var
  EBottom: Double;
  AState: TState;
  PIndex: Integer;
  function bfp(T, E: Double): Double;
  begin
    Result := bfp(T,E);
    if Result < 0 then begin
      if not Assigned(Overflow) then
        Overflow := EReportException.CreateFmt('Overflow in Bolzmann for %s: E=%g',
                           [Eigenvalues.DominantBasis[i].StateLabel(Mixture.QuantumNumberFormat), E]);
      Result := 0;
    end;
  end;
begin
  EBottom := Molecule.LowestOrigin;
  if Simulation.Tvib > 0 then begin
    { Vibrational temperature calc }
    AState := Eigenvalues.DominantBasis[i].State;
    Result := bf(Simulation.Tvib, AState.StateOrigin-EBottom);
    EBottom := AState.StateOrigin;
  end else
    Result := 1;
  if Simulation.Tspin >= 0 then
    case IncludeTspin of
      0: Result := Result * Molecule.SpinSpeciesAbundance(Simulation, EigenValues.SpinSpecies(i),Overflow);
      1: begin
           { Calculating at Tspin }
           Result := Result * bf(Simulation.Tspin, Eigenvalues.E[i] - EBottom);
           Exit; { Skip normal population calculation }
         end;
    { else skip use of SpinSpeciesAbundance }
    end;
  if Simulation.Temperature > 0 then
    { Normal Boltzman calculation }
    Result := Result * bf(Simulation.Temperature, Eigenvalues.E[i] - EBottom)
  else
    with Eigenvalues do begin
      { TODO: This may need speeding up - optimize J range in AddManifoldToSpectrum, for example }
      PIndex := PopulationIndex(J,Sym,i); { TODO -cField: Fix with Field }
      if PIndex > 0 then
        Result := Result * Parameters[PIndex]
      else
        Result := 0;
    end;
end;

procedure TManifold.DecodeParameterName(s: string; out J: TQuantumNumber;
  out Sym, EvNo: Integer; AQuantumNumberFormat: TQuantumNumberFormat);
var
  p: Integer;
begin
  p := Pos(' ',s);
  if (p > 1) and TryStrToQuantumNumber(Copy(s,1,p-1),J,AQuantumNumberFormat) then begin
    Delete(s,1,p);
    p := Pos(' ',s);
    if Molecule.TryStrToAnySymName(J,Trim(Copy(s,1,p-1)),Sym)
       and TryStrToInt(Copy(s,p+1,255),EvNo) then
      Exit;
  end;
  raise EReportException.Create('Bad population element:'+s);
end;

function TManifold.ReadProperties(Node: TDOMNode; CheckUnread: Boolean):Integer;
var
  Child, NextChild: TDOMElement;
  Attr: TDOMAttr;
  J: TQuantumNumber;
  Sym, Index, EvNo: Integer;
  QnFmt: TQuantumNumberFormat;
begin
  Result := inherited ReadProperties(Node, False);
  Child := Node.FirstChild as TDOMElement;
  if Mixture.ReadVersion < 5002528 then
    QnFmt := qft2J
  else
    QnFmt := QuantumNumberFormat;
  if Assigned(Child) then
    Molecule.FixupProperties.Free; { Required for parameter name decoding to work }
  { Read (and create) dynamic parameters containing population }
  while Assigned(Child) do begin
    NextChild := Child.NextSibling as TDOMElement;
    if CompareText(Child.NodeName, 'Parameter') = 0 then begin
      Attr := (Child as TDOMElement).GetAttributeNode('Name');
      if Assigned(Attr) then begin
        DecodeParameterName(Attr.NodeValue,J,Sym,EvNo,QnFmt);
        Index := PopulationIndex(J,Sym,EvNo,True);
        Result := Result + ReadFloatableDOMNode(Index,'0',Child);
        Node.RemoveChild(Child);
      end;
    end;
    Child := NextChild;
  end;
  if CheckUnread then
    Result := Result + nUnreadProperties(Node)
end;

function TManifold.WriteProperties(Parent: TDOMNode; AllParameters: Boolean; WithChildren: Boolean): TDOMElement;
//var
//  i: Integer;
//  Written: array of Boolean;
//  Child: TDOMElement;
//  Attr: TDOMAttr;
//  J: TQuantumNumber;
//  Sym, Index, EvNo: Integer;
begin
  Result := inherited WriteProperties(Parent, True, WithChildren);
// No longer needed, as all dynamic parameters now written by default
//  { Force write of all populations, even if set to then default }
//  { Cycle through all parameters, checking for those written }
//  SetLength(Written, nParameters+1);
//  FillChar(Written[0],Sizeof(Boolean)*(nParameters+1),0);
//  Child := Result.FirstChild as TDOMElement;
//  while Assigned(Child) do begin
//    if CompareText(Child.NodeName, 'Parameter') = 0 then begin
//      Attr := (Child as TDOMElement).GetAttributeNode('Name');
//      if Assigned(Attr) then begin
//        DecodeParameterName(Attr.NodeValue,J,Sym,EvNo,QuantumNumberFormat);
//        Index := PopulationIndex(J,Sym,EvNo,False);
//        if Index < 0 then
//          raise Exception.Create('TManifold.WriteProperties'); { Help! }
//        Written[Index] := True;
//      end;
//    end;
//    Child := Child.NextSibling as TDOMElement;
//  end;
//  { Now write all those that haven't been written already }
//  for i := FirstDynamicParameter to nParameters do
//    if not Written[i] then
//      WriteFloatableValue(i,'1',Result);
end;

function TManifold.HelpOnParameter(AIndex: Integer): string;
var
  Eig: TEigenvalues;
  J: TQuantumNumber;
  Sym, EvNo, LastSearched: Integer;
begin
  if (AIndex=0) or (AIndex > nParameters) then
    Exit;
  Result := 'Population for J=' + ParameterNames[AIndex] + ' level';
  DecodeParameterName(ParameterNames[AIndex],J,Sym,EvNo,QuantumNumberFormat);
  Eig := FindEigenvalues(0,LastSearched,J,Sym,Mixture.Simulations[0].ExField);
  if Assigned(Eig) then
    if Eig.Size < EvNo then
      Result := Result + ' (non existent level)'
    else
      Result := Result + ':' + Eig.GetDominantBasisState(EvNo).StateLabel(Mixture.QuantumNumberFormat);
end;

{ TDistribution }

{ Special purpose transition moment that shows a population
  distribution only }

type
  TDistribution = class(TTransitionMoments)
  public
    procedure AddToSpectrum(Spectrum: TSpectrum; Ev1, Ev2: TEigenvalues; ATransitionMoment: TTransitionMoment; Simulation: TSimulation; Wait: Boolean; PrintLevel: TPrintLevel; const Excitation: TExcitation; var Problem: Exception; IncludeZero: Boolean = False); override;
    function MakeLine(out Line: TLine; D: TMatrix; Ev1, Ev2: TEigenvalues; Index1, Index2: Integer; const Excitation: TExcitation; IncludeZero: Boolean = false):Boolean; override;
    function Allowed(J1: TQuantumNumber; Sym1: Integer; J2: TQuantumNumber; Sym2: Integer): Boolean; override;
    function MaxRank: TQuantumNumber; override;
  end;

var
  PopulationConnection: TDistribution;

procedure TDistribution.AddToSpectrum(Spectrum: TSpectrum; Ev1,
  Ev2: TEigenvalues; ATransitionMoment: TTransitionMoment;
  Simulation: TSimulation;
  Wait: Boolean; PrintLevel: TPrintLevel;
  const Excitation: TExcitation; var Problem: Exception; IncludeZero: Boolean);
var
  i, ColourI: Integer;
  te, wt, str: Double;
  Line: TLine;
begin
  if not Wait then
    Exit;
  if Ev1 <> Ev2 then
    raise Exception.Create('TDistribution.AddToSpectrum');
  TStart(AddMatrixToSpectrumTime);
  Line.EigenvaluesL := Ev1;
  Line.EigenvaluesU := nil;
  for i := 1 to Ev1.Size do begin
    { For each state }
    wt := Ev1.StatisticalWeight(i) * Ev1.Manifold.Molecule.Abundance * Ev1.Manifold.Molecule.Species.Concentration;
    Line.IndexL := i;
    Line.IndexU := -1;
    ColourI := Colour;
    if ColourI = clNone then
      ColourI := Ev1.DominantBasis[i].State.FindColour;
    te := Ev1.E[i];
    Str := wt;
    Line.Frequency := te;
    Line.Strength := Str;
    Line.Width := 0;
    if Assigned(Ev1.Widths) then
      Line.Width := Line.Width + Ev1.Widths[i];
    if ColourI = clNone then
      ColourI := clBlack;
    Line.Colour := ColourI;
    if LineListWanted(PrintLevel) then begin
      Line.CachedName := '';
      Spectrum.ShowLine(Line, Log, PrintLevel, Mixture.Simulations[0].OThreshold, False, Problem);
      Line.CachedName := '';
    end;
    Spectrum.AddLine(Line, Ev1.Manifold.Mixture.Simulations[0], Ev1.Manifold.Mixture.Units, Problem);
  end;
  TEnd(AddMatrixToSpectrumTime);
end;

function TDistribution.Allowed(J1: TQuantumNumber; Sym1: Integer;
  J2: TQuantumNumber; Sym2: Integer): Boolean;
begin
  Result := (J1 = J2) and (Sym1 = Sym2);
end;

function TDistribution.MakeLine(out Line: TLine; D: TMatrix; Ev1, Ev2: TEigenvalues; Index1, Index2: Integer; const Excitation: TExcitation; IncludeZero: Boolean = false): Boolean;
begin
  raise NotYet;
end;

function TDistribution.MaxRank: TQuantumNumber;
begin
  Result := 0;
end;

{ TPartitionFunction }

constructor TPartitionFunction.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Active=True';
  inherited Create(AOwner);
end;

function TPartitionFunction.EnergyZero(LowestOrigin: Double): Double;
begin
  Result := LowestOrigin;
end;

function TPartitionFunction.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'If true, use this to generate the partition function'
  else
    Result:=inherited HelpOnSetting(AIndex);
end;

{ TMolecule }

constructor TMolecule.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Colour=None,JAdjustSym=True,Abundance=1,Jmin=-1,Jmax=-1,BlockMatrix=True';
  DimensionlessParameters := DimensionlessParameters + ',Abundance';
  FLastPartitionFunction := -99;
  FnSpinSpecies := 1;
  inherited Create(AOwner);
end;

destructor TMolecule.Destroy;
begin
  FreeAndNil(FDefaultPartitionFunction);
  FActivePartitionFunction := nil;
  inherited Destroy;
end;

function TMolecule.EinsteinA(const Line: TLine; AUnits: TResonanceUnits): Double;
begin
  { TODO: May need adjustment in the presence of a field; needs documentation either way }
  Result := Line.Strength * IntensityFactor(Line.Frequency, AUnits, HzperMolecule)
            / Line.EigenvaluesU.Degeneracy(Line.IndexU)
end;

function TMolecule.GetManifold(AIndex: Integer): TManifold;
begin
  {$IFOPT R+}
  Result := FManifolds[AIndex] as TManifold;
  {$ELSE}
  Result := TManifold(FManifolds[AIndex]);
  {$ENDIF}
end;

function TMolecule.ReadProperties(Node: TDOMNode; CheckUnread: Boolean):Integer;
begin
  Result := inherited ReadProperties(Node, CheckUnread);
  FixupProperties.Free; { Fixup required for some post processing (which?) }
end;

function TMolecule.FixupProperties:Exception;
var
  n, i, ii, j, nLower, nNuc: Integer;
  Carry, New: TQuantumNumber;
  HalfSpinDetermined, Reject: Boolean;
  LowerLim, UpperLim: array of TQuantumNumber;
begin
  Result := nil;
  FnTransitionMoments := 0;
  FMaxRank := 0;
  FActivePartitionFunction := nil;
  FnStates := 0;
  { Compute nManifolds and check for duplicate names }
  for i := 0 to ComponentCount-1 do begin
    if Items[i] is TManifold then begin
      for j := i+1 to ComponentCount-1 do
        if (Items[j] is TManifold) and (Items[i].Name = Items[j].Name) then
          Result := EReportException.Create('Duplicate Manifold Names:'+ Items[i].Name)
    end else if Items[i] is TTransitionMoments then begin
      Inc(FnTransitionMoments);
      FMaxRank := Max(MaxRank, TTransitionMoments(Items[i]).MaxRank);
    end;
  end;
  if not Assigned(Result) then
    Result := inherited FixupProperties;
  if nManifolds = 0 then Exit;
  nLower := 0;
  FLowestOrigin := Manifold[0].LowestStateOrigin; { Not necessarily valid, but avoid awkward numbers }
  FLongestManifoldName := '';
  HalfSpinDetermined := False;
  FAllSinglet := True;
  for i := 0 to nManifolds-1 do begin
    Inc(FnStates, Manifold[i].nStates);
    if Manifold[i].nStates > 0 then
      if not HalfSpinDetermined then begin
        HalfSpinDetermined := True;
        FHalfSpin := Manifold[i].HalfSpin;
      end else if (Result = nil) and (FHalfSpin <> Manifold[i].HalfSpin) then
        Result := EReportException.Create('Odd and even numbers of electrons in ' + Name);
    if Manifold[i].Initial and ((nLower=0) or (Manifold[i].LowestStateOrigin < FLowestOrigin)) then
      FLowestOrigin := Manifold[i].LowestStateOrigin;
    if Manifold[i].Initial then
      Inc(nLower);
    if Length(Manifold[i].Name) > Length(LongestManifoldName) then
      FLongestManifoldName := Manifold[i].Name;
    for j := 0 to Manifold[i].nStates - 1 do
      FAllSinglet := FAllSinglet and (Manifold[i].State[j].S=0);
  end;
  FLowestOrigin := ActivePartitionFunction.EnergyZero(FLowestOrigin);
  { Make up HyperfineBasisStates }
  { First Determine total number of spin combinations }
  n := 1;
  FSumNuclearSpin := 0;
  SetLength(HyperfineBasisStates,nNuclei+1);
  SetLength(HyperfineBasisStates[0],1); { nNuclei = 0 -> Empty array }
  SetLength(LowerLim, nNuclei);
  SetLength(UpperLim, nNuclei);
  for i := 0 to nNuclei-1 do begin
    n := n * (Nuclei[i].Spin+1);
    Inc(FSumNuclearSpin,Nuclei[i].Spin);
    if Nuclei[i].AsNext then begin
      { Two coupled nuclei - must treat as pair }
      SetLength(HyperfineBasisStates[i+1],0); { Flag invalid as can't omit just one of the spins }
      if not Assigned(Result) then
        if i = nNuclei-1 then
          Result := EReportException.Create('Last nucleus for '+ Name + ' must not be set equivalent')
        else if Nuclei[i+1].AsNext then
          Result := EReportException.Create('More than two equivalent nuclei not implemented in '+ Name);
      { For I12 = I1 + I2, just record quantum number rather than offset }
      LowerLim[i] := Abs(Nuclei[i].Spin-Nuclei[i+1].Spin);
      UpperLim[i] := Nuclei[i].Spin+Nuclei[i+1].Spin;
    end else begin
      if (i > 0) and Nuclei[i-1].AsNext then
        { Two equivalent nuclei require special handling }
        UpperLim[i] := UpperLim[i-1]
      else
        UpperLim[i] := Nuclei[i].Spin;
      LowerLim[i] := -UpperLim[i];
      SetLength(HyperfineBasisStates[i+1],n,i+1);
    end;
  end;
  if FSumNuclearSpin = 0 then
    SetLength(HyperfineBasisStates,1)
  else
    { Now make up different combinations }
    for nNuc := 1 to nNuclei do
      if not Nuclei[nNuc-1].AsNext then begin
        for i := 0 to nNuc-1 do
          if (i > 0) and Nuclei[i-1].AsNext then
            HyperfineBasisStates[nNuc,0,i] := Max(LowerLim[i], -HyperfineBasisStates[nNuc,0,i-1])
          else
            HyperfineBasisStates[nNuc,0,i] := LowerLim[i];
        for i := 1 to High(HyperfineBasisStates[nNuc]) do begin
          ii := i-1;
          repeat
            Carry := 2;
            Reject := False;
            for j := nNuc-1 downto 0 do begin
              New := HyperfineBasisStates[nNuc,ii,j] + Carry;
              if New > UpperLim[j] then
                New := LowerLim[j]
              else
                Carry := 0;
              HyperfineBasisStates[nNuc,i,j] := New;
              Reject := Reject or (Nuclei[j].AsNext and (Abs(HyperfineBasisStates[nNuc,i,j+1]) > New));
            end;
            ii := i;
          until not Reject;
        end;
      end;
  FSymLength := 1;
  for i := 0 to LastSym do begin
    FSymLength := Max(FSymLength, Length(JAdjustedSymName(0,i)));
    FSymLength := Max(FSymLength, Length(JAdjustedSymName(2,i)));
    FSymLength := Max(FSymLength, Length(SymName(i)));
  end;
end;

function TMolecule.PartitionFunction(Simulation: TSimulation; var Overflow: Exception):Double;
var
  LocalOverflow: Exception;
begin
  { Note - relies on ClearCache being triggered on TPartitionFunction change }
  if (FLastPartitionFunction >= 0) and (Simulation.Temperature = FLastTemperature)
     and (Simulation.Tvib = FLastTvib) and (Simulation.Tspin = FLastTspin) then begin
    Result := FLastPartitionFunction;
    if (Length(FLastProblem) > 0) and not Assigned(Overflow) then
      Overflow := EReportException.Create(FLastProblem);
  end else begin
    FLastTemperature := Simulation.Temperature;
    FLastTvib := Simulation.Tvib;
    FLastTspin := Simulation.Tspin;
    if (nSpinSpecies > 1) and (FLastTspin >= 0) then
      SetLength(FSpinSpeciesAbundance,nSpinSpecies)
    else
      SetLength(FSpinSpeciesAbundance,0);
    FLastPartitionFunction := ActivePartitionFunction.CalculatePartitionFunction(Self,Simulation,LocalOverflow,FSpinSpeciesAbundance);
    Result := FLastPartitionFunction;
    if Result = 0 then
      raise EReportException.Create('Zero partion function');
    if Assigned(LocalOverflow) then begin
      FLastProblem := LocalOverflow.Message;
      if Assigned(Overflow) then
        LocalOverflow.Free
      else
        Overflow := LocalOverflow;
    end else
      FLastProblem := '';
  end;
end;

procedure TMolecule.PrecomputeSymNames;
begin
  { No action by default }
end;

function TMolecule.QnoInfo(QnoType: TQnoType): TQnoInfo;
begin
  with Result do begin
    MaxValue := High(Integer);
    if QnoType = qtM then
      MinValue := -MaxValue
    else
      MinValue := 0; { TODO: Switch on F number }
    StepSize := 2;
    Name := QnoTypeNames[Ord(QnoType)];
  end;
end;

function TMolecule.AddToSpectrum(
  Spectrum: TSpectrum; PrintLevel: TPrintLevel;
  const Excitation: TExcitation;
  var Problem: Exception;
  Simulation: TSimulation;
  { Note that all the following parameters have a default, which implies all possible values }
  J1Sel, Sym1Sel, J2Sel, Sym2Sel: Integer;
  Manifold1Sel, Manifold2Sel: TManifold;
  DeltaJ: TQuantumNumber;
  TransitionMomentsSel: TTransitionMoments;
  TransitionMomentSel: TTransitionMoment;
  State1Sel, State2Sel: TState;
  IncludePerturbations: Boolean;
  IncludeZero: Boolean):Boolean;
var
  i, j, Sym1, Sym2: Integer;
  J1, J1Low, J1Up: TQuantumNumber;
  J2, J2Low, J2Loww, J2Up, J2Upp: TQuantumNumber;
  Prog1, Prog1Max, Prog2, Prog2Max: Integer; { for progress display }
  MaxDj: TQuantumNumber;
  Manifold1, Manifold2: TManifold;
  Connection: TTransitionMoments;
  Ev1, Ev2: TEigenvalues;
  nInitial, nConnections: Integer;
  HasFields: Boolean;
  PassMin, Pass: Integer;
  Status: string;
  MatElemTypes: TMatElemTypes;
begin
  if Abundance = 0 then Exit;
  nInitial := 0;
  nConnections := 0;
  HasFields := Simulation.HasFields;
  if not HasFields then begin
    if J1Sel < 0 then
      J1Sel := Low(Integer);
    if J2Sel < 0 then
      J2Sel := Low(Integer);
  end;
  MatElemTypes := [meHrot, meWidth, meAllDerivatives, meFloatedDerivatives];
  { If parallel processing, make a prepass to fill eigenvalue and transition matrix caches }
  if (nWorkers > 0) and (PrintLevel <= plLineList) then
    PassMin := 1
  else
    PassMin := 3;
  for Pass := PassMin to 3 do begin
    for i := 0 to nManifolds-1 do
      if (Manifold[i] = Manifold1Sel)
         or (not Assigned(Manifold1Sel) and Manifold[i].Initial ) then begin
        { For each manifold marked as initial }
        Manifold1 := Manifold[i];
        Inc(nInitial);
        for j := 0 to nManifolds-1 do
          if (Manifold[j] = Manifold2Sel)
             or (not Assigned(Manifold2Sel) and ((j >= i) or not Manifold[j].Initial)) then begin
            { For every other manifold, except those we have already started from }
            Manifold2 := Manifold[j];
            if Spectrum.IsDistribution then
              if Manifold1 = Manifold2 then begin
                Connection := PopulationConnection;
                { Ignore selection of "other" state }
                TransitionMomentSel := nil;
                J2Sel := J1Sel;
                Sym2Sel := Sym1Sel;
                State2Sel := State1Sel;
              end else
                Connection := nil { Only need one state for a population distribution }
            else
              Connection := TransitionMoments[i, j];
            if Assigned(Connection)
              and ((Connection = TransitionMomentsSel) or not Assigned(TransitionMomentsSel)) then begin
              Inc(nConnections);
              if Assigned(TransitionMomentSel) then
                MaxDj := TransitionMomentSel.Rank
              else
                MaxDj := Connection.MaxRank;
              if J1Sel <> Low(Integer) then begin
                J1Low := J1Sel;
                J1Up := J1Sel;
                if J2Sel <> Low(Integer) then begin
                  J2Low := J2Sel;
                  J2Up := J2Sel;
                end else if Abs(DeltaJ) < 100 then begin
                  J2Low := J1Low + DeltaJ;
                  J2Up := J2Low;
                  if not HasFields and (J2Low < 0) then Exit;
                end else begin
                  J2Low := J1Sel - MaxDj;
                  if not HasFields then
                    J2Low := SmallestProjection(J2Low);
                  J2Up := J1Sel + MaxDj;
                end;
              end else if J2Sel <> Low(Integer) then begin
                J2Low := J2Sel;
                J2Up := J2Sel;
                if Abs(DeltaJ) < 100 then begin
                  J1Low := J2Low - DeltaJ;
                  J1Up := J1Low;
                  if not HasFields and (J1Low < 0) then Exit;
                end else begin
                  J1Low := J2Sel - MaxDj;
                  if not HasFields then
                    J1Low := SmallestProjection(J2Low);
                  J1Up := J2Sel + MaxDj;
                end;
              end else begin
                J1Up := Manifold1.FindJmax(Manifold1.Molecule.Species.Jmax);
                if HasFields then
                  J1Low := -J1Up
                else
                  J1Low := Manifold1.FindJmin(HalfSpin);
                J2Up := Manifold2.FindJmax(J1Up + MaxDj);
                if HasFields then
                  J2Low := -J2Up
                else
                  J2Low := Manifold2.FindJmin(J1Low - MaxDj);
              end;
              J1 := J1Low;
              case Pass of
                1: Status := 'Energy Levels ';
                2: Status := 'Transition Matrices ';
                3: Status := '';
              end;
              PushProgress(Status + Manifold2.Name + ' - ' + Manifold1.Name, J1Up div 2);
              try
                while J1 <= J1Up do begin
                  Prog1 := (J1 - J1Low) div 2;
                  Prog1Max := (J1Up - J1Low + 2) div 2;
                  if HasFields then
                    Sym1 := 0
                  else if Sym1Sel >= 0 then begin
                    Sym1 := Sym1Sel;
                    if JAdjustSym then
                      Sym1 := JAdjustedSym(J1, Sym1);
                  end else begin
                    Sym1 := Manifold1.FirstSym(J1);
                    Prog1 := Prog1*Manifold1.nSym + Sym1;
                    Prog1Max := Prog1Max*Manifold1.nSym;
                  end;
                  while Sym1 >= 0 do begin
                    Ev1 := Manifold1.Eigenvalues(J1, Sym1, MatElemTypes, Simulation.ExField, Pass >= 2, PrintLevel, State1Sel, IncludePerturbations, 0);
                    if Ev1.Size > 0 then begin
                      if Abs(DeltaJ) < 100 then begin
                        J2 := J1 + DeltaJ;
                        J2Upp := J2;
                      end else begin
                        J2 := J1-MaxDj;
                        J2Upp := Min(J2Up,J1+MaxDj);
                      end;
                      if J2 < J2Low then J2 := J2Low;
                      J2Loww := J2;
                      while J2 <= J2Upp do begin
                        Prog2 := (J2 - J2Loww) div 2;
                        Prog2Max := (J2Upp - J2Loww + 2 ) div 2;
                        if HasFields then
                          Sym2 := 0
                        else if Sym2Sel >= 0 then begin
                          Sym2 := Sym2Sel;
                          if JAdjustSym then
                            Sym2 := JAdjustedSym(J2, Sym2);
                        end else begin
                          Sym2 := Manifold2.FirstSym(J2);
                          Prog2 := Prog2*Manifold2.nSym + Sym2;
                          Prog2Max := Prog2Max*Manifold2.nSym;
                        end;
                        SetProgress((Prog1*Prog1Max+Prog2)*2,Prog1Max*Prog2Max*2);
                        while Sym2 >= 0 do begin
                          if HasFields or Connection.Allowed(J1, Sym1, J2, Sym2) then begin
                            SetProgress((Prog1*Prog1Max+Prog2)*2,Prog1Max*Prog2Max*2);
                            Ev2 := Manifold2.Eigenvalues(J2, Sym2, MatElemTypes, Simulation.ExField, Pass >= 2, PrintLevel, State2Sel, IncludePerturbations, 0);
                            if (Ev2.Size > 0) and (Pass >= 2) then
                              Connection.AddToSpectrum(Spectrum, Ev1, Ev2, TransitionMomentSel, Simulation, Pass = 3, PrintLevel, Excitation, Problem, IncludeZero);
                            SetProgress((Prog1*Prog1Max+Prog2)*2+1,Prog1Max*Prog2Max*2);
                          end;
                          Inc(Prog2);
                          if HasFields or (Sym2Sel >= 0) then
                            Break
                          else
                            Sym2 := Manifold2.NextSym(J2, Sym2);
                        end;
                        Inc(J2, 2);
                      end;
                    end;
                    Inc(Prog1);
                    if HasFields or (Sym1Sel >= 0) then
                      Break
                    else
                      Sym1 := Manifold1.NextSym(J1, Sym1);
                  end;
                  Inc(J1, 2);
                end;
              finally
                PopProgress;
              end;
            end;
          end;
      end;
  end;
  if (nManifolds > 0) and (nInitial = 0) then
      raise EReportException.Create('No Initial States set for '+ Name);
  Result := nConnections > 0;
end;

function TMolecule.ManifoldWithName(AName: string): TManifold;
var
  i: Integer;
begin
  Result := nil;
  { NB: can't use FindComponent as Name property has been overridden }
  for i := 0 to ComponentCount-1 do
    if (Items[i] is TManifold) and (TManifold(Items[i]).Name = AName) then begin
      Result := TManifold(Items[i]);
      Exit;
    end;
end;

function TMolecule.FindTransitionMoments(ManifoldA, ManifoldB: TManifold): TTransitionMoments;
var
  i: Integer;
  tms: TTransitionMoments;
begin
  Result := nil;
  for i := 0 to ComponentCount-1 do
    if Items[i] is TTransitionMoments then begin
      tms := TTransitionMoments(Items[i]);
      if ((tms.Bra = ManifoldA) and (tms.Ket = ManifoldB))
         or ((tms.Bra = ManifoldB) and (tms.Ket = ManifoldA)) then begin
          if Assigned(Result) then
            raise EReportException.Create('Duplicate transition moments <' + ManifoldA.Name + '|' + Result.PairName + '|' + ManifoldB.Name + '> in '+Name);
          Result := TTransitionMoments(Items[i]);
      end;
    end;
end;

function TMolecule.GetTransitionMomentsFunction(AIndex, BIndex: Integer): TTransitionMoments;
begin
  { TODO: Speed up search? }
  Result := FindTransitionMoments(Manifold[AIndex], Manifold[BIndex]);
end;

function TMolecule.Species: TSpecies;
begin
  Result := Owner as TSpecies;
end;

function TMolecule.SpinSpeciesAbundance(Simulation: TSimulation; Index: Integer; var Overflow: Exception): Double;
begin
  if nSpinSpecies > 1 then begin
    PartitionFunction(Simulation,Overflow); { Ensure caches filled }
    Result := FSpinSpeciesAbundance[Index-1];
  end else
    Result := 1;
end;

function TMolecule.GetMixture: TMixture;
begin
  Result := Owner.Owner as TMixture;
end;

function TMolecule.GetTransitionMoments(AIndex: Integer): TTransitionMoments;
var
  i: Integer;
begin
  Result := nil;
  if nTransitionMoments < 1 then Exit;
  for i := 0 to ComponentCount-1 do
    if Items[i] is TTransitionMoments then begin
      Dec(AIndex);
      if AIndex < 0 then begin
        Result := TTransitionMoments(Items[i]);
        Exit;
      end;
    end;
end;

procedure TMolecule.SetModified(const Value: Boolean);
begin
  inherited;
  if Value then
    ClearCache;
end;

procedure TMolecule.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  FnManifolds := AdjustNodeIndexes(TManifold,FManifolds,AComponent,Operation);
end;

function TMolecule.SymName(Sym: Integer): string;
begin
  if Sym < 0 then
    Result := '-' { Symmetry not used, typically for external field }
  else
    Result := IntToStr(Sym);
end;

function TMolecule.JAdjustedSymName(J: TQuantumNumber; Sym: Integer): string;
begin
  if JAdjustSym then
    Result := JAdjustedSymNameBase(J, Sym)
  else
    Result := SymName(Sym);
end;

function TMolecule.JAdjustedSymNameBase(J: TQuantumNumber; Sym: Integer): string;
begin
  if Sym < 0 then
    Result := '-' { Symmetry not used, typically for external field }
  else
    Result := IntToStr(Sym); { Default to number }
end;

function TMolecule.BothSymNames(J: TQuantumNumber; Sym: Integer): string;
var
  Other: string;
begin
  if JAdjustSym then begin
    Result := JAdjustedSymName(J, Sym);
    Other := SymName(Sym);
    if Result <> Other then
      Result := Result + ' (' + Other + ')';
  end else
    Result := SymName(Sym);
end;

function TMolecule.JDeAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer;
var
  i: Integer;
begin
  { Reverse effects of JAdjustedSym }
  { This can be overridden at the individual molecule level to make it more efficient }
  if JAdjustedSym(J, Sym) = Sym then
    Result := Sym { No adjustment - easy }
  else begin
    for i := 0 to LastSym do
      if (i <> Sym) and (JAdjustedSym(J, i) = Sym) then begin
        Result := i;
        Exit;
      end;
    raise Exception.CreateFmt('JDeAdjustedSym fails for %s: %d %d', [Name, J, Sym]);
  end;
end;

procedure TMolecule.SetnNuclei(const Index, Value: Integer);
var
  i, j: Integer;
begin
  inherited SetSetting(Index, Value);
  SetLength(Nuclei,Value);
  { Make sure each state has the right number of nuclei }
  for i := 0 to ComponentCount-1 do
    if Items[i] is TManifold then
      with TManifold(Items[i]) do
        for j := 0 to ComponentCount-1 do
          if Items[j] is TState then
            TState(Items[j]).AdjustNuclei(True);
end;

function TMolecule.SettingToPrettyStr(AIndex: Integer): string;
begin
  Result := '';
  case AIndex of
    1: if nNuclei <= 0 then Exit;
    2,3: if Settings[AIndex] <= 0 then Exit; { Jmin, Jmax }
  end;
  Result := inherited SettingToPrettyStr(AIndex);
end;

function TMolecule.GetActivePartitionFunction: TPartitionFunction;
var
  i: Integer;
begin
  if not Assigned(FActivePartitionFunction) then
    for i := 0 to ComponentCount-1 do
      if (Components[i] is TPartitionFunction)
         and TPartitionFunction(Components[i]).Active then begin
        FActivePartitionFunction := TPartitionFunction(Components[i]);
        Break;
      end;
  if not Assigned(FActivePartitionFunction) then begin
    if not Assigned(FDefaultPartitionFunction) then
      FDefaultPartitionFunction := TDefaultPartitionFunction.Create(nil);
    FActivePartitionFunction := FDefaultPartitionFunction;
  end;
  Result := FActivePartitionFunction;
end;

function TMolecule.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := AIndex = 4;
  if not Result then
    Result := inherited IsSettingColour(AIndex);
end;

function TMolecule.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex in [2,3];
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

function TMolecule.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Number of nuclei to calculate hyperfine structure for.'
  else if AIndex = 2 then
    Result := 'Minimum J to use in calculation - set to -1 (default) to use value from species.'
  else if AIndex = 3 then
    Result := 'Maximum J to use in calculation - set to -1 (default) to use value from species.'
  else if AIndex = 4 then
    Result := ColourHelpString
  else if AIndex = 5 then
    Result := 'If set, energy level plots and Fortrat diagrams will take account of the alternation in symmetry with J that is often found. For example in linear molecules an energy level will typically alternate between + and - parity with J (hence the e/f notation).'
  else if AIndex = 6 then
    Result := 'Select to split Hamiltonian matrix into unconnected blocks, if any exist.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TMolecule.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Abundance of this isotope; default is 1'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

class function TMolecule.GetChildTypes(AIndex: Integer): TMNodeClass;
begin
  if AIndex = nChildTypes - 1 then
    Result := TInterpolatedPartitionFunction
  else if AIndex = nChildTypes - 2 then
    Result := TTransitionMoments
  else
    Result := inherited GetChildTypes(AIndex);
end;

class function TMolecule.nChildTypes: Integer;
begin
  Result := inherited nChildTypes + 2;
end;

procedure TMolecule.ClearCache;
begin
  inherited;
  FLastPartitionFunction := -1;
end;

function TMolecule.TryStrToAnySymName(J: TQuantumNumber; const s: string; out Sym: Integer): Boolean;
var
  i: Integer;
begin
  for i := 0 to LastSym do
    if CompareText(s,SymName(i)) = 0 then begin
      Result := True;
      Sym := i;
      Exit;
    end;
  Result := TryStrToJAdjustedSymName(J, s, Sym);
end;

function TMolecule.TryStrToJAdjustedSymName(J: TQuantumNumber; const s: string; out Sym: Integer): Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 0 to LastSym do
    if CompareText(JAdjustedSymNameBase(J,i), s) = 0 then begin
      Result := True;
      Sym := i;
      Exit;
    end;
  if s = '-' then begin
    Result := True;
    Sym := -1;
  end;
end;

function TMolecule.FixSym(J: TQuantumNumber; Sym: Integer; InputPgopherVersion: Integer): Integer;
begin
  { Fix symmetry as required for older pgopher version .lin files }
  Result := Sym;
end;

{ TSpecies }

function TSpecies.GetIsotopomer(AIndex: Integer): TMolecule;
begin
  Result := TMolecule(Items[AIndex]);
end;

constructor TSpecies.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Colour=None,Concentration=1,Jmin=-1,Jmax=50';
  DimensionlessParameters := 'Concentration,' + DimensionlessParameters;
  inherited Create(AOwner);
end;

function TSpecies.AddToSpectrum(Spectrum: TSpectrum; var Problem: Exception; Simulation: TSimulation):Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 0 to Count-1 do
    { Note that we want to call AddToSpectrum for every species }
    Result := Isotopomer[i].AddToSpectrum(Spectrum, (Owner as TMixture).PrintLevel, RandomExcitation, Problem, Simulation)
              or Result ;
end;

function TSpecies.FindMolecule(AName: string): TMolecule;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to Count-1 do
    if CompareText(Isotopomer[i].Name, AName) = 0 then
      if Assigned(Result) then
        raise EReportException.CreateFmt('Duplicate Isotopomer name in %s: %s', [AName, Result.Name])
      else
        Result := Isotopomer[i];
end;

function TSpecies.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := AIndex = 3;
  if not Result then
    Result := inherited IsSettingColour(AIndex);
end;

function TSpecies.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex in [1,2];
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

function TSpecies.QnoInfo(QnoType: TQnoType): TQnoInfo;
var
  i: Integer;
begin
  if Count = 0 then
    Result := NullQnoInfo
  else
    Result := Isotopomer[0].QnoInfo(QnoType);
  for i := 1 to Count - 1 do
    MergeQno(Result, Isotopomer[i].QnoInfo(QnoType), QnoType);
end;

procedure TSpecies.SetModified(const Value: Boolean);
begin
  inherited;
  if Value then
    ClearCache;
end;

function TSpecies.SettingToPrettyStr(AIndex: Integer): string;
begin
  Result := '';
  if IsSettingQuantum(AIndex) and (Settings[AIndex] < 0) then
    Exit;
  Result := inherited SettingToPrettyStr(Aindex);
end;

function TSpecies.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Minimum J to use in calculation - set to -1 to use lowest J.'
  else if AIndex = 2 then
    Result := 'Maximum J to use in calculation.'
  else if AIndex = 3 then
    Result := ColourHelpString
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TSpecies.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Relative concentration of species'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

{ TMixture }

function TMixture.GetSettingsNames(AIndex: Integer;
  out LowestValue: Integer): PPChar;
begin
  Result := inherited GetSettingsNames(AIndex,LowestValue);
  if AIndex = 5 then
    Result := @QuantumNumberFormatNames
  else if AIndex = 2 then
    Result := @PrintLevelNames;
end;

function TMixture.GetSimulation(AIndex: Integer): TSimulation;
var
  i: Integer;
begin
  for i := 0 to Count-1 do
    if Items[i] is TSimulation then begin
      if AIndex = 0 then begin
        Result := TSimulation(Items[i]);
        Exit;
      end;
      Dec(AIndex);
    end;
  Result := nil;
end;

function TMixture.GetSpecies(AIndex: Integer): TSpecies;
var
  i: Integer;
begin
  for i := 0 to Count-1 do
    if Items[i] is TSpecies then begin
      if AIndex = 0 then begin
        Result := TSpecies(Items[i]);
        Exit;
      end;
      Dec(AIndex);
    end;
  Result := nil;
end;

function TMixture.GetQuantumNumberFormat(AIndex: Integer): TQuantumNumberFormat;
begin
  Result := TQuantumNumberFormat(GetSetting(AIndex));
end;

constructor TMixture.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + 'Precision=4,PrintLevel=plMinimal,SVDThresh=1e-15';
  SavedProperties := SavedProperties + ',AutoReplot=True,QuantumNumberFormat=J,FitCycles=1';
  SavedProperties := SavedProperties + ',ScaleChanges=1,';
  DimensionlessParameters := DimensionlessParameters + ',ScaleChanges,SVDThresh,';
  inherited;
  ClearUndo;
  TSimulation.Create(Self); { Must have at least one simulation }
  FnSimulations := 1;
  Modified := False;
end;

destructor TMixture.Destroy;
begin
  SetLineList(nil);
  FreeAndNil(FLoadedFormSettings);
  inherited Destroy;
end;

procedure TMixture.PreCalculatePartitionFunctions(Simulation: TSimulation; var Overflow: Exception);
var
  i, j: Integer;
begin
  for i := 0 to nSpecies-1 do
    for j := 0 to Species[i].Count-1 do
      Species[i].Isotopomer[j].PartitionFunction(Simulation, Overflow);
end;

procedure TMixture.PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string);
var
  i: Integer;
begin
  WriteLn(F, Indent, VersionInfo);
  PrettyPrintHeader(F,PrintLevel,Indent);
  Indent := Indent + '  ';
  PrettyPrintSettings(F, PrintLevel, Indent);
  PrettyPrintParameters(F, PrintLevel, Indent);
  for i := 0 to nSpecies - 1 do
    Species[i].PrettyPrint(F, PrintLevel, Indent);
  if Assigned(FCorrelations) and (Correlations.nEntries > 1) then
    Correlations.PrettyPrint(F,PrintLevel,Indent);
  { Do not print Simulation or form parameters }
  WriteLn(F);
end;

procedure TMixture.ConvertToPerturbations;
var
  i, j, k: Integer;
begin
  for i := 0 to nSpecies-1 do
    with Species[i] do
      for j := 0 to Count-1 do
        with Isotopomer[j] do
          for k := 0 to nManifolds-1 do
            Manifold[k].ConvertToPerturbations;
  FixupProperties;
end;

function TMixture.Correlations: TCorrelations;
begin
  if not Assigned(FCorrelations) then begin
    { FCorrelations := } TCorrelations.Create(Self);
    FCorrelations.Name := 'Correlations';
  end;
  Result := FCorrelations
end;

function TMixture.FixupProperties:Exception;
var
  i, j: Integer;
  LManName: string;
begin
  TStart(FixupPropertiesTime);
  FLargestIndex := 1;
  Result := inherited FixupProperties;
  FnMolecules := 0;
  FnSpecies := 0;
  FLongestMoleculeName := '';
  FLongestSymLength := 1;
  LManName := '';
  FnSimulations := 0;
  for i := 0 to Count - 1 do
    if Items[i] is TSpecies then
      with TSpecies(Items[i]) do begin
        Inc(FnSpecies);
        Inc(FnMolecules, Count);
        for j := 0 to Count-1 do
          with Isotopomer[j] do begin
            if Length(LongestMoleculeName) < Length(Name) then
              FLongestMoleculeName := Name;
            if Length(LongestManifoldName) > Length(LManName) then
              LManName := LongestManifoldName;
            if SymLength > FLongestSymLength then
              FLongestSymLength := SymLength;
          end;
      end else if Items[i] is TSimulation then
        Inc(FnSimulations);
  FLongestManifoldName := LManName;
  TEnd(FixupPropertiesTime);
end;

procedure TMixture.SetModified(const Value: Boolean);
begin
  inherited;
  if not Value then
    Exit;
  LineList := nil;
end;

procedure TMixture.SetQuantumNumberFormat(const Index: Integer; const Value: TQuantumNumberFormat);
begin
  SetSetting(Index, Ord(Value));
end;

procedure TMixture.SetSetting(AIndex: Integer; const AValue: Integer);
begin
  if ((AIndex = 4 { BasisOrder }) or (AIndex = 11 {AllowComplex}))
     and (Settings[AIndex] <> AValue) then
    ClearCache;
  inherited;
end;

function TMixture.SubPropInfo(const Token: string; out MNode: TMNode;  out ErrMsg: string): PPropInfo;
var
  AltMsg: string;
begin
  Result := inherited SubPropInfo(Token, MNode, ErrMsg);
  { Backward compatibility - treat properties of first Simulation object
    as properties of mixture }
  if not Assigned(Result) then
    Result := Simulations[0].SubPropInfo(Token, MNode, AltMsg);
  if not Assigned(Result) and (Pos('.',Token) < 0) then
    ErrMsg := AltMsg; { May be more meaningfull }
end;

function TMixture.FindMolecule(AName: string): TMolecule;
var
  Molecule: TMolecule;
  i: Integer;
begin
  Result := nil;
  for i := 0 to nSpecies-1 do begin
    Molecule := Species[i].FindMolecule(AName);
    if Assigned(Molecule) then begin
      if Assigned(Result) then
        raise EReportException.Create('Duplicate molecule name: ' + Result.Name);
      Result := Molecule;
    end;
  end;
end;

const
  MixtureSettingNames: array[1..12] of string = (
'Units for calculation - valid values are cm1, MHz, Kelvin and eV. The key change this makes is in calculating Boltzmann factors from temperature.',
'Controls the amount of printout in the Log window.',
'Precision of printout (decimal places after the point).',
'How to order the basis states making up a basis. Valid values are: boDefault, boDiagonal and boStateDiagonal.',
'How to display quantum numbers.',
'Number of least squares fit cycles to do on each press of a "fit" button.',
'True to show more info on fit',
'If false, automatic recalculation of the simulation is disabled in some circumstances. Use if the program is running slowly.',
'Set to use complex arithmetic if necessary.',
'Disable use of line lists; may be required for very large calculations.',
'Show estimated uncertainties of line positions in line lists.',
'Show derivatives in line listings.'
);

function TMixture.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(MixtureSettingNames))
     and (AIndex <= High(MixtureSettingNames)) then
    Result := MixtureSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TMixture.IsValidChildClass(MNodeClass: TMNodeClass): Boolean;
begin
  Result := inherited IsValidChildClass(MNodeClass);
  if Result and (MNodeClass = TCorrelations) then
    Result := not Assigned(FCorrelations);
end;

const
 MixtureParameterNames: array [1..2] of string = (
'Discard Singular Value Decomposition eigenvalues below this fraction of the largest eigenvalue',
'Scale changes produced by fitting by this - normally 1'
);

function TMixture.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(MixtureParameterNames))
     and (AIndex <= High(MixtureParameterNames)) then
    Result := MixtureParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

procedure TMixture.SetBasisOrder(const Index: Integer; const AValue: TBasisOrder);
begin
  SetSetting(Index, Ord(AValue));
end;

function TMixture.GetBasisOrder(const Index: Integer): TBasisOrder;
begin
  Result := TBasisOrder(GetSetting(Index));
end;

procedure TMixture.SetPrintLevel(const Index: Integer; const AValue: TPrintLevel);
begin
  SetSetting(Index, Ord(AValue));
end;

function TMixture.GetPrintLevel(const Index: Integer): TPrintLevel;
begin
  Result := TPrintLevel(GetSetting(Index));
end;

function TMixture.AddToSpectrum(Spectrum: TSpectrum; var Problem: Exception; Simulation: TSimulation): Boolean;
var
  i: Integer;
begin
  CheckProperties;
  Result := False;
  Spectrum.HeadingPrinted := False;
  for i := 0 to nSpecies-1 do
    Result := Species[i].AddToSpectrum(Spectrum, Problem, Simulation)
              or Result;
  if LineListWanted(PrintLevel) then
    Spectrum.ShowFooter(Log,PrintLevel);
  if not Result then begin
    if not Assigned(Problem) and (nSpecies > 0) then
      Problem := EReportException.Create('No transition moments set');
  end;
end;

function TMixture.QnoInfo(QnoType: TQnoType): TQnoInfo;
var
  i: Integer;
begin
  if nSpecies = 0 then
    Result := NullQnoInfo
  else
    Result := Species[0].QnoInfo(QnoType);
  for i := 1 to nSpecies-1 do
    MergeQno(Result, Species[i].QnoInfo(QnoType), QnoType);
end;

function TMixture.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  Child, NextChild: TDOMNode;
  s: string;
  Simulation: TSimulation;
  Value, I, J: Integer;
  ReadSimulation: Boolean;
  CorrelationNode: TDOMDocument; { Correlation node must be read last, and there should only be one }
begin
  { Version info first written in 5.2.275 }
  FReadVersion :=  DecodePgopherVersion(ReadValue(Node, 'Version', ''));
  { PrintLevel was an integer before 5.2.823, and was written as such until 6.0.207; read it here,
    with the side effect of removing it from the Node }
  s := ReadValue(Node, 'PrintLevel', '');
  ClearFormSettings;
  CorrelationNode := nil;
  { On writing, the parameters and settings of the first simulation object
    are written to the Mixture object, so reverse this logic here }
  Simulation := Simulations[0];
  ReadSimulation := False;
  Result := 0;
  Child := Node.FirstChild;
  while Assigned(Child) do begin
    NextChild := Child.NextSibling;
    if not ReadSimulation and (CompareText(Child.NodeName, 'Simulation') = 0) then begin
      Result := Result + Simulation.ReadProperties(Child, CheckUnread);
      Node.RemoveChild(Child);
      ReadSimulation := True; { Flag first simulation read }
    end else if (CompareText(Child.NodeName, 'Polarization') = 0) then begin
      Simulation.ReadMNode(Child,Value);
      if CheckUnread then
        Result := Result + nUnreadProperties(Child);
      Node.RemoveChild(Child);
    end else if (CompareText(Child.NodeName, 'FormSettings') = 0) then begin
      { Read form settings into LoadedFormSettings for loading to forms later if required }
      LoadedFormSettings.AppendChild(Child.CloneNode(True, LoadedFormSettings));
      Node.RemoveChild(Child);
    end else if (CompareText(Child.NodeName, 'Correlations') = 0) then begin
      if Assigned(CorrelationNode) then
        raise EReportException.Create('Duplicate Correlations Object');
      CorrelationNode := TDOMDocument.Create;
      CorrelationNode.AppendChild(Child.CloneNode(True,CorrelationNode));
      Node.RemoveChild(Child);
    end;
    Child := NextChild;
  end;
  Result := Result + inherited ReadProperties(Node, False);
  { Compatibility: Some properties have moved into Simulations[0] }
  Simulation.ReadProperties(Node, False);
  { Now use earlier PrintLevel string }
  if (s <> '') then
    if TryStrToInt(s, Value) then begin
      Inc(Value,3); { Read as integer offset from plMinimal, offset by -3 }
      if Value <= 0 then
        PrintLevel := TPrintLevel(Min(Max(0,Ord(plMinimal) + Value),Ord(High(TPrintLevel))))
      else if Value = 1 then
        PrintLevel := plMinimal { Old plBasic, which no longer exists }
      else { skip over plCSV, plText, plBinary }
        PrintLevel := TPrintLevel(Min(Ord(plDetail) + Value-2,Ord(High(TPrintLevel))))
    end else
      SetSettingFromStr(IndexOfSetting('PrintLevel'),s);
  s := ReadValue(Node, 'DoubleQnos', '');
  if s <> '' then
    if StrToBool(s) then
      QuantumNumberFormat := qft2J
    else
      QuantumNumberFormat := qftJ;
  { ContourFile has moved into TOverlay }
  s := ReadValue(Node,'ContourFile','0');
  if s <> '0' then
    ShowInfo('ContourFile setting in Mixture must now be set in Overlay properties instead');
  { JAdjustSym has moved into TMolecule }
  s := ReadValue(Node,'JAdjustSym','True');
  if (s <> '') and not StrToBool(s) then
    for i := 0 to Count - 1 do
      if Items[i] is TSpecies then
        with TSpecies(Items[i]) do
          for j := 0 to Count-1 do
            if Items[j] is TMolecule then
              TMolecule(Items[j]).JAdjustSym := False;
  { Read only TCorrelations object }
  if Assigned(CorrelationNode) then begin
    Result := Result + Correlations.ReadProperties(CorrelationNode.FirstChild, CheckUnread);
    CorrelationNode.Free;
  end;
  if CheckUnread then
    Result := Result + nUnreadProperties(Node)
end;

function TMixture.WriteProperties(Parent: TDOMNode; AllParameters: Boolean; WithChildren: Boolean): TDOMElement;
var
  Child: TDOMNode;
  i: Integer;
begin
  { For backwards compatibility, write Simulation parameters as
    parameters to the mixture object if there is only one simulation
    object }
  for i := 0 to nSimulations - 1 do
    Simulations[i].FWritten := False;
  if nSimulations = 1 then
    Simulations[0].FWritten := True;
  Result := inherited WriteProperties(Parent, AllParameters, WithChildren);
  WriteValue(Result, 'Version', VersionInfo, '');
  Simulations[0].TransferProperties(Result, True, AllParameters, WithChildren);
  if Assigned(LoadedFormSettings) and WithChildren then begin
    Child := LoadedFormSettings.FirstChild;
    while Assigned(Child) do begin
      Result.AppendChild(Child.CloneNode(True,Parent.OwnerDocument));
      Child := Child.NextSibling;
    end;
  end;
  { PrintLevel was an integer before 5.2.823, and was written as such until 6.0.207 }
  { No longer used as plCSV etc now introduced at 6.0.207, so better compatibility
    with 6.0 using normal text, with pl prefix }
  WriteValue(Result,'PrintLevel','pl'+SettingToStr(IndexOfSetting('PrintLevel')),'plMinimal');
  { Compatibility - do not write correlation matrix if ShowEstUnc not set }
  if not ShowEstUnc then begin
    Child := Result.FindNode('Correlations');
    if Assigned(Child) then
      Result.RemoveChild(Child);
  end;
end;

procedure TMixture.ClearCache;
begin
  PerformMNodeNotification(Self, mnCacheCleared, 0);
  LineList := nil;
  inherited;
end;

procedure TMixture.ClearFormSettings;
begin
  FreeAndNil(FLoadedFormSettings);
  FLoadedFormSettings := TDOMDocument.Create;
end;

procedure TMixture.PrepareForFit(FitType: TFitType; OtherNode: TRootMNode);
begin
  FillParameters;
  if Assigned(OtherNode) then
    OtherNode.AddNodeParameters(FittingParameters);
  PushParameters(FitType, False);
  ClearStdDev;
  if Assigned(OtherNode) then
    OtherNode.ClearStdDev;
  if Assigned(FCorrelations) then
    Correlations.Clear;
end;

procedure TMixture.ClearUndo;
begin
  FUndoPosition := 0;
  SetLength(FSavedParameters, 0);
  ClearConstraints;
  if Assigned(FCorrelations) then
    Correlations.Clear;
end;

procedure TMixture.DiscardFit;
begin
  Dec(FUndoPosition);
  if FUndoPosition < 0 then
    FUndoPosition := 0;
  SetLength(FSavedParameters, FUndoPosition);
end;

procedure TMixture.RedoFit;
begin
  if CanRedoFit then begin
    Inc(FUndoPosition);
    LoadCurrentSavedParameters;
  end;
end;

procedure TMixture.UndoFit;
begin
  if CanUndoFit then begin
    Dec(FUndoPosition);
    LoadCurrentSavedParameters;
  end;
end;

function TMixture.CanRedoFit: Boolean;
begin
  Result := FUndoPosition < High(FSavedParameters);
end;

function TMixture.CanUndoFit: Boolean;
begin
  Result := FUndoPosition > 0;
end;

procedure TMixture.FillParameters;
begin
  SetLength(FittingParameters, 0);
  AddNodeParameters(FittingParameters);
end;

procedure TMixture.PushParameters(FitType: TFitType; SetCurrent: Boolean);
begin
  if (Length(FSavedParameters) > 0) and (FSavedParameters[0].FitType <> FitType) then
    ClearUndo;
  SetLength(FSavedParameters, FUndoPosition+1);
  FSavedParameters[FUndoPosition].FitType := FitType;
  FSavedParameters[FUndoPosition].Parameters := Copy(FittingParameters);
  FSavedParameters[FUndoPosition].Correlation := Copy(Correlations.FitDetails.Correlation);
  if not SetCurrent then
    Inc(FUndoPosition);
end;

procedure TMixture.LoadCurrentSavedParameters;
var
  i: Integer;
  {$IFNDEF CONSOLE}
  DummyParameters: TParameters;
  {$ENDIF}
begin
  Correlations.FitDetails.Correlation := Copy(FSavedParameters[FUndoPosition].Correlation);
  with FSavedParameters[FUndoPosition] do
    for i := 0 to High(Parameters) do
      with Parameters[i] do begin
        MNode.SetParameter(Index, Value);
        MNode.StdDev[Index] := StdDev;
      end;
  ApplyConstraints;
  {$IFNDEF CONSOLE}
  ResetProgress;
  with MainForm do
    case FSavedParameters[FUndoPosition].FitType of
      ftLine:       if not Split.Visible and Mixture.Simulations[0].HasPlotted then
                      ConsiderRePlot;
      ftContour:    FitContour(Mixture, Overlays, {$IFDEF FPC}@{$ENDIF}ShowContour, DummyParameters, True);
      ftOverlaySum: FitOverlaySum(Mixture, Overlays, {$IFDEF FPC}@{$ENDIF}ShowOverlaySum, DummyParameters, True);
    end;
  {$ENDIF}
end;

procedure TMixture.SetEstUnc(const Index: Integer; const Value: Boolean);
begin
  SetFlag(Index, Value);
  if Value then
    Correlations; { Ensure correlation object exists }
end;

procedure TMixture.SetEUnits(const Index: Integer; const AValue: TEnergyUnits);
begin
  SetUnits(Index, AValue);
end;

procedure TMixture.SetLineList(const Value: TLineList);
begin
  FLineList.Free;
  FLineList := Value;
end;

function TMixture.GetEUnits(const Index: Integer): TEnergyUnits;
begin
  Result := GetUnits(Index);
end;

procedure TMixture.CycleUnits;
var
  NewUnits: TEnergyUnits;
begin
  if Units = High(TEnergyUnits) then
    NewUnits := Low(TEnergyUnits)
  else
    NewUnits := Succ(Units);
  ScaleParameters(Units,NewUnits);
  Units := NewUnits;
end;

{ TState }

function TState.AddBasis(Eigenvalues: TEigenvalues; F, M: TQuantumNumber; Sym: Integer; out StatWtGroup: Integer): Integer;
{ Note that Sym <> Eigenvalues.Sym for non zero external fields, where all fields used }
var
  F1, J, EffectiveS, Ieff: TQuantumNumber;
  i, ii, nuc, nNucleiEffective, nStates, StatWt, LastStatWtGroup: Integer;
begin
  Result := -1;
  StatWtGroup := -1;
  if RveSelect >=0 then
    if Molecule.JAdjustedSym(F,Sym) <> RveSelect then
      Exit;

  nNucleiEffective := Max(Molecule.nNuclei-Eigenvalues.RemovedSpins,0);
  if Molecule.SumNuclearSpin = 0 then
    nNucleiEffective := 0;
  if Eigenvalues.RemovedSpins > Molecule.nNuclei then
    EffectiveS := 0
  else
    EffectiveS := S;
  for i := 0 to High(Molecule.HyperfineBasisStates[nNucleiEffective]) do begin
    F1 := F;
    nuc := nNucleiEffective-1;
    while nuc >= 0 do begin
      J := F1 - Molecule.HyperfineBasisStates[nNucleiEffective,i,nuc];
      if (nuc > 0) and Molecule.Nuclei[nuc-1].AsNext then begin
        { This is the J I12 F case; J is correct, but need to use total
          nuclear spin I12 rather then an individual spin in checking validity.
          In this case HyperfineBasisStates contains the true quantum number
          rather than an offset }
        Dec(nuc);
        Ieff := Molecule.HyperfineBasisStates[nNucleiEffective,i,nuc]
      end else
        Ieff := Molecule.Nuclei[nuc].Spin;
      if (J < 0) or not IsTriangle(J,Ieff,F1) then begin
        F1 := -1;
        Break;
      end;
      Dec(nuc);
      F1 := J;
    end;
    if F1 >= 0 then begin
      nStates := Eigenvalues.ComponentCount;
      StatWt := AddOneJ(Eigenvalues,F1,Sym,Molecule.HyperfineBasisStates[nNucleiEffective,i],EffectiveS,LastStatWtGroup);
      for ii := nStates to Eigenvalues.ComponentCount-1 do begin
        (Eigenvalues.Components[ii] as TBasisState).F := F;
        (Eigenvalues.Components[ii] as TBasisState).M := M;
      end;
      if StatWtGroup < 0 then begin
        StatWtGroup := LastStatWtGroup;
        Result := StatWt;
      end else if StatWtGroup <> LastStatWtGroup then
        raise Exception.Create('AddBasis: Inconsistent Statistical Weights in '+Name);
    end;
  end;
end;

procedure TState.AdjustNuclei(FixNumber: Boolean);
{ Fixup Nucleus links within the state;
  Optionally Add or remove Nuclei to agree with molecule count }
begin
  if FixNumber then
    AdjustNodeIndexes(TNucleus, FNuclei, nil, opInsert, Molecule.nNuclei)
  else
    AdjustNodeIndexes(TNucleus, FNuclei)
end;

function TState.AllowedSphericalComponent(TransitionMoment: TSphericalTransitionMoment; State: TState; out Component: TQuantumNumber): Exception;
begin
  Result := EReportException.Create('auto not implemented for this molecule type');
end;

constructor TState.Create(AOwner: TComponent);
var
  i: Integer;
begin
  if Assigned(AOwner) and Assigned(AOwner.Owner) then
    FMolecule := AOwner.Owner as TMolecule;
  SavedProperties := SavedProperties + ',Colour=None,RveSelect=all';
  inherited;
  FFullRotationalSymmetry := True;
  { To avoid problems when adding states, make default spin
    the same as the first state in the same manifold }
  if Assigned(Owner) then
    for i := 0 to Owner.ComponentCount- 1 do
      if (Owner.Components[i] <> Self)
         and (Owner.Components[i] is TState) then begin
        S := TState(Owner.Components[i]).S;
        Break;
      end;
end;

destructor TState.Destroy;
begin
  FreeAndNil(VariableList);
  inherited Destroy;
end;

function TState.FindColour: TColor;
begin
  Result := Colour;
  if Result = clNone then
    Result := Manifold.FindColour;
end;

function TState.FixupProperties: Exception;
begin
  AllowComplex := Molecule.Mixture.AllowComplex;
  ClearCache; { Clears symbolic matrix elements }
  AdjustNuclei(True);
  Result := inherited FixupProperties;
end;

function TState.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  if AIndex = 2 then
    Result := SymSettingsNames(Self, LowestValue)
  else
    Result := inherited GetSettingsNames(AIndex, LowestValue);
end;

function TState.HasWidth: Boolean;
begin
  Result := Width <> 0;
end;

function IsNegative(const za, zb: Complex):Boolean;
begin
  Result := (za.re = -zb.re) and (za.im = -zb.im);
end;

function IsSameNegative(const za, zb: Complex):Boolean;
var
  z: Complex;
begin
  CMul(z,zb,-1);
  Result := MathSup.SameValue(za,z);
end;

function TState.SymmetrizedRotationalMatrixElement(A, B: TBasisState; MNode: TMnode; Flag: Integer;
  MatrixElementFunction: TMatrixElementFunction; SymCheck: TSymCheck): Complex;
  function MF(PhaseA, PhaseB, Mult: Integer):Complex;
  begin
    Result := MatrixElementFunction(A, B, PhaseA, PhaseB, MNode, Flag);
    if Mult > 0 then
      Exit;
    Result.re := - Result.re;
    Result.im := - Result.im;
  end;
var
  r: Complex;
  procedure FullCheck(Id: Integer);
  begin
    { Slower test here, but allow for slight rounding errors }
    if not MathSup.SameValue(r, Result)
       and not ((SymCheck = MayCancel) and IsSameNegative(r,Result)) then begin
       Write(Log, 'Symmetry mismatch ', Id, ':', r.re, ',', r.im, ' ', Result.re, ',', Result.im, ' ', A.BasisLabel, ' ', B.BasisLabel);
       if Assigned(MNode) then
         Write(Log, ' ', MNode.Name);
       WriteLn(Log);
       //WriteLn(Log,MF(1,1,1).re);
       //WriteLn(Log,MF(-1,-1,1).re);
       //WriteLn(Log,MF(1,-1,1).re);
       //WriteLn(Log,MF(-1,1,1).re);
       raise EReportException.Create('Symmetry mismatch ' + IntToStr(Id));
    end;
  end;
begin
  Result := MF(1, 1, 1);
  if A.Phase = 0 then begin
    if B.Phase <> 0 then begin
      CAdd(Result, MF(1, -1, B.Phase));
      CMul(Result, Result,1/Sqrt(Two));
    end;
  end else if B.Phase = 0 then begin
    CAdd(Result, MF(-1, 1, A.Phase));
    CMul(Result, Result, 1/Sqrt(Two));
  end else begin
    if not FullRotationalSymmetry and (SymCheck = MustEqual) then
      SymCheck := MayCancel;
    {$IFOPT R-}
    if SymCheck = MustEqual then
      SymCheck := NoCheck;
    {$ENDIF}
    // SymCheck := FullSum;
    if SymCheck <> NoCheck then begin
      r := MF(-1, -1, A.Phase*B.Phase);
      if (SymCheck <> FullSum) and not IsEqual(r,Result) and not ((SymCheck = MayCancel) and IsNegative(r, Result)) then begin
        // MatrixElementFunction(A, B, 1, 1, MNode, Flag);
        // MatrixElementFunction(A, B, -1, -1, MNode, Flag);
        FullCheck(1);
      end;
    end;
    CAdd(Result, MF(1, -1, B.Phase));
    if SymCheck <> NoCheck then begin
      CAdd(r, MF(-1, 1, A.Phase));
      if (SymCheck <> FullSum) and not IsEqual(r,Result) and not ((SymCheck = MayCancel) and IsNegative(r,Result)) then
        FullCheck(2);
    end;
    if (SymCheck = MayCancel) { or (SymCheck = FullSum) } then begin
      CAdd(Result, r);
      CMul(Result, Result, 1/Two);
    end;
  end;
  if not AllowComplex then
    if Result.im <> 0 then
      if Result.re = 0 then begin
        { Purely complex matrix element; in most cases changing the phase of one
          of the wavefunctions by -i has no net effect }
        Result.re := Result.im;
        Result.im := 0;
      end else
        raise EReportException.Create('Complex matrix element');
end;

procedure TState.SymmetryTable(var F: Text);
begin
  raise NotYet;
end;

function TState.SymmetrizedHyperfineMatrixElement(A, B: TBasisState;
  Rank: TQuantumNumber;
  MatrixElementFunction: TMatrixElementFunction;
  MNode: TMNode): Complex;
begin
  Result := z0;
  if Abs(A.J-B.J) > Rank then
    Exit;
  Result := SymmetrizedRotationalMatrixElement(A, B, MNode, Rank, MatrixElementFunction);
end;

function TState.ConvertParameter(AIndex: Integer; Zero: Boolean): TPerturbation;
begin
  Result := nil; { By default, can't convert }
end;

function TState.nParameterVariables: Integer;
{ Normally nParameters, but Vib has a special case to allow for subsidiary parameters }
begin
  Result := nParameters;
end;

function TState.ParameterVariable(AIndex: Integer): TSortedVariableNode;
var
  i: Integer;
begin
  { Note that this implementation relies on FixupProperties regenerating the
    symbolic expressions whenever a parameter value changes }
  if not Assigned(VariableList) then begin
    VariableList := TVariableList.Create;
    for i := 1 to nParameters do
      VariableList.Add(TSortedVariable.Create(ParameterNames[i],i));
  end;
  i := VariableList.IndexOf(ParameterNames[AIndex]);
  Result := (VariableList.Objects[i] as TSortedVariable).Template;
  Result.Value := Parameters[AIndex];
end;

procedure TState.ConvertToPerturbations;
var
  i: Integer;
begin
  { Ignore failures }
  for i := 2 { Do not convert origin } to nParameterVariables do
    if ParameterVariable(i).Value <> 0 then
      ConvertParameter(i, True);
end;

function TState.Manifold: TManifold;
begin
  Result := Owner as TManifold;
end;

function TState.MatrixElement(A, B: TBasisState; MatElemType: TMatElemType): Complex;
{ Matrix element of normal Hamiltonian }
var
  FFp, FF, FF1p, FF1: TQuantumNumber; { Partial F quantum numbers }
  Ieffp, Ieff: TQuantumNumber; { Nuclear spin, possibly a sum where equivalent nuclei used }
  I1, I2: TQuantumNumber;
  nuc: Integer;
  Rank1Sum, Rank2Sum, Rank1Extra, Rank2Extra: Complex;
  MatrixElementFunction: TMatrixElementFunction;
  SpinsMatch: Boolean; { Keep track of nuclear spin states matching }
  Nucleus1, Nucleus2: TNucleus;

  function CallHyperfineMatrixElement(Rank: TQuantumNumber; Nucleus: TNucleus): Complex;
  begin
    CMul(Result,SymmetrizedHyperfineMatrixElement(A, B, Rank,
                                                  {$IFDEF FPC}@{$ENDIF}HyperfineMatrixElement,
                                                  Nucleus),
                Minus(FF1+Ieffp+FF)*sixj(FF,Ieffp,FF1p,Rank,FF1,Ieff));
  end;

  function CallTwoHyperfineMatrixElements(Rank: TQuantumNumber): Complex;
  var
    z: Complex;
  begin
    CMul(Result, CallHyperfineMatrixElement(Rank, Nucleus1),
                 Minus(I1+I2+Ieff+Rank) * Sqrt((Ieffp+1)*(Ieff+1))
                 *sixj(I1, Ieffp, I2, Ieff, I1, Rank));
    CMul(z, CallHyperfineMatrixElement(Rank, Nucleus2),
            Minus(I1+I2+Ieffp+Rank) * Sqrt((Ieffp+1)*(Ieff+1))
            *sixj(I2, Ieffp, I1, Ieff, I2, Rank));
    CAdd(Result, z);
  end;
begin
  Result := z0;
  if A.State <> B.State then
    Exit;
  { Handle multiple nuclear spin case here, by decomposing into simple calls
    for a single nucleus }
  if (MatElemType = meHrot) { TODO: Other matrix element types }
     and (Molecule.nNuclei > 0) and (Abs(A.J-B.J) <= 4) then begin
    { Work from last nucleus }
    FFp := A.F;
    FF := B.F;
    Rank1Sum := z0;
    Rank2Sum := z0;
    SpinsMatch := True;
    nuc := High(A.HyperfineState);
    while nuc >= 0 do begin
      SpinsMatch := SpinsMatch and (FF = FFp);
      FF1p := FFp - A.HyperfineState[nuc];
      FF1 := FF - B.HyperfineState[nuc];
      Rank1Extra := z0;
      Rank2Extra := z0;
      Nucleus1 := A.State.Nucleus(nuc);
      if (nuc > 0) and Molecule.Nuclei[nuc-1].AsNext then begin
        Dec(nuc);
        Ieffp := A.HyperfineState[nuc];
        Ieff := B.HyperfineState[nuc];
        if Nucleus1.HasIdentical then
          Nucleus2 := Nucleus1
        else begin
          Nucleus2 := Nucleus1;
          Nucleus1 := A.State.Nucleus(nuc);
        end;
        I1 := Nucleus1.spin;
        I2 := Nucleus2.spin;
        Rank1Extra := CallTwoHyperfineMatrixElements(2);
        if (I1 > 1) or (I2 > 1) then
          Rank2Extra := CallTwoHyperfineMatrixElements(4);
        { Check I12 for match }
        if Ieffp <> Ieff then begin
          { Change in I12 - other spin operators can't change this }
          SpinsMatch := False;
          Rank1Sum := z0;
          Rank2Sum := z0;
        end;
      end else begin
        Ieff := Molecule.Nuclei[nuc].spin;
        Ieffp := Ieff;
        if SpinsMatch then begin
          { This nucleus contributes }
          Rank1Extra := CallHyperfineMatrixElement(2, Nucleus1);
          if Ieff > 1 then
            Rank2Extra := CallHyperfineMatrixElement(4, Nucleus1);
        end;
      end;
      if not MathSup.IsZero(Rank1Sum) then
        CMul(Rank1Sum,Rank1Sum, Minus(FF1p+Ieff+FF+2)*Sqrt((FFp+1)*(FF+1))
                                * sixj(FF1p, FFp, Ieff, FF, FF1, 2));
      if not MathSup.IsZero(Rank2Sum) then
        CMul(Rank2Sum,Rank2Sum, Minus(FF1p+Ieff+FF+4)*Sqrt((FFp+1)*(FF+1))
                                * sixj(FF1p, FFp, Ieff, FF, FF1, 4));
      CAdd(Rank1Sum, Rank1Extra);
      CAdd(Rank2Sum, Rank2Extra);
      FFp := FF1p;
      FF := FF1;
      Dec(nuc);
    end;
    SpinsMatch := SpinsMatch and (FF = FFp);
    CAdd(Result, Rank1Sum, Rank2Sum);
  end else
    SpinsMatch := A.HyperfineState = B.HyperfineState;
  if SpinsMatch then begin
    if ExprChecking then
      MatrixElementFunction := {$IFDEF FPC}@{$ENDIF}CheckedRotationalMatrixElement
    else
      MatrixElementFunction := {$IFDEF FPC}@{$ENDIF}RotationalMatrixElement;
    CAdd(Result, SymmetrizedRotationalMatrixElement(A, B, nil, Ord(MatElemType), MatrixElementFunction));
  end;
end;

function TState.PerturbationMatrixElement(Perturbation: TPerturbation; A, B: TBasisState; MatElemType: TMatElemType): Complex;
var
  FFp, FF, FF1p, FF1: TQuantumNumber; { Partial F quantum numbers }
  nuc: Integer;
  MatrixElementFunction: TMatrixElementFunction;
begin
  Result := z0;
  if MatElemType <> meHrot then Exit; { TODO: Perturbation meWidth }
  if Perturbation.SymSelect >= 0 then
    if Molecule.JAdjustedSym(A.J,A.Sym) <> Perturbation.SymSelect then
      Exit;

  if ExprChecking then
    MatrixElementFunction := {$IFDEF FPC}@{$ENDIF}CheckedRotationalPerturbationMatrixElement
  else
    MatrixElementFunction := {$IFDEF FPC}@{$ENDIF}RotationalPerturbationMatrixElement;
  { Handle multiple nuclear spin case here, by decomposing into simple calls
    for a single nucleus }
  if (Molecule.nNuclei > 0) and (Perturbation.Nucleus > 0) and (Abs(A.J-B.J) <= Perturbation.NuclearRank) then begin
    { Work from last nucleus }
    FFp := A.F;
    FF := B.F;
    for nuc := High(A.HyperfineState) downto 0 do begin
      FF1p := FFp - A.HyperfineState[nuc];
      FF1 := FF - B.HyperfineState[nuc];
      if (nuc = Perturbation.Nucleus-1) then
        if FF <> FFp then
          Exit
        else
          CMul(Result, SymmetrizedHyperfineMatrixElement(A, B, Perturbation.NuclearRank, MatrixElementFunction, Perturbation),
                       Minus(FF1+Molecule.Nuclei[nuc].spin+FF)*sixj(FF,Molecule.Nuclei[nuc].spin,FF1p,Perturbation.NuclearRank,FF1,Molecule.Nuclei[nuc].spin))
      else if nuc < Perturbation.Nucleus-1 then
        CMul(Result, Result, Minus(FF1p+Molecule.Nuclei[nuc].spin+FF+2)*Sqrt((FFp+1)*(FF+1))
                             * sixj(FF1p, FFp, Molecule.Nuclei[nuc].spin, FF, FF1, Perturbation.NuclearRank));
      FFp := FF1p;
      FF := FF1;
    end;
  end else if (A.HyperfineState = B.HyperfineState) and (A.J = B.J) then
    Result := SymmetrizedRotationalMatrixElement(A, B, Perturbation, Ord(MatElemType), MatrixElementFunction);
end;

procedure TState.PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string);
var
  Elements: TMatrixElements;
  i, j: Integer;
begin
  inherited;
  if PrintLevel < plDetail then
    Exit;
  for i := 2 { Skip Origin } to nParameterVariables do
    if ParameterVariable(i).Value <> 0 then
      try
        Elements := SymbolicRotationalMatrixElement(i);
        for j := 0 to High(Elements) do
          WriteLn(F, Indent, MatrixElementType(Elements[j].Signature, ParameterVariable(i).Text()), ' = ', Elements[j].Expr.Text(True));
        DiscardMatrixElements(Elements);
      except
        WriteLn(F, Indent, 'Matrix elements for ', ParameterVariable(i).Text(), ' not available');
      end;
end;

function TState.SymbolicRotationalMatrixElement(AIndex: Integer): TMatrixElements;
begin
  raise NotYet;
end;

threadvar
  PSignature: TIntegerArray; { Minimize overhead }

var
  EvaluatorLock: TRTLCriticalSection;

procedure TState.CheckedMatrixElement(
  ABra, AKet: TBasisState;
  SignA, SignB: Integer;
  const AMatrixElements: TMatrixElements;
  Numeric: Complex; const AName: string);
var
  i, j: Integer;
  V: Double;
  Match: Boolean;
  Expr: TExprNode;
begin
  if not SameValue(Numeric.im,0) then
    raise NotYet; { TODO: Add complex to symbolic }
  PSignature := ABra.MakeSignature(SignA, AKet, SignB);
  V := 0;
  Expr := nil;
  for i := 0 to High(AMatrixElements) do begin
    Match := True;
    for j := 0 to High(PSignature) do
      if AMatrixElements[i].Signature[j] <> PSignature[j] then begin
        Match := False;
        Break;
      end;
    if Match then begin
      Expr := AMatrixElements[i].Expr;
      EnterCriticalsection(EvaluatorLock);
      try
        ABra.SetSymbolicVariables(SignA, AKet, SignB);
        V := AMatrixElements[i].Evaluator.Value;
        { Additional (very slow) check:
          if not SameValue(V, Expr.RealValue) then
            WriteLn(Log, Expr.Text, ' ', V, Expr.RealValue);
        }
      finally
        LeaveCriticalsection(EvaluatorLock);
      end;
      Break;
    end;
  end;
  if not SameValue(Numeric.re, V) then begin
    Write(Log, '<', ABra.State.UniqueName, ':', ABra.BasisLabel, ' J=', ABra.J, '(', SignA, ')|', AName, '|',
                    AKet.State.UniqueName, ':', AKet.BasisLabel, ' J=', AKet.J, '(', SignB, ')>',
               ' SymbolicMatrixElement: Signature:');
    for i := 0 to High(PSignature) do
      Write(Log, PSignature[i]:3);
    WriteLn(Log, ' Numeric=', Numeric.re, ' Symbolic=', V);
    if Assigned(Expr) then
      WriteLn(Log, Expr.Text(True));
  end;
end;

function TState.CheckedRotationalPerturbationMatrixElement(
  ABra, AKet: TBasisState;
  SignA, SignB: Integer;
  MNode: TMNode; Flag: Integer
  ): Complex;
var
  Perturbation: TPerturbation;
begin
  Result := RotationalPerturbationMatrixElement(ABra,AKet,SignA,SignB,MNode,Flag);
  Perturbation := MNode as TPerturbation;
  if Perturbation.CheckMatrixElementsAvailable then
    CheckedMatrixElement(ABra, AKet, SignA, SignB, Perturbation.MatrixElements, Result, Perturbation.PairName);
end;

function TState.CheckedRotationalMatrixElement(ABra, AKet: TBasisState; SignA,
  SignB: Integer; MNode: TMnode; Flag: Integer): Complex;
begin
  CheckMatrixElementsAvailable;
  Result := RotationalMatrixElement(ABra,AKet,SignA,SignB,MNode,Flag);
  CheckedMatrixElement(ABra, AKet, SignA, SignB, MatrixElements, Result, 'Hrot');
end;

procedure TState.CheckMatrixElementsAvailable;
var
  M, ME: TMatrixElements;
  i, j, k, l, w: Integer;
  Match: Boolean;
begin
  if (Length(MatrixElements) > 0) or MatrixElementsEmpty then
    Exit;
  MustBeSerial;
  { The SymbolicRotationalMatrixElement() routine runs rather slowly
    when using threads due to overheads in the memory allocation routine.
    (HundA will also not be thread safe) }
  { Construct local copy in ME for locking purposes }
  for i := 1 to nParameterVariables do
    if ParameterVariable(i).Value <> 0 then begin
      TStart(IndividualSymbolicRotationalMatrixElementTime);
      M := SymbolicRotationalMatrixElement(i);
      TEnd(IndividualSymbolicRotationalMatrixElementTime);
      TStart(SymbolicRotationalMatrixElementTime);
      for j := 0 to High(M) do begin { For each set of changes of quantum numbers }
        w := Length(ME);
        for k := 0 to High(ME) do begin
          { Do we already have some matching changes? }
          Match := High(ME[k].Signature) = High(M[j].Signature);
          if Match then
            for l := 0 to High(ME[k].Signature) do
              Match := Match and (ME[k].Signature[l] = M[j].Signature[l]);
          if Match then begin
            w := k;
            Break;
          end;
        end;
        M[j].Expr := OperatorTimes.CreateExprNode(ParameterVariable(i).Duplicate,M[j].Expr);
        if w = Length(ME) then begin
          { No Match - add new }
          SetLength(ME, w+1);
          ME[w].Signature := M[j].Signature;
          ME[w].Expr := M[j].Expr;
        end else
          { Same set of changes - add expression, but do not simplify }
          with ME[w] do begin
            if Expr.Base <> OperatorPlus then
              Expr := OperatorPlus.CreateExprNode(Expr);
            Expr.AddNode(M[j].Expr);
          end;
      end;
      TEnd(SymbolicRotationalMatrixElementTime);
    end;
  FMatrixElementsEmpty := Length(ME) = 0;
  TStart(SymbolicRotationalMatrixElementTime);
  for i := 0 to High(ME) do
    ME[i].Evaluator := ME[i].Expr.Compile;
  TEnd(SymbolicRotationalMatrixElementTime);
  FMatrixElements := ME;
end;

procedure TState.ClearCache;
begin
  inherited;
  DiscardMatrixElements(FMatrixElements);
  FMatrixElementsEmpty := False;
end;

function TState.TransitionMatrixElement(TransitionMoment: TTransitionMoment; A, B: TBasisState): Complex;
{ Reduced matrix element of transition dipole }
{ Similar logic to TState.MatrixElement, but assuming transition moment does not
  act on any nuclear spins }
{ TODO: Implement nuclear magnetic dipole transitions }
var
  FFp, FF, FF1p, FF1, Ieff: TQuantumNumber; { Partial F quantum numbers }
  nuc: Integer;
  IsNuclearDipole: Boolean;
  Rank: TQuantumNumber;
begin
  Result := z0;
  if Abs(A.J-B.J) > TransitionMoment.Rank then
    Exit;
  Result := z1;
  if Molecule.nNuclei > 0 then begin
    { Work from last nucleus }
    IsNuclearDipole := TransitionMoment is TNuclearMagneticMoment;
    FFp := A.F;
    FF := B.F;
    nuc := High(A.HyperfineState);
    if IsNuclearDipole and ((TNuclearMagneticMoment(TransitionMoment)).Nucleus-1 > nuc) then begin
      Result := z0; { Removed nucleus }
      Exit;
    end;
    Rank := TransitionMoment.Rank;
    while nuc >= 0 do begin
      FF1p := FFp - A.HyperfineState[nuc];
      FF1 := FF - B.HyperfineState[nuc];
      if (nuc > 0) and Molecule.Nuclei[nuc-1].AsNext then begin
        if IsNuclearDipole then
          raise NotYet;
        { Coupled spins; need to use total nuclear spin for pair }
        Dec(nuc);
        Ieff := A.HyperfineState[nuc];
        if Ieff <> B.HyperfineState[nuc] then begin
          Result := z0; { Can't change hyperfine state }
          Exit;
        end;
      end else
        Ieff := Molecule.Nuclei[nuc].Spin;
      if IsNuclearDipole and ((TNuclearMagneticMoment(TransitionMoment)).Nucleus-1 = nuc) then begin
        if FF1p = FF1 then begin
          CMul(Result, Result, -TNuclearMagneticMoment(TransitionMoment).gN
                               * Minus(FF1+Ieff+FF+Rank)*Sqrt((FFp+1)*(FF+1))
                               * sixj(Ieff, FFp, FF1, FF, Ieff, Rank)
                               * ReducedMatrixElement(2,Ieff));
          Rank := 0; { Flag need to use identity operator for everything else }
        end else
          Result := z0;
      end else if IsNuclearDipole and (Rank = 0) then begin
        if FF1p <> FF1 then
          Result := z0;
      end else
        CMul(Result, Result, Minus(FF1p+Ieff+FF+Rank)*Sqrt((FFp+1)*(FF+1))
                             * sixj(FF1p, FFp, Ieff, FF, FF1, Rank));
      if MathSup.IsZero(Result) then
        Exit;
      FFp := FF1p;
      FF := FF1;
      Dec(nuc);
    end;
  end;
  CMul(Result, Result, SymmetrizedRotationalMatrixElement(A, B, TransitionMoment, -1, {$IFDEF FPC}@{$ENDIF}RotationalTransitionMatrixElement, MayCancel));
  { The following line can be added if a consistent sign is wanted for the reduced matrix elements;
    without it states differening in total angular momentum by an odd number will have a different
    sign to this matrix element depending on the order of calculation.
    Computes phase factor for smallest M }
  // CMul(Result, Result, Minus(Max(A.F,B.F)-SmallestProjection(A.F)));
end;

function TState.Nucleus(Index: Integer): TNucleus;
begin
  Result := FNuclei[Index] as TNucleus;
end;

function TState.HalfSpin(RemovedSpins: Byte): TQuantumNumber;
var
  i: Integer;
begin
  if RemovedSpins > Molecule.nNuclei then
    Result := 0
  else begin
    Result := S;
    for i := 0 to Molecule.nNuclei - 1 - RemovedSpins do
      Inc(Result,Molecule.Nuclei[i].Spin);
  end;
  Result := Result and 1;
end;

function TState.ReadBranchFormat(Text: string; var Obs: TObservation;
  UpperState: TState; Reader: TObservationsReaderBase): Boolean;
begin
  raise EReportException.Create('ReadBranchFormat not implemented for '+Name);
end;

function TState.ReadHyperfineBranch(Text: string; var Obs: TObservation; var HyperfineU, HyperfineL: THyperfineBasis; Reader: TObservationsReaderBase; HasFields:Boolean): Boolean;
  function ReadOneHyperfine(var J: TQuantumNumber; var Hyperfine: THyperfineBasis):Boolean;
  var
    p, i: Integer;
  begin
    Result := False;
    p := Molecule.nNuclei-Reader.RemoveSpins;
    if HasFields then
      Inc(p);
    SetLength(Hyperfine,p);
    for i := 0 to High(Hyperfine) do begin
      p := Pos(',',Text);
      if p <= 0 then
        p := Length(Text)+1;
      if Reader.ReadQuantumNumber(Hyperfine[i],Copy(Text,1,p-1)) then
        Exit; { Fails }
      if i >= Molecule.nNuclei - Reader.RemoveSpins then
        J := Hyperfine[i] { M }
      else if not Molecule.Nuclei[i].AsNext then begin
        Dec(Hyperfine[i],J);
        J := Hyperfine[i]+J;
      end;
      Delete(Text,1,p);
    end;
    if HasFields then
      SetLength(Hyperfine,Length(Hyperfine)-1);
    { TODO: May need modifying of for hyperfine and field }
    Result := True;
  end;
begin
  Result := not (ReadOneHyperfine(Obs.JU,HyperfineU)
                 and ReadOneHyperfine(Obs.JL,HyperfineL)
                 and (Text = ''));
end;

function TState.ReadProperties(Node: TDOMNode; CheckUnread: Boolean):Integer;
var
  Child, NextChild: TDOMNode;
  Index: Integer;
  NucName: string;
begin
  { Ensure Nuclei read in correct order, and discard any un-matched nuclei }
  Index := 0;
  Child := Node.FirstChild;
  while Assigned(Child) do begin
    NextChild := Child.NextSibling;
    if Pos('Nucleus', Child.NodeName) = Length(Child.NodeName)-6 then begin
      NucName := ReadValue(Child, 'Name', '', '', False);
      if Index < Molecule.nNuclei then begin
        if (Molecule.Nuclei[Index].Name <> '') and (NucName <> Molecule.Nuclei[Index].Name) then
//          for i := 0 to Molecule.nNuclei-1 do
//            if Molecule.Nuclei[i].Name = NucName then
          raise EReportException.Create('Nucleus ' + IntToStr(Index) + ' out of order for ' + Owner.Name + ' - ' + Node.NodeName + '; Expected ' + Molecule.Nuclei[Index].Name + ' found ' + NucName);
        Inc(Index);
      end else
        Node.RemoveChild(Child); { Assume paste or merge from system with fewer nuclei }
    end;
    Child := NextChild;
  end;
  Result := inherited ReadProperties(Node, CheckUnread);
end;

function TState.SettingToPrettyStr(AIndex: Integer): string;
begin
  Result := '';
  case AIndex of
    2: if RveSelect < 0 then Exit;
    3: if S <= 0 then Exit;
  end;
  Result := inherited SettingToPrettyStr(AIndex);
end;

function TState.SpinSpin(Nuc1, Nuc2: Integer): TSpinSpin;
var
  i: Integer;
begin
  for i := 0 to ComponentCount - 1 do
    if Components[i] is TSpinSpin then
      with TSpinSpin(Components[i]) do
        if ((Bra.Name = Molecule.Nuclei[Nuc1].Name)
            and (Ket.Name = Molecule.Nuclei[Nuc2].Name))
           or
           ((Bra.Name = Molecule.Nuclei[Nuc2].Name)
            and (Ket.Name = Molecule.Nuclei[Nuc1].Name)) then begin
              Result := Self.Components[i] as TSpinSpin;
              Exit;
            end;
  Result := SpinSpinType.Create(Self);
  Result.SetBraKetName(Molecule.Nuclei[Nuc1].Name, Molecule.Nuclei[Nuc2].Name);
end;

function TState.StateOrigin: Double;
begin
  Result := Origin;
end;

function TState.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := AIndex = 1;
  if not Result then
    Result := inherited IsSettingColour(AIndex);
end;

function TState.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex = 3;
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

function TState.HelpOnSetting(AIndex: Integer): string;
begin
  case AIndex of
    1: Result := ColourHelpString;
    2: Result := 'Only include given rovibronic symmetry - for normal use set to ''all''';
    3: Result := 'Electron Spin';
  else
    Result := inherited HelpOnSetting(AIndex);
  end;
end;

function TState.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'State Origin.'
  else if AIndex = 2 then
    Result := 'Linewidth (rotation independent) for state.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

{ TTransitionMoments }

function TTransitionMoments.GetTransitionMoment(AIndex: Integer): TTransitionMoment;
begin
  Result := Items[AIndex] as TTransitionMoment;
end;

constructor TTransitionMoments.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Colour=None';
  inherited Create(AOwner);
end;

constructor TTransitionMoments.Create(AOwner: TComponent; Manifold1Name, Manifold2Name: string);
begin
  Create(AOwner);
  FBraName := Manifold1Name;
  FKetName := Manifold2Name;
end;

destructor TTransitionMoments.Destroy;
begin
  FreeAndNil(CacheLock);
  inherited;
end;

function TTransitionMoments.FixupProperties:Exception;
var
  i: Integer;
  Swapped: Boolean;
begin
  if (nWorkers > 0) and not Assigned(CacheLock) then
    CacheLock := TMultiReadExclusiveWriteSynchronizer.Create;
  DisableRankCheck := False;
  Result := inherited FixupProperties;
  if Assigned(Result) then Exit;
  { Sort by Rank }
  repeat
    Swapped := False;
    for i := 0 to ComponentCount-2 do
      if TransitionMoment[i].Rank > TransitionMoment[i+1].Rank then begin
        TransitionMoment[i].ComponentIndex := i+1;
        Swapped := True;
      end;
  until not Swapped;
  if ComponentCount > 0 then
    FnRanks := 1
  else
    FnRanks := 0;
  for i := 0 to ComponentCount-2 do
    if TransitionMoment[i].Rank <> TransitionMoment[i+1].Rank then
      Inc(FnRanks);
  for i := 0 to ComponentCount-1 do
    TransitionMoment[i].FNodeIndex := i; { Repair NodeIndex }
end;

function TTransitionMoments.MaxRank: Integer;
begin
  if ComponentCount > 0 then
    Result := TransitionMoment[ComponentCount-1].Rank
  else
    Result := -2;
end;

function TTransitionMoments.Allowed(J1: TQuantumNumber; Sym1: Integer; J2: TQuantumNumber; Sym2: Integer): Boolean;
{ Do any of the transition moments give rise to an allowed transition? }
var
  i: Integer;
  TM: TTransitionMoment;
begin
  Result := True;
  for i := 0 to ComponentCount-1 do begin
    TM := TransitionMoment[i];
    if (DisableRankCheck or IsTriangle(J1,J2,TM.Rank)) and TM.Allowed(Sym1, Sym2) then
      Exit;
  end;
  Result := False;
end;

procedure TTransitionMoments.AddToSpectrum(Spectrum: TSpectrum; Ev1, Ev2: TEigenvalues; ATransitionMoment: TTransitionMoment; Simulation: TSimulation; Wait: Boolean; PrintLevel: TPrintLevel; const Excitation: TExcitation; var Problem: Exception; IncludeZero: Boolean);
var
  D: TMatrix;
begin
  D := TransitionMatrix(Ev1, Ev2, Wait, not Assigned(ATransitionMoment), ATransitionMoment, Simulation, PrintLevel);
  if Wait then
    try
      AddMatrixToSpectrum(D, Spectrum, Ev1, Ev2, PrintLevel, Excitation, Problem, IncludeZero);
    finally
      if Assigned(ATransitionMoment) then
        D.Free;
    end;
end;

procedure TTransitionMoments.AddMatrixToSpectrum(D: TMatrix; Spectrum: TSpectrum; Ev1, Ev2: TEigenvalues; PrintLevel: TPrintLevel; const Excitation: TExcitation; var Problem: Exception; IncludeZero: Boolean = false);
var
  i, f, ColourI: Integer;
  te, wt, str: Double;
  Line, SLine: TLine;
  Swap: Boolean;
  procedure DoLine(var Line: TLine);
  var
    ALine: TLine;
  begin
    if LineListWanted(PrintLevel) then begin
      ALine := Line;
      ALine.CachedName := '';
      if Mixture.Simulations[0].PlotUnits <> Mixture.Units then
        ConvertLine(ALine, Mixture.Units, Mixture.Simulations[0].PlotUnits);
      Spectrum.ShowLine(ALine, Log, PrintLevel, Mixture.Simulations[0].OThreshold, False, Problem);
    end;
    Spectrum.AddLine(Line, Mixture.Simulations[0], Mixture.Units, Problem);
  end;
begin
  TStart(AddMatrixToSpectrumTime);
  Line.EigenvaluesL := Ev1;
  Line.EigenvaluesU := Ev2;
  {$IFDEF OLD}
  for i := 1 to Ev1.Size do begin
    { For each lower state }
    wt := Ev1.StatisticalWeight(i) * (Owner as TMolecule).Abundance * (Owner.Owner as TSpecies).Concentration;
    Line.IndexL := i;
    ColourI := Colour;
    if ColourI = clNone then
      ColourI := Ev1.DominantBasis[i].State.FindColour;
    for f := 1 to Ev2.Size do begin
      { For each upper state }
      Line.IndexU := f;
      te := Ev2.E[f] - Ev1.E[i];
      Swap := False;
      { TODO: Call MakeLine here instead, but check skiped line logic carefully }
      if Ev2.Manifold.Initial then
        if (Ev1.Manifold = Ev2.Manifold) and (te <= 0) then
          Continue { Avoid handling transitions twice if within same manifold }
          { Note that transitions with exactly 0 frequency will be ignored completely,
            rather than counting them twice }
        else begin
          if te < 0 then begin
            te := -te;
            Swap := True;
          end;
        end
      else if (Ev1.Manifold = Ev2.Manifold) { Relative energies within manifold must be OK }
              or (Ev2.Manifold.LowestStateOrigin < Ev1.Manifold.LowestStateOrigin) then begin
        te := -te; { Presumably Emission spectrum }
        Swap := True;
      end;
      Str := D[i,f] * wt;
      if not IncludeZero and (Str = 0) then Continue; { Zero intensity transitions are boring }
      Line.Frequency := te;
      Line.Strength := Str*IntensityCorrection(Ev2.J, Ev1.J, Excitation);
      Line.Width := 0;
      if Assigned(Ev1.Widths) then
        Line.Width := Line.Width + Ev1.Widths[i];
      if Assigned(Ev2.Widths) then
        Line.Width := Line.Width + Ev2.Widths[f];
      Line.Colour := ColourI;
      if Line.Colour = clNone then
        Line.Colour := Ev2.DominantBasis[f].State.FindColour;
      if Line.Colour = clNone then
        Line.Colour := clBlack;
      if Swap then begin
// TODO:        if Ev2.Manifold.Initial and (Ev1.Manifold = Ev2.Manifold) then
// TODO:          Continue; { Avoid handling transitions twice if within same manifold }
        SLine := Line;
        SLine.EigenvaluesU := Ev1;
        SLine.EigenvaluesL := Ev2;
        Sline.IndexU := i;
        Sline.IndexL := f;
        DoLine(SLine);
      end else
        DoLine(Line);
    end;
  end;
  {$ELSE}
  for i := 1 to Ev1.Size do begin
    { For each lower state }
    for f := 1 to Ev2.Size do begin
      { For each upper state }
      Swap := MakeLine(Line,D,Ev1,Ev2,i,f,Excitation,IncludeZero);
      if not IncludeZero and (Line.Strength = 0) then Continue; { Zero intensity transitions are boring }
      if Swap then begin
        if Ev2.Manifold.Initial and (Ev1.Manifold = Ev2.Manifold) then
          Continue; { Avoid handling transitions twice if within same manifold }
        SLine := Line;
        SLine.EigenvaluesU := Ev1;
        SLine.EigenvaluesL := Ev2;
        Sline.IndexU := i;
        Sline.IndexL := f;
        DoLine(SLine);
      end else begin
        if (Ev1.Manifold = Ev2.Manifold) and (Line.Frequency <= 0) then
          Continue; { Discard zero frequency transitions within the same manifold }
        DoLine(Line);
      end;
    end;
  end;
  {$ENDIF}
  TEnd(AddMatrixToSpectrumTime);
end;

function TTransitionMoments.MakeLine(out Line: TLine; Ev1, Ev2: TEigenvalues; Index1, Index2: Integer; Simulation: TSimulation): Boolean;
var
  D: TMatrix;
begin
  D := TransitionMatrix(Ev1, Ev2, True, True, nil, Simulation);
  Result := MakeLine(Line, D, Ev1, Ev2, Index1, Index2, RandomExcitation, True);
end;

function TTransitionMoments.MakeLine(out Line: TLine; D: TMatrix; Ev1, Ev2: TEigenvalues; Index1,
  Index2: Integer; const Excitation: TExcitation; IncludeZero: Boolean): Boolean;
var
  ColourI: Integer;
  te, wt: Double;
begin
  Line.EigenvaluesL := Ev1;
  Line.EigenvaluesU := Ev2;
  { For Linear TVibratingMolecules only, statistical weights can be different for the two states;
    For cases like Pi <- Sigma, scale strength up by 2;
    Symmetric/Spherical top point groups may require additional handling }
  wt := Max(Ev1.StatisticalWeight(Index1),Ev2.StatisticalWeight(index2)) * (Owner as TMolecule).Abundance * (Owner.Owner as TSpecies).Concentration;
  Line.IndexL := Index1;
  ColourI := Colour;
  if ColourI = clNone then
    ColourI := Ev1.DominantBasis[Index1].State.FindColour;
  Line.IndexU := Index2;
  te := Ev2.E[Index2] - Ev1.E[Index1];
  Result := False;
  if Ev2.Manifold.Initial then begin
    if te < 0 then begin
      te := -te;
      Result := True;
    end;
  end else if (Ev2.Manifold.LowestStateOrigin < Ev1.Manifold.LowestStateOrigin) then begin
    te := -te; { Presumably Emission spectrum }
    Result := True;
  end;
  Line.Strength := D[Index1,Index2] * wt;
  Line.Frequency := te;
  if not IncludeZero and (Line.Strength = 0) then Exit; { Zero intensity transitions are boring }
  Line.Strength := Line.Strength*IntensityCorrection(Ev2.J, Ev1.J, Excitation);
  Line.Width := 0;
  if Assigned(Ev1.Widths) then
    Line.Width := Line.Width + Ev1.Widths[Index1];
  if Assigned(Ev2.Widths) then
    Line.Width := Line.Width + Ev2.Widths[Index2];
  Line.Colour := ColourI;
  if Line.Colour = clNone then
    Line.Colour := Ev2.DominantBasis[Index2].State.FindColour;
  if Line.Colour = clNone then
    Line.Colour := clBlack;
end;

function TTransitionMoments.OriginalTransitionMatrix(
            Ev1, Ev2: TEigenvalues; Rank: TQuantumNumber; ATransitionMoment: TTransitionMoment;
            Intensity, Electric, Magnetic: Boolean; Simulation: TSimulation): TComplexMatrix;
{ Transition Moments in original Basis for the given rank only }
var
  tm: TTransitionMoment;
  State1, State2: TState;
  b, bb, t, nTM: Integer;
  Sum, V: Complex;
  TMs: array of TTransitionMoment;
begin
  TStart(MakeTransitionMatrixTime);
  if not Ev1.HasFields and (Ev1.Size > 0) and (Ev2.Size > 0) and (Ev1.SpinSpecies(1) <> Ev2.SpinSpecies(1)) then
    raise EreportException.CreateFmt('Inconsistent statistical weights for %s in %s', [Name, Molecule.UniqueName]);
  Result := TComplexMatrix.Create(Ev1.Size, Ev2.Size);
  Result.Zero;
  SetLength(TMs, Count);
  nTM := 0;
  { Optimization - pick out relevant transition moments first }
  for t := 0 to Count-1 do
    if ((TransitionMoment[t].Rank = Rank)
        and (not Assigned(ATransitionMoment) or (TransitionMoment[t] = ATransitionMoment))
        and ((Electric and not TransitionMoment[t].Magnetic)
             or (Magnetic and TransitionMoment[t].Magnetic)
             )
       )
     then begin
      TMs[nTM] := TransitionMoment[t];
      Inc(nTM);
    end;
  if nTM = 0 then
    Exit;
  try
    for b := 1 to Ev1.Size do begin
      State1 := Ev1.Basis[b].State;
      for bb := 1 to Ev2.Size do begin
        Sum := z0;
        State2 := Ev2.Basis[bb].State;
        for t := 0 to nTM-1 do begin { for each transition moment }
          tm := TMs[t];
          if (tm.Bra = State1) and (tm.Ket = State2) then begin
            V := tm.MatrixElement(Ev1.Basis[b], Ev2.Basis[bb],Intensity);
            if Ev1.HasFields then { Can't apply to sum, as complex conjugate of matix sometimes evaluated }
              CMul(V, Mfactor(Ev1.Basis[b].F, Ev1.J, Ev2.Basis[bb].F, Ev2.J, Simulation));
            Cadd(Sum, V)
          end else if (tm.Bra = State2) and (tm.Ket = State1) then begin
            { Work out complex conjugate }
            V := tm.MatrixElement(Ev2.Basis[bb], Ev1.Basis[b],Intensity);
            if Ev1.HasFields then
              CMul(V, Mfactor(Ev2.Basis[bb].F, Ev2.J, Ev1.Basis[b].F, Ev1.J, Simulation));
            V.im := -V.im;
            Cadd(Sum, V);
          end;
        end;
        if not Ev1.HasFields then
          CMul(Sum, Sqrt(One/(Rank+1))); { Sum over M }
        Result.AutoSetValue(b,bb,Sum);
      end;
    end;
  except
    Result.Free;
    raise;
  end;
  TEnd(MakeTransitionMatrixTime);
end;

function TTransitionMoments.TransitionMatrix(
  Ev1, Ev2: TEigenvalues; Wait, Cache: Boolean;
  TransitionMomentSel: TTransitionMoment;
  Simulation: TSimulation;
  APrintLevel: TPrintLevel): TComplexMatrix;
{ Transformed Transition Moments squared }
  function FindTransitionMatrix: TCachedTransitionMatrix;
  var
    i: Integer;
  begin
    if Assigned(CacheLock) then
      CacheLock.BeginRead;
    Result := nil;
    try
      for i := 0 to High(Ev1.CachedTransitionMatrices) do
        if (Ev1.CachedTransitionMatrices[i].OtherEigenvalues = Ev2) and (Ev1.CachedTransitionMatrices[i].Simulation = Simulation) then begin
          Result := Ev1.CachedTransitionMatrices[i];
          Exit;
        end;
    finally
      if Assigned(CacheLock) then
        CacheLock.EndRead;
    end;
  end;
var
  D: TComplexMatrix;
  i: Integer;
  Rank: Integer;
  DoneRank: array of Boolean;
  CacheEntry: TCachedTransitionMatrix;
begin
  if not Wait and (not Cache or Assigned(TransitionMomentSel)) then
    { Wait makes no sense if not caching }
    raise Exception.Create('TTransitionMoments.TransitionMatrix');
  if not Assigned(TransitionMomentSel) and Cache then begin
    CacheEntry := FindTransitionMatrix;
    if not Assigned(CacheEntry) then
      try
        if Assigned(CacheLock) then
          CacheLock.BeginWrite;
        { Check that it hasn't been created by another thread }
        CacheEntry := FindTransitionMatrix;
        if not Assigned(CacheEntry) then begin
          SetLength(Ev1.CachedTransitionMatrices, Length(Ev1.CachedTransitionMatrices)+1);
          CacheEntry := TCachedTransitionMatrix.Create(nil);
          Ev1.CachedTransitionMatrices[High(Ev1.CachedTransitionMatrices)] := CacheEntry;
          if APrintLevel = plDefault then
            APrintLevel := Mixture.PrintLevel;
          with CacheEntry do begin
            TransitionMoments := Self;
            EigenValues := Ev1;
            OtherEigenvalues := Ev2;
            PrintLevel := APrintLevel;
          end;
          CacheEntry.Simulation := Simulation;
        end;
      finally
        if Assigned(CacheLock) then
          CacheLock.EndWrite;
      end;
    { We now have the structure, but not necessarily complete }
    CacheEntry.Make(Wait,nil);
    if Wait then
      Result := CacheEntry.TransitionMatrix
    else
      Result := nil;
    Exit;
  end;
  { Check that required matrices are available }
  Ev1.Make(True, [meHrot, meWidth]);
  Ev2.Make(True, [meHrot, meWidth]);
  TStart(TransitionMatrixTime);
  Result := nil;
  SetLength(DoneRank, MaxRank + 1);
  for i := 0 to MaxRank do
    DoneRank[i] := False;
  try
    for i := 0 to Count-1 do
      { Each rank must be done separately }
      if (TransitionMoment[i] = TransitionMomentSel)
         or (not Assigned(TransitionMomentSel) and not DoneRank[TransitionMoment[i].Rank]) then begin
        Rank := TransitionMoment[i].Rank;
        DoneRank[Rank] := True;
        D := OriginalTransitionMatrix(Ev1, Ev2, Rank, TransitionMomentSel, True, True, True, Simulation);
        if APrintLevel >= plMatrices then
          Ev1.ShowTransitionMatrix(Ev2, D, Format('Rank %s Transition Matrix for ', [Mixture.QuantumNumberToStr(Rank)]), 0);
        try
          TransformTransitionMatrix(D, Ev1, Ev2);
          if APrintLevel >= plMatrices then
            Ev1.ShowTransitionMatrix(Ev2, D, Format('Transformed Rank %s Transition Matrix for ', [Mixture.QuantumNumberToStr(Rank)]), 0, False);
          D.Sqr;
          if APrintLevel >= Succ(plMatrices) then
            Ev1.ShowTransitionMatrix(Ev2, D, Format('Squared Rank %s Transition Matrix for ', [Mixture.QuantumNumberToStr(Rank)]), 0, False);
        except
          D.Free;
          raise;
        end;
        if Assigned(Result) then begin
          Result.Add(D);
          D.Free;
        end else
          Result := D;
      end;
  except
    Result.Free;
    raise;
  end;
  TEnd(TransitionMatrixTime);
end;

procedure TTransitionMoments.TransformTransitionMatrix(D: TComplexMatrix; Ev1, Ev2: TEigenvalues);
{ Transform Matrix D using eigenvectors - overwrites D }
begin
  TStart(MatrixMultiplyTime);
  Ev1.Multiply(D);
  TEnd(MatrixMultiplyTime);
  TStart(PostMultiplyTime);
  D.PostMultiply(Ev2.C);
  TEnd(PostMultiplyTime);
end;

function TTransitionMoments.Mixture: TMixture;
begin
  Result := Owner.Owner.Owner as TMixture;
end;

class function TTransitionMoments.BraKetClass: TMNodeClass;
begin
  Result := TManifold;
end;

function TTransitionMoments.Molecule: TMolecule;
begin
  Result := Owner as TMolecule;
end;

function TTransitionMoments.PairName: string;
begin
  Result := 'mu';
end;

function TTransitionMoments.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := AIndex = 1;
  if not Result then
    Result := inherited IsSettingColour(AIndex);
end;

function TTransitionMoments.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := ColourHelpString
  else
    Result := inherited HelpOnSetting(AIndex);
end;

{ TTransitionMoment }

constructor TTransitionMoment.Create(AOwner: TComponent);
begin
  FRank := 2;
  inherited Create(AOwner);
end;

constructor TTransitionMoment.Create(AOwner: TComponent; State1Name, State2Name: string);
begin
  Create(AOwner);
  FBraName := State1Name;
  FKetName := State2Name;
end;

function TTransitionMoment.FixupProperties: Exception;
begin
  FVibronicallyAllowed := vaMaybe; { This variable otherwise handled at molecule level only }
  Result := inherited FixupProperties;
  if not Assigned(Result) and Odd(Rank) then
    Result := EReportException.Create('Transition Moment Rank must be integral in '+PairName);
end;

function TTransitionMoment.Allowed(Sym1, Sym2: Integer): Boolean;
begin
  Result := (Owner.Owner as TMolecule).Allowed(Self, Sym1, Sym2);
end;

function TTransitionMoment.MatrixElement(ABra, AKet: TBasisState; Intensity: Boolean): Complex;
begin
  Result := ABra.State.TransitionMatrixElement(Self, ABra, AKet);
//  if Magnetic and HasUnits((RootNode as TMixture).Simulations[0].IntensityUnits)  then
  if Magnetic then begin
    if Intensity then
      CMul(Result, Result, 1e21 * BohrMagneton);
    if NuclearMagneton then
      CMul(Result,MeOverMp);
  end;
end;

function TTransitionMoment.BraOwner: TMNode;
begin
  Result := (Owner as TTransitionMoments).Bra;
end;

function TTransitionMoment.KetOwner: TMNode;
begin
  Result := (Owner as TTransitionMoments).Ket;
end;

function TTransitionMoment.CheckProjections(K1, K2: TQuantumNumber): Boolean;
begin
  { Check change in given projections for consistency with transition moment;
    only return False if the matrix element is definately zero.
    Only used for optimization }
  Result := (Abs(K1-K2) <= Rank) or (Abs(K1+K2) <= Rank);
end;

class function TTransitionMoment.BraKetClass: TMNodeClass;
begin
  Result := TState;
end;

{ TCartesianTransitionMoment }

const
  AxisNames: array [0..3] of PChar = ('a', 'b', 'c', nil);

function AxisName(AAxis: TabcAxis): string;
begin
  Result := AxisNames[Ord(AAxis)];
end;

function TCartesianTransitionMoment.GetSettingsNames(AIndex: Integer; out LowestValue:Integer): PPChar;
begin
  Result := inherited GetSettingsNames(AIndex,LowestValue);
  if AIndex = 1 then
    Result := @AxisNames[0];
end;

function TCartesianTransitionMoment.FixupProperties:Exception;
begin
  Result := inherited FixupProperties;
  if not Assigned(Result) and (Rank <> 2) then
    Result := EReportException.CreateFmt('%s-%s: Transition Moment Rank Not 1: %d', [Name, AxisName(Axis), Rank]);
end;

function TCartesianTransitionMoment.GetAxis(const Index: Integer): TabcAxis;
begin
  Result := TabcAxis(GetSetting(Index));
end;

procedure TCartesianTransitionMoment.SetAxis(const Index: Integer; const AValue: TabcAxis);
begin
  SetSetting(Index, Ord(AValue));
end;

constructor TCartesianTransitionMoment.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Axis=a';
  inherited;
end;

function TCartesianTransitionMoment.PairName: string;
begin
  Result := AxisName(Axis);
end;

function TCartesianTransitionMoment.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Axis of transtion moment - a, b or c.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

{ TSphericalTransitionMoment }

function TSphericalTransitionMoment.PairName: string;
begin
  if Component <= -9999 then
    Result := Format('T(%s)', [TMixture(RootNode).QuantumNumberToStr(Rank)])
  else
    Result := Format('T(%s,%s)', [TMixture(RootNode).QuantumNumberToStr(Rank), TMixture(RootNode).QuantumNumberToStr(Component)]);
end;

procedure TSphericalTransitionMoment.SetSetting(AIndex: Integer; const AValue: Integer);
begin
  inherited SetSetting(AIndex, AValue);
  FRank := Rank;
end;

procedure TSphericalTransitionMoment.SetSettingFromStr(AIndex: Integer; const Str: string);
begin
  if (AIndex = 2) and (Str = 'auto') then
    Component := -9999
  else
    inherited;
end;

function TSphericalTransitionMoment.SettingToStr(AIndex: Integer): string;
begin
  if (AIndex = 2) and (Component <= -9999) then
    Result := 'auto'
  else
    Result := inherited SettingToStr(AIndex);
end;

constructor TSphericalTransitionMoment.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Rank=2,Component=auto';
  inherited;
end;

function TSphericalTransitionMoment.FixupProperties: Exception;
begin
  Result := inherited FixupProperties;
  if not Assigned(Result) and (Component <= -9999) then
    if Rank = 0 then
      FActualComponent := 0 { No Choice! }
    else
      Result := (Bra as TState).AllowedSphericalComponent(Self, Ket as TState, FActualComponent)
  else
    FActualComponent := Component;
  if not Assigned(Result) and (Odd(FActualComponent) or (Abs(FActualComponent) > Rank)) then
    Result := EReportException.Create('Invalid Transition Moment component in '+UniqueName);
end;

function TSphericalTransitionMoment.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
begin
  if ((RootNode as TMixture).ReadVersion < 5002300)
     and not Assigned((Node as TDOMElement).GetAttributeNode('Component')) then
    { Back compatibility - default changes from "0" to "auto" for component }
    (Node as TDOMElement).SetAttribute('Component', '0');
  Result:=inherited ReadProperties(Node, CheckUnread);
end;

function TSphericalTransitionMoment.HelpOnSetting(AIndex: Integer): string;
var
  E: Exception;
begin
  if AIndex = 1  then
    Result := 'Rank of transition - ' + GlobalMixture.QuantumNumberToStr(2) + ' for a normal electric dipole transition.'
  else if AIndex = 2 then
    if (Component <= -9999) and Assigned(Bra) and Assigned(Ket) then begin
      try
        E := (Bra as TState).AllowedSphericalComponent(Self, Ket as TState, FActualComponent)
      except
        on Ex: Exception do
          E := EReportException.Create(Ex.Message);
      end;
      if Assigned(E) then begin
        Result := 'Problem with auto setting: ' + E.Message
      end else
        Result := 'Current value = ' + (Owner.Owner as TMolecule).QuantumNumberToStr(FActualComponent);
    end else
      Result := 'Projection quantum number of transition moment - auto works in many cases'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TSphericalTransitionMoment.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex in [1,2];
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

{ TEigenvalues }

function TEigenvalues.FieldLabel: string;
begin
  Result := '';
  if ExField.E <> 0 then
    Result := Format('Electric Field = %g', [ExField.E]);
  if ExField.B <> 0 then
    if Result <> '' then begin
      Result := Result + ', ';
    Result := Result + Format('Magnetic Field = %g', [ExField.B]);
    if ExField.E <> 0  then
      Result := Result + Format(', Angle = %g', [ExField.Angle]);
  end
end;

function TEigenvalues.FindDerivatives(const Parameter: TParameter): Integer;
var
  i: Integer;
begin
  Result := -1;
  for i := 0 to High(Derivatives) do
    with Derivatives[i] do
      if (MNode = Parameter.MNode) and (Index = Parameter.Index) then begin
        Result := i;
        Exit;
      end;
end;

procedure TEigenvalues.Fixup;
var
  i: Integer;
begin
  { Make up Basis array }
  SetLength(Basis, ComponentCount+1); { Element 0 is unused }
  for i := 1 to ComponentCount do
    Basis[i] := Components[i-1] as TBasisState;
  Manifold.Mixture.LargestIndex := Max(Manifold.Mixture.LargestIndex,Size);
end;

function TEigenvalues.Size: Integer;
begin
  Result := ComponentCount;
end;

constructor TEigenvalues.Create(AManifold: TManifold; AJ: TQuantumNumber;
  ASym: Integer; const AExField: TStaticFields; IncludePerturbations: Boolean;
  AState: TState; RemoveSpins: Byte);
  procedure DoAddBasis(State: TState; F, M: TQuantumNumber; Sym: Integer);
  var
    wt, LastStatWtGroup, i, LastCount: Integer;
    procedure AddWeights;
    var
      i: Integer;
    begin
      if ComponentCount > Length(FStatisticalWeights) then
        SetLength(FStatisticalWeights, ComponentCount*2); { Guess at size required }
      SetLength(FStatWtGroups,Length(FStatisticalWeights));
      { Record statistical weights of latest states added }
      for i := LastCount to ComponentCount - 1 do begin
        FStatisticalWeights[i] := wt;
        FStatWtGroups[i] := LastStatWtGroup;
      end;
    end;
  begin
    LastCount := ComponentCount;
    wt := State.AddBasis(Self, F, M, Sym, LastStatWtGroup);
    if FStatWtGroup < 0 then begin
      FStatisticalWeight := wt;
      FStatWtGroup := LastStatWtGroup;
    end else if Length(FStatisticalWeights) > 0 then
      AddWeights
    else if (LastStatWtGroup >= 0) and (LastStatWtGroup <> FStatWtGroup) then
      if HasFields then begin
        { Different statistical weights in basis }
        AddWeights; { Also allocates the array }
        for i := 0 to LastCount-1 do begin { Fill in weights so far }
          FStatisticalWeights[i] := FStatisticalWeight;
          FStatWtGroups[i] := FStatWtGroup;
        end;
      end else
        raise Exception.CreateFmt('Inconsistent Statistical Weight groups for %s : %d <> %d', [State.UniqueName, FStatWtGroup, LastStatWtGroup]);
  end;
  procedure AddBasis(State: TState);
  var
    i: Integer;
    F: TQuantumNumber;
  begin
    if HasFields or (Sym < 0) then begin
      F := Max(Abs(AJ),Manifold.FindJmin(0));
      while F <= Manifold.FindJmax(F+4) do begin
        for i := 0 to Manifold.nSym-1 do { TODO -cField: May not need every symmetry with Field in place }
          DoAddBasis(State, F, AJ, i);
        Inc(F,2);
      end;
    end else
      DoAddBasis(State, AJ, 0, Sym);
  end;
var
  i: Integer;
begin
  TStart(MakeBasisTime);
  inherited Create(nil);
  Manifold := AManifold;
  J := AJ; { M if HasFields }
  FStatisticalWeight := -1;
  FStatWtGroup := -1;
  SelectedState := AState;
  LastStateIndex := 0;
  RemovedSpins := RemoveSpins;
  NoPerturbations := not IncludePerturbations;
  ExField := AExField;
  FHasFields := IsNonZero(ExField);
  if not Assigned(Manifold) then
    Exit; { Making dummy Eigenvalues object }
  if HasFields then
    Sym := -1 { Do not use symmetry }
  else begin
    Sym := ASym;
    if (Sym < 0) or (Sym >= Manifold.nSym) then
      raise Exception.CreateFmt('%s: Invalid symmetry number %d', [Manifold.Name, Sym]);
  end;
  if Assigned(AState) then begin
    { If asked, just use specified state }
    if AState.Manifold <> Manifold then
      raise Exception.Create('State does not belong to manifold');
    AddBasis(AState);
  end else
    { Make up Basis }
    for i := 0 to Manifold.nStates-1 do
      AddBasis(Manifold.State[i]);
  Fixup;
  TEnd(MakeBasisTime);
end;

function TEigenvalues.Degeneracy(AIndex: Integer): Integer;
begin
  Result := StatisticalWeight(AIndex);
  if not HasFields then
    Result := Result * (J+1);
end;

destructor TEigenvalues.Destroy;
var
  i: Integer;
begin
  for i := 0 to High(CachedTransitionMatrices) do
    FreeAndNil(CachedTransitionMatrices[i]);
  FreeAndNil(FE);
  FreeAndNil(FC);
  FreeAndNil(FWidths);
  FreeAndNil(FDominantBasisIndex);
  for i := 0 to High(Derivatives) do
    FreeAndNil(Derivatives[i].Values);
  inherited Destroy;
end;

function TEigenvalues.DiagonalMatrixElements(MatElemType: TMatElemType; Base: Integer = 0; IncludePerturbations: Boolean = False): TDArray;
var
  i, p: Integer;
begin
  SetLength(Result, Size-Base);
  for i := Base+1 to Size do begin
    Result[i-1-Base] := Basis[i].State.MatrixElement(Basis[i], Basis[i], MatElemType).re; { TODO: Check for non-zero im }
    if IncludePerturbations then
      for p := 0 to Manifold.nPerturbations-1 do
        if Manifold.Perturbation[p].Connects(Basis[i].State, Basis[i].State) then
          Result[i-1-Base] := Result[i-1-Base] + Manifold.Perturbation[p].MatrixElement(Basis[i], Basis[i], MatElemType).re;
  end;
end;

{
function TEigenvalues.DiagonalMatrixElements(MatElemType: TMatElemType): array of Double;
var
  i: Integer;
begin
  SetLength(Result, Size+1);
  Result := TColumnVector.Create(Size);
  try
    for i := 1 to Size do
      Result[i] := Basis[i].State.MatrixElement(Basis[i], Basis[i], MatElemType);
  except
    Result.Free;
    raise;
  end;
end;
}
{ Implementation of a simple arbitrary sized set }
type
  TBigSet = array of Integer;

procedure RotateLeft(var s: TBigSet);
var
  i, c, nc: Integer;
begin
  if Length(s) = 1 then
    s[0] := s[0] shl 1
  else begin
    c := 0;
    for i := 0 to High(s) do begin
      if s[i] < 0 then
        nc := 1
      else
        nc := 0;
      s[i] := (s[i] shl 1) or c;
      c := nc;
    end;
  end;
end;

procedure RotateRight(var s: TBigSet);
var
  i, c, nc: Integer;
begin
  if Length(s) = 1 then
    s[0] := s[0] shr 1
  else begin
    c := 0;
    for i := High(s) downto 0 do begin
      nc := s[i] and 1;
      s[i] := (s[i] shr 1) or (c shl (SizeOf(Integer)*8-1));
      c := nc;
    end;
  end;
end;

procedure OrIn(var s: TBigSet; const p: TBigSet);
var
  i: Integer;
begin
  for i := 0 to High(s) do
    s[i] := s[i] or p[i];
end;

procedure CopySet(var s: TBigSet; const p: TBigSet);
var
  i: Integer;
begin
  for i := 0 to High(s) do
    s[i] := p[i];
end;

function AndNonZero(const s, p: TBigSet):Boolean;
var
  i: Integer;
begin
  if Length(p) <> 0 then begin
    Result := True;
    for i := 0 to High(s) do
      if (s[i] and p[i]) <> 0 then
        Exit;
  end;
  Result := False;
end;

function OrInChanged(var s: TBigSet; const p: TBigSet):Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 0 to High(s) do
    if (s[i] or p[i]) <> s[i] then begin
      s[i] := s[i] or p[i];
      Result := True;
    end;
end;

procedure SetBitn(var s: TBigSet; n: Cardinal);
var
  i: Integer;
begin
  i := n div (SizeOf(Integer)*8);
  n := n mod (SizeOf(Integer)*8);
  s[i] := s[i] or (1 shl n);
end;

procedure Clear(var s: TBigSet);
var
  i: Integer;
begin
  for i := 0 to High(s) do
    s[i] := 0;
end;

procedure TEigenvalues.MatrixElements(out Matrices: TComplexMatrices; BlockMatrix: Boolean; MatElemType: TMatElemType; IncludePerturbations: Boolean);
var
  {$IFDEF SMALLSET}
  { Track non-zero matrix elements to allow matrix to be blocked if possible }
  Mask: Int64;
  nSet: Integer;
  Connections: array of Int64;
  {$ENDIF}
  { Track non-zero matrix elements to allow matrix to be blocked if possible }
  Mask: TBigSet;
  nSet: Integer;
  Connections: array of TBigSet;
  nBlocks: Integer;

  procedure FactorMatrix(var Matrices: TComplexMatrices; Connections: array of TBigSet);
    { Given the Hamiltonian Matrix and a bitmap of non-zero elements, rearrange elements to block-diagonalize }
  var
    b, bb, bbb, bNext, i, k, CIndex, BlockSize: Integer;
    CurMask, Mask, bMask: TBigSet;
    Matrix, Block: TComplexMatrix;
  begin
    Matrix := Matrices[0];
    try
      { Now split matrix }
      SetLength(Matrices,nBlocks);
      for i := 0 to nBlocks - 1 do
        Matrices[i] := nil;
      b := Size;
      CIndex := Size-1;
      SetLength(Mask, Length(Connections[Size]));
      bMask := Copy(Mask); { Allocation only }
      repeat
        BlockSize := 0;
        bNext := 0;
        Clear(Mask);
        SetBitn(Mask,b-1);
        CurMask := Connections[b];
        for bb := b downto 1 do begin
          if AndNonZero(CurMask,Mask) then begin
            { Rearrange basis by block }
            Basis[bb].ComponentIndex := CIndex;
            Dec(CIndex);
            Inc(BlockSize);
          end else if (bNext = 0) and (Length(Connections[bb]) <> 0) then
            bNext := bb; { Highest unused basis state not involved in the current matrix }
          RotateRight(Mask);
        end;
        { Create next block matrix }
        Block := TComplexMatrix.Create(BlockSize,BlockSize);
        Matrices[nBlocks-1] := Block;
        Dec(nBlocks);
        Block.Mode := Matrix.Mode;
        { Copy elements }
        Clear(Mask);
        SetBitn(Mask,b-1);
        i := BlockSize;
        for bb := b downto 1 do begin
          if AndNonZero(CurMask,Mask) then begin
            { This basis state contributes }
            CopySet(bMask,Mask);
            k := i;
            for bbb := bb downto 1 do begin
              { For each other basis state }
              if AndNonZero(CurMask,bMask) then begin
                Block[i,k] := Matrix[bb,bbb];
                if i <> k then
                  Block[k,i] := Matrix[bbb,bb];
                Dec(k);
              end;
              RotateRight(bMask);
            end;
            Dec(i);
          end;
          RotateRight(Mask);
        end;
        b := bNext;
      until b = 0;
      { Sanity check }
      bb := 0;
      for b := 0 to High(Matrices) do
        Inc(bb,Matrices[b].RowCount);
      if bb <> Size then begin
        raise Exception.Create('MatrixElements.FactorMatrix');
        //WriteLn(Log, bb);
        //Clear(Mask);
        //for b := 1 to High(Connections) do
        //  if High(Connections[b]) > 0 then
        //    OrIn(Mask,Connections[b]);
        //MatrixElements(Matrices, True, meHrot);
        //for b := 1 to High(Connections) do begin
        //  Write(Log, b:4, ' ');
        //  for bb := Size downto 1 do
        //    Write(Log, ((Connections[b][0] shr (bb-1)) and 1));
        //  WriteLn(Log);
        //end;
      end;
    finally
      Matrix.Free;
    end;
    Fixup;
  end;

  function CountBlocks:Integer;
  { Count how many blocks to factor into }
  var
    b, bb, bNext: Integer;
    MaskChanged: Boolean;
    CurMask: TBigSet;
  begin
    { Work backwards from last row, as this will have most connections set }
    b := Size;
    Result := 0;
    repeat
      CurMask := Connections[b];
      Inc(Result);
      repeat
        bNext := 0;
        MaskChanged := False;
        for bb := b-1 downto 1 do begin
          if AndNonZero(CurMask,Connections[bb]) then begin
            MaskChanged := OrInChanged(CurMask,Connections[bb]) or MaskChanged;
            Connections[bb] := nil
          end else if (bNext = 0) and (Length(Connections[bb]) <> 0) then
            bNext := bb; { Highest unused basis state not involved in the current matrix }
        end;
      until not MaskChanged;
      b := bNext;
    until bNext = 0;
    nBlocks  := Result;
  end;

  {$IFDEF SMALLSET}
  function CountBlocks:Integer;
  { Count how many blocks to factor into }
  var
    b, bb, bNext: Integer;
    MaskChanged: Boolean;
    CurMask: Int64;
  begin
    { Work backwards from last row, as this will have most connections set }
    b := Size;
    Result := 0;
    repeat
      CurMask := Connections[b];
      Inc(Result);
      repeat
        bNext := 0;
        MaskChanged := False;
        for bb := b-1 downto 1 do begin
          if (CurMask and Connections[bb]) <> 0 then begin
            if (CurMask or Connections[bb]) <> CurMask then begin
              MaskChanged := True;
              CurMask := CurMask or Connections[bb];
            end;
            { Flag this basis state has been copied }
            Connections[bb] := 0
          end else if (bNext = 0) and (Connections[bb] <> 0) then
            bNext := bb; { Highest unused basis state not involved in the current matrix }
        end;
      until not MaskChanged;
      Connections[b] := CurMask;
      b := bNext;
    until bNext = 0;
    Connections[0] := Result;
  end;
  {$ENDIF}
var
  p, b, bb, EBField: Integer;
  Current: TState;
  Conv: Double;
  TransitionMoments: TTransitionMoments;
  Matrix, EMat: TComplexMatrix;
  Sum: Complex;
begin
  BlockMatrix := BlockMatrix and (Size > 1);
  SetLength(Matrices,1);
  Matrices[0] := nil;
  if Size = 0 then Exit;
  Matrices[0] := TComplexMatrix.Create(Size, Size);
  Matrix := Matrices[0];
  Matrix.Zero;
  if BlockMatrix then begin
    SetLength(Connections,Size+1);
    SetLength(Mask, Size div (SizeOf(Integer)*8) + 1);
    SetBitn(Mask,0);
  end;
  try
    for b := 1 to Size do begin
      Current := Basis[b].State;
      if BlockMatrix then begin
        Connections[b] := Copy(Mask); { Diagonal elements set }
        RotateLeft(Mask);
      end;
      for bb := 1 to b do
        if (Current = Basis[bb].State) and (not HasFields or ((Basis[b].F = Basis[bb].F) and (Basis[b].Sym = Basis[bb].Sym))) then begin
          Sum := Current.MatrixElement(Basis[b], Basis[bb], MatElemType);
          if not MathSup.IsZero(Sum) then begin
            if BlockMatrix and (b <> bb) then
              OrIn(Connections[b],Connections[bb]);
            Matrix.AutoSetValue(b,bb,Sum);
            if b <> bb then begin
              Sum.im := -Sum.im;
              Matrix[bb,b] := Sum;
            end;
          end;
        end
    end;
    { Shift perturbations to separate loop for timing purposes }
    if IncludePerturbations and (Manifold.nPerturbations > 0) then begin
      TStart(PerturbationMatrixElementTime);
      for b := 1 to Size do begin
        Current := Basis[b].State;
        for bb := 1 to b do begin
          Sum := z0;
          nSet := 0;
          if not HasFields or ((Basis[b].F = Basis[bb].F) and (Basis[b].Sym = Basis[bb].Sym)) then
            for p := 0 to Manifold.nPerturbations-1 do
              if Manifold.Perturbation[p].Connects(Current, Basis[bb].State) then begin
                CAdd(Sum, Manifold.Perturbation[p].MatrixElement(Basis[b], Basis[bb], MatElemType));
                Inc(nSet);
              end;
          if (nSet > 0) and not MathSup.IsZero(Sum) then begin
            if BlockMatrix then
              OrIn(Connections[b],Connections[bb]);
            CAdd(Sum, Matrix[b,bb]);
            Matrix.AutoSetValue(b,bb,Sum);
            if b <> bb then begin
              Sum.im := -Sum.im;
              Matrix[bb,b] := Sum;
            end;
          end;
        end;
      end;
      TEnd(PerturbationMatrixElementTime);
    end;
    if HasFields then begin
      { Transition moments diagonal in manifold }
      TransitionMoments := Manifold.Molecule.TransitionMoments[Manifold.NodeIndex,Manifold.NodeIndex];
      if Assigned(TransitionMoments) then begin
        for EBField := 0 to 1 do begin
          if EBField = 0 then
            { This factor is correct for electric fields - Debye = 1e-21/c Cm}
            Conv := ExField.E * ConversionFactor(MHz,Manifold.Mixture.Units)*1e-27/SpeedOfLight/Planck
          else
            { function returns BohrMagneton, so convert from J to required energy unit }
            Conv := ExField.B * ConversionFactor(MHz,Manifold.Mixture.Units) * 1e-6 * BohrMagneton / Planck;
          { TODO -cField: Caching likely to help a lot, as part matrices field independent }
          { TODO -cField: Check/Handle Rank <> 1 case; Sum over ranks }
          if Conv <> 0 then begin
            EMat := TransitionMoments.OriginalTransitionMatrix(Self,Self,TransitionMoments.MaxRank,nil,False,EBField=0,EBField=1,nil);
            for b := 1 to Size do
              for bb := 1 to b do begin
                CMul(Sum, EMat[b,bb], Conv);
                if not IsZero(Sum) then begin
                  CAdd(Sum, Matrix[b,bb]);
                  {$IFOPT R+}
                  if not MathSup.SameValue(EMat[b,bb],EMat[bb,b]) then
                    WriteLn(Log, 'Transition Matrix not symmetric:', EMat[b,bb].Re, ' ', EMat[bb,b].Re);
                  {$ENDIF}
                  if BlockMatrix then
                    OrIn(Connections[b],Connections[bb]);
                  Matrix.AutoSetValue(b,bb,Sum);
                  if b <> bb then begin
                    Sum.im := - Sum.im;
                    Matrix[bb,b] := Sum;
                  end;
                end;
              end;
            EMat.Free;
          end;
        end;
      end;
    end;
    if BlockMatrix and (CountBlocks <> 1) then
      FactorMatrix(Matrices, Connections);
  except
    for b := 0 to High(Matrices) do
      FreeAndNil(Matrices[b]);
    raise;
  end;
end;

procedure TEigenvalues.ShowBasisMatrix(M: TComplexMatrix; const Title: string; Offset: Integer);
begin
  ShowTransitionMatrix(Self, M, Title, Offset);
end;

procedure TEigenvalues.ShowTransitionMatrix(Ev: TEigenValues; M: TComplexMatrix; const Title: string; Offset: Integer; ShowBasis: Boolean);
var
  b, bb, sw: Integer;
  LWidths: array of Integer;
  s: string;
  IsComplex: Boolean;
  procedure ShowState(Ev: TEigenValues);
  var
    s: string;
  begin
    if Ev.HasFields then
      s := ', ' + Ev.FieldLabel
    else
      s := '';
    Write(Log, Ev.Basis[1].State.Manifold.Name,', J = ', Ev.Manifold.Mixture.QuantumNumberToStr(Ev.J), ', Symmetry = ', Manifold.Molecule.SymName(Ev.Sym), s);
  end;
  function SLabel(i: Integer; Ev: TEigenValues): string;
  begin
    if ShowBasis then
      Result := Ev.Basis[i+Offset].BasisLabel
    else
      Result := Ev.DominantBasis[i+Offset].StateLabel(Ev.Manifold.Mixture.QuantumNumberFormat);
  end;
begin
  { Print Hamiltonian or other Matrix, including basis state labels }
  if Size = 0 then Exit;
  IsComplex := IsComplexMatrix(M);
  Write(Log, Title,' ');
  if Ev <> Self then Write(Log, '<');
  ShowState(Self);
  if Ev <> Self then begin
    Write(Log, '| |');
    ShowState(Ev);
    Write(Log, '>');
  end;
  sw := 4; { Work out width of Row labels }
  for b := 1 to M.RowCount do
    sw := Max(sw, Length(SLabel(b,Self))+2);
  WriteLn(Log);
  Write(Log, ' ':sw);
  SetLength(LWidths, Ev.Size+1); { Index 0 is not used }
  for b := 1 to M.RowCount do begin
    s := '|'+SLabel(b,Ev)+'>';
    LWidths[b] := Length(s);
    Write(Log, s);
    if IsComplex then
      Write(Log, ' i', ' ':Length(s)-2);
  end;
  WriteLn(Log);
  for b := 1 to M.RowCount do begin
    Write(Log, '<'+SLabel(b,Self)+'|':sw);
    for bb := 1 to M.RowCount do begin
      Write(Log, DtosMA(M.Re[b,bb],LWidths[bb],-1));
      if M.Mode <> cmReal then
        Write(Log, DtosMA(M.Im[b,bb],LWidths[bb],-1));
    end;
    WriteLn(Log);
  end;
  WriteLn(Log);
end;

procedure TEigenvalues.Multiply(M: TComplexMatrix);
{ Multiply matrix M by eigenvalues
  ( Actually Pre multiply by transpose of eigenvector matrix )
    so requires match of Rows of M with size of Basis }
var
  i: Integer;
  N: Integer;
  A: PDouble;
  No, Yes: Char;
  alpha, beta: Double;
  zAlpha, zBeta: Complex;
begin
  if M.RowCount <> Size then
    raise Exception.Create('TEigenvalues.Multiply: Size mismatch');
  N := Size;
  {$DEFINE BLAS}
  {$IFDEF BLAS}
    if (C.Mode = cmReal) and (M.Mode = cmReal) then begin
      M.Size := M.Size * 2; { Allocate workspace }
      A := PDouble(PChar(M.FirstItem) + (M.Size div 2)*SizeOf(Double));
      System.Move(M.FirstItem^, A^, (M.Size div 2)*SizeOf(Double));
      No := 'N';
      Yes := 'C';
      alpha := 1;
      beta := 0;
      i := M.ColCount;
      dgemm_(Yes, No, N, i, N, alpha, C.FirstItem^, N, A^, N, beta, M.FirstItem^, N, 1, 1);
      M.Size := M.Size div 2;
    end else begin
      C.Mode := cmNComplex;
      M.Mode := cmNComplex;
      M.Size := M.Size * 2; { Allocate workspace }
      A := PDouble(PChar(M.FirstItem) + (M.Size div 2)*SizeOf(Double));
      System.Move(M.FirstItem^, A^, (M.Size div 2)*SizeOf(Double));
      No := 'N';
      Yes := 'C';
      zAlpha := z1;
      zBeta := z0;
      i := M.ColCount;
      zgemm_(Yes, No, N, i, N, zAlpha, PComplex(C.FirstItem)^, N, PComplex(A)^, N, zBeta, PComplex(M.FirstItem)^, N, 1, 1);
      M.Size := M.Size div 2;
    end;
  {$ELSE}
  //W: array of Double;
  //Sum: Double;
  //PM, PC: PDouble;
  {$DEFINE ALT}
  {$IFDEF ALT}
  { This ordering is significantly quicker, as transition matrices are typically
    banded and thus not very dense }
  SetLength(W, Size+1); { Index 0 is not used }
  for i := 1 to M.ColCount do begin
    for ii := 1 to M.RowCount do begin
      W[ii] := M[ii,i];
      M[ii,i] := 0;
    end;
    for iii := 1 to M.RowCount do begin
      Sum := W[iii]; { = original M[iii,i] }
      if Sum <> 0 then begin
        PM := M.Item[1,i];
        PC := C.Item[iii,1];
        for ii := 1 to M.RowCount do begin
          {$IFDEF SIMPLE}
          M[ii,i] := M[ii,i] + Sum*C[iii,ii];
          {$ELSE}
          PM^ := PM^ + Sum*PC^;
          Inc(PM);
          Inc(PC,N);
          {$ENDIF}
        end;
      end;
    end;
  end;
  {$ELSE}
  SetLength(W, Size+1); { Index 0 is not used }
  for i := 1 to M.ColCount do begin
    for ii := 1 to Size do begin
      Sum := 0;
      for iii := 1 to Size do
        Sum := Sum + C[iii,ii] * M[iii,i];
      W[ii] := Sum;
    end;
    for ii := 1 to Size do
      M[ii,i] := W[ii];
  end;
  {$ENDIF}
  {$ENDIF}
end;

function TEigenvalues.Qno(AIndex: Integer; QnoType: TQnoType): TQuantumNumber;
begin
  Result := DominantBasis[AIndex].Qno(QnoType);
end;

function TEigenvalues.Transform(M: TComplexMatrix): TColumnVector;
{ Return diagnonal elements of M transformed by eigenvectors }
{ M is overwritten }
var
  i, ii: Integer;
  Sum: Double;
  Csum, T: Complex;
begin
  Multiply(M);
  Result := TColumnVector.Create(Size);
  try
    if M.Mode <> cmReal then begin
      if C.Mode = cmReal then
        raise NotYet; { if M is Complex, C should be also }
      for i := 1 to Size do begin
        Csum := z0;
        for ii := 1 to Size do begin
          CMul(T, M[i,ii], C[ii,i]);
          CAdd(Csum,T);
        end;
        Result[i] := Sqrt(SqrMod(Csum));
      end;
    end else
      for i := 1 to Size do begin
        Sum := 0;
        for ii := 1 to Size do
          Sum := Sum + M.Re[i, ii] * C.Re[ii, i];
        Result[i] := Sum;
      end;
  except
    Result.Free;
    raise;
  end;
end;

procedure TEigenvalues.SortFrom(Start: Integer; Emax: Double);
var
  V: TDarray;
  i: Integer;
begin
  Fixup;
  if (Size-Start <= 1) and (Emax >= 1e99) then
    Exit { 1 element must be sorted }
  else begin
    V := DiagonalMatrixElements(meHrot, Start, True);
    SortBasisOn(V, False, Start);
    if Emax < 1e99 then
      for i := 0 to High(V) do
        if V[i] > Emax then begin
          { Delete unwanted basis states }
          while Size > Start+i do
            Components[i+Start].Free;
          Fixup;
          Exit;
        end;
  end;
end;

function TEigenvalues.StatisticalWeight(Index: Integer): Integer;
begin
  { This allows a basis with mixed nuclear statistical weights;
    only normally required where HasFields }
  if Length(FStatisticalWeights) = 0 then
    Result := FStatisticalWeight
  else
    Result := FStatisticalWeights[Index-1];
end;

function TEigenvalues.SpinSpecies(Index: Integer): Integer;
begin
  { This allows a basis with mixed nuclear statistical weights;
    only normally required where HasFields }
  if Length(FStatWtGroups) = 0 then
    Result := FStatWtGroup
  else
    Result := FStatWtGroups[Index-1];
end;

type
  TBasisShellSorter = class(TShellSorter)
    ByState: Boolean;
    Eigenvalues: TEigenvalues;
    Base: Integer;
    V: TDArray;
    function InOrder(I,J:Integer):Boolean; override;
    procedure Swap(I,J: Integer); override;
  end;

function TBasisShellSorter.InOrder(I,J:Integer):Boolean;
var
  Diff: Integer;
begin
  if ByState then begin
    Diff := Eigenvalues.Basis[I+Base+1].StateIndex - Eigenvalues.Basis[J+Base+1].StateIndex;
    if Diff <> 0 then begin
      Result := Diff <= 0;
      Exit;
    end;
  end;
  Result := V[I] <= V[J];
end;

procedure TBasisShellSorter.Swap(I,J: Integer);
begin
  SwapValues(V[I], V[J], SizeOf(Double));
  SwapValues(Eigenvalues.Basis[I+Base+1], Eigenvalues.Basis[J+Base+1], SizeOf(TBasisState));
end;

procedure TEigenvalues.SortBasisOn(V: TDarray; ByState: Boolean; Base: Integer);
{ Sort the basis states starting from Base;
  Given V[b] for each basis state b, shuffle so that V[b] is in ascending order }
var
  Sorter: TBasisShellSorter;
  i: Integer;
begin
  Sorter := TBasisShellSorter.Create;
  Sorter.ByState := ByState;
  Sorter.Eigenvalues := Self;
  Sorter.Base := Base;
  Sorter.V := V;
  Sorter.Sort(Size-Base);
  for i := Base to Size-1 do
    Basis[i+1].ComponentIndex := i;
  Sorter.Free;
end;

procedure TEigenvalues.DoLimitedEigenSearch(Co: TComplexMatrix; Offset: Integer);
  { Must be called with Offset = 0 first }
var
  BasisStateIndex, nBasisStates, nContr: Integer;
  Ind: array of Integer;
  Contr: array of Double;
  IContr: array of Integer;
  Sum: Double;
  i, ii, k, N: Integer;
begin
  { Leave Individual basis states in original order,
    but make sure states are correct }
  if (LastStateIndex <= 1) or (Size <= 1) then begin
    FreeAndNil(FDominantBasisIndex); { Leave as is }
    Exit;
  end;
  if Assigned(FDominantBasisIndex) and (Offset = 0) then
    Exit; { Probably never called }
  if not Assigned(FDominantBasisIndex) then
    FDominantBasisIndex := TIntColumnVector.Create(Size);
  N := Co.RowCount;
  { For each basis state index:
    - find out ouw many basis states it corresponds to
    - select exactly that many on the basis of largest coefficients }
  SetLength(Ind, N);   { Index into basis for this BasisStateIndex }
  SetLength(Contr, N);
  SetLength(IContr, N);
  for BasisStateIndex := 0 to LastStateIndex-1 do begin
    { Ind is Index into basis for this BasisStateIndex }
    { Logically: SetLength(Ind, N); }
    nBasisStates := 0;
    for k := 1 to N do
      if Basis[k+Offset].StateIndex = BasisStateIndex then begin
        Ind[nBasisStates] := k;
        Inc(nBasisStates);
      end else if (Basis[k+Offset].StateIndex < 0) or (Basis[k+Offset].StateIndex >= LastStateIndex) then
        raise Exception.Create(Basis[k].BasisLabel); { TODO: Remove }
    if nBasisStates = 0 then
      Continue;
    { Ind is Index into basis for this BasisStateIndex }
    { Logically, but causes unecessary calls to realloc:
      SetLength(Ind, nBasisStates); SetLength(Contr, nBasisStates); SetLength(IContr, nBasisStates); }
    nContr := 0; { Contr and IContr keep track of the most siginificant contributions }
    for k := 1 to N do
      if FDominantBasisIndex[k+Offset] = 0 then begin
        { Work out how much eigenvalue k has from this particular BasisStateIndex }
        Sum := 0;
        for i := 0 to nBasisStates-1 do
          Sum := Sum + Co.SqrMod[Ind[i],k];
        { Insert into sorted list of the nBasisStates largest contributions }
        // WriteLn(Log,k, ' ', Sum);
        { Find insert position in list }
        i := 0;
        while (i < nContr) and (Sum > Contr[i]) do
          Inc(i);
        if nContr < nBasisStates then begin
          { Not enough numbers in list yet, so grow list
            - Some values to shift up }
          for ii := nContr downto i+1 do begin
            Contr[ii] := Contr[ii-1];
            IContr[ii] := IContr[ii-1];
          end;
          Contr[i] := Sum;
          IContr[i] := k;
          Inc(nContr);
        end else if i > 0 then begin
          Dec(i);
          { Replace a current entry, shifting lower values down }
          for ii := 0 to i-1 do begin
            Contr[ii] := Contr[ii+1];
            IContr[ii] := IContr[ii+1];
          end;
          Contr[i] := Sum;
          IContr[i] := k;
        end;
      end;
    if nContr <> nBasisStates then
      raise Exception.CreateFmt('DoLimitedEigenSearch %d %d', [nContr, nBasisStates]);
    Sort(IContr,nBasisStates);
    for i := 0 to nBasisStates-1 do
      FDominantBasisIndex[IContr[i]+Offset] := Ind[i]+Offset;
  end;
end;

procedure TEigenvalues.DoEigenSearch(Co: TComplexMatrix; Offset: Integer);
  { Must be called with Offset = 0 first }
var
  i, ii, iimax, N: Integer;
  cmax, v: Double;
  Taken: array of Boolean;
begin
  if Assigned(FDominantBasisIndex) and (Offset = 0) then
    Exit; { Probably never called }
  if not Assigned(FDominantBasisIndex) then
    FDominantBasisIndex := TIntColumnVector.Create(Size);
  N := Co.RowCount; { Size of this block }
  SetLength(Taken,N+1);
  FillChar(Taken[0], SizeOf(Boolean)*(N+1),0);
  { Now scan eigenvalues for largest coefficient of each basis state }
  { Any coefficients > 1/Sqrt(2) are definite assignments }
  { Note that the assignment will be unstable for coefficients that are exactly 1/sqrt(2);
    loop below has bias towards assigning in basis order, but doesn't
    in the case that the coefficents**2 are just less than 0.5 }
  for i := 1 to N do
    for ii := 1 to N do
      if not Taken[ii] then begin
        v := Co.SqrMod[ii,i];
        if v >= 0.5 then begin
          FDominantBasisIndex[i+Offset] := ii + Offset;
          Taken[ii] := True;
          Break;
        end;
      end;
  { For any states not assigned above }
  for i := 1 to N do
    if FDominantBasisIndex[i+Offset] = 0 then begin
      cmax := 0;
      iimax := 0;
      for ii := 1 to N do
        { Find largest remaining coefficient }
        if not Taken[ii] then begin
          v := Co.SqrMod[ii,i];
          if v > cmax then begin
            iimax := ii;
            cmax := v;
          end;
        end;
      if iimax = 0 then begin
        { Horrible problem - all C zero or Nan; just take first }
        for ii := 1 to N do
          if not Taken[ii] then begin
            iimax := ii;
            Break;
          end;
      end;
      FDominantBasisIndex[i+Offset] := iimax + Offset;
      Taken[iimax] := True;
    end;
end;

function TEigenvalues.GetDominantBasisState(AIndex: Integer): TBasisState; {$IFDEF FPCINLINE} inline; {$ENDIF}
begin
  if Assigned(FDominantBasisIndex) then
    AIndex := FDominantBasisIndex[AIndex];
  Result := Basis[AIndex];
end;

function TEigenvalues.GetEnergy(AIndex: Integer): Double;
begin
  Result := E[AIndex];
end;

function TEigenvalues.HasCorrectAttributes(Attributes: Pointer): Boolean;
var
  MatElemTypes: TMatElemTypes;
begin
  Result := Size = 0;
  if not Result then begin
    MatElemTypes := TMatElemTypes({$IFDEF FPC}Integer{$ELSE}Byte{$ENDIF}(Attributes));
    Result := Assigned(E);
    if Result and Manifold.HasWidth and (meWidth in MatElemTypes) then
      Result := Assigned(Widths);
    if Result and (MatElemTypes <> []) then
      Result := Assigned(C);
    if Result and (meFloatedDerivatives in MatElemTypes) or (meAllDerivatives in MatElemTypes) then
      Result := FDerivativesCalculated;
  end;
end;

procedure TEigenvalues.ShowEigenvalues;
var
  b, bb, sw: Integer;
  s: string;
  BWidths: array of Integer;
begin
  if HasFields then
    s := ', ' + FieldLabel
  else
    s := '';
  WriteLn(Log, 'Eigenvalues for ' + Manifold.Name + ', J = ', Manifold.Mixture.QuantumNumberToStr(J), ', Symmetry = ', Manifold.Molecule.SymName(Sym), s);
  sw := 1; { Maximum width of state label }
  for b := 1 to Size do
    sw := Max(sw, Length(DominantBasis[b].StateLabel(Manifold.Mixture.QuantumNumberFormat)));
  Write(Log, ' ':18+sw);
  if Assigned(Widths) then Write(Log, ' ':10);
  SetLength(BWidths, Size+1);
  for b := 1 to Size do begin
    s := Basis[b].BasisLabel;
    BWidths[b] := Max(10, Length(s)+2);
    Write(Log, '|'+s+'>':10);
    if C.Mode <> cmReal then
      Write(Log, '  i', ' ':7);
  end;
  WriteLn(Log);
  for b := 1 to Size do begin
    Write(Log, b:3, ' ', DominantBasis[b].StateLabel(Manifold.Mixture.QuantumNumberFormat):sw+1, E[b]:13:3);
    if Assigned(Widths) then Write(Log, Widths[b]:10:5);
    for bb := 1 to Size do
      if C.Mode <> cmReal then begin
        Write(Log, C[bb,b].re:BWidths[bb]:6);
        Write(Log, C[bb,b].im:10:6);
      end else
        Write(Log, C.Re[bb,b]:BWidths[bb]:6);
    WriteLn(Log);
  end;
  WriteLn(Log);
end;

procedure TEigenvalues.CalculateDerivatives(MatElemTypes: TMatElemTypes);
var
  nDerivatives: Integer;
  Matrices: TComplexMatrices;
  procedure HandleParameters(MNode: TMNode);
  var
    I: Integer;
  begin
    for I := 1 to MNode.nParameters do
      if ((meAllDerivatives in MatElemTypes) and MNode.DerivativeRequired(I))
         or ((meFloatedDerivatives in MatElemTypes) and MNode.Floating[I]) then
        try
          if nDerivatives >= High(Derivatives) then
            SetLength(Derivatives, nDerivatives + MNode.nParameters);
          MNode.ModifyParameter(I,1,False, False);
          Manifold.CheckProperties;
          MatrixElements(Matrices, False, meHrot, not NoPerturbations);
          Derivatives[nDerivatives].Values := Transform(Matrices[0]);
          Derivatives[nDerivatives].Index := I;
          Derivatives[nDerivatives].MNode := MNode;
          Inc(nDerivatives);
        finally
          MNode.ModifyParameter(I,0,False, False);
          Matrices[0].Free;
        end;
    for I := 0 to MNode.Count - 1 do
      HandleParameters(MNode.Components[I] as TMNode);
  end;
begin
  if FDerivativesCalculated then
    Exit;
  nDerivatives := 0;
  Manifold.StartDerivatives;
  try
    HandleParameters(Manifold);
    SetLength(Derivatives,nDerivatives);
    FDerivativesCalculated := True;
  finally
    Manifold.StopDerivatives;
  end;
end;

procedure TEigenvalues.DoMake(RequiredAttributes: Pointer);
var
  CalculateVectors, CalculateWidths: Boolean;
  MatElemTypes: TMatElemTypes;
  Matrices: TComplexMatrices;
  i, Base: Integer;
begin
  if Size = 0 then
    Exit;
  MatElemTypes := TMatElemTypes({$IFDEF FPC}Integer{$ELSE}Byte{$ENDIF}(RequiredAttributes));
  CalculateWidths := Manifold.HasWidth and (meWidth in MatElemTypes) and not Assigned(Widths);
  CalculateVectors := ((meHrot in MatElemTypes) or CalculateWidths or (meAllDerivatives in MatElemTypes) or (meFloatedDerivatives in MatElemTypes)) and not Assigned(C);
  TStart(EigenvaluesTime);
  if not Assigned(E) or CalculateVectors then begin
    if Assigned(E) then
      WriteLn(Log,'**********', ThreadName); { TODO: Remove }
    try
      TStart(MakeHamiltonianTime);
      MakeHamiltonian(Matrices, Manifold.Molecule.BlockMatrix, Manifold.Mixture.BasisOrder);
      TEnd(MakeHamiltonianTime);
      if PrintLevel >= plHamiltonian then begin
        WriteLn(Log);
        if Length(Matrices) = 1 then
          ShowBasisMatrix(Matrices[0], 'Hamiltonian Matrix for ', 0)
        else begin
          Base := 0;
          for i := 0 to High(Matrices) do begin
            ShowBasisMatrix(Matrices[i], 'Block ' + IntToStr(i+1) + ' of Hamiltonian Matrix for ', Base);
            Inc(Base,Matrices[i].RowCount);
          end;
        end;
      end;
      Diagonalize(Matrices,
                  (meHrot in MatElemTypes) or CalculateWidths or (meFloatedDerivatives in MatElemTypes) or (meAllDerivatives in MatElemTypes)
                  or Manifold.EigenSearch or (PrintLevel >= plEigenVectors),
                  Manifold.EigenSearch, Manifold.LimitSearch);
    finally
      for i := 0 to High(Matrices) do
        Matrices[i].Free;
    end;
    {$IFDEF EXPLICITTEST}
    DoExplicitTest(Result);
    {$ENDIF}
  end;
  if (meFloatedDerivatives in MatElemTypes) or (meAllDerivatives in MatElemTypes) then
    CalculateDerivatives(MatElemTypes);
  if (MatElemTypes <> []) or (PrintLevel >= plEigenVectors) then begin
    if CalculateWidths then
      MakeWidths(PrintLevel);
    if PrintLevel >= plEigenVectors then
      ShowEigenvalues;
  end;
  TEnd(EigenvaluesTime);
end;

function TEigenvalues.DominantLabel(AIndex: Integer; QuantumNumberFormat: TQuantumNumberFormat): string;
begin
  Result := DominantBasis[AIndex].StateLabel(QuantumNumberFormat);
end;

procedure TEigenvalues.Make(Wait: Boolean; MatElemTypes: TMatElemTypes);
begin
  Make(Wait, Pointer({$IFDEF FPC}PtrInt(Integer{$ELSE}Byte({$ENDIF}(MatElemTypes))));
end;

procedure TEigenvalues.MakeHamiltonian(out Matrices: TComplexMatrices; BlockMatrix: Boolean; BasisOrder: TBasisOrder);
var
  V: TDarray;
begin
  if BasisOrder <> boDefault then begin
    V := DiagonalMatrixElements(meHrot, 0, not NoPerturbations);
    SortBasisOn(V, BasisOrder = boStateDiagonal);
  end;
  MatrixElements(Matrices, BlockMatrix, meHrot, not NoPerturbations);
end;

procedure TEigenvalues.MakeWidths(APrintLevel: TPrintLevel);
var
  ComplexMatrices: TComplexMatrices;
begin
  { TODO -cField: How should this handle complex eigenvectors? }
  MatrixElements(ComplexMatrices, False, meWidth);
  try
    if APrintLevel >= plMatrices then
      ShowBasisMatrix(ComplexMatrices[0], 'Width Matrix for ', 0);
    FWidths := Transform(ComplexMatrices[0]);
  finally
    ComplexMatrices[0].Free;
  end;
end;

procedure TEigenvalues.Diagonalize(var Matrices: TComplexMatrices; EigenvectorsWanted, EigenSearch, LimitSearch: Boolean);
var
  TempE, TE: TColumnVector;
  i, k, kk, Base, Block, BlockSize, BlockNo: Integer;
  BlockStarts, BlockP, BlockEnds, Ind, BlockSource: array of Integer;
  Emin: Double;
  TI: TIntColumnVector;
begin
  { Do do not overwrite E if it already exists }
  if not Assigned(E) then begin
    FE := TColumnVector.Create(Size);
    TempE := E;
  end else
    TempE := TColumnVector.Create(Size);
  try
    if Length(Matrices) = 1 then begin
      FC := Matrices[0];
      Matrices[0] := nil;
      CEigen(TempE, C, EigenvectorsWanted);
      if EigenSearch then begin
        TStart(EigenSearchTime);
        if LimitSearch then
          DoLimitedEigenSearch(C)
        else
          DoEigenSearch(C);
        TEnd(EigenSearchTime);
      end;
    end else begin
      Base := 0;
      SetLength(BlockStarts, Length(Matrices));
      SetLength(BlockEnds, Length(Matrices));
      for i := 0 to High(Matrices) do begin
        BlockStarts[i] := Base+1;
        BlockSize := Matrices[i].RowCount;
        BlockEnds[i] := Base+BlockSize+1;
        CEigen(TempE, Matrices[i], EigenvectorsWanted, Base);
        if EigenSearch then begin
          TStart(EigenSearchTime);
          if LimitSearch then
            DoLimitedEigenSearch(Matrices[i],Base)
          else
            DoEigenSearch(Matrices[i],Base);
          TEnd(EigenSearchTime);
        end;
        Inc(Base, BlockSize);
      end;
      { Now merge sort eigenvalues, recording indices as we go }
      SetLength(Ind, Size);
      SetLength(BlockSource, Size);
      BlockP := Copy(BlockStarts);
      i := 0;
      while i < Size do begin
        BlockNo := -1;
        { Find next lowest energy by searching individual blocks }
        for Block := 0 to High(Matrices) do
          if (BlockP[Block] < BlockEnds[Block])
             and ((BlockNo < 0) or (TempE[BlockP[Block]] < Emin)) then begin
            BlockNo := Block;
            Emin := TempE[BlockP[Block]];
          end;
        Ind[i] := BlockP[BlockNo];
        BlockSource[i] := BlockNo;
        Inc(BlockP[BlockNo]);
        Inc(i);
      end;
      { Permute energies }
      TE := TColumnVector.Create(Size);
      for i := 1 to Size do
        TE[i] := TempE[Ind[i-1]];
      TempE.AssignAndNil(TVector(TE));
      if Assigned(FDominantBasisIndex) then begin
        { Permute Basis indices }
        TI := TIntColumnVector.Create(Size);
        for i := 1 to Size do
          TI[i] := FDominantBasisIndex[Ind[i-1]];
        FDominantBasisIndex.AssignAndNil(TGHandle(TI));
      end else begin
        { No permutation of individual matrices }
        FDominantBasisIndex := TIntColumnVector.Create(Size);
        for i := 1 to Size do
          FDominantBasisIndex[i] := Ind[i-1];
      end;
      if EigenvectorsWanted then begin
        { Copy eigenvectors into position }
        FC := TComplexMatrix.Create(Size,Size,Matrices[0].Mode);
        FC.Zero;
        for i := 1 to Size do begin
          Block := BlockSource[i-1];
          Base := BlockStarts[Block]-1;
          k := Ind[i-1]-Base;
          for kk := 1 to Matrices[Block].RowCount do
            C[kk+Base,i] :=  Matrices[Block][kk,k];
        end;
      end;
    end;
  finally
    for i := 0 to High(Matrices) do
      FreeAndNil(Matrices[i]);
    if E <> TempE then
      TempE.Free;
  end;
end;

{ TBasisState }

constructor TBasisState.Create(AOwner: TComponent; AF: TQuantumNumber; ASym: Integer);
begin
  inherited Create(AOwner);
  F := AF;
  Sym := ASym;
end;

function TBasisState.BasisLabelKey: string;
begin
  Result := 'Name  J +-';
end;

function TBasisState.BasisLabel: string;
begin
  if State.Manifold.Molecule.nManifolds > 1 then
    Result := State.Manifold.Name + ' '
  else
    Result := '';
  Result := Result + State.Name
            + Format(' %2s', [State.Manifold.Mixture.QuantumNumberToStr(J)]);
  if Phase > 0 then
    Result := Result + ' +'
  else if Phase < 0 then
    Result := Result + ' -'
  else
    Result := Result + '  ';
end;

function TBasisState.StateLabel(QuantumNumberFormat: TQuantumNumberFormat): string;
begin
  if State.Manifold.Molecule.nManifolds > 1 then
    Result := State.Manifold.Name + ' '
  else
    Result := '';
  Result := Result + State.Name;
end;

function TBasisState.StateLabelKey: string;
begin
  Result := 'Name';
end;

function TBasisState.HyperfineStateLabel(WithBlanks: Boolean): string;
var
  FF, Fn: TQuantumNumber;
  i: Integer;
  procedure AddQno(Q: TQuantumNumber);
  begin
    if WithBlanks then
      Result := Format(' %2s', [State.Manifold.Mixture.QuantumNumberToStr(Q)]) + Result
    else begin
      if Length(Result) > 0 then
        Result := ',' + Result;
      Result := State.Manifold.Mixture.QuantumNumberToStr(Q) + Result;
    end;
  end;
begin
  Result := '';
  with Owner as TEigenvalues do
    if HasFields then
      AddQno(M);
  if Length(HyperfineState) > 0 then begin
    FF := F;
    for i := High(HyperfineState) downto 0 do begin
      if State.Nucleus(i).AsNext then
        Fn := HyperfineState[i]
      else begin
        Fn := FF;
        Dec(FF,HyperfineState[i]);
      end;
      AddQno(Fn);
    end;
  end;
end;

function TBasisState.HyperfineStateLabelKey(WithBlanks: Boolean; ExtraChar: Char): string;
var
  i: Integer;
  procedure AddQno(const Q: string);
  begin
    if WithBlanks then
      Result := ' ' + Result
    else if Result <> '' then
      Result := ',' + Result;
    if ExtraChar <> ' ' then
      Result := ExtraChar + Result;
    Result := Q + Result;
 end;
begin
  Result := '';
  with Owner as TEigenvalues do
    if HasFields then
      AddQno('M');
  if Length(HyperfineState) > 0 then begin
    AddQno('F');
    for i := Length(HyperfineState)-2 downto 0 do begin
      if State.Nucleus(i).AsNext then
        AddQno(Format('I%d%d', [i+1,i+2]))
      else
        AddQno(Format('F%d', [i+1]));
     end;
  end;
end;

function TBasisState.HyperfineTxNameKey(WithBlanks: Boolean): string;
begin
  if (Length(HyperfineState) > 0) or (Owner as TEigenvalues).HasFields then begin
    Result := HyperfineStateLabelKey(WithBlanks, '''');
    if WithBlanks then
      Result := Result + ' '
    else
      Result := Result + ',';
    Result := Result + HyperfineStateLabelKey(WithBlanks,'"')
  end else
    Result := '';
end;

function TBasisState.TxName(Upper, Lower: TBasisState): string;
begin
  Result := '';
end;

function TBasisState.TxNameKey(Upper, Lower: TBasisState): string;
begin
  Result := '';
end;

function TBasisState.TrackIndex: Integer;
begin
  Result := 0; { Default is all states in same class }
end;

function TBasisState.Qno(QnoType: TQnoType): TQuantumNumber;
var
  i, StopAt: Integer;
begin
  case QnoType of
    qtJ: Result := J;
    qtN: Result := N;
    qtF: Result := F;
    qtM: Result := M;
    qtA, qtB: raise Exception.Create('Unimplemeted quantum number');
  else
    Result := F;
    StopAt := Ord(QnoType) - Ord(qtF);
    if (StopAt < Length(HyperfineState)) and State.Nucleus(StopAt).AsNext then
      Result := HyperfineState[StopAt]
    else
      for i := Length(HyperfineState)-2 downto StopAt do
        if not State.Nucleus(i).AsNext then
          Dec(Result,HyperfineState[i+1]);
  end;
end;

{ TSpectrum }

procedure TSpectrum.Clear;
begin
  FnLines := 0;
  ITotal := 0;
end;

procedure TSpectrum.AddIn(LineList: TLineList);
var
  i: Integer;
begin
  for i := 0 to LineList.nLines - 1 do
    with LineList.Line[i] do
      AddLine(Frequency,Strength,Width,Colour,LineList.Units);
end;

procedure TSpectrum.AddIn(LineList: TLineList; Simulation: TSimulation; var Problem: Exception; FirstLine: Integer; Increment: Integer);
var
  i, c: Integer;
begin
  { FirstLine and Increment allow the core loop to be parallelized }
  { Currently this is only implemented in TFitPlot, as it is only here the loop
    is particularly time consuming }
  if FirstLine = 0 then
    PushProgress('Accumulating Spectrum', LineList.nLines div 100);
  try
    if (Increment = 1) and IsNormalized(Simulation.IntensityUnits) then
      Simulation.Mixture.PreCalculatePartitionFunctions(Simulation, Problem);
    c := 0;
    TStart(AddInLinesTime);
    i := FirstLine;
    while i < LineList.nLines do begin
      AddLine(LineList.Line[i], Simulation, LineList.Units, Problem);
      Dec(c);
      if (c < 0) and (FirstLine = 0) then begin
        SetProgress(i div 100);
        c := 100;
      end else if Aborted then
        Break;
      Inc(i,Increment);
    end;
    TEnd(AddInLinesTime);
  finally
    if FirstLine = 0 then
      PopProgress;
  end;
end;

constructor TSpectrum.Create(AUnits: TResonanceUnits; AIsDistribution: Boolean);
begin
  inherited Create;
  Colour := clNone;
  Clear;
  Units := AUnits;
  FIsDistribution := AIsDistribution;
end;

procedure TSpectrum.AddLine(Frequency, Strength, Width: Double; AColour: TColor; AUnits: TResonanceUnits);
var
  Line: TLine;
  var Problem: Exception;
begin
  Line.Frequency := Frequency;
  Line.Strength := Strength;
  Line.Width := Width;
  Line.EigenvaluesL := nil;
  Line.EigenvaluesU := nil;
  Line.Colour := AColour;
//  if AUnits <> Units then
//    ConvertLine(Line, AUnits, Units);
  Problem := nil;
  AddLine(Line, nil, AUnits, Problem);
  if Assigned(Problem) then
    raise Problem; { With current design, should never happen }
end;

procedure TSpectrum.SetHeadingPrinted(const Value: Boolean);
begin
  if not Value then begin
    ITotal := 0;
    nPrinted := 0;
  end;
  FHeadingPrinted := Value;
end;

function EstErr(EvU, EvL: TEigenvalues; IndexU, IndexL: Integer):Double;
var
  Mixture: TMixture;
  Molecule: TMolecule;
  VarCovar: TMatrix;
  Derivs: array of Double;
  i, j, nPar: Integer;
  v: Double;
begin
  Result := 0;
  if not Assigned(EvL) then
    Exit;
  Mixture := EvL.Manifold.Mixture;
  Molecule := EvL.Manifold.Molecule;
  nPar := Mixture.Correlations.nEntries;
  { Make up variance-covariance matrix }
  VarCovar := TMatrix.Create(nPar, nPar);
  with Mixture.Correlations.FitDetails do
    for i := 0 to nPar - 1 do begin
      VarCovar[i+1,i+1] := Sqr(Parameters[i].StdDev);
      for j := 0 to i - 1 do begin
        VarCovar[i+1,j+1] := Mixture.Correlations.FitDetails.Correlation[i,j]*Parameters[i].StdDev*Parameters[j].StdDev;
        VarCovar[j+1,i+1] := VarCovar[i+1,j+1];
      end;
    end;
  { Now need dObs/dCalc, and Match }
  SetLength(Derivs, nPar);
  for i := 0 to nPar - 1 do begin
    v := 0;
    j := EvL.FindDerivatives(Mixture.Correlations.FitDetails.Parameters[i]);
    if j >= 0 then
      v := EvL.Derivatives[j].Values[IndexL];
    if Assigned(EvU) then begin
      v := -v;
      j := EvU.FindDerivatives(Mixture.Correlations.FitDetails.Parameters[i]);
      if j >= 0 then
        v := v + EvU.Derivatives[j].Values[IndexU];
    end;
    Derivs[i] := v;
  end;
  Result := 0;
  for i := 0 to nPar - 1 do
    for j := 0 to nPar - 1 do
      Result := Result + Derivs[i]*VarCovar[j+1,i+1]*Derivs[j];
  VarCovar.Free;
  Result := Sqrt(Result);
end;

procedure TSpectrum.ShowLine(var Line: TLine; var F: TextFile; PrintLevel: TPrintLevel; Threshold: Double; LinFormat: Boolean; var Problem: Exception; Footer: Boolean);
var
  Intensity: Double;
  l: Integer;
  Ev1, Ev2: TEigenvalues;
  Mixture: TMixture;
  Molecule: TMolecule;
  FirstField: Boolean;
  Lab: string;
  bbuf: shortstring;
  Separator: Char;
  procedure ShowField(const Field: string);
  begin
    if Separator = #0 then begin
      Write(F,Char(Length(Field)));
      Write(F,Field);
      SetLength(bbuf,20);
    end else begin
      if FirstField then
        FirstField := False
      else
        Write(F,Separator);
      if Separator = ',' then
        if Pos('"', Field) <> 0 then
          Write(F,AnsiQuotedStr(Field,'"'))
        else
          Write(F,Field)
      else
        Write(F,Field);
    end;
  end;
  procedure ShowEOL;
  begin
    if Separator = #0 then
      ShowField(#10) { Must write end of line to avoid text file mis-handling }
    else
      WriteLn(F);
  end;
  procedure ShowBuffer;
  begin
    SetLength(bbuf, (Ord(bbuf[1])-200) mod 16 + 1);
    Write(F,bbuf);
  end;
  procedure ShowInt(V: Integer);
  begin
    if Separator = #0 then begin
      bbuf[1] := Char(Sizeof(Integer)+200);
      PInteger(@bbuf[2])^ := V;
      ShowBuffer;
    end else
      ShowField(IntToStr(V));
  end;
  procedure ShowQno(V: TQuantumNumber);
  begin
    if Separator = #0 then
      ShowInt(V)
    else
      ShowField(Mixture.QuantumNumberToStr(V));
  end;
  procedure ShowSym(J: TQuantumNumber; Sym: Integer);
  begin
    if Separator = #0 then
      ShowInt(Molecule.JAdjustedSym(J,Sym))
    else
      ShowField(Molecule.JAdjustedSymName(J,Sym));
  end;
  procedure ShowD(V: Double);
  begin
    if Separator = #0 then begin
      bbuf[1] := Char(SizeOf(Double)+216);
      PDouble(@bbuf[2])^ := V;
      ShowBuffer;
    end else
      ShowField(FloatToStr(V));
  end;
  function SpaceQno(J: TQuantumNumber):string;
  var
    W: Integer;
  begin
    if Ev1.HasFields or (J >= 0) then
      Result := Mixture.QuantumNumberToStr(J)
    else
      Result := '-';
    if Mixture.QuantumNumberFormat = qftJ then
      W := 5
    else
      W := 3;
    if Length(Result) < W then
      Result := StringOfChar(' ', W-Length(Result)) + Result
    else if Result[1] <> ' ' then
      Result := ' ' + Result;
  end;
  procedure ShowDerivatives;
    procedure ShowDerivative(const Derivative: TDerivative; Value: Double);
    begin
      ShowField(Derivative.MNode.FullParameterName(Derivative.Index));
      ShowD(Value);
    end;
  var
    i: Integer;
  begin
    if not Mixture.ShowDerivatives or (Length(Ev1.Derivatives) <= 0) then
      Exit;
    if Separator = ' ' then
      Write(F, ' : ');
    if IsDistribution then
      for i := 0 to High(Ev1.Derivatives) do
        ShowDerivative(Ev1.Derivatives[i],Ev1.Derivatives[i].Values[Line.IndexL])
    else if Ev1.Manifold = Ev2.Manifold then
      for i := 0 to High(Ev1.Derivatives) do
        ShowDerivative(Ev2.Derivatives[i], Ev2.Derivatives[i].Values[Line.IndexU]-Ev1.Derivatives[i].Values[Line.IndexL])
    else begin
      for i := 0 to High(Ev2.Derivatives) do
        ShowDerivative(Ev2.Derivatives[i], Ev2.Derivatives[i].Values[Line.IndexU]);
      for i := 0 to High(Ev1.Derivatives) do
        ShowDerivative(Ev1.Derivatives[i], -Ev1.Derivatives[i].Values[Line.IndexL]);
    end;
  end;
var
  SL, nL: Integer;
  JHeader, JorM: string;
  A, StdDev: Double;
begin
  case PrintLevel of
    plCSV: Separator := ',';
    plText: Separator := #9;
    plBinary: Separator := #0;
  else
    Separator := ' ';
  end;
  SetLength(bbuf,20);
  FirstField := True;
  if Footer then begin
    ShowEOL;
    if Separator = ' ' then
      WriteLn(F, nPrinted, ' Lines, Total intensity = ', dtosm(ITotal,14,-1))
    else begin
      ShowInt(nPrinted);
      ShowD(ITotal);
      ShowEOL;
    end;
    Exit;
  end;
  Ev2 := Line.EigenvaluesU;
  Ev1 := Line.EigenvaluesL;
  if not Assigned(Ev1) or (not Assigned(Ev2) and not IsDistribution) then
    Exit;
  Mixture := Ev1.Manifold.Mixture;
  Molecule := Ev1.Manifold.Molecule;
  Intensity := Mixture.Simulations[0].LineIntensity(Line, Mixture.Units, Problem);
  A := Molecule.EinsteinA(Line,Mixture.Simulations[0].PlotUnits);
  ITotal := ITotal + Intensity;
  if Intensity < Threshold then
    Exit;
  Inc(nPrinted);
  l := Length(Mixture.LongestManifoldName);
  if Separator = ' ' then
    nL := Length(IntToStr(Mixture.LargestIndex))+1;
  if not HeadingPrinted then begin
    if Ev1.HasFields then
      JorM := 'M'
    else
      JorM := 'J';
    if Separator = ' ' then begin
      Write(F, Copy('Molecule',1,Length(Mixture.LongestMoleculeName)):Length(Mixture.LongestMoleculeName), ' ');
      if Mixture.QuantumNumberFormat = qftJ then
        JHeader := '   '
      else
        JHeader := ' ';
      JHeader := JHeader+JorM;
      if IsDistribution then begin
        SL := Mixture.LongestSymLength;
        Write(F,'M':l, ' ', JHeader, ' ', 'S':SL, '#':nL, '     Position ')
      end else begin
        SL := Max(Mixture.LongestSymLength,2);
        Write(F,'M':l, '''', JHeader, ''' ', 'S''':SL, '#''':nL,
                'M':l, '"', JHeader, '" ', 'S"':SL, '#"':nL,
                '    Position ');
      end;
      if Mixture.ShowEstUnc then
        Write(F, ' Est Unc ');
      if LinFormat then
        Write(F, ' Wt I      ')
      else if IsDistribution then
        Write(F, ' I               E      Spol   ')
      else
        Write(F, ' I               Eupper        Elower S        A      ');
      WriteLn(F, '  Width');
    end else begin
      if Separator <> ' ' then begin
        ShowField('Line list produced by '+VersionInfo);
        ShowInt(DecodePgopherVersion(VersionInfo));
        ShowEOL;
      end;
      ShowField('Molecule');
      if IsDistribution then begin
        ShowField('Manifold');
        ShowField(JorM);
        ShowField('Sym');
        ShowField('#');
      end else begin
        ShowField('Upper Manifold');
        ShowField(JorM+'''');
        ShowField('Sym''');
        ShowField('#''');
        ShowField('Lower Manifold');
        ShowField(JorM+'"');
        ShowField('Sym"');
        ShowField('#"');
      end;
      ShowField('Position');
      if Mixture.ShowEstUnc then
        ShowField('Est Unc');
      if LinFormat then
        ShowField('Wt');
      ShowField('Intensity');
      if not LinFormat then begin
        if IsDistribution then
          ShowField('E')
        else begin
          ShowField('Eupper');
          ShowField('Elower');
        end;
        ShowField('Strength');
        if not IsDistribution then
          ShowField('A');
      end;
      ShowField('Width');
      ShowField('Branch');
      ShowField('Label');
      ShowEOL;
    end;
  end;
  HeadingPrinted := True;
  FirstField := True;

  if Mixture.ShowEstUnc then
    if IsDistribution then
      StdDev := EstErr(nil, Ev1, -1, Line.IndexL)
    else
      StdDev := EstErr(Ev2, Ev1, Line.IndexU, Line.IndexL);

  if Separator = ' ' then begin
    Write(F, Molecule.Name:Length(Mixture.LongestMoleculeName), ' ');
    SL := Mixture.LongestSymLength;
    if not IsDistribution then begin
      SL := Max(SL,2);
      Write(F,Ev2.Manifold.Name:l, SpaceQno(Ev2.J), ' ', Molecule.JAdjustedSymName(Ev2.J,Ev2.Sym):SL, AtLeast(Line.IndexU,nL), ' ');
    end;
    Write(F,Ev1.Manifold.Name:l, SpaceQno(Ev1.J), ' ', Molecule.JAdjustedSymName(Ev1.J,Ev1.Sym):SL, AtLeast(Line.IndexL,nL),
            AtLeast(Line.Frequency,14,Mixture.Precision));
    if Mixture.ShowEstUnc then
        Write(F, DtosMA(StdDev,9,-1));
    if LinFormat then
      Write(F, ' - ');
    Write(F, DtosMA(Intensity,9,-4));
    if not LinFormat then begin
      if not IsDistribution then
        Write(F, AtLeast(Ev2.Energy[Line.IndexU],14,Mixture.Precision));
      Write(F, AtLeast(Ev1.Energy[Line.IndexL],14,Mixture.Precision),
               DtosMA(Line.Strength,9,-4));
      if not IsDistribution then
        Write(F, DtosMA(A,9,-4));
    end;
    Write(F, DtosMA(Line.Width,9,-4), ' : ', MakeTxName(Line));
    ShowDerivatives;
    WriteLn(F);
  end else begin
    ShowField(Ev1.Manifold.Molecule.Name);
    if not IsDistribution then begin
      ShowField(Ev2.Manifold.Name);
      ShowQno(Ev2.J);
      ShowSym(Ev2.J,Ev2.Sym);
      ShowInt(Line.IndexU);
    end;
    ShowField(Ev1.Manifold.Name);
    ShowQno(Ev1.J);
    ShowSym(Ev1.J,Ev1.Sym);
    ShowInt(Line.IndexL);
    ShowD(Line.Frequency);
    if Mixture.ShowEstUnc then
      ShowD(StdDev);
    if LinFormat then
      ShowField('-');
    ShowD(Intensity);
    if not LinFormat then begin
      if not IsDistribution then
        ShowD(Ev2.E[Line.IndexU]);
      ShowD(Ev1.E[Line.IndexL]);
      ShowD(Line.Strength);
      if not IsDistribution then
        ShowD(A);
    end;
    ShowD(Line.Width);
    Lab := MakeTxName(Line);
    l := Pos(':',Lab);
    if l > 0 then begin
      ShowField(Trim(Copy(Lab,1,l-1)));
      ShowField(Trim(Copy(Lab,l+1,Length(Lab))));
    end else begin
      ShowField('');
      ShowField(Trim(Lab));
    end;
    ShowDerivatives;
    ShowEOL;
  end;
end;

procedure TSpectrum.ShowFooter(var F: TextFile; PrintLevel: TPrintLevel);
var
  Line:TLine; { Unused }
  Execpt: Exception; { Unused }
begin
  ShowLine(Line,F,PrintLevel,0,False,Execpt,True);
end;

{ TLineList }

function TLineList.GetLine(AIndex: Integer): TLine;
begin
  Result := FLines[AIndex];
end;

procedure TLineList.AddLine(const ALine: TLine; Simulation: TSimulation; AUnits: TResonanceUnits; var Problem: Exception);
{ Note that problem is never touched here }
begin
  if nLines >= High(FLines) then begin
    { Optimization - if small buffer fills up, take larger steps }
    if High(FLines) > 0 then
      SetLength(FLines, High(FLines)+50000)
    else
      SetLength(FLines, High(FLines)+10000);
  end;
  FLines[nLines] := ALine;
  if AUnits <> Units then
    ConvertLine(FLines[nLines], AUnits, Units);
  Inc(FnLines);
  if ALine.Frequency < FMin then
    FMin := ALine.Frequency;
  if ALine.Frequency > FMax then
    FMax := ALine.Frequency;
end;

procedure TLineList.Clear;
begin
  inherited;
  FMax := -1.0e99;
  FMin := 1.0e99;
end;

{$IFNDEF CONSOLE}
procedure TLineList.FindLines(Centre, Range: Double; AUnits: TResonanceUnits; AColour: TColor; OnCheckLine: TOnCheckLine);
var
  i: Integer;
  p1, p2: Double;
begin
  p1 := Convert(Centre-Range, AUnits, Units);
  p2 := Convert(Centre+Range, AUnits, Units);
  Centre := (p1+p2)/2;
  Range := Abs(p2-Centre);
  for i := 0 to nLines-1 do
    with FLines[i] do
      { clHide is used to flag two different things:
        - in the linelist itself, clHide lines are ignored
        - for matching colours, clHide implies match all lines }
      if (Colour <> clHide) and ((AColour = clHide) or (Colour = AColour)) then
        if Abs(Frequency-Centre) <= Range then begin
          if Assigned(OnCheckLine) and not OnCheckLine(FLines[i]) then
            Continue;
          LineForm.AddLine(FLines[i],Units);
        end;
end;
{$ENDIF}

function TLineList.FindPeak(Left, Right: Double; out Position: Double;
  AUnits: TResonanceUnits; AColour: TColor): Boolean;
var
  SumI, SumIX, Intensity: Double;
  i: Integer;
  Source: TMNode;
  Problem: Exception;
begin
  Problem := nil;
  Left := Convert(Left, AUnits, Units);
  Right := Convert(Right, AUnits, Units);
  InOrder(Left, Right);
  SumI := 0;
  SumIX :=0;
  Source := nil;
  for i := 0 to nLines-1 do
    with Flines[i] do
      if ((AColour = clHide) or (Colour = AColour))
         and (Frequency >= Left) and (Frequency <= Right) then begin
        if Assigned(EigenvaluesU) then
          Source := EigenvaluesU.Manifold.CommonAncestor(Source);
        if Assigned(EigenvaluesL) then
          Source := EigenvaluesL.Manifold.CommonAncestor(Source);
        Intensity := GlobalMixture.Simulations[0].LineIntensity(FLines[i],Units,Problem);
        SumI := SumI + Intensity;
        SumIX := SumIX + Intensity*Frequency;
      end;
  Problem.Free; { Ignore any propblems for now }
  Result := SumI > 0;
  if Result then begin
    Position := SumIX/SumI;
    { and consider assigning it to a calibration spectrum }
    if Assigned(Source) then
      PeakMeasured(Convert(Position,Units,AUnits),AUnits,'Calculated',Source);
  end;
end;

procedure TLineList.ShowLine(LineNumber: Integer; var F: TextFile;  PrintLevel: TPrintLevel; Threshold:Double; LinFormat: Boolean; var Problem: Exception);
begin
  ShowLine(FLines[LineNumber], F, PrintLevel, Threshold, LinFormat, Problem);
end;

function TLineList.LargestIntensity: Double;
var
  Mixture: TMixture;
  Intensity: Double;
  i: Integer;
  Problem: Exception;
begin
  Result := 0;
  Problem := nil;
  if nLines = 0 then
    Exit;
  if Assigned(Line[0].EigenvaluesU) then
    Mixture := Line[0].EigenvaluesU.Manifold.Mixture
  else
    Mixture := Line[0].EigenvaluesL.Manifold.Mixture;
  for i := 0 to nLines-1 do begin
    Intensity := Mixture.Simulations[0].LineIntensity(Line[i], Units, Problem);
    if Intensity > Result then
      Result := Intensity;
  end;
  Problem.Free;
end;

procedure TLineList.Delete(n: Integer);
var
  i: Integer;
begin
  if n = nLines-1 then
    SetLength(FLines, n) { Delete last line }
  else
    for i := n to nLines-2 do
      FLines[i] := Line[i+1];
  Dec(FnLines);
end;

procedure TLineList.DeleteSelected(UpStateSel, LoStateSel: Integer);
var
  w, i: Integer;
  Keep: Boolean;
begin
  if (UpStateSel <= 0) and (LoStateSel <= 0) then
    Exit; { Nothing to do }
  w := 0; { Write pointer }
  for i := 0 to nLines-1 do
    with FLines[i] do begin
      Keep := ((UpStateSel <= 0) or (IndexU = UpStateSel))
              and ((LoStateSel <= 0) or (IndexL = LoStateSel));
      if Keep then begin
        FLines[w] := FLines[i];
        Inc(w);
      end;
    end;
  FnLines := w;
  SetLength(FLines,w);
end;

{ TStreamableMNode }

procedure TStreamableMNode.LoadFromStream(Stream: TStream; AFileName: string);
begin
  DoDirSeparators(AFileName);
  MergeFromStream(Stream, AFileName);
  Name := ExtractFileName(AFileName);
  FDirectoryName := ExtractFilePath(ExpandFileName(AFileName));
  Modified := False;
end;

procedure TStreamableMNode.LoadFromFile(AFileName: string);
var
  Stream: TStream;
begin
  DoDirSeparators(AFileName);
  Stream := TFileStream.Create(AFilename, fmOpenRead or fmShareDenyWrite);
  try
    LoadFromStream(Stream, AFileName);
  finally
    Stream.Free;
  end;
end;

procedure TStreamableMNode.MergeFromStream(Stream: TStream; AFileName: string);
var
  XMLDocument: TXMLDocument;
  MNodeClass: TMNodeClass;
  nUnread: Integer;
  E: Exception;
begin
  ReadXMLFile(XMLDocument, Stream, AFileName);
  try
    MNodeClass := ReadMNodeClass(XMLDocument.FirstChild,nil);
    if MNodeClass = nil then
      { Ignore - message already presented }
    else if MNodeClass.InheritsFrom(ClassType) then
      nUnread := ReadProperties(XMLDocument.FirstChild)
    else if ClassType.InheritsFrom(MNodeClass.OwnerType) then
      nUnread := ReadChildProperties(XMLDocument)
    else begin
      nUnread := 1;
      WriteLn(Log, 'Unrecognized item - ', MNodeClass.ClassName);
    end;
    if nUnread > 0 then
      ShowInfo('Warning - '+IntToStr(nUnread)+' unread item(s); see log for details');
    E := FixupProperties;
    if Assigned(E) then
      ShowInfo('Error:'+E.Message);
    E.Free;
  finally
    XMLDocument.Free;
  end;
end;

procedure TStreamableMNode.SaveToFile(AFileName: string);
var
  XMLDocument: TXMLDocument;
  OldName: string;
begin
  FixupProperties; { Discard error }
  XMLDocument := TXMLDocument.Create;
  try
    OldName := FName;
    FName := ExtractFileName(AFileName);
    FDirectoryName := ExtractFilePath(AFileName);
    WriteProperties(XMLDocument);
    WriteXMLFile(XMLDocument, AFilename);
    Modified := False;
    if FName <> OldName then
      PerformMNodeNotification(Self, mnNameChange, -1);
  finally
    XMLDocument.Free;
  end;
end;

{ TUnitsMNode }

function TUnitsMNode.GetIntensityUnits(const Index: Integer): TIntensityUnits;
begin
  Result := TIntensityUnits(GetSetting(Index));
end;

procedure TUnitsMNode.SetIntensityUnits(const Index: Integer;
  const AValue: TIntensityUnits);
begin
  ModifySetting(Index, Ord(AValue), IsPopDist(AValue)<>IsPopDist(GetIntensityUnits(Index)));
end;

function TUnitsMNode.GetUnits(const Index: Integer): TResonanceUnits;
begin
  Result := TResonanceUnits(GetSetting(Index));
end;

procedure TUnitsMNode.SetUnits(const Index: Integer; const AValue: TResonanceUnits);
begin
  SetSetting(Index, Ord(AValue));
end;

function TUnitsMNode.UnitsName: string;
begin
  Result := NameOfUnit(GetUnits(1));
end;

function TUnitsMNode.UnitsTypeName: string;
begin
  Result := TypeOfUnit(GetUnits(1));
end;

{ TPerturbation }

destructor TPerturbation.Destroy;
begin
  DiscardMatrixElements(FMatrixElements);
  ParameterVariable.Free;
  inherited Destroy;
end;

class function TPerturbation.BraKetClass: TMNodeClass;
begin
  Result := TState;
end;

function TPerturbation.Connects(State1, State2: TState): Boolean;
begin
  Result := ((State1 = Bra)and (State2 = Ket))
            or ((State2 = Bra)and (State1 = Ket));
end;

constructor TPerturbation.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',SymSelect=all';
  inherited;
end;

function TPerturbation.FixupProperties: Exception;
begin
  DiscardMatrixElements(FMatrixElements);
  FMatrixElementsEmpty := False;
  if not Assigned(ParameterVariable) then
    FParameterVariable := TSortedVariable.Create(Name,500); { TODO: Name needs work }
  FParameterVariable.Value := Value;
  Result := inherited FixupProperties;
end;

function TPerturbation.CheckMatrixElementsAvailable: Boolean;
var
  i: Integer;
begin
  Result := True; { Default is that they can be implemented; Return false if not implementable }
  if FMatrixElementsEmpty or (Length(FMatrixElements) > 0) then
    Exit;
  FMatrixElements := SymbolicMatrixElements;
  for i := 0 to High(MatrixElements) do begin
    MatrixElements[i].Expr := OperatorTimes.CreateExprNode(ParameterVariable.Template.Duplicate,MatrixElements[i].Expr);
    MatrixElements[i].Evaluator := MatrixElements[i].Expr.Compile;
  end;
  FMatrixElementsEmpty := Length(FMatrixElements) = 0;
end;

function TPerturbation.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  if AIndex = LPI-2 then
    Result := SymSettingsNames(Self, LowestValue)
  else
    Result := inherited GetSettingsNames(AIndex, LowestValue);
end;

function TPerturbation.MatrixElement(State1, State2: TBasisState; MatElemType: TMatElemType): Complex;
begin
  Result := State1.State.PerturbationMatrixElement(Self, State1, State2, MatElemType)
end;

function TPerturbation.PairName: string;
begin
  Result := 'H';
end;

procedure TPerturbation.PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string);
var
  Elements: TMatrixElements;
  j: Integer;
begin
  inherited;
  if (PrintLevel < plDetail) or (Value = 0) then
    Exit;
  try
    Elements := SymbolicMatrixElements;
    for j := 0 to High(Elements) do
      WriteLn(F, Indent, (Bra as TState).MatrixElementType(Elements[j].Signature, PairName), ' = ', Elements[j].Expr.Text(True));
    DiscardMatrixElements(Elements);
  except
    { Not yet implemented }
  end;
end;

procedure TPerturbation.PrettyPrintSettings(var F: Text; PrintLevel: TPrintLevel; Indent: string);
begin
  if Nucleus > 0 then
    WriteLn(F, Indent, 'Nucleus ', Nucleus);
  if ScalePrev <> 0 then
    WriteLn(F, Indent, 'Scale from previous ', FloatToStr(ScalePrev));
end;

function TPerturbation.HelpOnSetting(AIndex: Integer): string;
begin
  case AIndex of
    LPI-3: Result := 'Index, starting from 1, of nuclear spin involved in perturbation; 0 (default) for those not involving a nuclear spin.';
    LPI-2: Result := 'Symmetry select';
    LPI-1: Result := 'Scale factor with respect to preceding perturbation';
  else
    Result := inherited HelpOnSetting(AIndex);
  end;
end;

function TPerturbation.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Size of perturbation.'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

procedure TPerturbation.SetParameter(AIndex: Integer; const Value: Double);
var
  i: Integer;
begin
  { Need to test for Nan/Inf, as this special case forces
    restoration of the default value }
  if not IsNanOrInf(Value) and (Parameters[AIndex] = Value) then
    Exit;
  inherited SetParameter(AIndex, Value);
  for i := ComponentIndex+1 to Owner.ComponentCount - 1 do
    if Owner.Components[i] is TPerturbation then begin
      if TPerturbation(Owner.Components[i]).ScalePrev <> 0 then
        TPerturbation(Owner.Components[i]).Value := Value*TPerturbation(Owner.Components[i]).ScalePrev;
      Break;
    end;
end;

function TPerturbation.SymbolicMatrixElements: TMatrixElements;
begin
  raise NotYet;
end;

{ TMixtureMNode }

function TMixtureMNode.BoltzmannConstant: Double;
var
  Units: TEnergyUnits;
begin
  if RootNode is TMixture then
    Units := TMixture(RootNode).Units
  else
    Units := cm1;
  Result := BoltzmannConstants[Units];
end;

function TMixtureMNode.Precision: Integer;
begin
  if RootNode is TMixture then
    Result := TMixture(RootNode).Precision
  else
    Result := 4; { TODO: Raise }
end;

function TMixtureMNode.QuantumNumberFormat: TQuantumNumberFormat;
begin
  if RootNode is TMixture then
    Result := TMixture(RootNode).QuantumNumberFormat
  else
    Result := GlobalMixture.QuantumNumberFormat;
end;

function TMixtureMNode.QuantumNumberToStr(AQuantumNumber: TQuantumNumber): string;
begin
  if Assigned(Self) then
    Result := QuantumNumber.QuantumNumberToStr(AQuantumNumber,QuantumNumberFormat)
  else { Take default; means that GlobalMixture.QuantumNumberToStr will always work }
    Result := QuantumNumber.QuantumNumberToStr(AQuantumNumber,qftJ);
end;

function TMixtureMNode.StrToQuantumNumber(s: string): TQuantumNumber;
var
  OK: Boolean;
begin
  OK := TryStrToQuantumNumber(s,Result,QuantumNumberFormat);
  if not OK then
    raise EReportException('Bad quantum number format:'+s);
end;

{ TNucleus }

procedure TNucleus.AddNodeParameters(var FittingParameters: TParameters);
begin
  if not AsNext then
    inherited;
  { else Parameters constrained to be identical }
end;

constructor TNucleus.Create(AOwner: TComponent);
begin
  { Default mechanism overwrites name, spin and AsNext so defeat during creation }
  SavedProperties := SavedProperties + ',Spin=0,AsNext=false';
  inherited Create(AOwner);
  if Owner = nil then Exit; { Documentation purposes only }
  FState := AOwner as TState;
  State.AdjustNuclei(False);
  { Spin and AsNext should be per molecule settings, but are actually
    in the per state object. Copy from the per molecule setting on
    initialisation }
  ModifySetting(1, Molecule.Nuclei[NodeIndex].Spin, False);
  ModifySetting(2, Ord(Molecule.Nuclei[NodeIndex].AsNext), False);
end;

function TNucleus.MatchingNucleus: TNucleus;
begin
  Result := Self;
  if not HasIdentical then
    Exit;
  if Result.AsNext then begin
    { Same as next nucleus }
    if High(State.FNuclei) > NodeIndex then
      Result := State.Nucleus(NodeIndex+1);
  end else
    { Check for match to previous; if found return first nucleus in list }
    while (Result.NodeIndex > 0) and State.Nucleus(Result.NodeIndex-1).AsNext do
        Result := State.Nucleus(Result.NodeIndex-1)
end;

function TNucleus.Molecule: TMolecule;
begin
  Result := Owner.Owner.Owner as TMolecule;
end;

procedure TNucleus.SetComponentIndex(const Value: Integer);
begin
  inherited;
  State.AdjustNuclei(False);
end;

procedure TNucleus.SetFloating(AIndex: Integer; const Value: Boolean);
var
  Other: TNucleus;
begin
  if Floating[Aindex] = Value then
    Exit;
  inherited SetFloating(AIndex, Value);
  Other := MatchingNucleus;
  if Other <> Self then
    Other.SetFloating(AIndex, Value)
end;

procedure TNucleus.SetIncrement(AIndex: Integer; const Value: Double);
var
  Other: TNucleus;
begin
  { TODO -cStdDev: Is this logic required for StdDev also? }
  if Increment[Aindex] = Value then
    Exit;
  inherited SetIncrement(AIndex, Value);
  Other := MatchingNucleus;
  if Other <> Self then
    Other.SetIncrement(AIndex, Value)
end;

procedure TNucleus.SetName(const Value: TComponentName);
begin
  inherited;
  if NodeIndex >= 0 then
    Molecule.Nuclei[NodeIndex].Name := Value;
end;

procedure TNucleus.SetParameter(AIndex: Integer; const Value: Double);
var
  Other: TNucleus;
begin
  { Need to test for Nan/Inf, as this special case forces
    restoration of the default value }
  if not IsNanOrInf(Value) and (Parameters[AIndex] = Value) then
    Exit;
  inherited SetParameter(AIndex, Value);
  Other := MatchingNucleus;
  if Other <> Self then
    Other.SetParameter(AIndex, GetParameter(AIndex)) { Use GetParameter to avoid infinte loop on setting default value }
end;

function TNucleus.GetName: TComponentName;
begin
  Result := Molecule.Nuclei[NodeIndex].Name;
  if Result = '' then
    Result := Format('Nucleus%d',[NodeIndex+1]);
end;

procedure TNucleus.AddHandlers;
begin
  inherited;
  AddHandler('Global Shift Up', 'Shift values and names in all states', {$IFDEF FPC}@{$ENDIF}GlobalMoveUp);
  AddHandler('Global Shift Down', 'Shift values and names in all states', {$IFDEF FPC}@{$ENDIF}GlobalMoveDown);
end;

procedure TNucleus.GlobalMoveDown(Sender: TObject);
begin
  GlobalShift(1);
end;

procedure TNucleus.GlobalMoveUp(Sender: TObject);
begin
  GlobalShift(-1);
end;

procedure TNucleus.GlobalShift(Shift: Integer);
var
  s: string;
  Source, Target: TNucleus;
  i, j, MyIndex: Integer;
begin
  if (NodeIndex+Shift < 0) or (NodeIndex+Shift >= Molecule.nNuclei) then
    Exit;
  Target := State.Nucleus(NodeIndex+Shift);
  { Move global values once }
  s := Name;
  Name := Target.Name;
  Target.Name := s;
  AsNext := False;        { Probably not right, but probably safest option }
  Target.AsNext := False;
  MyIndex := NodeIndex;
  { Now move all other settings }
  for i := 0 to Molecule.nManifolds - 1 do
    with Molecule.Manifold[i] do
      for j := 0 to nStates - 1 do
        with State[j] do begin
          Source := State[j].Nucleus(MyIndex);
          Target := State[j].Nucleus(MyIndex+Shift);
          if Shift > 0 then
            Source.ComponentIndex := Target.ComponentIndex
          else
            Target.ComponentIndex := Source.ComponentIndex;
          AdjustNuclei(False);
        end;
  { Reset spin, so copied to global array }
  SetSetting(1,GetSetting(1));
  Target.SetSetting(1,Target.GetSetting(1));
end;

procedure TNucleus.SetSetting(AIndex: Integer; const AValue: Integer);
var
  i, j, MyIndex: Integer;
  Changed: Boolean;
begin
  inherited SetSetting(AIndex, AValue);
  { See notes in Create method; set per molecule settting also }
  if NodeIndex < 0 then
    Exit;
  Changed := False;
  case AIndex of
    1: begin Changed := Molecule.Nuclei[NodeIndex].Spin <> AValue; Molecule.Nuclei[NodeIndex].Spin := AValue; end;
    2: begin Changed := Molecule.Nuclei[NodeIndex].AsNext <> (AValue <> 0); Molecule.Nuclei[NodeIndex].AsNext := AValue <> 0; end;
  end;
  if not Changed then
    Exit;
  MyIndex := NodeIndex;
  { As GetSetting is not virtual, changes must be copied to all other nueclei }
  for i := 0 to Molecule.nManifolds - 1 do
    with Molecule.Manifold[i] do
      for j := 0 to nStates - 1 do
        State[j].Nucleus(MyIndex).ModifySetting(AIndex, AValue, True);
end;

function TNucleus.HelpOnSetting(AIndex: Integer): string;
begin
  case AIndex of
    1: Result := 'Nuclear spin';
    2: Result := 'Nucleus is equivalent to following nucleus';
  else
    Result := inherited HelpOnSetting(AIndex);
  end;
end;

function TNucleus.HasIdentical: Boolean;
begin
  Result := False; { By default, all nuclei must be different from any other }
end;

function TNucleus.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex = 1;
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

{ TMagneticTransitionMoment }

function TMagneticTransitionMoment.PairName: string;
begin
  if Rank = 2 then
    Result := Format('(%s)', [TMixture(RootNode).QuantumNumberToStr(Component)])
  else
    Result := Format('(%s,%s)', [TMixture(RootNode).QuantumNumberToStr(Rank), TMixture(RootNode).QuantumNumberToStr(Component)]);
end;

constructor TMagneticTransitionMoment.Create(AOwner: TComponent);
begin
  FMagnetic := True;
  inherited;
end;

{ TSpinTransitionMoment }

function TSpinTransitionMoment.PairName: string;
begin
  Result := 'gS' + inherited PairName;
end;

constructor TSpinTransitionMoment.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Strength=' + FloatToStr(ge);
  inherited;
end;

{ TOrbitalTransitionMoment }

function TOrbitalTransitionMoment.PairName: string;
begin
  Result := 'gL' + inherited PairName;
end;

{ TRotationalTransitionMoment }

constructor TRotationalTransitionMoment.Create(AOwner: TComponent);
begin
  FNuclearMagneton := True;
  FMagnetic := True;
  inherited;
end;

function TRotationalTransitionMoment.PairName: string;
begin
  Result := 'gr';
end;

function TRotationalTransitionMoment.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
begin
  if (RootNode as TMixture).ReadVersion < 5003335 then
    ShowInfo('Warning - change in definition of Strength of Rotational Transition Moments');
  Result := inherited ReadProperties(Node, CheckUnread);
end;

{ TSpinSpin }

class function TSpinSpin.BraKetClass: TMNodeClass;
begin
  Result := TNucleus;
end;

function TSpinSpin.PairName: string;
begin
  Result := 'II';
end;

{ TEigenValueCache }

destructor TEigenValueCache.Destroy;
begin
  FLock.Free;
  inherited;
end;

{ TSimulation }

procedure TSimulation.ChangePlotUnits(NewPlotUnits: TResonanceUnits);
var
  MidPoint, d: Double; { Centre plot position - needed for scaling widths }
begin
  { Change plot units, scaling affected values also }
  if NewPlotUnits = PlotUnits then
    Exit;
  if IsWavelengthUnit(NewPlotUnits) and (Min(FMin, FMax) <= 0) then begin
    d := Max(Abs(Fmin), Abs(Fmax));
    FMin := d/10;
    FMax := d;
  end;
  MidPoint := (FMin + FMax) / 2;
  FMin := Convert(FMin,PlotUnits,NewPlotUnits);
  FMax := Convert(FMax,PlotUnits,NewPlotUnits);
  Gaussian := ConvertWidth(Gaussian, MidPoint, PlotUnits, NewPlotUnits);
  Lorentzian := ConvertWidth(Lorentzian, MidPoint, PlotUnits, NewPlotUnits);
  { Change sign of displacement if taking reciprocal }
  Foffset := IfThen(IsWavelengthUnit(PlotUnits) xor IsWavelengthUnit(NewPlotUnits), -1, 1)
             * ConvertWidth(Foffset, MidPoint, PlotUnits, NewPlotUnits);
  PlotUnits := NewPlotUnits;
end;

constructor TSimulation.Create(AOwner: TComponent);
var
  i: Integer;
begin
  SavedProperties := SavedProperties + ',OThreshold=1e-6,WidthMult=6,SMargin=-1,Temperature=300,FMin=0,FMax=50000,nDF=2000,ShowSum=True,Tvib=-1,Tspin=-1';
  SavedProperties := SavedProperties + ',ScaleMarkSize=True,ShowSymmetry=True,Name=Simulation';
  SavedProperties := SavedProperties + ',IntensityUnits=Normalized,FortratQno=J';

  DimensionlessParameters := DimensionlessParameters + ',OThreshold,SMargin,Temperature,MinI,Tvib,Saturation,EField,BField,Tspin,,Doppler';
  { The next five are not dimensionless, but are only affected by PlotUnits }
  DimensionlessParameters := DimensionlessParameters + ',Foffset,Fmin,Fmax,Gaussian,Lorentzian';
  { ... in fact the only parameter with units is RefWidth! }

  inherited;
  HasPlotted := False;
  SetLength(FSettingOverridden, nSettings+1);
  SetLength(FParameterOverridden, nParameters+1);
  FillChar(FSettingOverridden[0], Length(FSettingOverridden)*SizeOf(Boolean),0);
  FillChar(FParameterOverridden[0], Length(FParameterOverridden)*SizeOf(Boolean),0);
  if Assigned(AOwner) then
    for i := 0 to AOwner.ComponentCount - 1 do
      if AOwner.Components[i] is TSimulation then begin
        if AOwner.Components[i] <> Self then
          FBaseNode := AOwner.Components[i] as TSimulation;
        Break;
      end;
end;

function TSimulation.GetDefaultSettings(AIndex: Integer): string;
begin
  if Assigned(FBaseNode) then
    Result := 'Track'
  else
    Result := inherited GetDefaultSettings(AIndex);
end;

function TSimulation.GetExField: TStaticFields;
begin
  Result.E := EField;
  Result.B := BField;
  Result.Angle := 0;
end;

function TSimulation.GetFortratQno(const Index: Integer): TQnoType;
begin
  Result := TQnoType(GetSetting(Index));
end;

function TSimulation.GetLifeModel(const Index: Integer): TLifeModel;
begin
  Result := TLifeModel(GetSetting(Index));
end;

function TSimulation.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  Result := inherited GetSettingsNames(AIndex,LowestValue);
  if SettingNames[AIndex] = 'FortratQno' then
    Result := @QnoTypeNames[0];
end;

const
 SimulationParameterNames: array [1..16] of string = (
'Left edge of plot range in main window.',
'Right edge of plot range in main window.',
'Rotational temeperature (Kelvin).',
'Gaussian contribution to linewidth (full width half maximum). If both this and a Lorentzian width (below) is set the result is a convolution of the two, a Voigt profile.',
'Lorentzian contribution to linewidth (full width half maximum). This and the Gaussian width can also be set from the main window.',
'Frequency offset to simulation.',
'Extra points to calculate off each end of required spectrum.',
'Ignore peaks smaller than this fraction of the maximum peak intensity.',
'Reference width in linewidth (predissociation) calculations.',
'Vibrational Temperature (Kelvin); set to -1 (default) to use rotational temperature for all Boltzmann factors.',
'Set the scaling of mark sizes in Fortrat plots.',
'Zero for normal calculation; positive values progesively switch strength to population only',
'Spin Temperature; set to -1 (default) to equilibrate different spinstates',
'Static Electric Field (V/m).',
'Static Magnetic Field (Tesla).',
'Doppler Splitting'
//'Angle between fields (degrees).'
);

function TSimulation.HasFields: Boolean;
begin
  Result := (EField <> 0) or (BField <> 0);
end;

function TSimulation.HasWidths: Boolean;
begin
  Result := (Gaussian <> 0) or (Lorentzian <> 0) or (Doppler <> 0);
end;

function TSimulation.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(SimulationParameterNames))
     and (AIndex <= High(SimulationParameterNames)) then
    Result := SimulationParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

const
  SimulationSettingNames: array[1..15] of string = (
'Units for intensity. For those with units, transition moments should be in Debye.',
'Model to use for state dependent lifetimes and linewidths. Valid values are lmNone, lmWidth, lmProduct, lmProductWidth, lmParent, lmParentWidth, lmGate, lmGateWidth.',
'Units for the horizontal scale for spectrum plots',
'Number of points (between Fmin and Fmax) to calculate the spectrum at. Note that if a peak width is less than 3*(Fmax-Fmin)./nDF, i.e. only a few points wide, then it is shown as a stick, rather than a peak.',
'Multiple of linewidth to extend convolution over.',
'Plot overall sum of individual spectra.',
'Plot individual spectra making up overall spectrum. The individual spectra are grouped by colour, so you will need to set colours to see something different to the sum. Colours can be set at the transition moment, state, molecule or species level.',
'Plot a Fortrat diagram, i.e. J against frequency, in the main window.',
'Set to use upper rather than lower state J, symmetry and state number in the Fortrat diagram.',
'Show symmetry in Fortrat plots.',
'Show change in J in Fortrat plots.',
'Scale Mark Size with intensity in Fortrat plots.',
'Set to show different symmetries in separate Fortrat plots.',
'Set to show different state numbers in separate Fortrat plots.',
'Select quantum number to use in Fortrat plots.'
);

function TSimulation.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(SimulationSettingNames))
     and (AIndex <= High(SimulationSettingNames)) then
    Result := SimulationSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TSimulation.LineIntensity(const Line: TLine; AUnits: TResonanceUnits; var Problem: Exception): Double;
var
  g: Integer;
  Str: Double;
begin
  if not Assigned(Line.EigenvaluesL) or not IncludesPopulation(IntensityUnits) then begin
    Result := Line.Strength;
    if IntensityUnits = Squared then
      Result := Sqr(Result);
    if (IntensityUnits = EinsteinA) and Assigned(Line.EigenvaluesU)  then
      Result := Result / Line.EigenvaluesU.Degeneracy(Line.IndexU);
  end else begin
    if not IsPopDist(IntensityUnits) and Assigned(Line.EigenvaluesU) and Line.EigenvaluesU.Manifold.Initial then
      Result := Line.EigenvaluesU.Manifold.Population(Line.EigenvaluesU, Line.IndexU, Self, Problem, 0)
    else
      Result := 0;
    if Line.EigenvaluesL.Manifold.Initial then
      { Use population difference - essential for microwave transitions }
      Result := Abs(Line.EigenvaluesL.Manifold.Population(Line.EigenvaluesL,Line.IndexL,Self,Problem,0)-Result);
    if IsPopDist(IntensityUnits) then
      Str := Line.Strength * (Line.EigenvaluesL.J+1)
    else if Saturation > 0 then begin
      { Basic allowance for saturation (assuming z polarized light) }
      g := Line.EigenvaluesL.J;
      if Assigned(Line.EigenvaluesU) then
        g := Min(Line.EigenvaluesU.J, g);
      g := (g+1)*Line.EigenvaluesL.StatisticalWeight(Line.IndexL);
      Str := (1-Exp(-Line.Strength*Saturation/g)) * g;
    end else
      Str := Line.Strength;
    Result := Result * Str;
    if IntensityUnits = Squared then
      Result := Sqr(Result);
    if IsNormalized(IntensityUnits) then
      Result := Result / Line.EigenvaluesL.Manifold.Molecule.PartitionFunction(Self, Problem);
  end;
  { TODO: different scaling factor in IntensityFactor for Quadrupole transitions }
  Result := Result * IntensityFactor(Line.Frequency,AUnits,IntensityUnits);
end;

function TSimulation.MakeLineList(var Problem: Exception): TLineList;
begin
  TStart(MakeLineListTime);
  Result := TLineList.Create(Mixture.Units,IsPopDist(IntensityUnits));
  Result.Temperature := Temperature;
  try
    if not Mixture.AddToSpectrum(Result, Problem, Self) then
      FreeAndNil(Result);
  except
    Result.Free;
    raise;
  end;
  Mixture.LineList := Result;
  TEnd(MakeLineListTime);
end;

function TSimulation.Mixture: TMixture;
begin
  Result := Owner as TMixture;
end;

function TSimulation.ParameterOverridden(AIndex: Integer): Boolean;
begin
  Result := FParameterOverridden[AIndex];
end;

function TSimulation.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
begin
  Result := inherited ReadProperties(Node, CheckUnread);
  { EinsteinASum was introduced after 6.0.228; before that IntensityUnits = EinsteinA gave current EinsteinASum behaviour. }
  if ((RootNode as TMixture).ReadVersion <= 6000228) and (IntensityUnits = EinsteinA) then
    IntensityUnits := EinsteinASum;
end;

procedure TSimulation.SetFortratQno(const Index: Integer; const Value: TQnoType);
begin
  SetSetting(Index, Ord(Value));
end;

procedure TSimulation.SetLifeModel(const Index: Integer; const AValue: TLifeModel);
begin
  SetSetting(Index, Ord(AValue));
end;

{ Changing frequency limits and so on does not count as a proper change
  Add mechanism to track individual changes }
procedure TSimulation.SetParameter(AIndex: Integer; const Value: Double);
begin
  if ((AIndex in [14..16] {ExField}) and ((Parameters[AIndex] <> Value)) or IsNanOrInf(Value)) then
    Mixture.ClearCache;
  if Assigned(FBaseNode) then
    if IsNanOrInf(Value) then begin
      FParameterOverridden[AIndex] := False;
      inherited ModifyParameter(AIndex, FBaseNode.Parameters[AIndex], False);
      Exit;
    end else if GetParameter(AIndex) <> Value then
      FParameterOverridden[AIndex] := True;
  inherited ModifyParameter(AIndex, Value, False);
end;

procedure TSimulation.SetSetting(AIndex: Integer; const AValue: Integer);
begin
  if Assigned(FBaseNode) and (GetSetting(AIndex) <> AValue) then
    FSettingOverridden[AIndex] := True;
  inherited ModifySetting(AIndex, AValue, False);
end;

procedure TSimulation.SetSettingFromStr(AIndex: Integer; const s: string);
begin
  if Assigned(FBaseNode) and (s = 'Track') then begin
    FSettingOverridden[AIndex] := False;
    PerformMNodeNotification(Self, mnSettingChange, AIndex);
  end else
    inherited;
end;

function TSimulation.SettingToStr(AIndex: Integer): string;
begin
  if Assigned(FBaseNode) and not FSettingOverridden[AIndex] then
    Result := 'Track'
  else
    Result := inherited SettingToStr(AIndex);
end;

procedure TSimulation.UpdateBaseNode(ABaseNode: TSimulation);
var
  i: Integer;
begin
  FBaseNode := ABaseNode;
  if not Assigned(FBaseNode) then
    Exit;
  for i := 1 to nParameters do
    if not FParameterOverridden[i] then
      ModifyParameter(i, FBaseNode.GetParameter(i), False);
  for i := 0 to nSettings do
    if not FSettingOverridden[i] then
      ModifySetting(i, FBaseNode.GetSetting(i),False);
end;

procedure TSimulation.WriteFloatableValue(AIndex: Integer; Deflt: string; DOMNode: TDOMNode; DefaultCheck: Boolean);
begin
  if Assigned(FBaseNode) and not FParameterOverridden[AIndex] then
    Exit; { Skip - not overriden }
  inherited;
end;

function TSimulation.WriteProperties(Parent: TDOMNode; AllParameters: Boolean; WithChildren: Boolean): TDOMElement;
begin
  if not FWritten then
    Result := inherited WriteProperties(Parent, AllParameters, WithChildren)
  else
    Result := nil;
end;

{ TCachedTransitionMatrix }

destructor TCachedTransitionMatrix.Destroy;
begin
  FreeAndNil(TransitionMatrix);
  inherited;
end;

procedure TCachedTransitionMatrix.DoMake(RequiredAttributes: Pointer);
begin
  EigenValues.Make(True,[meHrot]);
  OtherEigenValues.Make(True,[meHrot]);
  TransitionMatrix := TransitionMoments.TransitionMatrix(Eigenvalues,OtherEigenValues,True,False,nil,Simulation,PrintLevel);
end;

function TCachedTransitionMatrix.HasCorrectAttributes(Attributes: Pointer): Boolean;
begin
  Result := Assigned(TransitionMatrix);
end;

{ TBaseRotationalTransitionMoment }

constructor TBaseRotationalTransitionMoment.Create(AOwner: TComponent);
begin
  DimensionlessParameters := DimensionlessParameters + ',C,D';
  inherited;
end;

{ TSingleTransitionMoment }

constructor TSingleTransitionMoment.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Strength=1';
  DimensionlessParameters := DimensionlessParameters + ',Strength';
  inherited;
end;

function TSingleTransitionMoment.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Relative transition (dipole) moment - note that the relative intensity is proportional to the square of this value.'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

{ TNuclearMagneticMoment }

constructor TNuclearMagneticMoment.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',gN=1,Nucleus=1';
  DimensionlessParameters := DimensionlessParameters + ',gN';
  FNuclearMagneton := True;
  FMagnetic := True;
  inherited;
end;

function TNuclearMagneticMoment.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Nuclear g factor in terms of Nuclear Magnetons.'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

function TNuclearMagneticMoment.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1  then
    Result := 'Index of nucleus'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TNuclearMagneticMoment.PairName: string;
begin
  Result := 'gN(';
  with Owner.Owner as TMolecule do
    if (nNuclei >= Nucleus) and (Nucleus > 0) then
      Result := Result + Nuclei[Nucleus-1].Name
    else
      Result := Result + '?';
  Result := Result + ')';
end;

{ TMNodeWithParameters }

procedure TMNodeWithParameters.AddNodeParameters(MNode: TMNode; var FittingParameters: TParameters);
var
  i, n: Integer;
  s: string;
begin
  for i := 1 to MNode.nParameters do
    if MNode.Floating[i] then begin
      n := Length(FittingParameters);
      SetLength(FittingParameters, n+1);
      FittingParameters[n].MNode := MNode;
      FittingParameters[n].Index := i;
      FittingParameters[n].Value := MNode.GetParameter(i);
      FittingParameters[n].Increment := MNode.Increment[i];
      FittingParameters[n].StdDev := MNode.StdDev[i];
      FittingParameters[n].Inverse := False;
      s := MNode.UniqueName;
      if s <> '' then
        s := s + ' ';
      FittingParameters[n].Name := s + MNode.ParameterNames[i];
    end;
  for i := 0 to MNode.Count-1 do
    if MNode.Items[i] is TMNodeWithParameters then
      TMNodeWithParameters(MNode.Items[i]).AddNodeParameters(FittingParameters)
    else
      AddNodeParameters(MNode.Items[i], FittingParameters);
end;

procedure TMNodeWithParameters.AddNodeParameters(var FittingParameters: TParameters);
begin
  AddNodeParameters(Self, FittingParameters);
end;

{ TRootMNode }

procedure TRootMNode.BeginPaste;
begin
  FReadVersion := DecodePgopherVersion(VersionInfo);
end;

{ TCorrelations }

procedure TCorrelations.AddHandlers;
begin
  inherited;
  AddHandler('Resync', 'Adjust matrix so only floated parameters are included', {$IFDEF FPC}@{$ENDIF}Resync);
end;

procedure TCorrelations.AdjustSize;
var
  i, j, n: Integer;
begin
  { Construct empty correlation array based on size of Parameters }
  n := Length(FitDetails.Parameters);
  SetLength(FitDetails.Correlation, n);
  for i := 0 to n - 1 do begin
    SetLength(FitDetails.Correlation[i],i);
    for j := 0 to i - 1 do
      FitDetails.Correlation[i,j] := 0;
  end;
end;

procedure TCorrelations.Clear;
begin
  FitDetails.Parameters := nil;
end;

procedure TCorrelations.UnpackCorrelations(var UnpackedCorrelations: TUnpackedCorrelations);
var
  i, j, n: Integer;
begin
  SetLength(UnpackedCorrelations, Sqr(nEntries) div 2 + 1); { Guess capacity }
  n := 0;
  for i := 0 to nEntries-1 do
    for j := 0 to i - 1 do
      if FitDetails.Correlation[i][j] <> 0 then
        with UnpackedCorrelations[n] do begin
          Row := EntryName(i,True);
          Col := EntryName(j,True);
          Value := FitDetails.Correlation[i][j];
          Inc(n);
        end;
  SetLength(UnpackedCorrelations, n);
end;

constructor TCorrelations.Create(AOwner: TComponent);
var
  Mixture: TMixture;
begin
  inherited;
  Mixture := RootNode as TMixture;
  if Assigned(Mixture.FCorrelations) then
    raise Exception.Create('Duplicate correlations object')
  else
    Mixture.FCorrelations := Self;
end;

destructor TCorrelations.Destroy;
begin
  if RootNode <> Self then
    (RootNode as TMixture).FCorrelations := nil;
  { otherwise being destroyed }
  inherited;
end;

function TCorrelations.EntryName(i: Integer; Quote: Boolean): string;
begin
  with FitDetails.Parameters[i] do begin
    Result := MNode.FullParameterName(Index,Quote);
    if not (MNode.RootNode is TMixture) then
      Result := Copy(MNode.RootNode.ClassName,2,255)+'.'+Result;
  end;
end;

function TCorrelations.nEntries: Integer;
begin
  { Allow for inconsistent value of result during update }
  with FitDetails do
    Result := Min(Length(Parameters),Length(Correlation));
end;

procedure TCorrelations.PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string);
var
  i, j, l: Integer;
begin
  if nEntries <= 1 then
    Exit;
  inherited;
  l := 9;
  for i := 0 to nEntries-1 do
    l := Max(l,Length(EntryName(i,False)));
  Write(F,' ':l);
  for i := 0 to nEntries-1 do
    Write(F, ' ', EntryName(i,False));
  WriteLn(F);
  for i := 0 to nEntries-1 do begin
    Write(F, EntryName(i,False):l);
    for j := 0 to i-1 do
      Write(F, ' ', FitDetails.Correlation[i,j]:Length(EntryName(j,False)):6);
    WriteLn(F, ' ', 1.0:Length(EntryName(i,False)):6);
  end;
end;

function TCorrelations.ReadData(UnpackedCorrelations: TUnpackedCorrelations; FirstPass: Integer; CheckUnread: Boolean; OtherNode:TRootMNode): Integer;
var
  UnRead: TStringList; { Record unrecognised parameters, only one error message per parameter }
  OtherStart: string;

  function FindParameter(Name: string):Integer;
  var
    PropInfo: PPropInfo;
    MNode: TMNode;
    Index, I: Integer;
    ErrMsg: string;
  begin
   PropInfo := RootNode.SubPropInfo(Name, MNode, ErrMsg);
   if not Assigned(PropInfo) and Assigned(OtherNode) and (Pos(OtherStart,Name)=1) then
     PropInfo := OtherNode.SubPropInfo(Copy(Name,Length(OtherStart)+1,Length(Name)), MNode, ErrMsg);
   if Assigned(PropInfo) then
     if Cardinal(PropInfo) <= MNode.nParameters then
       Index := Cardinal(PropInfo)
     else
       Index := PropInfo^.Index
   else
     Index := 0;
   Result := -1;
   if Index = 0 then begin
     if (Pos('Overlays.',Name) = 1) and not Assigned(UnRead) and not Assigned(OtherNode) then
       Exit; { Sliently discard entries from Overlay parameters, if no overlay }
     if not Assigned(UnRead) then begin
       UnRead := TStringList.Create;
       if CheckUnread then
         WriteLn(Log, 'Unrecognized parameter(s) in correlation matrix:');
     end;
     if UnRead.IndexOf(Name) < 0 then begin
       UnRead.Add(Name);
       if CheckUnread then
         WriteLn(Log, '  ', Name, ' ', ErrMsg);
     end;
   end else begin
     for I := High(FitDetails.Parameters) downto 0 do
       if (FitDetails.Parameters[i].Index = Index) and (FitDetails.Parameters[i].MNode = MNode) then begin
         Result := I;
         Break;
       end;
     if (Result < 0) and (FirstPass = 0) then begin
       Result := Length(FitDetails.Parameters);
       SetLength(FitDetails.Parameters, Result+1);
       FitDetails.Parameters[Result].MNode := MNode;
       FitDetails.Parameters[Result].Index := Index;
       FitDetails.Parameters[Result].StdDev := MNode.StdDev[Index];
     end;
   end;
  end;
var
  Pass, i, j, e:Integer;
begin
  if FirstPass = 0 then
    Clear;
  UnRead := nil;
  Result := 0;
  if Assigned(OtherNode) then
    OtherStart := Copy(OtherNode.ClassName,2,255)+'.'
  else
    OtherStart := '';
  for Pass := FirstPass to 1 do begin
    for e := 0 to High(UnpackedCorrelations) do
      with UnpackedCorrelations[e] do begin
        i := FindParameter(Row);
        if i >= 0 then begin
          j := FindParameter(Col);
          if (j >= 0) and (Pass = 1) then
            if i > j then
              FitDetails.Correlation[i,j] := Value
            else
              FitDetails.Correlation[j,i] := Value
        end;
      end;
    if Pass = 0 then
      AdjustSize;
  end;
  if Assigned(UnRead) then begin
    Result := UnRead.Count;
    Unread.Free;
  end;
end;

function TCorrelations.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  Child, NextChild: TDOMNode;
  C: TUnpackedCorrelations;
  n: Integer;
begin
  n := 0;
  Child := Node.FirstChild;
  while Assigned(Child) do
    if CompareText(Child.NodeName, 'Correlation') = 0 then begin
      Inc(n);
      Child := Child.NextSibling;
    end;
  SetLength(C, n);
  n := 0;
  Child := Node.FirstChild;
  while Assigned(Child) do
    if CompareText(Child.NodeName, 'Correlation') = 0 then
      with C[n] do begin
        Row := ReadValue(Child, 'Row', '');
        Col := ReadValue(Child, 'Col', '');
        Value := StrToFloatEx(ReadValue(Child, 'Value', '0'));
        Inc(n);
        NextChild := Child.NextSibling;
        if CheckUnread then
          Result := Result + nUnreadProperties(Child);
        Node.RemoveChild(Child);
        Child := NextChild;
      end;
  Result := inherited ReadProperties(Node, False);
  Result := Result + ReadData(C, 0, CheckUnread, GlobalOverlays);
  if CheckUnread then
    Result := Result + nUnreadProperties(Node);
end;

procedure TCorrelations.Resync(Sender: TObject);
var
  Mixture: TMixture;
  Original: TUnpackedCorrelations;
begin
  UnpackCorrelations(Original);
  Mixture := RootNode as TMixture;
  Mixture.FillParameters;
  if GlobalOverlays <> nil then
    GlobalOverlays.AddNodeParameters(Mixture.FittingParameters);
  FitDetails.Parameters := Copy(Mixture.FittingParameters);
  AdjustSize;
  ReadData(Original,1,False,GlobalOverlays);
  PerformMNodeNotification(RootNode, mnParameterChange, 1); { Spurious change notification - triggers refresh of Correlation amtrix display }
end;

function TCorrelations.WriteProperties(Parent: TDOMNode; AllParameters: Boolean; WithChildren: Boolean): TDOMElement;
var
  Node: TDOMElement;
  i, j: Integer;
begin
  Result := inherited WriteProperties(Parent, AllParameters, WithChildren);
  for i := 0 to nEntries-1 do
    for j := 0 to i - 1 do
      if FitDetails.Correlation[i][j] <> 0 then begin
        Node := Parent.OwnerDocument.CreateElement('Correlation');
        WriteValue(Node, 'Row', EntryName(i,True), '');
        WriteValue(Node, 'Col', EntryName(j,True), '');
        WriteValue(Node, 'Value', FloatToStr(FitDetails.Correlation[i][j]), '');
        Result.AppendChild(Node);
      end;
end;

initialization
  TSpecies.RegisterClass(TMixture);
  TTransitionMoments.RegisterClass(TMolecule);
  TSphericalTransitionMoment.RegisterClass(TTransitionMoments);
  TCartesianTransitionMoment.RegisterClass(TTransitionMoments);
  TSpinTransitionMoment.RegisterClass(TTransitionMoments);
  TOrbitalTransitionMoment.RegisterClass(TTransitionMoments);
  TRotationalTransitionMoment.RegisterClass(TTransitionMoments);
  TNuclearMagneticMoment.RegisterClass(TTransitionMoments);
  TSimulation.RegisterClass(TMixture);
  TMixture.RegisterClass(nil);
  TCorrelations.RegisterClass(TMixture);
  BoltzMax := Ln(MaxDouble)/2;
  PopulationConnection := TDistribution.Create(nil);
  InitCriticalSection(EvaluatorLock);
finalization
  DoneCriticalsection(EvaluatorLock);
  FreeAndNil(PopulationConnection);
  SetLength(SymSelNames,0);
end.

